module cm3_dap_ahb_ap
  (// Inputs
   hclk, hreset_n, dap_en_i, master_enable_i, dap_clken_i, dap_clk, dap_reset_n,
   dap_sel_i, dap_enable_i, dap_write_i, dap_abort_i, dap_addr_i, dap_wdata_i,
   dap_ahb_hready_i, dap_ahb_haddracc_i, dap_ahb_hresp_i, dap_ahb_hrdata_i,
   fix_master_type_i,
   dap_ready_o, dap_slverr_o, dap_rdata_o, dap_ahb_haddr_o, dap_ahb_hmaster_o,
   dap_ahb_htrans_o, dap_ahb_hwrite_o, dap_ahb_hsize_o, dap_ahb_hburst_o,
   dap_ahb_hprot_o, dap_ahb_hwdata_o, dap_ahb_habort_o
  );


  parameter DEBUG_LVL      = 3;
  parameter RESET_ALL_REGS = 0;



  input         hclk;                // AHB domain clock
  input         hreset_n;            // AHB domain reset
  input         dap_clk;             // DAP domain clock
  input         dap_reset_n;         // DAP domain reset
  input         dap_clken_i;         // DAP domain clock enable

  input         dap_en_i;            // DAP enable - enables AHB transfers
  input         master_enable_i;     // Master enable - enabled core HMASTER

  input         dap_sel_i;           // DAP select (address decode)
  input   [7:0] dap_addr_i;          // DAP address
  input         dap_enable_i;        // DAP transfer enable
  input         dap_write_i;         // DAP transfer direction
  input         dap_abort_i;         // DAP transfer abort
  input  [31:0] dap_wdata_i;         // DAP write data
  output        dap_ready_o;         // DAP slave ready
  output        dap_slverr_o;        // DAP slave error response
  output [31:0] dap_rdata_o;         // DAP slave read data

  output [31:0] dap_ahb_haddr_o;     // AHB address
  output        dap_ahb_hmaster_o;   // AHB master [0]core, [1]debug
  output  [1:0] dap_ahb_htrans_o;    // AHB transfer type (IDLE or NONSEQ)
  output        dap_ahb_hwrite_o;    // AHB transfer direction
  output  [1:0] dap_ahb_hsize_o;     // AHB transfer size
  output  [2:0] dap_ahb_hburst_o;    // AHB burst type (always SINGLE)
  output  [3:0] dap_ahb_hprot_o;     // AHB privilege information
  output [31:0] dap_ahb_hwdata_o;    // AHB write data
  output        dap_ahb_habort_o;    // AHB-AP bus matrix HREADY abort
  input         dap_ahb_hready_i;    // AHB slave ready
  input         dap_ahb_haddracc_i;  // AHB address ready
  input         dap_ahb_hresp_i;     // AHB slave response
  input  [31:0] dap_ahb_hrdata_i;    // Read data from AHB slave
  input         fix_master_type_i;   // Fix HMASTER output to 1



  wire          dap_ready;
  wire          dap_slverr;
  wire   [31:0] dap_rdata;

  wire   [31:0] dap_ahb_haddr;
  wire          dap_ahb_hmaster;
  wire    [1:0] dap_ahb_htrans;
  wire          dap_ahb_hwrite;
  wire    [1:0] dap_ahb_hsize;
  wire    [2:0] dap_ahb_hburst;
  wire    [3:0] dap_ahb_hprot;
  wire   [31:0] dap_ahb_hwdata;
  wire          dap_ahb_habort;
  wire          dap_ahb_habort_sync;  // Synchronized HABORT

  wire   [31:0] slv_haddr;
  wire          slv_hmaster;
  wire          slv_hwrite;
  wire    [1:0] slv_hsize;
  wire          slv_hprot1;
  wire          mst_resp_err;
  wire   [31:0] mst_hrdata;

  wire          slv_req_trans;        // Transfer request from debug access port
  wire          slv_req_trans_rtn;    // slv_req_trans returned back to dap_clk
  wire          mst_end_trans;        // Transfer completed on master interface

  wire          slv_req_trans_sync;   // Synchronized version of slv_req_trans
  wire          mst_end_trans_sync;   // Synchronized version of mst_end_trans

  wire          slv_tra_update;
  wire          slv_tra_update_sync;

  wire          slv_idle;
  wire          slv_idle_sync;        // sync to hclk domain to clear error resp

  wire          opt_dap_ready;
  wire          opt_dap_slverr;
  wire   [31:0] opt_dap_rdata;
  wire   [31:0] opt_dap_ahb_haddr;
  wire          opt_dap_ahb_hmaster;
  wire    [1:0] opt_dap_ahb_htrans;
  wire          opt_dap_ahb_hwrite;
  wire    [1:0] opt_dap_ahb_hsize;
  wire    [2:0] opt_dap_ahb_hburst;
  wire    [3:0] opt_dap_ahb_hprot;
  wire   [31:0] opt_dap_ahb_hwdata;
  wire          opt_dap_ahb_habort;

`ifdef ARM_DSM



  wire [1:0]    dsm_debug_lvl;

  assign opt_dap_ready       = (dsm_debug_lvl > 0) ? dap_ready           : 1'b1;
  assign opt_dap_slverr      = (dsm_debug_lvl > 0) ? dap_slverr          : 1'b0;
  assign opt_dap_rdata       = (dsm_debug_lvl > 0) ? dap_rdata           : {32{1'b0}};
  assign opt_dap_ahb_haddr   = (dsm_debug_lvl > 0) ? dap_ahb_haddr       : {32{1'b0}};
  assign opt_dap_ahb_hmaster = (dsm_debug_lvl > 0) ? dap_ahb_hmaster     : 1'b0;
  assign opt_dap_ahb_htrans  = (dsm_debug_lvl > 0) ? dap_ahb_htrans      : 2'b00;
  assign opt_dap_ahb_hwrite  = (dsm_debug_lvl > 0) ? dap_ahb_hwrite      : 1'b0;
  assign opt_dap_ahb_hsize   = (dsm_debug_lvl > 0) ? dap_ahb_hsize       : 2'b00;
  assign opt_dap_ahb_hburst  = (dsm_debug_lvl > 0) ? dap_ahb_hburst      : 3'b000;
  assign opt_dap_ahb_hprot   = (dsm_debug_lvl > 0) ? dap_ahb_hprot       : 4'b0000;
  assign opt_dap_ahb_hwdata  = (dsm_debug_lvl > 0) ? dap_ahb_hwdata      : {32{1'b0}};
  assign opt_dap_ahb_habort  = (dsm_debug_lvl > 0) ? dap_ahb_habort_sync : 1'b0;

`else


  assign opt_dap_ready       = (DEBUG_LVL > 0) ? dap_ready           : 1'b1;
  assign opt_dap_slverr      = (DEBUG_LVL > 0) ? dap_slverr          : 1'b0;
  assign opt_dap_rdata       = (DEBUG_LVL > 0) ? dap_rdata           : {32{1'b0}};
  assign opt_dap_ahb_haddr   = (DEBUG_LVL > 0) ? dap_ahb_haddr       : {32{1'b0}};
  assign opt_dap_ahb_hmaster = (DEBUG_LVL > 0) ? dap_ahb_hmaster     : 1'b0;
  assign opt_dap_ahb_htrans  = (DEBUG_LVL > 0) ? dap_ahb_htrans      : 2'b00;
  assign opt_dap_ahb_hwrite  = (DEBUG_LVL > 0) ? dap_ahb_hwrite      : 1'b0;
  assign opt_dap_ahb_hsize   = (DEBUG_LVL > 0) ? dap_ahb_hsize       : 2'b00;
  assign opt_dap_ahb_hburst  = (DEBUG_LVL > 0) ? dap_ahb_hburst      : 3'b000;
  assign opt_dap_ahb_hprot   = (DEBUG_LVL > 0) ? dap_ahb_hprot       : 4'b0000;
  assign opt_dap_ahb_hwdata  = (DEBUG_LVL > 0) ? dap_ahb_hwdata      : {32{1'b0}};
  assign opt_dap_ahb_habort  = (DEBUG_LVL > 0) ? dap_ahb_habort_sync : 1'b0;

`endif



  cm3_dap_ahb_ap_slv #(DEBUG_LVL, RESET_ALL_REGS)
    u_cm3_dap_ahb_ap_slv
      (// Inputs
       .dap_clk                   (dap_clk),
       .dap_reset_n               (dap_reset_n),
       .dap_clken_i               (dap_clken_i),
       .master_enable_i           (master_enable_i),
       .dap_en_i                  (dap_en_i),
       .dap_sel_i                 (dap_sel_i),
       .dap_enable_i              (dap_enable_i),
       .dap_write_i               (dap_write_i),
       .dap_addr_i                (dap_addr_i),
       .dap_wdata_i               (dap_wdata_i),
       .dap_abort_i               (dap_abort_i),
       .mst_resp_err_async_i      (mst_resp_err),
       .mst_hrdata_async_i        (mst_hrdata),
       .mst_end_trans_i           (mst_end_trans_sync),
       .slv_req_trans_rtn_i       (slv_req_trans_rtn),
       .dap_ready_o               (dap_ready),
       .dap_slverr_o              (dap_slverr),
       .dap_rdata_o               (dap_rdata),
       .dap_ahb_habort_o          (dap_ahb_habort),
       .slv_haddr_o               (slv_haddr),
       .slv_hwrite_o              (slv_hwrite),
       .slv_hsize_o               (slv_hsize),
       .slv_hmaster_o             (slv_hmaster),
       .slv_hprot1_o              (slv_hprot1),
       .slv_req_trans_o           (slv_req_trans),
       .slv_idle_o                (slv_idle),
       .slv_tra_update_o          (slv_tra_update)
      );



  cm3_dap_ahb_ap_mst #(DEBUG_LVL, RESET_ALL_REGS)
    u_cm3_dap_ahb_ap_mst
      (// Inputs
       .hclk                      (hclk),
       .hreset_n                  (hreset_n),
       .dap_ahb_hready_i          (dap_ahb_hready_i),
       .dap_ahb_haddracc_i        (dap_ahb_haddracc_i),
       .dap_ahb_hresp_i           (dap_ahb_hresp_i),
       .dap_ahb_hrdata_i          (dap_ahb_hrdata_i),
       .slv_haddr_async_i         (slv_haddr),
       .slv_hsize_async_i         (slv_hsize),
       .slv_hwrite_async_i        (slv_hwrite),
       .slv_hprot1_async_i        (slv_hprot1),
       .slv_hmaster_async_i       (slv_hmaster),
       .fix_master_type_i         (fix_master_type_i),
       .slv_req_trans_i           (slv_req_trans_sync),
       .slv_idle_i                (slv_idle_sync),
       .slv_tra_update_i          (slv_tra_update_sync),
       .dap_wdata_async_i         (dap_wdata_i),
       .dap_ahb_htrans_o          (dap_ahb_htrans),
       .dap_ahb_haddr_o           (dap_ahb_haddr),
       .dap_ahb_hsize_o           (dap_ahb_hsize),
       .dap_ahb_hwrite_o          (dap_ahb_hwrite),
       .dap_ahb_hprot_o           (dap_ahb_hprot),
       .dap_ahb_hburst_o          (dap_ahb_hburst),
       .dap_ahb_hwdata_o          (dap_ahb_hwdata),
       .dap_ahb_hmaster_o         (dap_ahb_hmaster),
       .mst_resp_err_o            (mst_resp_err),
       .mst_hrdata_o              (mst_hrdata),
       .mst_end_trans_o           (mst_end_trans)
      );



  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 0)  // No glitch check
    u_cm3_dap_ahb_ap_sync_habort
      (// Inputs
       .clk                       (hclk),
       .reset_n                   (hreset_n),
       .d_async_i                 (dap_ahb_habort),
       .q_o                       (dap_ahb_habort_sync)
      );

  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 1)  // Glitch check
    u_cm3_dap_ahb_ap_sync_htrans
      (// Inputs
       .clk                       (hclk),
       .reset_n                   (hreset_n),
       .d_async_i                 (slv_req_trans),
       .q_o                       (slv_req_trans_sync)
      );

  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 0)  // No glitch check
    u_cm3_dap_ahb_ap_sync_idle
      (// Inputs
       .clk                       (hclk),
       .reset_n                   (hreset_n),
       .d_async_i                 (slv_idle),
       .q_o                       (slv_idle_sync)
      );

  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 0)  // No glitch check
    u_cm3_dap_ahb_ap_sync_tra
      (// Inputs
       .clk                       (hclk),
       .reset_n                   (hreset_n),
       .d_async_i                 (slv_tra_update),
       .q_o                       (slv_tra_update_sync)
      );

  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 1)  // Glitch check
    u_cm3_dap_ahb_ap_sync_trans_rtn
      (// Inputs
       .clk                       (dap_clk),
       .reset_n                   (dap_reset_n),
       .d_async_i                 (slv_req_trans_sync),
       .q_o                       (slv_req_trans_rtn)
      );

  cm3_dap_ahb_ap_sync #(DEBUG_LVL, 1)  // Glitch check
    u_cm3_dap_ahb_ap_sync_end_trans
      (// Inputs
       .clk                       (dap_clk),
       .reset_n                   (dap_reset_n),
       .d_async_i                 (mst_end_trans),
       .q_o                       (mst_end_trans_sync)
      );



  assign dap_ready_o        = opt_dap_ready;
  assign dap_slverr_o       = opt_dap_slverr;
  assign dap_rdata_o        = opt_dap_rdata;

  assign dap_ahb_haddr_o    = opt_dap_ahb_haddr;
  assign dap_ahb_hmaster_o  = opt_dap_ahb_hmaster;
  assign dap_ahb_htrans_o   = opt_dap_ahb_htrans;
  assign dap_ahb_hwrite_o   = opt_dap_ahb_hwrite;
  assign dap_ahb_hsize_o    = opt_dap_ahb_hsize;
  assign dap_ahb_hburst_o   = opt_dap_ahb_hburst;
  assign dap_ahb_hprot_o    = opt_dap_ahb_hprot;
  assign dap_ahb_hwdata_o   = opt_dap_ahb_hwdata;
  assign dap_ahb_habort_o   = opt_dap_ahb_habort;



`ifdef ARM_MONITOR_ON

  cm3_dap_ahb_ap_monitor
    u_cm3_dap_ahb_ap_monitor
      (// Inputs
       .hclk                      (hclk),
       .hreset_n                  (hreset_n),
       .dap_clk                   (dap_clk),
       .dap_reset_n               (dap_reset_n),
       .dap_clken_i               (dap_clken_i),
       .master_enable_i           (master_enable_i),
       .dap_en_i                  (dap_en_i),
       .dap_sel_i                 (dap_sel_i),
       .dap_enable_i              (dap_enable_i),
       .dap_write_i               (dap_write_i),
       .dap_abort_i               (dap_abort_i),
       .dap_addr_i                (dap_addr_i),
       .dap_wdata_i               (dap_wdata_i),
       .dap_ready_i               (opt_dap_ready),
       .dap_slverr_i              (opt_dap_slverr),
       .dap_rdata_i               (opt_dap_rdata),
       .dap_ahb_hready_i          (dap_ahb_hready_i),
       .dap_ahb_haddracc_i        (dap_ahb_haddracc_i),
       .dap_ahb_hresp_i           (dap_ahb_hresp_i),
       .dap_ahb_hrdata_i          (dap_ahb_hrdata_i),
       .dap_ahb_haddr_i           (opt_dap_ahb_haddr),
       .dap_ahb_hmaster_i         (opt_dap_ahb_hmaster),
       .dap_ahb_htrans_i          (opt_dap_ahb_htrans),
       .dap_ahb_hwrite_i          (opt_dap_ahb_hwrite),
       .dap_ahb_hsize_i           (opt_dap_ahb_hsize),
       .dap_ahb_hburst_i          (opt_dap_ahb_hburst),
       .dap_ahb_hprot_i           (opt_dap_ahb_hprot),
       .dap_ahb_hwdata_i          (opt_dap_ahb_hwdata),
       .dap_ahb_habort_i          (opt_dap_ahb_habort),
       .dap_ahb_habort_dapclk_i   (dap_ahb_habort)
      );

`endif



`ifdef ARM_ASSERT_ON

  wire  ovl_hs_start;          // Handshake start
  wire  ovl_hs_end;            // Handshake end
  reg   ovl_slv_req_trans_reg; // Delayed transfer request
  reg   ovl_trans_reg;         // Delayed transfer request
  wire  nxt_ovl_trans_reg;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        ovl_slv_req_trans_reg <= 1'b0;
        ovl_trans_reg         <= 1'b0;
      end
    else
      begin
        ovl_slv_req_trans_reg <= slv_req_trans;
        ovl_trans_reg         <= nxt_ovl_trans_reg;
      end

  assign nxt_ovl_trans_reg = (opt_dap_ahb_htrans[1] & dap_ahb_hready_i) |
                             (ovl_trans_reg & ~dap_ahb_hready_i);

  assign ovl_hs_start = (slv_req_trans ^ ovl_slv_req_trans_reg); // Pulse
  assign ovl_hs_end   = ovl_trans_reg & dap_ahb_hready_i;

  assert_handshake #(0,0,0,0,0,0,0,"Master/slave handshake must be complete.")
    ovl_ap_mstslv_handshake(hclk, hreset_n, slv_req_trans, mst_end_trans);

  assert_never #(0,0,"DAP transfer cannot occur if DEBUG_LVL == 0")
    ovl_cfg_dap(hclk, hreset_n, ((DEBUG_LVL == 0) & (|opt_dap_ahb_htrans)));

`endif

endmodule
