`include "cm3_dap_ahb_ap_defs.vp"

module cm3_dap_ahb_ap_mst
  (// Inputs
   hclk, hreset_n, dap_ahb_hready_i, dap_ahb_haddracc_i, dap_ahb_hresp_i,
   dap_ahb_hrdata_i, slv_haddr_async_i, slv_hsize_async_i, slv_hwrite_async_i,
   slv_hprot1_async_i, slv_hmaster_async_i, fix_master_type_i,
   dap_wdata_async_i, slv_req_trans_i, slv_idle_i, slv_tra_update_i,
   dap_ahb_htrans_o, dap_ahb_haddr_o, dap_ahb_hsize_o, dap_ahb_hwrite_o,
   dap_ahb_hprot_o, dap_ahb_hburst_o, dap_ahb_hwdata_o, dap_ahb_hmaster_o,
   mst_resp_err_o, mst_hrdata_o, mst_end_trans_o
  );


  parameter DEBUG_LVL      = 3;  // Debug support level
  parameter RESET_ALL_REGS = 0;  // Reset all registers if 1.



  input         hclk;               // Clock in master domain
  input         hreset_n;           // Reset in master domain

  output [31:0] dap_ahb_haddr_o;    // Transfer address
  output        dap_ahb_hmaster_o;  // AHB Master type
  output  [1:0] dap_ahb_htrans_o;   // Transfer type (IDLE or NONSEQ)
  output  [1:0] dap_ahb_hsize_o;    // transfer size
  output        dap_ahb_hwrite_o;   // transfer direction
  output  [3:0] dap_ahb_hprot_o;    // protection control
  output  [2:0] dap_ahb_hburst_o;   // Burst type (always SINGLE)
  output [31:0] dap_ahb_hwdata_o;   // Write data
  input         dap_ahb_hready_i;   // AHB slave ready
  input         dap_ahb_haddracc_i; // AHB address ready
  input         dap_ahb_hresp_i;    // AHB slave response
  input  [31:0] dap_ahb_hrdata_i;   // Read data from AHB slave
  input         fix_master_type_i;  // Fix Master Type to 1 is asserted

  input  [31:0] slv_haddr_async_i;  // AHB address in DAP clock domain
  input   [1:0] slv_hsize_async_i;  // AHB size in DAP clock domain
  input         slv_hwrite_async_i; // AHB write in DAP clock domain
  input         slv_hprot1_async_i; // AHB PROT[1] in DAP clock domain
  input         slv_hmaster_async_i;// AHB master type
  output        mst_resp_err_o;     // AHB slave responded with an ERROR
  output [31:0] mst_hrdata_o;       // Registered dap_ahb_hrdata_i

  input  [31:0] dap_wdata_async_i;  // DAP write data

  input         slv_req_trans_i;    // Transfer request from degug access port
  output        mst_end_trans_o;    // Transfer completed on master interface

  input         slv_tra_update_i;   // TraUpdated handshake

  input         slv_idle_i;         // DAP is in idle state, clear mst_resp_err_o



  wire          full_hreset_n;        // Full register reset, when enabled

  reg     [1:0] mst_state;            // Master state machine
  reg     [1:0] nxt_mst_state;        // Next state

  reg           slv_req_trans_reg;    // Registered slv_req_trans
  reg           slv_tra_update_reg;   // Registered slv_tra_update
  wire          req_trans_pulse;      // slv_req_trans_i high for a single cycle
  wire          tra_update_pulse;     // Detect address update

  wire          end_trans_pulse;      // Pulse when transfer completed
  reg           mst_end_trans;        // End of transfer on master interface
  wire          mst_end_trans_we;
  wire          nxt_mst_end_trans;

  reg    [31:0] slv_haddr_reg;        // Sampling register for HADDR
  wire          slv_haddr_reg_we;     // Write enable for HADDR reg
  reg           slv_hprot1_reg;       // Sampling register for HPROT1
  reg     [1:0] slv_hsize_reg;        // Sampling register for HSIZE
  reg           slv_hwrite_reg;       // Sampling register for HWRITE
  reg           slv_hmaster_reg;      // Sampling register for HMASTER

  reg     [7:0] shared_rwdata0;       // Read/write data, byte lane 0
  reg     [7:0] shared_rwdata1;       // Read/write data, byte lane 1
  reg     [7:0] shared_rwdata2;       // Read/write data, byte lane 2
  reg     [7:0] shared_rwdata3;       // Read/write data, byte lane 3
  wire          shared_rwdata0_we;    // Write enable for byte lane 0
  wire          shared_rwdata1_we;    // Write enable for byte lane 1
  wire          shared_rwdata2_we;    // Write enable for byte lane 2
  wire          shared_rwdata3_we;    // Write enable for byte lane 3
  wire   [31:0] nxt_shared_rwdata;    // Next shared read/write data value
  wire          rdata_sample0;        // Sample HRDATA byte lane 0
  wire          rdata_sample1;        // Sample HRDATA byte lane 1
  wire          rdata_sample2;        // Sample HRDATA byte lane 2
  wire          rdata_sample3;        // Sample HRDATA byte lane 3
  wire   [31:0] rdata_mask;           // Read data mask
  wire          wdata_sample;         // Sample HWDATA
  wire   [31:0] wdata_mask;           // Write data mask

  wire   [31:0] dap_ahb_haddr;        // Masked AHB address
  reg     [1:0] dap_ahb_htrans;       // AHB HTRANS
  wire    [2:0] dap_ahb_hburst;       // AHB HBURST
  wire    [3:0] dap_ahb_hprot;        // AHB HPROT
  wire   [31:0] dap_ahb_hwdata;       // AHB HWDATA
  wire          dap_ahb_hmaster;      // AHB HMASTER
  wire    [1:0] nxt_dap_ahb_htrans;   // Next HTRANS

  wire   [31:0] mst_hrdata;           // Read data output
  reg           mst_resp_err;         // Error response
  wire          nxt_mst_resp_err;
  wire          mst_resp_err_we;

  reg           fix_master_type_reg;
  wire          fix_master_type_reg_we;
  wire          nxt_fix_master_type_reg;

  wire    [1:0] opt_dap_ahb_htrans;
  wire   [31:0] opt_dap_ahb_haddr;
  wire    [2:0] opt_dap_ahb_hburst;
  wire    [3:0] opt_dap_ahb_hprot;
  wire    [1:0] opt_dap_ahb_hsize;
  wire          opt_dap_ahb_hwrite;
  wire   [31:0] opt_dap_ahb_hwdata;
  wire          opt_dap_ahb_hmaster;
  wire          opt_mst_resp_err;
  wire   [31:0] opt_mst_hrdata;
  wire          opt_mst_end_trans;
  wire          opt_mst_en;

`ifdef ARM_DSM


  wire       dsm_reset_all_regs;
  wire [1:0] dsm_debug_lvl;

  assign full_hreset_n = dsm_reset_all_regs ? hreset_n : 1'b1;


  assign opt_dap_ahb_htrans  = (dsm_debug_lvl > 0) ? dap_ahb_htrans  : 2'b00;
  assign opt_dap_ahb_haddr   = (dsm_debug_lvl > 0) ? dap_ahb_haddr   : {32{1'b0}};
  assign opt_dap_ahb_hburst  = (dsm_debug_lvl > 0) ? dap_ahb_hburst  : 3'b000;
  assign opt_dap_ahb_hprot   = (dsm_debug_lvl > 0) ? dap_ahb_hprot   : 4'b0000;
  assign opt_dap_ahb_hsize   = (dsm_debug_lvl > 0) ? slv_hsize_reg   : 2'b00;
  assign opt_dap_ahb_hwrite  = (dsm_debug_lvl > 0) ? slv_hwrite_reg  : 1'b0;
  assign opt_dap_ahb_hwdata  = (dsm_debug_lvl > 0) ? dap_ahb_hwdata  : {32{1'b0}};
  assign opt_dap_ahb_hmaster = (dsm_debug_lvl > 0) ? dap_ahb_hmaster : 1'b0;
  assign opt_mst_resp_err    = (dsm_debug_lvl > 0) ? mst_resp_err    : 1'b0;
  assign opt_mst_hrdata      = (dsm_debug_lvl > 0) ? mst_hrdata      : {32{1'b0}};
  assign opt_mst_end_trans   = (dsm_debug_lvl > 0) ? mst_end_trans   : 1'b0;
  assign opt_mst_en          = (dsm_debug_lvl > 0) ? 1'b1            : 1'b0;

`else


  assign full_hreset_n = (RESET_ALL_REGS == 1) ? hreset_n : 1'b1;


  assign opt_dap_ahb_htrans  = (DEBUG_LVL > 0) ? dap_ahb_htrans  : 2'b00;
  assign opt_dap_ahb_haddr   = (DEBUG_LVL > 0) ? dap_ahb_haddr   : {32{1'b0}};
  assign opt_dap_ahb_hburst  = (DEBUG_LVL > 0) ? dap_ahb_hburst  : 3'b000;
  assign opt_dap_ahb_hprot   = (DEBUG_LVL > 0) ? dap_ahb_hprot   : 4'b0000;
  assign opt_dap_ahb_hsize   = (DEBUG_LVL > 0) ? slv_hsize_reg   : 2'b00;
  assign opt_dap_ahb_hwrite  = (DEBUG_LVL > 0) ? slv_hwrite_reg  : 1'b0;
  assign opt_dap_ahb_hwdata  = (DEBUG_LVL > 0) ? dap_ahb_hwdata  : {32{1'b0}};
  assign opt_dap_ahb_hmaster = (DEBUG_LVL > 0) ? dap_ahb_hmaster : 1'b0;
  assign opt_mst_resp_err    = (DEBUG_LVL > 0) ? mst_resp_err    : 1'b0;
  assign opt_mst_hrdata      = (DEBUG_LVL > 0) ? mst_hrdata      : {32{1'b0}};
  assign opt_mst_end_trans   = (DEBUG_LVL > 0) ? mst_end_trans   : 1'b0;
  assign opt_mst_en          = (DEBUG_LVL > 0) ? 1'b1            : 1'b0;

`endif


  always @ (req_trans_pulse or dap_ahb_hready_i or dap_ahb_haddracc_i or
            mst_state)
    case (mst_state)
      `DAP_MST_IDLE: nxt_mst_state = req_trans_pulse ? `DAP_MST_ADDR : `DAP_MST_IDLE;

      `DAP_MST_ADDR: nxt_mst_state = (dap_ahb_hready_i & dap_ahb_haddracc_i) ?
                                      `DAP_MST_DATA : `DAP_MST_ADDR;

      `DAP_MST_DATA: nxt_mst_state = dap_ahb_hready_i ? `DAP_MST_IDLE : `DAP_MST_DATA;

      default:      nxt_mst_state = `DAP_MST_IDLE;
    endcase


  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      mst_state <= `DAP_MST_IDLE;
    else
      mst_state <= nxt_mst_state;



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      slv_req_trans_reg <= 1'b0;
    else
      slv_req_trans_reg <= slv_req_trans_i;

  assign req_trans_pulse = (slv_req_trans_i ^ slv_req_trans_reg);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      slv_tra_update_reg <= 1'b0;
    else
      slv_tra_update_reg <= slv_tra_update_i;

  assign tra_update_pulse = (slv_tra_update_i ^ slv_tra_update_reg);



  assign end_trans_pulse = (mst_state == `DAP_MST_DATA) & dap_ahb_hready_i;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      mst_end_trans <= 1'b0;
    else if (mst_end_trans_we)
      mst_end_trans <= nxt_mst_end_trans;

  assign mst_end_trans_we  = end_trans_pulse & opt_mst_en;
  assign nxt_mst_end_trans = ~mst_end_trans;  // Toggle



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        slv_hprot1_reg  <= 1'b0;
        slv_hsize_reg   <= 2'b00;
        slv_hwrite_reg  <= 1'b0;
        slv_hmaster_reg <= 1'b0;
      end
    else if (req_trans_pulse)
      begin
        slv_hprot1_reg  <= slv_hprot1_async_i;
        slv_hsize_reg   <= slv_hsize_async_i;
        slv_hwrite_reg  <= slv_hwrite_async_i;
        slv_hmaster_reg <= slv_hmaster_async_i;
      end



  assign nxt_dap_ahb_htrans = (nxt_mst_state == `DAP_MST_ADDR) ?
                                `DAP_TRN_NONSEQ : `DAP_TRN_IDLE;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      dap_ahb_htrans <= `DAP_TRN_IDLE;
    else
      dap_ahb_htrans <= nxt_dap_ahb_htrans;

  assign dap_ahb_hburst = `DAP_BUR_SINGLE;  // Only single bursts supported
  assign dap_ahb_hprot  = {2'b00, slv_hprot1_reg, 1'b1};



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      fix_master_type_reg <= 1'b0;
    else if (fix_master_type_reg_we)
      fix_master_type_reg <= nxt_fix_master_type_reg;

  assign fix_master_type_reg_we  = dap_ahb_hready_i & opt_mst_en;

  assign nxt_fix_master_type_reg = dap_ahb_hready_i ? fix_master_type_i :
                                                      fix_master_type_reg;

  assign dap_ahb_hmaster = slv_hmaster_reg | fix_master_type_reg;



  assign slv_haddr_reg_we = (req_trans_pulse | tra_update_pulse) & opt_mst_en;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      slv_haddr_reg <= {32{1'b0}};
    else if (slv_haddr_reg_we)
      slv_haddr_reg <= slv_haddr_async_i[31:0];


  assign rdata_sample0 = (end_trans_pulse & ~slv_hwrite_reg) &
                         (// Word
                          (slv_hsize_reg == 2'b10) |
                          ((slv_hsize_reg == 2'b01) &
                            ((slv_haddr_reg[1:0] == 2'b00)  |
                             (slv_haddr_reg[1:0] ==2'b11))) |
                          ((slv_hsize_reg == 2'b00) &
                            (slv_haddr_reg[1:0] == 2'b00)));

  assign rdata_sample1 = (end_trans_pulse & ~slv_hwrite_reg) &
                         ((slv_hsize_reg == 2'b10) |
                          ((slv_hsize_reg == 2'b01) &
                            ((slv_haddr_reg[1:0] == 2'b01)   |
                             (slv_haddr_reg[1:0] == 2'b00))) |
                          ((slv_hsize_reg == 2'b00) &
                            (slv_haddr_reg[1:0] == 2'b01)));

  assign rdata_sample2 = (end_trans_pulse & ~slv_hwrite_reg) &
                         ((slv_hsize_reg == 2'b10) |
                          ((slv_hsize_reg == 2'b01) &
                            ((slv_haddr_reg[1:0] == 2'b10)   |
                             (slv_haddr_reg[1:0] == 2'b01))) |
                          ((slv_hsize_reg == 2'b00) &
                            (slv_haddr_reg[1:0] == 2'b10)));

  assign rdata_sample3 = (end_trans_pulse & ~slv_hwrite_reg) &
                         ((slv_hsize_reg == 2'b10) |
                          ((slv_hsize_reg == 2'b01) &
                            ((slv_haddr_reg[1:0] == 2'b11)   |
                             (slv_haddr_reg[1:0] == 2'b10))) |
                          ((slv_hsize_reg == 2'b00) &
                            (slv_haddr_reg[1:0] == 2'b11)));


  assign rdata_mask = {{8{rdata_sample3}},
                       {8{rdata_sample2}},
                       {8{rdata_sample1}},
                       {8{rdata_sample0}}};

  assign wdata_sample = (mst_state == `DAP_MST_ADDR) &
                        dap_ahb_hready_i             &
                        dap_ahb_haddracc_i           &
                        slv_hwrite_reg;

  assign wdata_mask   = {32{wdata_sample}};

  assign nxt_shared_rwdata = (rdata_mask & dap_ahb_hrdata_i) |
                             (wdata_mask & dap_wdata_async_i);

  assign shared_rwdata0_we = (rdata_sample0 | wdata_sample) & opt_mst_en;
  assign shared_rwdata1_we = (rdata_sample1 | wdata_sample) & opt_mst_en;
  assign shared_rwdata2_we = (rdata_sample2 | wdata_sample) & opt_mst_en;
  assign shared_rwdata3_we = (rdata_sample3 | wdata_sample) & opt_mst_en;

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      shared_rwdata0 <= {8{1'b0}};
    else if (shared_rwdata0_we)
      shared_rwdata0 <= nxt_shared_rwdata[7:0];

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      shared_rwdata1 <= {8{1'b0}};
    else if (shared_rwdata1_we)
      shared_rwdata1 <= nxt_shared_rwdata[15:8];

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      shared_rwdata2 <= {8{1'b0}};
    else if (shared_rwdata2_we)
      shared_rwdata2 <= nxt_shared_rwdata[23:16];

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      shared_rwdata3 <= {8{1'b0}};
    else if (shared_rwdata3_we)
      shared_rwdata3 <= nxt_shared_rwdata[31:24];


  assign dap_ahb_hwdata = {shared_rwdata3,
                           shared_rwdata2,
                           shared_rwdata1,
                           shared_rwdata0};

  assign dap_ahb_haddr  = {slv_haddr_reg[31:2],
                           ({2{mst_state == `DAP_MST_ADDR}} & slv_haddr_reg[1:0])};



  assign mst_hrdata = {shared_rwdata3,
                       shared_rwdata2,
                       shared_rwdata1,
                       shared_rwdata0};

  always @ (negedge hreset_n or posedge hclk)
    if (!hreset_n)
      mst_resp_err <= 1'b0;
    else if (mst_resp_err_we)
      mst_resp_err <= nxt_mst_resp_err;

  assign nxt_mst_resp_err = (dap_ahb_hresp_i != `DAP_RSP_OKAY) & ~slv_idle_i;
  assign mst_resp_err_we  = (end_trans_pulse | (mst_resp_err & slv_idle_i)) & opt_mst_en;



  assign dap_ahb_haddr_o   = opt_dap_ahb_haddr;
  assign dap_ahb_hmaster_o = opt_dap_ahb_hmaster;
  assign dap_ahb_htrans_o  = opt_dap_ahb_htrans;
  assign dap_ahb_hsize_o   = opt_dap_ahb_hsize;
  assign dap_ahb_hwrite_o  = opt_dap_ahb_hwrite;
  assign dap_ahb_hprot_o   = opt_dap_ahb_hprot;
  assign dap_ahb_hburst_o  = opt_dap_ahb_hburst;
  assign dap_ahb_hwdata_o  = opt_dap_ahb_hwdata;

  assign mst_resp_err_o    = opt_mst_resp_err;
  assign mst_hrdata_o      = opt_mst_hrdata;

  assign mst_end_trans_o   = opt_mst_end_trans;



`ifdef ARM_ASSERT_ON

  reg           ovl_req_trans_pulse_reg; // Delay request pulse
  reg           ovl_end_trans_pulse_reg; // Delay end pulse

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        ovl_req_trans_pulse_reg <= 1'b0;
        ovl_end_trans_pulse_reg <= 1'b0;
      end
    else
      begin
        ovl_req_trans_pulse_reg <= req_trans_pulse;
        ovl_end_trans_pulse_reg <= end_trans_pulse;
      end

  assert_never #(0,0,"req_trans_pulse must be a single cycle pulse.")
    ovl_pulse_reqtrans(hclk, hreset_n, (req_trans_pulse & ovl_req_trans_pulse_reg));

  assert_never #(0,0,"end_trans_pulse must be a single cycle pulse.")
    ovl_pulse_endtrans(hclk, hreset_n, (end_trans_pulse & ovl_end_trans_pulse_reg));

`endif

endmodule

`include "cm3_dap_ahb_ap_undefs.vp"
