`include "cm3_dap_ahb_ap_defs.vp"

module cm3_dap_ahb_ap_slv
  (// Inputs
   dap_clk, dap_reset_n, dap_en_i, master_enable_i, dap_clken_i, dap_sel_i,
   dap_enable_i, dap_write_i, dap_addr_i, dap_wdata_i, dap_abort_i,
   mst_resp_err_async_i, mst_hrdata_async_i, mst_end_trans_i,
   slv_req_trans_rtn_i,
   dap_ready_o, dap_slverr_o, dap_rdata_o, slv_haddr_o, slv_hwrite_o,
   slv_hsize_o, dap_ahb_habort_o, slv_hprot1_o, slv_hmaster_o, slv_req_trans_o,
   slv_idle_o, slv_tra_update_o
  );


  parameter DEBUG_LVL      = 3;  // Debug support level
  parameter RESET_ALL_REGS = 0;  // Reset all registers if 1



  input         dap_clk;              // DAP APB clock
  input         dap_reset_n;          // DAP APB reset
  input         dap_clken_i;          // DAP APB clock enable

  input         dap_en_i;             // Debug enable - Enables AHB transfers
  input         master_enable_i;      // Master enable - Enabled Core HMASTER

  input         dap_sel_i;            // Module select signal
  input         dap_enable_i;         // Data valid strobe
  input         dap_write_i;          // Write/nRead signal
  input   [7:0] dap_addr_i;           // Address (used bits only)
  input  [31:0] dap_wdata_i;          // Write data
  input         dap_abort_i;          // Transaction abort
  output        dap_ready_o;          // Read data ready signal
  output        dap_slverr_o;         // Read data error from this slave
  output [31:0] dap_rdata_o;          // Read data

  output [31:0] slv_haddr_o;          // Address
  output        slv_hwrite_o;         // Transfer direction
  output  [1:0] slv_hsize_o;          // Transfer size
  output        dap_ahb_habort_o;     // DAP -> bus matrix HREADY abort
  output        slv_hprot1_o;         // HPROT[1]
  output        slv_hmaster_o;        // HMASTER
  input         mst_resp_err_async_i; // AHB slave responded with an error
  input  [31:0] mst_hrdata_async_i;   // Registered HRDATA

  output        slv_req_trans_o;      // Request transfer to AHB master interface
  input         slv_req_trans_rtn_i;  // Return of slv_req_trans from HCLK domain
  input         mst_end_trans_i;      // Transfer completed on master interface

  output        slv_tra_update_o;     // Target address register updated handshake

  output        slv_idle_o;           // DAP is in IDLE state



  wire          full_dap_reset_n;       // Full register reset term

  wire          dap_trans_valid;        // Transfer on DAP interface
  reg     [7:0] dap_addr_reg;           // Registered DAP address
  reg           dap_en_reg;             // Registered debug enable
  wire          dap_addr_reg_we;        // DAP address register write enable

  wire          dap_reg_access;         // Internal register access
  wire          ahb_bd_access;          // Bank data access
  wire          ahb_drw_access;         // Data read/write access
  wire          dap_csw_wr;             // Write to Control/Status Word
  wire          dap_tra_wr;             // Write to Transfer Address Register

  wire          slv_trans_start;        // Start the next transfer
  reg     [2:0] slv_state;              // State Machine - current state
  reg     [2:0] nxt_slv_state;          // State Machine - next state
  reg           slv_state_enable;       // State Machine - state enable
  wire          slv_state_we;           // State Machine - reg write enable
  reg           slv_idle;               // State machine in idle state
  wire          slv_idle_we;
  wire          nxt_slv_idle;
  wire          slv_trans_active;       // Transfer is in progress

  reg           dap_csw_reg_hmaster;    // CSW HMASTER field
  reg           dap_csw_reg_hprot1;     // CSW HPROT1 field
  reg     [1:0] dap_csw_reg_addrinc;    // CSW AddrInc (Address increment) field
  reg     [1:0] dap_csw_reg_size;       // CSW HSIZE field (bits[1:0])
  wire          nxt_dap_csw_reg_hmaster;
  wire    [1:0] nxt_dap_csw_reg_addrinc;
  wire    [1:0] nxt_dap_csw_reg_size;
  wire          dap_csw_reg_we;
  wire   [31:0] dap_csw_reg;            // CSW register read value

  reg    [31:0] dap_tra_reg;            // Transfer Address register
  wire   [31:0] nxt_dap_tra_reg;        // Next value of dap_tra_reg
  wire          dap_tra_reg_we;         // Write enable for dap_tra_reg
  wire          dap_tra_autoinc;        // Auto Increment Enable
  reg     [2:0] addr_inc_size;          // Increment value for the transfer address

  reg           dap_ready;
  wire          dap_slverr;
  reg    [31:0] dap_rdata;
  reg           dap_abort_reg;          // Registered DAP abort
  reg           nxt_dap_ready;          // Logic for DAPREADY
  wire          dap_ready_we;           // Write enable for dap_ready register

  reg           dap_ahb_habort;
  wire          dap_ahb_habort_set;     // Bus matrix abort set
  wire          dap_ahb_habort_clr;     // Bus matrix abort clear
  wire          dap_ahb_habort_we;      // Bus matrix abort write enable
  wire          nxt_dap_ahb_habort;     // Bus matrix abort next value
  reg           trans_abort_reg;        // Registered transaction abort
  wire          nxt_trans_abort_reg;    // Next transaction abort register
  wire          trans_abort_reg_we;     // Transaction abort register write enable

  reg           slv_req_trans;     // Transfer request
  wire          nxt_slv_req_trans;
  wire          slv_req_trans_we;
  wire          req_trans_pulse;   // Pulse to toggle ReqTranReg
  reg           end_trans_reg;     // Transfer completed on AHB interface
  wire          end_trans_reg_we;
  wire          end_trans_pulse;   // Pulse generated when mst_end_trans_i toggles

  wire   [31:0] slv_haddr;
  wire          slv_hwrite;
  wire    [1:0] slv_hsize;
  wire          slv_hprot1;
  wire          slv_hmaster;
  reg           slv_tra_update;
  wire          nxt_slv_tra_update;

  wire          opt_dap_ready;
  wire          opt_dap_slverr;
  wire   [31:0] opt_dap_rdata;
  wire   [31:0] opt_slv_haddr;
  wire          opt_slv_hwrite;
  wire    [1:0] opt_slv_hsize;
  wire          opt_slv_hprot1;
  wire          opt_slv_hmaster;
  wire          opt_slv_req_trans;
  wire          opt_slv_idle;
  wire          opt_dap_ahb_habort;
  wire          opt_mst_en;

`ifdef ARM_DSM


  wire          dsm_reset_all_regs;
  wire [1:0]    dsm_debug_lvl;

  assign full_dap_reset_n = dsm_reset_all_regs  ? dap_reset_n   : 1'b1;


  assign opt_dap_ready      = (dsm_debug_lvl > 0) ? dap_ready      : 1'b0;
  assign opt_dap_slverr     = (dsm_debug_lvl > 0) ? dap_slverr     : 1'b0;
  assign opt_dap_rdata      = (dsm_debug_lvl > 0) ? dap_rdata      : {32{1'b0}};
  assign opt_slv_haddr      = (dsm_debug_lvl > 0) ? slv_haddr      : {32{1'b0}};
  assign opt_slv_hwrite     = (dsm_debug_lvl > 0) ? slv_hwrite     : 1'b0;
  assign opt_slv_hsize      = (dsm_debug_lvl > 0) ? slv_hsize      : 2'b00;
  assign opt_slv_hprot1     = (dsm_debug_lvl > 0) ? slv_hprot1     : 1'b0;
  assign opt_slv_hmaster    = (dsm_debug_lvl > 0) ? slv_hmaster    : 1'b0;
  assign opt_slv_req_trans  = (dsm_debug_lvl > 0) ? slv_req_trans  : 1'b0;
  assign opt_slv_idle       = (dsm_debug_lvl > 0) ? slv_idle       : 1'b1;
  assign opt_dap_ahb_habort = (dsm_debug_lvl > 0) ? dap_ahb_habort : 1'b0;
  assign opt_mst_en         = (dsm_debug_lvl > 0) ? 1'b1           : 1'b0;

`else


  assign full_dap_reset_n = (RESET_ALL_REGS == 1) ? dap_reset_n : 1'b1;


  assign opt_dap_ready      = (DEBUG_LVL > 0) ? dap_ready      : 1'b0;
  assign opt_dap_slverr     = (DEBUG_LVL > 0) ? dap_slverr     : 1'b0;
  assign opt_dap_rdata      = (DEBUG_LVL > 0) ? dap_rdata      : {32{1'b0}};
  assign opt_slv_haddr      = (DEBUG_LVL > 0) ? slv_haddr      : {32{1'b0}};
  assign opt_slv_hwrite     = (DEBUG_LVL > 0) ? slv_hwrite     : 1'b0;
  assign opt_slv_hsize      = (DEBUG_LVL > 0) ? slv_hsize      : 2'b00;
  assign opt_slv_hprot1     = (DEBUG_LVL > 0) ? slv_hprot1     : 1'b0;
  assign opt_slv_hmaster    = (DEBUG_LVL > 0) ? slv_hmaster    : 1'b0;
  assign opt_slv_req_trans  = (DEBUG_LVL > 0) ? slv_req_trans  : 1'b0;
  assign opt_slv_idle       = (DEBUG_LVL > 0) ? slv_idle       : 1'b1;
  assign opt_dap_ahb_habort = (DEBUG_LVL > 0) ? dap_ahb_habort : 1'b0;
  assign opt_mst_en         = (DEBUG_LVL > 0) ? 1'b1           : 1'b0;

`endif




  assign dap_trans_valid = dap_clken_i & dap_sel_i & ~dap_enable_i;

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      begin
        dap_addr_reg  <= {8{1'b0}};
        dap_en_reg    <= 1'b0;
      end
    else if (dap_addr_reg_we)
      begin
        dap_addr_reg  <= dap_addr_i[7:0];
        dap_en_reg    <= dap_en_i;
      end

  assign dap_addr_reg_we = dap_trans_valid & opt_mst_en;



  assign dap_csw_wr = (dap_addr_i[7:2] == `DAP_CSWADDR) & dap_trans_valid & dap_write_i;

  assign dap_tra_wr = (dap_addr_i[7:2] == `DAP_TRAADDR) & dap_trans_valid & dap_write_i;

  assign dap_reg_access = (dap_addr_i[7:2] == `DAP_CSWADDR) |
                          (dap_addr_i[7:2] == `DAP_TRAADDR) |
                          (dap_addr_i[7:2] == `DAP_ROMADDR) |
                          (dap_addr_i[7:2] == `DAP_IDRADDR);

  assign ahb_bd_access  = (dap_addr_i[7:2] == `DAP_BD0ADDR) |
                          (dap_addr_i[7:2] == `DAP_BD1ADDR) |
                          (dap_addr_i[7:2] == `DAP_BD2ADDR) |
                          (dap_addr_i[7:2] == `DAP_BD3ADDR);

  assign ahb_drw_access = (dap_addr_i[7:2] == `DAP_DRWADDR);



  assign slv_trans_start = dap_sel_i &
                           (dap_reg_access | ahb_bd_access | ahb_drw_access);

  always @ (slv_state or dap_clken_i or dap_en_i or slv_trans_start or
            dap_reg_access or ahb_bd_access or mst_resp_err_async_i or
            dap_csw_reg_addrinc or dap_csw_reg_size or end_trans_pulse or
            trans_abort_reg)

    case (slv_state)

      `DAP_SLV_IDLE:
        begin

          case ({dap_reg_access, dap_en_i, ahb_bd_access})
            3'b100, 3'b101, 3'b110, 3'b111:
              nxt_slv_state = `DAP_SLV_REG;
            3'b000, 3'b001:
              nxt_slv_state = `DAP_SLV_READY;
            3'b011:
              nxt_slv_state = `DAP_SLV_AHBSING;
            3'b010:
              nxt_slv_state = `DAP_SLV_AHB0INC;
            default:
              nxt_slv_state = 3'bXXX;
          endcase

          slv_state_enable = dap_clken_i & slv_trans_start;
        end

      `DAP_SLV_REG:
        begin
          nxt_slv_state    = `DAP_SLV_IDLE;
          slv_state_enable = dap_clken_i;
        end

      `DAP_SLV_AHBSING:
        begin
          nxt_slv_state    = `DAP_SLV_READY;
          slv_state_enable = dap_clken_i & (end_trans_pulse | trans_abort_reg);
        end

      `DAP_SLV_AHB0INC:
        begin
          nxt_slv_state    = ((dap_csw_reg_addrinc == `DAP_INCPACKED) &
                              ~mst_resp_err_async_i                   &
                               ((dap_csw_reg_size    == `DAP_SZ_BYTE) |
                                (dap_csw_reg_size == `DAP_SZ_HALF))) ?
                             `DAP_SLV_AHB1INC : `DAP_SLV_READY;

          slv_state_enable = dap_clken_i & (end_trans_pulse|trans_abort_reg);
        end

      `DAP_SLV_AHB1INC:
        begin
          nxt_slv_state    = ((dap_csw_reg_size == `DAP_SZ_BYTE) & ~mst_resp_err_async_i) ?
                             `DAP_SLV_AHB2INC : `DAP_SLV_READY;

          slv_state_enable = dap_clken_i & (end_trans_pulse|trans_abort_reg);
        end

      `DAP_SLV_AHB2INC:
        begin
          nxt_slv_state    = (~mst_resp_err_async_i) ? `DAP_SLV_AHB3INC : `DAP_SLV_READY;
          slv_state_enable = dap_clken_i & (end_trans_pulse | trans_abort_reg);
        end

      `DAP_SLV_AHB3INC:
        begin
          nxt_slv_state    = `DAP_SLV_READY;
          slv_state_enable = dap_clken_i & (end_trans_pulse | trans_abort_reg);
        end

      `DAP_SLV_READY:
        begin
          nxt_slv_state    = `DAP_SLV_IDLE;
          slv_state_enable = dap_clken_i;
        end

      default:
        begin
          nxt_slv_state    = 3'bXXX;
          slv_state_enable = 1'bX;
        end

    endcase


  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      slv_state <= `DAP_SLV_IDLE;
    else if (slv_state_we)
      slv_state <= nxt_slv_state;

  assign slv_state_we = slv_state_enable & opt_mst_en;


  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      slv_idle <= 1'b1;
    else if (slv_idle_we)
      slv_idle <= nxt_slv_idle;

  assign slv_idle_we  = dap_clken_i & opt_mst_en;
  assign nxt_slv_idle = slv_state_enable ? (nxt_slv_state == `DAP_SLV_IDLE) :
                                           (slv_state     == `DAP_SLV_IDLE);

  assign slv_trans_active = (slv_state==`DAP_SLV_AHBSING) |
                            (slv_state==`DAP_SLV_AHB0INC) |
                            (slv_state==`DAP_SLV_AHB1INC) |
                            (slv_state==`DAP_SLV_AHB2INC) |
                            (slv_state==`DAP_SLV_AHB3INC) ;



  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      begin
        dap_csw_reg_hmaster <= 1'b1;
        dap_csw_reg_hprot1  <= 1'b1;
        dap_csw_reg_addrinc <= 2'b00;
        dap_csw_reg_size    <= 2'b00;
      end
    else if (dap_csw_reg_we)
      begin
        dap_csw_reg_hmaster <= nxt_dap_csw_reg_hmaster;
        dap_csw_reg_hprot1  <= dap_wdata_i[25];
        dap_csw_reg_addrinc <= nxt_dap_csw_reg_addrinc;
        dap_csw_reg_size    <= nxt_dap_csw_reg_size;
      end

  assign dap_csw_reg_we          = dap_csw_wr & opt_mst_en;
  assign nxt_dap_csw_reg_hmaster = master_enable_i ? dap_wdata_i[29] : dap_csw_reg_hmaster;
  assign nxt_dap_csw_reg_addrinc = &dap_wdata_i[5:4] ? 2'b10 : dap_wdata_i[5:4];
  assign nxt_dap_csw_reg_size    = &dap_wdata_i[1:0] ? 2'b10 : dap_wdata_i[1:0];

  assign dap_csw_reg = {1'b0,                   // [31]    Reserved
                        1'b0,                   // [30]    Reserved
                        opt_slv_hmaster,        // [29]    HMASTER
                        3'b000,                 // [28:26] HPROT[4:2]
                        opt_slv_hprot1,         // [25]    Hprot1
                        1'b1,                   // [24]    Reserved
                        12'b000000000000,       // [23:12] Reserved
                        4'b0000,                // [11:8]  Mode
                        slv_trans_active,       // [7]     Transfer in progress
                        dap_en_reg,             // [6]     DbgStatus bit
                        dap_csw_reg_addrinc,    // [5:4]   AddrInc
                        1'b0,                   // [3]     Reserved
                        1'b0,dap_csw_reg_size}; // [2:0]   Size

  assign slv_hprot1  = dap_csw_reg_hprot1;
  assign slv_hmaster = dap_csw_reg_hmaster;



  always @ (posedge dap_clk or negedge full_dap_reset_n)
    if (!full_dap_reset_n)
      dap_tra_reg <= {32{1'b0}};
    else if (dap_tra_reg_we)
      dap_tra_reg <= nxt_dap_tra_reg;

  assign nxt_dap_tra_reg[31:12] = dap_tra_wr ? dap_wdata_i[31:12] : dap_tra_reg[31:12];
  assign nxt_dap_tra_reg[11:0]  = dap_tra_wr ? dap_wdata_i[11:0] :
                                               (dap_tra_reg[11:0] + {{9{1'b0}},addr_inc_size});

  assign dap_tra_reg_we = (dap_tra_wr | dap_tra_autoinc) & opt_mst_en;

  assign dap_tra_autoinc = dap_clken_i & (
                           (((slv_state == `DAP_SLV_AHB0INC) |
                             (slv_state == `DAP_SLV_AHB1INC) |
                             (slv_state == `DAP_SLV_AHB2INC) |
                             (slv_state == `DAP_SLV_AHB3INC))  &
                            (slv_state_enable & (nxt_slv_state != slv_state)) &
                            (dap_csw_reg_addrinc == `DAP_INCPACKED)) |
                           ((slv_state == `DAP_SLV_AHB0INC) &
                            (slv_state_enable & (nxt_slv_state == `DAP_SLV_READY)) &
                            (dap_csw_reg_addrinc == `DAP_INCSINGLE)));

  always @ (dap_csw_reg_size)
    case (dap_csw_reg_size)
      `DAP_SZ_BYTE: addr_inc_size = 3'b001;
      `DAP_SZ_HALF: addr_inc_size = 3'b010;
      `DAP_SZ_WORD: addr_inc_size = 3'b100;
      `DAP_SZ_DWRD: addr_inc_size = 3'b100;
      default:      addr_inc_size = 3'bXXX;
    endcase


  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      slv_tra_update <= 1'b0;
    else if (dap_tra_reg_we)
      slv_tra_update <= nxt_slv_tra_update;

  assign nxt_slv_tra_update = ~slv_tra_update;



  always @ (nxt_slv_state or dap_abort_i)
    case (nxt_slv_state)
      `DAP_SLV_IDLE, `DAP_SLV_REG, `DAP_SLV_READY:
        nxt_dap_ready = 1'b1;

      `DAP_SLV_AHBSING, `DAP_SLV_AHB0INC, `DAP_SLV_AHB1INC, `DAP_SLV_AHB2INC,
      `DAP_SLV_AHB3INC:
        nxt_dap_ready = dap_abort_i;

      default:
        nxt_dap_ready = 1'bX;
    endcase

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      begin
        dap_ready     <= 1'b1;
        dap_abort_reg <= 1'b0;
      end
    else if (dap_ready_we)
      begin
        dap_ready     <= nxt_dap_ready;
        dap_abort_reg <= dap_abort_i;
      end

  assign dap_ready_we = slv_state_enable & opt_mst_en;

  assign dap_slverr = (slv_state == `DAP_SLV_READY) &
                      (mst_resp_err_async_i | dap_abort_reg | ~dap_en_reg | trans_abort_reg);


  always @ (dap_addr_reg or dap_csw_reg or dap_tra_reg or mst_hrdata_async_i)
    case (dap_addr_reg[7:2])
      `DAP_CSWADDR: dap_rdata = dap_csw_reg;        // Control/Status Word
      `DAP_TRAADDR: dap_rdata = dap_tra_reg;        // Transfer Address
      `DAP_DRWADDR: dap_rdata = mst_hrdata_async_i; // Data read
      `DAP_BD0ADDR: dap_rdata = mst_hrdata_async_i; // Banked data 0
      `DAP_BD1ADDR: dap_rdata = mst_hrdata_async_i; // Banked data 1
      `DAP_BD2ADDR: dap_rdata = mst_hrdata_async_i; // Banked data 2
      `DAP_BD3ADDR: dap_rdata = mst_hrdata_async_i; // Banked data 3
      `DAP_ROMADDR: dap_rdata = `DAP_ROMVALUE;      // ROM Address
      `DAP_IDRADDR: dap_rdata = `DAP_IDRVALUE;      // ID Register
      default:      dap_rdata = {32{1'b0}};         // Default RAZ
    endcase



  assign dap_ahb_habort_set = dap_abort_i & slv_trans_active;
  assign dap_ahb_habort_clr = end_trans_pulse | req_trans_pulse;

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      dap_ahb_habort <= 1'b0;
    else if (dap_ahb_habort_we)
      dap_ahb_habort <= nxt_dap_ahb_habort;

  assign dap_ahb_habort_we  = dap_clken_i &
                              opt_mst_en  &
                              (dap_ahb_habort_set |  dap_ahb_habort_clr);

  assign nxt_dap_ahb_habort = dap_ahb_habort_set & ~dap_ahb_habort_clr;


  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      trans_abort_reg <= 1'b0;
    else if (trans_abort_reg_we)
      trans_abort_reg <= nxt_trans_abort_reg;

  assign trans_abort_reg_we  = dap_clken_i & opt_mst_en;
  assign nxt_trans_abort_reg = ~dap_ahb_habort_clr &
                               (trans_abort_reg | opt_dap_ahb_habort);



  assign req_trans_pulse = (nxt_slv_state != `DAP_SLV_IDLE)  &
                           (nxt_slv_state != `DAP_SLV_READY) &
                           (nxt_slv_state != `DAP_SLV_REG)   &
                           (nxt_slv_state != slv_state)      &
                           slv_state_enable;

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      slv_req_trans <= 1'b0;
    else if (slv_req_trans_we)
      slv_req_trans <= nxt_slv_req_trans;

  assign slv_req_trans_we  = req_trans_pulse & opt_mst_en;
  assign nxt_slv_req_trans = ~slv_req_trans_rtn_i;  // Toggle slv_req_trans



  assign slv_haddr  = ((dap_addr_reg[7:2] == `DAP_BD0ADDR) |
                       (dap_addr_reg[7:2] == `DAP_BD1ADDR) |
                       (dap_addr_reg[7:2] == `DAP_BD2ADDR) |
                       (dap_addr_reg[7:2] == `DAP_BD3ADDR)) ?
                      {dap_tra_reg[31:4], dap_addr_reg[3:0]} : dap_tra_reg[31:0];

  assign slv_hsize  = ((dap_addr_i[7:2] == `DAP_BD0ADDR) |
                       (dap_addr_i[7:2] == `DAP_BD1ADDR) |
                       (dap_addr_i[7:2] == `DAP_BD2ADDR) |
                       (dap_addr_i[7:2] == `DAP_BD3ADDR)) ?
                      `DAP_SZ_WORD : dap_csw_reg_size;

  assign slv_hwrite = dap_write_i;




  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      end_trans_reg <= 1'b0;
    else if (end_trans_reg_we)
      end_trans_reg <= mst_end_trans_i;

  assign end_trans_reg_we = dap_clken_i & opt_mst_en;

  assign end_trans_pulse = mst_end_trans_i ^ end_trans_reg;



  assign dap_ready_o      = opt_dap_ready;
  assign dap_slverr_o     = opt_dap_slverr;
  assign dap_rdata_o      = opt_dap_rdata;

  assign dap_ahb_habort_o = opt_dap_ahb_habort;

  assign slv_haddr_o      = opt_slv_haddr;
  assign slv_hwrite_o     = opt_slv_hwrite;
  assign slv_hsize_o      = opt_slv_hsize;
  assign slv_hprot1_o     = opt_slv_hprot1;
  assign slv_hmaster_o    = opt_slv_hmaster;

  assign slv_idle_o       = opt_slv_idle;
  assign slv_req_trans_o  = opt_slv_req_trans;
  assign slv_tra_update_o = slv_tra_update;



`ifdef ARM_ASSERT_ON

  reg  [8*20:0] ovl_slv_state_mnemonic;

  always @ (slv_state)
    case (slv_state)
      `DAP_SLV_IDLE:    ovl_slv_state_mnemonic = {"IDLE"};
      `DAP_SLV_READY:   ovl_slv_state_mnemonic = {"READY"};
      `DAP_SLV_REG:     ovl_slv_state_mnemonic = {"REG"};
      `DAP_SLV_AHBSING: ovl_slv_state_mnemonic = {"SING"};
      `DAP_SLV_AHB0INC: ovl_slv_state_mnemonic = {"INC0"};
      `DAP_SLV_AHB1INC: ovl_slv_state_mnemonic = {"INC1"};
      `DAP_SLV_AHB2INC: ovl_slv_state_mnemonic = {"INC2"};
      `DAP_SLV_AHB3INC: ovl_slv_state_mnemonic = {"INC3"};
      default:          ovl_slv_state_mnemonic = {"UNKNOWN"};
    endcase

  reg ovl_csw_reg_chk;  // Flag that the CSW register has been accessed
  reg ovl_tra_reg_chk;  // Flag that the TRA register has been accessed

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      ovl_csw_reg_chk <= 1'b0;
    else if (dap_csw_wr)
      ovl_csw_reg_chk <= 1'b1;

  always @ (posedge dap_clk or negedge dap_reset_n)
    if (!dap_reset_n)
      ovl_tra_reg_chk <= 1'b0;
    else if (dap_tra_wr)
      ovl_tra_reg_chk <= 1'b1;

  assert_never #(2,1,"csw_reg should be programmed before accessing Slave AHB.")
    ovl_apslv_cswreg(dap_clk, dap_reset_n, (slv_req_trans & ~ovl_csw_reg_chk));

  assert_never #(2,1,"dap_tra_reg should be programmed before accessing Slave AHB.")
    ovl_apslv_trareg(dap_clk, dap_reset_n, (slv_req_trans & ~ovl_tra_reg_chk));

  assert_zero_one_hot #(0,2,0,"Register enable must be exclusive.")
    ovl_apslv_enable(dap_clk, dap_reset_n, {dap_csw_wr, dap_tra_wr});

  assert_zero_one_hot #(0,3,0,"Register access must be exclusive.")
    ovl_apslv_access(dap_clk, dap_reset_n,
      {dap_reg_access, ahb_bd_access, ahb_drw_access});

`endif

endmodule

`include "cm3_dap_ahb_ap_undefs.vp"
