module cm3_dpu_alu_bshift
  (// Inputs
   instr_immed_valid_ex_i, shift_ctl_ex_i, shift_in1_ex_i, shift_in2_ex_i,
   psr_ex_i,
   shift_res_ex_o, shift_cout_ex_o
   );

  input          instr_immed_valid_ex_i;   // Immediate valid
  input [1:0]    shift_ctl_ex_i;           // Shifter control
  input [31:0]   shift_in1_ex_i;           // Shifter input 1
  input [7:0]    shift_in2_ex_i;           // Shifter input 2
  input [31:0]   psr_ex_i;                 // PSR

  output [31:0]  shift_res_ex_o;           // Shifter result
  output         shift_cout_ex_o;          // Shifter result carry out


  wire           shift_lsr_ex;             // Shift type is LSR
  wire           shift_lsl_ex;             // Shift type is LSL
  wire           shift_ror_ex;             // Shift type is ROR
  wire           shift_asr_ex;             // Shift type is ASR
  wire [1:0]     ror_value_ex;             // ROR amount value
  wire [31:0]    a_rol_0_ex;               // A (shift_in1) ROL by 0
  wire [31:0]    a_rol_8_ex;               // A (shift_in1) ROL by 8
  wire [31:0]    a_rol_16_ex;              // A (shift_in1) ROL by 16
  wire [31:0]    a_rol_24_ex;              // A (shift_in1) ROL by 24
  wire [31:0]    a_rol_res_ex;             // A (shift_in1) ROL result
  wire [31:0]    a_ror_0_ex;               // A (shift_in1) ROR by 0
  wire [31:0]    a_ror_8_ex;               // A (shift_in1) ROR by 8
  wire [31:0]    a_ror_16_ex;              // A (shift_in1) ROR by 16
  wire [31:0]    a_ror_24_ex;              // A (shift_in1) ROR by 24
  wire [31:0]    a_ror_res_ex;             // A (shift_in1) ROR result
  wire [31:0]    b_rol_0_ex;               // B ROL by 0
  wire [31:0]    b_rol_4_ex;               // B ROL by 4
  wire [31:0]    b_ror_0_ex;               // B ROR by 0
  wire [31:0]    b_ror_4_ex;               // B ROR by 4
  wire [31:0]    b_rot_res_ex;             // B rotated result (ROR or ROL)
  wire [31:0]    c_ror_0_ex;               // C ROR by 0
  wire [31:0]    c_ror_1_ex;               // C ROR by 1
  wire [31:0]    c_ror_2_ex;               // C ROR by 2
  wire [31:0]    c_ror_3_ex;               // C ROR by 3
  wire [31:0]    c_ror_res_ex;             // C ROR result
  wire           in2_ne_0_ex;              // shift_in2_ex != 0
  wire           in2_ge_32_ex;             // shift_in2_ex >= 32
  wire           in2_ge_64_ex;             // shift_in2_ex >= 64
  wire           shift_gt_32_ex;           // Shift value  >  32
  wire           shift_ge_32_ex;           // Shift value  >= 32
  wire           imm_ge_32_ex;             // Immediate    >= 32
  wire           mask_force_1_ex;          // Force shift mask value to 1's
  wire           mask_invert_ex;           // Invert shift mask value
  wire           mask_a_ex;                // Value to use for shifted mask A
  wire [6:0]     mask_a_lsr_0_ex;          // Mask A - LSR 0
  wire [6:0]     mask_a_lsr_2_ex;          // Mask A - LSR 2
  wire [6:0]     mask_a_lsr_4_ex;          // Mask A - LSR 4
  wire [6:0]     mask_a_lsr_6_ex;          // Mask A - LSR 6
  wire [6:0]     mask_b_ex;                // Value to use for shifted mask B
  wire [30:0]    mask_b_lsr_0_ex;          // Mask B - LSR 0
  wire [30:0]    mask_b_lsr_8_ex;          // Mask B - LSR 8
  wire [30:0]    mask_b_lsr_16_ex;         // Mask B - LSR 16
  wire [30:0]    mask_b_lsr_24_ex;         // Mask B - LSR 24
  wire [30:0]    mask_c_ex;                // Mask C - mask B LSR result
  wire [30:0]    mask_c_rev_ex;            // Mask C reversed
  wire [31:0]    mask_d_ex;                // Mask D based on mask C selection
  wire [31:0]    mask_pos_res_ex;          // Final mask result - +ve version
  wire [31:0]    mask_neg_res_ex;          // Final mask result - -ve version
  wire [31:0]    shift_sign_ex;            // Shift sign value if required
  wire           carry_bypass_ex;          // Ignore carry
  wire [31:0]    shift_res_ex;             // Shifter result
  wire           shift_cout_ex;            // Shifter result carry out

  assign shift_lsl_ex = (shift_ctl_ex_i == 2'b00);
  assign shift_lsr_ex = (shift_ctl_ex_i == 2'b01);
  assign shift_asr_ex = (shift_ctl_ex_i == 2'b10);
  assign shift_ror_ex = (shift_ctl_ex_i == 2'b11);


  assign ror_value_ex = {2{shift_lsl_ex}} ^ shift_in2_ex_i[1:0];





  assign a_rol_0_ex   = shift_in1_ex_i;
  assign a_rol_8_ex   = {shift_in1_ex_i[23:0], shift_in1_ex_i[31:24]};
  assign a_rol_16_ex  = {shift_in1_ex_i[15:0], shift_in1_ex_i[31:16]};
  assign a_rol_24_ex  = {shift_in1_ex_i[ 7:0], shift_in1_ex_i[31: 8]};

  assign a_rol_res_ex = (({32{shift_in2_ex_i[4:3]==2'b00}} & a_rol_0_ex)  |
                         ({32{shift_in2_ex_i[4:3]==2'b01}} & a_rol_8_ex)  |
                         ({32{shift_in2_ex_i[4:3]==2'b10}} & a_rol_16_ex) |
                         ({32{shift_in2_ex_i[4:3]==2'b11}} & a_rol_24_ex));

  assign a_ror_0_ex   = shift_in1_ex_i;
  assign a_ror_8_ex   = {shift_in1_ex_i[ 7:0], shift_in1_ex_i[31: 8]};
  assign a_ror_16_ex  = {shift_in1_ex_i[15:0], shift_in1_ex_i[31:16]};
  assign a_ror_24_ex  = {shift_in1_ex_i[23:0], shift_in1_ex_i[31:24]};

  assign a_ror_res_ex = (({32{shift_in2_ex_i[4:3]==2'b00}} & a_ror_0_ex)  |
                         ({32{shift_in2_ex_i[4:3]==2'b01}} & a_ror_8_ex)  |
                         ({32{shift_in2_ex_i[4:3]==2'b10}} & a_ror_16_ex) |
                         ({32{shift_in2_ex_i[4:3]==2'b11}} & a_ror_24_ex));



  assign b_rol_0_ex = {a_rol_res_ex[28:0],a_rol_res_ex[31:29]};
  assign b_rol_4_ex = {a_rol_res_ex[24:0],a_rol_res_ex[31:25]};
  assign b_ror_0_ex = a_ror_res_ex;
  assign b_ror_4_ex = {a_ror_res_ex[3:0], a_ror_res_ex[31:4]};

  assign b_rot_res_ex = (({32{{shift_lsl_ex,shift_in2_ex_i[2]}==2'b00}} & b_ror_0_ex) |
                         ({32{{shift_lsl_ex,shift_in2_ex_i[2]}==2'b01}} & b_ror_4_ex) |
                         ({32{{shift_lsl_ex,shift_in2_ex_i[2]}==2'b10}} & b_rol_0_ex) |
                         ({32{{shift_lsl_ex,shift_in2_ex_i[2]}==2'b11}} & b_rol_4_ex));

  assign c_ror_0_ex = b_rot_res_ex;
  assign c_ror_1_ex = {b_rot_res_ex[  0], b_rot_res_ex[31:1]};
  assign c_ror_2_ex = {b_rot_res_ex[1:0], b_rot_res_ex[31:2]};
  assign c_ror_3_ex = {b_rot_res_ex[2:0], b_rot_res_ex[31:3]};

  assign c_ror_res_ex = (({32{ror_value_ex[1:0]==2'b00}} & c_ror_0_ex) |
                         ({32{ror_value_ex[1:0]==2'b01}} & c_ror_1_ex) |
                         ({32{ror_value_ex[1:0]==2'b10}} & c_ror_2_ex) |
                         ({32{ror_value_ex[1:0]==2'b11}} & c_ror_3_ex));


  assign in2_ne_0_ex    = (|shift_in2_ex_i[4:0]);
  assign in2_ge_32_ex   = ~instr_immed_valid_ex_i & (|(shift_in2_ex_i[7:5]));
  assign in2_ge_64_ex   = ~instr_immed_valid_ex_i & (|(shift_in2_ex_i[7:6]));
  assign shift_gt_32_ex = (~shift_ror_ex &
                           (in2_ge_64_ex | (in2_ge_32_ex & in2_ne_0_ex)));
  assign shift_ge_32_ex = ((~shift_ror_ex & in2_ge_32_ex) |
                           ((shift_lsr_ex | shift_asr_ex) &
                            instr_immed_valid_ex_i & ~in2_ne_0_ex));

  assign imm_ge_32_ex    = (instr_immed_valid_ex_i & (|shift_in2_ex_i[7:5]) &
                            shift_lsl_ex);
  assign mask_force_1_ex = shift_ge_32_ex | shift_ror_ex | imm_ge_32_ex;
  assign mask_invert_ex  = (shift_ge_32_ex | (shift_lsl_ex & in2_ne_0_ex) |
                            imm_ge_32_ex);


  assign mask_a_ex = ~shift_in2_ex_i[0];

  assign mask_a_lsr_0_ex = {            mask_a_ex, {6{1'b1}} };
  assign mask_a_lsr_2_ex = { {2{1'b0}}, mask_a_ex, {4{1'b1}} };
  assign mask_a_lsr_4_ex = { {4{1'b0}}, mask_a_ex, {2{1'b1}} };
  assign mask_a_lsr_6_ex = { {6{1'b0}}, mask_a_ex            };

  assign mask_b_ex = (({7{shift_in2_ex_i[2:1]==2'b00}} & mask_a_lsr_0_ex) |
                      ({7{shift_in2_ex_i[2:1]==2'b01}} & mask_a_lsr_2_ex) |
                      ({7{shift_in2_ex_i[2:1]==2'b10}} & mask_a_lsr_4_ex) |
                      ({7{shift_in2_ex_i[2:1]==2'b11}} & mask_a_lsr_6_ex));

  assign mask_b_lsr_0_ex  = {             mask_b_ex, {24{1'b1}} };
  assign mask_b_lsr_8_ex  = { { 8{1'b0}}, mask_b_ex, {16{1'b1}} };
  assign mask_b_lsr_16_ex = { {16{1'b0}}, mask_b_ex, { 8{1'b1}} };
  assign mask_b_lsr_24_ex = { {24{1'b0}}, mask_b_ex             };

  assign mask_c_ex = (({31{shift_in2_ex_i[4:3]==2'b00}} & mask_b_lsr_0_ex)  |
                      ({31{shift_in2_ex_i[4:3]==2'b01}} & mask_b_lsr_8_ex)  |
                      ({31{shift_in2_ex_i[4:3]==2'b10}} & mask_b_lsr_16_ex) |
                      ({31{shift_in2_ex_i[4:3]==2'b11}} & mask_b_lsr_24_ex));

  assign mask_c_rev_ex = ~ { mask_c_ex[ 0], mask_c_ex[ 1], mask_c_ex[ 2],
                             mask_c_ex[ 3], mask_c_ex[ 4], mask_c_ex[ 5],
                             mask_c_ex[ 6], mask_c_ex[ 7], mask_c_ex[ 8],
                             mask_c_ex[ 9], mask_c_ex[10], mask_c_ex[11],
                             mask_c_ex[12], mask_c_ex[13], mask_c_ex[14],
                             mask_c_ex[15], mask_c_ex[16], mask_c_ex[17],
                             mask_c_ex[18], mask_c_ex[19], mask_c_ex[20],
                             mask_c_ex[21], mask_c_ex[22], mask_c_ex[23],
                             mask_c_ex[24], mask_c_ex[25], mask_c_ex[26],
                             mask_c_ex[27], mask_c_ex[28], mask_c_ex[29],
                             mask_c_ex[30] };

  assign mask_d_ex = ((shift_lsl_ex & ~mask_c_ex[30]) ? {1'b0, mask_c_rev_ex} :
                                                        {mask_c_ex, 1'b1});


  assign mask_pos_res_ex = ((mask_d_ex | {32{mask_force_1_ex}}) ^
                            {32{ mask_invert_ex}});
  assign mask_neg_res_ex = ((mask_d_ex | {32{mask_force_1_ex}}) ^
                            {32{~mask_invert_ex}});

  assign shift_sign_ex   = {32{shift_asr_ex & shift_in1_ex_i[31]}};
  assign shift_res_ex    = ((mask_pos_res_ex & c_ror_res_ex) |
                            (mask_neg_res_ex & shift_sign_ex));


  assign carry_bypass_ex = ~(shift_ge_32_ex | in2_ge_32_ex | in2_ne_0_ex);

  assign shift_cout_ex   = (shift_gt_32_ex ?
                            shift_sign_ex[0] :
                            (carry_bypass_ex ?
                             psr_ex_i[29] : (shift_lsl_ex ? c_ror_res_ex[0] :
                                                            c_ror_res_ex[31])));


  assign shift_res_ex_o  = shift_res_ex;
  assign shift_cout_ex_o = shift_cout_ex;

endmodule
