`include "cm3_dpu_defs.vp"

module cm3_dpu_alu_dp
  (// Inputs
   hreset_n, hclk, aclk, alu_cg_en_i, instr_immed_ex_i, instr_immed_valid_ex_i,
   instr_algn_ex_i, instr_sign_ex_i, instr_data_size_ex_i, instr_alu_ctl_ex_i,
   alu_mcyc_state_ex_i, mul_input_sel_ex_i, mul_neg_in2_ex_i,
   mul_add_valid_ex_i, mul_mla_track_ex_i, mul_cin_ex_i, mul_mull_extend_ex_i,
   div_fast_zero_ex_i, div_srt_init_ex_i, div_srt_sign_ex_i, div_cnt_ex_i,
   div_cnt_reg_ex_i, shift_in2_fast_ex_i, shift_l_mask_ex_i, shift_r_mask_ex_i,
   shift_lr_mask_ex_i, shift_sel_ex_i, rf_rd_a_data_ex_i, rf_rd_b_data_ex_i,
   e_simd_res_ex_i, psr_ex_i, privileged_i, process_sp_ex_i, user_thread_ex_i,
   fpu_active_ex_i, int_isr_ex_i, nvic_basepri_i, nvic_fmask_i, nvic_pmask_i,
   shift_in1_ex_o, shift_res_ex_o, clz_res_ex_o, alu_cout_ex_o,
   alu_flag_res_ex_o, alu_ge_res_ex_o, alu_mul_flag_res_ex_o,
   alu_rf_wr_d_data_ex_o
   );

  parameter      RESET_ALL_REGS = 0;       // Reset all registers if 1
  parameter      EXT_PRESENT    = 0;       // Extensions present if 1


  input          hreset_n;
  input          hclk;
  input          aclk;                     // ALU multi-cycle operation clock
  input          alu_cg_en_i;              // ALU clock gate enable
  input [31:0]   instr_immed_ex_i;         // Immediate value
  input          instr_immed_valid_ex_i;   // Immediate valid
  input          instr_algn_ex_i;          // Align data
  input          instr_sign_ex_i;          // Signed operation or sign extend
  input [1:0]    instr_data_size_ex_i;     // Data size
  input [20:0]   instr_alu_ctl_ex_i;       // ALU unit control
  input [3:0]    alu_mcyc_state_ex_i;      // ALU multi-cycle operation state
  input [5:0]    mul_input_sel_ex_i;       // Multiplier input selects
  input          mul_neg_in2_ex_i;         // Invert multiplier input 2
  input          mul_add_valid_ex_i;       // Multiply result input to adder
  input          mul_mla_track_ex_i;       // MLA instr in progress
  input          mul_cin_ex_i;             // Adder carry in for multiply
  input          mul_mull_extend_ex_i;     // Extend *MULL operation
  input          div_fast_zero_ex_i;       // Early zero detection for DIV
  input [1:0]    div_srt_init_ex_i;        // DIV SRT unit intialisation value
  input          div_srt_sign_ex_i;        // DIV SRT unit sign value
  input [5:0]    div_cnt_ex_i;             // DIV operation cycle count
  input [1:0]    div_cnt_reg_ex_i;         // Registered div_cnt_ex
  input [7:0]    shift_in2_fast_ex_i;      // Shifter input 2 (fast version)
  input [31:0]   shift_l_mask_ex_i;        // Shifter mask (left shift only)
  input [31:0]   shift_r_mask_ex_i;        // Shifter mask (right shift only)
  input [31:0]   shift_lr_mask_ex_i;       // Shifter mask (left & right shift)
  input [6:0]    shift_sel_ex_i;           // Shifter result select
  input [31:0]   rf_rd_a_data_ex_i;        // RF A port read data
  input [31:0]   rf_rd_b_data_ex_i;        // RF B port read data
  input [31:0]   e_simd_res_ex_i;          // SIMD unit result
  input [31:0]   psr_ex_i;                 // PSR
  input          privileged_i;             // Privileged mode (vs user)
  input [1:0]    process_sp_ex_i;          // Current SP is process (vs main)
  input          user_thread_ex_i;         // Thread is (or will be) user mode
  input          fpu_active_ex_i;          // FPU active in current context
  input [8:0]    int_isr_ex_i;             // Current ISR number
  input [7:0]    nvic_basepri_i;           // BASEPRI value
  input          nvic_fmask_i;             // FAULTMASK value
  input          nvic_pmask_i;             // PRIMASK value

  output [31:0]  shift_in1_ex_o;           // ALU shifter input 1
  output [31:0]  shift_res_ex_o;           // ALU shifter result
  output [5:0]   clz_res_ex_o;             // ALU count leading zeros result
  output         alu_cout_ex_o;            // ALU carry out
  output [4:0]   alu_flag_res_ex_o;        // ALU flag result
  output [3:0]   alu_ge_res_ex_o;          // ALU GE bits result
  output [4:0]   alu_mul_flag_res_ex_o;    // ALU MUL flag result (critical)
  output [31:0]  alu_rf_wr_d_data_ex_o;    // ALU RF D port write data


  wire           opt_ext_en;               // Master enable for extensions
  wire           bwise_and_ex;             // Bitwise AND
  wire           bwise_eor_ex;             // Bitwise EOR
  wire           bwise_orr_ex;             // Bitwise OR
  wire           bwise_valid_ex;           // Bitwise operation valid
  wire           alu_rev_ex;               // REV
  wire           alu_rbit_ex;              // RBIT
  wire           alu_movt_ex;              // MOVT
  wire           alu_mov_ex;               // MOV
  wire           alu_mov_rev_ex;           // MOV or REV
  wire           alu_clz_ex;               // Count leading zeros
  wire           alu_msr_flag_ex;          // MSR flag write
  wire           alu_mrs_ctrl_ex;          // MRS control bits read
  wire           alu_mrs_base_ex;          // MRS BASEPRI read
  wire           alu_mrs_pri_ex;           // MRS PRIMASK read
  wire           alu_mrs_fault_ex;         // MRS FAULTMASK read
  wire           alu_mrs_psr_ex;           // MRS flag/intnum read operation
  wire           alu_mrs_flag_ex;          // MRS flag read
  wire           alu_mrs_intnum_ex;        // MRS current interrupt number read
  wire           alu_muldiv_ex;            // Multiply or divide
  wire           alu_ml_ex;                // MLA/MLS multiply
  wire           alu_mlal_ex;              // SMLAL/UMLAL multiply
  wire           alu_smlal_ex;             // SMLAL multiply
  wire           alu_mul_ex;               // MUL multiply
  wire           alu_mull_ex;              // SMULL/UMULL multiply
  wire           alu_smull_ex;             // SMULL multiply
  wire           alu_div_ex;               // SDIV/UDIV divide
  wire           alu_sdiv_ex;              // SDIV divide
  wire           alu_shift_ex;             // Shifter required
  wire           alu_adder_ex;             // ALU adder operation
  wire           alu_arith_ex;             // ALU arithmetic op
  wire [31:0]    mov_in_ex;                // MOV input
  wire [3:0]     mov_ctrl_ex;              // MOV control
  wire [31:0]    clz_in_ex;                // Count leading zeros input
  wire           clz_ctrl_ex;              // CLZ control (count 0's vs 1's)
  wire [31:0]    clz_data_ex;              // CLZ data to count
  wire [31:0]    rbit_in_ex;               // RBIT input
  wire [31:0]    mul_in1_ex;               // 32 bit multiplier input 1
  wire [31:0]    mul_in2_ex;               // 32 bit multiplier input 2
  wire           mcyc_a_signed_ex;         // Signed SMULL/SMLAL required
  wire           mcyc_b_signed_ex;         // Signed SMULL/SMLAL required
  reg  [31:0]    mcyc_alu_in1_ex;          // Multi-cycle operation ALU input 1
  reg  [31:0]    mcyc_alu_in2_ex;          // Multi-cycle operation ALU input 2
  wire           alu_inv_in1_ex;           // Invert ALU input 1
  wire           alu_inv_in2_ex;           // Invert ALU input 2
  wire [31:0]    alu_in1_ex;               // Generic ALU input 1
  wire [31:0]    alu_in2_ex;               // Generic ALU input 2
  wire           alu_cin_valid_ex;         // ALU carry in valid
  wire           alu_cin_ex;               // ALU carry in
  wire [31:0]    adder_in1_ex;             // Adder input 1
  wire [31:0]    adder_in2_ex;             // Adder input 2
  wire           adder_cin_ex;             // Adder carry in
  wire           alu_add_in1_valid_ex;     // ALU operation adder input 1 valid
  wire           alu_add_in2_valid_ex;     // ALU operation adder input 2 valid
  wire           adder_in1_valid_ex;       // Adder input 1 valid
  wire           adder_in2_valid_ex;       // Adder input 2 valid
  wire           adder_cin_valid_ex;       // Carry in valid for adder operation
  wire [31:0]    div_srt_in1_ex;           // DIV SRT unit input 1
  wire [31:0]    div_srt_in2_ex;           // DIV SRT unit input 2
  wire           shift1_sel_b_ex;          // Shifter input 1 select B port data
  wire           shift2_sel_a_ex;          // Shifter input 2 select A port data
  wire           shift2_sel_i_ex;          // Shifter input 2 select immediate
  wire [31:0]    shift_in1_ex;             // Shifter input 1
  wire [7:0]     shift_in2_ex;             // Shifter input 2
  wire [1:0]     shift_ctl_ex;             // Shifter control
  wire [1:0]     shift_inlinec_ex;         // Shifter has in-line carry
  wire           shift_rrx_valid_ex;       // Shifter RRX result valid
  wire [31:0]    imm_shift_mux_ex;         // Select shifter result or immediate
  wire           top_adder_in1_valid_ex;   // Top 16 of 48 bit adder in 1 valid
  wire           top_adder_in2_valid_ex;   // Top 16 of 48 bit adder in 2 valid
  wire [15:0]    top_adder_in1_ex;         // Top 16 of 48 bit adder in 1
  wire [15:0]    top_adder_in2_ex;         // Top 16 of 48 bit adder in 2
  wire           top_adder_cin_ex;         // Top 16 of 48 bit adder carry in
  wire [15:0]    top_adder_res_ex;         // Top 16 of 48 bit adder result
  wire [47:0]    full_adder_res_ex;        // Full 48 bit adder result
  wire [3:0]     nxt_mcyc_temp_low_sel_ex; // nxt_mcyc_temp_low_ex mux selector
  wire [31:0]    nxt_mcyc_temp_low_ex;     // Next mcyc_temp_low_ex
  wire           mcyc_temp_low_we_ex;      // mcyc_temp_low_ex write enable
  reg  [31:0]    mcyc_temp_low_ex;         // Lower word of multi-cycle storage
  wire [2:0]     nxt_mcyc_temp_upp_sel_ex; // nxt_mcyc_temp_upp_ex mux selector
  wire [31:0]    nxt_mcyc_temp_upp_ex;     // Next mcyc_temp_upp_ex
  wire           mcyc_temp_upp_we_ex;      // mcyc_temp_upp_ex write enable
  reg  [31:0]    mcyc_temp_upp_ex;         // upper word of multi-cycle storage
  wire [63:0]    mcyc_temp_ex;             // Combined multi-cycle storage
  wire [34:0]    nxt_mcyc_temp2_ex;        // Next mcyc_temp2_ex
  wire           mcyc_temp2_we_ex;         // mcyc_temp2_ex write enable
  reg  [34:0]    mcyc_temp2_ex;            // Multi-cycle secondary storage
  wire [34:0]    mcyc_temp2_shift_ex;      // Shifter version of mcyc_temp2_ex
  wire [31:0]    div_srt_rem_ex;           // DIV SRT unit remainder
  wire [3:0]     div_srt_res_ex;           // DIV SRT unit result
  wire [31:0]    bwise_res_ex;             // Bitwise operation result
  wire           bwise_zero_ex;            // Bitwise result is zero
  wire           clz_sel_b3_1_ex;          // CLZ byte 3 result select 1
  wire           clz_sel_b3_2_ex;          // CLZ byte 3 result select 2
  wire           clz_sel_b3_3_ex;          // CLZ byte 3 result select 3
  wire           clz_sel_b3_4_ex;          // CLZ byte 3 result select 4
  wire           clz_sel_b3_5_ex;          // CLZ byte 3 result select 5
  wire           clz_sel_b3_6_ex;          // CLZ byte 3 result select 6
  wire           clz_sel_b3_7_ex;          // CLZ byte 3 result select 7
  wire [2:0]     clz_b3_res_ex;            // CLZ result for byte 3
  wire           clz_sel_b2_1_ex;          // CLZ byte 2 result select 1
  wire           clz_sel_b2_2_ex;          // CLZ byte 2 result select 2
  wire           clz_sel_b2_3_ex;          // CLZ byte 2 result select 3
  wire           clz_sel_b2_4_ex;          // CLZ byte 2 result select 4
  wire           clz_sel_b2_5_ex;          // CLZ byte 2 result select 5
  wire           clz_sel_b2_6_ex;          // CLZ byte 2 result select 6
  wire           clz_sel_b2_7_ex;          // CLZ byte 2 result select 7
  wire [2:0]     clz_b2_res_ex;            // CLZ result for byte 2
  wire           clz_sel_b1_1_ex;          // CLZ byte 1 result select 1
  wire           clz_sel_b1_2_ex;          // CLZ byte 1 result select 2
  wire           clz_sel_b1_3_ex;          // CLZ byte 1 result select 3
  wire           clz_sel_b1_4_ex;          // CLZ byte 1 result select 4
  wire           clz_sel_b1_5_ex;          // CLZ byte 1 result select 5
  wire           clz_sel_b1_6_ex;          // CLZ byte 1 result select 6
  wire           clz_sel_b1_7_ex;          // CLZ byte 1 result select 7
  wire [2:0]     clz_b1_res_ex;            // CLZ result for byte 1
  wire           clz_sel_b0_1_ex;          // CLZ byte 0 result select 1
  wire           clz_sel_b0_2_ex;          // CLZ byte 0 result select 2
  wire           clz_sel_b0_3_ex;          // CLZ byte 0 result select 3
  wire           clz_sel_b0_4_ex;          // CLZ byte 0 result select 4
  wire           clz_sel_b0_5_ex;          // CLZ byte 0 result select 5
  wire           clz_sel_b0_6_ex;          // CLZ byte 0 result select 6
  wire           clz_sel_b0_7_ex;          // CLZ byte 0 result select 7
  wire [2:0]     clz_b0_res_ex;            // CLZ result for byte 0
  wire           clz_sel_b3_ex;            // CLZ select byte 3 CLZ result
  wire           clz_sel_b2_ex;            // CLZ select byte 2 CLZ result
  wire           clz_sel_b1_ex;            // CLZ select byte 1 CLZ result
  wire           clz_sel_b0_ex;            // CLZ select byte 0 CLZ result
  wire           clz_sel_w0_ex;            // CLZ select word is 0 CLZ result
  wire [5:0]     clz_res_ex;               // Count leading zeros result
  wire [31:0]    alu_clz_res_ex;           // ALU count leading zeros result
  wire [31:0]    mrs_psr_res_ex;           // MRS PSR result
  wire [31:0]    mrs_ctrl_res_ex;          // MRS control result
  reg  [31:0]    mov_res_ex;               // MOV result
  wire [31:0]    movt_res_ex;              // MOVT result
  wire [31:0]    mul_res_ex;               // Multiplier result
  wire [31:0]    rbit_res_ex;              // RBIT result
  wire [3:0]     arith_sel_ex;             // Arithmetic result mux select
  wire [31:0]    arith_res_ex;             // Arithmetic result
  wire [31:0]    shift_res_ex;             // Shifter result
  wire [31:0]    shift_rrx_res_ex;         // Shifter result modified for RRX
  wire [31:0]    bf_sat_res_ex;            // Bitfield/saturation result
  wire [32:0]    adder_res_ex;             // 32 bit adder result
  wire [31:0]    adder_cs_sout_ex;         // Adder carry-save sum out
  wire [30:0]    adder_cs_cout_ex;         // Adder carry-save carry out
  wire           adder_zero_ex;            // Adder result is zero
  wire [31:0]    alu_res_ex;               // ALU result (adder or bwise)
  wire           shift_cout_ex;            // Shifter result carry out
  wire           alu_cout_ex;              // ALU carry out
  wire           alu_oflow_ex;             // ALU overflow
  wire [4:0]     alu_nzcvq_erly_ex;        // ALU NZCVQ result (early, excl MSR)
  wire [4:0]     alu_nzcvq_ex;             // ALU NZCVQ result (excl MSR)
  wire [4:0]     alu_flag_res_ex;          // ALU flag result (incl MSR)
  wire [3:0]     alu_ge_res_ex;            // ALU GE bits result
  wire [4:0]     alu_mul_flag_res_ex;      // ALU MUL flag result (critical)
  wire [31:0]    alu_rf_wr_d_data_ex;      // ALU RF D port write data
  wire           full_hreset_n;            // Apply hreset_n to all registers

  `ifdef ARM_DSM

  wire           dsm_reset_all_regs;

  assign full_hreset_n = (dsm_reset_all_regs)  ? hreset_n : 1'b1;

`else

  assign full_hreset_n = (RESET_ALL_REGS == 1) ? hreset_n : 1'b1;

`endif

  assign opt_ext_en    = (EXT_PRESENT == 1)    ? 1'b1 : 1'b0;



  assign bwise_valid_ex    = (instr_alu_ctl_ex_i[17:15] == 3'b000 &
                              (|instr_alu_ctl_ex_i[14:13]));
  assign bwise_and_ex      = instr_alu_ctl_ex_i[17:13] == 5'b00001;
  assign bwise_eor_ex      = instr_alu_ctl_ex_i[17:13] == 5'b00011;
  assign bwise_orr_ex      = instr_alu_ctl_ex_i[17:13] == 5'b00010;

  assign alu_rev_ex        = (instr_alu_ctl_ex_i[17:13] == 5'b10110);

  assign alu_rbit_ex       = (instr_alu_ctl_ex_i[17:13] == 5'b10101);

  assign alu_movt_ex       = (instr_alu_ctl_ex_i[17:13] == 5'b10111);
  assign alu_mov_ex        = (instr_alu_ctl_ex_i[17:13] == 5'b10100);

  assign alu_mov_rev_ex    = alu_mov_ex | alu_rev_ex;

  assign alu_clz_ex        = (instr_alu_ctl_ex_i[17:13] == 5'b11111);

  assign alu_msr_flag_ex   = (instr_alu_ctl_ex_i[17:13] == 5'b01000);

  assign alu_mrs_ctrl_ex   = (instr_alu_ctl_ex_i[17:13] == 5'b01010);
  assign alu_mrs_base_ex   = ((instr_alu_ctl_ex_i[17:13] == 5'b01011) &
                              privileged_i);
  assign alu_mrs_pri_ex    = ((instr_alu_ctl_ex_i[17:15] == 3'b011) &
                              instr_alu_ctl_ex_i[13] & privileged_i);
  assign alu_mrs_fault_ex  = ((instr_alu_ctl_ex_i[17:14] == 4'b0111) &
                              privileged_i);
  assign alu_mrs_psr_ex    = (instr_alu_ctl_ex_i[17:15] == 3'b100);
  assign alu_mrs_flag_ex   = (alu_mrs_psr_ex & instr_alu_ctl_ex_i[13]);
  assign alu_mrs_intnum_ex = (alu_mrs_psr_ex & instr_alu_ctl_ex_i[14]);

  assign alu_muldiv_ex     = (instr_alu_ctl_ex_i[17:16] == 2'b11 & ~alu_clz_ex);
  assign alu_mlal_ex       = (instr_alu_ctl_ex_i[17:13] == 5'b11001);
  assign alu_smlal_ex      = alu_mlal_ex & instr_sign_ex_i;
  assign alu_mul_ex        = (instr_alu_ctl_ex_i[17:13] == 5'b11011);
  assign alu_mull_ex       = ((instr_alu_ctl_ex_i[17:13] == 5'b11010) |
                              mul_mull_extend_ex_i);
  assign alu_smull_ex      = alu_mull_ex & instr_sign_ex_i;
  assign alu_div_ex        = (instr_alu_ctl_ex_i[17:13] == 5'b11000);
  assign alu_sdiv_ex       = alu_div_ex & instr_sign_ex_i;
  assign alu_ml_ex         = ((instr_alu_ctl_ex_i[17:14] == 4'b1110) &
                              ~mul_mull_extend_ex_i);

  assign alu_shift_ex      = (instr_alu_ctl_ex_i[17:15] == 3'b001);

  assign alu_adder_ex      = alu_add_in2_valid_ex & ~alu_mov_ex;
  assign alu_arith_ex      = alu_adder_ex | bwise_valid_ex;


  assign shift1_sel_b_ex    = instr_alu_ctl_ex_i[9];
  assign shift2_sel_a_ex    = instr_alu_ctl_ex_i[11];
  assign shift2_sel_i_ex    = instr_alu_ctl_ex_i[10];
  assign shift_inlinec_ex   = instr_alu_ctl_ex_i[5:4];
  assign shift_rrx_valid_ex = instr_alu_ctl_ex_i[12];
  assign shift_ctl_ex       = instr_alu_ctl_ex_i[7:6];


  assign shift_in1_ex = ({32{shift1_sel_b_ex}} & rf_rd_b_data_ex_i);
  assign shift_in2_ex = (({8{shift2_sel_a_ex}} & rf_rd_a_data_ex_i[7:0]) |
                         ({8{shift2_sel_i_ex}} & shift_in2_fast_ex_i));

  cm3_dpu_alu_bshift
    u_cm3_dpu_alu_bshift
      (// Inputs
       .instr_immed_valid_ex_i     (instr_immed_valid_ex_i),
       .shift_ctl_ex_i             (shift_ctl_ex),
       .shift_in1_ex_i             (shift_in1_ex),
       .shift_in2_ex_i             (shift_in2_ex),
       .psr_ex_i                   (psr_ex_i),
       .shift_res_ex_o             (shift_res_ex),
       .shift_cout_ex_o            (shift_cout_ex)
       );

  assign shift_rrx_res_ex = { shift_rrx_valid_ex ? psr_ex_i[29] :
                                                   shift_res_ex[31],
                              shift_res_ex[30:0] };

  assign imm_shift_mux_ex = (instr_immed_valid_ex_i ? instr_immed_ex_i :
                                                      shift_rrx_res_ex);

  assign mov_in_ex   = ({32{alu_mov_rev_ex}} &
                        (imm_shift_mux_ex ^ {32{alu_inv_in2_ex}}));

  assign mov_ctrl_ex = ({4{alu_mov_rev_ex}} &
                        {alu_rev_ex, instr_sign_ex_i, instr_data_size_ex_i});

  always @ (mov_ctrl_ex or mov_in_ex)
    case (mov_ctrl_ex)
      4'b0000 : mov_res_ex = {{24{1'b0}},mov_in_ex[7:0]};
      4'b0100 : mov_res_ex = {{24{mov_in_ex[7]}},mov_in_ex[7:0]};
      4'b0001 : mov_res_ex = {{16{1'b0}},mov_in_ex[15:0]};
      4'b0101 : mov_res_ex = {{16{mov_in_ex[15]}},mov_in_ex[15:0]};
      4'b0010,
      4'b0110 : mov_res_ex = mov_in_ex;
      4'b1000 : mov_res_ex = {mov_in_ex[ 7: 0],mov_in_ex[15: 8],
                              mov_in_ex[23:16],mov_in_ex[31:24]};
      4'b1001 : mov_res_ex = {mov_in_ex[23:16],mov_in_ex[31:24],
                              mov_in_ex[ 7: 0],mov_in_ex[15: 8]};
      4'b1101 : mov_res_ex = {{16{mov_in_ex[7]}},
                              mov_in_ex[7:0],mov_in_ex[15:8]};
      4'b0011,
      4'b0111,
      4'b1010,
      4'b1011,
      4'b1100,
      4'b1110,
      4'b1111 : mov_res_ex = {32{1'bx}}; // unused, see ovl
      default : mov_res_ex = {32{1'bx}};
    endcase


  assign movt_res_ex = ({32{alu_movt_ex}} &
                        {instr_immed_ex_i[31:16],rf_rd_b_data_ex_i[15:0]});

  assign rbit_in_ex  = rf_rd_b_data_ex_i & {32{alu_rbit_ex}};

  assign rbit_res_ex = { rbit_in_ex[0],  rbit_in_ex[1],  rbit_in_ex[2],
                         rbit_in_ex[3],  rbit_in_ex[4],  rbit_in_ex[5],
                         rbit_in_ex[6],  rbit_in_ex[7],  rbit_in_ex[8],
                         rbit_in_ex[9],  rbit_in_ex[10], rbit_in_ex[11],
                         rbit_in_ex[12], rbit_in_ex[13], rbit_in_ex[14],
                         rbit_in_ex[15], rbit_in_ex[16], rbit_in_ex[17],
                         rbit_in_ex[18], rbit_in_ex[19], rbit_in_ex[20],
                         rbit_in_ex[21], rbit_in_ex[22], rbit_in_ex[23],
                         rbit_in_ex[24], rbit_in_ex[25], rbit_in_ex[26],
                         rbit_in_ex[27], rbit_in_ex[28], rbit_in_ex[29],
                         rbit_in_ex[30], rbit_in_ex[31] };



  assign mul_in1_ex[15:0]  = (({16{mul_input_sel_ex_i[0]}} & rf_rd_a_data_ex_i[15:0]) |
                              ({16{mul_input_sel_ex_i[1]}} & rf_rd_a_data_ex_i[31:16]));

  assign mul_in1_ex[31:16] =  ({16{mul_input_sel_ex_i[2]}} & rf_rd_a_data_ex_i[31:16]);

  assign mul_in2_ex[15:0]  = (({16{mul_input_sel_ex_i[3]}} & rf_rd_b_data_ex_i[15:0]) |
                              ({16{mul_input_sel_ex_i[4]}} & rf_rd_b_data_ex_i[31:16]));

  assign mul_in2_ex[31:16] =  ({16{mul_input_sel_ex_i[5]}} & rf_rd_b_data_ex_i[31:16]);

  assign mul_res_ex = (mul_in1_ex * mul_in2_ex) & {32{~opt_ext_en}};

  assign mcyc_a_signed_ex = rf_rd_a_data_ex_i[31] & (alu_smull_ex | alu_smlal_ex);
  assign mcyc_b_signed_ex = rf_rd_b_data_ex_i[31] & (alu_smull_ex | alu_smlal_ex);

  always @ (alu_mcyc_state_ex_i or mcyc_b_signed_ex or mcyc_temp2_ex or
            rf_rd_a_data_ex_i)
    case (alu_mcyc_state_ex_i)
      `CORE_MUL_MLx_2,        `CORE_MUL_xMLAL_1,      `CORE_MUL_xMLAL_2 :
        mcyc_alu_in1_ex = rf_rd_a_data_ex_i;
      `CORE_MUL_xMULL_1 :
        mcyc_alu_in1_ex = ({32{mcyc_b_signed_ex}} & rf_rd_a_data_ex_i);
      `CORE_MUL_xMULL_2,      `CORE_MUL_xMULL_3,      `CORE_MUL_xMULL_4 :
        mcyc_alu_in1_ex = mcyc_temp2_ex[31:0];
      `CORE_MUL_xMULL_5,      `CORE_MUL_MLA_P :
        mcyc_alu_in1_ex = {mcyc_temp2_ex[15:0],{16{1'b0}}};
      `CORE_DIV_CLZ_1,        `CORE_DIV_WREG,         `CORE_MUL_NORMAL,
      `CORE_MUL_MLx_1,        `CORE_DIV_CLZ_2,        `CORE_DIV_INIT,
      `CORE_DIV_SRT :
        mcyc_alu_in1_ex = {32{1'b0}};
      default :
        mcyc_alu_in1_ex = {32{1'bx}};
    endcase

  always @ (alu_mcyc_state_ex_i or mcyc_a_signed_ex or mcyc_temp_ex or
            rf_rd_a_data_ex_i or rf_rd_b_data_ex_i or mcyc_temp2_shift_ex)
    case (alu_mcyc_state_ex_i)
      `CORE_MUL_MLx_2 :
        mcyc_alu_in2_ex = mcyc_temp_ex[31:0];
      `CORE_MUL_xMULL_1 :
        mcyc_alu_in2_ex = ({32{mcyc_a_signed_ex}} &  rf_rd_b_data_ex_i);
      `CORE_MUL_xMULL_2 :
        mcyc_alu_in2_ex = mcyc_temp_ex[63:32];
      `CORE_MUL_xMULL_3,      `CORE_MUL_xMULL_4,      `CORE_MUL_xMULL_5,
      `CORE_MUL_MLA_P :
        mcyc_alu_in2_ex = mcyc_temp_ex[47:16];
      `CORE_MUL_xMLAL_1 :
        mcyc_alu_in2_ex = mcyc_temp_ex[31:0];
      `CORE_MUL_xMLAL_2 :
        mcyc_alu_in2_ex = mcyc_temp_ex[63:32];
      `CORE_DIV_CLZ_1 :
        mcyc_alu_in2_ex = rf_rd_a_data_ex_i;
      `CORE_DIV_WREG :
        mcyc_alu_in2_ex = mcyc_temp2_shift_ex[31:0];
      `CORE_MUL_NORMAL,       `CORE_MUL_MLx_1,        `CORE_DIV_CLZ_2,
      `CORE_DIV_INIT,         `CORE_DIV_SRT :
        mcyc_alu_in2_ex = {32{1'b0}};
      default :
        mcyc_alu_in2_ex = {32{1'bx}};
    endcase


  assign alu_inv_in1_ex       = instr_alu_ctl_ex_i[0];
  assign alu_inv_in2_ex       = instr_alu_ctl_ex_i[1];
  assign alu_add_in1_valid_ex = instr_alu_ctl_ex_i[2];
  assign alu_add_in2_valid_ex = instr_alu_ctl_ex_i[3];
  assign alu_cin_valid_ex     = mul_cin_ex_i | adder_cin_valid_ex;
  assign alu_cin_ex           = ((alu_cin_valid_ex & psr_ex_i[29]) |
                                 mul_cin_ex_i);

  assign alu_in1_ex = ((({32{alu_arith_ex}} &
                         {rf_rd_a_data_ex_i[31:2],
                          {2{~instr_algn_ex_i}} & rf_rd_a_data_ex_i[1:0]}) |
                        ({32{alu_muldiv_ex}} & mcyc_alu_in1_ex)) ^
                       {32{alu_inv_in1_ex}});

  assign alu_in2_ex = ((({32{alu_arith_ex}} & imm_shift_mux_ex) |
                        ({32{alu_muldiv_ex}} & mcyc_alu_in2_ex)) ^
                       {32{alu_inv_in2_ex | mul_neg_in2_ex_i}});


  assign adder_cin_valid_ex = instr_alu_ctl_ex_i[20];
  assign adder_in1_valid_ex = alu_add_in1_valid_ex | mul_add_valid_ex_i;
  assign adder_in2_valid_ex = alu_add_in2_valid_ex | mul_add_valid_ex_i;

  assign adder_in1_ex = {32{adder_in1_valid_ex}} & alu_in1_ex;
  assign adder_in2_ex = {32{adder_in2_valid_ex}} & alu_in2_ex;

  assign adder_cin_ex = (((alu_add_in2_valid_ex | mul_add_valid_ex_i) &
                          (alu_cin_valid_ex & alu_cin_ex)) |
                         ((alu_inv_in2_ex | mul_neg_in2_ex_i | alu_inv_in1_ex) &
                          (~alu_cin_valid_ex | alu_cin_ex) & ~bwise_valid_ex));

  assign adder_res_ex = ({1'b0,adder_in1_ex} +
                         {1'b0,adder_in2_ex} +
                         {{32{1'b0}},adder_cin_ex});


  assign top_adder_in1_valid_ex = (((alu_mcyc_state_ex_i == `CORE_MUL_xMULL_5) |
                                    (alu_mcyc_state_ex_i == `CORE_MUL_MLA_P  )) &
                                   ~opt_ext_en);

  assign top_adder_in2_valid_ex = (((alu_mcyc_state_ex_i == `CORE_MUL_xMULL_3) |
                                    (alu_mcyc_state_ex_i == `CORE_MUL_xMULL_4) |
                                    (alu_mcyc_state_ex_i == `CORE_MUL_xMULL_5) |
                                    (alu_mcyc_state_ex_i == `CORE_MUL_MLA_P  )) &
                                   ~opt_ext_en);

  assign top_adder_in1_ex = {16{top_adder_in1_valid_ex}} & mcyc_temp2_ex[31:16];
  assign top_adder_in2_ex = {16{top_adder_in2_valid_ex}} & mcyc_temp_ex[63:48];
  assign top_adder_cin_ex = top_adder_in2_valid_ex & alu_cout_ex;

  assign top_adder_res_ex  = (top_adder_in1_ex +
                              top_adder_in2_ex +
                              {{15{1'b0}}, top_adder_cin_ex});

  assign full_adder_res_ex = { top_adder_res_ex[15:0], adder_res_ex[31:0] };


  assign div_srt_in1_ex  = mcyc_temp_low_ex;
  assign div_srt_in2_ex  = mcyc_temp_upp_ex;

  cm3_dpu_alu_srtdiv
    u_cm3_dpu_alu_srtdiv
      (// Inputs
       .div_srt_init_ex_i          (div_srt_init_ex_i),
       .div_srt_sign_ex_i          (div_srt_sign_ex_i),
       .div_srt_in1_ex_i           (div_srt_in1_ex),
       .div_srt_in2_ex_i           (div_srt_in2_ex),
       .div_srt_rem_ex_o           (div_srt_rem_ex),
       .div_srt_res_ex_o           (div_srt_res_ex)
       );


  assign nxt_mcyc_temp_low_sel_ex = { (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_4) |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_3),
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_1) |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_MLx_1)   |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_MLA_P),
                                      (alu_mcyc_state_ex_i==`CORE_DIV_CLZ_1),
                                      (alu_mcyc_state_ex_i==`CORE_DIV_SRT) &
                                      ~div_cnt_ex_i[5]};

  assign nxt_mcyc_temp_upp_sel_ex = { (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_5 &
                                       (alu_mlal_ex | mul_mla_track_ex_i)) |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_4) |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_3),
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_2) |
                                      (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_1),
                                      (alu_mcyc_state_ex_i==`CORE_DIV_INIT) |
                                      (alu_mcyc_state_ex_i==`CORE_DIV_SRT &
                                       ~div_cnt_ex_i[5])};

  assign mcyc_temp_low_we_ex  = (|nxt_mcyc_temp_low_sel_ex);

  assign nxt_mcyc_temp_low_ex = (({32{nxt_mcyc_temp_low_sel_ex[0]}} &
                                  div_srt_rem_ex)                    |
                                 ({32{nxt_mcyc_temp_low_sel_ex[1]}} &
                                  full_adder_res_ex[31:0])           |
                                 ({32{nxt_mcyc_temp_low_sel_ex[2]}} &
                                  mul_res_ex)                       |
                                 ({32{nxt_mcyc_temp_low_sel_ex[3]}} &
                                  {full_adder_res_ex[15:0],
                                   mcyc_temp_low_ex[15:0]}));

  always @ (posedge aclk or negedge full_hreset_n)
    if (!full_hreset_n)
      mcyc_temp_low_ex <= {32{1'b0}};
    else if (mcyc_temp_low_we_ex)
      mcyc_temp_low_ex <= nxt_mcyc_temp_low_ex;

  assign mcyc_temp_upp_we_ex  = (|nxt_mcyc_temp_upp_sel_ex);

  assign nxt_mcyc_temp_upp_ex = (({32{nxt_mcyc_temp_upp_sel_ex[0]}} &
                                  shift_res_ex)                      |
                                 ({32{nxt_mcyc_temp_upp_sel_ex[1]}} &
                                  full_adder_res_ex[31:0])           |
                                 ({32{nxt_mcyc_temp_upp_sel_ex[2]}} &
                                  full_adder_res_ex[47:16]));

  always @ (posedge aclk or negedge full_hreset_n)
    if (!full_hreset_n)
      mcyc_temp_upp_ex <= {32{1'b0}};
    else if (mcyc_temp_upp_we_ex)
      mcyc_temp_upp_ex <= nxt_mcyc_temp_upp_ex;

  assign mcyc_temp_ex = { mcyc_temp_upp_ex, mcyc_temp_low_ex };

  assign mcyc_temp2_we_ex  = ((alu_mcyc_state_ex_i==`CORE_DIV_CLZ_1   &
                               div_fast_zero_ex_i)                     |
                              (alu_mcyc_state_ex_i==`CORE_DIV_CLZ_2)   |
                              (alu_mcyc_state_ex_i==`CORE_DIV_SRT)     |
                              (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_1) |
                              (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_2) |
                              (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_3) |
                              (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_4) |
                              (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_5 &
                               (alu_mlal_ex | mul_mla_track_ex_i)));

  assign nxt_mcyc_temp2_ex = (({35{alu_mcyc_state_ex_i==`CORE_DIV_SRT}} &
                               {mcyc_temp2_ex[30:0],div_srt_res_ex[3:0]}) |
                              ({35{alu_mcyc_state_ex_i!=`CORE_MUL_NORMAL  &
                                   alu_mcyc_state_ex_i!=`CORE_DIV_INIT    &
                                   alu_mcyc_state_ex_i!=`CORE_DIV_SRT     &
                                   alu_mcyc_state_ex_i!=`CORE_DIV_CLZ_1   &
                                   alu_mcyc_state_ex_i!=`CORE_MUL_xMULL_1 &
                                   alu_mcyc_state_ex_i!=`CORE_DIV_CLZ_2}} &
                               {3'b000,mul_res_ex}));

  always @ (posedge aclk or negedge full_hreset_n)
    if (!full_hreset_n)
      mcyc_temp2_ex <= {35{1'b0}};
    else if (mcyc_temp2_we_ex)
      mcyc_temp2_ex <= nxt_mcyc_temp2_ex;

  assign mcyc_temp2_shift_ex = (mcyc_temp2_ex[34:0] >> div_cnt_reg_ex_i);


  assign clz_ctrl_ex     = ((alu_mcyc_state_ex_i==`CORE_DIV_CLZ_1) &
                            alu_sdiv_ex & rf_rd_b_data_ex_i[31]);

  assign clz_in_ex       = (({32{alu_mcyc_state_ex_i==`CORE_DIV_CLZ_1}} &
                             rf_rd_b_data_ex_i)                          |
                            ({32{alu_mcyc_state_ex_i==`CORE_DIV_CLZ_2}} &
                             mcyc_temp_low_ex)                           |
                            ({32{alu_clz_ex}} & rf_rd_b_data_ex_i));

  assign clz_data_ex     = {32{clz_ctrl_ex}} ^ clz_in_ex;

  assign clz_sel_b3_1_ex = ~( clz_data_ex[31])    & clz_data_ex[30];
  assign clz_sel_b3_2_ex = ~(|clz_data_ex[31:30]) & clz_data_ex[29];
  assign clz_sel_b3_3_ex = ~(|clz_data_ex[31:29]) & clz_data_ex[28];
  assign clz_sel_b3_4_ex = ~(|clz_data_ex[31:28]) & clz_data_ex[27];
  assign clz_sel_b3_5_ex = ~(|clz_data_ex[31:27]) & clz_data_ex[26];
  assign clz_sel_b3_6_ex = ~(|clz_data_ex[31:26]) & clz_data_ex[25];
  assign clz_sel_b3_7_ex = ~(|clz_data_ex[31:25]);

  assign clz_b3_res_ex   = (({3{clz_sel_b3_1_ex}} & 3'b001) |
                            ({3{clz_sel_b3_2_ex}} & 3'b010) |
                            ({3{clz_sel_b3_3_ex}} & 3'b011) |
                            ({3{clz_sel_b3_4_ex}} & 3'b100) |
                            ({3{clz_sel_b3_5_ex}} & 3'b101) |
                            ({3{clz_sel_b3_6_ex}} & 3'b110) |
                            ({3{clz_sel_b3_7_ex}} & 3'b111));

  assign clz_sel_b2_1_ex = ~( clz_data_ex[23])    & clz_data_ex[22];
  assign clz_sel_b2_2_ex = ~(|clz_data_ex[23:22]) & clz_data_ex[21];
  assign clz_sel_b2_3_ex = ~(|clz_data_ex[23:21]) & clz_data_ex[20];
  assign clz_sel_b2_4_ex = ~(|clz_data_ex[23:20]) & clz_data_ex[19];
  assign clz_sel_b2_5_ex = ~(|clz_data_ex[23:19]) & clz_data_ex[18];
  assign clz_sel_b2_6_ex = ~(|clz_data_ex[23:18]) & clz_data_ex[17];
  assign clz_sel_b2_7_ex = ~(|clz_data_ex[23:17]);

  assign clz_b2_res_ex   = (({3{clz_sel_b2_1_ex}} & 3'b001) |
                            ({3{clz_sel_b2_2_ex}} & 3'b010) |
                            ({3{clz_sel_b2_3_ex}} & 3'b011) |
                            ({3{clz_sel_b2_4_ex}} & 3'b100) |
                            ({3{clz_sel_b2_5_ex}} & 3'b101) |
                            ({3{clz_sel_b2_6_ex}} & 3'b110) |
                            ({3{clz_sel_b2_7_ex}} & 3'b111));

  assign clz_sel_b1_1_ex = ~( clz_data_ex[15])    & clz_data_ex[14];
  assign clz_sel_b1_2_ex = ~(|clz_data_ex[15:14]) & clz_data_ex[13];
  assign clz_sel_b1_3_ex = ~(|clz_data_ex[15:13]) & clz_data_ex[12];
  assign clz_sel_b1_4_ex = ~(|clz_data_ex[15:12]) & clz_data_ex[11];
  assign clz_sel_b1_5_ex = ~(|clz_data_ex[15:11]) & clz_data_ex[10];
  assign clz_sel_b1_6_ex = ~(|clz_data_ex[15:10]) & clz_data_ex[09];
  assign clz_sel_b1_7_ex = ~(|clz_data_ex[15:09]);

  assign clz_b1_res_ex   = (({3{clz_sel_b1_1_ex}} & 3'b001) |
                            ({3{clz_sel_b1_2_ex}} & 3'b010) |
                            ({3{clz_sel_b1_3_ex}} & 3'b011) |
                            ({3{clz_sel_b1_4_ex}} & 3'b100) |
                            ({3{clz_sel_b1_5_ex}} & 3'b101) |
                            ({3{clz_sel_b1_6_ex}} & 3'b110) |
                            ({3{clz_sel_b1_7_ex}} & 3'b111));

  assign clz_sel_b0_1_ex = ~( clz_data_ex[7])   & clz_data_ex[6];
  assign clz_sel_b0_2_ex = ~(|clz_data_ex[7:6]) & clz_data_ex[5];
  assign clz_sel_b0_3_ex = ~(|clz_data_ex[7:5]) & clz_data_ex[4];
  assign clz_sel_b0_4_ex = ~(|clz_data_ex[7:4]) & clz_data_ex[3];
  assign clz_sel_b0_5_ex = ~(|clz_data_ex[7:3]) & clz_data_ex[2];
  assign clz_sel_b0_6_ex = ~(|clz_data_ex[7:2]) & clz_data_ex[1];
  assign clz_sel_b0_7_ex = ~(|clz_data_ex[7:1]);

  assign clz_b0_res_ex   = (({3{clz_sel_b0_1_ex}} & 3'b001) |
                            ({3{clz_sel_b0_2_ex}} & 3'b010) |
                            ({3{clz_sel_b0_3_ex}} & 3'b011) |
                            ({3{clz_sel_b0_4_ex}} & 3'b100) |
                            ({3{clz_sel_b0_5_ex}} & 3'b101) |
                            ({3{clz_sel_b0_6_ex}} & 3'b110) |
                            ({3{clz_sel_b0_7_ex}} & 3'b111));

  assign clz_sel_b3_ex   =  (|clz_data_ex[31:24]);
  assign clz_sel_b2_ex   = ~(|clz_data_ex[31:24]) & (|clz_data_ex[23:16]);
  assign clz_sel_b1_ex   = ~(|clz_data_ex[31:16]) & (|clz_data_ex[15:08]);
  assign clz_sel_b0_ex   = ~(|clz_data_ex[31:08]) & (|clz_data_ex[07:00]);
  assign clz_sel_w0_ex   = ~(|clz_data_ex[31:00]);

  assign clz_res_ex      = (({6{clz_sel_b3_ex}} & {3'b000, clz_b3_res_ex}) |
                            ({6{clz_sel_b2_ex}} & {3'b001, clz_b2_res_ex}) |
                            ({6{clz_sel_b1_ex}} & {3'b010, clz_b1_res_ex}) |
                            ({6{clz_sel_b0_ex}} & {3'b011, clz_b0_res_ex}) |
                            ({6{clz_sel_w0_ex}} & {3'b100, 3'b000}));

  assign alu_clz_res_ex  = ({32{alu_clz_ex}} & {{26{1'b0}},clz_res_ex});


  assign bwise_res_ex   = (({32{bwise_and_ex}} & (alu_in1_ex & alu_in2_ex)) |
                           ({32{bwise_eor_ex}} & (alu_in1_ex ^ alu_in2_ex)) |
                           ({32{bwise_orr_ex}} & (alu_in1_ex | alu_in2_ex)));

  assign mrs_psr_res_ex = {{5{alu_mrs_flag_ex}} & psr_ex_i[31:27], {7{1'b0}},
                           {4{alu_mrs_flag_ex}} & psr_ex_i[19:16], {7{1'b0}},
                           {9{alu_mrs_intnum_ex}} & int_isr_ex_i};


  assign mrs_ctrl_res_ex = { {24{1'b0}},
                             {{5{1'b0}}, {3{alu_mrs_ctrl_ex}} &
                              {fpu_active_ex_i,
                               process_sp_ex_i[0],
                               user_thread_ex_i}}   |
                             {{7{1'b0}}, alu_mrs_pri_ex & nvic_pmask_i} |
                             ({8{alu_mrs_base_ex}} & nvic_basepri_i)    |
                             {{7{1'b0}}, alu_mrs_fault_ex & nvic_fmask_i} };


  assign bf_sat_res_ex = (// SSAT -ve saturation
                          ({32{shift_sel_ex_i[5]}} & shift_l_mask_ex_i)    |
                          ({32{shift_sel_ex_i[4]}} & shift_r_mask_ex_i)    |
                          ({32{shift_sel_ex_i[3]}} & shift_res_ex)         |
                          ({32{shift_sel_ex_i[2]}} &
                           ((rf_rd_a_data_ex_i & shift_lr_mask_ex_i)       |
                            (shift_res_ex & ~shift_lr_mask_ex_i)))         |
                          ({32{shift_sel_ex_i[1]}} &
                           ((shift_res_ex & shift_r_mask_ex_i)             |
                            shift_l_mask_ex_i))                            |
                          ({32{shift_sel_ex_i[0]}} & shift_rrx_res_ex));


  assign alu_res_ex       = ({32{alu_arith_ex}} &
                             (adder_res_ex[31:0] | bwise_res_ex));

  assign bwise_zero_ex    = (~|bwise_res_ex) & bwise_valid_ex;

  assign adder_cs_sout_ex = adder_in1_ex ^ adder_in2_ex;
  assign adder_cs_cout_ex = adder_in1_ex[30:0] | adder_in2_ex[30:0];
  assign adder_zero_ex    = (adder_cs_sout_ex[31:0] ==
                             {adder_cs_cout_ex[30:0],adder_cin_ex});


  assign arith_sel_ex = {(alu_mull_ex & (alu_mcyc_state_ex_i!=`CORE_MUL_xMULL_5)),
                         (alu_mull_ex & (alu_mcyc_state_ex_i==`CORE_MUL_xMULL_5)) |
                         (alu_ml_ex & (alu_mcyc_state_ex_i==`CORE_MUL_MLA_P)),
                         (alu_ml_ex & (alu_mcyc_state_ex_i!=`CORE_MUL_MLA_P)) |
                         (alu_div_ex & (alu_mcyc_state_ex_i==`CORE_DIV_WREG)) |
                         alu_mlal_ex,
                         alu_mul_ex};

  assign arith_res_ex = (({32{arith_sel_ex[3]}} & {full_adder_res_ex[15:0],
                                                   mcyc_temp_ex[15:0]})     |
                         ({32{arith_sel_ex[2]}} & full_adder_res_ex[47:16]) |
                         ({32{arith_sel_ex[1]}} & full_adder_res_ex[31:0])  |
                         ({32{arith_sel_ex[0]}} & mul_res_ex));


  assign alu_rf_wr_d_data_ex = (e_simd_res_ex_i | rbit_res_ex     |
                                alu_clz_res_ex  | alu_res_ex      |
                                mov_res_ex      | mrs_psr_res_ex  |
                                mrs_ctrl_res_ex | movt_res_ex     |
                                arith_res_ex    | bf_sat_res_ex);

  assign alu_cout_ex  = (adder_res_ex[32] &
                         (alu_add_in2_valid_ex | mul_add_valid_ex_i));

  assign alu_oflow_ex = (( adder_in1_ex[31] &  adder_in2_ex[31] & ~alu_res_ex[31]) |
                         (~adder_in1_ex[31] & ~adder_in2_ex[31] &  alu_res_ex[31]));

  assign alu_nzcvq_erly_ex[`CORE_N] = ((bwise_valid_ex & bwise_res_ex[31]) |
                                       (alu_msr_flag_ex & rf_rd_b_data_ex_i[31]) |
                                       (alu_mov_rev_ex & mov_res_ex[31]) |
                                       (alu_shift_ex   & shift_rrx_res_ex[31]));

  assign alu_nzcvq_erly_ex[`CORE_Z] = ((bwise_valid_ex  & bwise_zero_ex) |
                                       (alu_msr_flag_ex & rf_rd_b_data_ex_i[30]) |
                                       (alu_mov_rev_ex & ~(|mov_res_ex)) |
                                       (alu_shift_ex   & ~(|shift_rrx_res_ex)));

  assign alu_nzcvq_erly_ex[`CORE_C] = ((alu_msr_flag_ex & rf_rd_b_data_ex_i[29]) |
                                       (alu_shift_ex & shift_cout_ex)         |
                                       (shift_inlinec_ex[0] & shift_cout_ex)  |
                                       (shift_inlinec_ex[1] & instr_immed_ex_i[31]));

  assign alu_nzcvq_erly_ex[`CORE_V] = (alu_msr_flag_ex & rf_rd_b_data_ex_i[28]);

  assign alu_nzcvq_ex[`CORE_N] = (alu_adder_ex ? adder_res_ex[31] :
                                                 alu_nzcvq_erly_ex[`CORE_N]);

  assign alu_nzcvq_ex[`CORE_Z] = (alu_adder_ex ? adder_zero_ex :
                                                 alu_nzcvq_erly_ex[`CORE_Z]);

  assign alu_nzcvq_ex[`CORE_C] = (alu_adder_ex ? (~(|shift_inlinec_ex) & alu_cout_ex) :
                                                 alu_nzcvq_erly_ex[`CORE_C]);

  assign alu_nzcvq_ex[`CORE_V] = (alu_adder_ex ? alu_oflow_ex :
                                                 alu_nzcvq_erly_ex[`CORE_V]);

  assign alu_nzcvq_ex[`CORE_Q] = alu_msr_flag_ex ? rf_rd_b_data_ex_i[27] : 1'b1;

  assign alu_flag_res_ex       = alu_nzcvq_ex;

  assign alu_ge_res_ex         = rf_rd_b_data_ex_i[19:16] & {4{opt_ext_en}};


  assign alu_mul_flag_res_ex[`CORE_N] = mul_res_ex[31];
  assign alu_mul_flag_res_ex[`CORE_Z] = ~(|mul_res_ex) & alu_mul_ex;
  assign alu_mul_flag_res_ex[`CORE_C] = 1'b0;
  assign alu_mul_flag_res_ex[`CORE_V] = 1'b0;
  assign alu_mul_flag_res_ex[`CORE_Q] = 1'b0;

  assign shift_in1_ex_o        = shift_in1_ex;
  assign shift_res_ex_o        = shift_res_ex;
  assign clz_res_ex_o          = clz_res_ex;
  assign alu_cout_ex_o         = alu_cout_ex;
  assign alu_flag_res_ex_o     = alu_flag_res_ex;
  assign alu_mul_flag_res_ex_o = alu_mul_flag_res_ex;
  assign alu_ge_res_ex_o       = alu_ge_res_ex;
  assign alu_rf_wr_d_data_ex_o = alu_rf_wr_d_data_ex;


`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Must never enter unused mov_ctrl_ex state")
    ovl_mov_ctrl_check(hclk, hreset_n, ((mov_ctrl_ex == 4'b0011) |
                                        (mov_ctrl_ex == 4'b0111) |
                                        (mov_ctrl_ex == 4'b1010) |
                                        (mov_ctrl_ex == 4'b1011) |
                                        (mov_ctrl_ex == 4'b1100) |
                                        (mov_ctrl_ex == 4'b1110) |
                                        (mov_ctrl_ex == 4'b1111)));

  assert_never #(0,0,"alu_mcyc_state_ex!=idle & alu_clz_ex should be excl")
     ovl_clz_excl(hclk, hreset_n,
                  alu_clz_ex & (alu_mcyc_state_ex_i==`CORE_DIV_CLZ_1 |
                                alu_mcyc_state_ex_i==`CORE_DIV_CLZ_2));

  assert_never #(0,0,"Multi-cycle controls outside of valid multiply states")
     ovl_mcyc_ctrl_check1(hclk, hreset_n, alu_ml_ex &
                          ~(alu_mcyc_state_ex_i==`CORE_MUL_MLx_1 |
                            alu_mcyc_state_ex_i==`CORE_MUL_MLx_2 |
                            alu_mcyc_state_ex_i==`CORE_MUL_MLA_P |
                            alu_mcyc_state_ex_i==`CORE_MUL_NORMAL));

  assert_never #(0,0,"ALU Clock Gate control has been incorrectly formed")
     ovl_alu_cg_check(hclk, hreset_n, ~alu_cg_en_i &
                      (mcyc_temp_low_we_ex |
                       mcyc_temp_upp_we_ex |
                       mcyc_temp2_we_ex));

  assert_zero_one_hot #(0,4,0,"arith_sel_ex must be exclusive")
     ovl_arith_sel_excl(hclk, hreset_n, arith_sel_ex);

`endif

`include "cm3_dpu_undefs.vp"

endmodule
