`include "cm3_dpu_defs.vp"

module cm3_dpu_br_dec
  (// Inputs
   hclk, hreset_n, state_fe_i, fe_buf_has_valid_instr_i, fe_address_add_res_i,
   instr_de_i, instr_valid_de_i, instr_ccpass_de_i, instr_itblock_de_i,
   instr_use_word_de_i, instr_16_mul_ex_i, instr_lr_ilock_ex_i,
   instr_br_pflush_ex_i, instr_detect_exit_ex_i, set_instr_redo_ex_i,
   ahb_htransi_i, mtx_ahb_hreadyi_i, fe_haddracci_i, nvic_slp_asleep_i,
   ext_iflush_i, nvic_int_invoke_i, nvic_int_tail_i, lsu_flag_wr_ex_i,
   lsu_pflush_ex_i, lsu_rf_wr_d_data_ex_i, rf_rd_lr_de_i, rf_rd_b_data_ex_i,
   rf_rd_b_zero_ex_i, fault_dblfault_ex_i, dbg_halted_i,
   instr_br_lr_sel_de_o, instr_br_cond_de_o, instr_br_ccpass_de_o,
   instr_br_imm_de_o, instr_br_req_de_o, instr_br_pflush_fast_de_o,
   instr_br_pflush_de_o, instr_brchstat_de_o, instr_de_br_det_de_o,
   instr_br_imm_ex_o, instr_br_type_ex_o, instr_pflush_ex_o,
   instr_lsu_br_det_ex_o, instr_det_erly_br_trk_o, instr_det_erly_br_trk_v_o,
   rf_pc_wr_data_de_o
   );

  input          hclk;
  input          hreset_n;
  input [3:0]    state_fe_i;                // Fetch state
  input          fe_buf_has_valid_instr_i;  // Fetch buffer has valid instr
  input [31:1]   fe_address_add_res_i;      // Fetch address adder result
  input [31:0]   instr_de_i;                // Instruction in decode
  input          instr_valid_de_i;          // Instruction in decode is valid
  input          instr_ccpass_de_i;         // Condition code pass
  input          instr_itblock_de_i;        // In IT block
  input          instr_use_word_de_i;       // Consumed 32 bits of instruction
  input          instr_16_mul_ex_i;         // Instruction is a T1 MUL
  input          instr_lr_ilock_ex_i;       // LR interlock
  input [2:0]    instr_br_pflush_ex_i;      // Branch pipeline flush
  input          instr_detect_exit_ex_i;    // ISR exit detected
  input          set_instr_redo_ex_i;       // Set redo instr
  input [1:0]    ahb_htransi_i;             // AHB I-side HTRANS
  input          mtx_ahb_hreadyi_i;         // AHB I-side HREADY
  input          fe_haddracci_i;            // Modified I-side HADDRACC
  input          nvic_slp_asleep_i;         // Core asleep
  input          ext_iflush_i;              // IFLUSH request
  input          nvic_int_invoke_i;         // Interrupt stack request
  input          nvic_int_tail_i;           // Interrupt tail-chain request
  input          lsu_flag_wr_ex_i;          // LSU flag result write enable
  input          lsu_pflush_ex_i;           // LSU pipe flush
  input [31:0]   lsu_rf_wr_d_data_ex_i;     // LSU RF D port write data
  input [31:0]   rf_rd_lr_de_i;             // RF read LR
  input [31:0]   rf_rd_b_data_ex_i;         // RF B port read data
  input [1:0]    rf_rd_b_zero_ex_i;         // RF B port read data HW zero
  input          fault_dblfault_ex_i;       // Double fault
  input          dbg_halted_i;              // Core halted

  output         instr_br_lr_sel_de_o;      // Select LR for BLX and MOV PC LR
  output [3:0]   instr_br_cond_de_o;        // Condition code of branch
  output         instr_br_ccpass_de_o;      // Branch passed condition code
  output [31:0]  instr_br_imm_de_o;         // Branch immediate offset
  output         instr_br_req_de_o;         // Branch fetch request
  output [1:0]   instr_br_pflush_fast_de_o; // Branch pipeline flush (fast)
  output [1:0]   instr_br_pflush_de_o;      // Branch pipeline flush
  output [3:0]   instr_brchstat_de_o;       // Branch status information
  output         instr_de_br_det_de_o;      // Detected a branch in decode
  output [31:0]  instr_br_imm_ex_o;         // Branch immediate offset
  output [1:0]   instr_br_type_ex_o;        // Type of branch in execute
  output         instr_pflush_ex_o;         // Pipeline flush due to instr
  output         instr_lsu_br_det_ex_o;     // LSU branch detected
  output         instr_det_erly_br_trk_o;   // Early branch detected
  output         instr_det_erly_br_trk_v_o; // Validated early branch detected
  output [31:1]  rf_pc_wr_data_de_o;        // PC write data


  wire           i16_b1_de;                 // 16 bit B (1)
  wire           i16_b2_de;                 // 16 bit B (2)
  wire           i16_mov3_de;               // 16 bit MOV
  wire           i32_bl_de;                 // 32 bit BL
  wire           i32_b0_de;                 // 32 bit B (0)
  wire           instr_br_de_valid_de;      // Branch decode enable
  wire           instr_de_br_b0_de;         // Accepted B (0)
  wire           instr_de_br_b1_de;         // Accepted B (1)
  wire           instr_de_br_blx_de;        // Accepted BLX
  wire           instr_de_br_mov_de;        // Accepted MOV
  wire           instr_de_br_de;            // Decode time branch
  wire           instr_de_br_det_de;        // Branch detected
  wire [3:0]     instr_br_cond_de;          // Branch condition code
  wire           instr_br_lr_sel_de;        // Select LR for branch
  wire           instr_det_erly_br_trk_set; // Set track branch speculation
  wire           instr_det_erly_br_trk_we;  // Track B write enable
  wire           instr_det_erly_br_trk_clr; // Track B write clear
  reg            instr_det_erly_br_trk;     // Early branch detected
  wire           instr_det_erly_br_trk_v;   // Validated early branch detected
  wire           instr_br_ilock_de;         // Branch interlock in decode
  wire           instr_br_ilock_we;         // Branch interlock write enable
  reg            instr_br_ilock_ex;         // Branch interlock in execute
  wire           instr_br_ccpass_de;        // Branch condition code passed
  wire           instr_ex_br_blx_de;        // Execute time BLX decoded
  wire           instr_ex_br_czb_de;        // Execute time CZB decoded
  wire           instr_ex_br_isb_de;        // Execute time ISB decoded
  wire           instr_ex_br_mov_de;        // Execute time MOV decoded
  wire [2:0]     instr_ex_br_det_de;        // Execute time branch decoded
  wire           instr_ex_br_det_we;        // Execute time branch write enable
  reg  [2:0]     instr_ex_br_det_ex;        // Execute time branch in execute
  wire [6:0]     instr_ex_br_imm_a_de;      // Execute time branch immediate A
  wire [6:0]     instr_ex_br_imm_b_de;      // Execute time branch immediate B
  wire [6:0]     instr_ex_br_imm_de;        // Execute time branch immediate
  wire           instr_ex_br_imm_we;        // Execute time imm write enable
  reg  [6:0]     instr_ex_br_imm_ex;        // Execute time branch immediate
  wire           instr_ex_br_add_de;        // Execute time LSU ADD PC decoded
  wire           instr_ex_br_pop_de;        // Execute time LSU POP decoded
  wire           instr_ex_br_ldr_de;        // Execute time LSU LDR decoded
  wire           instr_ex_br_ldm_de;        // Execute time LSU LDM decoded
  wire           instr_ex_br_tbl_de;        // Execute time LSU TB decoded
  wire           instr_lsu_br_det_de;       // LSU branch dectected in decode
  wire           instr_lsu_br_det_we;       // LSU branch write enable
  reg            instr_lsu_br_det_ex;       // LSU branch dectected in execute
  wire [31:0]    instr_br_imm_de;           // Branch immediate offset
  wire [5:0]     instr_br_type_de;          // Branch type
  wire [4:0]     instr_br_ccinfo_de;        // Branch condition code info
  wire           nxt_instr_br_ccpass_ex;    // Next instr_br_ccpass_ex
  reg            instr_br_ccpass_ex;        // Branch condition code passed
  wire [3:0]     instr_brchstat_de;         // Branch info and status
  wire [31:0]    instr_br_imm_ex;           // Branch immediate offset
  wire [1:0]     instr_br_type_ex;          // Branch type
  wire           instr_br_req_de;           // Branch request
  wire           instr_alu_pflush_de;       // ALU pipeline flush in decode
  wire           instr_alu_pflush_we;       // ALU pipeline flush write enable
  reg            instr_alu_pflush_ex;       // ALU pipeline flush in execute
  wire           instr_lsu_pflush_de;       // LSU pipeline flush in decode
  wire           instr_lsu_pflush_we;       // LSU pipeline flush write enable
  reg            instr_lsu_pflush_ex;       // LSU pipeline flush in execute
  wire           instr_pflush_ex;           // Pipeline flush due to instr
  wire           instr_halt_pflush_de;      // Debug halt pipeline flush (De)
  wire           instr_halt_pflush_we;      // Debug pipeline flush write enable
  reg            instr_halt_pflush_ex;      // Debug halt pipeline flush (Ex)
  wire [1:0]     instr_br_pflush_fast_de;   // Branch pipeline flush (fast)
  wire [1:0]     instr_br_pflush_de;        // Branch pipeline flush
  wire [31:1]    rf_pc_wr_data_de;          // PC write data


  assign i16_b1_de    = (instr_de_i[15:12] == 4'b1101 &
                         (instr_de_i[11:9] != 3'b111));
  assign i16_b2_de    = (instr_de_i[15:11] == 5'b11100);
  assign i16_mov3_de  = (instr_de_i[15:8]  == 8'b01000110);

  assign i32_bl_de    = ((instr_de_i[15:11] == 5'b11110) &
                         instr_de_i[31] & instr_de_i[28]);

  assign i32_b0_de    = ((instr_de_i[15:11] == 5'b11110) &
                         (instr_de_i[9:7] != 3'b111) &
                         (instr_de_i[31:30] == 2'b10) & ~instr_de_i[28]);

  assign instr_br_de_valid_de = (fe_buf_has_valid_instr_i &
                                 ~instr_br_ilock_ex & ~instr_det_erly_br_trk &
                                 ~(|instr_br_type_ex) & ~instr_lsu_br_det_ex &
                                 (state_fe_i!=`CORE_FTCH_PREFETCH_1));


  assign instr_de_br_b1_de  = i16_b1_de & ~instr_16_mul_ex_i;

  assign instr_de_br_b0_de  = i32_b0_de & ~instr_16_mul_ex_i;

  assign instr_de_br_blx_de = ((instr_de_i[15:8] == 8'b01000111) &
                               (instr_de_i[6:3] == 4'b1110) &
                               ~instr_lr_ilock_ex_i);

  assign instr_de_br_mov_de = (i16_mov3_de &
                               (instr_de_i[7:0] == 8'b11110111) &
                               ~instr_lr_ilock_ex_i);

  assign instr_de_br_de     = (instr_de_br_b1_de  | i16_b2_de |
                               instr_de_br_b0_de  | i32_bl_de |
                               instr_de_br_mov_de | instr_de_br_blx_de);

  assign instr_de_br_det_de = instr_br_de_valid_de & instr_de_br_de;

  assign instr_br_cond_de   = (({4{instr_de_br_b1_de}} & instr_de_i[11:8]) |
                               ({4{instr_de_br_b0_de}} & instr_de_i[9:6])  |
                               ({4{~instr_de_br_b1_de  & ~instr_de_br_b0_de}} &
                                4'b1110));

  assign instr_br_lr_sel_de = (instr_br_de_valid_de &
                               (instr_de_br_blx_de | instr_de_br_mov_de));


  assign instr_det_erly_br_trk_set = (instr_de_br_det_de & ~instr_valid_de_i &
                                      instr_br_ccpass_de &
                                      ~instr_det_erly_br_trk_clr);

  assign instr_det_erly_br_trk_clr = (// Interrupt
                                      (state_fe_i==`CORE_FTCH_PREFETCH_1) |
                                      instr_br_pflush_ex_i[0]             |
                                      instr_br_pflush_ex_i[1]             |
                                      ext_iflush_i                        |
                                      dbg_halted_i);

  assign instr_det_erly_br_trk_we  = (// Set
                                      (instr_det_erly_br_trk_set &
                                       ~instr_det_erly_br_trk) |
                                      (instr_det_erly_br_trk_clr &
                                        instr_det_erly_br_trk));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_det_erly_br_trk <= 1'b0;
    else if (instr_det_erly_br_trk_we)
      instr_det_erly_br_trk <= instr_det_erly_br_trk_set;

  assign instr_det_erly_br_trk_v = (instr_det_erly_br_trk_set |
                                    (instr_det_erly_br_trk &
                                     ~instr_det_erly_br_trk_clr));

  assign instr_br_ilock_de = (// Branch not taken
                              (instr_de_br_det_de & ~instr_valid_de_i &
                               ~(|instr_br_pflush_ex_i[1:0]) & ~ext_iflush_i &
                               ~instr_br_ccpass_de & ~instr_br_ilock_ex &
                               ~dbg_halted_i) |
                              (instr_detect_exit_ex_i & ~nvic_int_invoke_i &
                               ~nvic_int_tail_i));

  assign instr_br_ilock_we = (// Set when interlock occurs
                              instr_br_ilock_de |
                              (instr_br_ilock_ex &
                               (// Clr on execute
                                (instr_valid_de_i) |
                                (instr_br_pflush_ex_i[1] &
                                 state_fe_i==`CORE_FTCH_FETCH) |
                                (state_fe_i==`CORE_FTCH_PREFETCH_1 |
                                 (state_fe_i==`CORE_FTCH_PRELOAD_1 &
                                  nvic_int_tail_i)) |
                                (ext_iflush_i &
                                 ~(state_fe_i==`CORE_FTCH_PRELOAD_1 |
                                   state_fe_i==`CORE_FTCH_PRELOAD_2)) |
                                (dbg_halted_i) |
                                (lsu_flag_wr_ex_i &
                                 (state_fe_i==`CORE_FTCH_PRELOAD_2 |
                                  state_fe_i==`CORE_FTCH_INT_RETURN)))));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_br_ilock_ex <= 1'b0;
    else if (instr_br_ilock_we)
      instr_br_ilock_ex <= instr_br_ilock_de;


  assign instr_br_ccpass_de  = (// Conditional or Unconditional in IT
                                (instr_ccpass_de_i & instr_de_br_de) |
                                ((|instr_br_type_ex) & instr_br_pflush_ex_i[0]));


  assign instr_ex_br_blx_de  = ((instr_de_i[15:8] == 8'b01000111) &
                                ((instr_de_i[6:3] != 4'b1110) |
                                 instr_lr_ilock_ex_i));

  assign instr_ex_br_czb_de  = ((instr_de_i[15:12] == 4'b1011) &
                                ~instr_de_i[10] & instr_de_i[8]);

  assign instr_ex_br_isb_de  = ((instr_de_i[15:4] == 12'b111100111011) &
                                (instr_de_i[31:30] == 2'b10) &
                                ~instr_de_i[28] &
                                (instr_de_i[23:20] == 4'b0110));

  assign instr_ex_br_mov_de  = ((instr_de_i[15:8] == 8'b01000110) &
                                instr_de_i[7] & (instr_de_i[2:0] == 3'b111) &
                                ((instr_de_i[6:3] != 4'b1110) |
                                 instr_lr_ilock_ex_i));

  assign instr_ex_br_det_de  = ({3{ instr_valid_de_i         &
                                    ~instr_br_pflush_ex_i[0] &
                                    ~instr_det_erly_br_trk}} &
                                {instr_ex_br_czb_de &  instr_de_i[11],
                                 instr_ex_br_czb_de & ~instr_de_i[11],
                                 instr_ccpass_de_i &
                                 (instr_ex_br_blx_de |
                                  instr_ex_br_isb_de |
                                  instr_ex_br_mov_de)});

  assign instr_ex_br_det_we  = instr_valid_de_i | (|instr_br_type_ex);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_ex_br_det_ex <= 3'b000;
    else if (instr_ex_br_det_we)
      instr_ex_br_det_ex <= instr_ex_br_det_de;

  assign instr_ex_br_imm_a_de = (instr_valid_de_i & instr_ex_br_czb_de ?
                                 {1'b0,instr_de_i[9],instr_de_i[7:3]} :
                                 {7{1'b0}});
  assign instr_ex_br_imm_b_de = (instr_valid_de_i & instr_ex_br_czb_de ?
                                 (instr_use_word_de_i ?
                                  7'b0000010 : 7'b0000001) :
                                 {7{1'b0}});
  assign instr_ex_br_imm_de   = instr_ex_br_imm_a_de - instr_ex_br_imm_b_de;
  assign instr_ex_br_imm_we   = (instr_valid_de_i & instr_ex_br_czb_de &
                                  ~instr_br_pflush_ex_i[0]);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_ex_br_imm_ex <= {7{1'b0}};
    else if (instr_ex_br_imm_we)
      instr_ex_br_imm_ex <= instr_ex_br_imm_de;


  assign instr_ex_br_add_de = ((instr_de_i[15:7] == 9'b010001001) &
                               (instr_de_i[2:0] == 3'b111));

  assign instr_ex_br_pop_de = (instr_de_i[15:8] == 8'b10111101);

  assign instr_ex_br_ldr_de = ((instr_de_i[15:9] == 7'b1111100) &
                               (instr_de_i[6:4] == 3'b101) &
                               ~(instr_de_i[8] &
                                 instr_de_i[6:4] == 3'b101) &
                               ~(instr_de_i[8] & ~instr_de_i[4]) &
                               (instr_de_i[31:28] == 4'b1111) &
                               (// Format 0
                                (instr_de_i[3:0] == 4'b1111) |
                                (instr_de_i[7]) |
                                (~instr_de_i[7] & instr_de_i[27] &
                                 (instr_de_i[26] | instr_de_i[24])) |
                                (~instr_de_i[7] &
                                 (instr_de_i[27:22]==6'b000000))));

  assign instr_ex_br_ldm_de = (instr_de_i[31] & instr_de_i[15:9] == 7'b1110100 &
                               instr_de_i[8] != instr_de_i[7] & ~instr_de_i[6] &
                               instr_de_i[4]);

  assign instr_ex_br_tbl_de = (instr_de_i[15:4] == 12'b111010001101 &
                               ~(|instr_de_i[23:21]));

  assign instr_lsu_br_det_de = (instr_valid_de_i & instr_ccpass_de_i &
                                ~instr_det_erly_br_trk &
                                (instr_ex_br_add_de | instr_ex_br_pop_de |
                                 instr_ex_br_ldr_de | instr_ex_br_ldm_de |
                                 instr_ex_br_tbl_de));

  assign instr_lsu_br_det_we = (instr_lsu_br_det_de |
                                (instr_lsu_br_det_ex &
                                 (// instr_br_pflush_ex_i
                                  (instr_br_pflush_ex_i[1]) |
                                  (state_fe_i==`CORE_FTCH_PREFETCH_1))));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_lsu_br_det_ex <= 1'b0;
    else if (instr_lsu_br_det_we)
      instr_lsu_br_det_ex <= instr_lsu_br_det_de;


  assign instr_br_imm_de = {(// 16 bit B (1)
                             ({31{instr_de_br_b1_de}} &
                              {{23{instr_de_i[7]}},instr_de_i[7:0]})           |
                             ({31{i16_b2_de}} &
                              {{20{instr_de_i[10]}},instr_de_i[10:0]})         |
                             ({31{instr_de_br_b0_de}} &
                              {{11{instr_de_i[10]}},instr_de_i[10],
                               instr_de_i[27],instr_de_i[29],instr_de_i[5:0],
                               instr_de_i[26:16]})                             |
                             ({31{i32_bl_de}} &
                              {{8{instr_de_i[10]}},
                               ~(instr_de_i[29]^instr_de_i[10]),
                              ~(instr_de_i[27]^instr_de_i[10]),
                               instr_de_i[9:0],instr_de_i[26:16]})),
                            1'b0 };

  assign instr_br_imm_ex = {{24{instr_ex_br_imm_ex[6]}},
                            instr_ex_br_imm_ex[6:0],
                            1'b0};

  assign instr_br_type_de   = { // LSU Unconditional Execute
                                instr_ex_br_pop_de | instr_ex_br_tbl_de |
                                instr_ex_br_ldr_de | instr_ex_br_ldm_de,
                                instr_ex_br_add_de | instr_ex_br_blx_de |
                                instr_ex_br_isb_de | instr_ex_br_mov_de,
                                instr_ex_br_czb_de,
                                i16_b2_de  | instr_de_br_blx_de |
                                i32_bl_de  | instr_de_br_mov_de,
                                (instr_de_br_b1_de & ~instr_de_i[7]) |
                                (instr_de_br_b0_de & ~instr_de_i[10]),
                                (instr_de_br_b1_de &  instr_de_i[7]) |
                                (instr_de_br_b0_de &  instr_de_i[10]) };

  assign instr_br_ccinfo_de = { // Unconditional branch in execute
                                ~instr_itblock_de_i & (|instr_br_type_de[5:4]),
                                ~instr_itblock_de_i & instr_br_type_de[2],
                                instr_itblock_de_i &
                                (|instr_br_type_de[5:4]) | instr_br_type_de[3],
                                instr_itblock_de_i &
                                (|instr_br_type_de[2:0]) | instr_br_type_de[1],
                                ~instr_itblock_de_i & instr_br_type_de[0]};

  assign nxt_instr_br_ccpass_ex = (instr_br_de_valid_de &
                                   (|instr_br_ccinfo_de[1:0]) &
                                   instr_br_ccpass_de);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_br_ccpass_ex <= 1'b0;
    else
      instr_br_ccpass_ex <= nxt_instr_br_ccpass_ex;

  assign instr_brchstat_de = ({instr_br_ccpass_ex,
                               {3{instr_br_de_valid_de}} &
                               {instr_br_ccinfo_de[3] | instr_br_ccinfo_de[4],
                                instr_br_ccinfo_de[1] | instr_br_ccinfo_de[2],
                                instr_br_ccinfo_de[0] | instr_br_ccinfo_de[2] |
                                instr_br_ccinfo_de[4] }});

  assign instr_br_type_ex  = {((instr_ex_br_det_ex[2] & ~(&rf_rd_b_zero_ex_i)) |
                               (instr_ex_br_det_ex[1] &  (&rf_rd_b_zero_ex_i))),
                              instr_ex_br_det_ex[0]};

  assign instr_br_req_de   = instr_de_br_det_de | (|instr_br_type_ex);


  assign instr_alu_pflush_de = (instr_br_pflush_de[0] &
                                ~instr_detect_exit_ex_i &
                                ~(mtx_ahb_hreadyi_i & fe_haddracci_i));

  assign instr_alu_pflush_we = (// Set on decode
                                instr_alu_pflush_de |
                                (instr_alu_pflush_ex &
                                 (// Clr on interrupt
                                  (state_fe_i==`CORE_FTCH_PREFETCH_1) |
                                  (instr_detect_exit_ex_i &
                                   ~fault_dblfault_ex_i) |
                                  (mtx_ahb_hreadyi_i & fe_haddracci_i &
                                   ~instr_detect_exit_ex_i &
                                   ~nvic_slp_asleep_i))));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_alu_pflush_ex <= 1'b0;
    else if (instr_alu_pflush_we)
      instr_alu_pflush_ex <= instr_alu_pflush_de;

  assign instr_lsu_pflush_de = ((instr_br_pflush_ex_i[1] & ~dbg_halted_i &
                                 (((state_fe_i==`CORE_FTCH_FETCH) & ~nvic_int_invoke_i) |
                                  (state_fe_i==`CORE_FTCH_INT_ENTRY) |
                                  (state_fe_i==`CORE_FTCH_INT_RETURN))) |
                                ((state_fe_i==`CORE_FTCH_RST_PC) & ~ext_iflush_i));

  assign instr_lsu_pflush_we = (// Set on decode
                                instr_lsu_pflush_de |
                                (instr_lsu_pflush_ex &
                                 (// Clr on interrupt
                                  (state_fe_i==`CORE_FTCH_PREFETCH_1) |
                                  (instr_detect_exit_ex_i &
                                   ~fault_dblfault_ex_i) |
                                  (mtx_ahb_hreadyi_i & fe_haddracci_i &
                                   ~instr_detect_exit_ex_i) |
                                  dbg_halted_i)));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_lsu_pflush_ex <= 1'b0;
    else if (instr_lsu_pflush_we)
      instr_lsu_pflush_ex <= instr_lsu_pflush_de;

  assign instr_pflush_ex = (instr_lsu_pflush_ex |
                            instr_alu_pflush_ex |
                            instr_halt_pflush_ex);

  assign instr_halt_pflush_de = dbg_halted_i & ~instr_halt_pflush_ex;
  assign instr_halt_pflush_we = (instr_halt_pflush_de |
                                 (instr_halt_pflush_ex & ahb_htransi_i[1] &
                                  mtx_ahb_hreadyi_i & fe_haddracci_i));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      instr_halt_pflush_ex <= 1'b0;
    else if (instr_halt_pflush_we)
      instr_halt_pflush_ex <= instr_halt_pflush_de;

  assign instr_br_pflush_fast_de = {(instr_br_pflush_ex_i[1]),
                                    instr_de_br_det_de | (|instr_br_type_ex)};
  assign instr_br_pflush_de      = {(instr_br_pflush_ex_i[1]),
                                    (instr_de_br_det_de & instr_br_ccpass_de) |
                                    ((|instr_br_type_ex) &
                                     instr_br_pflush_ex_i[0])};

  assign rf_pc_wr_data_de = ((instr_br_pflush_fast_de[0] & ~lsu_pflush_ex_i) ?
                             (instr_br_type_ex[0] ?
                              rf_rd_b_data_ex_i[31:1] :
                              (((instr_br_de_valid_de &
                                 (instr_de_br_blx_de | instr_de_br_mov_de)) ?
                                rf_rd_lr_de_i[31:1] :
                                fe_address_add_res_i[31:1]))) :
                             lsu_rf_wr_d_data_ex_i[31:1]);

  assign instr_br_lr_sel_de_o      = instr_br_lr_sel_de;
  assign instr_br_cond_de_o        = instr_br_cond_de;
  assign instr_br_ccpass_de_o      = instr_br_ccpass_de;
  assign instr_br_imm_de_o         = instr_br_imm_de;
  assign instr_br_req_de_o         = instr_br_req_de;
  assign instr_br_pflush_fast_de_o = instr_br_pflush_fast_de;
  assign instr_br_pflush_de_o      = instr_br_pflush_de;
  assign instr_brchstat_de_o       = instr_brchstat_de;
  assign instr_de_br_det_de_o      = instr_de_br_det_de;
  assign instr_br_imm_ex_o         = instr_br_imm_ex;
  assign instr_br_type_ex_o        = instr_br_type_ex;
  assign instr_pflush_ex_o         = instr_pflush_ex;
  assign instr_lsu_br_det_ex_o     = instr_lsu_br_det_ex;
  assign instr_det_erly_br_trk_o   = instr_det_erly_br_trk;
  assign instr_det_erly_br_trk_v_o = instr_det_erly_br_trk_v;
  assign rf_pc_wr_data_de_o        = rf_pc_wr_data_de;


`ifdef ARM_ASSERT_ON

  reg  [1:0]     ovl_br_pflush_ex_reg;      // Reg of instr_br_pflush_ex
  wire           ovl_cond_br_de;            // Cond br in decode
  reg            ovl_cond_br_reg;           // Reg of cond br in decode
  reg            ovl_ccpass_reg;            // CC pass de registered

  assign ovl_cond_br_de = ((instr_brchstat_de == 4'b0001) |
                           (instr_brchstat_de == 4'b0010));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        ovl_br_pflush_ex_reg <= 2'b00;
        ovl_cond_br_reg      <= 1'b0;
        ovl_ccpass_reg       <= 1'b0;
      end
    else
      begin
        ovl_br_pflush_ex_reg <= instr_br_pflush_ex_i[1:0];
        ovl_cond_br_reg      <= ovl_cond_br_de;
        ovl_ccpass_reg       <= instr_ccpass_de_i;
      end

  assert_never #(0,0,"instr_br_pflush_ex_i from LSU/ALU must be exclusive")
    ovl_evict_chk(hclk, hreset_n, (&instr_br_pflush_ex_i[1:0]));

  assert_next #(0,1,0,0,0,"instr_br_type_ex must be pulsed")
    ovl_br_type_ex_pulsed(hclk, hreset_n,
                          (|instr_br_type_ex) & instr_br_ccpass_de,
                          ~(|instr_br_type_ex));

  assert_never #(0,0,"instr_de_br_det_de and instr_br_type_ex must be exclusive")
    ovl_br_det(hclk, hreset_n, (|instr_br_type_ex) & instr_de_br_det_de);

  assert_never #(0,0,"instr_lsu_br_det_ex should not be set under and ALU branch")
    ovl_br_det_2(hclk, hreset_n,
                 (instr_lsu_br_det_ex & instr_br_pflush_ex_i[0]));

  assert_never #(0,0,"instr_br_type_ex has been incorrectly formed")
    ovl_instr_br_type_ex(hclk, hreset_n,
                         (instr_br_type_ex[0] & ~instr_br_pflush_ex_i[0]));

  assert_never #(0,0,
                 "ALU flush cannot be followed immediately by an LSU flush")
    ovl_pflush_1(hclk, hreset_n, (ovl_br_pflush_ex_reg[0] &
                                  instr_br_pflush_de[1] &
                                  ~lsu_pflush_ex_i));

  assert_never #(0,0,"instr_br_pflush_de must be exclusive")
    ovl_pflush_2(hclk, hreset_n, ((&instr_br_pflush_de) & ~lsu_pflush_ex_i));

  assert_zero_one_hot #(0,5,0,"instr_br_ccinfo_de should be exclusive")
    ovl_br_ccinfo_excl(hclk, hreset_n, (instr_br_ccinfo_de &
                                        {5{instr_br_de_valid_de}}));

  assert_zero_one_hot #(0,6,0,"instr_br_type_de should be exclusive.")
    ovl_br_type_de_excl(hclk, hreset_n, (instr_br_type_de &
                                         {6{instr_br_de_valid_de}}));

  assert_never #(0,0,"instr_brchstat_de has been incorrectly formed")
    ovl_instr_brchstat(hclk, hreset_n, instr_brchstat_de[2:0] > 5);

  assert_never_unknown #(0,5,0,"X propagate protect")
    ovl_x_chk(hclk, hreset_n, 1'b1, { instr_br_pflush_fast_de,
                                      instr_br_pflush_de,
                                      instr_det_erly_br_trk });

  assert_never #(0,0,"Cannot CC fail a non-conditional branch")
    ovl_brchstat_chk1(hclk, hreset_n,
                      (instr_brchstat_de == 4'b0100) &
                      ~set_instr_redo_ex_i & ~instr_ccpass_de_i);

  assert_never #(0,0,"If we can CC fail then it must be a conditional branch")
    ovl_brchstat_chk2(hclk, hreset_n,
                      instr_br_de_valid_de & ~instr_ccpass_de_i &
                      instr_de_br_de & ~set_instr_redo_ex_i & ~ovl_cond_br_de);

  assert_never #(0,0,"Cannot take a conditional branch that did not exist")
    ovl_brchstat_chk3(hclk, hreset_n,
                      (instr_brchstat_de == 4'b1000) & ~ovl_cond_br_reg);

  assert_never #(0,0,"Cannot say branch in decode was taken if it was not")
    ovl_brchstat_chk4(hclk, hreset_n,
                      (instr_brchstat_de == 4'b1000) & ~ovl_ccpass_reg);

  assert_never #(0,0,"Should say conditional branch was taken if it was")
    ovl_brchstat_chk5(hclk, hreset_n,
                      ovl_ccpass_reg & ovl_cond_br_reg &
                      (instr_brchstat_de != 4'b1000));

`endif

`include "cm3_dpu_undefs.vp"

endmodule
