`include "cm3_dpu_defs.vp"

module cm3_dpu_fetch_ctl
  (// Inputs
   hreset_n, hclk, ext_iflush_i, instr_ext_pflush_fe_i, instr_br_pflush_ex_i,
   instr_detect_exit_ex_i, instr_dblfault_ex_i, lsu_stackpush_rdy_ex_i,
   nvic_int_invoke_i, nvic_int_newarrival_i, nvic_int_tail_i,
   nvic_int_in_nmihf_i, int_entry_i, int_exit_i, int_return_i,
   int_vector_err_fe_i, mtx_ahb_hreadyi_i, mtx_ahb_hreadyd_i,
   fe_haddracci_i, nvic_slp_asleep_i, nvic_slp_sleeponexit_i, dbg_halted_i,
   state_fe_o, rf_wr_d_en_fe_o, rf_wr_d_addr_fe_o
   );

  input          hreset_n;
  input          hclk;
  input          ext_iflush_i;              // Instruction flush request
  input          instr_ext_pflush_fe_i;     // External event caused pipe flush
  input [2:0]    instr_br_pflush_ex_i;      // Branh in Ex flushes pipe
  input          instr_detect_exit_ex_i;    // Exit dectected
  input          instr_dblfault_ex_i;       // Double fault has occured
  input          lsu_stackpush_rdy_ex_i;    // LSU ready to push stack
  input          nvic_int_invoke_i;         // Interrupt stack request
  input          nvic_int_newarrival_i;     // Interrupt new arrival
  input          nvic_int_tail_i;           // Interrupt tail-chain request
  input [2:0]    nvic_int_in_nmihf_i;       // In NMI/HF ISR
  input          int_entry_i;               // ISR entry
  input          int_exit_i;                // ISR exit
  input          int_return_i;              // Unstacking complete
  input          int_vector_err_fe_i;       // Vector fetch error
  input          mtx_ahb_hreadyi_i;         // AHB I-side HREADY
  input          mtx_ahb_hreadyd_i;         // AHB D-side HREADY
  input          fe_haddracci_i;            // Fetch version of HADDRACCI
  input          nvic_slp_asleep_i;         // DPU asleep
  input          nvic_slp_sleeponexit_i;    // Sleep-on-exit request
  input          dbg_halted_i;              // Core halted

  output [3:0]   state_fe_o;                // Fetch state
  output         rf_wr_d_en_fe_o;           // Fetch RF D port write enable
  output [3:0]   rf_wr_d_addr_fe_o;         // Fetch RF D port write address

  reg            state_we;                  // Fetch state write enable
  reg  [3:0]     nxt_state;                 // Next fetch state
  reg  [3:0]     state;                     // Fetch state
  wire           set_allow_pc_write;        // Set allow PC write
  wire           clr_allow_pc_write;        // Clear allow PC write
  wire           allow_pc_write_we;         // Allow PC write enable
  reg            allow_pc_write;            // Allow fetch to write to PC
  reg            rf_wr_d_en_fe;             // Fetch RF D port write enable
  reg  [3:0]     rf_wr_d_addr_fe;           // Fetch RF D port write address

  always @ (state or ext_iflush_i or mtx_ahb_hreadyi_i or nvic_slp_asleep_i or
            mtx_ahb_hreadyd_i or instr_br_pflush_ex_i or instr_dblfault_ex_i or
            nvic_slp_sleeponexit_i or instr_detect_exit_ex_i or dbg_halted_i or
            nvic_int_invoke_i or nvic_int_tail_i or int_entry_i or int_exit_i or
            int_return_i or nvic_int_newarrival_i or lsu_stackpush_rdy_ex_i or
            nvic_int_in_nmihf_i or int_vector_err_fe_i or
            instr_ext_pflush_fe_i or fe_haddracci_i)
    case (state)
      `CORE_FTCH_RESET : begin
        state_we  = 1'b1;
        nxt_state = `CORE_FTCH_RST_START;
      end
      `CORE_FTCH_RST_START : begin
        state_we  = ~ext_iflush_i & mtx_ahb_hreadyi_i & fe_haddracci_i;
        nxt_state = `CORE_FTCH_RST_LR;
      end
      `CORE_FTCH_RST_LR : begin
        state_we  = ext_iflush_i | (mtx_ahb_hreadyi_i & fe_haddracci_i);
        nxt_state = ext_iflush_i ? `CORE_FTCH_RST_START : `CORE_FTCH_RST_SP;
      end
      `CORE_FTCH_RST_SP : begin
        state_we  = ext_iflush_i | (mtx_ahb_hreadyi_i & fe_haddracci_i);
        nxt_state = ext_iflush_i ? `CORE_FTCH_RST_START : `CORE_FTCH_RST_PC;
      end
      `CORE_FTCH_RST_PC : begin
        state_we  = 1'b1;
        nxt_state = ext_iflush_i ? `CORE_FTCH_RST_START : `CORE_FTCH_FETCH;
      end
      `CORE_FTCH_FETCH : begin
        state_we  = (~dbg_halted_i &
                     ((nvic_int_invoke_i & mtx_ahb_hreadyd_i) |
                      (instr_detect_exit_ex_i & ~instr_dblfault_ex_i)));
        case ({nvic_int_invoke_i | nvic_int_tail_i,
               instr_detect_exit_ex_i & ~instr_dblfault_ex_i})
          2'b11,
          2'b10   : nxt_state = `CORE_FTCH_PREFETCH_1;
          2'b01   : nxt_state = (nvic_slp_sleeponexit_i ? `CORE_FTCH_SLEEP :
                                                          `CORE_FTCH_PRELOAD_1);
          2'b00   : nxt_state = `CORE_FTCH_PREFETCH_1; // unused
          default : nxt_state = 4'bxxxx;
        endcase
      end
      `CORE_FTCH_SLEEP : begin
        state_we  = nvic_int_invoke_i | ~(nvic_slp_asleep_i | int_exit_i);
        nxt_state = (nvic_int_invoke_i ? `CORE_FTCH_PREFETCH_1 :
                                         `CORE_FTCH_PRELOAD_1);
      end
      `CORE_FTCH_PREFETCH_1 : begin
        state_we  = ((mtx_ahb_hreadyi_i & fe_haddracci_i &
                      ~(instr_ext_pflush_fe_i & nvic_int_newarrival_i)) |
                     ~(nvic_int_invoke_i | nvic_int_tail_i));
        nxt_state = ((nvic_int_invoke_i | nvic_int_tail_i) ?
                     `CORE_FTCH_PREFETCH_2 : `CORE_FTCH_PRELOAD_1);
      end
      `CORE_FTCH_PREFETCH_2 : begin
        state_we  = (nvic_int_newarrival_i | ext_iflush_i |
                     ~(nvic_int_invoke_i | nvic_int_tail_i) |
                     (mtx_ahb_hreadyi_i & lsu_stackpush_rdy_ex_i &
                      (nvic_int_in_nmihf_i[2] | ~int_vector_err_fe_i)));
        nxt_state = ((nvic_int_newarrival_i | ext_iflush_i) ?
                     `CORE_FTCH_PREFETCH_1 : `CORE_FTCH_PREFETCH_3);
      end
      `CORE_FTCH_PREFETCH_3 : begin
        state_we  = (nvic_int_newarrival_i | ext_iflush_i |
                     (mtx_ahb_hreadyi_i & fe_haddracci_i));
        nxt_state = (nvic_int_newarrival_i | ext_iflush_i ?
                    `CORE_FTCH_PREFETCH_1 :  `CORE_FTCH_INT_ENTRY);
      end
      `CORE_FTCH_INT_ENTRY: begin
        state_we  = nvic_int_newarrival_i | int_entry_i;
        nxt_state = nvic_int_newarrival_i ? `CORE_FTCH_PREFETCH_1 :
                                            `CORE_FTCH_FETCH;
      end
      `CORE_FTCH_PRELOAD_1: begin
        state_we  = (nvic_int_invoke_i | nvic_int_tail_i |
                    instr_br_pflush_ex_i[1]);
        nxt_state = ((nvic_int_invoke_i | nvic_int_tail_i) ?
                    `CORE_FTCH_PREFETCH_1 : `CORE_FTCH_PRELOAD_2);
      end
      `CORE_FTCH_PRELOAD_2 : begin
        state_we  = ((mtx_ahb_hreadyi_i & fe_haddracci_i &
                      ~ext_iflush_i) |
                     nvic_int_invoke_i);
        nxt_state = nvic_int_invoke_i ? `CORE_FTCH_PREFETCH_1 :
                                        `CORE_FTCH_INT_RETURN;
      end
      `CORE_FTCH_INT_RETURN : begin
        state_we  = nvic_int_invoke_i | int_return_i;
        nxt_state = (nvic_int_invoke_i ? `CORE_FTCH_PREFETCH_1 :
                                         `CORE_FTCH_FETCH);
      end
      default : begin
        state_we  = 1'bx;
        nxt_state = 4'bxxxx;
      end
    endcase

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      state <= `CORE_FTCH_RESET;
    else if (state_we)
      state <= nxt_state;

  assign set_allow_pc_write = (~allow_pc_write &
                               ((state==`CORE_FTCH_RST_LR)     |
                                (state==`CORE_FTCH_INT_ENTRY)));
  assign clr_allow_pc_write = (allow_pc_write &
                               ((state==`CORE_FTCH_RST_START)  |
                                (state==`CORE_FTCH_RST_SP)     |
                                (state==`CORE_FTCH_PREFETCH_1) |
                                int_entry_i));
  assign allow_pc_write_we  = (set_allow_pc_write | clr_allow_pc_write);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      allow_pc_write <= 1'b0;
    else if (allow_pc_write_we)
      allow_pc_write <= set_allow_pc_write;

  always @ (state or allow_pc_write or ext_iflush_i)
    case(state)
      `CORE_FTCH_RST_SP:
        rf_wr_d_en_fe = ~ext_iflush_i &  allow_pc_write;
      `CORE_FTCH_RST_LR:
        rf_wr_d_en_fe = ~ext_iflush_i & ~allow_pc_write;
      `CORE_FTCH_RST_PC:
        rf_wr_d_en_fe = ~ext_iflush_i;
      `CORE_FTCH_INT_ENTRY:
        rf_wr_d_en_fe = ~allow_pc_write;
      `CORE_FTCH_RESET,       `CORE_FTCH_PREFETCH_3,  `CORE_FTCH_PRELOAD_1,
      `CORE_FTCH_RST_START,   `CORE_FTCH_PREFETCH_1,  `CORE_FTCH_PREFETCH_2,
      `CORE_FTCH_PRELOAD_2,   `CORE_FTCH_INT_RETURN,  `CORE_FTCH_FETCH,
      `CORE_FTCH_SLEEP :
        rf_wr_d_en_fe = 1'b0;
      default:
        rf_wr_d_en_fe = 1'bx;
    endcase

  always @ (state)
    case(state)
      `CORE_FTCH_RST_SP:
        rf_wr_d_addr_fe = 4'b1101;
      `CORE_FTCH_RST_LR:
        rf_wr_d_addr_fe = 4'b1110;
      `CORE_FTCH_RST_PC,      `CORE_FTCH_INT_ENTRY,   `CORE_FTCH_RESET,
      `CORE_FTCH_PREFETCH_3,  `CORE_FTCH_PRELOAD_1,   `CORE_FTCH_RST_START,
      `CORE_FTCH_PREFETCH_1,  `CORE_FTCH_PREFETCH_2,  `CORE_FTCH_PRELOAD_2,
      `CORE_FTCH_INT_RETURN,  `CORE_FTCH_FETCH,       `CORE_FTCH_SLEEP :
        rf_wr_d_addr_fe = 4'b1111;
      default:
        rf_wr_d_addr_fe = 4'bxxxx;
    endcase

  assign state_fe_o        = state;
  assign rf_wr_d_en_fe_o   = rf_wr_d_en_fe;
  assign rf_wr_d_addr_fe_o = rf_wr_d_addr_fe;


`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0,4,0,"Invalid state transition - X propogate.")
    ovl_opcode_excl(hclk, hreset_n, 1'b1, state);

  assert_never_unknown #(0,1,0,"Invalid state_we - X propogate.")
    ovl_state_we_xcheck(hclk, hreset_n, 1'b1, state_we);

  assert_never #(0,0,"Execute is out of sync with fetch.")
    ovl_exesync_en(hclk, hreset_n, dbg_halted_i & (state!=`CORE_FTCH_FETCH));

  assert_never #(0,0,"int_tail must never be asserted with int_invoke.")
    ovl_int_tail(hclk, hreset_n, nvic_int_tail_i & nvic_int_invoke_i);

  assert_never #(0,0,"int_newarrival must be asserted with int_invoke.")
    ovl_int_newarrival(hclk, hreset_n, nvic_int_newarrival_i & ~nvic_int_invoke_i);

  assert_never #(0,0,
    "allow_pc_write should not be asserted outside of specified states.")
    ovl_allow_pc_write(hclk, hreset_n,
                       allow_pc_write &
                       ~((state==`CORE_FTCH_RST_START)   |
                         (state==`CORE_FTCH_INT_ENTRY)   |
                         (state==`CORE_FTCH_PREFETCH_1)  |
                         (state==`CORE_FTCH_PREFETCH_3)  |
                         (state==`CORE_FTCH_RST_SP)      |
                         (state==`CORE_FTCH_RST_LR)));

  assert_never #(0,0,"allow_pc_write set/clear should be exclusive")
    ovl_allow_pc_set_clr(hclk, hreset_n,
                         set_allow_pc_write & clr_allow_pc_write);

`endif

`include "cm3_dpu_undefs.vp"

endmodule
