module cm3_dpu_regbank
  (// Inputs
   hreset_n, hclk, instr_valid_de_i, instr_used_word_de_i, rf_rd_a_en_de_i,
   rf_rd_a_addr_de_i, rf_rd_b_en_de_i, rf_rd_b_addr_de_i, rf_sp_rd_de_i,
   rf_wr_d_en_ex_i, rf_wr_d_addr_ex_i, rf_wr_d_data_ex_i, rf_wr_d_data_erly_ex_i,
   rf_sp_wr_ex_i, rf_sp_sel_ex_i, rf_pc_ctl_ex_i, rf_inc_pc_ex_i,
   lsu_fix_pc_ex_i, instr_link_ex_i, fault_lsu_precise_ex_i,
   rf_rd_a_data_de_o, rf_rd_b_data_de_o, rf_rd_lr_de_o, rf_rd_a_data_ex_o,
   rf_rd_a_zero_ex_o, rf_rd_b_data_ex_o, rf_rd_b_zero_ex_o, rf_pc_fwd_ex_o,
   rf_observation_o
   );

  parameter RESET_ALL_REGS  = 0; // Reset all registers if 1
  parameter OBSERVATION     = 0; // Allow internal observation if 1

  input          hreset_n;
  input          hclk;
  input          instr_valid_de_i;         // Instruction in decode is valid
  input          instr_used_word_de_i;     // Consumed 32 bits of instruction
  input          rf_rd_a_en_de_i;          // RF A port read enable
  input [3:0]    rf_rd_a_addr_de_i;        // RF A port read address
  input          rf_rd_b_en_de_i;          // RF B port read enable
  input [3:0]    rf_rd_b_addr_de_i;        // RF B port read address
  input [1:0]    rf_sp_rd_de_i;            // SP read control
  input          rf_wr_d_en_ex_i;          // RF D port write enable
  input [3:0]    rf_wr_d_addr_ex_i;        // RF D port write address
  input [31:0]   rf_wr_d_data_ex_i;        // RF D port write data
  input [31:0]   rf_wr_d_data_erly_ex_i;   // RF D port write data (early)
  input [1:0]    rf_sp_wr_ex_i;            // SP write control
  input [1:0]    rf_sp_sel_ex_i;           // SP selection
  input [1:0]    rf_pc_ctl_ex_i;           // PC control
  input          rf_inc_pc_ex_i;           // Increment PC
  input          lsu_fix_pc_ex_i;          // LSU requires PC fix
  input [1:0]    instr_link_ex_i;          // Link LR to PC
  input          fault_lsu_precise_ex_i;   // Precise LSU fault

  output [31:0]  rf_rd_a_data_de_o;        // RF A port read data
  output [31:0]  rf_rd_b_data_de_o;        // RF B port read data
  output [31:0]  rf_rd_lr_de_o;            // R14
  output [31:0]  rf_rd_a_data_ex_o;        // RF A port read data
  output [1:0]   rf_rd_a_zero_ex_o;        // RF A port read data HW zero
  output [31:0]  rf_rd_b_data_ex_o;        // RF B port read data
  output [1:0]   rf_rd_b_zero_ex_o;        // RF B port read data HW zero
  output [31:0]  rf_pc_fwd_ex_o;           // PC advanced to decode-time
  output [108:0] rf_observation_o;         // Observation of register bank


  wire           full_hreset_n;            // Apply hreset_n to all registers
  wire           opt_obs_en;               // Master observation enable
  wire [3:0]     rd1_addr;                 // Read port 1 address
  wire [3:0]     rd2_addr;                 // Read port 2 address
  wire           wr1_en;                   // Write port 1 enable
  wire [3:0]     wr1_addr;                 // Write port 1 address
  wire [31:0]    wr1_data;                 // Write port 1 data
  wire [31:0]    rd1_data;                 // Read port 1 data
  wire [31:0]    rd2_data;                 // Read port 2 data
  wire [31:0]    r14_data;                 // R14 from register file
  wire [31:0]    rf_rd_lr_de;              // R14 mapped onto LR
  wire [31:0]    rf_rd_a_data_de;          // Decode time port A
  wire           pc_rd_a_de;               // A port read of PC
  wire           rf_rd_a_pc_de;            //  Transitioning PC & ~written
  wire           rf_rd_a_fwd_pc_de;        //  Transitioning PC &  written
  wire           rf_rd_a_stop_pc_de;       // ~Transitioning PC
  wire           rf_rd_a_rd1_de;           // Read from RF read  port 1
  wire           rf_rd_a_fwd_wr1_de;       // Read from RF write port 1
  wire [31:0]    nxt_rf_rd_a_data_ex;      // Next rf_rd_a_data_ex
  reg  [31:0]    rf_rd_a_data_ex;          // Read port A result
  wire [1:0]     rf_rd_a_zero_ex;          // A port read data HW zero detect
  wire [31:0]    rf_rd_b_data_de;          // Decode time port B
  wire           pc_rd_b_de;               // B port read of PC
  wire           rf_rd_b_pc_de;            //  Transitioning PC & ~written
  wire           rf_rd_b_fwd_pc_de;        //  Transitioning PC &  written
  wire           rf_rd_b_stop_pc_de;       // ~Transitioning PC
  wire           rf_rd_b_rd2_de;           // Read from RF read  port 2
  wire           rf_rd_b_fwd_wr1_de;       // Read from RF write port 1
  wire [31:0]    nxt_rf_rd_b_data_ex;      // Next rf_rd_b_data_ex
  reg  [31:0]    rf_rd_b_data_ex;          // Read port B result
  wire [1:0]     rf_rd_b_zero_ex;          // B port read data HW zero detect
  wire           pc_de_2_ex;               // Advance the pipeline
  wire           pc_inc_word_de;           // Increase PC by word versus HW
  wire           pc_inc_word_we;           // pc_inc_word_ex write enable
  reg            pc_inc_word_ex;           // Increase PC by word versus HW
  wire [31:1]    br_adder_res_ex;          // Branch PC result
  wire [31:1]    seq_adder_res_ex;         // Sequential PC result
  wire           instr_pc_wr_ex;           // Instruction writes to PC
  wire [31:0]    nxt_rf_pc_fwd_ex;         // Next forwarded PC
  wire           rf_pc_fwd_we;             // Forwarded PC write enable
  reg  [31:0]    rf_pc_fwd_ex;             // Forwarded PC
  wire [2:0]     pc_fix_sel_ex;            // LSU fix amount select
  wire [3:0]     pc_fix_ctl_ex;            // Select bewteen word and HW
  wire [2:0]     lsu_fix_size_ex;          // LSU PC fix amount
  wire [31:1]    pc_corr_size_ex;          // PC fix amount
  wire [31:0]    rf_pc_ex;                 // Execute time PC value
  reg  [108:0]   rf_observation;           // Observation of register bank

`ifdef ARM_DSM

  wire           dsm_reset_all_regs;
  wire           dsm_observation;

  assign full_hreset_n  = dsm_reset_all_regs   ? hreset_n  : 1'b1;
  assign opt_obs_en     = dsm_observation      ? 1'b1      : 1'b0;

`else

  assign full_hreset_n  = (RESET_ALL_REGS == 1) ? hreset_n : 1'b1;

  assign opt_obs_en     = (OBSERVATION != 0) ? 1'b1        : 1'b0;

`endif


  assign rd1_addr = {rf_rd_a_addr_de_i[3:2],
                     ((rf_rd_a_addr_de_i!=4'b1101) ?
                      rf_rd_a_addr_de_i[1] :
                     (rf_sp_rd_de_i[0] ? rf_sp_rd_de_i[1] : rf_sp_sel_ex_i[1])),
                     rf_rd_a_addr_de_i[0]};

  assign rd2_addr = {rf_rd_b_addr_de_i[3:2],
                     ((rf_rd_b_addr_de_i!=4'b1101) ?
                      rf_rd_b_addr_de_i[1] :
                     (rf_sp_rd_de_i[0] ? rf_sp_rd_de_i[1] : rf_sp_sel_ex_i[1])),
                     rf_rd_b_addr_de_i[0]};

  assign wr1_en   = (rf_wr_d_en_ex_i &
                     (instr_link_ex_i[0] | ~(&rf_wr_d_addr_ex_i)));

  assign wr1_addr = (instr_link_ex_i[0] ?
                     4'b1110 :
                     {rf_wr_d_addr_ex_i[3:2],
                      (rf_wr_d_addr_ex_i!=4'b1101) ?
                      rf_wr_d_addr_ex_i[1] :
                      (rf_sp_wr_ex_i[0] ? rf_sp_wr_ex_i[1] : rf_sp_sel_ex_i[1]),
                      rf_wr_d_addr_ex_i[0]});

  assign wr1_data = (instr_link_ex_i[0] ?
                     {rf_pc_ex[31:1],1'b1} :
                     {rf_wr_d_data_ex_i[31:2],
                      rf_wr_d_data_ex_i[1:0] & {2{rf_wr_d_addr_ex_i!=4'b1101}}});

  cm3_dpu_regfile #(RESET_ALL_REGS)
    u_cm3_dpu_regfile
      (// Inputs
       .hreset_n                   (hreset_n),
       .hclk                       (hclk),
       .rd1_addr_i                 (rd1_addr),
       .rd2_addr_i                 (rd2_addr),
       .wr1_en_i                   (wr1_en),
       .wr1_addr_i                 (wr1_addr),
       .wr1_data_i                 (wr1_data),
       .rd1_data_o                 (rd1_data),
       .rd2_data_o                 (rd2_data),
       .r14_data_o                 (r14_data)
       );

  assign rf_rd_lr_de         = r14_data;


  assign rf_rd_a_data_de     = (&rf_rd_a_addr_de_i) ? rf_pc_fwd_ex : rd1_data;

  assign pc_rd_a_de          = (&rf_rd_a_addr_de_i);

  assign rf_rd_a_pc_de       = pc_rd_a_de &  pc_de_2_ex & ~instr_pc_wr_ex;

  assign rf_rd_a_fwd_pc_de   = pc_rd_a_de &  pc_de_2_ex & instr_pc_wr_ex;

  assign rf_rd_a_stop_pc_de  = pc_rd_a_de & ~pc_de_2_ex;

  assign rf_rd_a_rd1_de      = (~pc_rd_a_de &
                                ~(wr1_en & (rd1_addr==wr1_addr)));

  assign rf_rd_a_fwd_wr1_de  = (~pc_rd_a_de & (wr1_en & (rd1_addr==wr1_addr)));

  assign nxt_rf_rd_a_data_ex =
         (({32{rf_rd_a_pc_de}}      & {rf_pc_fwd_ex[31:1],1'b0})      |
          ({32{rf_rd_a_fwd_pc_de}}  & {rf_wr_d_data_ex_i[31:1],1'b0}) |
          ({32{rf_rd_a_stop_pc_de}} & {rf_pc_ex[31:1],1'b0})          |
          ({32{rf_rd_a_rd1_de}}     & rd1_data)                       |
          ({32{rf_rd_a_fwd_wr1_de}} & wr1_data));

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      rf_rd_a_data_ex <= {32{1'b0}};
    else if (rf_rd_a_en_de_i)
      rf_rd_a_data_ex <= nxt_rf_rd_a_data_ex;

  assign rf_rd_a_zero_ex     = {~(|rf_rd_a_data_ex[31:16]),
                                ~(|rf_rd_a_data_ex[15:0])};


  assign rf_rd_b_data_de     = (&rf_rd_b_addr_de_i) ? rf_pc_fwd_ex : rd2_data;

  assign pc_rd_b_de          = (&rf_rd_b_addr_de_i);

  assign rf_rd_b_pc_de       = pc_rd_b_de &  pc_de_2_ex & ~instr_pc_wr_ex;

  assign rf_rd_b_fwd_pc_de   = pc_rd_b_de &  pc_de_2_ex & instr_pc_wr_ex;

  assign rf_rd_b_stop_pc_de  = pc_rd_b_de & ~pc_de_2_ex;

  assign rf_rd_b_rd2_de      = (~pc_rd_b_de &
                                ~(wr1_en & (rd2_addr==wr1_addr)));

  assign rf_rd_b_fwd_wr1_de  = (~pc_rd_b_de & (wr1_en & (rd2_addr==wr1_addr)));

  assign nxt_rf_rd_b_data_ex =
         (({32{rf_rd_b_pc_de}}      & {rf_pc_fwd_ex[31:1],1'b0})      |
          ({32{rf_rd_b_fwd_pc_de}}  & {rf_wr_d_data_ex_i[31:1],1'b0}) |
          ({32{rf_rd_b_stop_pc_de}} & {rf_pc_ex[31:1],1'b0})          |
          ({32{rf_rd_b_rd2_de}}     & rd2_data)                       |
          ({32{rf_rd_b_fwd_wr1_de}} & wr1_data));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      rf_rd_b_data_ex <= {32{1'b0}};
    else if (rf_rd_b_en_de_i)
      rf_rd_b_data_ex <= nxt_rf_rd_b_data_ex;

  assign rf_rd_b_zero_ex = {~(|rf_rd_b_data_ex[31:16]),
                            ~(|rf_rd_b_data_ex[15:0])};


  assign pc_de_2_ex     = (instr_valid_de_i | rf_inc_pc_ex_i);

  assign pc_inc_word_de = ((instr_pc_wr_ex & ~pc_de_2_ex) |
                           instr_used_word_de_i);
  assign pc_inc_word_we = instr_pc_wr_ex | pc_de_2_ex;

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      pc_inc_word_ex <= 1'b0;
    else if (pc_inc_word_we)
      pc_inc_word_ex <= pc_inc_word_de;



  assign br_adder_res_ex  = rf_wr_d_data_erly_ex_i[31:1] + {{29{1'b0}}, 2'b10};

  assign seq_adder_res_ex = (rf_pc_fwd_ex[31:1] +
                             {{29{1'b0}},
                              ( instr_used_word_de_i),   // +4
                              (~instr_used_word_de_i)}); // +2

  assign instr_pc_wr_ex   = rf_wr_d_en_ex_i & (&rf_wr_d_addr_ex_i);
  assign nxt_rf_pc_fwd_ex = { (instr_pc_wr_ex ? br_adder_res_ex[31:1] :
                                                seq_adder_res_ex[31:1]), 1'b0 };
  assign rf_pc_fwd_we     = (pc_de_2_ex | instr_pc_wr_ex);

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      rf_pc_fwd_ex <= {32{1'b0}};
    else if (rf_pc_fwd_we)
      rf_pc_fwd_ex <= nxt_rf_pc_fwd_ex;


  assign pc_fix_sel_ex    = {// 8
                             (rf_pc_ctl_ex_i[0] &  fault_lsu_precise_ex_i &
                              rf_pc_ctl_ex_i[1]),
                             (rf_pc_ctl_ex_i[0] &  fault_lsu_precise_ex_i &
                              ~rf_pc_ctl_ex_i[1]),
                             (rf_pc_ctl_ex_i[0] & ~fault_lsu_precise_ex_i)};

  assign pc_fix_ctl_ex    = (pc_inc_word_ex ? {pc_fix_sel_ex,1'b0} :
                                              {1'b0,pc_fix_sel_ex});

  assign lsu_fix_size_ex  = {(pc_fix_ctl_ex[3] |
                              pc_fix_ctl_ex[2] |
                              pc_fix_ctl_ex[1]),
                             (pc_fix_ctl_ex[3] |
                              pc_fix_ctl_ex[0] |
                              ~rf_pc_ctl_ex_i[0]),
                             (pc_fix_ctl_ex[2] |
                              pc_fix_ctl_ex[0])};

  assign pc_corr_size_ex = (lsu_fix_pc_ex_i ?
                            {{28{1'b0}}, lsu_fix_size_ex} :
                            {{29{1'b0}},
                              (pc_inc_word_ex | (&instr_link_ex_i)),
                             ~(pc_inc_word_ex | (&instr_link_ex_i))});

  assign rf_pc_ex        = { (rf_pc_fwd_ex[31:1] - pc_corr_size_ex), 1'b0 };

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        rf_observation[108:0]  <= {109{1'b0}};
      end
    else if (opt_obs_en)
      begin
        rf_observation[31:0]   <= wr1_data[31:0];
        rf_observation[32]     <= wr1_en;
        rf_observation[36:33]  <= wr1_addr[3:0];
        rf_observation[40:37]  <= rd1_addr[3:0];
        rf_observation[44:41]  <= rd2_addr[3:0];
        rf_observation[76:45]  <= rd1_data[31:0];
        rf_observation[108:77] <= rd2_data[31:0];
      end

  assign rf_rd_a_data_de_o = rf_rd_a_data_de;
  assign rf_rd_b_data_de_o = rf_rd_b_data_de;
  assign rf_rd_lr_de_o     = rf_rd_lr_de;
  assign rf_rd_a_data_ex_o = rf_rd_a_data_ex;
  assign rf_rd_a_zero_ex_o = rf_rd_a_zero_ex;
  assign rf_rd_b_data_ex_o = rf_rd_b_data_ex;
  assign rf_rd_b_zero_ex_o = rf_rd_b_zero_ex;
  assign rf_pc_fwd_ex_o    = rf_pc_fwd_ex;
  assign rf_observation_o  = opt_obs_en ? rf_observation : {109{1'b0}};


`ifdef ARM_ASSERT_ON

  wire [31:0]    ovl_r13_pr;               // PSP version of r13
  wire [31:0]    ovl_r13_mn;               // MSP version of r13
  wire [31:0]    ovl_sp;                   // Current SP
  wire [31:1]    ovl_pc_add_val_ex;        // Value to add to PC
  wire [31:1]    ovl_nxt_pc_ex;            // Next PC in execute
  reg  [31:0]    ovl_pc_ex;                // PC in execute
  wire           ovl_pc_we;                // PC write enable


  assign ovl_r13_mn        = { u_cm3_dpu_regfile.reg13, 2'b00 };
  assign ovl_r13_pr        = { u_cm3_dpu_regfile.reg15, 2'b00 };
  assign ovl_sp            = ((rf_sp_rd_de_i[0] ?
                               rf_sp_rd_de_i[1] :
                               rf_sp_sel_ex_i[1]) ?
                              {u_cm3_dpu_regfile.reg15,2'b00} :
                              {u_cm3_dpu_regfile.reg13,2'b00});


  assign ovl_pc_add_val_ex = (instr_link_ex_i[0] ?
                              {31{instr_link_ex_i[1]}} :
                              {{29{1'b0}},pc_inc_word_ex,~pc_inc_word_ex});

  assign ovl_nxt_pc_ex     = ((rf_wr_d_en_ex_i & (&rf_wr_d_addr_ex_i)) ?
                              rf_wr_d_data_erly_ex_i[31:1] :
                              (ovl_pc_ex[31:1] + ovl_pc_add_val_ex));

  assign ovl_pc_we         = ((rf_wr_d_en_ex_i & (&rf_wr_d_addr_ex_i)) |
                              pc_de_2_ex);

  always @ (posedge hclk)
    if (ovl_pc_we)
      ovl_pc_ex <= {ovl_nxt_pc_ex,1'b0};

  assert_never #(0,0,"PC write address must match assertion of instr_link_ex")
    ovl_link(hclk, hreset_n, (instr_link_ex_i[0] &
                              (rf_wr_d_addr_ex_i!=4'b1111)));

  assert_never #(0,0,"PC register increment does not match decode time value")
    ovl_pcreg(hclk, hreset_n, (pc_de_2_ex &
                               ~(instr_pc_wr_ex & pc_de_2_ex) &
                               (ovl_nxt_pc_ex[31:1]!=rf_pc_fwd_ex[31:1])));

  assert_never #(0,0,"instr_pc_wr_ex should not be asserted with decode")
    ovl_pc_wr_check(hclk, hreset_n, instr_pc_wr_ex & pc_de_2_ex);

`endif

endmodule
