module cm3_dwt
  (// Inputs
   clk, reset_n, dpu_trace_enable_i, ppb_hready_i, ppb_hsel_i, ppb_haddr_i,
   ppb_hwdata_i, ppb_hwrite_i, mpu_ppb_hresp_i, dpu_ahb_haddrd_i,
   dpu_ahb_htransd_i, dpu_ahb_hsized_i, dpu_ahb_hwrited_i, dpu_ahb_hwdatad_i,
   dpu_ahb_hrdatad_i, mpu_ahb_hreadyoutd_i, dpu_ahb_hrespd_i,
   dpu_ahb_haddraccd_i, mpu_canceld_i, dpu_etm_ivalid_i, dpu_etm_istall_i,
   dpu_etm_cancel_i, dpu_etm_pstat_i, dpu_etm_lsu_i, dpu_etm_fold_i,
   dpu_etm_ia_i, dpu_sleeping_i, dpu_sleep_hold_ack_n, dpu_etm_int_num_i,
   dpu_etm_int_stat_i, itm_dwt_ready_i,
   ppb_hreadyout_o, ppb_hresp_o, ppb_hrdata_o, dwt_size_o, dwt_overflow_o,
   dwt_valid_o, dwt_data_o, dwt_dist_sync_o, dwt_watchpoint_o,
   dwt_etm_trigger_o, dwt_etm_trig_i_not_d_o, dwt_observe_watchpoint_o
   );

  parameter TRACE_LVL      = 1;  // Trace support level. Between 0 and 3.
  parameter DEBUG_LVL      = 3;  // Debug support level. Between 0 and 3.
  parameter RESET_ALL_REGS = 0;  // Reset all registers if 1.
  parameter OBSERVATION    = 0;  // Allow internal observability if 1.
  parameter CP_PRESENT     = 0;  // Affects ID registers

  parameter INC_COMP0 = DEBUG_LVL == 0 ? 0 : 1;
  parameter INC_COMP1 = DEBUG_LVL >  1 ? 1 : 0;
  parameter INC_COMP2 = DEBUG_LVL >  1 ? 1 : 0;
  parameter INC_COMP3 = DEBUG_LVL >  1 ? 1 : 0;


  input         clk;                      // Clock
  input         reset_n;                  // Reset
  input         dpu_trace_enable_i;       // enable tracing
  input         ppb_hready_i;             // Ready (PPB)

  input         ppb_hsel_i;               // DWT select
  input         ppb_hwrite_i;             // PPB write
  input [12:2]  ppb_haddr_i;              // PPB address
  input [31:0]  ppb_hwdata_i;             // PPB write data
  output        ppb_hreadyout_o;          // DWT ready
  output        ppb_hresp_o;              // DWT response
  output [31:0] ppb_hrdata_o;             // DWT read data

  input [31:0]  dpu_ahb_haddrd_i;         // DCORE address
  input [1:0]   dpu_ahb_htransd_i;        // DCORE transaction
  input [1:0]   dpu_ahb_hsized_i;         // DCORE size
  input         dpu_ahb_hwrited_i;        // DCORE write
  input [31:0]  dpu_ahb_hwdatad_i;        // DCORE wdata
  input [31:0]  dpu_ahb_hrdatad_i;        // DCORE rdata
  input         mpu_ahb_hreadyoutd_i;     // DCORE ready
  input [1:0]   dpu_ahb_hrespd_i;         // DCORE response
  input         mpu_ppb_hresp_i;          // MPU response
  input         dpu_ahb_haddraccd_i;      // DCORE accept
  input         mpu_canceld_i;            // MPU cancel D

  input         dpu_etm_ivalid_i;         // Instruction Valid
  input         dpu_etm_istall_i;         // ETM opcode stall
  input         dpu_etm_cancel_i;         // Opcode cancelled
  input [1:0]   dpu_etm_pstat_i;          // Pipeline status
  input         dpu_etm_lsu_i;            // Processing LSU operation
  input         dpu_etm_fold_i;           // Folding an instruction
  input [31:1]  dpu_etm_ia_i;             // PC

  input         dpu_sleeping_i;           // Core is sleeping
  input         dpu_sleep_hold_ack_n;     // Core is in extended sleep

  input [8:0]   dpu_etm_int_num_i;        // Interrupt number
  input [1:0]   dpu_etm_int_stat_i;       // Interrupt status

  input         itm_dwt_ready_i;          // ITM Ready
  output [2:0]  dwt_size_o;               // Size
  output        dwt_overflow_o;           // Overflow
  output        dwt_valid_o;              // Valid
  output [39:0] dwt_data_o;               // Data
  output        dwt_dist_sync_o;          // Synchronising clock

  output        dwt_watchpoint_o;         // Watchpoint
  output [3:0]  dwt_etm_trigger_o;        // ETM Trigger
  output [3:0]  dwt_etm_trig_i_not_d_o;   // ETM triggers on PC match
  output        dwt_observe_watchpoint_o; // Watchpoint observability

  wire          sleep_state;              // Sleeping state

  wire [31:0]   dwt_pc_cycle;             // DWT PC Sampler Cycle Counter
  wire [31:0]   dwt_comp0;                // DWT Comparator 0
  wire [3:0]    dwt_mask0;                // DWT Mask 0
  wire          dwt_pc_sample_match0;     // DWT PS Count Match 0
  wire          dwt_emit_range0;          // DWT Emit Range 0
  wire [3:0]    dwt_function0;            // DWT Function 0

  wire          address_phase;            // Bus transfer address phase
  wire          qualify_data0;            // Data match qualify
  wire          qualify_data2;            // Data match qualify
  wire          qualify_data3;            // Data match qualify

  wire          comp1_good_data;          // Comp1 can attempt a data match
  wire          act_as_data_addr0;        // Comp0 is an addr comp for a data comp
  wire          act_as_data_addr2;        // Comp2 is an addr comp for a data comp
  wire          act_as_data_addr3;        // Comp3 is an addr comp for a data comp

  wire          emit_pc_periodic;         // Emit PCSample

  wire          emit_cyc;                 // Emit PSCYCCNT underflow event
  wire          emit_cpi;                 // Emit CPI Event
  wire          emit_int;                 // Emit Interrupt event
  wire          emit_sleep;               // Emit Sleep event
  wire          emit_fold;                // Emit fold event
  wire          emit_lsu;                 // Emit LSU event
  wire          int_trace_en;             // Enable interrupt trace packets

  wire [3:0]    etm_trigger;              // Emit ETM Trigger - bit per comparator
  wire [3:0]    emit_watchpoint;          // Emit Watchpoint - bit per comparator
  wire [3:0]    emit_pc_sample;           // Emit Sample PC - bit per comparator
  wire [3:0]    emit_data_sample;         // Emit Sample Data - bit per comparator

  wire          clear_etm_ivalid;         // Clear etm_ivalid_reg
  wire          allow_pc_match;           // Allow the PC to be compared

  wire          dpu_ahb_hwrite_reg;       // dpu_ahb_hwrited_i captured
  wire [1:0]    dpu_ahb_hsize_reg;        // dpu_ahb_hsized_i captured
  wire [1:0]    dpu_ahb_haddr_reg;        // dpu_ahb_haddrd_i[15:0] captured
  wire [31:0]   dpu_ahb_haddrd_masked;    // Masked to 0 when trace not enabled
  wire [31:0]   watchpoint_data;          // WPdata - mux of read and write data

  wire [31:0]   dwt_comp1;                // DWT Comparator 1
  wire [3:0]    dwt_mask1;                // DWT Mask 1
  wire          dwt_emit_range1;          // DWT Emit Range 1
  wire [1:0]    dwt_match_size1;          // DWT Data Size 1
  wire          dwt_match_on_data1;       // DWT Data Match 1
  wire [3:0]    dwt_function1;            // DWT Function 1
  wire [31:0]   dwt_comp2;                // DWT Comparator 2
  wire [3:0]    dwt_mask2;                // DWT Mask 2
  wire          dwt_emit_range2;          // DWT Emit Range 2
  wire [3:0]    dwt_function2;            // DWT Function 2
  wire [31:0]   dwt_comp3;                // DWT Comparator 3
  wire [3:0]    dwt_mask3;                // DWT Mask 3
  wire          dwt_emit_range3;          // DWT Emit Range 3
  wire [3:0]    dwt_function3;            // DWT Function 3

  wire [2:0]    dwt_size;
  wire          dwt_overflow;
  wire          dwt_valid;
  wire [39:0]   dwt_data;
  wire [3:0]    emit_etm_trigger;
  wire [3:0]    iemit_etm_trigger;
  wire [3:0]    ietm_trigger;
  wire [3:0]    iemit_watchpoint;
  wire [3:0]    iemit_pc_sample;
  wire [3:0]    iemit_data_sample;
  wire [3:0]    ietm_trig_i_not_d;
  wire [3:0]    etm_trig_i_not_d;
  wire          opt_qualify_data2;
  wire          opt_qualify_data3;
  wire          opt_trace_en;
  wire          opt_full_watchpoint_en;
  wire [31:0]   ppb_hrdata;
  wire          dwt_dist_sync;
  wire          dwt_watchpoint;
  wire          dwt_observe_watchpoint;
  wire          opt_ppb_hreadyout;
  wire          opt_ppb_hresp;

`ifdef ARM_DSM

  wire          dsm_observation;
  wire [1:0]    dsm_trace_lvl;
  wire [1:0]    dsm_debug_lvl;
  wire          dsm_reset_all_regs;

  assign opt_full_watchpoint_en = (dsm_debug_lvl > 1)  ? 1'b1 : 1'b0;
`else
  assign opt_full_watchpoint_en = (DEBUG_LVL > 1)      ? 1'b1 : 1'b0;
`endif


  cm3_dwt_apb_if  #(TRACE_LVL, DEBUG_LVL, RESET_ALL_REGS, OBSERVATION, CP_PRESENT)
    u_dwt_apb_if
      (// Inputs
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .ppb_hready_i               (ppb_hready_i),
       .ppb_hsel_i                 (ppb_hsel_i),
       .ppb_hwrite_i               (ppb_hwrite_i),
       .ppb_haddr_i                (ppb_haddr_i[12:2]),
       .ppb_hwdata_i               (ppb_hwdata_i[31:0]),
       .emit_etm_trigger_i         (emit_etm_trigger),
       .emit_watchpoint_i          (emit_watchpoint),
       .emit_pc_sample_i           (emit_pc_sample),
       .emit_data_sample_i         (emit_data_sample),
       .dpu_ahb_haddrd_i           (dpu_ahb_haddrd_i),
       .dpu_ahb_htransd1_i         (dpu_ahb_htransd_i[1]),
       .dpu_ahb_hsized_i           (dpu_ahb_hsized_i[1:0]),
       .dpu_ahb_hwrited_i          (dpu_ahb_hwrited_i),
       .dpu_ahb_hwdatad_i          (dpu_ahb_hwdatad_i[31:0]),
       .dpu_ahb_hrdatad_i          (dpu_ahb_hrdatad_i[31:0]),
       .mpu_ahb_hreadyoutd_i       (mpu_ahb_hreadyoutd_i),
       .dpu_ahb_haddraccd_i        (dpu_ahb_haddraccd_i),
       .mpu_canceld_i              (mpu_canceld_i),
       .dpu_etm_ivalid_i           (dpu_etm_ivalid_i),
       .dpu_etm_istall_i           (dpu_etm_istall_i),
       .dpu_etm_cancel_i           (dpu_etm_cancel_i),
       .dpu_etm_pstat_i            (dpu_etm_pstat_i[1:0]),
       .dpu_etm_lsu_i              (dpu_etm_lsu_i),
       .dpu_etm_fold_i             (dpu_etm_fold_i),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .dpu_sleeping_i             (dpu_sleeping_i),
       .dpu_sleep_hold_ack_n_i     (dpu_sleep_hold_ack_n),
       .qualify_data0_i            (qualify_data0),
       .qualify_data2_i            (qualify_data2),
       .qualify_data3_i            (qualify_data3),
       .ppb_hreadyout_o            (opt_ppb_hreadyout),
       .ppb_hresp_o                (opt_ppb_hresp),
       .ppb_hrdata_o               (ppb_hrdata),
       .dpu_ahb_hwrite_reg_o       (dpu_ahb_hwrite_reg),
       .dpu_ahb_haddr_reg_o        (dpu_ahb_haddr_reg),
       .dpu_ahb_haddrd_masked_o    (dpu_ahb_haddrd_masked),
       .dpu_ahb_hsize_reg_o        (dpu_ahb_hsize_reg),
       .dwt_dist_sync_o            (dwt_dist_sync),
       .dwt_watchpoint_o           (dwt_watchpoint),
       .dwt_observe_watchpoint_o   (dwt_observe_watchpoint),
       .clear_etm_ivalid_o         (clear_etm_ivalid),
       .allow_pc_match_o           (allow_pc_match),
       .address_phase_o            (address_phase),
       .dwt_pc_cycle_o             (dwt_pc_cycle),
       .dwt_pc_sample_match0_o     (dwt_pc_sample_match0),
       .act_as_data_addr0_o        (act_as_data_addr0),
       .dwt_emit_range0_o          (dwt_emit_range0),
       .dwt_mask0_o                (dwt_mask0),
       .dwt_function0_o            (dwt_function0),
       .dwt_comp0_o                (dwt_comp0),
       .dwt_match_on_data1_o       (dwt_match_on_data1),
       .dwt_match_size1_o          (dwt_match_size1),
       .dwt_emit_range1_o          (dwt_emit_range1),
       .dwt_mask1_o                (dwt_mask1),
       .dwt_function1_o            (dwt_function1),
       .dwt_comp1_o                (dwt_comp1),
       .act_as_data_addr2_o        (act_as_data_addr2),
       .dwt_emit_range2_o          (dwt_emit_range2),
       .dwt_mask2_o                (dwt_mask2),
       .dwt_function2_o            (dwt_function2),
       .dwt_comp2_o                (dwt_comp2),
       .act_as_data_addr3_o        (act_as_data_addr3),
       .dwt_emit_range3_o          (dwt_emit_range3),
       .dwt_mask3_o                (dwt_mask3),
       .dwt_function3_o            (dwt_function3),
       .dwt_comp3_o                (dwt_comp3),
       .comp1_good_data_o          (comp1_good_data),
       .sleep_state_o              (sleep_state),
       .emit_pc_periodic_o         (emit_pc_periodic),
       .emit_cyc_o                 (emit_cyc),
       .emit_cpi_o                 (emit_cpi),
       .emit_int_o                 (emit_int),
       .emit_sleep_o               (emit_sleep),
       .emit_fold_o                (emit_fold),
       .emit_lsu_o                 (emit_lsu),
       .int_trace_en_o             (int_trace_en),
       .watchpoint_data_o          (watchpoint_data)
       );


  cm3_dwt_comp #(TRACE_LVL, DEBUG_LVL, INC_COMP0)
    u_cm3_dwt_comp0
      (// Inputs
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .dwt_pc_sample_match_i      (dwt_pc_sample_match0),
       .dwt_match_on_data_i        (1'b0),
       .dwt_match_size_i           (2'b00),
       .act_as_data_addr_i         (act_as_data_addr0),
       .dwt_emit_range_i           (dwt_emit_range0),
       .dwt_mask_i                 (dwt_mask0),
       .dwt_function_i             (dwt_function0),
       .dwt_comp_i                 (dwt_comp0),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .etm_ivalid_i               (dpu_etm_ivalid_i),
       .clear_etm_ivalid_i         (clear_etm_ivalid),
       .allow_pc_match_i           (allow_pc_match),
       .address_phase_i            (address_phase),
       .dpu_ahb_haddrd_masked_i    (dpu_ahb_haddrd_masked),
       .dpu_ahb_htransd_i          (dpu_ahb_htransd_i),
       .dpu_ahb_hwrited_i          (dpu_ahb_hwrited_i),
       .dpu_ahb_hsized_i           (dpu_ahb_hsized_i),
       .dpu_ahb_hsize_reg_i        (2'b00),
       .dpu_ahb_haddr_reg_i        (2'b00),
       .dpu_ahb_hready_i           (mpu_ahb_hreadyoutd_i),
       .dpu_ahb_hresp_i            (dpu_ahb_hrespd_i),
       .mpu_hresp_i                (mpu_ppb_hresp_i),
       .dwt_pc_cycle_i             (dwt_pc_cycle),
       .watchpoint_data_i          (watchpoint_data),
       .comp_good_data_i           (1'b0),
       .dpu_ahb_hwrite_reg_i       (1'b0),
       .qualify_data_o             (qualify_data0),
       .emit_etm_trigger_o         (iemit_etm_trigger[0]),
       .etm_trigger_o              (ietm_trigger[0]),
       .emit_watchpoint_o          (iemit_watchpoint[0]),
       .emit_pc_sample_o           (iemit_pc_sample[0]),
       .emit_data_sample_o         (iemit_data_sample[0]),
       .etm_trig_i_not_d_o         (ietm_trig_i_not_d[0])
       );

  cm3_dwt_comp #(TRACE_LVL, DEBUG_LVL, INC_COMP1)
    u_cm3_dwt_comp1
      (
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .dwt_pc_sample_match_i      (1'b0),
       .dwt_match_on_data_i        (dwt_match_on_data1),
       .dwt_match_size_i           (dwt_match_size1),
       .act_as_data_addr_i         (1'b0),
       .dwt_emit_range_i           (dwt_emit_range1),
       .dwt_mask_i                 (dwt_mask1),
       .dwt_function_i             (dwt_function1),
       .dwt_comp_i                 (dwt_comp1),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .etm_ivalid_i               (dpu_etm_ivalid_i),
       .clear_etm_ivalid_i         (clear_etm_ivalid),
       .allow_pc_match_i           (allow_pc_match),
       .address_phase_i            (address_phase),
       .dpu_ahb_haddrd_masked_i    (dpu_ahb_haddrd_masked),
       .dpu_ahb_htransd_i          (dpu_ahb_htransd_i),
       .dpu_ahb_hwrited_i          (dpu_ahb_hwrited_i),
       .dpu_ahb_hsized_i           (dpu_ahb_hsized_i),
       .dpu_ahb_hsize_reg_i        (dpu_ahb_hsize_reg),
       .dpu_ahb_haddr_reg_i        (dpu_ahb_haddr_reg),
       .dpu_ahb_hready_i           (mpu_ahb_hreadyoutd_i),
       .dpu_ahb_hresp_i            (dpu_ahb_hrespd_i),
       .mpu_hresp_i                (mpu_ppb_hresp_i),
       .dwt_pc_cycle_i             ({32{1'b0}}),
       .watchpoint_data_i          (watchpoint_data),
       .comp_good_data_i           (comp1_good_data),
       .dpu_ahb_hwrite_reg_i       (dpu_ahb_hwrite_reg),
       .qualify_data_o             (/*Unused*/),
       .emit_etm_trigger_o         (iemit_etm_trigger[1]),
       .etm_trigger_o              (ietm_trigger[1]),
       .emit_watchpoint_o          (iemit_watchpoint[1]),
       .emit_pc_sample_o           (iemit_pc_sample[1]),
       .emit_data_sample_o         (iemit_data_sample[1]),
       .etm_trig_i_not_d_o         (ietm_trig_i_not_d[1])
       );

  cm3_dwt_comp #(TRACE_LVL, DEBUG_LVL, INC_COMP2)
    u_cm3_dwt_comp2
      (// Inputs
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .dwt_pc_sample_match_i      (1'b0),
       .dwt_match_on_data_i        (1'b0),
       .dwt_match_size_i           (2'b00),
       .act_as_data_addr_i         (act_as_data_addr2),
       .dwt_emit_range_i           (dwt_emit_range2),
       .dwt_mask_i                 (dwt_mask2),
       .dwt_function_i             (dwt_function2),
       .dwt_comp_i                 (dwt_comp2),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .etm_ivalid_i               (dpu_etm_ivalid_i),
       .clear_etm_ivalid_i         (clear_etm_ivalid),
       .allow_pc_match_i           (allow_pc_match),
       .address_phase_i            (address_phase),
       .dpu_ahb_haddrd_masked_i    (dpu_ahb_haddrd_masked),
       .dpu_ahb_htransd_i          (dpu_ahb_htransd_i),
       .dpu_ahb_hwrited_i          (dpu_ahb_hwrited_i),
       .dpu_ahb_hsized_i           (dpu_ahb_hsized_i),
       .dpu_ahb_hsize_reg_i        (2'b00),
       .dpu_ahb_haddr_reg_i        (2'b00),
       .dpu_ahb_hready_i           (mpu_ahb_hreadyoutd_i),
       .dpu_ahb_hresp_i            (dpu_ahb_hrespd_i),
       .mpu_hresp_i                (mpu_ppb_hresp_i),
       .dwt_pc_cycle_i             ({32{1'b0}}),
       .watchpoint_data_i          (watchpoint_data),
       .comp_good_data_i           (1'b0),
       .dpu_ahb_hwrite_reg_i       (1'b0),
       .qualify_data_o             (opt_qualify_data2),
       .emit_etm_trigger_o         (iemit_etm_trigger[2]),
       .etm_trigger_o              (ietm_trigger[2]),
       .emit_watchpoint_o          (iemit_watchpoint[2]),
       .emit_pc_sample_o           (iemit_pc_sample[2]),
       .emit_data_sample_o         (iemit_data_sample[2]),
       .etm_trig_i_not_d_o         (ietm_trig_i_not_d[2])
       );

  cm3_dwt_comp #(TRACE_LVL, DEBUG_LVL, INC_COMP3)
    u_cm3_dwt_comp3
      (// Inputs
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .dwt_pc_sample_match_i      (1'b0),
       .dwt_match_on_data_i        (1'b0),
       .dwt_match_size_i           (2'b00),
       .act_as_data_addr_i         (act_as_data_addr3),
       .dwt_emit_range_i           (dwt_emit_range3),
       .dwt_mask_i                 (dwt_mask3),
       .dwt_function_i             (dwt_function3),
       .dwt_comp_i                 (dwt_comp3),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .etm_ivalid_i               (dpu_etm_ivalid_i),
       .clear_etm_ivalid_i         (clear_etm_ivalid),
       .allow_pc_match_i           (allow_pc_match),
       .address_phase_i            (address_phase),
       .dpu_ahb_haddrd_masked_i    (dpu_ahb_haddrd_masked),
       .dpu_ahb_htransd_i          (dpu_ahb_htransd_i),
       .dpu_ahb_hwrited_i          (dpu_ahb_hwrited_i),
       .dpu_ahb_hsized_i           (dpu_ahb_hsized_i),
       .dpu_ahb_hsize_reg_i        (2'b00),
       .dpu_ahb_haddr_reg_i        (2'b00),
       .dpu_ahb_hready_i           (mpu_ahb_hreadyoutd_i),
       .dpu_ahb_hresp_i            (dpu_ahb_hrespd_i),
       .mpu_hresp_i                (mpu_ppb_hresp_i),
       .dwt_pc_cycle_i             ({32{1'b0}}),
       .watchpoint_data_i          (watchpoint_data),
       .comp_good_data_i           (1'b0),
       .dpu_ahb_hwrite_reg_i       (1'b0),
       .qualify_data_o             (opt_qualify_data3),
       .emit_etm_trigger_o         (iemit_etm_trigger[3]),
       .etm_trigger_o              (ietm_trigger[3]),
       .emit_watchpoint_o          (iemit_watchpoint[3]),
       .emit_pc_sample_o           (iemit_pc_sample[3]),
       .emit_data_sample_o         (iemit_data_sample[3]),
       .etm_trig_i_not_d_o         (ietm_trig_i_not_d[3])
       );

  assign emit_watchpoint[3:1] = opt_full_watchpoint_en ? iemit_watchpoint[3:1] : 3'b000;
  assign emit_watchpoint[0]   = iemit_watchpoint[0];
  assign qualify_data2        = opt_full_watchpoint_en ? opt_qualify_data2 : 1'b0;
  assign qualify_data3        = opt_full_watchpoint_en ? opt_qualify_data3 : 1'b0;

  assign etm_trigger[0]      = opt_trace_en ? ietm_trigger[0]         : 1'b0;
  assign emit_etm_trigger[0] = opt_trace_en ? iemit_etm_trigger[0]    : 1'b0;
  assign emit_pc_sample[0]   = opt_trace_en ? iemit_pc_sample[0]      : 1'b0;
  assign emit_data_sample[0] = opt_trace_en ? iemit_data_sample[0]    : 1'b0;
  assign etm_trig_i_not_d[0] = opt_trace_en ? ietm_trig_i_not_d[0] : 1'b0;

  assign etm_trigger[3:1]      = (opt_trace_en & opt_full_watchpoint_en) ? ietm_trigger[3:1]      : 3'b000;
  assign emit_etm_trigger[3:1] = (opt_trace_en & opt_full_watchpoint_en) ? iemit_etm_trigger[3:1] : 3'b000;
  assign emit_pc_sample[3:1]   = (opt_trace_en & opt_full_watchpoint_en) ? iemit_pc_sample[3:1]   : 3'b000;
  assign emit_data_sample[3:1] = (opt_trace_en & opt_full_watchpoint_en) ? iemit_data_sample[3:1] : 3'b000;
  assign etm_trig_i_not_d[3:1] = (opt_trace_en & opt_full_watchpoint_en) ? ietm_trig_i_not_d[3:1] : 3'b000;

  cm3_dwt_packet_gen #(TRACE_LVL, RESET_ALL_REGS)
    u_cm3_dwt_packet_gen
      (// Inputs
       .clk                        (clk),
       .reset_n                    (reset_n),
       .dpu_trace_enable_i         (dpu_trace_enable_i),
       .emit_data_sample_i         (emit_data_sample),
       .emit_pc_sample_i           (emit_pc_sample),
       .dwt_emit_range0_i          (dwt_emit_range0),
       .dwt_emit_range1_i          (dwt_emit_range1),
       .dwt_emit_range2_i          (dwt_emit_range2),
       .dwt_emit_range3_i          (dwt_emit_range3),
       .dpu_ahb_haddrd_masked_i    (dpu_ahb_haddrd_masked[15:0]),
       .itm_dwt_ready_i            (itm_dwt_ready_i),
       .dpu_etm_int_num_i          (dpu_etm_int_num_i),
       .dpu_etm_int_stat_i         (dpu_etm_int_stat_i),
       .dpu_etm_ivalid_i           (dpu_etm_ivalid_i),
       .dpu_etm_ia_i               (dpu_etm_ia_i),
       .sleeping_i                 (sleep_state),
       .emit_pc_periodic_i         (emit_pc_periodic),
       .emit_cyc_i                 (emit_cyc),
       .emit_cpi_i                 (emit_cpi),
       .emit_int_i                 (emit_int),
       .emit_sleep_i               (emit_sleep),
       .emit_fold_i                (emit_fold),
       .emit_lsu_i                 (emit_lsu),
       .int_trace_en_i             (int_trace_en),
       .dpu_ahb_hwrite_reg_i       (dpu_ahb_hwrite_reg),
       .dpu_ahb_hsize_reg_i        (dpu_ahb_hsize_reg),
       .dpu_ahb_haddr_reg_i        (dpu_ahb_haddr_reg),
       .watchpoint_data_i          (watchpoint_data),
       .dwt_valid_o                (dwt_valid),
       .dwt_size_o                 (dwt_size),
       .dwt_overflow_o             (dwt_overflow),
       .dwt_data_o                 (dwt_data)
       );

`ifdef ARM_DSM
  assign dwt_etm_trigger_o        = (dsm_trace_lvl > 0)    ? etm_trigger            : 4'b0000;
  assign dwt_etm_trig_i_not_d_o   = (dsm_trace_lvl > 0)    ? etm_trig_i_not_d       : 4'b0000;
  assign dwt_valid_o              = (dsm_trace_lvl > 0)    ? dwt_valid              : 1'b0;
  assign dwt_size_o               = (dsm_trace_lvl > 0)    ? dwt_size               : 3'b000;
  assign dwt_overflow_o           = (dsm_trace_lvl > 0)    ? dwt_overflow           : 1'b0;
  assign dwt_data_o               = (dsm_trace_lvl > 0)    ? dwt_data               : {40{1'b0}};
  assign opt_trace_en             = (dsm_trace_lvl > 0)    ? 1'b1                   : 1'b0;
  assign ppb_hrdata_o             = (dsm_trace_lvl > 0 |
                                     dsm_debug_lvl > 0)    ? ppb_hrdata             : {32{1'b0}};
  assign dwt_dist_sync_o          = (dsm_trace_lvl > 0 |
                                     dsm_debug_lvl > 0)    ? dwt_dist_sync          : 1'b0;
  assign dwt_watchpoint_o         = (dsm_trace_lvl > 0 |
                                     dsm_debug_lvl > 0)    ? dwt_watchpoint         : 1'b0;
  assign dwt_observe_watchpoint_o = ((dsm_trace_lvl > 0 |
                                      dsm_debug_lvl > 0) &
                                     dsm_observation != 0) ? dwt_observe_watchpoint : 1'b0;
  assign ppb_hreadyout_o          = (dsm_trace_lvl > 0 |
                                     dsm_debug_lvl > 0)    ? opt_ppb_hreadyout      : 1'b0;
  assign ppb_hresp_o              = (dsm_trace_lvl > 0 |
                                     dsm_debug_lvl > 0)    ? opt_ppb_hresp          : 1'b0;
`else

  assign dwt_etm_trigger_o        = (TRACE_LVL > 0)        ? etm_trigger            : 4'b0000;
  assign dwt_etm_trig_i_not_d_o   = (TRACE_LVL > 0)        ? etm_trig_i_not_d       : 4'b0000;
  assign dwt_valid_o              = (TRACE_LVL > 0)        ? dwt_valid              : 1'b0;
  assign dwt_size_o               = (TRACE_LVL > 0)        ? dwt_size               : 3'b000;
  assign dwt_overflow_o           = (TRACE_LVL > 0)        ? dwt_overflow           : 1'b0;
  assign dwt_data_o               = (TRACE_LVL > 0)        ? dwt_data               : {40{1'b0}};
  assign opt_trace_en             = (TRACE_LVL > 0)        ? 1'b1                   : 1'b0;
  assign ppb_hrdata_o             = (TRACE_LVL > 0 |
                                     DEBUG_LVL > 0)        ? ppb_hrdata             : {32{1'b0}};
  assign dwt_dist_sync_o          = (TRACE_LVL > 0 |
                                     DEBUG_LVL > 0)        ? dwt_dist_sync          : 1'b0;
  assign dwt_watchpoint_o         = (TRACE_LVL > 0 |
                                     DEBUG_LVL > 0)        ? dwt_watchpoint         : 1'b0;
  assign dwt_observe_watchpoint_o = ((TRACE_LVL > 0 |
                                      DEBUG_LVL > 0) &
                                     OBSERVATION != 0)     ? dwt_observe_watchpoint : 1'b0;
  assign ppb_hreadyout_o          = (TRACE_LVL > 0 |
                                     DEBUG_LVL > 0)        ? opt_ppb_hreadyout      : 1'b1;
  assign ppb_hresp_o              = (TRACE_LVL > 0 |
                                     DEBUG_LVL > 0)        ? opt_ppb_hresp          : 1'b0;

`endif

endmodule
