`include "cm3_dwt_defs.vp"

module cm3_dwt_comp
  (// Inputs
   clk, reset_n, dpu_trace_enable_i, dwt_pc_sample_match_i, dwt_match_on_data_i,
   dwt_match_size_i, act_as_data_addr_i, dwt_emit_range_i, dwt_function_i,
   comp_good_data_i, dpu_ahb_hwrite_reg_i, dwt_comp_i, dwt_mask_i, dpu_etm_ia_i,
   etm_ivalid_i, clear_etm_ivalid_i, allow_pc_match_i, address_phase_i,
   dpu_ahb_haddrd_masked_i, dpu_ahb_htransd_i, dpu_ahb_hwrited_i,
   dpu_ahb_hsized_i, dpu_ahb_haddr_reg_i, dpu_ahb_hsize_reg_i, dpu_ahb_hready_i,
   dpu_ahb_hresp_i, mpu_hresp_i, dwt_pc_cycle_i, watchpoint_data_i,
   qualify_data_o, emit_etm_trigger_o, etm_trigger_o, emit_watchpoint_o,
   emit_pc_sample_o, emit_data_sample_o, etm_trig_i_not_d_o
   );

  parameter TRACE_LVL = 1;
  parameter DEBUG_LVL = 3;
  parameter INC_COMP  = 1;


  input         clk;                      // Clock
  input         reset_n;                  // Reset

  input         dpu_trace_enable_i;       // Enable comparators
  input         dwt_pc_sample_match_i;    // match_result against cycle count value
  input         dwt_match_on_data_i;      // match_result against data value
  input [1:0]   dwt_match_size_i;         // Data size to match against
  input         act_as_data_addr_i;       // Act as address qualifier for data match
  input         dwt_emit_range_i;         // Emit 16 LSB's of matched address
  input [3:0]   dwt_function_i;           // Comparator function on match
  input         comp_good_data_i;         // Qualify for data match
  input         dpu_ahb_hwrite_reg_i;     // Perform data match on write

  input [31:0]  dwt_comp_i;               // Comparator value
  input [3:0]   dwt_mask_i;               // Address Mask

  input [31:1]  dpu_etm_ia_i;             // PC address for comparison
  input         etm_ivalid_i;             // PC address valid
  input         clear_etm_ivalid_i;       // Clear etm_ivalid register
  input         allow_pc_match_i;         // Allow PC matching based in ETM i/f

  input         address_phase_i;          // Data transaction
  input [31:0]  dpu_ahb_haddrd_masked_i;  // DCore address
  input [1:0]   dpu_ahb_htransd_i;        // DCore transaction
  input         dpu_ahb_hwrited_i;        // DCore write
  input [1:0]   dpu_ahb_hsized_i;         // DCore size
  input [1:0]   dpu_ahb_hsize_reg_i;      // DCore size registered
  input [1:0]   dpu_ahb_haddr_reg_i;      // DCore addr[1:0] registered
  input         dpu_ahb_hready_i;         // DCore ready
  input [1:0]   dpu_ahb_hresp_i;          // DCore response
  input         mpu_hresp_i;              // MPU response

  input [31:0]  dwt_pc_cycle_i;           // Cycle counter value
  input [31:0]  watchpoint_data_i;        // read or write data

  output        qualify_data_o;           // Qualify for data match
  output        etm_trig_i_not_d_o;       // ETM trigger on PC match
  output        emit_etm_trigger_o;       // ETM Trigger set
  output        etm_trigger_o;            // ETM Trigger
  output        emit_watchpoint_o;        // Watchpoint
  output        emit_pc_sample_o;         // fn_sample PC
  output        emit_data_sample_o;       // fn_sample data

  wire          fn_sample;                // fn_sample on match
  wire          fn_extended_sample;       // Extended fn_sample on match
  wire          fn_sample_pc;             // fn_sample PC on match
  wire          fn_sample_data;           // fn_sample data on match
  wire          fn_watchpoint;            // Watchpoint on match
  wire          fn_etm_trigger;           // ETM Trigger on match
  wire          fn_compare_pc;            // match_result against PC rather than data addr
  wire          fn_on_read;               // Action on data read
  wire          fn_on_write;              // Action on data write
  wire [31:0]   pc_addr_masked;           // Masked PC value
  wire          pc_match;                 // Potential PC match has occurred
  reg           pc_match_reg;             // Registered pc_match
  wire          pc_match_we;              // pc_match_reg write enable
  wire          nxt_pc_match_reg;         // pc_match masked with enable

  wire          match_result;             // Comparator match_result

  wire          match_data_sample_we;     // Enable emit_data_sample_o register
  reg           match_data_sample;        // Send data sample next cycle
  wire          match_data_sample_nxt;
  wire          emit_data_sample;

  wire [31:0]   data_to_match;            // Data to be matched for data match mode
  reg  [31:0]   swiz_data_to_match;       // Swizzled data_to_match
  wire [31:0]   dwt_comp_masked;          // dwt_comp_i masked with dwt_mask_i and LSBs
  reg  [31:0]   compare_value;            // Address for comparison
  wire [31:0]   compare_value_masked;     // compare_value masked by dwt_mask_i
  wire          address_comparator_en;    // Enable address comparison
  wire          value_match;              // Comparator found address match
  wire          mask_lsb;                 // Enabling masking of LSB's
  reg [15:0]    address_mask_expanded;    // Mask for address comparator
  wire [1:0]    trans_size_zeroed;        // dpu_ahb_hsize_reg_i masked
  wire          data_match_allowed;       // Data match enable
  wire          comp0_matches_byte0;      // COMP[7:0] matches DATA[7:0]
  wire          comp1_matches_byte1;      // COMP[15:8] matches DATA[15:8]
  wire          comp2_matches_byte2;      // COMP[23:16] matches DATA[23:16]
  wire          comp3_matches_byte3;      // COMP[31:24] matches DATA[31:24]
  wire          comp0_matches_byte1;      // COMP[7:0] matches DATA[15:8]
  wire          comp0_matches_byte2;      // COMP[7:0] matches DATA[23:16]
  wire          comp0_matches_byte3;      // COMP[7:0] matches DATA[31:24]
  wire          byte_in_halfword;         // COMP[7:0] matches DATA[7:0] or [15:8]
  wire          byte_in_full_word;        // COMP[7:0] matches any of the 4 bytes
  wire          comp10_matches_bytes10;   // COMP[15:0] matches DATA[15:0]
  wire          comp10_matches_bytes32;   // COMP[15:0] matches DATA[31:16]
  wire          halfword_in_word;         // COMP[15:0] matches DATA[15:0] or [31:16]
  reg           byte_matches;             // Data match for byte
  reg           half_word_matches;        // Data match for half word
  wire          full_word_matches;        // Data match for word
  reg           data_matches;             // Internal datamatch result (match occured)
  reg           data_match_reg;           // Registered data match
  wire          nxt_data_match_reg;       // Next data_match_reg
  wire          data_comp_valid;          // Enable data comparison
  wire          seq_transfer;             // seq_transfer transfer
  reg           ignore_seq;               // Only send the first address of a burst
  wire          ignore_seq_we;            // Enable ignore_seq register
  wire          nxt_ignore_seq;           // Next value of ignore_seq

  wire          emit_etm_trigger;
  reg           etm_trigger;
  wire          etm_trigger_we;
  wire          emit_pc_sample;
  reg           etm_trigger_i_not_d;
  wire          emit_watchpoint;
  wire          qualify_data;
  wire          opt_data_match_en;
  wire          opt_comp_en;

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  wire [1:0]    dsm_debug_lvl;
  wire          dsm_inc_comp;

  assign opt_data_match_en     = (dsm_debug_lvl > 2) ? 1'b1                : 1'b0;

  assign etm_trig_i_not_d_o    = (dsm_trace_lvl > 0) ? etm_trigger_i_not_d : 1'b0;

  assign emit_watchpoint_o     = (dsm_inc_comp == 1) ? emit_watchpoint  : 1'b0;
  assign qualify_data_o        = (dsm_inc_comp == 1) ? qualify_data     : 1'b0;
  assign opt_comp_en           = (dsm_inc_comp == 1) ? 1'b1             : 1'b0;

  assign emit_etm_trigger_o    = (dsm_trace_lvl > 0 &
                                  dsm_inc_comp)      ? emit_etm_trigger : 1'b0;
  assign etm_trigger_o         = (dsm_trace_lvl > 0 &
                                  dsm_inc_comp)      ? etm_trigger      : 1'b0;
  assign emit_pc_sample_o      = (dsm_trace_lvl > 0 &
                                  dsm_inc_comp)      ? emit_pc_sample   : 1'b0;
  assign emit_data_sample_o    = (dsm_trace_lvl > 0 &
                                  dsm_inc_comp)      ? emit_data_sample : 1'b0;

`else

  assign opt_data_match_en  = (DEBUG_LVL > 2) ? 1'b1 : 1'b0;

  assign etm_trig_i_not_d_o = (TRACE_LVL > 0) ? etm_trigger_i_not_d   : 1'b0;

  assign emit_watchpoint_o  = (INC_COMP == 1) ? emit_watchpoint : 1'b0;
  assign qualify_data_o     = (INC_COMP == 1) ? qualify_data    : 1'b0;
  assign opt_comp_en        = (INC_COMP == 1) ? 1'b1            : 1'b0;

  assign etm_trigger_o      = (TRACE_LVL > 0 & INC_COMP == 1) ? etm_trigger      : 1'b0;
  assign emit_etm_trigger_o = (TRACE_LVL > 0 & INC_COMP == 1) ? emit_etm_trigger : 1'b0;
  assign emit_pc_sample_o   = (TRACE_LVL > 0 & INC_COMP == 1) ? emit_pc_sample   : 1'b0;
  assign emit_data_sample_o = (TRACE_LVL > 0 & INC_COMP == 1) ? emit_data_sample : 1'b0;

`endif



  assign fn_sample      = (dpu_trace_enable_i &
                           dwt_function_i[`DWT_FUNC_ACTION+`DWT_FUNC_ACTION_SIZE-1:`DWT_FUNC_ACTION] ==
                           `DWT_FUNC_ACTION_SMPL);

  assign fn_extended_sample  = (dpu_trace_enable_i &
                                dwt_function_i[`DWT_FUNC_ACTION+`DWT_FUNC_ACTION_SIZE-1:`DWT_FUNC_ACTION] ==
                                `DWT_FUNC_ACTION_EXTSMPL);

  assign fn_watchpoint  = (opt_comp_en &
                           dwt_function_i[`DWT_FUNC_ACTION+`DWT_FUNC_ACTION_SIZE-1:`DWT_FUNC_ACTION] ==
                           `DWT_FUNC_ACTION_WP);

  assign fn_etm_trigger = (dpu_trace_enable_i &
                           dwt_function_i[`DWT_FUNC_ACTION+`DWT_FUNC_ACTION_SIZE-1:`DWT_FUNC_ACTION] ==
                           `DWT_FUNC_ACTION_ETM);

  assign fn_sample_pc   = ((fn_sample & (dwt_function_i[`DWT_FUNC_QUAL_PC] | dwt_emit_range_i)) |
                           (fn_extended_sample & (dwt_function_i[1] |  dwt_emit_range_i)));

  assign fn_sample_data = ((fn_sample & dwt_function_i[`DWT_FUNC_QUAL_DATA]) |
                           (fn_extended_sample & (dwt_function_i[1] | ~dwt_emit_range_i)));

  assign fn_on_read     = (fn_sample |
                           (fn_extended_sample ?
                            ~dwt_function_i[0] : dwt_function_i[`DWT_FUNC_QUAL_READ]));

  assign fn_on_write    = (fn_sample |
                           (fn_extended_sample ?
                            dwt_function_i[0] : dwt_function_i[`DWT_FUNC_QUAL_WRITE]));

  assign fn_compare_pc  = ~(fn_on_read | fn_on_write | dwt_pc_sample_match_i |
                            dwt_match_on_data_i | act_as_data_addr_i);


  assign mask_lsb = (dpu_trace_enable_i & ~fn_compare_pc &
                     ~dwt_pc_sample_match_i &
                     (dpu_ahb_hsized_i != `DWT_SIZE_BYTE) &
                     (dpu_ahb_haddrd_masked_i[1:0] <= dwt_comp_i[1:0]) &
                     ((dpu_ahb_hsized_i == `DWT_SIZE_WORD) |
                      ((dpu_ahb_hsized_i == `DWT_SIZE_HWORD) &
                       ((dwt_comp_i[1:0] == 2'b11 & ( dpu_ahb_haddrd_masked_i[1])) |
                        (dwt_comp_i[1:0] == 2'b10 & (^dpu_ahb_haddrd_masked_i[1:0])) |
                        (dwt_comp_i[1:0] == 2'b01 & (~dpu_ahb_haddrd_masked_i[1]))))));

  always @(dwt_mask_i)
    case (dwt_mask_i)
      4'b0000: address_mask_expanded = 16'b1111111111111111;
      4'b0001: address_mask_expanded = 16'b1111111111111110;
      4'b0010: address_mask_expanded = 16'b1111111111111100;
      4'b0011: address_mask_expanded = 16'b1111111111111000;
      4'b0100: address_mask_expanded = 16'b1111111111110000;
      4'b0101: address_mask_expanded = 16'b1111111111100000;
      4'b0110: address_mask_expanded = 16'b1111111111000000;
      4'b0111: address_mask_expanded = 16'b1111111110000000;
      4'b1000: address_mask_expanded = 16'b1111111100000000;
      4'b1001: address_mask_expanded = 16'b1111111000000000;
      4'b1010: address_mask_expanded = 16'b1111110000000000;
      4'b1011: address_mask_expanded = 16'b1111100000000000;
      4'b1100: address_mask_expanded = 16'b1111000000000000;
      4'b1101: address_mask_expanded = 16'b1110000000000000;
      4'b1110: address_mask_expanded = 16'b1100000000000000;
      4'b1111: address_mask_expanded = 16'b1000000000000000;
      default: address_mask_expanded = 16'bXXXXXXXXXXXXXXXX;
    endcase

  always @(dwt_pc_sample_match_i or dwt_pc_cycle_i or mask_lsb or dwt_comp_i or
           dpu_ahb_haddrd_masked_i)
    case ({dwt_pc_sample_match_i,mask_lsb})
      2'b00:   compare_value = dpu_ahb_haddrd_masked_i;
      2'b01:   compare_value = {dpu_ahb_haddrd_masked_i[31:2],dwt_comp_i[1:0]};
      2'b10,
      2'b11:   compare_value = dwt_pc_cycle_i;
      default: compare_value = {32{1'bx}};
    endcase

  assign compare_value_masked = (compare_value &
                                 {{16{1'b1}},address_mask_expanded});
  assign dwt_comp_masked      = (dwt_comp_i &
                                 {{16{1'b1}},address_mask_expanded});

  assign value_match          = (dwt_comp_masked == compare_value_masked);

  assign address_comparator_en = ((|dwt_function_i) & ~dwt_match_on_data_i &
                                  (dwt_pc_sample_match_i |
                                   (address_phase_i & ~fn_compare_pc &
                                    (dpu_ahb_hwrited_i ? fn_on_write :
                                                         fn_on_read))));


  assign pc_addr_masked = ({dpu_etm_ia_i,1'b0} &
                           {32{dpu_trace_enable_i}} &
                           {{16{1'b1}},address_mask_expanded});

  assign pc_match       = (dwt_comp_masked == pc_addr_masked);

  assign pc_match_we    = ((pc_match & etm_ivalid_i & fn_compare_pc) |
                           (pc_match_reg & clear_etm_ivalid_i));
  assign nxt_pc_match_reg = pc_match & etm_ivalid_i;

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      pc_match_reg <= 1'b0;
    else if (pc_match_we & opt_comp_en)
      pc_match_reg <= nxt_pc_match_reg;


  assign data_match_allowed = (dwt_match_on_data_i & comp_good_data_i &
                               opt_data_match_en);

  assign data_to_match      = watchpoint_data_i & {32{data_match_allowed}};

  always @ (dpu_ahb_haddr_reg_i or data_to_match)
    case (dpu_ahb_haddr_reg_i)
      2'b00   : swiz_data_to_match =  data_to_match[31:0];
      2'b01   : swiz_data_to_match = {data_to_match[7:0],  data_to_match[31:8] };
      2'b10   : swiz_data_to_match = {data_to_match[15:0], data_to_match[31:16]};
      2'b11   : swiz_data_to_match = {data_to_match[23:0], data_to_match[31:24]};
      default : swiz_data_to_match = {32{1'bx}};
    endcase

  assign trans_size_zeroed   = dpu_ahb_hsize_reg_i & {2{data_match_allowed}};

  assign comp0_matches_byte0 = (dwt_comp_i[7:0]   == swiz_data_to_match[7:0]);
  assign comp1_matches_byte1 = (dwt_comp_i[15:8]  == swiz_data_to_match[15:8]);
  assign comp2_matches_byte2 = (dwt_comp_i[23:16] == swiz_data_to_match[23:16]);
  assign comp3_matches_byte3 = (dwt_comp_i[31:24] == swiz_data_to_match[31:24]);
  assign comp0_matches_byte1 = (dwt_comp_i[7:0]   == swiz_data_to_match[15:8]);
  assign comp0_matches_byte2 = (dwt_comp_i[7:0]   == swiz_data_to_match[23:16]);
  assign comp0_matches_byte3 = (dwt_comp_i[7:0]   == swiz_data_to_match[31:24]);
  assign byte_in_halfword    = (comp0_matches_byte0 | comp0_matches_byte1);
  assign byte_in_full_word   = (byte_in_halfword    | comp0_matches_byte2 |
                                                      comp0_matches_byte3);

  assign comp10_matches_bytes10 = (comp0_matches_byte0 & comp1_matches_byte1);
  assign comp10_matches_bytes32 = (dwt_comp_i[15:0] == swiz_data_to_match[31:16]);
  assign halfword_in_word       = (comp10_matches_bytes10 |
                                   comp10_matches_bytes32 |
                                   (dwt_comp_i[15:0] == swiz_data_to_match[23:8]));

  always @ (trans_size_zeroed or comp0_matches_byte0 or byte_in_halfword or
            byte_in_full_word)
    case (trans_size_zeroed)
      2'b00   : byte_matches = comp0_matches_byte0;
      2'b01   : byte_matches = byte_in_halfword;
      2'b10,
      2'b11   : byte_matches = byte_in_full_word;
      default : byte_matches = 1'bx;
    endcase

  always @ (trans_size_zeroed or comp10_matches_bytes10 or halfword_in_word)
    case (trans_size_zeroed)
      2'b00   : half_word_matches = 1'b0;
      2'b01   : half_word_matches = comp10_matches_bytes10;
      2'b10,
      2'b11   : half_word_matches = halfword_in_word;
      default : half_word_matches = 1'bx;
    endcase

  assign full_word_matches = comp0_matches_byte0 & comp1_matches_byte1 &
                             comp2_matches_byte2 & comp3_matches_byte3;

  always @ (dwt_match_on_data_i or dwt_match_size_i or trans_size_zeroed or
            byte_matches or half_word_matches or full_word_matches)
    if (dwt_match_on_data_i & (trans_size_zeroed >= dwt_match_size_i))
        case (dwt_match_size_i)
          2'b00   : data_matches = byte_matches;
          2'b01   : data_matches = half_word_matches;
          2'b10,
          2'b11   : data_matches = full_word_matches;
          default : data_matches = 1'bx;
        endcase
    else
      data_matches = 1'b0;

  assign data_comp_valid = ((|dwt_function_i) & data_match_allowed &
                            dpu_ahb_hready_i &
                            (dpu_ahb_hwrite_reg_i ? fn_on_write :
                                                    fn_on_read));

  assign nxt_data_match_reg = (data_matches & opt_data_match_en &
                               data_comp_valid);

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      data_match_reg <= 1'b0;
    else if (opt_data_match_en & (nxt_data_match_reg | data_match_reg))
      data_match_reg <= nxt_data_match_reg;

  assign match_result = (opt_comp_en & ~act_as_data_addr_i &
                         ((data_match_reg) |
                          (pc_match_reg & allow_pc_match_i) |
                          (value_match & address_comparator_en)));

  assign seq_transfer = (&dpu_ahb_htransd_i);


  assign emit_etm_trigger      = match_result & fn_etm_trigger;
  assign emit_pc_sample        = (match_result & fn_sample_pc &
                                  ~(ignore_seq & seq_transfer));
  assign match_data_sample_nxt = match_result & fn_sample_data;

  assign emit_watchpoint       = match_result & fn_watchpoint;

  assign qualify_data          = (act_as_data_addr_i & dpu_trace_enable_i &
                                  value_match & address_phase_i);

  assign etm_trigger_we        = ((dpu_trace_enable_i | (etm_trigger)) &
                                  opt_comp_en);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        etm_trigger         <= 1'b0;
        etm_trigger_i_not_d <= 1'b0;
      end
    else if (etm_trigger_we)
      begin
        etm_trigger         <= emit_etm_trigger;
        etm_trigger_i_not_d <= fn_compare_pc;
      end

  assign match_data_sample_we = (dpu_trace_enable_i | match_data_sample) &
                                dpu_ahb_hready_i & opt_comp_en;

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      match_data_sample <= 1'b0;
    else if(match_data_sample_we)
      match_data_sample <= match_data_sample_nxt;

  assign emit_data_sample = ((match_data_sample & dpu_ahb_hready_i) &
                              ~dpu_ahb_hresp_i[0] & ~mpu_hresp_i);


  assign nxt_ignore_seq = match_result & fn_sample_pc;
  assign ignore_seq_we  = dpu_ahb_hready_i & dpu_trace_enable_i;

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      ignore_seq <= 1'b0;
    else if (ignore_seq_we & opt_comp_en)
      ignore_seq <= nxt_ignore_seq;


`ifdef ARM_ASSERT_ON

  assert_never_unknown  #(0,32,0,
                          "Function register not to be set before address")
  AddrRegister_check
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (|dwt_function_i),
     .test_expr (dwt_comp_i)
     );

  assert_never_unknown  #(0,4,0,"Function register not to be set before mask")
  MaskRegister_check
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (|dwt_function_i),
     .test_expr (dwt_mask_i)
     );

  assert_never #(0,0,"emit_etm_trigger only asserted if fn_etm_trigger set")
  emit_etm_trigger_o_check
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (etm_trigger & ~fn_etm_trigger)
     );

  assert_never #(0,0,"emit_watchpoint only asserted if Trigger disabled")
  emit_watchpoint_o_check1
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (emit_watchpoint & fn_etm_trigger)
     );

  assert_never #(0,0,
                 "emit_watchpoint only asserted if dwt_function_i set to Watchpoint")
  emit_watchpoint_o_check2mask_lsb1
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (emit_watchpoint & ~fn_watchpoint)
     );

  assert_never #(0,0,"emit_pc_sample_o only asserted if Trigger disabled")
  emit_pc_sample_o_check1
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (emit_pc_sample & fn_etm_trigger)
     );

  assert_never #(0,0,
                 "emit_pc_sample only asserted if dwt_function_i set to fn_sample PC")
  emit_pc_sample_o_check2mask_lsb1
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (emit_pc_sample & ~fn_sample_pc)
     );

  assert_never #(0,0,
                 "emit_data_sample only asserted if Trigger disabled")
  emit_data_sample_o_check1
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (match_data_sample_nxt & fn_etm_trigger)
     );

  assert_never #(0,0,
                 "emit_data_sample only asserted if dwt_function_i set to fn_sample Data")
  emit_data_sample_o_check2
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (match_data_sample_nxt & ~fn_sample_data));

  assert_never #(0,0,"mask_lsb must not be asserted for byte accesses.")
  mask_lsb1
    (
     .clk       (clk),
     .reset_n   (reset_n & dpu_trace_enable_i & (|dwt_function_i)),
     .test_expr (mask_lsb & (dpu_ahb_hsized_i == `DWT_SIZE_BYTE))
     );

  assert_never #(0,0,"mask_lsb must not be asserted for address > Comparator.")
  mask_lsb2
    (
     .clk       (clk),
     .reset_n   (reset_n & dpu_trace_enable_i & (|dwt_function_i)),
     .test_expr (mask_lsb & (dpu_ahb_haddrd_masked_i[1:0] > dwt_comp_i[1:0]))
     );

  assert_never #(0,0,
                 "mask_lsb must not be asserted for addresses lower than range.")
  mask_lsb3
    (
     .clk       (clk),
     .reset_n   (reset_n & dpu_trace_enable_i & (|dwt_function_i)),
     .test_expr (mask_lsb &
                 ( (dwt_comp_i[1:0] == 2'b11 &
                    dpu_ahb_hsized_i == `DWT_SIZE_HWORD &
                    ~dpu_ahb_haddrd_masked_i[1]) ||
                   (dwt_comp_i[1:0] == 2'b10 &
                    dpu_ahb_hsized_i == `DWT_SIZE_HWORD &
                    dpu_ahb_haddrd_masked_i[1:0] == 2'b00)))
     );

  assert_never #(0,0,
                 "mask_lsb must only be asserted for data address comparisons.")
  mask_lsb4
    (
     .clk       (clk),
     .reset_n   (reset_n & dpu_trace_enable_i),
     .test_expr (mask_lsb & fn_compare_pc)
     );

`endif

endmodule

`include "cm3_dwt_undefs.vp"
