`include "cm3_dwt_defs.vp"

module cm3_dwt_packet_gen
  (// Inputs
   clk, reset_n, dpu_trace_enable_i, emit_data_sample_i, emit_pc_sample_i,
   dwt_emit_range0_i, dwt_emit_range1_i, dwt_emit_range2_i, dwt_emit_range3_i,
   dpu_ahb_haddrd_masked_i, itm_dwt_ready_i, dpu_etm_int_num_i,
   dpu_etm_int_stat_i, dpu_etm_ivalid_i, dpu_etm_ia_i, sleeping_i,
   emit_pc_periodic_i, emit_cyc_i, emit_cpi_i, emit_int_i, emit_sleep_i,
   emit_fold_i, emit_lsu_i, int_trace_en_i, dpu_ahb_hwrite_reg_i,
   dpu_ahb_hsize_reg_i, dpu_ahb_haddr_reg_i, watchpoint_data_i,
   dwt_valid_o, dwt_size_o, dwt_overflow_o, dwt_data_o
   );

  parameter TRACE_LVL      = 1;
  parameter RESET_ALL_REGS = 0;


  input         clk;                     // Clock
  input         reset_n;                 // Reset
  input         dpu_trace_enable_i;      // enable tracing
  input [3:0]   emit_data_sample_i;      // Emit Sample Data - bit per comparator
  input [3:0]   emit_pc_sample_i;        // Emit Sample PC - bit per comparator
  input         dwt_emit_range0_i;       // DWT Emit Range 0
  input         dwt_emit_range1_i;       // DWT Emit Range 1
  input         dwt_emit_range2_i;       // DWT Emit Range 2
  input         dwt_emit_range3_i;       // DWT Emit Range 3
  input [15:0]  dpu_ahb_haddrd_masked_i; // DCore address
  input         itm_dwt_ready_i;         // ITM Ready
  input [8:0]   dpu_etm_int_num_i;       // Interrupt number
  input [1:0]   dpu_etm_int_stat_i;      // Interrupt status
  input         dpu_etm_ivalid_i;        // Instruction Valid
  input [31:1]  dpu_etm_ia_i;            // PC
  input         sleeping_i;              // Core is sleeping
  input         emit_pc_periodic_i;      // Emit pc_periodic
  input         emit_cyc_i;              // Emit PSCYCCNT underflow event
  input         emit_cpi_i;              // Emit CPI Event
  input         emit_int_i;              // Emit Interrupt event
  input         emit_sleep_i;            // Emit Sleep event
  input         emit_fold_i;             // Emit fold event
  input         emit_lsu_i;              // Emit LSU event
  input         int_trace_en_i;          // Interrupt trace event enabled
  input         dpu_ahb_hwrite_reg_i;    // ppb_hwrite_iDCore register
  input [1:0]   dpu_ahb_hsize_reg_i;     // HSIZEDCore register
  input [1:0]   dpu_ahb_haddr_reg_i;     // dpu_ahb_haddrd_i[15:0] register
  input [31:0]  watchpoint_data_i;       // WPdata - mux of read and write data

  output [2:0]  dwt_size_o;              // Size
  output        dwt_overflow_o;          // Overflow
  output        dwt_valid_o;             // Valid
  output [39:0] dwt_data_o;              // Data

  wire [2:0]    dwt_size;                // Size
  wire          dwt_overflow;            // Overflow
  wire          dwt_valid;               // Valid
  wire [31:0]   packet_data;             // Data
  wire [7:0]    packet_header;           // Header
  wire [39:0]   dwt_data;                // Data

  wire          emit_int_trace;          // Emit Interrupt Trace

  wire          events_event_valid;      // OR of the above 6 events
  wire [5:0]    events;                  // Vector of the above 6 events

  reg           range_address_select;    // Select address LSB's rather than PC
  wire [31:0]   address_value;           // Address value to be sent for WP address
  reg [31:1]    dpu_etm_ia_reg;          // Latest held PC
  wire [31:1]   current_pc;              // Current valid PC
  wire          dpu_etm_ia_reg_we;       // Register current PC

  wire [2:0]    periodic_at_size;        // Size of PC sample packet
  wire [31:0]   periodic_packet_data;    // PC sample packet payload
  wire [1:0]    periodic_packet_size;    // PC sample ITM size code

  wire [4:0]    granted;                 // granted requesting ITM

  wire [1:0]    pc_sampler_status;       // WP PC Sample StateMachine
  wire          capture_pc_sample;       // enable WP PC samplig

  reg [31:0]    nxt_data_packet_data;    // Next Sample Data
  wire [1:0]    data_sampler_status;     // WP Data Sample State Machine
  wire          capture_data_sample;     // WP's have matched for data sample

  reg [1:0]     pc_packet_header;        // packet header ID for PC sample
  reg [1:0]     nxt_data_packet_header;  // packet header ID for data sample
  reg [31:0]    pc_packet_data;          // PC packet (data only)
  reg           pc_packet_option;        // PC packet option (address LSBs or PC)
  reg [1:0]     pc_packet_header_reg;    // PC packet qualifiers (ID)
  wire [1:0]    pc_packet_size;          // PC packet size (2 or 4 bits)
  reg [31:0]    data_packet_data;        // Data packet (data only)
  reg [1:0]     data_packet_type;        // Data packet qualifiers (ID)
  reg [1:0]     data_packet_size;        // Data packet size
  reg           data_packet_write;       // Data packet write signal
  reg [2:0]     data_at_size;            // dwt_size_o encoding for data packets
  wire [2:0]    offset_at_size;          // dwt_size_o encoding for offset packets
  reg [1:0]     data_packet_size_bits;   // dwt_data_o[1:0] encoding for PC/data pkts

  reg [5:0]     event_packet_data;       // Event Sample register
  wire [5:0]    nxt_event_packet_data;   // Event Sample register - next
  wire [1:0]    event_sampler_status;    // Event Sample StateMachine

  reg  [10:0]   int_value_reg;           // Interrupt Trace Sample register
  wire [10:0]   nxt_int_value_reg;       // Interrupt Trace Sample register
  wire [1:0]    int_trace_sampler_status;// Intterupt Trace Sample StateMachine

  reg [31:1]    pc_periodic;             // PC Sample register
  reg           periodic_was_sleeping;   // Core sleeping at PC sample
  wire [1:0]    periodic_sampler_status; // PC Sample StateMachine
  wire          opt_reset_all_n;         // Optional reset all flops
  wire          opt_gen_enable;          // Logic removal term

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  wire          dsm_reset_all_regs;

  assign opt_reset_all_n = dsm_reset_all_regs  ? reset_n      : 1'b1;
  assign opt_gen_enable  = (dsm_trace_lvl > 0) ? 1'b1         : 1'b0;
`else

  assign opt_reset_all_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;
  assign opt_gen_enable  = (TRACE_LVL > 0) ? 1'b1        : 1'b0;
`endif




  assign capture_pc_sample   = |(emit_pc_sample_i[3:0])   & opt_gen_enable;
  assign capture_data_sample = |(emit_data_sample_i[3:0]) & opt_gen_enable;

  always @(emit_data_sample_i)
    case (emit_data_sample_i[2:0])
      3'b001, 3'b011,
      3'b101, 3'b111: nxt_data_packet_header = `ITM_WP0;
      3'b010, 3'b110: nxt_data_packet_header = `ITM_WP1;
      3'b100:         nxt_data_packet_header = `ITM_WP2;
      3'b000:         nxt_data_packet_header = `ITM_WP3;
      default:        nxt_data_packet_header = 2'bxx;
    endcase

  always @(emit_pc_sample_i)
    case (emit_pc_sample_i[2:0])
      3'b001, 3'b011,
      3'b101, 3'b111: pc_packet_header = `ITM_WP0;
      3'b010, 3'b110: pc_packet_header = `ITM_WP1;
      3'b100:         pc_packet_header = `ITM_WP2;
      3'b000:         pc_packet_header = `ITM_WP3;
      default:        pc_packet_header = 2'bxx;
    endcase

  always @(emit_pc_sample_i or dwt_emit_range0_i or dwt_emit_range1_i or
           dwt_emit_range2_i or dwt_emit_range3_i)
    case(emit_pc_sample_i[2:0])
      3'b001, 3'b011,
      3'b101, 3'b111: range_address_select = dwt_emit_range0_i;
      3'b010, 3'b110: range_address_select = dwt_emit_range1_i;
      3'b100:         range_address_select = dwt_emit_range2_i;
      3'b000:         range_address_select = dwt_emit_range3_i & emit_pc_sample_i[3];
      default:        range_address_select = 1'bX;
    endcase

  assign address_value = (range_address_select ?
                          {{16{1'b0}},dpu_ahb_haddrd_masked_i[15:0]} :
                          {current_pc,1'b0});


  cm3_dwt_packet_state #(TRACE_LVL)
  u_dwt_state_pc_sample
    (// Inputs
     .clk              (clk),
     .reset_n          (reset_n),
     .event_valid_i    (capture_pc_sample),
     .granted_i        (1'b1), // Always granted if valid
     .itm_dwt_ready_i  (itm_dwt_ready_i),
     .sampler_status_o (pc_sampler_status)
     );

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      begin
        pc_packet_data       <= {32{1'b0}};
        pc_packet_option     <= 1'b0;
        pc_packet_header_reg <= 2'b00;
      end
    else if(capture_pc_sample)
      begin
        pc_packet_data       <= address_value;
        pc_packet_option     <= range_address_select;
        pc_packet_header_reg <= pc_packet_header;
      end

  assign pc_packet_size = pc_packet_option ? `ITM_SIZE_2_BYTE : `ITM_SIZE_4_BYTE;
  assign offset_at_size = pc_packet_option ? `AT_SIZE_2BYTE   : `AT_SIZE_4BYTE;


  cm3_dwt_packet_state #(TRACE_LVL)
  u_dwt_state_data
    (// Inputs
     .clk              (clk),
     .reset_n          (reset_n),
     .event_valid_i    (capture_data_sample),
     .granted_i        (granted[1]),
     .itm_dwt_ready_i  (itm_dwt_ready_i),
     .sampler_status_o (data_sampler_status)
     );

  always @(dpu_ahb_haddr_reg_i or dpu_ahb_hsize_reg_i or watchpoint_data_i)
    case({dpu_ahb_hsize_reg_i,dpu_ahb_haddr_reg_i[1:0]})
      {`DWT_SIZE_BYTE,  2'b00}: nxt_data_packet_data = {{24{1'b0}},watchpoint_data_i[7:0]};
      {`DWT_SIZE_BYTE,  2'b01}: nxt_data_packet_data = {{24{1'b0}},watchpoint_data_i[15:8]};
      {`DWT_SIZE_BYTE,  2'b10}: nxt_data_packet_data = {{24{1'b0}},watchpoint_data_i[23:16]};
      {`DWT_SIZE_BYTE,  2'b11}: nxt_data_packet_data = {{24{1'b0}},watchpoint_data_i[31:24]};
      {`DWT_SIZE_HWORD, 2'b00}: nxt_data_packet_data = {{16{1'b0}},watchpoint_data_i[15:0]};
      {`DWT_SIZE_HWORD, 2'b01}: nxt_data_packet_data = {{16{1'b0}},watchpoint_data_i[23:8]};
      {`DWT_SIZE_HWORD, 2'b10}: nxt_data_packet_data = {{16{1'b0}},watchpoint_data_i[31:16]};
      {`DWT_SIZE_HWORD, 2'b11}: nxt_data_packet_data = {{16{1'b0}},watchpoint_data_i[7:0],
                                                        watchpoint_data_i[31:24]};
      {`DWT_SIZE_WORD,  2'b00}: nxt_data_packet_data = watchpoint_data_i[31:0];
      {`DWT_SIZE_WORD,  2'b01}: nxt_data_packet_data = {watchpoint_data_i[7:0],
                                                        watchpoint_data_i[31:8]};
      {`DWT_SIZE_WORD,  2'b10}: nxt_data_packet_data = {watchpoint_data_i[15:0],
                                                        watchpoint_data_i[31:16]};
      {`DWT_SIZE_WORD,  2'b11}: nxt_data_packet_data = {watchpoint_data_i[23:0],
                                                        watchpoint_data_i[31:24]};
      {`DWT_SIZE_DWORD, 2'b00}: nxt_data_packet_data = 32'h00000000;
      {`DWT_SIZE_DWORD, 2'b01}: nxt_data_packet_data = 32'h00000000;
      {`DWT_SIZE_DWORD, 2'b10}: nxt_data_packet_data = 32'h00000000;
      {`DWT_SIZE_DWORD, 2'b11}: nxt_data_packet_data = 32'h00000000;
      default:                  nxt_data_packet_data = 32'hxxxxxxxx;
    endcase

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      begin
        data_packet_data   <= {32{1'b0}};
        data_packet_type   <= 2'b00;
        data_packet_write  <= 1'b0;
        data_packet_size   <= 2'b00;
      end
    else if(capture_data_sample)
      begin
        data_packet_data   <= nxt_data_packet_data;
        data_packet_type   <= nxt_data_packet_header;
        data_packet_write  <= dpu_ahb_hwrite_reg_i;
        data_packet_size   <= dpu_ahb_hsize_reg_i;
      end


  assign events             = {emit_cyc_i,  emit_fold_i,
                               emit_lsu_i,  emit_sleep_i,
                               emit_int_i,  emit_cpi_i};
  assign events_event_valid = (|events);

  cm3_dwt_packet_state #(TRACE_LVL)
  u_dwt_state_events
    (// Inputs
     .clk              (clk),
     .reset_n          (reset_n),
     .event_valid_i    (events_event_valid),
     .granted_i        (granted[2]),
     .itm_dwt_ready_i  (itm_dwt_ready_i),
     .sampler_status_o (event_sampler_status)
     );

  assign nxt_event_packet_data = events | ({6{~((event_sampler_status == `ES_INVALID) |
                                                (granted[2] & itm_dwt_ready_i))}} &
                                           event_packet_data);

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      event_packet_data <= {6{1'b0}};
    else if(events_event_valid & opt_gen_enable)
      event_packet_data <= nxt_event_packet_data;

  assign emit_int_trace = ((dpu_etm_int_stat_i != 2'b00) &
                           int_trace_en_i & opt_gen_enable);

  cm3_dwt_packet_state #(TRACE_LVL)
  u_dwt_state_int_trace
    (// Inputs
     .clk              (clk),
     .reset_n          (reset_n),
     .event_valid_i    (emit_int_trace),
     .granted_i        (granted[3]),
     .itm_dwt_ready_i  (itm_dwt_ready_i),
     .sampler_status_o (int_trace_sampler_status)
     );

  assign nxt_int_value_reg = {dpu_etm_int_num_i,dpu_etm_int_stat_i};

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      int_value_reg <= {11{1'b0}};
    else if(emit_int_trace & opt_gen_enable)
      int_value_reg <= nxt_int_value_reg;

  cm3_dwt_packet_state #(TRACE_LVL)
  u_dwt_state_pc_periodic
    (// Inputs
     .clk              (clk),
     .reset_n          (reset_n),
     .event_valid_i    (emit_pc_periodic_i),
     .granted_i        (granted[4]),
     .itm_dwt_ready_i  (itm_dwt_ready_i),
     .sampler_status_o (periodic_sampler_status)
     );

  assign dpu_etm_ia_reg_we = (dpu_trace_enable_i & dpu_etm_ivalid_i &
                              opt_gen_enable);

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      dpu_etm_ia_reg <= {31{1'b0}};
    else if(dpu_etm_ia_reg_we)
      dpu_etm_ia_reg <= dpu_etm_ia_i;

  assign current_pc = dpu_etm_ivalid_i ? dpu_etm_ia_i : dpu_etm_ia_reg;

  always @(posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      begin
        pc_periodic           <= {31{1'b0}};
        periodic_was_sleeping <= 1'b0;
      end
    else if(emit_pc_periodic_i & opt_gen_enable)
      begin
        pc_periodic           <= current_pc;
        periodic_was_sleeping <= sleeping_i;
      end


  assign granted[0] = pc_sampler_status[1];
  assign granted[1] = data_sampler_status[1]      & ~granted[0];
  assign granted[2] = event_sampler_status[1]     & ~(|granted[1:0]);
  assign granted[3] = int_trace_sampler_status[1] & ~(|granted[2:0]);
  assign granted[4] = periodic_sampler_status[1]  & ~(|granted[3:0]);

  assign dwt_valid = (pc_sampler_status[1]        |
                      data_sampler_status[1]      |
                      event_sampler_status[1]     |
                      int_trace_sampler_status[1] |
                      periodic_sampler_status[1]);

  always @(data_packet_size)
    case(data_packet_size)
      `DWT_SIZE_BYTE:
        begin
          data_at_size          = `AT_SIZE_1BYTE;
          data_packet_size_bits = `ITM_SIZE_1_BYTE;
        end
      `DWT_SIZE_HWORD:
        begin
          data_at_size          = `AT_SIZE_2BYTE;
          data_packet_size_bits = `ITM_SIZE_2_BYTE;
        end
      `DWT_SIZE_WORD:
        begin
          data_at_size          = `AT_SIZE_4BYTE;
          data_packet_size_bits = `ITM_SIZE_4_BYTE;
        end
      `DWT_SIZE_DWORD:
        begin
          data_at_size          = `AT_SIZE_0BYTE;
          data_packet_size_bits = `ITM_SIZE_0_BYTE;
        end
      default:
        begin
          data_at_size          = 3'bxxx;
          data_packet_size_bits = 2'bxx;
        end
    endcase // case(data_packet_size)

  assign periodic_at_size     = periodic_was_sleeping ? `AT_SIZE_1BYTE:`AT_SIZE_4BYTE;
  assign periodic_packet_data = {32{~periodic_was_sleeping}} & {pc_periodic[31:1],1'b0};
  assign periodic_packet_size = periodic_was_sleeping ? `ITM_SIZE_1_BYTE:`ITM_SIZE_4_BYTE;

  assign dwt_size      = ((offset_at_size   & {3{granted[0]}}) |
                          (data_at_size     & {3{granted[1]}}) |
                          (`AT_SIZE_1BYTE   & {3{granted[2]}}) |
                          (`AT_SIZE_2BYTE   & {3{granted[3]}}) |
                          (periodic_at_size & {3{granted[4]}}));


  assign dwt_overflow  = (((pc_sampler_status        == `ES_OFLOW)) | // highest priority always grants
                          ((data_sampler_status      == `ES_OFLOW) & granted[1]) |
                          ((event_sampler_status     == `ES_OFLOW) & granted[2]) |
                          ((int_trace_sampler_status == `ES_OFLOW) & granted[3]) |
                          ((periodic_sampler_status  == `ES_OFLOW) & granted[4]));

  assign packet_data   = ((pc_packet_data                  & {32{granted[0]}}) |
                          (data_packet_data                & {32{granted[1]}}) |
                          ({{26{1'b0}},event_packet_data}  & {32{granted[2]}}) |
                          ({{18{1'b0}},int_value_reg[1:0],
                            {3{1'b0}},int_value_reg[10:2]} & {32{granted[3]}}) |
                          (periodic_packet_data            & {32{granted[4]}}));

  assign packet_header = (((`ITM_WP_PC_HEADER |
                            {2'b00,pc_packet_header_reg,pc_packet_option,1'b0,pc_packet_size}) &
                           {8{granted[0]}}) |
                          ((`ITM_WP_DATA_HEADER |
                            {2'b00,data_packet_type,data_packet_write,1'b0,data_packet_size_bits}) &
                           {8{granted[1]}})|
                          (`ITM_EVENT_HEADER     & {8{granted[2]}})|
                          (`ITM_INT_HEADER       & {8{granted[3]}})|
                          ((`ITM_PC_HEADER | {6'b000000,periodic_packet_size}) &
                           {8{granted[4]}}));

  assign dwt_data      = {packet_data,packet_header};

`ifdef ARM_DSM
  assign dwt_size_o      = (dsm_trace_lvl > 0) ? dwt_size     : 3'b000;
  assign dwt_overflow_o  = (dsm_trace_lvl > 0) ? dwt_overflow : 1'b0;
  assign dwt_valid_o     = (dsm_trace_lvl > 0) ? dwt_valid    : 1'b0;
  assign dwt_data_o      = (dsm_trace_lvl > 0) ? dwt_data     : {40{1'b0}};
`else
  assign dwt_size_o      = (TRACE_LVL > 0) ? dwt_size     : 3'b000;
  assign dwt_overflow_o  = (TRACE_LVL > 0) ? dwt_overflow : 1'b0;
  assign dwt_valid_o     = (TRACE_LVL > 0) ? dwt_valid    : 1'b0;
  assign dwt_data_o      = (TRACE_LVL > 0) ? dwt_data     : {40{1'b0}};

`endif


`ifdef ARM_ASSERT_ON

  assert_zero_one_hot #(0,5,0,"granted must be one-hot")
  onehot_master
    (
     .clk       (clk),
     .reset_n    (reset_n),
     .test_expr (granted[4:0])
     );


  wire [4:0] ovl_packet;
  wire [4:0] ovl_status_invalid;
  reg [4:0]  ovl_request_reg;
  wire [4:0] nxt_ovl_request_reg;
  reg        ovl_check_wp_data;
  reg        ovl_check_wp_pc;
  reg        ovl_dpu_ahb_hsize_reg_wr;

  assign ovl_packet[0] = (dwt_valid &
                          ((((dwt_data[7:0] & 8'b11001111) == 8'b01000111) &
                            (dwt_size == `AT_SIZE_4BYTE)) |
                           (((dwt_data[7:0] & 8'b11001111) == 8'b01001110) &
                            (dwt_size == `AT_SIZE_2BYTE))));
  assign ovl_packet[1] = (dwt_valid &
                          ((dwt_data[7:0] & 8'b11000100) == 8'b10000100));
  assign ovl_packet[2] = (dwt_valid &
                          ((dwt_data[7:0] == 8'b00000101) &
                           (dwt_size == `AT_SIZE_1BYTE)));
  assign ovl_packet[3] = (dwt_valid &
                          ((dwt_data[7:0] == 8'b00001110) &
                           (dwt_size == `AT_SIZE_2BYTE)));
  assign ovl_packet[4] = (dwt_valid &
                          (((dwt_data[7:0] == 8'b00010111) &
                            (dwt_size == `AT_SIZE_4BYTE)) |
                           ((dwt_data[7:0] == 8'b00010101) &
                            (dwt_size == `AT_SIZE_1BYTE))));

  assign ovl_status_invalid[0] = (pc_sampler_status        == `ES_INVALID);
  assign ovl_status_invalid[1] = (data_sampler_status      == `ES_INVALID);
  assign ovl_status_invalid[2] = (event_sampler_status     == `ES_INVALID);
  assign ovl_status_invalid[3] = (int_trace_sampler_status == `ES_INVALID);
  assign ovl_status_invalid[4] = (periodic_sampler_status  == `ES_INVALID);

  assign nxt_ovl_request_reg[4:0] = (opt_gen_enable ?
                                     {emit_pc_periodic_i,
                                      emit_int_trace,
                                      events_event_valid,
                                      capture_data_sample,
                                      capture_pc_sample} :
                                     {5{1'b0}});

  always @(posedge clk or negedge reset_n)
  if( ~reset_n )
    ovl_request_reg <= {5{1'b0}};
  else
    ovl_request_reg <= nxt_ovl_request_reg[4:0];

  always @(posedge clk or negedge reset_n)
  if( ~reset_n )
    begin
      ovl_check_wp_data        <= 1'b0;
      ovl_check_wp_pc          <= 1'b0;
      ovl_dpu_ahb_hsize_reg_wr <= 1'b0;
    end
  else
    begin
      ovl_check_wp_data        <= capture_data_sample;
      ovl_check_wp_pc          <= capture_pc_sample;
      ovl_dpu_ahb_hsize_reg_wr <= ovl_dpu_ahb_hsize_reg_wr;
    end

  assert_never #(0, 0, "Unrecognised ITM packet type")
  ITM_packet_Check
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (dwt_valid & ~(|ovl_packet))
     );

  assert_never #(0, 0, "Incorrect packet request status")
  ITM_Request_Status
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (|(ovl_request_reg & ovl_status_invalid))
     );

  assert_never_unknown #(0,2,0,"Invalid PC type information")
  WPPCSize_not_x_or_z
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (ovl_check_wp_pc),
     .test_expr (pc_packet_header_reg)
     );

  assert_never_unknown #(0,2,0,"Invalid data type information")
  data_packet_type_not_x_or_z
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (ovl_check_wp_data),
     .test_expr (data_packet_type)
     );

  assert_never_unknown #(0,32,0,"Invalid data packet information")
  data_packet_data_not_x_or_z
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (ovl_check_wp_data),
     .test_expr (data_packet_data)
     );

  assert_never_unknown #(0,1,0,"Invalid data write information")
  data_packet_write_not_x_or_z
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (ovl_check_wp_data),
     .test_expr (data_packet_write)
     );

  assert_never_unknown #(0,2,0,"Invalid data size information")
  data_packet_size_not_x_or_z
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (ovl_check_wp_data),
     .test_expr (data_packet_size)
     );

  assert_never #(0,0,"data_packet_size must never be double-word")
  data_packet_size_not_double_word
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .test_expr (ovl_check_wp_data & ovl_dpu_ahb_hsize_reg_wr &
                 (data_packet_size == `DWT_SIZE_DWORD))
     );

  assert_never_unknown #(0,4,0,"emit_pc_sample_i is x")
  ovl_emit_pc_unknown
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (emit_pc_sample_i)
     );

  assert_never_unknown #(0,4,0,"emit_data_sample_i is x")
  ovl_emit_data_unknown
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (emit_data_sample_i)
     );

`endif

endmodule
`include "cm3_dwt_undefs.vp"
