`include "cm3_etm_defs.vp"

module CM3ETM (
  FCLK, PORESETn, NIDEN, ETMIA, ETMIVALID, ETMISTALL, ETMICCFAIL, ETMIBRANCH,
  ETMIINDBR, ETMISB, ETMINTSTAT, ETMCANCEL, ETMFOLD, ETMINTNUM,
  COREHALT, EXTIN, MAXEXTIN, FIFOFULLEN, DWTMATCH, DWTINOTD, ETMDVALID,
  ATREADYM, PSEL, PENABLE, PADDR, PWRITE, PWDATA, TSVALUEB, TSCLKCHANGE,
  SE, CGBYPASS,
  ETMPWRUP, ETMEN, PRDATA, ATDATAM, ATVALIDM, AFREADYM, ETMTRIGOUT,
  ATIDM, ETMDBGRQ, FIFOFULL
`ifdef ARM_DSM
  , dsm_ID, dsm_no_etm, dsm_trace_lvl, dsm_clkgate_present, dsm_reset_all_regs
`endif
  );

  parameter TRACE_LVL          = 2;
  parameter CLKGATE_PRESENT    = 0;
  parameter RESET_ALL_REGS     = 0;

  parameter ETMCLKGATE_PRESENT = (CLKGATE_PRESENT != 0 & TRACE_LVL > 1) ? 1 : 0;


  input         FCLK;              // Main Clock
  input         PORESETn;          // APB Reset

  input         NIDEN;             // Non-invasive debug enable - must be synchronous

  input [31:1]  ETMIA;             // Instruction Address
  input         ETMIVALID;         // Instruction data is valid
  input         ETMISTALL;         // Stall IVALID from prev. cycle
  input         ETMDVALID;         // Data Matches are valid
  input         ETMFOLD;           // 2 instructions this cycle
  input         ETMICCFAIL;        // Instruction failed condition code
  input         ETMIBRANCH;        // Instruction is Branch
  input         ETMIINDBR;         // Instruction is indirect branch
  input         ETMISB;            // Instruction is ISB
  input [2:0]   ETMINTSTAT;        // Branch is due to int entry[x] or exit[y]
  input         ETMCANCEL;         // Previous started instruction did not complete

  input [8:0]   ETMINTNUM;         // Interrupt Source
  input         COREHALT;          // Core is in DEBUG state
  input [1:0]   EXTIN;             // ETM External Inputs
  input [1:0]   MAXEXTIN;          // Maximum External Inputs Supported
  input         FIFOFULLEN;        // System supports FIFOFULL
  input [3:0]   DWTMATCH;          // Comparator Matches
  input [3:0]   DWTINOTD;          // Comparators are I or D
  input         ATREADYM;          // ATDATA can be accepted
  input         PSEL;              // APB Slave Select Signal
  input         PENABLE;           // APB Transfer Type
  input [11:2]  PADDR;             // APB Address Bus
  input         PWRITE;            // APB !Read/Write Signal
  input [31:0]  PWDATA;            // APB Write Data Bus
  input [47:0]  TSVALUEB;           // natural timestamp count input
  input         TSCLKCHANGE;       // timestamp clock ratio changed
  input         SE;                // Scan Enable
  input         CGBYPASS;          // Architectural clock gate bypass


  output        ETMPWRUP;          // The ETM is powered-up
  output        ETMEN;             // ETM Traceport Enabled
  output [31:0] PRDATA;            // APB Read Data Bus
  output [7:0]  ATDATAM;           // Trace Bus Data
  output        ATVALIDM;          // Trace Bus Data Valid
  output        AFREADYM;          // Fifo Flush Finished
  output        ETMTRIGOUT;        // ETM Trigger Status Signal to TPIU
  output [6:0]  ATIDM;             // ATB Trace Source ID
  output        ETMDBGRQ;          // Debug request to core
  output        FIFOFULL;          // FIFO depth exceeds pre-set level

`ifdef ARM_DSM
  input [2:0]   dsm_ID;
  input         dsm_no_etm;
  input [1:0]   dsm_trace_lvl;
  input         dsm_clkgate_present;
  input         dsm_reset_all_regs;
`endif

  wire [3:0]    etm_rev_and;            // RevAnd for eco changes
  wire          register_write;         // Register Write Signal
  wire [6:0]    ATIDM;                  // From u_cm3_etm_trigger
  wire          prg_branch_broadcast;   // From u_cm3_etm_control_reg
  wire          prg_ts_enable;          // From u_cm3_etm_control_reg
  wire          prg_stall_processor;    // Enable FifoFull output
  wire [3:0]    prg_port_size;          // Control register port size
  wire [2:0]    prg_port_mode;          // Control register port mode
  wire          prg_etm_en;             // Control register enable bit
  wire          prg_etm_power_down;     // Control register power down
  wire          prg_etm_power_up;       // inverted power down signal

  wire          clk_for_etm;            // Parameterised gated clock
  wire          gated_clock;            // From u_cm3_etm_clk_gen
  wire          prg_may_assert_dbgreq;  // From u_cm3_etm_control_reg
  wire          power_down_reset;       // From u_cm3_etm_control_reg
  wire          prg_prog_bit;           // From u_cm3_etm_control_reg
  wire          prg_prog_bit_fall;      // From u_cm3_etm_control_reg
  wire          prg_prog_bit_and_en;
  wire [31:0]   read_data;              // From u_cm3_etm_trigger
  wire [11:2]   apb_addr_reg;           // Registered apb address
  wire          trigger_req;            // From u_cm3_etm_trigger
  wire          trace_enable;           // From u_cm3_etm_trigger
  wire          timestamp_event;        // From u_cm3_etm_trigger
  wire [7:0]    fifo_out;               // Fifo Output
  wire          reason_overflow;        // ETM not yet recovered from overflow
  wire          pop_fifo;               // Byte taken from fifo
  wire          fifo_empty;             // No bytes in fifo
  wire          fifo_full_support_reg;  // Buffered FIFOFULLEN
  wire          prg_lock_status;        // CoreSight lock status
  wire [3:0]    prg_claim_tag;          // CoreSight claim tags

  wire          integration_mode;       // Integration Test Mode
  wire          integration_trigger_we; // Integration Mode TRIGOUT Register Enable
  wire          integration_atb_cr0_we; // Integration Mode ATBCTR2 Register Enable

  wire [(`FIFO_DEPTH_SIZE-1):0] prg_fifo_full_level; // Fifo Full level register
  wire          fifo_stall;

  wire          trace_flushed;          // Trace has stopped and fifo drained
  wire          flush_trace;            // Request ot drain fifos
  wire          niden_t0;               // NIDEN input double synchronised to CLK
  wire          niden_t1;               // ditto, but aligned to instruction boundary.
  wire          niden_t2;               // NIDEN T1 registered into T2.
  wire          fake_ivalid_t1;         // Instruction data is valid, with faked excpn
  wire          etm_istall_t1;          // ETMISTALL in T1 stage
  wire          etm_ivalid_t1;          // ETMIVALID in T1 stage

  wire [7:0]    atb_atdatam;            // ATB trace data
  wire          atb_atvalidm;           // ATB data valid
  wire          atb_afreadym;           // ATB sink ready
  wire          etm_trigger_out;        // ETM trigger output
  wire [6:0]    atb_atidm;              // ATB trace source ID
  wire          etm_dbg_req;            // Debug request to core

`ifdef ARM_DSM

  assign ETMPWRUP   = (dsm_trace_lvl > 1) ? prg_etm_power_up : 1'b0;
  assign ETMEN      = (dsm_trace_lvl > 1) ? prg_etm_en       : 1'b0;
  assign PRDATA     = (dsm_trace_lvl > 1) ? read_data        : {32{1'b0}};
  assign ATDATAM    = (dsm_trace_lvl > 1) ? atb_atdatam      : {8{1'b0}};
  assign ATVALIDM   = (dsm_trace_lvl > 1) ? atb_atvalidm     : 1'b0;
  assign AFREADYM   = (dsm_trace_lvl > 1) ? atb_afreadym     : 1'b0;
  assign ETMTRIGOUT = (dsm_trace_lvl > 1) ? etm_trigger_out  : 1'b0;
  assign ATIDM      = (dsm_trace_lvl > 1) ? atb_atidm        : {7{1'b0}};
  assign ETMDBGRQ   = (dsm_trace_lvl > 1) ? etm_dbg_req      : 1'b0;
  assign FIFOFULL   = (dsm_trace_lvl > 1) ? fifo_stall       : 1'b0;

`else

  assign ETMPWRUP   = (TRACE_LVL > 1) ? prg_etm_power_up : 1'b0;
  assign ETMEN      = (TRACE_LVL > 1) ? prg_etm_en       : 1'b0;
  assign PRDATA     = (TRACE_LVL > 1) ? read_data        : {32{1'b0}};
  assign ATDATAM    = (TRACE_LVL > 1) ? atb_atdatam      : {8{1'b0}};
  assign ATVALIDM   = (TRACE_LVL > 1) ? atb_atvalidm     : 1'b0;
  assign AFREADYM   = (TRACE_LVL > 1) ? atb_afreadym     : 1'b0;
  assign ETMTRIGOUT = (TRACE_LVL > 1) ? etm_trigger_out  : 1'b0;
  assign ATIDM      = (TRACE_LVL > 1) ? atb_atidm        : {7{1'b0}};
  assign ETMDBGRQ   = (TRACE_LVL > 1) ? etm_dbg_req      : 1'b0;
  assign FIFOFULL   = (TRACE_LVL > 1) ? fifo_stall       : 1'b0;

`endif


  assign etm_rev_and = 4'b0000;

  cm3_etm_apb_if #(TRACE_LVL, RESET_ALL_REGS)
    u_etm_apb_if
    (
    .reset_n                 (PORESETn),
    .clk                     (FCLK),
    .apb_sel_i               (PSEL),
    .apb_enable_i            (PENABLE),
    .apb_write_i             (PWRITE),
    .apb_addr_i              (PADDR),
    .register_write_o        (register_write),
    .apb_addr_reg_o          (apb_addr_reg)
     );

  cm3_etm_control_reg #(TRACE_LVL)
    u_etm_control_reg
    (
    .clk                     (FCLK),
    .reset_n                 (PORESETn),
    .register_write_i        (register_write),
    .apb_wdata_i             (PWDATA[31:0]),
    .apb_addr_i              (apb_addr_reg),
    .fifo_full_en_i          (FIFOFULLEN),
    .niden_i                 (NIDEN),
    .etm_ivalid_i            (ETMIVALID),
    .fake_ivalid_t1_i        (fake_ivalid_t1),
    .etm_istall_i            (ETMISTALL),
    .etm_istall_t1_i         (etm_istall_t1),
    .etm_ivalid_t1_i         (etm_ivalid_t1),
    .etm_int_stat_i          (ETMINTSTAT),
    .power_down_reset_o      (power_down_reset),
    .prg_etm_en_o            (prg_etm_en),
    .prg_prog_bit_o          (prg_prog_bit),
    .prg_prog_bit_fall_o     (prg_prog_bit_fall),
    .prg_prog_bit_and_en_o   (prg_prog_bit_and_en),
    .prg_may_assert_dbgreq_o (prg_may_assert_dbgreq),
    .prg_branch_broadcast_o  (prg_branch_broadcast),
    .prg_ts_enable_o         (prg_ts_enable),
    .prg_port_mode_o         (prg_port_mode),
    .prg_port_size_o         (prg_port_size),
    .prg_etm_power_down_o    (prg_etm_power_down),
    .niden_t0_o              (niden_t0),
    .niden_t1_o              (niden_t1),
    .niden_t2_o              (niden_t2),
    .prg_stall_processor_o   (prg_stall_processor),
    .prg_etm_power_up_o      (prg_etm_power_up),
    .prg_lock_status_o       (prg_lock_status),
    .prg_claim_tag_o         (prg_claim_tag),
    .fifo_full_support_reg_o (fifo_full_support_reg)
     );

  cm3_etm_clk_gate #(ETMCLKGATE_PRESENT)
    u_etm_clk_gate
    (
    .clk                     (FCLK),
    .clk_enable_i            (prg_etm_power_up),
    .global_disable_i        (CGBYPASS),
    .clk_gated_o             (gated_clock)
     );

  assign clk_for_etm = (TRACE_LVL > 1) ? ((CLKGATE_PRESENT != 0) ? gated_clock : FCLK) : 1'b0;

  cm3_etm_trigger #(TRACE_LVL, RESET_ALL_REGS,0)
    u_etm_trigger
    (
    .clk                     (clk_for_etm),
    .reset_n                 (PORESETn),
    .trace_flushed_i         (trace_flushed),
    .prg_prog_bit_fall_i     (prg_prog_bit_fall),
    .prg_prog_bit_and_en_i   (prg_prog_bit_and_en),
    .etm_ivalid_i            (ETMIVALID),
    .etm_dvalid_i            (ETMDVALID),
    .prg_port_mode_i         (prg_port_mode),
    .prg_etm_en_i            (prg_etm_en),
    .prg_prog_bit_i          (prg_prog_bit),
    .prg_may_assert_dbgreq_i (prg_may_assert_dbgreq),
    .prg_ts_enable_i         (prg_ts_enable),
    .prg_branch_broadcast_i  (prg_branch_broadcast),
    .prg_stall_processor_i   (prg_stall_processor),
    .prg_port_size_i         (prg_port_size),
    .prg_etm_power_down_i    (prg_etm_power_down),
    .core_halt_i             (COREHALT),
    .niden_t0_i              (niden_t0),
    .niden_t1_i              (niden_t1),
    .ext_in_i                (EXTIN),
    .dwt_match_i             (DWTMATCH),
    .dwt_i_not_d_i           (DWTINOTD),
    .etm_rev_and_i           (etm_rev_and[3:0]),
    .reason_overflow_i       (reason_overflow),
    .max_ext_in_i            (MAXEXTIN),
    .fifo_full_support_reg_i (fifo_full_support_reg),
    .register_write_i        (register_write),
    .prg_lock_status_i       (prg_lock_status),
    .prg_claim_tag_i         (prg_claim_tag),
    .apb_wdata_i             (PWDATA[31:0]),
    .apb_addr_i              (apb_addr_reg),
    .power_down_reset_i      (power_down_reset),
    .atb_atreadym_i          (ATREADYM),
    .atb_atidm_o             (atb_atidm),
    .etm_dbg_req_o           (etm_dbg_req),
    .trigger_req_o           (trigger_req),
    .trace_enable_o          (trace_enable),
    .timestamp_event_o       (timestamp_event),
    .flush_trace_o           (flush_trace),
    .read_data_o             (read_data),
    .integration_mode_o      (integration_mode),
    .integration_atb_cr0_we_o(integration_atb_cr0_we),
    .integration_trigger_we_o(integration_trigger_we),
    .prg_fifo_full_level_o   (prg_fifo_full_level)
     );

  cm3_etm_trace #(TRACE_LVL, RESET_ALL_REGS)
    u_etm_trace
    (
    .clk                     (clk_for_etm),
    .reset_n                 (PORESETn),
    .prg_prog_bit_i          (prg_prog_bit),
    .prg_prog_bit_fall_i     (prg_prog_bit_fall),
    .etm_ia_i                (ETMIA[31:1]),
    .etm_fold_i              (ETMFOLD),
    .etm_cancel_i            (ETMCANCEL),
    .etm_ivalid_i            (ETMIVALID),
    .etm_istall_i            (ETMISTALL),
    .etm_cc_fail_i           (ETMICCFAIL),
    .etm_branch_i            (ETMIBRANCH),
    .etm_indirect_branch_i   (ETMIINDBR),
    .etm_isb_i               (ETMISB),
    .etm_int_stat_i          (ETMINTSTAT),
    .etm_int_num_i           (ETMINTNUM),
    .core_halt_i             (COREHALT),
    .trigger_req_i           (trigger_req),
    .pop_fifo_i              (pop_fifo),
    .prg_ts_enable_i         (prg_ts_enable),
    .prg_branch_broadcast_i  (prg_branch_broadcast),
    .prg_stall_processor_i   (prg_stall_processor),
    .trace_enable_i          (trace_enable),
    .flush_trace_i           (flush_trace),
    .power_down_reset_i      (power_down_reset),
    .integration_mode_i      (integration_mode),
    .apb_wdata_0_i           (PWDATA[0]),
    .integration_trigger_we_i(integration_trigger_we),
    .prg_fifo_full_level_i   (prg_fifo_full_level),
    .niden_t2_i              (niden_t2),
    .ts_value_i              (TSVALUEB),
    .ts_clk_change_i         (TSCLKCHANGE),
    .timestamp_event_i       (timestamp_event),
    .reason_overflow_o       (reason_overflow),
    .fifo_out_o              (fifo_out),
    .fifo_empty_o            (fifo_empty),
    .etm_trigger_out_o       (etm_trigger_out),
    .fifo_stall_o            (fifo_stall),
    .fake_ivalid_t1_o        (fake_ivalid_t1),
    .etm_istall_t1_o         (etm_istall_t1),
    .etm_ivalid_t1_o         (etm_ivalid_t1)
     );

  cm3_etm_trace_out #(TRACE_LVL)
    u_etm_trace_out
    (
    .clk                     (clk_for_etm),
    .reset_n                 (PORESETn),
    .fifo_out_i              (fifo_out),
    .fifo_empty_i            (fifo_empty),
    .atb_atreadym_i          (ATREADYM),
    .integration_mode_i      (integration_mode),
    .integration_atb_cr0_we_i(integration_atb_cr0_we),
    .apb_wdata_0_i           (PWDATA[0]),
    .atb_atdatam_o           (atb_atdatam[7:0]),
    .atb_atvalidm_o          (atb_atvalidm),
    .atb_afreadym_o          (atb_afreadym),
    .pop_fifo_o              (pop_fifo),
    .trace_flushed_o         (trace_flushed)
     );


`ifdef ARM_ASSERT_ON
 `include "std_ovl_defines.h"

`ifdef ETM_INTERFACE_TYPE
`else
`define ETM_INTERFACE_TYPE `OVL_ASSUME
`endif


  assert_never #(0,0,"ETM trace cannot be generated if TRACE_LVL == 0")
    cfg_etmtrace(clk_for_etm, PORESETn, (TRACE_LVL == 0) & ATVALIDM);

  wire   ovl_only_prog_bit;
  wire   ovl_progbit_ack;

  assign ovl_only_prog_bit = ((prg_ts_enable        == PWDATA[28]) &
                              (prg_port_size[3]     == PWDATA[21]) &
                              (prg_port_mode[1:0]   == PWDATA[17:16]) &
                              (prg_port_mode[2]     == PWDATA[13]) &
                              (prg_etm_en           == PWDATA[11]) &
                              (prg_may_assert_dbgreq== PWDATA[9]) &
                              (prg_branch_broadcast == PWDATA[8]) &
                              (prg_stall_processor  == PWDATA[7]) &
                              (prg_port_size[2:0]   == PWDATA[6:4]) &
                              (prg_etm_power_down   == PWDATA[0]));

  assert_implication
    #(`OVL_WARNING, `OVL_ASSUME,
      "When the prog bit is clear no other bit should be written")
      u_ovl_prog_bit_clear
        (.clk             (FCLK),
         .reset_n         (PORESETn),
         .antecedent_expr (register_write & (PADDR[11:2] == `AddrCtl) & ~trace_flushed),
         .consequent_expr (ovl_only_prog_bit));

`endif


`ifdef ARM_DSM

  assign u_etm_apb_if.dsm_reset_all_regs                = dsm_reset_all_regs;
  assign u_etm_trace.u_fifo.dsm_reset_all_regs          = dsm_reset_all_regs;
  assign u_etm_trace.u_control.dsm_reset_all_regs       = dsm_reset_all_regs;
  assign u_etm_trigger.u_res_control.dsm_reset_all_regs = dsm_reset_all_regs;

  `ifndef ARM_CMS_BUILD
    etmvaltracelogger_dsm
      u_etmvaltracelogger_dsm
        (
         .etm_log_off   (dsm_no_etm),
         .ID            (dsm_ID)
         );
  `endif
`endif

endmodule
