`include "cm3_etm_defs.vp"

module cm3_etm_control_reg(
  clk, reset_n, register_write_i, apb_wdata_i, apb_addr_i,
  fifo_full_en_i, fake_ivalid_t1_i,
  niden_i, etm_ivalid_i, etm_istall_i, etm_istall_t1_i, etm_ivalid_t1_i, etm_int_stat_i,
  power_down_reset_o, prg_etm_en_o, prg_prog_bit_o, prg_prog_bit_fall_o,
  prg_prog_bit_and_en_o, prg_may_assert_dbgreq_o, prg_branch_broadcast_o,
  prg_stall_processor_o, prg_etm_power_up_o, prg_etm_power_down_o,
  prg_ts_enable_o, prg_port_mode_o, prg_port_size_o, fifo_full_support_reg_o,
  prg_lock_status_o, prg_claim_tag_o, niden_t0_o, niden_t1_o, niden_t2_o
  );

  parameter TRACE_LVL = 1;

  input          clk;                     // un-gated ETM Clock
  input          reset_n;                 // ETM Reset

  input          register_write_i;        // Register Write Signal
  input [31:0]   apb_wdata_i;             // Register Write Data
  input [11:2]   apb_addr_i;              // Register Write Address
  input          fifo_full_en_i;          // System supports FIFO Full stalling
  input          niden_i;                 // Non-intrusive debug enable
  input          etm_ivalid_i;            // instruction valid
  input          etm_ivalid_t1_i;         // instruction valid
  input          fake_ivalid_t1_i;        // faked ivalid used with exceptions
  input          etm_istall_i;            // Previous instruction not yet confirmed
  input          etm_istall_t1_i;         // etm_istall in t1
  input [2:0]    etm_int_stat_i;          // Interrupt entry(1), exit(2) and return(3)

  output         power_down_reset_o;      // Initialise after powerdown
  output         prg_etm_en_o;            // Control reg enable bit
  output         prg_prog_bit_o;          // Programming Bit
  output         prg_prog_bit_fall_o;     // Falling edge of progbit
  output         prg_prog_bit_and_en_o;   // ProgBit with ETMEN factored in
  output         prg_may_assert_dbgreq_o; // Triggers set Dbgrq
  output         prg_branch_broadcast_o;  // Branch broadcasting enabled
  output         prg_stall_processor_o;   // Stall processor when Fifo fills
  output         prg_etm_power_up_o;      // ETM Powered Up
  output         prg_etm_power_down_o;    // Inverse of powered-up bit

  output         prg_ts_enable_o;         // Control register bit  {28}
  output [2:0]   prg_port_mode_o;         // Control register bits {17,16,13}
  output [3:0]   prg_port_size_o;         // Control register bits {21,6,5,4}
  output         fifo_full_support_reg_o; // Registered fifo full supported input pin
  output         prg_lock_status_o;       // Lock register statua
  output [3:0]   prg_claim_tag_o;         // Claim tag register
  output         niden_t0_o;              // Synchronised NIDEN input, T0
  output         niden_t1_o;              // Synchronised NIDEN input, T1
  output         niden_t2_o;              // Synchronised NIDEN input, T2

  reg            prg_ts_enable;
  reg [2:0]      prg_port_mode;
  reg            prg_etm_en;
  reg            prg_prog_bit;
  reg            prg_may_assert_dbgreq;
  reg            prg_branch_broadcast;
  reg            prg_stall_processor;
  reg [3:0]      prg_port_size;
  reg            prg_etm_power_down;
  reg            prg_etm_power_down_reg;

  wire           nxt_power_down_reset;
  wire           power_down_reset_we;     // Enable for power down reset flops

  wire           control_we;
  wire           nxt_prg_prog_bit_and_en;
  reg            fifo_full_support_reg;   // Fifo Full is supported in the system
  wire           fifo_full_support_we;

  wire           lock_status_we;
  wire           nxt_lock_status;
  wire           register_write_lock;

  wire           prg_claim_tag_we;
  wire [3:0]     nxt_prg_claim_tag;

  reg            power_down_reset;
  reg            prg_prog_bit_reg;
  reg            prg_prog_bit_and_en;
  wire           prg_prog_bit_fall;
  reg            prg_lock_status;
  reg [3:0]      prg_claim_tag;

  reg            niden_reg;
  wire           niden_t1;
  reg            niden_t2;
  reg            niden_t0;
  reg            niden_t0_reg;
  wire           opt_etm_en;

  assign opt_etm_en        = (TRACE_LVL > 1) ? 1'b1             : 1'b0;


  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        niden_t0  <= 1'b0;
        niden_reg <= 1'b0;
      end
    else if (opt_etm_en)
      begin
        niden_reg <= niden_i;
        niden_t0  <= niden_reg;
      end

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      niden_t0_reg <= 1'b0;
    else if (opt_etm_en)
      niden_t0_reg <= niden_t0;

  assign niden_t1 = (((niden_t0_reg | ~(etm_ivalid_i | fake_ivalid_t1_i)) & niden_t2) |
                         (niden_t0_reg &
                          (((etm_ivalid_t1_i | etm_istall_t1_i) & ~etm_istall_i &
                            ~(etm_int_stat_i == 3'b100)) |
                           fake_ivalid_t1_i)));

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      niden_t2 <= 1'b0;
    else if (opt_etm_en)
      niden_t2 <= niden_t1;

  assign fifo_full_support_we = (fifo_full_en_i ^ fifo_full_support_reg) & opt_etm_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      fifo_full_support_reg <= 1'b0;
    else if (fifo_full_support_we)
      fifo_full_support_reg <= fifo_full_en_i;

  assign nxt_power_down_reset = ~prg_etm_power_down & prg_etm_power_down_reg;
  assign power_down_reset_we  = (power_down_reset | nxt_power_down_reset |
                                 (prg_etm_power_down_reg ^ prg_etm_power_down)) &
                                  opt_etm_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        power_down_reset       <= 1'b0;
        prg_etm_power_down_reg <= 1'b1;
      end
    else if (power_down_reset_we)
      begin
        power_down_reset       <= nxt_power_down_reset;
        prg_etm_power_down_reg <= prg_etm_power_down;
      end

  always @(posedge clk or negedge reset_n)
    begin
      if (!reset_n)
        prg_prog_bit_reg <= 1'b0;
      else if (opt_etm_en)
        prg_prog_bit_reg <= prg_prog_bit;
    end

  assign prg_prog_bit_fall = prg_prog_bit_reg & ~prg_prog_bit;

  assign control_we = register_write_i & ~prg_lock_status &
                      (apb_addr_i[11:2] == `AddrCtl) & opt_etm_en;

  assign nxt_prg_prog_bit_and_en = ~prg_prog_bit & prg_etm_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      prg_prog_bit_and_en <= 1'b0;
    else if (opt_etm_en)
      prg_prog_bit_and_en <= nxt_prg_prog_bit_and_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        prg_ts_enable        <= 1'b0;
        prg_port_mode[2:0]   <= 3'b000;
        prg_etm_en           <= 1'b0;
        prg_prog_bit         <= 1'b1;
        prg_may_assert_dbgreq<= 1'b0;
        prg_branch_broadcast <= 1'b0;
        prg_stall_processor  <= 1'b0;
        prg_port_size[3:0]   <= 4'b0001;
        prg_etm_power_down   <= 1'b1;
      end
    else
      if (control_we)
        begin
          prg_ts_enable        <= apb_wdata_i[28];
          prg_port_size[3]     <= apb_wdata_i[21];
          prg_port_mode[1:0]   <= apb_wdata_i[17:16];
          prg_port_mode[2]     <= apb_wdata_i[13];
          prg_etm_en           <= apb_wdata_i[11];
          prg_prog_bit         <= apb_wdata_i[10];
          prg_may_assert_dbgreq<= apb_wdata_i[9];
          prg_branch_broadcast <= apb_wdata_i[8];
          prg_stall_processor  <= apb_wdata_i[7];
          prg_port_size[2:0]   <= apb_wdata_i[6:4];
          prg_etm_power_down   <= apb_wdata_i[0];
        end

  assign lock_status_we = (register_write_i & (apb_addr_i == `AddrLAccess)) & opt_etm_en;

  assign nxt_lock_status = ({32{lock_status_we}} & apb_wdata_i) != 32'hC5ACCE55;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      prg_lock_status <= 1'b1;
    else if (lock_status_we)
      prg_lock_status <= nxt_lock_status;

  assign register_write_lock = register_write_i & ~prg_lock_status;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      prg_claim_tag <= 4'h0;
    else if (prg_claim_tag_we)
      prg_claim_tag <= nxt_prg_claim_tag;

  assign prg_claim_tag_we   = (register_write_lock & opt_etm_en &
                               ( (apb_addr_i == `CM_ETM_CLAIMSET_ADD) |
                                 (apb_addr_i == `CM_ETM_CLAIMCLR_ADD) ) );
  assign nxt_prg_claim_tag  = (apb_addr_i == `CM_ETM_CLAIMSET_ADD) ?
                                (prg_claim_tag |  apb_wdata_i[3:0]) :
                                  (prg_claim_tag & ~apb_wdata_i[3:0]);

  assign power_down_reset_o      = (TRACE_LVL > 1) ? power_down_reset      : 1'b0;
  assign prg_etm_en_o            = (TRACE_LVL > 1) ? prg_etm_en            : 1'b0;
  assign prg_prog_bit_o          = (TRACE_LVL > 1) ? prg_prog_bit          : 1'b0;
  assign prg_prog_bit_and_en_o   = (TRACE_LVL > 1) ? prg_prog_bit_and_en   : 1'b0;
  assign prg_prog_bit_fall_o     = (TRACE_LVL > 1) ? prg_prog_bit_fall     : 1'b0;
  assign prg_may_assert_dbgreq_o = (TRACE_LVL > 1) ? prg_may_assert_dbgreq : 1'b0;
  assign prg_branch_broadcast_o  = (TRACE_LVL > 1) ? prg_branch_broadcast  : 1'b0;
  assign prg_stall_processor_o   = (TRACE_LVL > 1) ? prg_stall_processor   : 1'b0;
  assign prg_etm_power_up_o      = (TRACE_LVL > 1) ? ~prg_etm_power_down   : 1'b0;
  assign prg_etm_power_down_o    = (TRACE_LVL > 1) ? prg_etm_power_down    : 1'b0;
  assign prg_ts_enable_o         = (TRACE_LVL > 1) ? prg_ts_enable         : 1'b0;
  assign prg_port_mode_o         = (TRACE_LVL > 1) ? prg_port_mode         : 3'b000;
  assign prg_port_size_o         = (TRACE_LVL > 1) ? prg_port_size         : 4'b0000;
  assign fifo_full_support_reg_o = (TRACE_LVL > 1) ? fifo_full_support_reg : 1'b0;
  assign prg_lock_status_o       = (TRACE_LVL > 1) ? prg_lock_status       : 1'b0;
  assign prg_claim_tag_o         = (TRACE_LVL > 1) ? prg_claim_tag         : 4'b0000;
  assign niden_t0_o              = (TRACE_LVL > 1) ? niden_t0              : 1'b0;
  assign niden_t1_o              = (TRACE_LVL > 1) ? niden_t1              : 1'b0;
  assign niden_t2_o              = (TRACE_LVL > 1) ? niden_t2              : 1'b0;

endmodule
