`include "cm3_etm_defs.vp"

module cm3_etm_fifo (
  reset_n, clk, fifo_in_i, num_bytes_i, pop_fifo_i, async_req_i,
  prg_fifo_full_level_i, trigger_req_i, prg_stall_processor_i,
  prg_prog_bit_i,
  fifo_empty_o, fifo_overflow_o, fifo_out_o, fifo_stall_o, fifo_ts_holdoff_o
  );

  parameter TRACE_LVL      = 1;
  parameter RESET_ALL_REGS = 0;

  input         reset_n;           // ETM reset
  input         clk;               // Gated ETM Clock

  input [63:0]  fifo_in_i;         // 0-8 bytes of data to store (from FIFO in)
  input [3:0]   num_bytes_i;       // Number of bytes to place in FIFO
  input         pop_fifo_i;        // Request FIFO output
  input         async_req_i;       // Request Alignment Sync
  input         trigger_req_i;     // Request Trigger
  input  [(`FIFO_DEPTH_SIZE-1):0]   prg_fifo_full_level_i; // Fifo Full level register
  input         prg_stall_processor_i; // Enable FifoStall output
  input         prg_prog_bit_i;    // Programming state

  output        fifo_empty_o;      // asserted when FIFO is empty
  output        fifo_overflow_o;   // asserted when there is not enough
  output [7:0]  fifo_out_o;        // Fifo O/P
  output        fifo_stall_o;      // Fifo Depth is above watermark on prg_fifo_full_level_i
  output        fifo_ts_holdoff_o; // Fifo is 16 or more deep, or 8 bytes written (for timestamp)



  wire [63:0]   data_rotated;      // FIFO input (rotated)
  wire [ 7:0]   byte_we;           // 1 active bit per byte each byte

  reg [`FIFO_WIDTHROWS-1:0]     opt_write_row_ptr;
  reg [2:0]                     opt_write_column; // fifo byte write pointer
  wire [`FIFO_WIDTHROWS-1:0]    write_row_ptr;
  wire [2:0]                    write_column;

  wire [`FIFO_WIDTHROWS-1:0]    nxt_write_row_ptr;
  wire [2:0]                    nxt_write_column;

  wire [`FIFO_WIDTHROWS-1:0]     write_row_ptr_after;

  wire [(`FIFO_DEPTH_SIZE-1):0]  fifo_depth; // Number of bytes in FIFO

  wire           fifo_overflow;

  wire [(`FIFO_POINTER_SIZE-1):0] nxt_rd_byte_ptr_if_wrap;
  wire [(`FIFO_POINTER_SIZE-1):0] nxt_rd_byte_ptr;
  reg [(`FIFO_POINTER_SIZE-1):0]  opt_rd_byte_ptr;
  reg                             opt_rd_byte_ptr_msb;
  wire [(`FIFO_POINTER_SIZE-1):0] rd_byte_ptr;
  wire                            rd_byte_ptr_msb;
  wire                            nxt_rd_byte_ptr_msb;
  wire                            fifo_read_has_wrapped;

  reg  [(`FIFO_DEPTH_SIZE-1):0]   opt_async_count;
  wire [(`FIFO_DEPTH_SIZE-1):0]   async_count;
  wire                            async_count_we;
  wire [(`FIFO_DEPTH_SIZE-1):0]   nxt_async_count;
  wire [(`FIFO_DEPTH_SIZE-1):0]   async_count_dec;
  wire                            async_active; // Async is being output
  wire                            async_idle; // Async counter not running
  reg  [(`FIFO_DEPTH_SIZE-1):0]   opt_trigger_count;
  wire [(`FIFO_DEPTH_SIZE-1):0]   trigger_count;
  wire                            trigger_count_we;
  wire [(`FIFO_DEPTH_SIZE-1):0]   nxt_trigger_count;
  wire [(`FIFO_DEPTH_SIZE-1):0]   trigger_count_dec;
  wire                            trigger_active; // Trigger is being output
  wire                            trigger_idle; // Trigger counter idle

  reg  [7:0] fifo_r0_b0_reg;
  reg  [7:0] fifo_r0_b1_reg;
  reg  [7:0] fifo_r0_b2_reg;
  reg  [7:0] fifo_r0_b3_reg;
  reg  [7:0] fifo_r0_b4_reg;
  reg  [7:0] fifo_r0_b5_reg;
  reg  [7:0] fifo_r0_b6_reg;
  reg  [7:0] fifo_r0_b7_reg;
  reg  [7:0] fifo_r1_b0_reg;
  reg  [7:0] fifo_r1_b1_reg;
  reg  [7:0] fifo_r1_b2_reg;
  reg  [7:0] fifo_r1_b3_reg;
  reg  [7:0] fifo_r1_b4_reg;
  reg  [7:0] fifo_r1_b5_reg;
  reg  [7:0] fifo_r1_b6_reg;
  reg  [7:0] fifo_r1_b7_reg;
  reg  [7:0] fifo_r2_b0_reg;
  reg  [7:0] fifo_r2_b1_reg;
  reg  [7:0] fifo_r2_b2_reg;
  reg  [7:0] fifo_r2_b3_reg;
  reg  [7:0] fifo_r2_b4_reg;
  reg  [7:0] fifo_r2_b5_reg;
  reg  [7:0] fifo_r2_b6_reg;
  reg  [7:0] fifo_r2_b7_reg;

  wire [`FIFO_POINTER_SIZE-1:0]    wr_byte_ptr;
  reg  [`FIFO_POINTER_SIZE-1:0]    opt_wr_byte_ptr;
  wire [`FIFO_POINTER_SIZE-1:0]    full_wr_byte_ptr;
  wire          write_fifo_has_wrapped;
  wire [`FIFO_POINTER_SIZE-1:0]    wrap_wr_byte_ptr;
  wire       nxt_wr_wrap_fifo_msb;
  wire       wr_wrap_fifo_msb_we;
  reg        opt_wr_wrap_fifo_msb;
  wire       wr_wrap_fifo_msb;
  wire       write_pointer_we;
  wire [`FIFO_POINTER_SIZE-1:0]    nxt_wr_byte_ptr;
  wire       nxt_fifo_has_no_data; // Indicates there is data in the FIFO
  reg        opt_fifo_has_no_data; // Indicates there is data in the FIFO
  wire       fifo_has_no_data;
  wire       wr_ptr_greater;
  wire       fifo_r0_b0_we;
  wire       fifo_r0_b1_we;
  wire       fifo_r0_b2_we;
  wire       fifo_r0_b3_we;
  wire       fifo_r0_b4_we;
  wire       fifo_r0_b5_we;
  wire       fifo_r0_b6_we;
  wire       fifo_r0_b7_we;
  wire       fifo_r1_b0_we;
  wire       fifo_r1_b1_we;
  wire       fifo_r1_b2_we;
  wire       fifo_r1_b3_we;
  wire       fifo_r1_b4_we;
  wire       fifo_r1_b5_we;
  wire       fifo_r1_b6_we;
  wire       fifo_r1_b7_we;
  wire       fifo_r2_b0_we;
  wire       fifo_r2_b1_we;
  wire       fifo_r2_b2_we;
  wire       fifo_r2_b3_we;
  wire       fifo_r2_b4_we;
  wire       fifo_r2_b5_we;
  wire       fifo_r2_b6_we;
  wire       fifo_r2_b7_we;
  wire [7:0] read_mux_byte0;       // FIFO Data Mux 0
  wire [7:0] read_mux_byte1;       // FIFO Data Mux 1
  wire [7:0] read_mux_byte2;       // FIFO Data Mux 2
  wire [7:0] read_mux_byte3;       // FIFO Data Mux 3
  wire [7:0] read_mux_byte4;       // FIFO Data Mux 0
  wire [7:0] read_mux_byte5;       // FIFO Data Mux 1
  wire [7:0] read_mux_byte6;       // FIFO Data Mux 2
  wire [7:0] read_mux_byte7;       // FIFO Data Mux 3

  wire [1:0] read_mux_row;
  wire       rd_ptr_inc;           // Valid pop_fifo_i seen
  wire [3:0] last_column_to_write; // Un-wrapped MSB column to write
  wire [7:0] fifo_byte_preselect;  // Byte strobes before overflow
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select0;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select1;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select2;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select3;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select4;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select5;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select6;
  wire [`FIFO_WIDTHROWS-1:0] fifo_row_select7;
  wire [3:0] full_write_column;
  wire       write_wrap_row;
  wire       nxt_fifo_stall;

  wire       fifo_empty;
  reg [7:0]  opt_fifo_out;
  reg        opt_fifo_stall;
  wire       opt_fifo_empty;
  wire       opt_fifo_overflow;
  wire [63:0]   rotate_1;          // Rotate Stage 1
  wire [63:0]   rotate_2;          // Rotate Stage 2
  reg [2:0]     rotation;  // Rotation value registered

  wire       opt_etm_en;   // Enable term for logic removal
  wire       opt_reset_n;  // Optional reset for all registers
  wire       write_row_ptr_en;     // Row pointer flop enable

`ifdef ARM_DSM
  wire   dsm_reset_all_regs;

  assign opt_reset_n = dsm_reset_all_regs    ? reset_n : 1'b1;

`else

  assign opt_reset_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

`endif

  assign opt_etm_en       = (TRACE_LVL > 1) ? 1'b1              : 1'b0;
  assign fifo_overflow    = (TRACE_LVL > 1) ? opt_fifo_overflow : 1'b0;
  assign fifo_empty       = (TRACE_LVL > 1) ? opt_fifo_empty    : 1'b0;


  assign nxt_fifo_stall = prg_stall_processor_i & ~prg_prog_bit_i &
                          (fifo_depth > (`FIFO_SIZE - prg_fifo_full_level_i));

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_fifo_stall <= 1'b0;
    else if (opt_etm_en)
      opt_fifo_stall <= nxt_fifo_stall;

  assign write_row_ptr_after = (write_row_ptr == 2'b10) ? 2'b00 :
                               write_row_ptr + 2'b01;

  assign last_column_to_write = {1'b0,write_column} + num_bytes_i;

  assign fifo_byte_preselect[0]  = (3'd0 >= write_column) ?
                                   (4'd0 < last_column_to_write) :
                                     (4'd8 < last_column_to_write);
  assign fifo_byte_preselect[1]  = (3'd1 >= write_column) ?
                                   (4'd1 < last_column_to_write) :
                                     (4'd9 < last_column_to_write);
  assign fifo_byte_preselect[2]  = (3'd2 >= write_column) ?
                                   (4'd2 < last_column_to_write) :
                                     (4'd10 < last_column_to_write);
  assign fifo_byte_preselect[3]  = (3'd3 >= write_column) ?
                                   (4'd3 < last_column_to_write) :
                                     (4'd11 < last_column_to_write);
  assign fifo_byte_preselect[4]  = (3'd4 >= write_column) ?
                                   (4'd4 < last_column_to_write) :
                                     (4'd12 < last_column_to_write);
  assign fifo_byte_preselect[5]  = (3'd5 >= write_column) ?
                                   (4'd5 < last_column_to_write) :
                                     (4'd13 < last_column_to_write);
  assign fifo_byte_preselect[6]  = (3'd6 >= write_column) ?
                                   (4'd6 < last_column_to_write) :
                                     (4'd14 < last_column_to_write);
  assign fifo_byte_preselect[7]  =  (4'd7 < last_column_to_write);

  assign byte_we = {8{~fifo_overflow}} & fifo_byte_preselect[7:0];


  assign fifo_row_select0 = (3'd0 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select1 = (3'd1 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select2 = (3'd2 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select3 = (3'd3 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select4 = (3'd4 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select5 = (3'd5 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select6 = (3'd6 >= write_column) ? write_row_ptr : write_row_ptr_after;
  assign fifo_row_select7 = write_row_ptr;

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      rotation <= 3'b000;
    else if (opt_etm_en)
      rotation <= full_write_column[2:0];

  assign rotate_1  = rotation[0] ? {fifo_in_i[55:0], fifo_in_i[63:56]} :
                                  fifo_in_i[63:0];

  assign rotate_2  = rotation[1] ? {rotate_1[47:0], rotate_1[63:48]} :
                                  rotate_1[63:0];

  assign data_rotated  = rotation[2] ? {rotate_2[31:0], rotate_2[63:32]} :
                                      rotate_2[63:0];

  assign full_write_column = last_column_to_write;

  assign write_wrap_row = (full_write_column >= 4'd8);
  assign nxt_write_column = full_write_column[2:0];
  assign nxt_write_row_ptr = ((TRACE_LVL > 1) ?
                              write_row_ptr_after :
                              {`FIFO_WIDTHROWS{1'b0}});

  assign full_wr_byte_ptr = wr_byte_ptr + {1'b0, num_bytes_i};

  assign write_fifo_has_wrapped = full_wr_byte_ptr >= `FIFO_SIZE;
  assign wrap_wr_byte_ptr = (full_wr_byte_ptr > 5'd23) ?
                              (full_wr_byte_ptr - `FIFO_SIZE) : full_wr_byte_ptr;
  assign nxt_wr_wrap_fifo_msb = ((TRACE_LVL > 1) ?
                                 ~wr_wrap_fifo_msb :
                                 1'b0);
  assign wr_wrap_fifo_msb_we = opt_etm_en & ~fifo_overflow & write_fifo_has_wrapped;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_wr_wrap_fifo_msb <= 1'b0;
    else if (wr_wrap_fifo_msb_we)
      opt_wr_wrap_fifo_msb <= nxt_wr_wrap_fifo_msb;

  assign wr_wrap_fifo_msb = (TRACE_LVL > 1) ? opt_wr_wrap_fifo_msb : 1'b0;

  assign nxt_wr_byte_ptr = ((TRACE_LVL > 1) ?
                            (wrap_wr_byte_ptr[`FIFO_POINTER_SIZE-1:0]) :
                            {`FIFO_POINTER_SIZE{1'b0}});
  assign write_pointer_we = opt_etm_en & ~fifo_overflow;


  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        opt_write_column     <= 3'b000;
        opt_wr_byte_ptr      <= {`FIFO_POINTER_SIZE{1'b0}};
      end
    else if (write_pointer_we)
      begin
        opt_write_column     <= nxt_write_column[2:0];
        opt_wr_byte_ptr      <= nxt_wr_byte_ptr;
      end

  assign write_row_ptr_en = opt_etm_en & write_wrap_row & ~fifo_overflow;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_write_row_ptr    <= {`FIFO_WIDTHROWS{1'b0}};
    else if (write_row_ptr_en)
      opt_write_row_ptr    <= nxt_write_row_ptr;

  assign write_column  = (TRACE_LVL > 1) ? opt_write_column  : 3'b000;
  assign wr_byte_ptr   = (TRACE_LVL > 1) ? opt_wr_byte_ptr   : {`FIFO_POINTER_SIZE{1'b0}};
  assign write_row_ptr = (TRACE_LVL > 1) ? opt_write_row_ptr : {`FIFO_WIDTHROWS{1'b0}};

  assign nxt_fifo_has_no_data = ((TRACE_LVL > 1) ?
                                 (((wr_byte_ptr == nxt_rd_byte_ptr) & rd_ptr_inc) |
                                   fifo_has_no_data) &
                                 ~|num_bytes_i &
                                 (fifo_read_has_wrapped |
                                  (wr_wrap_fifo_msb == rd_byte_ptr_msb)) :
                                 1'b0);

  assign opt_fifo_empty = fifo_has_no_data & ~async_active & ~trigger_active;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_fifo_has_no_data <= 1'b1;
    else if (opt_etm_en & ~fifo_overflow)
      opt_fifo_has_no_data <= nxt_fifo_has_no_data;

  assign fifo_has_no_data = (TRACE_LVL > 1) ? opt_fifo_has_no_data : 1'b0;

  assign wr_ptr_greater = (wr_byte_ptr > rd_byte_ptr) | fifo_has_no_data;

  assign fifo_depth = ((wr_ptr_greater) ?
                      (wr_byte_ptr -
                       rd_byte_ptr) :
                      (`FIFO_SIZE - rd_byte_ptr +
                       wr_byte_ptr));

  assign opt_fifo_overflow = (((|fifo_depth) ?
                               (fifo_depth > (`FIFO_SIZE -
                                              {{(`FIFO_DEPTH_SIZE-4){1'b0}},
                                               num_bytes_i})) :
                               1'b0));

  assign async_idle = (async_count[(`FIFO_DEPTH_SIZE-1):1]=={(`FIFO_DEPTH_SIZE-1){1'b0}});

  assign async_count_dec = async_idle ? {(`FIFO_DEPTH_SIZE){1'b0}} :
                                        async_count -
                                        {{(`FIFO_DEPTH_SIZE-1){1'b0}},1'b1};

  assign nxt_async_count = (async_idle & async_req_i) ?
                             (fifo_depth + `ASYNC_COUNT_OFFSET+
                              {{`FIFO_DEPTH_SIZE - 1{1'b0}},(~pop_fifo_i | fifo_empty)}):
                               async_count_dec;
  assign async_count_we = opt_etm_en & (pop_fifo_i | (async_idle & async_req_i));

  assign async_active = ~((|async_count[(`FIFO_DEPTH_SIZE-1):3])
                          |(async_count[2:1] == 2'b00));

  assign trigger_idle = (trigger_count == {(`FIFO_DEPTH_SIZE){1'b0}});

  assign trigger_count_dec = trigger_idle ? {(`FIFO_DEPTH_SIZE){1'b0}} :
                               trigger_count -
                               {{(`FIFO_DEPTH_SIZE-1){1'b0}},1'b1};

  reg    trigger_reg;
  wire   nxt_trigger_reg;
  assign nxt_trigger_reg = trigger_idle & trigger_req_i;

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      trigger_reg <= 1'b0;
    else if(opt_etm_en)
      trigger_reg <= nxt_trigger_reg;

  assign nxt_trigger_count = (trigger_reg) ?
                               (fifo_depth +
                                {{`FIFO_DEPTH_SIZE - 1{1'b0}},
                                 ~pop_fifo_i | fifo_has_no_data}):
                                 trigger_count_dec;
  assign trigger_count_we = opt_etm_en & ((trigger_reg) |
                                          (pop_fifo_i & ~async_active));

  assign trigger_active = (trigger_count ==
                           {{(`FIFO_DEPTH_SIZE-1){1'b0}}, 1'b1});



  assign fifo_r0_b0_we = byte_we[ 0] & (fifo_row_select0  == 2'b00) & opt_etm_en;
  assign fifo_r0_b1_we = byte_we[ 1] & (fifo_row_select1  == 2'b00) & opt_etm_en;
  assign fifo_r0_b2_we = byte_we[ 2] & (fifo_row_select2  == 2'b00) & opt_etm_en;
  assign fifo_r0_b3_we = byte_we[ 3] & (fifo_row_select3  == 2'b00) & opt_etm_en;
  assign fifo_r0_b4_we = byte_we[ 4] & (fifo_row_select4  == 2'b00) & opt_etm_en;
  assign fifo_r0_b5_we = byte_we[ 5] & (fifo_row_select5  == 2'b00) & opt_etm_en;
  assign fifo_r0_b6_we = byte_we[ 6] & (fifo_row_select6  == 2'b00) & opt_etm_en;
  assign fifo_r0_b7_we = byte_we[ 7] & (fifo_row_select7  == 2'b00) & opt_etm_en;

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b0_reg <= {8{1'b0}};
    else if (fifo_r0_b0_we)
      fifo_r0_b0_reg <= data_rotated[7:0];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b1_reg <= {8{1'b0}};
    else if (fifo_r0_b1_we)
      fifo_r0_b1_reg <= data_rotated[15:8];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b2_reg <= {8{1'b0}};
    else if (fifo_r0_b2_we)
      fifo_r0_b2_reg <= data_rotated[23:16];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b3_reg <= {8{1'b0}};
    else if (fifo_r0_b3_we)
      fifo_r0_b3_reg <= data_rotated[31:24];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b4_reg <= {8{1'b0}};
    else if (fifo_r0_b4_we)
      fifo_r0_b4_reg <= data_rotated[39:32];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b5_reg <= {8{1'b0}};
    else if (fifo_r0_b5_we)
      fifo_r0_b5_reg <= data_rotated[47:40];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b6_reg <= {8{1'b0}};
    else if (fifo_r0_b6_we)
      fifo_r0_b6_reg <= data_rotated[55:48];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r0_b7_reg <= {8{1'b0}};
    else if (fifo_r0_b7_we)
      fifo_r0_b7_reg <= data_rotated[63:56];



  assign fifo_r1_b0_we = byte_we[ 0] & (fifo_row_select0  == 2'b01) & opt_etm_en;
  assign fifo_r1_b1_we = byte_we[ 1] & (fifo_row_select1  == 2'b01) & opt_etm_en;
  assign fifo_r1_b2_we = byte_we[ 2] & (fifo_row_select2  == 2'b01) & opt_etm_en;
  assign fifo_r1_b3_we = byte_we[ 3] & (fifo_row_select3  == 2'b01) & opt_etm_en;
  assign fifo_r1_b4_we = byte_we[ 4] & (fifo_row_select4  == 2'b01) & opt_etm_en;
  assign fifo_r1_b5_we = byte_we[ 5] & (fifo_row_select5  == 2'b01) & opt_etm_en;
  assign fifo_r1_b6_we = byte_we[ 6] & (fifo_row_select6  == 2'b01) & opt_etm_en;
  assign fifo_r1_b7_we = byte_we[ 7] & (fifo_row_select7  == 2'b01) & opt_etm_en;

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b0_reg <= {8{1'b0}};
    else if (fifo_r1_b0_we)
      fifo_r1_b0_reg <= data_rotated[7:0];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b1_reg <= {8{1'b0}};
    else if (fifo_r1_b1_we)
      fifo_r1_b1_reg <= data_rotated[15:8];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b2_reg <= {8{1'b0}};
    else if (fifo_r1_b2_we)
      fifo_r1_b2_reg <= data_rotated[23:16];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b3_reg <= {8{1'b0}};
    else if (fifo_r1_b3_we)
      fifo_r1_b3_reg <= data_rotated[31:24];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b4_reg <= {8{1'b0}};
    else if (fifo_r1_b4_we)
      fifo_r1_b4_reg <= data_rotated[39:32];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b5_reg <= {8{1'b0}};
    else if (fifo_r1_b5_we)
      fifo_r1_b5_reg <= data_rotated[47:40];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b6_reg <= {8{1'b0}};
    else if (fifo_r1_b6_we)
      fifo_r1_b6_reg <= data_rotated[55:48];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r1_b7_reg <= {8{1'b0}};
    else if (fifo_r1_b7_we)
      fifo_r1_b7_reg <= data_rotated[63:56];



  assign fifo_r2_b0_we = byte_we[ 0] & (fifo_row_select0  == 2'b10) & opt_etm_en;
  assign fifo_r2_b1_we = byte_we[ 1] & (fifo_row_select1  == 2'b10) & opt_etm_en;
  assign fifo_r2_b2_we = byte_we[ 2] & (fifo_row_select2  == 2'b10) & opt_etm_en;
  assign fifo_r2_b3_we = byte_we[ 3] & (fifo_row_select3  == 2'b10) & opt_etm_en;
  assign fifo_r2_b4_we = byte_we[ 4] & (fifo_row_select4  == 2'b10) & opt_etm_en;
  assign fifo_r2_b5_we = byte_we[ 5] & (fifo_row_select5  == 2'b10) & opt_etm_en;
  assign fifo_r2_b6_we = byte_we[ 6] & (fifo_row_select6  == 2'b10) & opt_etm_en;
  assign fifo_r2_b7_we = byte_we[ 7] & (fifo_row_select7  == 2'b10) & opt_etm_en;

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b0_reg <= {8{1'b0}};
    else if (fifo_r2_b0_we)
      fifo_r2_b0_reg <= data_rotated[7:0];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b1_reg <= {8{1'b0}};
    else if (fifo_r2_b1_we)
      fifo_r2_b1_reg <= data_rotated[15:8];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b2_reg <= {8{1'b0}};
    else if (fifo_r2_b2_we)
      fifo_r2_b2_reg <= data_rotated[23:16];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b3_reg <= {8{1'b0}};
    else if (fifo_r2_b3_we)
      fifo_r2_b3_reg <= data_rotated[31:24];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b4_reg <= {8{1'b0}};
    else if (fifo_r2_b4_we)
      fifo_r2_b4_reg <= data_rotated[39:32];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b5_reg <= {8{1'b0}};
    else if (fifo_r2_b5_we)
      fifo_r2_b5_reg <= data_rotated[47:40];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b6_reg <= {8{1'b0}};
    else if (fifo_r2_b6_we)
      fifo_r2_b6_reg <= data_rotated[55:48];

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      fifo_r2_b7_reg <= {8{1'b0}};
    else if (fifo_r2_b7_we)
      fifo_r2_b7_reg <= data_rotated[63:56];

  assign read_mux_row = rd_byte_ptr[4:3];

  assign read_mux_byte0 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b0_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b0_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b0_reg));

  assign read_mux_byte1 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b1_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b1_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b1_reg));

  assign read_mux_byte2 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b2_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b2_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b2_reg));

  assign read_mux_byte3 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b3_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b3_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b3_reg));

  assign read_mux_byte4 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b4_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b4_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b4_reg));

  assign read_mux_byte5 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b5_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b5_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b5_reg));

  assign read_mux_byte6 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b6_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b6_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b6_reg));

  assign read_mux_byte7 = (({8{(read_mux_row == 2'b00)}} & fifo_r0_b7_reg) |
                           ({8{(read_mux_row == 2'b01)}} & fifo_r1_b7_reg) |
                           ({8{(read_mux_row == 2'b10)}} & fifo_r2_b7_reg));

  always @ (async_count or rd_byte_ptr or read_mux_byte0
            or read_mux_byte1 or read_mux_byte2 or read_mux_byte3 or read_mux_byte4
            or read_mux_byte5 or read_mux_byte6 or read_mux_byte7 or async_active
            or trigger_active)
    if(async_active)
      case (async_count[2:0])
        3'b111 : opt_fifo_out = 8'h00;
        3'b110 : opt_fifo_out = 8'h00;
        3'b101 : opt_fifo_out = 8'h00;
        3'b100 : opt_fifo_out = 8'h00;
        3'b011 : opt_fifo_out = 8'h00;
        3'b010 : opt_fifo_out = 8'h80;
        3'b001 : opt_fifo_out = {8{1'bx}}; // Not used
        3'b000 : opt_fifo_out = {8{1'bx}}; // Not used
        default : opt_fifo_out = {8{1'bx}};
      endcase
    else if(trigger_active)
      opt_fifo_out = `Tr_HEADER;
    else
      case (rd_byte_ptr[2:0])
        3'b000 : opt_fifo_out = read_mux_byte0;
        3'b001 : opt_fifo_out = read_mux_byte1;
        3'b010 : opt_fifo_out = read_mux_byte2;
        3'b011 : opt_fifo_out = read_mux_byte3;
        3'b100 : opt_fifo_out = read_mux_byte4;
        3'b101 : opt_fifo_out = read_mux_byte5;
        3'b110 : opt_fifo_out = read_mux_byte6;
        3'b111 : opt_fifo_out = read_mux_byte7;
        default : opt_fifo_out = {8{1'bX}};
      endcase // case(RdBytePtr[1:0])


  assign rd_ptr_inc = pop_fifo_i & ~fifo_has_no_data & ~async_active & ~trigger_active;

  assign fifo_read_has_wrapped = (rd_ptr_inc) & (rd_byte_ptr == 5'd23) & opt_etm_en;

  assign nxt_rd_byte_ptr_if_wrap = {`FIFO_POINTER_SIZE{~fifo_read_has_wrapped}} &
                                   (rd_byte_ptr + 5'b00001);

  assign nxt_rd_byte_ptr = ((TRACE_LVL > 1) ?
                         (nxt_rd_byte_ptr_if_wrap[`FIFO_POINTER_SIZE-1:0]) :
                         {`FIFO_POINTER_SIZE{1'b0}});
  assign nxt_rd_byte_ptr_msb = ((TRACE_LVL > 1) ?
                                (~rd_byte_ptr_msb) :
                                1'b0);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        opt_rd_byte_ptr     <= {`FIFO_POINTER_SIZE{1'b0}};
      end
    else if (opt_etm_en & rd_ptr_inc)
      begin
        opt_rd_byte_ptr     <= nxt_rd_byte_ptr[`FIFO_POINTER_SIZE-1:0];
   end

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_rd_byte_ptr_msb <= 1'b0;
    else if (fifo_read_has_wrapped)
      opt_rd_byte_ptr_msb <= nxt_rd_byte_ptr_msb;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_async_count     <= {`FIFO_DEPTH_SIZE{1'b0}};
    else if (async_count_we)
      opt_async_count     <= nxt_async_count;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_trigger_count   <= {`FIFO_DEPTH_SIZE{1'b0}};
    else if (trigger_count_we)
      opt_trigger_count   <= nxt_trigger_count;

  assign rd_byte_ptr      = ((TRACE_LVL > 1) ?
                            opt_rd_byte_ptr : {`FIFO_POINTER_SIZE{1'b0}});
  assign rd_byte_ptr_msb  = ((TRACE_LVL > 1) ?
                            opt_rd_byte_ptr_msb : 1'b0);
  assign async_count      = ((TRACE_LVL > 1) ?
                            opt_async_count : {`FIFO_DEPTH_SIZE{1'b0}});
  assign trigger_count    = ((TRACE_LVL > 1) ?
                            opt_trigger_count : {`FIFO_DEPTH_SIZE{1'b0}});

  assign fifo_empty_o     = fifo_empty;
  assign fifo_overflow_o  = fifo_overflow;
  assign fifo_out_o       = (TRACE_LVL > 1) ? opt_fifo_out      : {8{1'b0}};
  assign fifo_stall_o     = (TRACE_LVL > 1) ? opt_fifo_stall    : 1'b0;
  assign fifo_ts_holdoff_o  = (TRACE_LVL > 1) ? fifo_depth[4]   : 1'b0;

`ifdef ARM_ASSERT_ON

  assert_never #(0, 0, "fifo_depth greater than size")
  ovl_fifo_depth_gt_max (
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (fifo_depth > `FIFO_SIZE));

  assert_never #(0, 0, "Write pointer overflow")
  ovl_write_pointer_gt_max (
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (wr_byte_ptr > `FIFO_SIZE));

  assert_never_unknown #(0, 1,0, "fifo_overflow must not be X")
  ovl_fifo_overflow_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (fifo_overflow));

  assert_never_unknown #(0, 1,0, "write_row_ptr_en must not be X")
  ovl_write_row_ptr_en_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (write_row_ptr_en));

  assert_never_unknown #(0, 1,0, "async_active must not be X")
  ovl_async_active_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (async_active));

  assert_never_unknown #(0, 1,0, "trigger_active must not be X")
  ovl_trigger_active_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (trigger_active));

  assert_never_unknown #(0, 1,0, "fifo_read_has_wrapped must not be X")
  ovl_fifo_read_has_wrapped_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (fifo_read_has_wrapped));

  assert_never_unknown #(0, 1,0, "wr_wrap_fifo_msb_we must not be X")
  ovl_wr_wrap_fifo_msb_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (wr_wrap_fifo_msb_we));

  assert_never_unknown #(0, 1,0, "write_pointer_we must not be X")
  ovl_write_pointer_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (write_pointer_we));


`endif

`include "cm3_etm_undefs.vp"

endmodule
