`include "cm3_etm_defs.vp"

module cm3_etm_gen (
  clk, reset_n, etm_ivalid_i, etm_istall_i, etm_ia_i, etm_cc_fail_i, etm_branch_i,
  etm_indirect_branch_i, etm_isb_i, etm_int_stat_i, etm_fold_i, etm_cancel_i,
  etm_int_num_i, core_halt_i, prg_prog_bit_fall_i, prg_branch_broadcast_i,
  prg_ts_enable_i, isync_req_i, trigger_req_t2_i, fifo_overflow_i, fifo_empty_i,
  fifo_ts_holdoff_i, trace_enable_i, niden_t2_i, ts_value_i,
  ts_clk_change_i, timestamp_event_i, flush_trace_i, power_down_reset_i,
  integration_mode_i, apb_wdata_0_i, integration_trigger_we_i,
  fifo_in_o, num_bytes_o, reason_overflow_o, etm_trigger_out_o, fake_ivalid_t1_o,
  etm_istall_t1_o, etm_ivalid_t1_o
  );

  parameter TRACE_LVL = 1;
  parameter RESET_ALL_REGS = 0;


  input         clk;                // Gated ETM Clock
  input         reset_n;            // ETM Reset

  input         etm_ivalid_i;       // Instruction data is valid
  input         etm_istall_i;       // IVALID from prev. cycle should be stalled
  input         etm_fold_i;         // Two instructions execute this cycle
  input [31:1]  etm_ia_i;           // Instruction Address
  input         etm_cc_fail_i;      // Instruction failed condition code
  input         etm_branch_i;       // Instruction is Branch
  input         etm_indirect_branch_i; // Branch is indirect branch
  input         etm_isb_i;          // Instruction is ISB
  input [2:0]   etm_int_stat_i;     // Interrupt entry(1), exit(2) and return(3)

  input         etm_cancel_i;       // Previous instruction did not execute

  input [8:0]   etm_int_num_i;      // Exception is ISRn
  input         core_halt_i;        // Core has entered debug state
  input         prg_prog_bit_fall_i;// Falling edge of progbit


  input         prg_branch_broadcast_i; // All branch addresses should be output
  input         prg_ts_enable_i;    // Enable timestamp insertion
  input         isync_req_i;        // Request for ISYNC Packet
  input         trigger_req_t2_i;   // Request for trigger
  input         fifo_overflow_i;    // Push was rejected by FIFO
  input         fifo_empty_i;       // Fifo is not outputting data
  input         fifo_ts_holdoff_i;  // Fifo is 16 bytes or more deep
  input         trace_enable_i;     // Trace control input
  input         flush_trace_i;      // Force trace to flush without needing IVALID
  input         power_down_reset_i; // ETM has been powered up and should reset

  input         integration_mode_i; // Integration mode control
  input         integration_trigger_we_i; // Write to trigger out register
  input         apb_wdata_0_i;      // APB write data

  input         niden_t2_i;         // NIDEN in T2 stage
  input [47:0]  ts_value_i;         // Binary coded timestamp clock value
  input         ts_clk_change_i;    // Timestamp clock ratio changed
  input         timestamp_event_i;  // Timestamp event


  output [63:0] fifo_in_o;          // 8 bytes to FIFO
  output [3:0]  num_bytes_o;        // Number of bytes to FIFO

  output        reason_overflow_o;  // Pending Overflow
  output        etm_trigger_out_o;  // Trigger has been generated
  output        fake_ivalid_t1_o;   // Instruction data is valid, with faked excpn
  output        etm_istall_t1_o;    // etm_istall_i in T1 stage
  output        etm_ivalid_t1_o;    // ETMIVALID in T1 stage


  wire          fake_ivalid_t0;     // Instruction data is valid, with faked excpn
  wire          etm_fold_t1;
  wire  [31:1]  etm_ia_t0;          // Instruction Address latch
  wire  [31:1]  etm_ia_t1;          // Instruction Address
  wire  [31:1]  stalled_address_t1; // Output of T1 address logic
  wire  [31:1]  stalled_address_t1_reg; // Instruction Address if etm_istall_i set
  wire  [31:1]  address_recirculate_for_turn_on; // Instruction Address if etm_istall_i and int pending
  wire          cc_fail_t0;
  wire          inst_cc_fail_t1;
  wire          cc_fail_t1;         // Instruction failed condition code
  wire          inst_cc_fail_t2;    // Instruction failed condition code from prev. opcode
  wire          cc_fail_after_stall_t1; // CC information if spec. opcode is discarded

  wire          branch_selected_t1; // Instruction is Branch, which could be traced
  wire          etm_int_pend_t0;
  wire          etm_int_entry_t0;
  wire          etm_int_exit_t0;
  reg           etm_int_pend_t1;
  wire          entering_debug_int_entry_t0;
  reg           entering_debug_int_entry_t1; // partial result for exception_entering_debug_t1
  reg           etm_int_exit_t1;

  wire  [8:0]   etm_int_num_t0;     // Input for ISR redister
  wire  [8:0]   etm_int_num_t1;     // Exception is ISRn
  wire          core_halt_t1;       // Core has entered debug state
  wire          core_halt_t2;

  wire          entering_debug_t1;
  wire          leaving_debug_t1;
  wire          exception_leaving_debug_t1;
  wire          exception_entering_debug_t1;
  wire          reset_done_t1;      //Core has been reset
  wire          exception_after_reset_t0; //Core is taking exception out of reset
  wire          exception_after_reset_t1; //Used to cancel the vaked Ivalid
  wire          reset_taken;        //Reset exception seen
  wire          nxt_reset_taken;

  wire          stalled_ivalid_t0;  // IVALID detected, not yet validated
  wire          stalled_ivalid_t1;
  wire          inst_valid_confirmed_t1; // IVALID is genuine
  wire          inst_valid_confirmed_t2;
  wire          isync_valid_t1;
  wire          turn_on_allowed_t1;
  wire          turn_on_allowed_t2;
  reg  [1:0]    trace_state_t1;
  wire [1:0]    trace_state_t2;
  wire          trace_active_t1;
  wire          trace_active_t2;
  wire          trace_active_and_started_t2; //Inst has been captured after trace on
  wire          trace_active_and_started_t3;
  wire          state_overflow_t2;
  wire          nxt_reason_overflow;
  wire          isync_pending_t1;
  wire          isync_pending_t2;
  wire          isync_periodic_t1;
  wire          isync_non_periodic_t1;
  wire          debug_reason_code;
  wire          debug_fake_inst;
  wire          debug_reason_code_reg;
  wire          debug_fake_inst_reg;
  wire          isync_generated_t1;
  wire          isync_valid_t2;
  wire          rfe_last_t1;
  wire          rfe_last_t2;
  wire          mid_instr_turn_on_t1;
  wire [1:0]    reason_code_t1;
  wire [47:0]   isync_packet_t2;
  wire          exception_cancel_t0;
  wire          exception_cancel_t1;
  wire          nxt_exception_cancel_t2; // Combined with Debug exit exception
  wire          exception_cancel_t2;
  reg  [8:0]    exception_code_t1;
  wire [8:0]    exception_code_t2;
  wire [15:0]   exception_packet_t2;
  wire          exception_valid_t1;
  wire          exception_valid_t2;
  wire [1:0]    exception_byte_valid_t2;
  wire          branch_selected_t0;
  wire          branch_traced_t1;
  wire          branch_traced_t2;
  wire [4:0]    address_byte_valid_t2;
  wire [35:1]   address_packet_t1;
  wire [35:1]   address_packet_t2;
  wire          address_traced_t1;
  wire [39:0]   compressed_address_t2;
  wire          trace_inst_t1;
  wire          trace_e_atom_t1;
  wire          trace_n_atom_t1;
  wire          traced_fold_t1;
  wire [3:0]    p_header_e_count_t1;
  wire [3:0]    p_header_e_count_t2;
  wire          p_header_n_count_t1;
  wire          p_header_n_count_t2;
  wire          format_2_p_header_t1;
  wire          format_2_nn_t1;             // Detect simple case of 2 cc fail back to back
  wire          nxt_format_2_nn_t3;         // Mask with trace active
  reg           opt_format_2_nn_t2;         // Detect simple case of 2 cc fail back to back
  reg           opt_format_2_nn_t3;         // Detect simple case of 2 cc fail back to back
  wire          format_2_nn_t2;             // Detect simple case of 2 cc fail back to back
  wire          format_2_nn_t3;             // Detect simple case of 2 cc fail back to back
  wire          format_2_hint_t1;
  wire          format_2_p_header_t2;
  wire          p_header_empty_t2;
  wire          p_header_valid_t2;
  wire [7:0]    p_header_t2;
  wire [2:0]    compressed_address_size_t2; // Address plus exception number of bytes
  wire          rfe_valid_t1;
  wire          rfe_valid_t2;
  wire          emit_rfe_t2;

  wire          fifo_overflow_t1;
  wire          pending_interrupt_t0; // IntEntry eithout ivalid seen
  wire          exception_pend_or_valid_t0; // pending_interrupt or etm_int_entry
  wire          exception_pend_or_valid_t1; // valid exception condition
  wire          nxt_pending_interrupt_t0; // IntEntry eithout ivalid seen
  wire          monitor_step_t0;   // Fake and cancel an Ivalid due to monstep
  wire          monitor_step_t1;

  wire          isync_overflow;    // Periodic Isync occurs whilst last is pended
  wire          fifo_overflow_t0;  // Fifo or Isync overflow.
  wire [53:0]   ts_packet_t1;      // timestamp packet generation
  reg  [53:0]   ts_packet_t2;      // timestamp packet register
  wire          ts_packet_we;      // timestamp packet write enable
  wire          ts_packet_valid_we;// timestamp packet write enable for C bits
  wire          ts_clk_changed_t1; // clock change bit
  reg           ts_clk_changed_t2; // clock change bit
  wire          ts_full_pend_t1;   // Full timestamp should be output
  reg           ts_full_pend_t2;   // Full timestamp should be output
  wire          ts_pend_t1;        // Timestamp should be captured
  reg           ts_pend_t2;        // Timestamp should be captured
  wire          ts_valid_t1;       // Instruction traced, timestamp is valid
  reg           ts_valid_t2;       // Instruction traced, timestamp is valid
  wire          ts_accepted_t2;    // No other data, TS presented to fifo
  wire [3:0]    ts_packet_size_t2; // num_bytes for timestamp
  reg           ts_holdoff_t2;     // Prevent ts insertion when busy
  wire          nxt_ts_holdoff_t2; // Prevent ts insertion when busy


  reg [63:0]    fifo_in1;          // Muxed data, before timestamp
  reg [3:0]     num_bytes1;        // Muxed data, before timestamp
  wire [63:0]   fifo_in;           // Muxed data, wit timestamp
  wire [3:0]    num_bytes;         // Muxed data, with timestamp
  wire          reason_overflow;
  reg           etm_trigger_out;
  wire          nxt_etm_trigger_out;
  wire          etm_trigger_out_we;
  wire          opt_etm_en;        // Logic optimisation term low if etm not present
  wire          opt_reset_n;       // Optional reset all flops

  reg           opt_etm_fold_t1;
  reg   [31:1]  opt_etm_ia_t1;                // instruction Address
  reg   [31:1]  opt_stalled_address_t1_reg;   // instruction Address if etm_istall_i set
  reg           opt_cc_fail_t1;               // Instruction failed condition code
  reg           opt_inst_cc_fail_t2;          // Instruction failed condition code from prev. opcode
  reg           opt_branch_selected_t1;       // Instruction is Branch, of type determined by branch broadcast
  reg   [8:0]   opt_etm_int_num_t1;           // Exception is ISRn
  reg           opt_core_halt_t1;             // Core has entered debug state
  reg           opt_core_halt_t2;
  reg           opt_exception_after_reset_t1; //Used to cancel the vaked Ivalid
  reg           opt_stalled_ivalid_t1;
  reg           opt_etm_ivalid_t1;
  reg           opt_etm_istall_t1;
  reg           fake_ivalid_t1;
  reg           opt_inst_valid_confirmed_t2;
  reg           opt_turn_on_allowed_t2;
  reg [1:0]     opt_trace_state_t2;
  reg           opt_trace_active_and_started_t3;
  reg           opt_reason_overflow;
  reg           opt_isync_pending_t2;
  reg           opt_debug_reason_code_reg;
  reg           opt_debug_fake_inst_reg;
  reg           opt_isync_valid_t2;
  reg           opt_rfe_last_t2;
  reg           opt_exception_cancel_t1;
  reg           opt_exception_cancel_t2;
  reg [8:0]     opt_exception_code_t2;
  reg           opt_exception_valid_t2;
  reg           opt_branch_traced_t2;
  reg [35:1]    opt_address_packet_t2;
  reg [3:0]     opt_p_header_e_count_t2;
  reg           opt_p_header_n_count_t2;
  reg           opt_format_2_p_header_t2;
  reg           opt_rfe_valid_t2;
  reg           opt_fifo_overflow_t1;
  reg           opt_pending_interrupt_t0;       // opt_ntEntry eithout ivalid seen
  reg           opt_exception_pend_or_valid_t1; // IntEntry eithout ivalid seen T1
  reg           opt_monitor_step_t1;
  reg           opt_reset_taken;

  assign opt_etm_en       = (TRACE_LVL > 1) ? 1'b1            : 1'b0;

`ifdef ARM_DSM
  wire   dsm_reset_all_regs;

  assign opt_reset_n = dsm_reset_all_regs    ? reset_n : 1'b1;

`else

  assign opt_reset_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

`endif


  assign nxt_etm_trigger_out  = integration_mode_i ? apb_wdata_0_i : trigger_req_t2_i;
  assign etm_trigger_out_we   = integration_mode_i ? integration_trigger_we_i :
                                                     etm_trigger_out ^ nxt_etm_trigger_out;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      etm_trigger_out                <= 1'b0;
    else if (etm_trigger_out_we)
      etm_trigger_out                <= nxt_etm_trigger_out;


  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        opt_stalled_ivalid_t1        <= 1'b0;
        opt_etm_ivalid_t1            <= 1'b0;
        fake_ivalid_t1               <= 1'b0;
        opt_monitor_step_t1          <= 1'b0;
        opt_etm_fold_t1              <= 1'b0;
        opt_etm_ia_t1                <= {31{1'b0}};
        opt_stalled_address_t1_reg   <= {31{1'b0}};
        opt_cc_fail_t1               <= 1'b0;
        opt_inst_cc_fail_t2          <= 1'b0;
        opt_branch_selected_t1       <= 1'b0;
        etm_int_pend_t1              <= 1'b0;
        entering_debug_int_entry_t1  <= 1'b0;
        etm_int_exit_t1              <= 1'b0;
        opt_etm_istall_t1            <= 1'b0;
        opt_exception_cancel_t1      <= 1'b0;
        opt_etm_int_num_t1           <= {9{1'b0}};
        opt_core_halt_t1             <= 1'b0;
        opt_core_halt_t2             <= 1'b0;
        opt_turn_on_allowed_t2       <= 1'b0;
        opt_trace_state_t2           <= 2'b00;
        opt_isync_pending_t2         <= 1'b0;
        opt_debug_reason_code_reg    <= 1'b0;
        opt_debug_fake_inst_reg      <= 1'b0;
        opt_address_packet_t2        <= {35{1'b0}};
        opt_branch_traced_t2         <= 1'b0;
        opt_rfe_valid_t2             <= 1'b0;
        opt_rfe_last_t2              <= 1'b0;
        opt_inst_valid_confirmed_t2  <= 1'b0;
        opt_isync_valid_t2           <= 1'b0;
        opt_exception_cancel_t2      <= 1'b0;
        opt_exception_valid_t2       <= 1'b0;
        opt_p_header_e_count_t2      <= 4'h0;
        opt_p_header_n_count_t2      <= 1'b0;
        opt_format_2_p_header_t2     <= 1'b0;
        opt_format_2_nn_t2           <= 1'b0;
        opt_format_2_nn_t3           <= 1'b0;
        opt_exception_code_t2        <= {9{1'b0}};
        opt_trace_active_and_started_t3<= 1'b0;
        opt_fifo_overflow_t1         <= 1'b0;
        opt_pending_interrupt_t0     <= 1'b0;
        opt_exception_pend_or_valid_t1 <= 1'b0;
        opt_reason_overflow          <= 1'b0;
        opt_exception_after_reset_t1 <= 1'b0;
      end
    else if (opt_etm_en)
      begin
        opt_stalled_ivalid_t1        <= stalled_ivalid_t0;
        opt_etm_ivalid_t1            <= etm_ivalid_i;
        fake_ivalid_t1               <= fake_ivalid_t0;
        opt_monitor_step_t1          <= monitor_step_t0;
        opt_etm_fold_t1              <= etm_fold_i;
        opt_etm_ia_t1                <= etm_ia_t0;
        opt_stalled_address_t1_reg   <= stalled_address_t1;
        opt_cc_fail_t1               <= cc_fail_t0;
        opt_inst_cc_fail_t2          <= inst_cc_fail_t1;
        opt_branch_selected_t1       <= branch_selected_t0;
        etm_int_pend_t1              <= etm_int_pend_t0;
        entering_debug_int_entry_t1  <= entering_debug_int_entry_t0;
        etm_int_exit_t1              <= etm_int_exit_t0;
        opt_etm_istall_t1            <= etm_istall_i;
        opt_exception_cancel_t1      <= exception_cancel_t0;
        opt_etm_int_num_t1           <= etm_int_num_t0;
        opt_core_halt_t1             <= core_halt_i;
        opt_core_halt_t2             <= core_halt_t1;
        opt_turn_on_allowed_t2       <= turn_on_allowed_t1;
        opt_trace_state_t2           <= trace_state_t1;
        opt_isync_pending_t2         <= isync_pending_t1;
        opt_debug_reason_code_reg    <= debug_reason_code;
        opt_debug_fake_inst_reg      <= debug_fake_inst;
        opt_address_packet_t2        <= address_packet_t1;
        opt_branch_traced_t2         <= branch_traced_t1;
        opt_rfe_valid_t2             <= rfe_valid_t1;
        opt_rfe_last_t2              <= rfe_last_t1;
        opt_inst_valid_confirmed_t2  <= inst_valid_confirmed_t1;
        opt_isync_valid_t2           <= isync_valid_t1;
        opt_exception_cancel_t2      <= nxt_exception_cancel_t2;
        opt_exception_valid_t2       <= exception_valid_t1;
        opt_p_header_e_count_t2      <= p_header_e_count_t1;
        opt_p_header_n_count_t2      <= p_header_n_count_t1;
        opt_format_2_p_header_t2     <= format_2_p_header_t1;
        opt_format_2_nn_t2           <= format_2_nn_t1;
        opt_format_2_nn_t3           <= nxt_format_2_nn_t3;
        opt_exception_code_t2        <= exception_code_t1;
        opt_trace_active_and_started_t3<= trace_active_and_started_t2;
        opt_fifo_overflow_t1         <= fifo_overflow_t0;
        opt_pending_interrupt_t0     <= nxt_pending_interrupt_t0;
        opt_exception_pend_or_valid_t1 <= exception_pend_or_valid_t0;
        opt_reason_overflow          <= nxt_reason_overflow;
        opt_exception_after_reset_t1 <= exception_after_reset_t0;
      end

  assign stalled_ivalid_t1      = (TRACE_LVL > 1) ? opt_stalled_ivalid_t1      : 1'b0;
  assign monitor_step_t1        = (TRACE_LVL > 1) ? opt_monitor_step_t1        : 1'b0;
  assign etm_fold_t1            = (TRACE_LVL > 1) ? opt_etm_fold_t1            : 1'b0;
  assign etm_ia_t1              = (TRACE_LVL > 1) ? opt_etm_ia_t1              : {31{1'b0}};
  assign stalled_address_t1_reg = (TRACE_LVL > 1) ? opt_stalled_address_t1_reg : {31{1'b0}};
  assign cc_fail_t1             = (TRACE_LVL > 1) ? opt_cc_fail_t1             : 1'b0;
  assign inst_cc_fail_t2        = (TRACE_LVL > 1) ? opt_inst_cc_fail_t2        : 1'b0;
  assign branch_selected_t1     = (TRACE_LVL > 1) ? opt_branch_selected_t1     : 1'b0;
  assign exception_cancel_t1    = (TRACE_LVL > 1) ? opt_exception_cancel_t1    : 1'b0;
  assign etm_int_num_t1         = (TRACE_LVL > 1) ? opt_etm_int_num_t1         : {9{1'b0}};
  assign core_halt_t1           = (TRACE_LVL > 1) ? opt_core_halt_t1           : 1'b0;
  assign core_halt_t2           = (TRACE_LVL > 1) ? opt_core_halt_t2           : 1'b0;
  assign turn_on_allowed_t2     = (TRACE_LVL > 1) ? opt_turn_on_allowed_t2     : 1'b0;
  assign trace_state_t2         = (TRACE_LVL > 1) ? opt_trace_state_t2         : 2'b00;
  assign isync_pending_t2       = (TRACE_LVL > 1) ? opt_isync_pending_t2       : 1'b0;
  assign debug_reason_code_reg  = (TRACE_LVL > 1) ? opt_debug_reason_code_reg  : 1'b0;
  assign debug_fake_inst_reg    = (TRACE_LVL > 1) ? opt_debug_fake_inst_reg    : 1'b0;
  assign address_packet_t2      = (TRACE_LVL > 1) ? opt_address_packet_t2      : {35{1'b0}};
  assign branch_traced_t2       = (TRACE_LVL > 1) ? opt_branch_traced_t2       : 1'b0;
  assign rfe_valid_t2           = (TRACE_LVL > 1) ? opt_rfe_valid_t2           : 1'b0;
  assign rfe_last_t2            = (TRACE_LVL > 1) ? opt_rfe_last_t2            : 1'b0;
  assign inst_valid_confirmed_t2= (TRACE_LVL > 1) ? opt_inst_valid_confirmed_t2: 1'b0;
  assign isync_valid_t2         = (TRACE_LVL > 1) ? opt_isync_valid_t2         : 1'b0;
  assign exception_cancel_t2    = (TRACE_LVL > 1) ? opt_exception_cancel_t2    : 1'b0;
  assign exception_valid_t2     = (TRACE_LVL > 1) ? opt_exception_valid_t2     : 1'b0;
  assign p_header_e_count_t2    = (TRACE_LVL > 1) ? opt_p_header_e_count_t2    : 4'h0;
  assign p_header_n_count_t2    = (TRACE_LVL > 1) ? opt_p_header_n_count_t2    : 1'b0;
  assign format_2_p_header_t2   = (TRACE_LVL > 1) ? opt_format_2_p_header_t2   : 1'b0;
  assign format_2_nn_t2         = (TRACE_LVL > 1) ? opt_format_2_nn_t2         : 1'b0;
  assign format_2_nn_t3         = (TRACE_LVL > 1) ? opt_format_2_nn_t3         : 1'b0;
  assign exception_code_t2      = (TRACE_LVL > 1) ? opt_exception_code_t2      : {9{1'b0}};
  assign trace_active_and_started_t3 = (TRACE_LVL > 1) ? opt_trace_active_and_started_t3 : 1'b0;
  assign fifo_overflow_t1       = (TRACE_LVL > 1) ? opt_fifo_overflow_t1       : 1'b0;
  assign pending_interrupt_t0   = (TRACE_LVL > 1) ? opt_pending_interrupt_t0   : 1'b0;
  assign exception_pend_or_valid_t1 = (TRACE_LVL > 1) ? opt_exception_pend_or_valid_t1   : 1'b0;
  assign reason_overflow        = (TRACE_LVL > 1) ? opt_reason_overflow        : 1'b0;
  assign exception_after_reset_t1 = (TRACE_LVL > 1) ? opt_exception_after_reset_t1 : 1'b0;

  assign isync_valid_t1 = isync_non_periodic_t1 | (isync_periodic_t1 & trace_active_t1);

  assign stalled_ivalid_t0 = etm_ivalid_i |
                             (stalled_ivalid_t1 &  etm_istall_i & ~etm_cancel_i);

  assign inst_valid_confirmed_t1 = fake_ivalid_t1 |
                                   (stalled_ivalid_t1 & (~etm_istall_i | etm_fold_t1));

  assign etm_int_pend_t0  = etm_int_stat_i == `ETM_INT_PENDING;
  assign etm_int_entry_t0 = etm_int_stat_i == `ETM_INT_ENTRY;
  assign etm_int_exit_t0  = etm_int_stat_i == `ETM_INT_EXIT;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_reset_taken                <= 1'b0;
    else if (opt_etm_en)
      opt_reset_taken                <= nxt_reset_taken;

  assign reset_taken = (TRACE_LVL > 1) ? opt_reset_taken : 1'b0;

  assign nxt_reset_taken = (((etm_int_entry_t0) &
                         (etm_int_num_i == 9'b000000000)) |
                         (reset_taken & ~inst_valid_confirmed_t1 &
                          ~monitor_step_t0 & ~core_halt_i));

  assign reset_done_t1    = reset_taken & inst_valid_confirmed_t1;

  assign exception_after_reset_t0 = (reset_taken & ~exception_cancel_t2 &
                                     (etm_int_pend_t0));

  assign fifo_overflow_t0 = (fifo_overflow_i | isync_overflow);

  assign entering_debug_t1 = core_halt_t1 & ~core_halt_t2;
  assign leaving_debug_t1  = core_halt_t2 & ~core_halt_t1;
  assign entering_debug_int_entry_t0 = core_halt_i & ~core_halt_t1 & etm_int_entry_t0;


  assign exception_pend_or_valid_t0 = pending_interrupt_t0 | etm_int_entry_t0;

  assign exception_leaving_debug_t1 = (debug_fake_inst & ~exception_cancel_t2 &
                               ((etm_int_pend_t1)));

  assign exception_entering_debug_t1 = (entering_debug_int_entry_t1 |
                                (pending_interrupt_t0 & core_halt_i));

  assign nxt_pending_interrupt_t0 = (~etm_ivalid_i & ~core_halt_i &
                          ~(etm_int_pend_t0) &
                          ((etm_int_entry_t0) | pending_interrupt_t0));

  assign monitor_step_t0 = (~etm_ivalid_i & ~core_halt_i & pending_interrupt_t0 &
                          (etm_int_pend_t0));

  assign exception_cancel_t0 = etm_cancel_i | monitor_step_t1 | exception_after_reset_t1;

  assign fake_ivalid_t0 = (exception_leaving_debug_t1 |
                           monitor_step_t0 | exception_after_reset_t0);

  assign turn_on_allowed_t1 = (inst_valid_confirmed_t1 |
                            (turn_on_allowed_t2 & ~core_halt_t1 & niden_t2_i));

  assign etm_int_num_t0 = etm_int_entry_t0 ? etm_int_num_i : etm_int_num_t1;

  always @(trace_state_t2 or fifo_overflow_t1 or trace_enable_i or
           inst_valid_confirmed_t1 or power_down_reset_i or fifo_empty_i or
           entering_debug_t1 or leaving_debug_t1 or turn_on_allowed_t1 or
           flush_trace_i)
    begin
      case (trace_state_t2)
        `ST_G_Idle:
          begin
            if (fifo_overflow_t1)
              trace_state_t1 = `ST_G_Overflow;
            else if (trace_enable_i & turn_on_allowed_t1)
              trace_state_t1 = `ST_G_Tracing;
            else
              trace_state_t1 = `ST_G_Idle;
          end

        `ST_G_Tracing:
          begin
            if (power_down_reset_i)
              trace_state_t1 = `ST_G_Idle;
            else if (fifo_overflow_t1)
              trace_state_t1 = `ST_G_Overflow;
            else if ((~trace_enable_i & (inst_valid_confirmed_t1)) |
                     entering_debug_t1 | flush_trace_i)
              trace_state_t1 = `ST_G_Idle;
            else
              trace_state_t1 = `ST_G_Tracing;
          end

        `ST_G_Overflow:
          begin
            if ((fifo_empty_i & leaving_debug_t1) | power_down_reset_i)
              trace_state_t1 = `ST_G_Idle;
            else if (fifo_empty_i & inst_valid_confirmed_t1 & trace_enable_i) //Starts cleanly.
              trace_state_t1 = `ST_G_Tracing;
            else
              trace_state_t1 = `ST_G_Overflow;
          end // if (State == `ST_G_Overflow)

        `ST_G_Unused: trace_state_t1 = `ST_G_Idle;
        default:      trace_state_t1 = `ST_G_X;
      endcase // case(State)
    end

  assign trace_active_t1  = trace_state_t1[0];
  assign trace_active_t2  = trace_state_t2[0];

  assign trace_active_and_started_t2 = trace_active_t2 & (trace_active_and_started_t3 |
                                              rfe_valid_t2 |
                                              inst_valid_confirmed_t1 |
                                              fake_ivalid_t0 |
                                              exception_entering_debug_t1);

  assign state_overflow_t2    = trace_state_t2[1];

  assign nxt_reason_overflow = (state_overflow_t2 |
                            (reason_overflow & ~trace_active_t2 & ~power_down_reset_i));

  assign isync_pending_t1 = (isync_req_i |
                             (isync_pending_t2 & ~isync_valid_t2 & trace_active_t2));

  assign isync_overflow = (isync_req_i & isync_pending_t2 &
                           ~isync_valid_t2 & trace_active_t2);

  assign isync_periodic_t1 = ((isync_pending_t2 & ~isync_valid_t2 & trace_active_t2) &
                              ((inst_valid_confirmed_t1 & ~branch_traced_t2) |
                               (branch_traced_t2 & ~stalled_ivalid_t1 & ~p_header_e_count_t2[1])) &
                              ~branch_selected_t1 &
                              ~opt_pending_interrupt_t0 &
                              ~monitor_step_t1 &
                              ~reset_taken &
                              ~exception_pend_or_valid_t1 &
                              ~p_header_n_count_t2 &
                              ~etm_int_exit_t1);

  assign isync_non_periodic_t1 = trace_active_t1 & ~trace_active_t2;

  assign isync_generated_t1 = isync_periodic_t1 | isync_non_periodic_t1;

  assign rfe_last_t1 = rfe_valid_t1 | (rfe_last_t2 & ~inst_valid_confirmed_t1 &
                                       ~entering_debug_t1);

  assign mid_instr_turn_on_t1 = isync_non_periodic_t1 & ~inst_valid_confirmed_t1 &
                                ~rfe_last_t1;

  assign debug_reason_code = (leaving_debug_t1 |
                              (debug_reason_code_reg &
                               ~(((inst_valid_confirmed_t2 | monitor_step_t0) &
                                  niden_t2_i) |
                                 core_halt_t2)));
  assign debug_fake_inst = (leaving_debug_t1 |
                           (debug_fake_inst_reg &
                            ~(stalled_ivalid_t1 | monitor_step_t0 | core_halt_t2)));

  assign reason_code_t1[1] = isync_non_periodic_t1 & (debug_reason_code |
                                                      state_overflow_t2);
  assign reason_code_t1[0] = isync_non_periodic_t1 & (debug_reason_code |
                                                      ~state_overflow_t2);

  assign isync_packet_t2 = {address_packet_t2[31:1], 1'b1,
                          1'b0, address_packet_t2[33:32], 5'b00001,
                          `ISync_HEADER};

  assign nxt_exception_cancel_t2 = (exception_cancel_t1 | exception_leaving_debug_t1 |
                                    (exception_cancel_t2 & ~exception_valid_t2));

  always @(etm_int_num_t1)
    begin
      if (|etm_int_num_t1[8:5])
        exception_code_t1 = etm_int_num_t1;
      else
        case (etm_int_num_t1[4:0])
          5'd00 : exception_code_t1 = 9'd17;
          5'd01 : exception_code_t1 = 9'd00; // not used by core
          5'd02 : exception_code_t1 = 9'd10;
          5'd03 : exception_code_t1 = 9'd19;
          5'd04 : exception_code_t1 = 9'd13;
          5'd05 : exception_code_t1 = 9'd21;
          5'd06 : exception_code_t1 = 9'd09;
          5'd07 : exception_code_t1 = 9'd22;
          5'd08 : exception_code_t1 = 9'd16;
          5'd09 : exception_code_t1 = 9'd20;
          5'd10 : exception_code_t1 = 9'd18;
          5'd11 : exception_code_t1 = 9'd11;
          5'd12 : exception_code_t1 = 9'd12;
          5'd13 : exception_code_t1 = 9'd23;
          5'd14 : exception_code_t1 = 9'd14;
          5'd15 : exception_code_t1 = 9'd15;
          5'd16 : exception_code_t1 = 9'd08;
          5'd17 : exception_code_t1 = 9'd01;
          5'd18 : exception_code_t1 = 9'd02;
          5'd19 : exception_code_t1 = 9'd03;
          5'd20 : exception_code_t1 = 9'd04;
          5'd21 : exception_code_t1 = 9'd05;
          5'd22 : exception_code_t1 = 9'd06;
          5'd23 : exception_code_t1 = 9'd07;
          5'd24 : exception_code_t1 = 9'd24;
          5'd25 : exception_code_t1 = 9'd25;
          5'd26 : exception_code_t1 = 9'd26;
          5'd27 : exception_code_t1 = 9'd27;
          5'd28 : exception_code_t1 = 9'd28;
          5'd29 : exception_code_t1 = 9'd29;
          5'd30 : exception_code_t1 = 9'd30;
          5'd31 : exception_code_t1 = 9'd31;
          default : exception_code_t1 = {9{1'bx}};
        endcase
    end

  assign exception_packet_t2[0] = 1'b0;
  assign exception_packet_t2[4:1] = exception_code_t2[3:0];
  assign exception_packet_t2[5] = exception_cancel_t2;
  assign exception_packet_t2[6] = 1'b0;
  assign exception_packet_t2[7] = |exception_code_t2[8:4]; // Continue bit
  assign exception_packet_t2[12:8] = exception_code_t2[8:4];
  assign exception_packet_t2[15:13] = 3'b000;

  assign exception_valid_t1 = (exception_entering_debug_t1 |
                               (monitor_step_t1 &
                                (~p_header_empty_t2 | rfe_valid_t2 |
                                 trace_active_and_started_t3 )) |
                               reset_done_t1 |
                               (exception_pend_or_valid_t1 &
                                inst_valid_confirmed_t1));

  assign exception_byte_valid_t2[0] = exception_valid_t2 & trace_active_and_started_t3;

  assign exception_byte_valid_t2[1] = exception_packet_t2[7] & exception_byte_valid_t2[0];

  assign branch_selected_t0 = prg_branch_broadcast_i ? etm_branch_i :
                                                       etm_indirect_branch_i;
  assign branch_traced_t1 = (branch_selected_t1 |
                             exception_valid_t1) & trace_active_t2;

  assign address_traced_t1 = branch_traced_t1 | isync_generated_t1;

  assign etm_ia_t0 = (etm_ivalid_i | fake_ivalid_t0 |
                      monitor_step_t0) ? etm_ia_i :
                                         address_recirculate_for_turn_on;

  assign cc_fail_t0 = (etm_ivalid_i | fake_ivalid_t0) ? etm_cc_fail_i :
                                                    (cc_fail_after_stall_t1 &
                                                     ~etm_fold_t1);
  assign inst_cc_fail_t1 = (etm_istall_i) ? inst_cc_fail_t2 :
                                         (cc_fail_t1 & ~etm_fold_t1);
  assign cc_fail_after_stall_t1 = (etm_istall_i & etm_cancel_i) ? inst_cc_fail_t2 :
                                                                    cc_fail_t1;

  assign stalled_address_t1[31:1] = (etm_istall_i & ~etm_fold_t1) ?
                                      stalled_address_t1_reg : etm_ia_t1[31:1];

  assign address_recirculate_for_turn_on [31:1] = (etm_istall_i & etm_cancel_i) ?
                                                    stalled_address_t1_reg[31:1] :
         (etm_ia_t1[31:1] + {{30{1'b0}},etm_fold_t1});




  reg [31:7] reference_address;    // Previous address output
  wire       reference_address_we; // register enable
  wire [35:1] compressed_address;  // Internal version of output

  assign          reference_address_we = address_traced_t1 & opt_etm_en;

  always @ (posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      reference_address              <= {25{1'b0}};
    else if (reference_address_we)
      reference_address              <= stalled_address_t1[31:7];

  assign compressed_address[31] = ~(stalled_address_t1[31:27] == reference_address[31:27]);
  assign compressed_address[23] = compressed_address[31] |
                                  (~(stalled_address_t1[26:20] == reference_address[26:20]));
  assign compressed_address[15] = compressed_address[23] |
                                  (~(stalled_address_t1[19:13] == reference_address[19:13]));
  assign compressed_address[ 7] = compressed_address[15] |
                                  ((exception_valid_t1 |
                                    ~(stalled_address_t1[13: 7] == reference_address[13: 7])));

  assign compressed_address[ 6: 1] = stalled_address_t1[ 6: 1];
  assign compressed_address[13: 8] = stalled_address_t1[12: 7];
  assign compressed_address[14]    = compressed_address[15] ? stalled_address_t1[13] :
                                       exception_valid_t1;
  assign compressed_address[21:16] = stalled_address_t1[19:14];
  assign compressed_address[22]    = compressed_address[23] ? stalled_address_t1[20] :
                                       exception_valid_t1;
  assign compressed_address[29:24] = stalled_address_t1[26:21];
  assign compressed_address[30]    = compressed_address[31] ? stalled_address_t1[27] :
                                       exception_valid_t1;
  assign compressed_address[35:32] = stalled_address_t1[31:28];




  assign address_packet_t1 = (isync_generated_t1) ?
                         {2'b00, reason_code_t1, stalled_address_t1[31:1]} :
                         compressed_address[35:1];

  assign compressed_address_t2 = {1'b0, exception_byte_valid_t2[0], 2'b01,
                       address_packet_t2[35:1], 1'b1};

  assign address_byte_valid_t2[0] = branch_traced_t2;
  assign address_byte_valid_t2[1] = address_packet_t2[7]  & branch_traced_t2;
  assign address_byte_valid_t2[2] = address_packet_t2[15] & branch_traced_t2;
  assign address_byte_valid_t2[3] = address_packet_t2[23] & branch_traced_t2;
  assign address_byte_valid_t2[4] = address_packet_t2[31] & branch_traced_t2;

  assign trace_inst_t1 = (inst_valid_confirmed_t1 & trace_active_t1);
  assign trace_e_atom_t1 = (trace_inst_t1 & ~cc_fail_t1) |
                             (mid_instr_turn_on_t1 & ~inst_cc_fail_t2);
  assign trace_n_atom_t1 = (trace_inst_t1 & ~etm_istall_i & cc_fail_t1) |
                             (mid_instr_turn_on_t1 & inst_cc_fail_t2);
  assign traced_fold_t1 = etm_fold_t1 & trace_active_t1;

  assign p_header_e_count_t1 = (p_header_e_count_t2 & {4{~p_header_valid_t2}}) +
                      {3'b000, trace_e_atom_t1} +
                      {3'b000, traced_fold_t1 & ~etm_istall_i};

  assign p_header_n_count_t1 = trace_n_atom_t1;

  assign format_2_p_header_t1 = trace_inst_t1 & format_2_hint_t1;
  assign format_2_hint_t1     = cc_fail_t1 & etm_fold_t1;

  assign format_2_nn_t1 = trace_inst_t1 & cc_fail_t1 & ~etm_fold_t1 &
         ~etm_istall_i & etm_cc_fail_i & (~(|p_header_e_count_t2)|
                                            p_header_n_count_t2) & ~format_2_nn_t2;
  assign nxt_format_2_nn_t3 = (format_2_nn_t2 & trace_active_t2);

  wire   exception_possible_t1;

  assign exception_possible_t1 = pending_interrupt_t0 |
                                 monitor_step_t1 |
                                 reset_taken |
                                 exception_pend_or_valid_t1;

  assign p_header_empty_t2 = ~(|p_header_e_count_t2) & ~format_2_p_header_t2;

  assign p_header_valid_t2 = ~p_header_empty_t2 &
                          ((~isync_pending_t2 & ts_pend_t2 & ~ts_valid_t2) |
                           trigger_req_t2_i |
                           (isync_pending_t2 & ~isync_valid_t2 &
                            (fake_ivalid_t1 | stalled_ivalid_t1) & ~branch_traced_t2) |
                           etm_int_exit_t1 |
                           monitor_step_t1 |
                           branch_selected_t1 |
                           exception_possible_t1 | // Atoms up to branch, not including destination
                           ~trace_active_t2 |
                           (&p_header_e_count_t2[3:0]) | // 15 Es
                           ((&p_header_e_count_t2[3:1]) & etm_fold_t1) | // 14 Es + 2 next
                           format_2_hint_t1) |
                           (p_header_n_count_t2 & ~format_2_nn_t2 & ~format_2_nn_t3) |
                           (format_2_nn_t3);

  assign p_header_t2 = {1'b1,
                        (p_header_n_count_t2 & ~format_2_nn_t2 & ~format_2_nn_t3 &
                         ~format_2_p_header_t2) |
                        (~trace_active_t2 & format_2_nn_t3),
                        p_header_e_count_t2[3:2],
                        p_header_e_count_t2[1] | (format_2_nn_t3 & trace_active_t2) |
                        format_2_p_header_t2,
                        (p_header_e_count_t2[0] & ~format_2_p_header_t2) |
                        (format_2_nn_t3 & trace_active_t2),
                        (format_2_nn_t3 & trace_active_t2) | format_2_p_header_t2,
                        1'b0};

  assign compressed_address_size_t2[2] = address_byte_valid_t2[3] |
         (address_byte_valid_t2[2] & exception_byte_valid_t2[0]) |
           exception_byte_valid_t2[1];

  assign compressed_address_size_t2[1] = (address_byte_valid_t2[3] &
                                          exception_byte_valid_t2[0] &
                                          (exception_byte_valid_t2[1] |
                                           address_byte_valid_t2[4])) |
                                         (~address_byte_valid_t2[2] &
                                          address_byte_valid_t2[1] &
                                          ~exception_byte_valid_t2[1]) |
                                         (~exception_byte_valid_t2[0] &
                                          ~address_byte_valid_t2[3] &
                                          address_byte_valid_t2[2]);
  assign compressed_address_size_t2[0] = ^{exception_byte_valid_t2,address_byte_valid_t2};

  assign rfe_valid_t1 = (etm_int_exit_t1 |
                         (rfe_valid_t2 & ~((opt_etm_ivalid_t1 | fake_ivalid_t0 ) |
                                           trace_active_t2 |
                                           state_overflow_t2 | core_halt_t1)));

  assign emit_rfe_t2 = rfe_valid_t2 &  trace_active_t2;

  assign ts_packet_we = (prg_ts_enable_i & trace_active_t2 &
                        inst_valid_confirmed_t2 & ~debug_fake_inst_reg);
  assign ts_packet_valid_we = ts_packet_we | ts_accepted_t2;

  assign ts_packet_t1[53:48] =  ts_value_i[47:42];
  assign ts_packet_t1[47]    =  |(ts_value_i[47:42] ^ ts_packet_t2[53:48])|
                                (ts_packet_t2[47] & ~ts_accepted_t2)  |
                                 ts_full_pend_t1;

  assign ts_packet_t1[46:40] =  ts_value_i[41:35];
  assign ts_packet_t1[39]    = |(ts_value_i[41:35] ^ ts_packet_t2[46:40]) |
                                (ts_packet_t2[39] & ~ts_accepted_t2) |
                                 ts_full_pend_t1 |
                                 ts_packet_t1[47];
  assign ts_packet_t1[38:32] =  ts_value_i[34:28];
  assign ts_packet_t1[31]    = |(ts_value_i[34:28] ^ ts_packet_t2[38:32]) |
                                (ts_packet_t2[31] & ~ts_accepted_t2) |
                                 ts_full_pend_t1 |
                                 ts_packet_t1[39] |
                                 ts_packet_t1[47];
  assign ts_packet_t1[30:24] =  ts_value_i[27:21];
  assign ts_packet_t1[23]    = |(ts_value_i[27:21] ^ ts_packet_t2[30:24]) |
                                (ts_packet_t2[23] & ~ts_accepted_t2) |
                                 ts_full_pend_t1 |
                                 ts_packet_t1[31] |
                                 ts_packet_t1[39] |
                                 ts_packet_t1[47];
  assign ts_packet_t1[22:16] =  ts_value_i[20:14];
  assign ts_packet_t1[15]    = |(ts_value_i[20:14] ^ ts_packet_t2[22:16]) |
                                 (ts_packet_t2[15] & ~ts_accepted_t2) |
                                   ts_full_pend_t1 |
                                   ts_packet_t1[23] |
                                   ts_packet_t1[31] |
                                   ts_packet_t1[39] |
                                   ts_packet_t1[47];
  assign ts_packet_t1[14: 8] =  ts_value_i[13: 7];
  wire   diff1 = |(ts_value_i[13: 7] ^ ts_packet_t2[14: 8]);

  assign ts_packet_t1[ 7]    = diff1 |
                                (ts_packet_t2[7] & ~ts_accepted_t2) |
                                 ts_full_pend_t1 |
                                 ts_packet_t1[15] |
                                 ts_packet_t1[23] |
                                 ts_packet_t1[31] |
                                 ts_packet_t1[39] |
                                 ts_packet_t1[47];
  assign ts_packet_t1[ 6: 0] =  ts_value_i[ 6: 0];

  always @ (posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      begin
        ts_packet_t2[6:0]   <= {7{1'b0}};
        ts_packet_t2[14:8]  <= {7{1'b0}};
        ts_packet_t2[22:16] <= {7{1'b0}};
        ts_packet_t2[30:24] <= {7{1'b0}};
        ts_packet_t2[38:32] <= {7{1'b0}};
        ts_packet_t2[46:40] <= {7{1'b0}};
        ts_packet_t2[53:48] <= {6{1'b0}};
      end
    else if (ts_packet_we)
      begin
        ts_packet_t2[6:0]   <= ts_packet_t1[6:0];
        ts_packet_t2[14:8]  <= ts_packet_t1[14:8];
        ts_packet_t2[22:16] <= ts_packet_t1[22:16];
        ts_packet_t2[30:24] <= ts_packet_t1[30:24];
        ts_packet_t2[38:32] <= ts_packet_t1[38:32];
        ts_packet_t2[46:40] <= ts_packet_t1[46:40];
        ts_packet_t2[53:48] <= ts_packet_t1[53:48];
      end

  always @ (posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      begin
        ts_packet_t2[7]  <= 1'b0;
        ts_packet_t2[15] <= 1'b0;
        ts_packet_t2[23] <= 1'b0;
        ts_packet_t2[31] <= 1'b0;
        ts_packet_t2[39] <= 1'b0;
        ts_packet_t2[47] <= 1'b0;
      end
    else if (ts_packet_valid_we)
      begin
        ts_packet_t2[7]  <= ts_packet_t1[7];
        ts_packet_t2[15] <= ts_packet_t1[15];
        ts_packet_t2[23] <= ts_packet_t1[23];
        ts_packet_t2[31] <= ts_packet_t1[31];
        ts_packet_t2[39] <= ts_packet_t1[39];
        ts_packet_t2[47] <= ts_packet_t1[47];
      end

  assign ts_packet_size_t2[3] = ts_packet_t2[47];
  assign ts_packet_size_t2[2:0] = ((3'b010 +
                                    {2'b00,ts_packet_t2[7]} +
                                    {2'b00,ts_packet_t2[15]} +
                                    {2'b00,ts_packet_t2[23]} +
                                    {2'b00,ts_packet_t2[31]}) |
                                   {3{ts_packet_t2[39]}}) &
                                    {3{~ts_packet_t2[47]}};



  assign nxt_ts_holdoff_t2 = fifo_ts_holdoff_i | isync_valid_t2 |
                     (trace_active_t2 & (branch_traced_t1 |
                                         rfe_valid_t2 | etm_int_exit_t1 |
                                         isync_pending_t1 |
                                         opt_etm_istall_t1 |
                     (opt_etm_ivalid_t1 & (cc_fail_t1 | cc_fail_t0))));

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      ts_holdoff_t2 <= 1'b0;
      else
    ts_holdoff_t2 <= nxt_ts_holdoff_t2;


  assign ts_accepted_t2 = ts_valid_t2 & ts_pend_t2 & ~ts_holdoff_t2 &
                          ~(|p_header_e_count_t2[3:1]) &
         ~isync_pending_t1;
  assign ts_pend_t1 = prg_ts_enable_i & ((~ts_accepted_t2 & (ts_pend_t2 |
                                                             ts_full_pend_t2 |
                                                            (trace_active_t2 &
                                                             (etm_isb_i |
                                                              rfe_valid_t2)) |
                                                            flush_trace_i |
                                                            timestamp_event_i)) |
                                         ts_clk_change_i |
                                         state_overflow_t2 |
                                         isync_req_i);
  assign ts_full_pend_t1 = prg_ts_enable_i & ((ts_full_pend_t2 & ~ts_accepted_t2) |
                                              state_overflow_t2 |
                                              prg_prog_bit_fall_i |
                                              isync_req_i);
  assign ts_valid_t1 = prg_ts_enable_i & ((ts_valid_t2 | ts_packet_we) &
                                          p_header_valid_t2 &
                        ~(ts_accepted_t2 | power_down_reset_i |
                          isync_non_periodic_t1 |
                          state_overflow_t2));
  assign ts_clk_changed_t1 = (ts_clk_changed_t2 & ~(ts_accepted_t2 |
                                                    power_down_reset_i |
                                        state_overflow_t2)) |
                         ts_clk_change_i;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        ts_pend_t2        <= 1'b0;
        ts_full_pend_t2   <= 1'b0;
        ts_valid_t2       <= 1'b0;
        ts_clk_changed_t2 <= 1'b0;
      end
    else
      begin
        ts_pend_t2        <= ts_pend_t1;
        ts_full_pend_t2   <= ts_full_pend_t1;
        ts_valid_t2       <= ts_valid_t1;
        ts_clk_changed_t2 <= ts_clk_changed_t1;
      end


  always @(exception_byte_valid_t2 or address_byte_valid_t2 or
           isync_valid_t2 or emit_rfe_t2 or p_header_t2 or isync_packet_t2 or
           rfe_valid_t2 or p_header_valid_t2 or compressed_address_t2 or
           exception_packet_t2 or compressed_address_size_t2 or
           fifo_overflow_t1 or state_overflow_t2 or ts_accepted_t2)
    begin
      if (fifo_overflow_t1 | state_overflow_t2 | ts_accepted_t2)
        begin
          fifo_in1[63:0] = {64{1'b0}};
          num_bytes1 = 4'b0000;
        end
      else if (isync_valid_t2)
        begin
          fifo_in1[63:0] = {p_header_t2,(rfe_valid_t2 ? `RFE_HEADER : p_header_t2), // last p-header not used, but for simplicity
                          isync_packet_t2[47:0]};
          num_bytes1 = (p_header_valid_t2 | rfe_valid_t2) ? 4'b0111 : 4'b0110;
        end
      else if (emit_rfe_t2)
        begin
          fifo_in1[63:0] = {{7{p_header_t2}}, `RFE_HEADER};
          num_bytes1 = p_header_valid_t2 ? 4'b0010 : 4'b0001;
        end
      else
        begin
          case ({exception_byte_valid_t2, address_byte_valid_t2})
            7'b00_00000 : fifo_in1[63:0] = {8{p_header_t2}};
            7'b00_00001 : fifo_in1[63:0] = {{7{p_header_t2}}, compressed_address_t2[7:0]};
            7'b00_00011 : fifo_in1[63:0] = {{6{p_header_t2}}, compressed_address_t2[15:0]};
            7'b00_00111 : fifo_in1[63:0] = {{5{p_header_t2}}, compressed_address_t2[23:0]};
            7'b00_01111 : fifo_in1[63:0] = {{4{p_header_t2}}, compressed_address_t2[31:0]};
            7'b00_11111 : fifo_in1[63:0] = {{3{p_header_t2}}, compressed_address_t2[39:0]};
            7'b01_00011 : fifo_in1[63:0] = {{5{p_header_t2}},
                                           exception_packet_t2[7:0], compressed_address_t2[15:0]};
            7'b01_00111 : fifo_in1[63:0] = {{4{p_header_t2}},
                                           exception_packet_t2[7:0], compressed_address_t2[23:0]};
            7'b01_01111 : fifo_in1[63:0] = {{3{p_header_t2}},
                                           exception_packet_t2[7:0], compressed_address_t2[31:0]};
            7'b01_11111 : fifo_in1[63:0] = {{2{p_header_t2}},
                                           exception_packet_t2[7:0], compressed_address_t2[39:0]};
            7'b11_00011 : fifo_in1[63:0] = {{4{p_header_t2}},
                                           exception_packet_t2[15:0], compressed_address_t2[15:0]};
            7'b11_00111 : fifo_in1[63:0] = {{3{p_header_t2}},
                                           exception_packet_t2[15:0], compressed_address_t2[23:0]};
            7'b11_01111 : fifo_in1[63:0] = {{2{p_header_t2}},
                                           exception_packet_t2[15:0], compressed_address_t2[31:0]};
            7'b11_11111 : fifo_in1[63:0] = {   p_header_t2  ,
                                           exception_packet_t2[15:0], compressed_address_t2[39:0]};
            default     : fifo_in1[63:0] = {64{1'bX}}; // not valid
          endcase
          num_bytes1 = (p_header_valid_t2 ? 4'b0001 : 4'b0000) +
                      {1'b0, compressed_address_size_t2};
        end
    end

  assign fifo_in = fifo_in1 |
         ({64{ts_accepted_t2}} & {2'b00, ts_packet_t2[53:0],
                                 5'b01000, ts_clk_changed_t2, 2'b10});
  assign num_bytes = num_bytes1 | ({4{ts_accepted_t2}} & ts_packet_size_t2);

  assign fifo_in_o             = (TRACE_LVL > 1) ? fifo_in          : {64{1'b0}};
  assign num_bytes_o           = (TRACE_LVL > 1) ? num_bytes        : 4'b0000;
  assign reason_overflow_o     = (TRACE_LVL > 1) ? reason_overflow  : 1'b0;
  assign etm_trigger_out_o     = (TRACE_LVL > 1) ? etm_trigger_out  : 1'b0;
  assign fake_ivalid_t1_o      = (TRACE_LVL > 1) ? fake_ivalid_t1   : 1'b0;
  assign etm_istall_t1_o       = (TRACE_LVL > 1) ? opt_etm_istall_t1: 1'b0;
  assign etm_ivalid_t1_o       = (TRACE_LVL > 1) ? opt_etm_ivalid_t1: 1'b0;


`ifdef ARM_DSM
`define  ARM_ETM_VALIDATION
`endif

`ifdef ARM_ETM_VALIDATION
  wire [9:0]    fifo_peek;
  wire          overflow_peek_t2;
  wire          isync_peek_t2;
  wire          rfe_peek_t1;       // RFE captured.
  assign        rfe_peek_t1 = rfe_valid_t1 & trace_active_t1;

  assign overflow_peek_t2 = ((trace_state_t1 != `ST_G_Overflow) &
                         (trace_state_t2 == `ST_G_Overflow));
  assign isync_peek_t2 = isync_valid_t2 & trace_active_t2 &
         ~(fifo_overflow_i | fifo_overflow_t1);
  wire dbg_trace_instr_main = trace_inst_t1 & ~fake_ivalid_t1;
  reg  dbg_trace_instr_fake;
  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      dbg_trace_instr_fake <= 1'b0;
      else
    dbg_trace_instr_fake <= (fake_ivalid_t1 & trace_active_t1);

  assign fifo_peek = {overflow_peek_t2, 1'b0, isync_peek_t2, trigger_req_t2_i,
                      4'b0000, (traced_fold_t1 & ~etm_istall_i) | (rfe_peek_t1),
                      trace_inst_t1 |
                      mid_instr_turn_on_t1};
`endif

`ifdef ARM_ASSERT_ON
`include "std_ovl_defines.h"
  reg    ovl_etm_ivalid_t1;
  reg    ovl_etm_ivalid_t2;
  reg    ovl_etm_ivalid_t3;
  reg    ovl_etm_ivalid_t4;
  reg    ovl_etm_ivalid_stall;
  wire   ovl_etm_ivalid_stall_nxt;
  reg    ovl_core_halt_t1;
  reg    ovl_core_halt_t2;
  reg    ovl_core_halt_t3;
  reg    ovl_pending_interrupt_t1;

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      ovl_pending_interrupt_t1 <= 1'b0;
      else
    ovl_pending_interrupt_t1   <= pending_interrupt_t0;

  assert_never #(0,0,"Can only output P header or RFE, not both.")
  ovl_etm_phdr_rfe_isync(
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (p_header_valid_t2 & rfe_valid_t2));


  assert_never #(0,0,"State variable has invalid value")
  ovl_etm_genstatevalid(
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (trace_state_t1 == `ST_G_Unused));

  assert_never_unknown #(0,2,0,"State variable must not be X")
  ovl_etm_genstatex(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (trace_state_t1));

  assert_never #(0,`ETM_INTERFACE_TYPE,"etm_branch_i signalled without ETMIBRANCH")
  ovl_etm_ibranchsideband (
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (etm_indirect_branch_i & ~etm_branch_i)
                       );
  assert_never #(0,`ETM_INTERFACE_TYPE,"etm_branch_i signalled without etm_ivalid_i")
  ovl_etm_branchisinstruction (
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (etm_branch_i & ~etm_ivalid_i)
                 );

  assert_never #(0,`ETM_INTERFACE_TYPE,"etm_cancel_i asserted but not INT ENTRY/PEND")
  ovl_etm_excancel (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_cancel_i & ~((etm_int_entry_t0) |
                                              (etm_int_pend_t0)
                                              ))
                  );

  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntIdle must not be followed by IntEntry")
  ovl_etm_intstat1 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (3'b000),
    .next_state  (`ETM_INT_ENTRY)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntEntry must not be followed by IntExit")
  ovl_etm_intstat2 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_ENTRY),
    .next_state  (`ETM_INT_EXIT)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntEntry must not be followed by IntReturn")
  ovl_etm_intstat3 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_ENTRY),
    .next_state  (`ETM_INT_RETURN)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntEntry must not be followed by IntReturn")
  ovl_etm_intstat3a (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_ENTRY),
    .next_state  (`ETM_INT_ENTRY)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntReturn must not be followed by IntEntry")
  ovl_etm_intstat4 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_RETURN),
    .next_state  (`ETM_INT_ENTRY)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntReturn must not be followed by IntExit")
  ovl_etm_intstat5 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_RETURN),
    .next_state  (`ETM_INT_EXIT)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntReturn must not be followed by IntReturn")
  ovl_etm_intstat5b (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_RETURN),
    .next_state  (`ETM_INT_RETURN)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntExit must not be followed by IntEntry")
  ovl_etm_intstat7 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_EXIT),
    .next_state  (`ETM_INT_ENTRY)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntExit must not be followed by IntReturn")
  ovl_etm_intstat8 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_EXIT),
    .next_state  (`ETM_INT_RETURN)
                  );
  assert_no_transition #(0,3,`ETM_INTERFACE_TYPE,"IntPend must not be followed by IntReturn")
  ovl_etm_intstat10 (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_int_stat_i),
    .start_state (`ETM_INT_PENDING),
    .next_state  (`ETM_INT_RETURN)
                  );
  assert_frame
    #(0,2,0,0,`ETM_INTERFACE_TYPE,"INTPEND state must last at least 2 cycles")
  ovl_etm_pend_min2
    (
     .clk        (clk),
     .reset_n    (reset_n),
     .start_event(etm_int_stat_i == `ETM_INT_PENDING),
     .test_expr  ((etm_int_stat_i != `ETM_INT_PENDING) &
                  (etm_int_stat_i != `ETM_INT_EXIT) &
                  (etm_int_stat_i != 3'b000))
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"IntStat can only take value in range 0-4")
  ovl_etm_int_stat_range
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i[2]),
     .consequent_expr (~(|etm_int_stat_i[1:0]))
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Int Exit must not be indicated with Ivalid")
  ovl_etm_int_exit_valid2
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_EXIT),
     .consequent_expr (~etm_ivalid_i & ~etm_istall_i)
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Int Exit must not be indicated with Ivalid")
  ovl_etm_int_pend_valid
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_PENDING),
     .consequent_expr (~etm_ivalid_i)
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Int Exit must not be followed by Ivalid")
  ovl_etm_int_exit_valid3
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_exit_t1),
     .consequent_expr (~etm_ivalid_i)
     );
  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"IntEntry must not immediately follow an instruction")
  ovl_etm_int_entry_valid3
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_ENTRY & (etm_int_num_i != 9'b000000000)),
     .consequent_expr (~opt_etm_ivalid_t1)
     );
  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"IntEntry must not immediately follow reset")
  ovl_etm_int_reset
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_ENTRY),
     .consequent_expr (~reset_taken)
     );
  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Debug Exception must not immediately follow reset")
  ovl_etm_dbg_except_reset
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr ((etm_int_stat_i==`ETM_INT_PENDING) & core_halt_i),
     .consequent_expr (~reset_taken)
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Reset exception must not occur with core_halt")
  ovl_etm_reset_halt
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_ENTRY & (etm_int_num_i == 9'b000000000)),
     .consequent_expr (~core_halt_i)
     );

  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"etm_fold_i asserted but not etm_ivalid_i")
  ovl_etm_foldvalid
    (
    .clk             (clk),
    .reset_n         (reset_n),
    .antecedent_expr (etm_fold_i),
    .consequent_expr (etm_ivalid_i)
     );
  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"ISB must be signaled with VALID")
  ovl_etm_isb_valid
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_isb_i),
     .consequent_expr (etm_ivalid_i)
     );
  assert_implication
    #(0,`ETM_INTERFACE_TYPE,"CC fail must only be signaled with VALID")
  ovl_etm_cc_valid
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_cc_fail_i),
     .consequent_expr (etm_ivalid_i)
     );

  assert_never #(0,`ETM_INTERFACE_TYPE,"etm_stall_i not valid with etm_ivalid_i")
  ovl_etm_stallvalid (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_istall_i & etm_ivalid_i)
                  );

    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Core halt must not be asserted at the same time as ivalid")
  ovl_etm_valid_ex_dbg
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_ivalid_i),
     .consequent_expr (~(core_halt_i | ovl_core_halt_t1 | ovl_core_halt_t2 | ovl_core_halt_t3))
     );
    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Stall must not be asserted at the same time as Halt")
  ovl_etm_halt_stall
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_istall_i),
     .consequent_expr (~core_halt_i)
     );
    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Stall must not be asserted at the same time as int entry")
  ovl_etm_entry_stall
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_istall_i),
     .consequent_expr (etm_int_stat_i != `ETM_INT_ENTRY)
     );

    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Exception cannot be signaled after halt")
  ovl_etm_int_dbg
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (ovl_core_halt_t1),
     .consequent_expr (etm_int_stat_i[1:0] == 2'b00)
     );

    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Exception Entry must be more than 3 cycles after debug exit")
  ovl_etm_irq_after_dbg
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_ENTRY),
     .consequent_expr (~(ovl_core_halt_t2 | ovl_core_halt_t3))
     );
    assert_implication

    #(0,`ETM_INTERFACE_TYPE,"Return from Exception can't occur during debug up to 2 cycles after")
  ovl_etm_rfe_in_dbg
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (etm_int_stat_i==`ETM_INT_EXIT),
     .consequent_expr (~(core_halt_i | ovl_core_halt_t1 | ovl_core_halt_t2 | debug_fake_inst))
     );

  assert_window
    #(0,`ETM_INTERFACE_TYPE,"Int Exit cannot be repeated without ivalid")
  ovl_etm_rfe_repeat
  (
   .clk             (clk),
   .reset_n         (reset_n),
   .start_event     ((etm_int_stat_i==`ETM_INT_EXIT) & ~etm_ivalid_i),
   .end_event       (etm_ivalid_i),
   .test_expr       (etm_int_stat_i!=`ETM_INT_EXIT)
   );
  assert_window
    #(0,`ETM_INTERFACE_TYPE,"Int Exit cannot follow Enter without ivalid")
 ovl_etm_rfe_enter
  (
   .clk             (clk),
   .reset_n         (reset_n),
   .start_event     ((etm_int_stat_i==`ETM_INT_ENTRY) & ~etm_ivalid_i),
   .end_event       (etm_ivalid_i),
   .test_expr       (etm_int_stat_i!=`ETM_INT_EXIT)
   );


  assign ovl_etm_ivalid_stall_nxt = ovl_etm_ivalid_t1 |
         (ovl_etm_ivalid_stall & (etm_int_stat_i==3'b000) & ~core_halt_i);

  always @(posedge clk or negedge reset_n)
  if(!reset_n)
    begin
      ovl_etm_ivalid_t1 <= 1'b0;
      ovl_etm_ivalid_t2 <= 1'b0;
      ovl_etm_ivalid_t3 <= 1'b0;
      ovl_etm_ivalid_t4 <= 1'b0;
      ovl_etm_ivalid_stall <= 1'b0;
      ovl_core_halt_t1  <= 1'b0;
      ovl_core_halt_t2  <= 1'b0;
      ovl_core_halt_t3  <= 1'b0;
    end
  else
    begin
      ovl_etm_ivalid_t1 <= etm_ivalid_i;
      ovl_etm_ivalid_t2 <= ovl_etm_ivalid_t1;
      ovl_etm_ivalid_t3 <= ovl_etm_ivalid_t2;
      ovl_etm_ivalid_t4 <= ovl_etm_ivalid_t3;
      ovl_etm_ivalid_stall <= ovl_etm_ivalid_stall_nxt;
      ovl_core_halt_t1  <= core_halt_i;
      ovl_core_halt_t2  <= ovl_core_halt_t1;
      ovl_core_halt_t3  <= ovl_core_halt_t2;
    end
    assert_implication
    #(0,`ETM_INTERFACE_TYPE,"Debug exit must occur 3 cycles after last Ivalid")
  ovl_etm_exitdbg_assume
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (ovl_core_halt_t1 & ~core_halt_i),
     .consequent_expr (~(etm_ivalid_i | ovl_etm_ivalid_t1 |
                         ovl_etm_ivalid_t2 | ovl_etm_ivalid_t3))
     );


  assert_never #(0,0,"Exit Debug Ivalid occured with E count > 0")
  ovl_etm_exitdbg (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (exception_leaving_debug_t1 & (|p_header_e_count_t1))
               );

  assert_transition #(0,1,`ETM_INTERFACE_TYPE,"Branch must not be immediately followed by branch")
  ovl_etm_br_back2back (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (etm_branch_i),
    .start_state (1'b1),
    .next_state  (1'b0)
                  );
  assert_transition #(0,2,0,"Back to back exceptions traced")
  ovl_etm_ex_back2back (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   ({p_header_valid_t2 & exception_valid_t2,exception_valid_t2}),
    .start_state (2'b01),
    .next_state  (2'b00)
                  );

  assert_zero_one_hot
    #(0,2,0,"Only one type of packet can be generated at once")
  ovl_etm_packet_select (
    .clk        (clk),
    .reset_n    (reset_n),
    .test_expr  ({isync_valid_t2 | rfe_valid_t2,
                  address_byte_valid_t2[0]})
  );

  assert_never
    #(0,0,"At least two address bytes must be output with an exception")
  ovl_etm_2addrex (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   ((|exception_byte_valid_t2) & ~(&address_byte_valid_t2[1:0]))
               );
  assert_never
        #(0,0,"Exception 2nd byte must come with 1st byte")
  ovl_etm_orderex (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (exception_byte_valid_t2==2'b10)
               );
  assert_never
        #(0,0,"Address bytes must come with all lower bytes")
  ovl_etm_orderad (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   ((address_byte_valid_t2[1:0]==2'b10) |
                  (address_byte_valid_t2[2:1]==2'b10) |
                  (address_byte_valid_t2[3:2]==2'b10) |
                  (address_byte_valid_t2[4:3]==2'b10)
                  )
               );

  assert_never
        #(0,0,"If exception bit is set, need exception byte")
  ovl_etm_expresent(
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   ((((address_byte_valid_t2[2:1]==2'b01) &
                    compressed_address_t2[14] ) |
                   ((address_byte_valid_t2[3:2]==2'b01) &
                    compressed_address_t2[22] ) |
                   ((address_byte_valid_t2[4:3]==2'b01) &
                    compressed_address_t2[30] ))
                  & ~exception_byte_valid_t2[0])
                );

  assert_never_unknown  #(0,7,0,"Address size generator inputs are x")
  ovl_etm_addrx(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr ({exception_byte_valid_t2, address_byte_valid_t2}));

  assert_never
        #(0,0,"Flush Trace should only occur when trace_enable_i is low")
  ovl_etm_flush_en (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (trace_enable_i & flush_trace_i));

  assert_implication
    #(0,0,"Timestamp should only be generated when enabled")
  ovl_etm_ts_enabled
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (ts_accepted_t2),
     .consequent_expr (prg_ts_enable_i)
     );

  assert_implication
    #(0,0,"Branch bytes marked valid, but no branch")
  ovl_etm_branch_bytes
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (|{exception_byte_valid_t2,address_byte_valid_t2}),
     .consequent_expr (branch_traced_t2 | isync_valid_t2)
     );

  assert_implication
    #(0,0,"Must not generate periodic Isync if branched_traced_t1")
  ovl_etm_no_periodic_isync_branch
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (isync_periodic_t1),
     .consequent_expr (~branch_traced_t1)
     );

  assert_implication
    #(0,0,"Must not generate periodic Isync if exception_valid_t1")
  ovl_etm_no_periodic_isync_exception
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (isync_periodic_t1),
     .consequent_expr (~exception_valid_t1)
     );

  assert_implication
    #(0,0,"Must flush P-header if exception is traced")
  ovl_etm_exception_possible
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .antecedent_expr (exception_valid_t1),
     .consequent_expr (exception_possible_t1)
     );

  wire [2:0]  ovl_compressed_address_size_t2;
  assign ovl_compressed_address_size_t2 = exception_byte_valid_t2[0] +
         exception_byte_valid_t2[1] +
         address_byte_valid_t2[4] +
         address_byte_valid_t2[3] +
         address_byte_valid_t2[2] +
         address_byte_valid_t2[1] +
         address_byte_valid_t2[0];

    assert_never
        #(0,0,"address byte counting failed")
  ovl_etm_address_size
    (
    .clk         (clk),
    .reset_n     (reset_n),
    .test_expr   (ovl_compressed_address_size_t2 != compressed_address_size_t2));

  assert_frame
    #(`OVL_ERROR, 0, 5, `OVL_IGNORE_NEW_START, `OVL_ASSERT,
      "Instruction Synchronisation stalled, but fifo is empty")
  u_ovl_stall_isync
    (.clk             (clk),
     .reset_n         (reset_n),
     .start_event     (fifo_empty_i & isync_pending_t2 & isync_valid_t2),
     .test_expr       (~isync_pending_t2 | ~fifo_empty_i));

  assert_frame
    #(`OVL_ERROR, 0, 5, `OVL_IGNORE_NEW_START, `OVL_ASSERT,
      "Timestamp Synchronisation stalled, but fifo is empty")
  u_ovl_stall_tsync
    (.clk             (clk),
     .reset_n         (reset_n),
     .start_event     (fifo_empty_i & ts_valid_t2 & ts_pend_t2),
     .test_expr       (~ts_pend_t2 | ~fifo_empty_i |isync_pending_t2));



  assert_never_unknown  #(0,1,0,"etm_trigger_out_we is x")
  ovl_etm_etm_trigger_out_we_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (etm_trigger_out_we));

  assert_never_unknown  #(0,1,0,"fifo_overflow_t1 is x")
  ovl_etm_fifo_overflow_t1_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (fifo_overflow_t1));

  assert_never_unknown  #(0,1,0,"trace_enable_i is x")
  ovl_etm_trace_enable_i_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (trace_enable_i));

  assert_never_unknown  #(0,1,0,"turn_on_allowed_t1 is x")
  ovl_etm_turn_on_allowed_t1_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (turn_on_allowed_t1));

  assert_never_unknown  #(0,1,0,"power_down_reset_i is x")
  ovl_etm_power_down_reset_i_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (power_down_reset_i));

  assert_never_unknown  #(0,1,0,"inst_valid_confirmed_t1 is x")
  ovl_etm_inst_valid_confirmed_t1_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (inst_valid_confirmed_t1));

  assert_never_unknown  #(0,1,0,"entering_debug_t1 is x")
  ovl_etm_entering_debug_t1_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (entering_debug_t1));

  assert_never_unknown  #(0,1,0,"flush_trace_i is x")
  ovl_etm_flush_trace_i_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (flush_trace_i));

  assert_never_unknown  #(0,1,0,"fifo_empty_i is x")
  ovl_etm_fifo_empty_i_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (fifo_empty_i));

  assert_never_unknown  #(0,1,0,"reference_address_we is x")
  ovl_etm_reference_address_we_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (reference_address_we));

  assert_never_unknown  #(0,1,0,"ts_packet_we is x")
  ovl_etm_ts_packet_we_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (ts_packet_we));

  assert_never_unknown  #(0,1,0,"ts_packet_valid_we is x")
  ovl_etm_ts_packet_valid_we_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (ts_packet_valid_we));

  assert_never_unknown  #(0,1,0,"isync_valid_t2 is x")
  ovl_etm_isync_valid_t2_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (isync_valid_t2));

  assert_never_unknown  #(0,1,0,"emit_rfe_t2 is x")
  ovl_etm_emit_rfe_t2_x(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (emit_rfe_t2));


`endif

endmodule
