module cm3_etm_sync_count (
    clk, reset_n, prg_prog_bit_i, prg_prog_bit_fall_i, pop_fifo_i,
    isync_req_o, async_req_o
    );

  parameter TRACE_LVL = 1;

  input         clk;               // Gated clock
  input         reset_n;           // ETM reset
  input         prg_prog_bit_i;    // Programming bit
  input         prg_prog_bit_fall_i; // Prog bit registered
  input         pop_fifo_i;        // A byte has been from fifo

  output        isync_req_o;       // Instruction sync request
  output        async_req_o;       // Alignment sync request

  wire  [8:0]   nxt_count;         // Count register input
  reg   [8:0]   opt_count;         // Count register
  wire  [8:0]   count;             // Count value
  wire          count_we;          // Count enable

  wire          reset_count;       // Pulsed when counter wraps

  wire          nxt_sync_type;     // 1'b1 is I-Sync, 1'b0 is A-Sync
  reg           opt_sync_type;     // 1'b1 is I-Sync, 1'b0 is A-Sync
  wire          sync_type;         // 1'b1 is I-Sync, 1'b0 is A-Sync
  wire          sync_set_en;       // Sync event ready
  wire          sync_type_we;      // Sync type flop enable

  wire          opt_isync_req;     // Instruction sync request
  wire          opt_async_req;     // Alignment sync request
  wire          opt_etm_en;        // Logic optimisation enable term

  assign opt_etm_en  = (TRACE_LVL > 1) ? 1'b1          : 1'b0;


  assign nxt_count = ((TRACE_LVL > 1) ?
                     ({9{reset_count}} | (count - 9'b000000001)) :
                      {9{1'b0}});

  assign count_we = (TRACE_LVL > 1) ? (~prg_prog_bit_i &
                                       (reset_count | pop_fifo_i)) : 1'b0;

  assign reset_count = !(|count) | prg_prog_bit_fall_i;

  assign nxt_sync_type = (TRACE_LVL > 1) ?
                           (~prg_prog_bit_i & ~sync_type) :
                             1'b0;

  assign sync_set_en = reset_count & ~prg_prog_bit_i;
  assign sync_type_we = (sync_set_en | prg_prog_bit_i) & opt_etm_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_count <= 9'b111111111;
    else if (count_we)
      opt_count <= nxt_count;

  assign count = (TRACE_LVL > 1) ? opt_count : {9{1'b0}};

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_sync_type <= 1'b0;
    else if (sync_type_we)
      opt_sync_type <= nxt_sync_type;

  assign sync_type = (TRACE_LVL > 1) ? opt_sync_type : 1'b0;

  assign opt_isync_req =  sync_type & sync_set_en;

  assign opt_async_req = ~sync_type & sync_set_en;

  assign isync_req_o = (TRACE_LVL > 1) ? opt_isync_req : 1'b0;
  assign async_req_o = (TRACE_LVL > 1) ? opt_async_req : 1'b0;


`ifdef ARM_ASSERT_ON
  assert_never_unknown #(0, 1,0, "count_we must not be X")
  ovl_count_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (count_we));

  assert_never_unknown #(0, 1,0, "sync_type_we must not be X")
  ovl_sync_type_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (sync_type_we));
`endif

endmodule
