`include "cm3_etm_defs.vp"

module cm3_etm_trace
  (
   clk, reset_n, prg_prog_bit_i, prg_prog_bit_fall_i, etm_ia_i, etm_ivalid_i,
   etm_istall_i, etm_cc_fail_i, etm_branch_i, etm_indirect_branch_i,
   etm_isb_i, etm_int_stat_i,
   etm_fold_i, etm_cancel_i, etm_int_num_i, core_halt_i, trigger_req_i,
   pop_fifo_i, prg_branch_broadcast_i, trace_enable_i, power_down_reset_i,
   integration_mode_i,  apb_wdata_0_i, integration_trigger_we_i,
   prg_fifo_full_level_i, prg_stall_processor_i, prg_ts_enable_i,
   flush_trace_i, niden_t2_i, ts_value_i, ts_clk_change_i, timestamp_event_i,
   reason_overflow_o, fifo_out_o, fifo_empty_o, etm_trigger_out_o, fifo_stall_o,
   fake_ivalid_t1_o, etm_istall_t1_o, etm_ivalid_t1_o
   );

  parameter TRACE_LVL      = 1;
  parameter RESET_ALL_REGS = 0;


  input         clk;                // Gated ETM Clock
  input         reset_n;            // ETM Reset
  input         prg_prog_bit_i;     // Progbit
  input         prg_prog_bit_fall_i;// Progbit falling edge
  input [31:1]  etm_ia_i;           // Instruction address from core
  input         etm_ivalid_i;       // Instruction address valid
  input         etm_istall_i;       // Previous instruction not yet confirmed
  input         etm_fold_i;         // 2 instructions executed in one cycle

  input         etm_cc_fail_i;      // Condition code result for instruction
  input         etm_branch_i;       // instruction is a branch
  input         etm_indirect_branch_i; // instruction is an indirect branch
  input         etm_isb_i;          // Instruction is ISB
  input [2:0]   etm_int_stat_i;     // Interrupt status
  input         etm_cancel_i;       // Previous instruction did not execute
  input [8:0]   etm_int_num_i;      // ISR number
  input         core_halt_i;        // Core in debug state
  input         trigger_req_i;      // Trigger must be inserted in trace
  input         pop_fifo_i;         // Fifo data taken
  input         prg_branch_broadcast_i; // Control reg branch broadcast enable
  input         prg_stall_processor_i;  // Control reg fifo-full enable
  input         prg_ts_enable_i;        // Control register timestamp enable

  input         trace_enable_i;     // Trace enable
  input         flush_trace_i;      // Request to flush trace
  input         power_down_reset_i; // Initialise state after power-down

  input         integration_mode_i; // Integration mode select
  input         apb_wdata_0_i;      // APB Write data
  input         integration_trigger_we_i; // Integration mode trigger output register write

  input  [(`FIFO_DEPTH_SIZE-1):0]   prg_fifo_full_level_i; // Fifo Full level register

  input         niden_t2_i;         // Non-invasive debug enable, aligned to instructions
  input [47:0]  ts_value_i;         // Binary coded timestamp clock value
  input         ts_clk_change_i;    // Timestamp clock ratio changed
  input         timestamp_event_i;  // Timestamp event


  output        reason_overflow_o;  // Un-traced overflow indicator
  output [7:0]  fifo_out_o;         // Output from trace fifo
  output        fifo_empty_o;       // Trace fifo has no data
  output        etm_trigger_out_o;  // Trigger has been inserted into trace
  output        fifo_stall_o;       // Fifo Depth exceeds preset level on prg_fifo_full_level_i
  output        fake_ivalid_t1_o;   // Instruction data is valid, with faked excpn
  output        etm_istall_t1_o;    // ISTALL input delayed one cycle
  output        etm_ivalid_t1_o;    // IVALID input delayed one cycle

  wire [63:0]   fifo_in;            // From uControl of etm_gen
  wire          fifo_overflow;      // From uFifo of etm_gen
  wire [3:0]    num_bytes;          // From uControl of etm_gen

  wire          isync_req;
  wire          async_req;

  wire          reason_overflow;
  wire [7:0]    fifo_out;
  wire          fifo_empty;
  wire          etm_trigger_out;
  wire          fifo_stall;
  wire          fifo_ts_holdoff;

  cm3_etm_gen #(TRACE_LVL, RESET_ALL_REGS)
    u_control
    (
    .clk                      (clk),
    .reset_n                  (reset_n),
    .etm_ivalid_i             (etm_ivalid_i),
    .etm_istall_i             (etm_istall_i),
    .etm_ia_i                 (etm_ia_i[31:1]),
    .etm_fold_i               (etm_fold_i),
    .etm_cancel_i             (etm_cancel_i),
    .etm_cc_fail_i            (etm_cc_fail_i),
    .etm_branch_i             (etm_branch_i),
    .etm_indirect_branch_i    (etm_indirect_branch_i),
    .etm_isb_i                (etm_isb_i),
    .etm_int_stat_i           (etm_int_stat_i),
    .etm_int_num_i            (etm_int_num_i),
    .core_halt_i              (core_halt_i),
    .prg_prog_bit_fall_i      (prg_prog_bit_fall_i),
    .prg_branch_broadcast_i   (prg_branch_broadcast_i),
    .prg_ts_enable_i          (prg_ts_enable_i),
    .isync_req_i              (isync_req),
    .trigger_req_t2_i         (trigger_req_i),
    .fifo_overflow_i          (fifo_overflow),
    .fifo_empty_i             (fifo_empty),
    .fifo_ts_holdoff_i        (fifo_ts_holdoff),
    .trace_enable_i           (trace_enable_i),
    .flush_trace_i            (flush_trace_i),
    .power_down_reset_i       (power_down_reset_i),
    .integration_mode_i       (integration_mode_i),
    .apb_wdata_0_i            (apb_wdata_0_i),
    .integration_trigger_we_i (integration_trigger_we_i),
    .niden_t2_i               (niden_t2_i),
    .ts_value_i               (ts_value_i),
    .ts_clk_change_i          (ts_clk_change_i),
    .timestamp_event_i        (timestamp_event_i),
    .fifo_in_o                (fifo_in),
    .num_bytes_o              (num_bytes),
    .reason_overflow_o        (reason_overflow),
    .etm_trigger_out_o        (etm_trigger_out),
    .fake_ivalid_t1_o         (fake_ivalid_t1_o),
    .etm_istall_t1_o          (etm_istall_t1_o),
    .etm_ivalid_t1_o          (etm_ivalid_t1_o));

  cm3_etm_fifo #(TRACE_LVL, RESET_ALL_REGS)
    u_fifo
     (// Inputs
    .reset_n                 (reset_n),
    .clk                     (clk),
    .fifo_in_i               (fifo_in),
    .num_bytes_i             (num_bytes),
    .pop_fifo_i              (pop_fifo_i),
    .async_req_i             (async_req),
    .trigger_req_i           (trigger_req_i),
    .prg_fifo_full_level_i   (prg_fifo_full_level_i),
    .prg_stall_processor_i   (prg_stall_processor_i),
    .prg_prog_bit_i          (prg_prog_bit_i),
    .fifo_empty_o            (fifo_empty),
    .fifo_overflow_o         (fifo_overflow),
    .fifo_out_o              (fifo_out),
    .fifo_stall_o            (fifo_stall),
    .fifo_ts_holdoff_o       (fifo_ts_holdoff)
     );

  cm3_etm_sync_count #(TRACE_LVL)
    u_sync_count
    (/*AUTOINST*/
    .clk                     (clk),
    .reset_n                 (reset_n),
    .prg_prog_bit_i          (prg_prog_bit_i),
    .prg_prog_bit_fall_i     (prg_prog_bit_fall_i),
    .pop_fifo_i              (pop_fifo_i),
    .isync_req_o             (isync_req),
    .async_req_o             (async_req)
    );

  assign reason_overflow_o = (TRACE_LVL > 1) ? reason_overflow : 1'b0;
  assign fifo_out_o        = (TRACE_LVL > 1) ? fifo_out        : {8{1'b0}};
  assign fifo_empty_o      = (TRACE_LVL > 1) ? fifo_empty      : 1'b0;
  assign etm_trigger_out_o = (TRACE_LVL > 1) ? etm_trigger_out     : 1'b0;
  assign fifo_stall_o      = (TRACE_LVL > 1) ? fifo_stall      : 1'b0;

`ifdef ARM_ASSERT_ON
`include "std_ovl_defines.h"
  reg    ovl_prg_prog_bit_reg;
  reg    ovl_prg_prog_bit_reg1;
  reg    ovl_prg_prog_bit_reg2;
  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        ovl_prg_prog_bit_reg <= 1'b0;
        ovl_prg_prog_bit_reg1 <= 1'b0;
        ovl_prg_prog_bit_reg2 <= 1'b0;
        end
    else
      begin
        ovl_prg_prog_bit_reg <= prg_prog_bit_i;
        ovl_prg_prog_bit_reg1 <= ovl_prg_prog_bit_reg;
        ovl_prg_prog_bit_reg2 <= ovl_prg_prog_bit_reg1;
      end // else: !if(!reset_n)
  wire ovl_prog_state = prg_prog_bit_i & ovl_prg_prog_bit_reg2 &
                        ovl_prg_prog_bit_reg1 & ovl_prg_prog_bit_reg;

  assert_no_transition
    #(0,2,0,"Fifo Should remain empty whilst ProgBit is asserted")
  ovl_etm_progbit_fifo_empty
    (
     .clk             (clk),
     .reset_n         (reset_n),
     .test_expr       ({ovl_prog_state, fifo_empty}),
     .start_state     (2'b11),
     .next_state      (2'b10)
     );
  assert_frame
    #(`OVL_ERROR, 0, 20, `OVL_IGNORE_NEW_START, `OVL_ASSERT,
      "TimeStamp holdoff must not remain high (except due to fifo full or Isync pend)")
  ovl_etm_ts_never_insert
    (
     .clk         (clk),
     .reset_n     (reset_n),
     .start_event (u_control.nxt_ts_holdoff_t2 & ~fifo_ts_holdoff & ~u_control.opt_etm_istall_t1 & prg_ts_enable_i),
     .test_expr   (~u_control.nxt_ts_holdoff_t2 | fifo_ts_holdoff |
                   prg_prog_bit_i | u_control.isync_pending_t2 | u_control.opt_etm_istall_t1)
     );
`endif
endmodule // cm3_etm_trace
