module cm3_etm_trace_out (
  clk, reset_n, fifo_out_i, fifo_empty_i, atb_atreadym_i, integration_mode_i,
  apb_wdata_0_i, integration_atb_cr0_we_i,
  atb_atdatam_o, atb_atvalidm_o, atb_afreadym_o, pop_fifo_o, trace_flushed_o
  );

  parameter TRACE_LVL = 1;

  input        clk;                // ATB Bus Clock
  input        reset_n;            // ATB Bus Reset
  input [7:0]  fifo_out_i;         // Output Data from Fifo
  input        fifo_empty_i;       // Current Fifo Depth
  input        atb_atreadym_i;     // ATB Bus Ready to accept data
  input        integration_mode_i; // Integration mode control
  input        apb_wdata_0_i;      // APB write data
  input        integration_atb_cr0_we_i; // Integration register write enable

  output [7:0] atb_atdatam_o;      // ATB Data
  output       atb_atvalidm_o;     // ATDATA Valid
  output       atb_afreadym_o;     // Fifo Flush Finished
  output       pop_fifo_o;         // Byte has been read from Fifo
  output       trace_flushed_o;    // Fifo has drained

  wire         atb_we;             // Enable for output registers

  wire [7:0]   nxt_atb_atdatam;    // D-input for output register
  wire [7:0]   atb_atdatam;        // Output Register
  reg  [7:0]   opt_atb_atdatam;    // ATB data
  wire         nxt_atb_atvalidm;   // D-input for output valid register
  wire         atb_atvalidm;       // Output Valid register
  reg          opt_atb_atvalidm;   // ATB data valid

  reg          opt_atb_afreadym;   // ATB flush acknowledge
  wire         opt_pop_fifo;       // Fifo data taken
  wire         opt_trace_flushed;  // Flush completed
  wire         pop_fifo;           // Fifo data taken

  assign pop_fifo   = (TRACE_LVL > 1) ? opt_pop_fifo  : 1'b0;


  assign opt_pop_fifo = (~fifo_empty_i & atb_we);

  assign nxt_atb_atdatam  = {8{~integration_mode_i}} & fifo_out_i;

  assign nxt_atb_atvalidm = (TRACE_LVL > 1) ? (integration_mode_i ?
                                               apb_wdata_0_i : ~fifo_empty_i) : 1'b0;

  assign atb_we = ((TRACE_LVL > 1) ?
                   (integration_mode_i ? integration_atb_cr0_we_i :
                    ((atb_atvalidm & atb_atreadym_i) | ~atb_atvalidm)) :
                   1'b0);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        opt_atb_atdatam  <= 8'h00;
        opt_atb_atvalidm <= 1'b0;
        opt_atb_afreadym <= 1'b1;
      end
    else if (atb_we)
      begin
        opt_atb_atdatam  <= nxt_atb_atdatam;
        opt_atb_atvalidm <= nxt_atb_atvalidm;
        opt_atb_afreadym <= fifo_empty_i;
      end

   assign atb_atdatam  = (TRACE_LVL > 1) ? opt_atb_atdatam  : 8'h00;
   assign atb_atvalidm = (TRACE_LVL > 1) ? opt_atb_atvalidm : 1'b0;

  assign opt_trace_flushed = ~atb_atvalidm & fifo_empty_i;

  assign atb_atdatam_o   = atb_atdatam;
  assign atb_atvalidm_o  = atb_atvalidm;
  assign atb_afreadym_o  = (TRACE_LVL > 1) ? opt_atb_afreadym  : 1'b1;
  assign trace_flushed_o = (TRACE_LVL > 1) ? opt_trace_flushed : 1'b0;
  assign pop_fifo_o      = pop_fifo;

`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Pop can never be asserted when Fifo is empty.")
    FifoDepth_lt_0
      (
    .clk       (clk),
    .reset_n   (reset_n),
    .test_expr (pop_fifo & fifo_empty_i)
       );

  assert_never_unknown #(0, 1,0, "atb_we must not be X")
  ovl_atb_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (atb_we));

`endif

endmodule
