module cm3_etm_trc_en
  (
   clk, reset_n, prg_prog_bit_i, prg_prog_bit_fall_i, prg_trc_en_event_i,
   prg_count_reload_i, prg_ts_event_i,
   stat_reg_write_i, stat_reg_written_i, apb_wdata_2_i,
   ext_in_i, dwt_match_i, dwt_i_not_d_i, prg_trc_en_start_stop_i,
   prg_trc_start_i, prg_trc_stop_i, niden_t1_i, etm_ivalid_i,
   etm_dvalid_i,
   trace_enable_o, flush_trace_o, rsrc_start_on_o, count_zero_o,
   timestamp_event_o
  );

  parameter TRACE_LVL = 1;
  parameter RESET_ALL_REGS = 0;

  input        clk;                // gated ETM Clock
  input        reset_n;            // Reset

  input        prg_prog_bit_i;     // Programming bit
  input        prg_prog_bit_fall_i;// Programming bit falling edge
  input [10:0] prg_trc_en_event_i; // Trace Enable Event Reg
  input [15:0] prg_count_reload_i; // Counter reload value
  input [10:0] prg_ts_event_i;     // timestamp Event Reg

  input        stat_reg_write_i;   // Write enable for status register
  input        stat_reg_written_i; // Status register written
  input        apb_wdata_2_i;      // Write data

  input [1:0]  ext_in_i;           // EXTIN Events
  input [3:0]  dwt_match_i;        // External address comparators
  input [3:0]  dwt_i_not_d_i;      // DWT compare was for instruction

  input        prg_trc_en_start_stop_i; // Trace Enable Ctrl 1 Reg[25]
  input [3:0]  prg_trc_start_i;    // Start/Stop control reg
  input [3:0]  prg_trc_stop_i;     // Start/Stop control reg
  input        etm_ivalid_i;       // PC address valid
  input        etm_dvalid_i;       // PC address valid
  input        niden_t1_i;         // Non-intrusive-debug enable

  output       trace_enable_o;     // Trace Enable O/P
  output       flush_trace_o;      // Request to drain all trace
  output       rsrc_start_on_o;    // Trc On/Off block Event O/P
  output       count_zero_o;       // Counter at zero
  output       timestamp_event_o;


  wire         trc_en_event;       // Trace enable event match
  wire [7:0]   event_rsrc_bus;     // Event generator input
  wire         rsrc_trc_start;     // Trc Start/Stop Rsrc On Events
  wire         rsrc_trc_stop;      // Trc Start/Stop Rsrc Off Events
  wire         trace_on;           // Trc On/Off signal used in TE calc
  reg          opt_rsrc_start_on;  // Trc On/Off rsrc
  wire         rsrc_start_on;      // Trc On/Off rsrc
  wire         nxt_rsrc_start_on;  // Trc On/Off rsrc
  wire [3:0]   opt_dwt_match;      // Inputs from DWT
  reg  [3:0]   opt_latched_dwt;    // Inputs from DWT, held between instructions
  wire [3:0]   latched_dwt;        // Inputs from DWT, held between instructions
  wire [3:0]   dwt_latch_we;       // Enable for DWT holding registers
  wire [3:0]   rsrc_eice;          // EICE resource from DWT inputs
  reg          ivalid_reg;         // Pipelined Instruction address valid
  reg          dvalid_reg;         // Pipelined Data address valid
  wire         opt_ivalid;         // Logic removal signals
  wire         opt_dvalid;         // Logic removal signals
  wire         count_we;
  wire         reload;
  wire [15:0]  count_sub;
  wire [15:0]  nxt_count_value;
  wire         nxt_count_zero;
  reg [15:0]   count_value;
  reg          count_zero;

  wire         timestamp_event;


  wire         opt_trace_enable;   // Trace enable output
  wire         opt_flush_trace;    // Trace flush request
  wire         opt_etm_en;         // Logic removal enable term
  wire         opt_reset_n;        // Optional reset all flops


  assign opt_etm_en     = (TRACE_LVL > 1) ? 1'b1         : 1'b0;
  assign opt_ivalid     = (TRACE_LVL > 1) ? etm_ivalid_i : 1'b0;
  assign opt_dvalid     = (TRACE_LVL > 1) ? etm_dvalid_i : 1'b0;
  assign opt_reset_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;



  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        ivalid_reg <= 1'b0;
        dvalid_reg <= 1'b0;
      end
    else if (opt_etm_en)
      begin
        ivalid_reg <= opt_ivalid;
        dvalid_reg <= opt_dvalid;
      end

  assign dwt_latch_we[0] = (dwt_i_not_d_i[0] ? ivalid_reg : dvalid_reg);
  assign dwt_latch_we[1] = (dwt_i_not_d_i[1] ? ivalid_reg : dvalid_reg);
  assign dwt_latch_we[2] = (dwt_i_not_d_i[2] ? ivalid_reg : dvalid_reg);
  assign dwt_latch_we[3] = (dwt_i_not_d_i[3] ? ivalid_reg : dvalid_reg);

  assign opt_dwt_match = (TRACE_LVL > 1) ? dwt_match_i : 4'b0000;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_latched_dwt[0] <= 1'b0;
    else if (dwt_latch_we[0])
      opt_latched_dwt[0] <= opt_dwt_match[0];

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_latched_dwt[1] <= 1'b0;
    else if (dwt_latch_we[1])
      opt_latched_dwt[1] <= opt_dwt_match[1];

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_latched_dwt[2] <= 1'b0;
    else if (dwt_latch_we[2])
      opt_latched_dwt[2] <= opt_dwt_match[2];

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_latched_dwt[3] <= 1'b0;
    else if (dwt_latch_we[3])
      opt_latched_dwt[3] <= opt_dwt_match[3];

  assign latched_dwt = (TRACE_LVL > 1) ? opt_latched_dwt : 4'b0000;
  assign rsrc_eice   = (latched_dwt & ~dwt_latch_we) | dwt_match_i;


  assign event_rsrc_bus = {ext_in_i,
                         rsrc_start_on,
                         count_zero,
                         rsrc_eice};

  cm3_etm_event_gen #(TRACE_LVL)
    u_trc_en_event
      (// Inputs
    .mux_in_i     (event_rsrc_bus),
    .mux_sel_i    (prg_trc_en_event_i),
    .event_o      (trc_en_event)
       );

  assign rsrc_trc_start = |(prg_trc_start_i & dwt_match_i);
  assign rsrc_trc_stop  = |(prg_trc_stop_i & dwt_match_i);

  assign nxt_rsrc_start_on = ((TRACE_LVL > 1) ?
                         ((apb_wdata_2_i & stat_reg_write_i) |
                         (prg_prog_bit_i & rsrc_start_on & ~stat_reg_write_i) |
                         (prg_prog_bit_fall_i & stat_reg_written_i &
                          rsrc_start_on ) |
                         (rsrc_trc_start & ~stat_reg_write_i & ~prg_prog_bit_i) |
                         (rsrc_start_on & ~rsrc_trc_stop &
                          ~prg_prog_bit_i & ~prg_prog_bit_fall_i)) :
                         1'b0);

  assign trace_on = ~(prg_prog_bit_fall_i | stat_reg_write_i) & rsrc_start_on;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      opt_rsrc_start_on <= 1'b0;
    else if (opt_etm_en)
      opt_rsrc_start_on <= nxt_rsrc_start_on;



  assign opt_trace_enable = (niden_t1_i &
                         (trc_en_event & ~prg_prog_bit_i) &
                         (trace_on | ~prg_trc_en_start_stop_i));

  assign opt_flush_trace  = ~niden_t1_i | prg_prog_bit_i;


  assign count_we = prg_prog_bit_fall_i | (~prg_prog_bit_i & niden_t1_i);

  assign reload = prg_prog_bit_fall_i | count_zero;

  assign count_sub = (count_value - {{15{1'b0}},1'b1});

  assign nxt_count_value = ({16{~prg_prog_bit_fall_i & ~count_zero}} &
                            count_sub[15:0]) |
                             ({16{reload}} & prg_count_reload_i);

  assign nxt_count_zero = ((count_value == 16'h0001) & ~prg_prog_bit_fall_i);

  always @(posedge clk or negedge opt_reset_n)
    if (!opt_reset_n)
      count_value <= {16{1'b1}};
    else if (count_we)
      count_value    <= nxt_count_value;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      count_zero  <= 1'b0;
    else if (count_we)
      count_zero  <= nxt_count_zero;

  cm3_etm_event_gen #(TRACE_LVL)
  u_timestamp_event
    (// Inputs
     .mux_in_i  (event_rsrc_bus),
    .mux_sel_i  (prg_ts_event_i),
    .event_o    (timestamp_event)
     );


  assign rsrc_start_on_o   = (TRACE_LVL > 1) ? rsrc_start_on     : 1'b0;
  assign trace_enable_o    = (TRACE_LVL > 1) ? opt_trace_enable  : 1'b0;
  assign flush_trace_o     = (TRACE_LVL > 1) ? opt_flush_trace   : 1'b0;
  assign rsrc_start_on     = (TRACE_LVL > 1) ? opt_rsrc_start_on : 1'b0;
  assign count_zero_o      = (TRACE_LVL > 1) ? count_zero        : 1'b0;
  assign timestamp_event_o = (TRACE_LVL > 1) ? timestamp_event   : 1'b0;

`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0, 4,0, "dwt_latch_we must not be X")
  ovl_dwt_latch_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (dwt_latch_we));

  assert_never_unknown #(0, 1,0, "count_we must not be X")
  ovl_count_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (count_we));

`endif

endmodule // cm3_etm_trc_en
