`include "cm3_etm_defs.vp"

module cm3_etm_trig_gen (
  clk, reset_n, apb_addr_i, register_write_i, apb_wdata_3_i, prg_prog_bit_i,
  prg_prog_bit_fall_i, prg_prog_bit_and_en_i,
  power_down_reset_i, prg_trigger_event_i, rsrc_start_on_i, count_zero_i,
  ext_in_i, dwt_match_i, prg_may_assert_dbgreq_i, core_halt_i, niden_t0_i,
  trigger_req_o, etm_dbg_req_o, rsrc_trig_detect_o
  );

  parameter TRACE_LVL = 1;


  input         clk;                   // gated ETM Clock
  input         reset_n;               // Sync'd ETM Reset
  input [11:2]  apb_addr_i;            // Register Write Addr
  input         register_write_i;      // Register Write Valid
  input         apb_wdata_3_i;         // Register Write Data
  input         prg_prog_bit_i;        // Programming Bit
  input         prg_prog_bit_fall_i;   // Prog Bit clear transition
  input         prg_prog_bit_and_en_i; // Prog Bit Qual'd with ETMEN
  input         power_down_reset_i;    // ETM has come out of powered down st
  input [10:0]  prg_trigger_event_i  ; // Trigger Event Register

  input         rsrc_start_on_i;       // Trc Start/Stop o/p
  input         count_zero_i;          // Counter at zero
  input [1:0]   ext_in_i;              // EXTIN Events
  input [3:0]   dwt_match_i;           // From DWT
  input         niden_t0_i;            // Non-intrusive debug enable, T0 stage

  input         prg_may_assert_dbgreq_i;// Control Reg bit 9
  input         core_halt_i;           // Core is in debug state

  output        trigger_req_o;         // Trigger signal (pipelined)
  output        etm_dbg_req_o;         // Dbgrq to put core into debug state
  output        rsrc_trig_detect_o;    // Trigger has been generated


  wire          trigger_event;         // Trigger event select output
  wire          nxt_trig_detect;       // nxt st trigger has been generated
  wire [7:0]    event_rsrc_bus;        // Resource bus

  reg [1:0]     opt_trigger_state;     // Trigger state machine state
  wire [1:0]    trigger_state;         // Trigger state machine
  reg [1:0]     nxt_trigger_state;     // Trigger state machine
  reg           nxt_dbg_req;           // nxt st for dbgrq o/p
  wire          trg_state_we;          // trig state wr enable
  wire          nxt_trg_state_written; // nxt st trigger st has been written
  reg           opt_trg_state_written; // trigger st has been written
  wire          trg_state_written;     // Trigger state written, will not reload
  wire          opt_trigger_req;       // Trigger request
  reg           opt_etm_dbg_req;       // Debug request to core
  reg           opt_rsrc_trig_detect;  // Trigger for status register read
  wire          opt_etm_en;            // Global enable term for logic removal

  assign opt_etm_en         = (TRACE_LVL > 1) ? 1'b1             : 1'b0;

  assign event_rsrc_bus = {ext_in_i,
                           rsrc_start_on_i,
                           count_zero_i,
                           dwt_match_i};

  cm3_etm_event_gen #(TRACE_LVL)
    u_trig_event
      (// inputs
    .mux_in_i    (event_rsrc_bus),
    .mux_sel_i   (prg_trigger_event_i),
    .event_o     (trigger_event)
       );


  assign trg_state_we    = (apb_addr_i[11:2] == `AddrStatus) && register_write_i;
  assign nxt_trg_state_written = ((TRACE_LVL > 1) ?
                                  ((prg_prog_bit_i & trg_state_we) |
                                   (trg_state_written & ~(prg_prog_bit_fall_i))) :
                                  1'b0);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      opt_trg_state_written <= 1'b0;
    else if (opt_etm_en)
      opt_trg_state_written <= nxt_trg_state_written;

   assign trg_state_written = (TRACE_LVL > 1) ? opt_trg_state_written : 1'b0;

  always @ (core_halt_i or prg_may_assert_dbgreq_i or power_down_reset_i
            or trigger_event or prg_prog_bit_i or prg_prog_bit_fall_i or trg_state_we
            or trg_state_written or trigger_state or apb_wdata_3_i or niden_t0_i)
    case (trigger_state)
            `ST_CMETM_TRIG_IDLE: // idle state (trigger occurs, or written as 'done'
              begin
                nxt_trigger_state = ({2{~prg_prog_bit_i &
                                        trigger_event & niden_t0_i}} & `ST_CMETM_TRIG_HIT) |
                                    ({2{trg_state_we & apb_wdata_3_i}}
                                                                     & `ST_CMETM_TRIG_DONE);
                nxt_dbg_req = 1'b0;
              end
             `ST_CMETM_TRIG_HIT: // trigger first detected.
              begin
                nxt_trigger_state = ({2{~power_down_reset_i &
                                        (~prg_may_assert_dbgreq_i |
                                         core_halt_i |
                                         (trg_state_we & apb_wdata_3_i))}} &
                                     `ST_CMETM_TRIG_DONE) |
                                    ({2{~power_down_reset_i &
                                        prg_may_assert_dbgreq_i &
                                        ~core_halt_i &
                                        ~trg_state_we}} &
                                     `ST_CMETM_TRIG_WAIT);
                nxt_dbg_req = prg_may_assert_dbgreq_i & ~(core_halt_i |
                                                          power_down_reset_i |
                                                          prg_prog_bit_i);
              end
            `ST_CMETM_TRIG_WAIT: // wait until DbgAck observed (as prev. state, but dbgreq known)
              begin
                nxt_trigger_state = ({2{~power_down_reset_i & (core_halt_i |
                                       (trg_state_we & apb_wdata_3_i))}} &
                                     `ST_CMETM_TRIG_DONE) |
                                    ({2{~power_down_reset_i &
                                        prg_may_assert_dbgreq_i &
                                        ~core_halt_i &
                                        ~trg_state_we}} & `ST_CMETM_TRIG_WAIT);
                nxt_dbg_req = ~(core_halt_i | power_down_reset_i | prg_prog_bit_i);
              end
            `ST_CMETM_TRIG_DONE: // trigger sequence completed.
              begin
                nxt_trigger_state = {2{(~power_down_reset_i &
                                        (~prg_prog_bit_fall_i |
                                         trg_state_written) &
                                        ~(trg_state_we &
                                          ~apb_wdata_3_i))}} &
                                    `ST_CMETM_TRIG_DONE;
                nxt_dbg_req = 1'b0;
              end // case: `ST_CMETM_TRIG_DONE
      default:
        begin
          nxt_trigger_state = 2'bXX;
          nxt_dbg_req       = 1'bX;
        end

    endcase // case(trigger_state)


  assign nxt_trig_detect   = (|nxt_trigger_state);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        opt_trigger_state    <= `ST_CMETM_TRIG_IDLE;
        opt_rsrc_trig_detect <= 1'b0;
        opt_etm_dbg_req      <= 1'b0;
      end
    else if (opt_etm_en)
      begin
        opt_trigger_state    <= nxt_trigger_state;
        opt_rsrc_trig_detect <= nxt_trig_detect;
        opt_etm_dbg_req      <= nxt_dbg_req;
      end

  assign trigger_state = (TRACE_LVL > 1) ? opt_trigger_state : `ST_CMETM_TRIG_IDLE;

  assign opt_trigger_req = (opt_trigger_state == `ST_CMETM_TRIG_HIT) &
                             prg_prog_bit_and_en_i;

  assign trigger_req_o      = (TRACE_LVL > 1) ? opt_trigger_req  : 1'b0;
  assign etm_dbg_req_o      = (TRACE_LVL > 1) ? opt_etm_dbg_req  : 1'b0;
  assign rsrc_trig_detect_o = (TRACE_LVL > 1) ? opt_rsrc_trig_detect : 1'b0;

`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0,2,0,"opt_trigger_state must not be X.")
  u_ovl_etm_trigger_state(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (trigger_state)
                  );
  assert_never_unknown #(0,1,0,"etm_dbg_req must not be X.")
  u_ovl_etm_dbgrq(
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (opt_etm_dbg_req)
                  );

  reg [(8*20)-1:0] trig_state; // Debug text for header decode

  always@(posedge clk)
    case (trigger_state)
      `ST_CMETM_TRIG_IDLE:   trig_state={"IDLE"};
      `ST_CMETM_TRIG_HIT:    trig_state={"Detect"};
      `ST_CMETM_TRIG_WAIT:   trig_state={"WAIT"};
      `ST_CMETM_TRIG_DONE:   trig_state={"Completed"};
      default: trig_state={"Undefined"};
    endcase // case(TriggerState)

`endif

endmodule
