`include "cm3_etm_defs.vp"

module cm3_etm_trigger
 (
  clk, reset_n, trace_flushed_i, prg_prog_bit_fall_i,
  prg_prog_bit_and_en_i, core_halt_i, ext_in_i, prg_port_mode_i,
  prg_etm_en_i, prg_prog_bit_i, prg_may_assert_dbgreq_i,
  prg_ts_enable_i,
  prg_branch_broadcast_i, prg_stall_processor_i, prg_port_size_i, prg_etm_power_down_i,
  dwt_match_i, dwt_i_not_d_i, etm_rev_and_i, niden_t0_i,
  niden_t1_i, reason_overflow_i, max_ext_in_i, register_write_i, prg_lock_status_i,
  prg_claim_tag_i, apb_wdata_i, apb_addr_i, power_down_reset_i, etm_ivalid_i,
  etm_dvalid_i, fifo_full_support_reg_i, atb_atreadym_i,
  atb_atidm_o, etm_dbg_req_o, trigger_req_o, trace_enable_o, timestamp_event_o,
  flush_trace_o, read_data_o, integration_mode_o, integration_trigger_we_o,
  integration_atb_cr0_we_o, prg_fifo_full_level_o
  );

  parameter TRACE_LVL      = 1;
  parameter RESET_ALL_REGS = 0;
  parameter CP_PRESENT     = 0;

  input         clk;                     // Gated ETM Clock
  input         reset_n;                 // ETM Reset
  input         trace_flushed_i;         // Programming Bit and Fifo has emptied
  input         prg_prog_bit_fall_i;     // Prog Bit (Reg'd)
  input         prg_prog_bit_and_en_i;   // Prog Bit Qual'd with ETMEN
  input         core_halt_i;             // Core is in debug state
  input [1:0]   ext_in_i;                // from user-defined external resources

  input [3:0]   dwt_match_i;             // From DWT
  input [3:0]   dwt_i_not_d_i;           // From DWT
  input [3:0]   etm_rev_and_i;           // ETM ID revision
  input [2:0]   prg_port_mode_i;         // Control register port mode (not used)
  input         prg_etm_en_i;            // Control register enable bit
  input         prg_prog_bit_i;          // Control register programming bit
  input         prg_may_assert_dbgreq_i; // Control register debug control
  input         prg_ts_enable_i;         // Control register timestamp control
  input         prg_branch_broadcast_i;  // Control register branch broadcast
  input         prg_stall_processor_i;   // Control register fifo full enable
  input [3:0]   prg_port_size_i;         // Control register port size (not used)
  input         prg_etm_power_down_i;    // Control register powerdown
  input         niden_t0_i;              // Non-intrusive-debug enable
  input         niden_t1_i;              // Non-intrusive-debug enable
  input         reason_overflow_i;       // Pending Overflow
  input [1:0]   max_ext_in_i;            // Maximum supported no. of EXTIN's

  input         register_write_i;        // Register Write Signal
  input         prg_lock_status_i;       // Lock Register status
  input [3:0]   prg_claim_tag_i;         // Claim tag register value

  input [31:0]  apb_wdata_i;             // APB Write Data
  input [11:2]  apb_addr_i;              // APB Address
  input         power_down_reset_i;      // ETM has come out of powered down st
  input         etm_ivalid_i;            // PC address valid
  input         etm_dvalid_i;            // Data match valid
  input         fifo_full_support_reg_i; // Buffered FifoFullEn

  input         atb_atreadym_i;          // For Integration Mode

  output [6:0]  atb_atidm_o;             // Value of ATID
  output        etm_dbg_req_o;           // Dbgrq to put core into debug state
  output        trigger_req_o;           // Trigger signal (to common blocks)
  output        trace_enable_o;          // Trace Enable O/P
  output        timestamp_event_o;       // Timestamp request event
  output        flush_trace_o;           // Flush trace since no longer enabled.
  output [31:0] read_data_o;             // Reg Read Mux Data

  output        integration_mode_o;      // Integration Test Mode
  output        integration_trigger_we_o;// Integration Mode TRIGOUT Register Enable
  output        integration_atb_cr0_we_o;// Integration Mode ATBCTR2 Register Enable

  output [(`FIFO_DEPTH_SIZE-1):0]   prg_fifo_full_level_o; // Fifo Full level register

  wire          rsrc_start_on;           // From uTrcEn
  wire          rsrc_trig_detect;        // From uTrigGen
  wire          prg_trc_en_start_stop;   // From uResControl
  wire [10:0]   prg_trc_en_event;        // From uResControl
  wire [3:0]    prg_trc_start;           // From uResControl
  wire [3:0]    prg_trc_stop;            // From uResControl
  wire [10:0]   prg_trigger_event;       // From uResControl
  wire          stat_reg_write;          // From uResControl
  wire          stat_reg_written;        // From uResControl
  wire          count_zero;
  wire [15:0]   prg_count_reload;
  wire [10:0]   prg_ts_event;
  wire          timestamp_event;



  wire [6:0]    opt_atb_atidm;           // ATB ATID
  wire          opt_etm_dbg_req;         // Debug request to CPU
  wire          opt_trigger_req;         // Trace trigger request
  wire          opt_trace_enable;        // Trace Enable control
  wire          opt_flush_trace;         // Flush trace when progbit is set
  wire [31:0]   opt_read_data;           // APB read data to apb interface
  wire          opt_integration_mode;    // Integration test mode
  wire          opt_integration_trigger_we; // integration mode trigger write
  wire          opt_integration_atb_cr0_we; // integration mode atb output write
  wire [(`FIFO_DEPTH_SIZE-1):0] opt_prg_fifo_full_level; // Fifo full level

  assign atb_atidm_o    = (TRACE_LVL > 1) ? opt_atb_atidm         : {7{1'b0}};
  assign etm_dbg_req_o  = (TRACE_LVL > 1) ? opt_etm_dbg_req    : 1'b0;
  assign trigger_req_o  = (TRACE_LVL > 1) ? opt_trigger_req    : 1'b0;
  assign trace_enable_o = (TRACE_LVL > 1) ? opt_trace_enable   : 1'b0;
  assign timestamp_event_o = (TRACE_LVL > 1) ? timestamp_event : 1'b0;
  assign flush_trace_o  = (TRACE_LVL > 1) ? opt_flush_trace    : 1'b0;
  assign read_data_o    = (TRACE_LVL > 1) ? opt_read_data      : {32{1'b0}};
  assign integration_mode_o       = (TRACE_LVL > 1) ? opt_integration_mode       : 1'b0;
  assign integration_trigger_we_o = (TRACE_LVL > 1) ? opt_integration_trigger_we : 1'b0;
  assign integration_atb_cr0_we_o = (TRACE_LVL > 1) ? opt_integration_atb_cr0_we : 1'b0;
  assign prg_fifo_full_level_o    = (TRACE_LVL > 1) ? opt_prg_fifo_full_level    : {`FIFO_DEPTH_SIZE{1'b0}};





  cm3_etm_res_control #(TRACE_LVL, RESET_ALL_REGS,CP_PRESENT)
    u_res_control
    (
    .clk                      (clk),
    .reset_n                  (reset_n),
    .apb_addr_i               (apb_addr_i),
    .apb_wdata_i              (apb_wdata_i[31:0]),
    .register_write_i         (register_write_i),
    .prg_lock_status_i        (prg_lock_status_i),
    .prg_claim_tag_i          (prg_claim_tag_i),
    .niden_t0_i               (niden_t0_i),
    .max_ext_in_i             (max_ext_in_i),
    .prg_port_mode_i          (prg_port_mode_i),
    .prg_etm_en_i             (prg_etm_en_i),
    .prg_prog_bit_i           (prg_prog_bit_i),
    .prg_may_assert_dbgreq_i  (prg_may_assert_dbgreq_i),
    .prg_ts_enable_i          (prg_ts_enable_i),
    .prg_branch_broadcast_i   (prg_branch_broadcast_i),
    .prg_stall_processor_i    (prg_stall_processor_i),
    .prg_port_size_i          (prg_port_size_i),
    .prg_etm_power_down_i     (prg_etm_power_down_i),
    .rsrc_start_on_i          (rsrc_start_on),
    .rsrc_trig_detect_i       (rsrc_trig_detect),
    .trace_flushed_i          (trace_flushed_i),
    .prg_prog_bit_fall_i      (prg_prog_bit_fall_i),
    .reason_overflow_i        (reason_overflow_i),
    .etm_rev_and_i            (etm_rev_and_i),
    .fifo_full_support_reg_i  (fifo_full_support_reg_i),
    .core_halt_i              (core_halt_i),
    .ext_in_i                 (ext_in_i),
    .atb_atreadym_i           (atb_atreadym_i),
    .prg_trigger_event_o      (prg_trigger_event),
    .prg_trc_en_start_stop_o  (prg_trc_en_start_stop),
    .prg_trc_start_o          (prg_trc_start),
    .prg_trc_stop_o           (prg_trc_stop),
    .prg_trc_en_event_o       (prg_trc_en_event),
    .prg_count_reload_o       (prg_count_reload),
    .prg_ts_event_o           (prg_ts_event),
    .read_data_o              (opt_read_data[31:0]),
    .atb_atidm_o              (opt_atb_atidm[6:0]),
    .stat_reg_write_o         (stat_reg_write),
    .stat_reg_written_o       (stat_reg_written),
    .integration_mode_o       (opt_integration_mode),
    .integration_trigger_we_o (opt_integration_trigger_we),
    .integration_atb_cr0_we_o (opt_integration_atb_cr0_we),
    .prg_fifo_full_level_o    (opt_prg_fifo_full_level[(`FIFO_DEPTH_SIZE-1):0])
   );


  cm3_etm_trig_gen #(TRACE_LVL)
    u_trig_gen
    (
    .clk                      (clk),
    .reset_n                  (reset_n),
    .apb_addr_i               (apb_addr_i),
    .register_write_i         (register_write_i),
    .apb_wdata_3_i            (apb_wdata_i[3]),
    .niden_t0_i               (niden_t0_i),
    .prg_prog_bit_i           (prg_prog_bit_i),
    .prg_prog_bit_fall_i      (prg_prog_bit_fall_i),
    .prg_prog_bit_and_en_i    (prg_prog_bit_and_en_i),
    .power_down_reset_i       (power_down_reset_i),
    .prg_trigger_event_i      (prg_trigger_event),
    .rsrc_start_on_i          (rsrc_start_on),
    .count_zero_i             (count_zero),
    .ext_in_i                 (ext_in_i),
    .dwt_match_i              (dwt_match_i),
    .prg_may_assert_dbgreq_i  (prg_may_assert_dbgreq_i),
    .core_halt_i              (core_halt_i),
    .trigger_req_o            (opt_trigger_req),
    .etm_dbg_req_o            (opt_etm_dbg_req),
    .rsrc_trig_detect_o       (rsrc_trig_detect)
     );


  cm3_etm_trc_en #(TRACE_LVL, RESET_ALL_REGS)
    u_trc_en
    (
    .clk                      (clk),
    .reset_n                  (reset_n),
    .prg_prog_bit_i           (prg_prog_bit_i),
    .prg_prog_bit_fall_i      (prg_prog_bit_fall_i),
    .prg_count_reload_i       (prg_count_reload),
    .prg_ts_event_i           (prg_ts_event),
    .etm_ivalid_i             (etm_ivalid_i),
    .etm_dvalid_i             (etm_dvalid_i),
    .niden_t1_i               (niden_t1_i),
    .prg_trc_en_event_i       (prg_trc_en_event),
    .stat_reg_write_i         (stat_reg_write),
    .stat_reg_written_i       (stat_reg_written),
    .apb_wdata_2_i            (apb_wdata_i[2]),
    .ext_in_i                 (ext_in_i),
    .dwt_match_i              (dwt_match_i),
    .dwt_i_not_d_i            (dwt_i_not_d_i),
    .prg_trc_en_start_stop_i  (prg_trc_en_start_stop),
    .prg_trc_start_i          (prg_trc_start),
    .prg_trc_stop_i           (prg_trc_stop),
    .trace_enable_o           (opt_trace_enable),
    .flush_trace_o            (opt_flush_trace),
    .rsrc_start_on_o          (rsrc_start_on),
    .count_zero_o             (count_zero),
    .timestamp_event_o        (timestamp_event)
     );

endmodule
