`define ARM_POWERED_UP      2'b00
`define ARM_POWERING_DOWN   2'b01
`define ARM_POWERED_DOWN    2'b10
`define ARM_POWERING_UP     2'b11

module cm3_example_pmu
  (// Inputs
   FCLK, nRESET, nSYSRESETIN, SLEEPHOLDREQn, PMUENABLE, WICENACK, SLEEPING,
   SLEEPDEEP, SLEEPHOLDACKn, WAKEUP, PWRDOWNHOLD, PWRUPREADY,
   PWRDNBYPASS, TESTIN, CDBGPWRUPREQ,
   WICENREQ, SLEEPHOLDREQPMUn, ISOLATEn, RETAINn, PWRDOWN, nSYSRESETOUT,
   CDBGPWRUPACK
  );


  input         FCLK;
  input         nRESET;           // Power on reset
  input         nSYSRESETIN;      // System reset from reset controller

  input         PMUENABLE;        // Enable PMU to carry out powerdown

  input         WICENACK;         // WIC mode hand handshaking ack
  output        WICENREQ;         // WIC mode handshaking request
  input         WAKEUP;           // Wakeup request from WIC

  input         SLEEPING;         // Processor in sleep
  input         SLEEPDEEP;        // Processor in deepsleep
  input         SLEEPHOLDACKn;    // Processor sleep hold i/f ack
  input         PWRDOWNHOLD;      // Hold PMU in powerdown state
  input         SLEEPHOLDREQn;    // Auxillary sleep extension request
  input         PWRUPREADY;       // Processor is powered up,
  input         PWRDNBYPASS;      // Prevent scan test to trigger powerdown
  input   [2:0] TESTIN;           // External powerdown control for DFT
  input         CDBGPWRUPREQ;     // Debug power up request

  output        SLEEPHOLDREQPMUn; // PMU requesting sleep extension
  output        ISOLATEn;         // Isolation control for power domain
  output        RETAINn;          // State Retent control
  output        PWRDOWN;          // Powerdown control
  output        nSYSRESETOUT;     // System reset to core
  output        CDBGPWRUPACK;     // Debug power up acknowledge


  wire          wic_en_req;        // WIC mode request to WIC
  wire          wic_mode_en;       // WIC mode enabled when set
  reg     [1:0] pmu_state;         // PMU state FSM
  reg     [1:0] nxt_pmu_state;     // Next state
  reg           pmu_state_we;      // FSM update

  wire          set_sleep_hold_req;
  wire          clr_sleep_hold_req;
  reg           sleep_hold_req;
  wire          nxt_sleep_hold_req;
  wire          sleep_hold_req_we;

  reg           isolate_n_reg;
  reg           retain_n_reg;
  reg           pwrdown_reg;
  wire          pwrdown_ctrl_we;    // Register write enable

  reg           nxt_isolate_n;
  reg           nxt_retain_n;
  reg           nxt_pwrdown;

  reg           reset_req;
  wire          reset_req_we;

  wire          dbg_pwrup_req_s;
  wire          nxt_dbg_pwrup_ack;
  wire          dbg_pwrup_ack_we;
  reg           dbg_pwrup_ack;


  assign reset_req_we = (pmu_state == `ARM_POWERED_UP) & reset_req;

  always @ (posedge FCLK or negedge nSYSRESETIN)
    if (!nSYSRESETIN)
      reset_req <= 1'b1;
    else if (reset_req_we)
      reset_req <= 1'b0;


  assign wic_en_req = (PMUENABLE |
                       (pmu_state != `ARM_POWERED_UP) |
                       (~SLEEPHOLDREQn & WICENACK));

  assign wic_mode_en = wic_en_req & WICENACK & ~dbg_pwrup_req_s;


  assign set_sleep_hold_req = (wic_mode_en & SLEEPDEEP & ~WAKEUP &
                               (pmu_state==`ARM_POWERED_UP));

  assign clr_sleep_hold_req =
       ((pmu_state==`ARM_POWERED_UP) & sleep_hold_req & WAKEUP) |
       ((pmu_state==`ARM_POWERING_UP) & isolate_n_reg) |
       (SLEEPING & ~SLEEPDEEP) |
       ~wic_mode_en;

  assign nxt_sleep_hold_req = (set_sleep_hold_req | sleep_hold_req) &
                              ~clr_sleep_hold_req;

  assign sleep_hold_req_we  = set_sleep_hold_req | clr_sleep_hold_req;

  always @ (posedge FCLK or negedge nRESET)
    if (!nRESET)
      sleep_hold_req <= 1'b0;
    else if (sleep_hold_req_we)
      sleep_hold_req <= nxt_sleep_hold_req;


  always @ (wic_mode_en or SLEEPDEEP or SLEEPHOLDACKn or WAKEUP or
            PWRDOWN or pmu_state or ISOLATEn or PWRDOWNHOLD or reset_req or
            PMUENABLE or sleep_hold_req or dbg_pwrup_req_s)
    begin
      case (pmu_state)
        `ARM_POWERED_UP:
          begin
            pmu_state_we  = (wic_mode_en & SLEEPDEEP & ~SLEEPHOLDACKn &
                             sleep_hold_req & ~WAKEUP);
            nxt_pmu_state = `ARM_POWERING_DOWN;
          end

        `ARM_POWERING_DOWN:
          begin
            pmu_state_we  = ~wic_mode_en | PWRDOWN;
            nxt_pmu_state = ~wic_mode_en ? `ARM_POWERING_UP : `ARM_POWERED_DOWN;
          end

        `ARM_POWERED_DOWN:
          begin
            pmu_state_we  = (WAKEUP | ~wic_mode_en | reset_req |
                             ~PMUENABLE | dbg_pwrup_req_s) &
                            ~PWRDOWNHOLD;
            nxt_pmu_state = `ARM_POWERING_UP;
          end

        `ARM_POWERING_UP:
          begin
            pmu_state_we  = ISOLATEn;
            nxt_pmu_state = `ARM_POWERED_UP;
          end

        default:
          begin
            pmu_state_we  = 1'bX;
            nxt_pmu_state = 2'bXX;
          end
      endcase
    end

  always @ (posedge FCLK or negedge nRESET)
    if (!nRESET)
      pmu_state <= `ARM_POWERED_UP;
    else if (pmu_state_we)
      pmu_state <= nxt_pmu_state;


  always @ (pmu_state or PWRDOWN or RETAINn or ISOLATEn or PWRUPREADY)
    case (pmu_state)
      `ARM_POWERING_DOWN, `ARM_POWERED_DOWN:
        begin
          nxt_isolate_n = 1'b0;
          nxt_retain_n  = ISOLATEn;
          nxt_pwrdown   = ~RETAINn;
        end

      `ARM_POWERING_UP, `ARM_POWERED_UP:
        begin
          nxt_pwrdown   = 1'b0;
          nxt_retain_n  = ~PWRDOWN & PWRUPREADY;
          nxt_isolate_n = RETAINn;
        end

      default:
        begin
          nxt_isolate_n = 1'bX;
          nxt_retain_n  = 1'bX;
          nxt_pwrdown   = 1'bX;
        end
    endcase

  always @ (posedge FCLK or negedge nRESET)
    if (!nRESET)
      begin
        isolate_n_reg <= 1'b1;
        retain_n_reg  <= 1'b1;
        pwrdown_reg   <= 1'b0;
      end
    else if (pwrdown_ctrl_we)
      begin
        isolate_n_reg <= nxt_isolate_n;
        retain_n_reg  <= nxt_retain_n;
        pwrdown_reg   <= nxt_pwrdown;
      end

  assign pwrdown_ctrl_we = (pmu_state == `ARM_POWERING_DOWN) |
                           (pmu_state == `ARM_POWERING_UP);


  cm3_sync
    u_dbgpwrupreq_sync
      (// Inputs
       .clk                       (FCLK),
       .reset_n                   (nRESET),
       .d_async_i                 (CDBGPWRUPREQ),
       .q_o                       (dbg_pwrup_req_s)
      );


  assign nxt_dbg_pwrup_ack  = (dbg_pwrup_req_s &
                               (pmu_state_we ?
                                (pmu_state == `ARM_POWERED_UP) :
                                (pmu_state == `ARM_POWERED_UP)));

  assign dbg_pwrup_ack_we   = ((dbg_pwrup_req_s   ^ CDBGPWRUPACK) &
                               (nxt_dbg_pwrup_ack ^ CDBGPWRUPACK));

  always @ (posedge FCLK or negedge nRESET)
    if (!nRESET)
      dbg_pwrup_ack <= 1'b0;
    else if (dbg_pwrup_ack_we)
      dbg_pwrup_ack <= nxt_dbg_pwrup_ack;


  assign WICENREQ         = wic_en_req;
  assign SLEEPHOLDREQPMUn = ~sleep_hold_req;

  assign ISOLATEn         = PWRDNBYPASS ? TESTIN[0] : isolate_n_reg;
  assign RETAINn          = PWRDNBYPASS ? TESTIN[1] : retain_n_reg;
  assign PWRDOWN          = PWRDNBYPASS ? TESTIN[2] : pwrdown_reg;

  assign nSYSRESETOUT     = ~reset_req;
  assign CDBGPWRUPACK     = dbg_pwrup_ack;


`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0,2,0,"Invalid value on pmu_state.")
    ovl_pmu_state_check(FCLK, nRESET, 1'b1, pmu_state);

`endif

endmodule

