`include "cm3_fpb_defs.vp"

module cm3_fpb
  (// Inputs
   pclk, hclk, hreset_n, ppb_hready_i, mtx_fpb_active_i, mpu_canceli_i,
   mpu_canceld_i, mtx_mpu_canceli_reg_i, mtx_mpu_canceld_reg_i, ppb_hsel_i,
   ppb_haddr_i, ppb_hwdata_i, ppb_hwrite_i, dpu_ahb_haddri_i, dpu_ahb_htransi_i,
   dpu_ahb_hproti_i, mpu_mem_attri_i, dpu_ahb_hreadyi_i,
   dpu_ahb_haddracci_i, dpu_ahb_haddrd_i, dpu_ahb_htransd_i,
   dpu_ahb_hwrited_i, dpu_ahb_hprotd_i, mpu_mem_attrd_i, dpu_exreqd_i,
   dpu_ahb_hsized_i, dpu_ahb_hreadyd_i, dpu_ahb_haddraccd_i,
   ppb_hreadyout_o, ppb_hresp_o, ppb_hrdata_o, fpb_matched_o, fpb_matchedi_o,
   fpb_matched_addri_o, fpb_matched_addrd_o, fpb_matched_proti_o,
   fpb_matched_mem_attri_o, fpb_matched_protd_o, fpb_matched_mem_attrd_o,
   fpb_matched_exreqd_o, fpb_matched_sized_o, fpb_remap_addr_o,
   fpb_remap_func_o, fpb_observe_matched_o
  );


  parameter DEBUG_LVL      = 3;  // Debug support level, between 0 and 3.
  parameter RESET_ALL_REGS = 0;  // Reset all registers if 1.
  parameter OBSERVATION    = 0;  // Allow internal observability if 1.



  input         hclk;                      // AHB system clock
  input         pclk;                      // PPB clock
  input         hreset_n;                  // AHB system reset

  input  [31:0] dpu_ahb_haddri_i;
  input         dpu_ahb_htransi_i;
  input   [3:0] dpu_ahb_hproti_i;
  input         dpu_ahb_hreadyi_i;
  input         dpu_ahb_haddracci_i;

  input  [31:0] dpu_ahb_haddrd_i;
  input         dpu_ahb_htransd_i;
  input         dpu_ahb_hwrited_i;
  input   [3:0] dpu_ahb_hprotd_i;
  input   [1:0] dpu_ahb_hsized_i;
  input         dpu_exreqd_i;
  input         dpu_ahb_hreadyd_i;
  input         dpu_ahb_haddraccd_i;

  input         mtx_fpb_active_i;

  input         mpu_canceli_i;
  input         mpu_canceld_i;
  input   [1:0] mpu_mem_attri_i;
  input   [1:0] mpu_mem_attrd_i;
  input         mtx_mpu_canceli_reg_i;
  input         mtx_mpu_canceld_reg_i;

  input         ppb_hsel_i;                // FPB selected
  input  [11:2] ppb_haddr_i;               // Word-aligned address
  input         ppb_hwrite_i;
  input         ppb_hready_i;              // PPB ready feedback
  input  [31:0] ppb_hwdata_i;
  output        ppb_hresp_o;
  output [31:0] ppb_hrdata_o;
  output        ppb_hreadyout_o;           // PPB FPB ready

  output  [1:0] fpb_matched_o;             // FPB matched
  output        fpb_matchedi_o;            // Instruction matched, data phase
  output  [4:0] fpb_matched_addri_o;       // Matched instruction address
  output  [3:0] fpb_matched_proti_o;       // Matched instruction protection
  output  [1:0] fpb_matched_mem_attri_o;   // Matched instruction attributes
  output  [2:0] fpb_matched_addrd_o;       // Matched data address
  output  [1:0] fpb_matched_sized_o;       // Matched data size
  output  [3:0] fpb_matched_protd_o;       // Matched data protection
  output  [1:0] fpb_matched_mem_attrd_o;   // Matched data attributes
  output        fpb_matched_exreqd_o;      // Matched data exclusive request
  output [23:0] fpb_remap_addr_o;          // Remap base address
  output  [1:0] fpb_remap_func_o;          // Remap function

  output  [2:0] fpb_observe_matched_o;     // View of matched signals



  wire          full_hreset_n;             // Optional reset, using RESET_ALL_REGS

  wire          fp_reg_addr;               // FP register access
  wire          fp_id_addr;                // FP ID register access
  wire          fpb_access_valid;          // Register access valid
  wire          fpb_dphase_valid;          // Dataphase valid
  reg           fp_id_addr_reg;            // FP ID register access, data phase
  reg           fp_reg_addr_reg;           // FP register access, data phase
  reg     [5:2] ppb_haddr_reg;             // Registered address
  reg           ppb_hwrite_reg;            // Registered write signal
  wire          ppb_hwrite_reg_we;
  wire          nxt_ppb_hwrite_reg;
  wire          fpb_trans_reg_we;          // Write enable for registered trans
  wire    [5:0] fp_reg_sel;                // Register select
  wire          fpb_write_valid;           // Data phase write valid
  wire          ppb_hreadyout;             // FPB PPB HREADYOUT signal
  wire          ppb_hresp;                 // FPB PPB HRESP signal
  reg    [31:0] ppb_hrdata;                // FPB PPB HRDATA signal

  reg           fp_ctrl_enable;            // R/W ENABLE field
  wire          fpb_ctrl_we;               // Write enable for ENABLE field
  wire          fpb_en;                    // FPB enable
  wire   [31:0] fp_ctrl;                   // Full FP_CTRL register

  reg    [23:0] fp_remap_remap;            // FP_REMAP register REMAP field
  wire          fp_remap_we;               // FP_REMAP write enable
  wire   [23:0] opt_fp_remap;              // REMAP removal term
  wire   [31:0] fp_remap;                  // Full FP_REMAP register

  reg           fp_comp0_enable;           // FP_COMPx ENABLE fields
  reg           fp_comp1_enable;
  reg           fp_comp2_enable;
  reg           fp_comp3_enable;
  reg           fp_comp4_enable;
  reg           fp_comp5_enable;
  reg           fp_comp6_enable;
  reg           fp_comp7_enable;
  reg    [26:0] fp_comp0_comp;             // FP_COMPx COMP fields
  reg    [26:0] fp_comp1_comp;
  reg    [26:0] fp_comp2_comp;
  reg    [26:0] fp_comp3_comp;
  reg    [26:0] fp_comp4_comp;
  reg    [26:0] fp_comp5_comp;
  reg    [26:0] fp_comp6_comp;
  reg    [26:0] fp_comp7_comp;
  reg     [1:0] fp_comp0_replace;          // FP_COMPx REPLACE fields
  reg     [1:0] fp_comp1_replace;
  reg     [1:0] fp_comp2_replace;
  reg     [1:0] fp_comp3_replace;
  reg     [1:0] fp_comp4_replace;
  reg     [1:0] fp_comp5_replace;
  wire          fp_comp0_we;               // Write enable terms
  wire          fp_comp1_we;
  wire          fp_comp2_we;
  wire          fp_comp3_we;
  wire          fp_comp4_we;
  wire          fp_comp5_we;
  wire          fp_comp6_we;
  wire          fp_comp7_we;
  wire   [31:0] fp_comp0;                  // Complete FP_COMPx registers
  wire   [31:0] fp_comp1;
  wire   [31:0] fp_comp2;
  wire   [31:0] fp_comp3;
  wire   [31:0] fp_comp4;
  wire   [31:0] fp_comp5;
  wire   [31:0] fp_comp6;
  wire   [31:0] fp_comp7;

  wire          i_htransi;                 // I HTRANS gated with MPU Cancel
  wire          i_htransd;                 // D HTRANS gated with MPU Cancel
  wire   [31:2] i_haddri;                  // Masked I HADDR
  wire   [31:2] i_haddrd;                  // Masked D HADDR
  wire    [7:0] comp_en;                   // Comparator enable
  wire    [7:0] comp_match;                // FPB comparator match
  reg     [2:0] nxt_matched_addri;         // I comparator that matched
  wire          nxt_matched_addrd;         // D comparator that matched
  reg     [1:0] nxt_remap_func;            // Remap function (REMAP, hi/lo BKPT)

  wire          fpb_matched_instr;         // FPB matched on I-Side
  wire    [1:0] fpb_matched;
  reg           fpb_matchedi;
  wire          fpb_matchedi_we;
  wire          nxt_fpb_matchedi;
  reg     [4:0] fpb_matched_addri;
  reg     [3:0] fpb_matched_proti;
  reg     [1:0] fpb_matched_mem_attri;
  wire    [4:0] nxt_fpb_matched_addri;
  wire          instr_reg_we;
  reg     [2:0] fpb_matched_addrd;
  reg     [3:0] fpb_matched_protd;
  reg     [1:0] fpb_matched_mem_attrd;
  reg           fpb_matched_exreqd;
  reg     [1:0] fpb_matched_sized;
  wire    [2:0] nxt_fpb_matched_addrd;
  wire          data_reg_we;
  reg     [1:0] fpb_remap_func;
  wire   [23:0] fpb_remap_addr;

  reg     [2:0] fpb_observe_matched;       // View of matched signals
  wire    [2:0] nxt_fpb_observe_matched;

  wire    [1:0] opt_fpb_matched;
  wire    [4:0] opt_fpb_matched_addri;
  wire    [3:0] opt_fpb_matched_proti;
  wire    [1:0] opt_fpb_matched_mem_attri;
  wire    [2:0] opt_fpb_matched_addrd;
  wire    [3:0] opt_fpb_matched_protd;
  wire    [1:0] opt_fpb_matched_mem_attrd;
  wire          opt_fpb_matched_exreqd;
  wire    [1:0] opt_fpb_matched_sized;
  wire   [23:0] opt_fpb_remap_addr;
  wire          opt_fpb_matchedi;
  wire    [1:0] opt_fpb_remap_func;
  wire          opt_ppb_hreadyout;
  wire          opt_ppb_hresp;
  wire   [31:0] opt_ppb_hrdata;
  wire          opt_fpb_en;
  wire          opt_mst_en;
  wire    [2:0] opt_fpb_observe_matched;
  wire          opt_mst_observe_en;
  wire          opt_bp_available;          // Break point feature available

`ifdef ARM_DSM


  wire          dsm_reset_all_regs;
  wire [1:0]    dsm_debug_lvl;
  wire          dsm_observation;

  assign full_hreset_n             = dsm_reset_all_regs  ? hreset_n              : 1'b1;

  assign opt_fpb_matched           = (dsm_debug_lvl > 1) ? fpb_matched           : 2'b00;
  assign opt_fpb_matched_addri     = (dsm_debug_lvl > 1) ? fpb_matched_addri     : 5'b00000;
  assign opt_fpb_matched_proti     = (dsm_debug_lvl > 1) ? fpb_matched_proti     : 4'b0000;
  assign opt_fpb_matched_mem_attri = (dsm_debug_lvl > 1) ? fpb_matched_mem_attri : 2'b00;
  assign opt_fpb_matched_addrd     = (dsm_debug_lvl > 1) ? fpb_matched_addrd     : 3'b000;
  assign opt_fpb_matched_protd     = (dsm_debug_lvl > 1) ? fpb_matched_protd     : 4'b0000;
  assign opt_fpb_matched_mem_attrd = (dsm_debug_lvl > 1) ? fpb_matched_mem_attrd : 2'b00;
  assign opt_fpb_matched_exreqd    = (dsm_debug_lvl > 1) ? fpb_matched_exreqd    : 1'b0;
  assign opt_fpb_matched_sized     = (dsm_debug_lvl > 1) ? fpb_matched_sized     : 2'b00;
  assign opt_fpb_remap_addr        = (dsm_debug_lvl > 1) ? fpb_remap_addr        : {24{1'b0}};
  assign opt_fpb_en                = (dsm_debug_lvl > 1) ? 1'b1                  : 1'b0;

  assign opt_bp_available          = (dsm_debug_lvl > 0) ? 1'b1                  : 1'b0;
  assign opt_fpb_matchedi          = (dsm_debug_lvl > 0) ? fpb_matchedi          : 1'b0;
  assign opt_fpb_remap_func        = (dsm_debug_lvl > 0) ? fpb_remap_func        : 2'b00;
  assign opt_ppb_hreadyout         = (dsm_debug_lvl > 0) ? ppb_hreadyout         : 1'b1;
  assign opt_ppb_hresp             = (dsm_debug_lvl > 0) ? ppb_hresp             : 1'b0;
  assign opt_ppb_hrdata            = (dsm_debug_lvl > 0) ? ppb_hrdata            : {32{1'b0}};
  assign opt_mst_en                = (dsm_debug_lvl > 0) ? 1'b1                  : 1'b0;

  assign opt_fpb_observe_matched = (dsm_observation & (dsm_debug_lvl > 0)) ?
                                     fpb_observe_matched : 3'b000;

  assign opt_mst_observe_en      = ((OBSERVATION != 0) & (DEBUG_LVL > 0))  ? 1'b1 : 1'b0;

`else

  assign full_hreset_n = (RESET_ALL_REGS == 1) ? hreset_n : 1'b1;


  assign opt_fpb_matched           = (DEBUG_LVL > 1) ? fpb_matched           : 2'b00;
  assign opt_fpb_matched_addri     = (DEBUG_LVL > 1) ? fpb_matched_addri     : 5'b00000;
  assign opt_fpb_matched_proti     = (DEBUG_LVL > 1) ? fpb_matched_proti     : 4'b0000;
  assign opt_fpb_matched_mem_attri = (DEBUG_LVL > 1) ? fpb_matched_mem_attri : 2'b00;
  assign opt_fpb_matched_addrd     = (DEBUG_LVL > 1) ? fpb_matched_addrd     : 3'b000;
  assign opt_fpb_matched_protd     = (DEBUG_LVL > 1) ? fpb_matched_protd     : 4'b0000;
  assign opt_fpb_matched_mem_attrd = (DEBUG_LVL > 1) ? fpb_matched_mem_attrd : 2'b00;
  assign opt_fpb_matched_exreqd    = (DEBUG_LVL > 1) ? fpb_matched_exreqd    : 1'b0;
  assign opt_fpb_matched_sized     = (DEBUG_LVL > 1) ? fpb_matched_sized     : 2'b00;
  assign opt_fpb_remap_addr        = (DEBUG_LVL > 1) ? fpb_remap_addr        : {24{1'b0}};
  assign opt_fpb_en                = (DEBUG_LVL > 1) ? 1'b1                  : 1'b0;

  assign opt_fpb_matchedi          = (DEBUG_LVL > 0) ? fpb_matchedi   : 1'b0;
  assign opt_fpb_remap_func        = (DEBUG_LVL > 0) ? fpb_remap_func : 2'b00;
  assign opt_ppb_hreadyout         = (DEBUG_LVL > 0) ? ppb_hreadyout  : 1'b1;
  assign opt_ppb_hresp             = (DEBUG_LVL > 0) ? ppb_hresp      : 1'b0;
  assign opt_ppb_hrdata            = (DEBUG_LVL > 0) ? ppb_hrdata     : {32{1'b0}};
  assign opt_mst_en                = (DEBUG_LVL > 0) ? 1'b1           : 1'b0;
  assign opt_bp_available          = (DEBUG_LVL > 0) ? 1'b1           : 1'b0;


  assign opt_fpb_observe_matched = (((OBSERVATION != 0) & (DEBUG_LVL > 0)) ?
                                    fpb_observe_matched : 3'b000);

  assign opt_mst_observe_en      = (((OBSERVATION != 0) & (DEBUG_LVL > 0)) ?
                                    1'b1 : 1'b0);

`endif


  assign fp_reg_addr = ( ppb_haddr_i[11:6] == 6'b0000_00) &
                       (~ppb_haddr_i[5] |
                         (~ppb_haddr_i[4] & ~ppb_haddr_i[3]));

  assign fp_id_addr  = (ppb_haddr_i[11:6] == 6'b1111_11) &
                       (ppb_haddr_i[5:4]  != 2'b00);

  assign fpb_access_valid = ppb_hsel_i & ppb_hready_i;

  assign fpb_dphase_valid = ~mtx_fpb_active_i & ppb_hready_i;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        ppb_haddr_reg   <= 4'b0000;
        fp_reg_addr_reg <= 1'b0;
        fp_id_addr_reg  <= 1'b0;
      end
    else if (fpb_trans_reg_we)
      begin
        ppb_haddr_reg   <= ppb_haddr_i[5:2];
        fp_reg_addr_reg <= fp_reg_addr;
        fp_id_addr_reg  <= fp_id_addr;
      end

  assign fpb_trans_reg_we = fpb_access_valid & opt_mst_en;

  assign fp_reg_sel = {fp_id_addr_reg, fp_reg_addr_reg, ppb_haddr_reg};


  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      ppb_hwrite_reg <= 1'b0;
    else if (ppb_hwrite_reg_we)
      ppb_hwrite_reg <= nxt_ppb_hwrite_reg;

  assign ppb_hwrite_reg_we  = (opt_mst_en &
                               (fpb_access_valid |
                                (ppb_hready_i & ppb_hwrite_reg)));
  assign nxt_ppb_hwrite_reg = ppb_hwrite_i & ppb_hsel_i;

  assign fpb_write_valid = ppb_hwrite_reg & fpb_dphase_valid;

  assign ppb_hreadyout = ~ppb_hwrite_reg | ~mtx_fpb_active_i;
  assign ppb_hresp     = 1'b0;


  always @ (fp_ctrl or fp_remap or fp_comp0 or fp_comp1 or fp_comp2 or
            fp_comp3 or fp_comp4 or fp_comp5 or fp_comp6 or fp_comp7 or
            fp_reg_sel)
    case (fp_reg_sel)
      `FPB_CTRL_ADDR:  ppb_hrdata = fp_ctrl;      // Flash Patch control register
      `FPB_REMAP_ADDR: ppb_hrdata = fp_remap;     // Flash Patch remap register

      `FPB_COMP0_ADDR: ppb_hrdata = fp_comp0;     // Flash Patch comparator 0
      `FPB_COMP1_ADDR: ppb_hrdata = fp_comp1;     // Flash Patch comparator 1
      `FPB_COMP2_ADDR: ppb_hrdata = fp_comp2;     // Flash Patch comparator 2
      `FPB_COMP3_ADDR: ppb_hrdata = fp_comp3;     // Flash Patch comparator 3
      `FPB_COMP4_ADDR: ppb_hrdata = fp_comp4;     // Flash Patch comparator 4
      `FPB_COMP5_ADDR: ppb_hrdata = fp_comp5;     // Flash Patch comparator 5
      `FPB_COMP6_ADDR: ppb_hrdata = fp_comp6;     // Flash Patch comparator 6
      `FPB_COMP7_ADDR: ppb_hrdata = fp_comp7;     // Flash Patch comparator 7

      `FPB_PID4_ADDR:  ppb_hrdata = 32'h00000004; // PID4
      `FPB_PID5_ADDR:  ppb_hrdata = 32'h00000000; // PID5
      `FPB_PID6_ADDR:  ppb_hrdata = 32'h00000000; // PID6
      `FPB_PID7_ADDR:  ppb_hrdata = 32'h00000000; // PID7
      `FPB_PID0_ADDR:  ppb_hrdata = 32'h00000003; // PID0
      `FPB_PID1_ADDR:  ppb_hrdata = 32'h000000B0; // PID1
      `FPB_PID2_ADDR:  ppb_hrdata = 32'h0000002B; // PID2
      `FPB_PID3_ADDR:  ppb_hrdata = 32'h00000000; // PID3
      `FPB_CID0_ADDR:  ppb_hrdata = 32'h0000000D; // CID0
      `FPB_CID1_ADDR:  ppb_hrdata = 32'h000000E0; // CID1
      `FPB_CID2_ADDR:  ppb_hrdata = 32'h00000005; // CID2
      `FPB_CID3_ADDR:  ppb_hrdata = 32'h000000B1; // CID3

      default:         ppb_hrdata = 32'h00000000;
    endcase



  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_ctrl_enable <= 1'b0;
    else if (fpb_ctrl_we)
      fp_ctrl_enable <= ppb_hwdata_i[0];

  assign fpb_en = fp_ctrl_enable;

  assign fpb_ctrl_we = ((fp_reg_sel == `FPB_CTRL_ADDR) &
                        fpb_write_valid                &
                        ppb_hwdata_i[1]                & // KEY field set
                        opt_mst_en);

  assign fp_ctrl = {{17{1'b0}},                     // [31:15] Reserved
                    3'b000,                         // [14:12] NUM_CODE2
                    opt_fpb_en ? 4'b0010 : 4'b0000, //  [11:8] NUM_LIT
                    opt_fpb_en ? 4'b0110 : 4'b0010, //   [7:4] NUM_CODE1
                    2'b00,                          //   [3:2] Reserved
                    1'b0,                           //     [1] KEY
                    fpb_en};                        //     [0] ENABLE


  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      fp_remap_remap <= {24{1'b0}};
    else if (fp_remap_we)
      fp_remap_remap <= ppb_hwdata_i[28:5];

  assign fp_remap_we = ((fp_reg_sel == `FPB_REMAP_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign opt_fp_remap = fp_remap_remap & {24{opt_fpb_en}};

  assign fp_remap = {3'b001,       // [31:29] (Hardwired to SYS space)
                     opt_fp_remap, //  [28:5] REMAP
                     3'b000,       //   [4:2] (COMP[2:0])
                     2'b00};       //   [1:0] (HADDR[1:0])


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp0_enable <= 1'b0;
    else if (fp_comp0_we)
      fp_comp0_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp0_comp    <= {27{1'b0}};
        fp_comp0_replace <= 2'b00;
      end
    else if (fp_comp0_we)
      begin
        fp_comp0_comp    <= ppb_hwdata_i[28:2];
        fp_comp0_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp0_we = (fp_reg_sel == `FPB_COMP0_ADDR) &
                       fpb_write_valid                 &
                       opt_mst_en;

  assign fp_comp0 = {fp_comp0_replace[1:0], 1'b0,
                     fp_comp0_comp[26:0], 1'b0,
                     fp_comp0_enable};


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp1_enable <= 1'b0;
    else if (fp_comp1_we)
      fp_comp1_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp1_comp    <= {27{1'b0}};
        fp_comp1_replace <= 2'b00;
      end
    else if (fp_comp1_we)
      begin
        fp_comp1_comp    <= ppb_hwdata_i[28:2];
        fp_comp1_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp1_we = ((fp_reg_sel == `FPB_COMP1_ADDR) &
                        fpb_write_valid & opt_mst_en);

  assign fp_comp1    = { fp_comp1_replace[1:0], 1'b0,
                         fp_comp1_comp[26:0], 1'b0,
                         fp_comp1_enable };


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp2_enable <= 1'b0;
    else if (fp_comp2_we)
      fp_comp2_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp2_comp    <= {27{1'b0}};
        fp_comp2_replace <= 2'b00;
      end
    else if (fp_comp2_we)
      begin
        fp_comp2_comp    <= ppb_hwdata_i[28:2];
        fp_comp2_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp2_we = ((fp_reg_sel == `FPB_COMP2_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp2    = ({32{opt_fpb_en}} &
                        { fp_comp2_replace[1:0], 1'b0,
                          fp_comp2_comp[26:0], 1'b0,
                          fp_comp2_enable });


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp3_enable <= 1'b0;
    else if (fp_comp3_we)
      fp_comp3_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp3_comp    <= {27{1'b0}};
        fp_comp3_replace <= 2'b00;
      end
    else if (fp_comp3_we)
      begin
        fp_comp3_comp    <= ppb_hwdata_i[28:2];
        fp_comp3_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp3_we = ((fp_reg_sel == `FPB_COMP3_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp3    = ({32{opt_fpb_en}} &
                        { fp_comp3_replace[1:0], 1'b0,
                          fp_comp3_comp[26:0], 1'b0,
                          fp_comp3_enable });


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp4_enable <= 1'b0;
    else if (fp_comp4_we)
      fp_comp4_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp4_comp    <= {27{1'b0}};
        fp_comp4_replace <= 2'b00;
      end
    else if (fp_comp4_we)
      begin
        fp_comp4_comp    <= ppb_hwdata_i[28:2];
        fp_comp4_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp4_we = ((fp_reg_sel == `FPB_COMP4_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp4    = ({32{opt_fpb_en}} &
                        {fp_comp4_replace[1:0], 1'b0,
                         fp_comp4_comp[26:0], 1'b0,
                         fp_comp4_enable});


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp5_enable <= 1'b0;
    else if (fp_comp5_we)
      fp_comp5_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fp_comp5_comp    <= {27{1'b0}};
        fp_comp5_replace <= 2'b00;
      end
    else if (fp_comp5_we)
      begin
        fp_comp5_comp    <= ppb_hwdata_i[28:2];
        fp_comp5_replace <= ppb_hwdata_i[31:30];
      end

  assign fp_comp5_we = ((fp_reg_sel == `FPB_COMP5_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp5    = ({32{opt_fpb_en}} &
                        { fp_comp5_replace[1:0], 1'b0,
                          fp_comp5_comp[26:0], 1'b0,
                          fp_comp5_enable });


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp6_enable <= 1'b0;
    else if (fp_comp6_we)
      fp_comp6_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      fp_comp6_comp <= {27{1'b0}};
    else if (fp_comp6_we)
      fp_comp6_comp <= ppb_hwdata_i[28:2];

  assign fp_comp6_we = ((fp_reg_sel == `FPB_COMP6_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp6    = ({32{opt_fpb_en}} &
                        { 3'b000, fp_comp6_comp[26:0],
                          1'b0, fp_comp6_enable });


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      fp_comp7_enable <= 1'b0;
    else if (fp_comp7_we)
      fp_comp7_enable <= ppb_hwdata_i[0];

  always @ (posedge pclk or negedge full_hreset_n)
    if (!full_hreset_n)
      fp_comp7_comp <= {27{1'b0}};
    else if (fp_comp7_we)
      fp_comp7_comp <= ppb_hwdata_i[28:2];

  assign fp_comp7_we = ((fp_reg_sel == `FPB_COMP7_ADDR) &
                        fpb_write_valid & opt_fpb_en);

  assign fp_comp7    = ({32{opt_fpb_en}} &
                        { 3'b000, fp_comp7_comp[26:0],
                          1'b0, fp_comp7_enable });


  assign i_htransi  = dpu_ahb_htransi_i & ~mtx_mpu_canceli_reg_i;
  assign i_htransd  = dpu_ahb_htransd_i & ~mtx_mpu_canceld_reg_i;

  assign i_haddri   = {30{fpb_en}} & dpu_ahb_haddri_i[31:2];
  assign i_haddrd   = {30{fpb_en}} & dpu_ahb_haddrd_i[31:2];

  assign comp_en[0] = fpb_en & fp_comp0_enable & i_htransi;
  assign comp_en[1] = fpb_en & fp_comp1_enable & i_htransi;
  assign comp_en[2] = fpb_en & fp_comp2_enable & i_htransi;
  assign comp_en[3] = fpb_en & fp_comp3_enable & i_htransi;
  assign comp_en[4] = fpb_en & fp_comp4_enable & i_htransi;
  assign comp_en[5] = fpb_en & fp_comp5_enable & i_htransi;
  assign comp_en[6] = fpb_en & fp_comp6_enable & i_htransd & ~dpu_ahb_hwrited_i;
  assign comp_en[7] = fpb_en & fp_comp7_enable & i_htransd & ~dpu_ahb_hwrited_i;

  assign comp_match[0] = comp_en[0] & ({3'b000, fp_comp0_comp} == i_haddri);
  assign comp_match[1] = comp_en[1] & ({3'b000, fp_comp1_comp} == i_haddri);
  assign comp_match[2] = comp_en[2] & ({3'b000, fp_comp2_comp} == i_haddri);
  assign comp_match[3] = comp_en[3] & ({3'b000, fp_comp3_comp} == i_haddri);
  assign comp_match[4] = comp_en[4] & ({3'b000, fp_comp4_comp} == i_haddri);
  assign comp_match[5] = comp_en[5] & ({3'b000, fp_comp5_comp} == i_haddri);
  assign comp_match[6] = comp_en[6] & ({3'b000, fp_comp6_comp} == i_haddrd);
  assign comp_match[7] = comp_en[7] & ({3'b000, fp_comp7_comp} == i_haddrd);


  always @ (comp_match or fp_comp0_replace or fp_comp1_replace or
            fp_comp2_replace or fp_comp3_replace or fp_comp4_replace or
            fp_comp5_replace)
    if (comp_match[0])
      begin
        nxt_matched_addri = 3'b000;
        nxt_remap_func    = fp_comp0_replace;
      end
    else if (comp_match[1])
      begin
        nxt_matched_addri = 3'b001;
        nxt_remap_func    = fp_comp1_replace;
      end
    else if (comp_match[2])
      begin
        nxt_matched_addri = 3'b010;
        nxt_remap_func    = fp_comp2_replace;
      end
    else if (comp_match[3])
      begin
        nxt_matched_addri = 3'b011;
        nxt_remap_func    = fp_comp3_replace;
      end
    else if (comp_match[4])
      begin
        nxt_matched_addri = 3'b100;
        nxt_remap_func    = fp_comp4_replace;
      end
    else if (comp_match[5])
      begin
        nxt_matched_addri = 3'b101;
        nxt_remap_func    = fp_comp5_replace;
      end
    else
      begin
        nxt_matched_addri = 3'b000;
        nxt_remap_func    = 2'b00;
      end

  assign nxt_matched_addrd = ~comp_match[6];


  assign fpb_matched_instr = (|(comp_match[5:0]) & dpu_ahb_haddracci_i &
                              ~mpu_canceli_i);

  assign fpb_matched[0]    = (fpb_matched_instr & (nxt_remap_func == 2'b00) &
                              dpu_ahb_hreadyi_i);
  assign fpb_matched[1]    = ((|comp_match[7:6]) & dpu_ahb_hreadyd_i &
                              dpu_ahb_haddraccd_i & ~mpu_canceld_i);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      fpb_matchedi <= 1'b0;
    else if (fpb_matchedi_we)
      fpb_matchedi <= nxt_fpb_matchedi;

  assign fpb_matchedi_we  = dpu_ahb_hreadyi_i & opt_mst_en;
  assign nxt_fpb_matchedi = fpb_matched_instr & (nxt_remap_func != 2'b00);

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fpb_matched_addri     <= 5'b00000;
        fpb_matched_proti     <= 4'b0000;
        fpb_matched_mem_attri <= 2'b00;
      end
    else if (instr_reg_we)
      begin
        fpb_matched_addri     <= nxt_fpb_matched_addri;
        fpb_matched_proti     <= dpu_ahb_hproti_i[3:0];
        fpb_matched_mem_attri <= mpu_mem_attri_i[1:0];
      end

  assign nxt_fpb_matched_addri = {nxt_matched_addri, dpu_ahb_haddri_i[1:0]};

  assign instr_reg_we = (dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
                         fpb_matched_instr & opt_mst_en);

  always @ (posedge hclk or negedge full_hreset_n)
    if (!full_hreset_n)
      begin
        fpb_matched_addrd     <= 3'b000;
        fpb_matched_protd     <= 4'b0000;
        fpb_matched_mem_attrd <= 2'b00;
        fpb_matched_exreqd    <= 1'b0;
        fpb_matched_sized     <= 2'b00;
      end
    else if (data_reg_we)
      begin
        fpb_matched_addrd     <= nxt_fpb_matched_addrd;
        fpb_matched_protd     <= dpu_ahb_hprotd_i[3:0];
        fpb_matched_mem_attrd <= mpu_mem_attrd_i[1:0];
        fpb_matched_exreqd    <= dpu_exreqd_i;
        fpb_matched_sized     <= dpu_ahb_hsized_i[1:0];
      end

  assign nxt_fpb_matched_addrd = {nxt_matched_addrd, dpu_ahb_haddrd_i[1:0]};

  assign data_reg_we = (dpu_ahb_hreadyd_i & dpu_ahb_haddraccd_i  &
                        fpb_matched[1] & opt_mst_en);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      fpb_remap_func <= 2'b00;
    else if (instr_reg_we)
      fpb_remap_func <= nxt_remap_func;

  assign fpb_remap_addr[23:0] = opt_fp_remap[23:0];


  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      fpb_observe_matched <= 3'b000;
    else if (opt_mst_observe_en)
      fpb_observe_matched <= nxt_fpb_observe_matched;

  assign nxt_fpb_observe_matched = {opt_fpb_matchedi, opt_fpb_matched};



  assign fpb_matched_o           = opt_fpb_matched;
  assign fpb_matched_addri_o     = opt_fpb_matched_addri;
  assign fpb_matched_proti_o     = opt_fpb_matched_proti;
  assign fpb_matched_mem_attri_o = opt_fpb_matched_mem_attri;
  assign fpb_matched_addrd_o     = opt_fpb_matched_addrd;
  assign fpb_matched_protd_o     = opt_fpb_matched_protd;
  assign fpb_matched_mem_attrd_o = opt_fpb_matched_mem_attrd;
  assign fpb_matched_exreqd_o    = opt_fpb_matched_exreqd;
  assign fpb_matched_sized_o     = opt_fpb_matched_sized;
  assign fpb_remap_addr_o        = opt_fpb_remap_addr;
  assign fpb_matchedi_o          = opt_fpb_matchedi;
  assign fpb_remap_func_o        = opt_fpb_remap_func;

  assign ppb_hreadyout_o         = opt_ppb_hreadyout;
  assign ppb_hresp_o             = opt_ppb_hresp;
  assign ppb_hrdata_o            = opt_ppb_hrdata;

  assign fpb_observe_matched_o   = opt_fpb_observe_matched;


`ifdef ARM_ASSERT_ON


  assert_zero_one_hot #(0,2,0,"fp_id_addr_reg and fp_reg_addr_reg must be exclusive")
    ovl_addr_reg_exclusive(hclk, hreset_n, {fp_id_addr_reg, fp_reg_addr_reg});

  assert_zero_one_hot #(0,2,0,"fp_id_addr and fp_reg_addr must be exclusive")
    ovl_fp_addr_exclusive(hclk, hreset_n, {fp_id_addr, fp_reg_addr});

  assert_never_unknown #(0,6,0,"fp_reg_sel goes to x.")
    ovl_fp_reg_sel_x (hclk, hreset_n, 1'b1, fp_reg_sel);

  assert_never #(0,0,"fpb_matched[0] must not be asserted if mpu_canceli_i is asserted")
    ovl_mpu_canceli(hclk, hreset_n, (fpb_matched[0] && mpu_canceli_i));

  assert_never #(0,0,"fpb_matched[1] must not be asserted if mpu_canceld_i is asserted")
    ovl_mpu_canceld(hclk, hreset_n, (fpb_matched[1] && mpu_canceld_i));

  assert_always #(0,0,"fpb_matched[0] must only be asserted if i_htransi is asserted")
    ovl_match_htransi(hclk, hreset_n,
      (fpb_matched[0] ? (i_htransi & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i) : 1'b1));

  assert_always #(0,0,"fpb_matched[1] must only be asserted if i_htransd is asserted")
    ovl_match_htransd(hclk, hreset_n,
      (fpb_matched[1] ? (i_htransd & dpu_ahb_hreadyd_i & dpu_ahb_haddraccd_i) : 1'b1));

  assert_always #(0,0,"fpb_matched must only be asserted if FPB enabled")
    ovl_matched_enable(hclk, hreset_n,
      ((fpb_matched[0] || fpb_matched[1]) ? fp_ctrl_enable : 1'b1));

  assert_always #(0,0,
    "fpb_matched[0] must only be asserted if at least one instruction comparator enabled")
    ovl_fpb_instr_match_enable(hclk, hreset_n,
      (fpb_matched[0] ?
        (fp_comp0_enable | fp_comp1_enable | fp_comp2_enable | fp_comp3_enable |
         fp_comp4_enable | fp_comp5_enable) : 1'b1));

  assert_always #(0,0,
    "fpb_matched[1] must only be asserted if at least one literal comparator enabled")
    ovl_fpb_data_match_enable(hclk, hreset_n,
      (fpb_matched[1] ? (fp_comp6_enable | fp_comp7_enable) : 1'b1));

  assert_always #(0,0,
    "fpb_matched[0] must only be asserted if at least one instruction comparator matched")
    ovl_fpb_instr_match_check(hclk, hreset_n,
      (fpb_matched[0] ? ((|comp_match[5:0]) & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i) : 1'b1));

  assert_always #(0,0,
    "fpb_matched[1] must only be asserted if at least one literal comparator matched")
    ovl_fpb_data_match_check(hclk, hreset_n,
      (fpb_matched[1] ? ((|comp_match[7:6]) & dpu_ahb_hreadyd_i & dpu_ahb_haddraccd_i) : 1'b1));

  assert_always #(0,0,"fpb_matched[0] must only be asserted for I access to ICode space")
    ovl_fpb_instr_icode(hclk, hreset_n,
      (fpb_matched[0] ? (dpu_ahb_haddri_i[31:29] == 3'b000) : 1'b1));

  assert_always #(0,0,"fpb_matched[1] must only be asserted for D access to DCode space")
    ovl_fpb_data_dcode(hclk, hreset_n,
      (fpb_matched[1] ? (dpu_ahb_haddrd_i[31:29] == 3'b000) : 1'b1));

  assert_next #(0,1,1,0,0,"comp_match[7] must map to fpb_matched_addrd[2] == b1")
    ovl_match7_map(hclk, hreset_n,
      (comp_match[7] & ~comp_match[6] & dpu_ahb_hreadyd_i & dpu_ahb_haddraccd_i &
       ~mpu_canceld_i),
        (fpb_matched_addrd[2] == 1'b1));

  assert_next #(0,1,1,0,0,"comp_match[6] must map to fpb_matched_addrd[2] == b0")
    ovl_match6_map(hclk, hreset_n,
      (comp_match[6] & dpu_ahb_hreadyd_i & dpu_ahb_haddraccd_i & ~mpu_canceld_i),
        (fpb_matched_addrd[2] == 1'b0));

  assert_next #(0,1,1,0,0,"comp_match[5] must map to fpb_matched_addri[4:2] == b101")
    ovl_match5_map(hclk, hreset_n,
      (comp_match[5] & ~(|comp_match[4:0]) & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
        ~mpu_canceli_i),
          (fpb_matched_addri[4:2] == 3'b101));

  assert_next #(0,1,1,0,0,"comp_match[4] must map to fpb_matched_addri[4:2] == b100")
    ovl_match4_map(hclk, hreset_n,
      (comp_match[4] & ~(|comp_match[3:0]) & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
        ~mpu_canceli_i),
          (fpb_matched_addri[4:2] == 3'b100));

  assert_next #(0,1,1,0,0,"comp_match[3] must map to fpb_matched_addri[4:2] == b011")
    ovl_match3_map(hclk, hreset_n,
      (comp_match[3] & ~(|comp_match[2:0]) & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
        ~mpu_canceli_i),
          (fpb_matched_addri[4:2] == 3'b011));

  assert_next #(0,1,1,0,0,"comp_match[2] must map to fpb_matched_addri[4:2] == b010")
    ovl_match2_map(hclk, hreset_n,
      (comp_match[2] & ~(|comp_match[1:0]) & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
        ~mpu_canceli_i),
          (fpb_matched_addri[4:2] == 3'b010));

  assert_next #(0,1,1,0,0,"comp_match[1] must map to fpb_matched_addri[4:2] == b001")
    ovl_match1_map(hclk, hreset_n,
      (comp_match[1] & ~comp_match[0] & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i &
        ~mpu_canceli_i),
          (fpb_matched_addri[4:2] == 3'b001));

  assert_next #(0,1,1,0,0,"comp_match[0] must map to fpb_matched_addri[4:2] == b000")
    ovl_match0_map(hclk, hreset_n,
      (comp_match[0] & dpu_ahb_hreadyi_i & dpu_ahb_haddracci_i & ~mpu_canceli_i),
        (fpb_matched_addri[4:2] == 3'b000));


  assert_always #(0,0,"FP_CTRL[31:12] must always read as zero")
    ovl_fp_ctrl_31_12(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CTRL_ADDR) ? (ppb_hrdata[31:12] == {20{1'b0}}) : 1'b1));

  assert_always #(0,0,"FP_CTRL[11:8] must always read as 4'b0010")
    ovl_fp_ctrl_11_8(hclk, hreset_n,
      (((fp_reg_sel == `FPB_CTRL_ADDR) & opt_fpb_en) ? (ppb_hrdata[11:8] == 4'b0010) : 1'b1));

  assert_always #(0,0,"FPB_CTRL[7:4] must always read as 4'b0110")
    ovl_fp_ctrl_7_4(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CTRL_ADDR) & opt_fpb_en) ? (ppb_hrdata[7:4] == 4'b0110) : 1'b1);

  assert_always #(0,0,"FPB_CTRL[3:2] must always read as 2'b00")
    ovl_fp_ctrl_3_2(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CTRL_ADDR) ? (ppb_hrdata[3:2] == 2'b00) : 1'b1));

  assert_always #(0,0,"FPB_CTRL can only be written when KEY is set")
    ovl_fp_ctrl_key(hclk, hreset_n,
      (fpb_ctrl_we ? ppb_hwdata_i[1] : 1'b1));


  reg [31:0] ovl_ppb_hwdata_reg;
  reg        ovl_remap_addr_reg;

  always @ (posedge hclk)
    begin
      ovl_remap_addr_reg <= fpb_remap_addr;
      ovl_ppb_hwdata_reg <= ppb_hwdata_i;
    end

  assert_next #(0,1,1,0,0,"PPB HWDATA maps to correct FP_CTRL bits")
    ovl_fp_ctrl_map(hclk, hreset_n,
    ((fp_reg_sel == `FPB_CTRL_ADDR) & ppb_hwrite_reg & ppb_hwdata_i[1] & fpb_dphase_valid),
         (fp_ctrl_enable == ovl_ppb_hwdata_reg[0]));

  assert_always #(0,0,"FP_REMAP[31:29] must always read as 3'b001")
    ovl_fp_remap_31_29(hclk, hreset_n,
      (((fp_reg_sel == `FPB_REMAP_ADDR) & opt_fpb_en) ?
        (ppb_hrdata[31:29] == 3'b001) : 1'b1));

  assert_always #(0,0,"FP_REMAP[4:0] must always read as 5'b00000")
    ovl_fp_remap_4_0(hclk, hreset_n,
      ((fp_reg_sel == `FPB_REMAP_ADDR) ? (ppb_hrdata[4:0] == 5'b00000) : 1'b1));

  assert_implication #(0,0,"fpb_remap_addr should be the same as FP_REMAP[28:5]")
    ovl_fp_remap_output(hclk, hreset_n, (fpb_remap_addr[23:0] != ovl_remap_addr_reg),
      (opt_fp_remap[23:0] == fpb_remap_addr[23:0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_REMAP bits")
    ovl_fp_remap_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_REMAP_ADDR) & ppb_hwrite_reg & fpb_dphase_valid &
        opt_fpb_en), (opt_fp_remap == ovl_ppb_hwdata_reg[28:5]));


  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP0 (comp) bits")
    ovl_fp_comp0_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP0_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_bp_available),
          (fp_comp0_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP0 (enable) bits")
    ovl_fp_comp0_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP0_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_bp_available),
          (fp_comp0_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP1 (comp) bits")
    ovl_fp_comp1_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP1_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_bp_available),
          (fp_comp1_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP1 (enable) bits")
    ovl_fp_comp1_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP1_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_bp_available),
          (fp_comp1_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP2 (comp) bits")
    ovl_fp_comp2_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP2_ADDR) & ppb_hwrite_reg  & fpb_dphase_valid & opt_fpb_en),
          (fp_comp2_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP2 (enable) bits")
    ovl_fp_comp2_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP2_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp2_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP3 (comp) bits")
    ovl_fp_comp3_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP3_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp3_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP3 (enable) bits")
    ovl_fp_comp3_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP3_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp3_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP4 (comp) bits")
    ovl_fp_comp4_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP4_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp4_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP4 (enable) bits")
    ovl_fp_comp4_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP4_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp4_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP5 (comp) bits")
    ovl_fp_comp5_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP5_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp5_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP5 (enable) bits")
    ovl_fp_comp5_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP5_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp5_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP6 (comp) bits")
    ovl_fp_comp6_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP6_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp6_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP6 (enable) bits")
    ovl_fp_comp6_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP6_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp6_enable == ovl_ppb_hwdata_reg[0]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP7 (comp) bits")
    ovl_fp_comp7_comp_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP7_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp7_comp == ovl_ppb_hwdata_reg[28:2]));

  assert_next #(0,1,1,0,0,"ppb_hwdata_i maps to correct FP_COMP7 (enable) bits")
    ovl_fp_comp7_enable_map(hclk, hreset_n,
      ((fp_reg_sel == `FPB_COMP7_ADDR) & ppb_hwrite_reg & fpb_dphase_valid & opt_fpb_en),
          (fp_comp7_enable == ovl_ppb_hwdata_reg[0]));


  assert_always #(0,0,"FP_PID4 must always read as 32'h00000004")
    ovl_fp_pid4_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID4_ADDR) ? (ppb_hrdata == 32'h00000004) : 1'b1));

  assert_always #(0,0,"FP_PID5 must always read as 32'h00000000")
    ovl_fp_pid5_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID5_ADDR) ? (ppb_hrdata == 32'h00000000) : 1'b1));

  assert_always #(0,0,"FP_PID6 must always read as 32'h00000000")
    ovl_fp_pid6_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID6_ADDR) ? (ppb_hrdata == 32'h00000000) : 1'b1));

  assert_always #(0,0,"FP_PID7 must always read as 32'h00000000")
    ovl_fp_pid7_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID7_ADDR) ? (ppb_hrdata == 32'h00000000) : 1'b1));

  assert_always #(0,0,"FP_PID0 must always read as 32'h00000003")
    ovl_fp_pid0_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID0_ADDR) ? (ppb_hrdata == 32'h00000003) : 1'b1));

  assert_always #(0,0,"FP_PID1 must always read as 32'h000000B0")
    ovl_fp_pid1_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID1_ADDR) ? (ppb_hrdata == 32'h000000B0) : 1'b1));

  assert_always #(0,0,"FP_PID2 must always read as 32'h0000002B")
    ovl_fp_pid2_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID2_ADDR) ? (ppb_hrdata == 32'h0000002B) : 1'b1));

  assert_always #(0,0,"FP_PID3 must always read as 32'h00000000")
    ovl_fp_pid3_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_PID3_ADDR) ? (ppb_hrdata == 32'h00000000) : 1'b1));

  assert_always #(0,0,"FP_CID0 must always read as 32'h0000000D")
    ovl_fp_cid0_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CID0_ADDR) ? (ppb_hrdata == 32'h0000000D) : 1'b1));

  assert_always #(0,0,"FP_CID1 must always read as 32'h000000E0")
    ovl_fp_cid1_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CID1_ADDR) ? (ppb_hrdata == 32'h000000E0) : 1'b1));

  assert_always #(0,0,"FP_CID2 must always read as 32'h00000005")
    ovl_fp_cid2_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CID2_ADDR) ? (ppb_hrdata == 32'h00000005) : 1'b1));

  assert_always #(0,0,"FP_CID3 must always read as 32'h000000B1")
    ovl_fp_cid3_read(hclk, hreset_n,
      ((fp_reg_sel == `FPB_CID3_ADDR) ? (ppb_hrdata == 32'h000000B1) : 1'b1));


  assert_never #(0,0,"FPB comparators 2 to 7 cannot match if DEBUG_LVL == 1")
    ovl_cfg_comp2to7(hclk, hreset_n, ((DEBUG_LVL == 1) & (|comp_match[7:2])));

  assert_never #(0,0,"FPB comparators 0 to 7 cannot match if DEBUG_LVL == 0")
    ovl_cfg_comp0to7(hclk, hreset_n, ((DEBUG_LVL == 0) & (|comp_match[7:0])));

  assert_never #(0,0,"FPB flash patch cannot occur if DEBUG_LVL < 2")
    ovl_cfg_fpatch(hclk, hreset_n, ((DEBUG_LVL < 2) & (|opt_fp_remap[23:0])));

`endif

endmodule

`include "cm3_fpb_undefs.vp"