module cm3_htm_port
  (// Inputs
   hclk, hreset_n, dpu_ahb_hreadyd_i, dpu_ahb_haddraccd_i, dpu_ahb_hrespd_i,
   dpu_ahb_hrdatad_i, mpu_ahb_hrespd_i, dpu_ahb_htransd_i, dpu_ahb_hwdatad_i,
   dpu_ahb_haddrd_i, dpu_ahb_hwrited_i, dpu_ahb_hsized_i, dpu_ahb_hprotd_i,
   trace_enable_i,
   htm_haddrd_o, htm_htransd_o, htm_hsized_o, htm_hburstd_o, htm_hprotd_o,
   htm_hwdatad_o, htm_hwrited_o, htm_hrdatad_o, htm_hreadyd_o, htm_hrespd_o
  );


  parameter TRACE_LVL = 1;



  input         hclk;               // System clock
  input         hreset_n;           // System reset

  input         dpu_ahb_hreadyd_i;
  input         dpu_ahb_haddraccd_i;
  input   [1:0] dpu_ahb_hrespd_i;
  input  [31:0] dpu_ahb_hrdatad_i;
  input         mpu_ahb_hrespd_i;
  input   [1:0] dpu_ahb_htransd_i;
  input  [31:0] dpu_ahb_hwdatad_i;
  input  [31:0] dpu_ahb_haddrd_i;
  input         dpu_ahb_hwrited_i;
  input   [1:0] dpu_ahb_hsized_i;
  input   [1:0] dpu_ahb_hprotd_i;

  input         trace_enable_i;

  output [31:0] htm_haddrd_o;       // HTM data HADDR
  output  [1:0] htm_htransd_o;      // HTM data HTRANS
  output  [2:0] htm_hsized_o;       // HTM data HSIZE
  output  [2:0] htm_hburstd_o;      // HTM data HBURST
  output  [3:0] htm_hprotd_o;       // HTM data HPROT
  output [31:0] htm_hwdatad_o;      // HTM data HWDATA
  output        htm_hwrited_o;      // HTM data HWRITE
  output [31:0] htm_hrdatad_o;      // HTM data HRDATA
  output        htm_hreadyd_o;      // HTM data HREADY
  output  [1:0] htm_hrespd_o;       // HTM data HRESP



  wire          htm_reg_we;         // Register write enable
  wire          htm_resp_reg_we;    // Write enable for HREADY/HRESP registers
  reg    [31:0] htm_haddrd;         // HADDR reg
  reg     [1:0] htm_htransd;        // HTRANS reg
  reg     [2:0] htm_hsized;         // HSIZE reg
  wire    [2:0] htm_hburstd;        // HBURST reg
  reg     [3:0] htm_hprotd;         // HPROT reg
  reg           htm_hwrited;        // HWRITE reg
  reg    [31:0] htm_hdatad;         // Shared HRDATA/HWDATA reg
  reg           htm_hreadyd;        // HREADY reg
  reg     [1:0] htm_hrespd;         // HRESP reg
  wire    [2:0] nxt_htm_hsized;     // Next value for HSIZE reg
  wire    [3:0] nxt_htm_hprotd;     // Next value for HPROT reg
  wire    [1:0] nxt_htm_htransd;    // Next value for HTRANS reg
  wire    [1:0] nxt_htm_hrespd;     // Next value for HRESP reg, including MPU
  wire   [31:0] rw_data_mux;        // MUX to merge read data and write data

  reg           data_rot8;          // Unaligned data rotate by 8: data phase
  reg           data_rot16;         // Unaligned data rotate by 16: data phase
  wire          nxt_data_rot8;      // Next rotate by 8
  wire          nxt_data_rot16;     // Next rotate by 16
  reg    [31:0] htm_hdata_rot;      // Data rotate mux

  wire   [31:0] opt_htm_haddrd;
  wire    [1:0] opt_htm_htransd;
  wire    [2:0] opt_htm_hsized;
  wire    [2:0] opt_htm_hburstd;
  wire    [3:0] opt_htm_hprotd;
  wire   [31:0] opt_htm_hwdatad;
  wire          opt_htm_hwrited;
  wire   [31:0] opt_htm_hrdatad;
  wire          opt_htm_hreadyd;
  wire    [1:0] opt_htm_hrespd;
  wire          opt_mst_en;

  wire [1:0]    opt_trace_lvl;

`ifdef ARM_DSM


  wire [1:0]    dsm_trace_lvl;

  assign opt_trace_lvl   = dsm_trace_lvl;

`else

  assign opt_trace_lvl   = TRACE_LVL;

`endif


  assign opt_htm_haddrd  = (opt_trace_lvl > 2) ? htm_haddrd    : {32{1'b0}};
  assign opt_htm_htransd = (opt_trace_lvl > 2) ? htm_htransd   : 2'b00;
  assign opt_htm_hsized  = (opt_trace_lvl > 2) ? htm_hsized    : 3'b000;
  assign opt_htm_hburstd = (opt_trace_lvl > 2) ? htm_hburstd   : 3'b000;
  assign opt_htm_hprotd  = (opt_trace_lvl > 2) ? htm_hprotd    : 4'b0000;
  assign opt_htm_hwrited = (opt_trace_lvl > 2) ? htm_hwrited   : 1'b0;
  assign opt_htm_hreadyd = (opt_trace_lvl > 2) ? htm_hreadyd   : 1'b0;
  assign opt_htm_hrespd  = (opt_trace_lvl > 2) ? htm_hrespd    : 2'b00;
  assign opt_htm_hwdatad = (opt_trace_lvl > 2) ? htm_hdata_rot : {32{1'b0}};
  assign opt_htm_hrdatad = (opt_trace_lvl > 2) ? htm_hdata_rot : {32{1'b0}};
  assign opt_mst_en      = (opt_trace_lvl > 2) ? 1'b1          : 1'b0;



  assign htm_reg_we = dpu_ahb_hreadyd_i & trace_enable_i & opt_mst_en;



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        htm_haddrd  <= {32{1'b0}};
        htm_htransd <= 2'b00;
        htm_hsized  <= 3'b000;
        htm_hprotd  <= 4'b0000;
        htm_hwrited <= 1'b0;
        htm_hdatad  <= {32{1'b0}};
      end
    else if(htm_reg_we)
      begin
        htm_haddrd  <= dpu_ahb_haddrd_i;
        htm_htransd <= nxt_htm_htransd;
        htm_hsized  <= nxt_htm_hsized;
        htm_hprotd  <= nxt_htm_hprotd;
        htm_hwrited <= dpu_ahb_hwrited_i;
        htm_hdatad  <= rw_data_mux;
      end

  assign nxt_htm_hsized = {1'b0, dpu_ahb_hsized_i};
  assign nxt_htm_hprotd = {2'b00,dpu_ahb_hprotd_i};

  assign nxt_htm_htransd = dpu_ahb_htransd_i & {2{dpu_ahb_haddraccd_i}};

  assign htm_hburstd = {2'b00, trace_enable_i & opt_mst_en};

  assign rw_data_mux = htm_hwrited ? dpu_ahb_hwdatad_i : dpu_ahb_hrdatad_i;



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        htm_hreadyd <= 1'b0;
        htm_hrespd  <= 2'b00;
      end
    else if (htm_resp_reg_we)
      begin
        htm_hreadyd <= dpu_ahb_hreadyd_i;
        htm_hrespd  <= nxt_htm_hrespd;
      end

  assign htm_resp_reg_we = trace_enable_i & opt_mst_en;

  assign nxt_htm_hrespd = {1'b0, mpu_ahb_hrespd_i} | dpu_ahb_hrespd_i;



  assign nxt_data_rot8  = htm_haddrd[0] & (htm_hsized[1:0]!=2'b00) & htm_htransd[1];
  assign nxt_data_rot16 = htm_haddrd[1] & (htm_hsized[1:0]==2'b10) & htm_htransd[1];

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        data_rot8  <= 1'b0;
        data_rot16 <= 1'b0;
      end
    else if (htm_reg_we)
      begin
        data_rot8  <= nxt_data_rot8;
        data_rot16 <= nxt_data_rot16;
      end

  always @ (data_rot8 or data_rot16 or htm_hdatad)
    case ({data_rot16, data_rot8})
      2'b00:   htm_hdata_rot = htm_hdatad[31:0];                       // None
      2'b01:   htm_hdata_rot = {htm_hdatad[7:0],  htm_hdatad[31:8]};  // Rot-8
      2'b10:   htm_hdata_rot = {htm_hdatad[15:0], htm_hdatad[31:16]}; // Rot-16
      2'b11:   htm_hdata_rot = {htm_hdatad[23:0], htm_hdatad[31:24]};  // Rot-24
      default: htm_hdata_rot = 32'hXXXXXXXX;
    endcase



  assign htm_haddrd_o  = opt_htm_haddrd;
  assign htm_htransd_o = opt_htm_htransd;
  assign htm_hsized_o  = opt_htm_hsized;
  assign htm_hburstd_o = opt_htm_hburstd;
  assign htm_hprotd_o  = opt_htm_hprotd;
  assign htm_hwdatad_o = opt_htm_hwdatad;
  assign htm_hwrited_o = opt_htm_hwrited;
  assign htm_hrdatad_o = opt_htm_hrdatad;
  assign htm_hreadyd_o = opt_htm_hreadyd;
  assign htm_hrespd_o  = opt_htm_hrespd;

endmodule
