module cm3_itm
  (// Inputs
   reset_n, clk, dpu_trace_enable_i, ppb_hsel_itm_i, ppb_hprot_i, ppb_htrans_i,
   ppb_hwrite_i, ppb_hsize_i, ppb_haddr_i, ppb_hready_i, ppb_hwdata_i,
   dwt_dist_sync_i, dwt_valid_i, dwt_overflow_i, dwt_size_i, dwt_data_i,
   atb_ready_i, tpiu_active_i, tpiu_baud_i, dpu_halted_i, ts_value_i,
   ts_clk_change_i,
   ppb_hreadyout_o, ppb_hresp_o, ppb_hrdata_o, atb_id_o, itm_dwt_ready_o,
   atb_valid_o, atb_data_o
  );

  parameter     TRACE_LVL      = 1;
  parameter     RESET_ALL_REGS = 0;
  parameter     EXT_PRESENT    = 0;  // Affects ID registers



  input         reset_n;           // AHB reset
  input         clk;               // AHB clock
  input         dpu_trace_enable_i;// Enable Tracing
  input         ppb_hsel_itm_i;    // AHB select
  input   [3:0] ppb_hprot_i;       // AHB protection
  input   [1:0] ppb_htrans_i;      // AHB enable
  input         ppb_hwrite_i;      // AHB write
  input   [1:0] ppb_hsize_i;       // AHB size
  input  [11:0] ppb_haddr_i;       // AHB address
  input         ppb_hready_i;      // AHB ready
  input  [31:0] ppb_hwdata_i;      // AHB write data
  input         tpiu_active_i;     // TPIU active
  input         tpiu_baud_i;       // TPIU baud
  input         dpu_halted_i;      // Core is halted
  input [47:0]  ts_value_i;        // Binary coded timestamp value
  input         ts_clk_change_i;   // Clock ratio changed

  output        ppb_hreadyout_o;   // AHB ready
  output        ppb_hresp_o;       // AHB error response
  output [31:0] ppb_hrdata_o;      // AHB read data

  input         dwt_dist_sync_i;   // Debug Synchronisation

  output [6:0]  atb_id_o;          // ITM ID

  input         dwt_valid_i;       // ATB Valid - DWT
  input         dwt_overflow_i;    // ATB OverFlow - DWT
  input   [2:0] dwt_size_i;        // ATB Size - DWT
  input  [39:0] dwt_data_i;        // ATB Data - DWT
  output        itm_dwt_ready_o;   // ATB Ready - DWT

  input         atb_ready_i;       // ATB Ready - ITM

  output        atb_valid_o;       // ATB Valid - ITM
  output  [7:0] atb_data_o;        // ATB Data - ITM



  wire          arb_valid;         // ATB Valid - Arb
  wire          arb_valid_ts;      // ATB Valid for timestamp- Arb
  wire    [2:0] arb_size;          // ATB Size - Arb
  wire          arb_overflow;      // ATB OverFlow - Arb
  wire   [39:0] arb_data;          // ATB Data - Arb
  wire          fifo_ready;        // ATB Ready - Arb

  wire          ts_valid;          // ATB Valid - TS
  wire          ts_overflow;       // ATB OverFlow - TS
  wire    [2:0] ts_size;           // ATB Size - TS
  wire   [39:0] ts_data;           // ATB Data - TS
  wire          arb_ready_ts;      // ATB Ready - TS
  wire          arb_ts_invalid;    // Global timestamp not possible

  wire          if_valid;          // ATB valid - IF
  wire          if_overflow;       // ATB OverFlow - IF
  wire    [2:0] if_size;           // ATB size - IF
  wire   [39:0] if_data;           // ATB data - IF
  wire          arb_ready_if;      // ATB ready - IF

  wire          dwt_valid_enabled;
  wire          dwt_enable;        // Enable DWT

  wire          sync_enable;       // Sync Enable

  wire          timestamp_enable;  // TimeStamp enable
  wire          timestamp_pulse;   // TimeStamp RefClock
  wire    [1:0] timestamp_prescale;// Timestamp prescaler

  wire          emit_valid;
  wire          fifo_valid;
  wire          emit_ready;
  wire    [8:0] fifo_data;

  wire          fifo_empty;        // FIFO empty

  wire          atb_valid;         // ATB Valid - ITM

  wire          itm_no_tx;         // ITM not transmitting

  wire          ahb_ready_out;
  wire          ahb_resp;
  wire [31:0]   ahb_rdata;
  wire [6:0]    atb_id;
  wire          itm_dwt_ready;
  wire [7:0]    atb_data;
  wire [1:0]    prg_ts_freq;
  wire          prg_ts_sync;


  cm3_itm_if #(TRACE_LVL, RESET_ALL_REGS, EXT_PRESENT)
    u_itm_if
      (// Inputs
       .reset_n              (reset_n),
       .clk                  (clk),
       .dpu_trace_enable_i   (dpu_trace_enable_i),
       .ahb_sel_i            (ppb_hsel_itm_i),
       .ahb_prot_i           (ppb_hprot_i),
       .ahb_trans_i          (ppb_htrans_i),
       .ahb_write_i          (ppb_hwrite_i),
       .ahb_size_i           (ppb_hsize_i),
       .ahb_addr_i           (ppb_haddr_i),
       .ahb_ready_in_i       (ppb_hready_i),
       .ahb_wdata_i          (ppb_hwdata_i),
       .ready_if_i           (arb_ready_if),
       .atb_ready_i          (atb_ready_i),
       .fifo_empty_i         (fifo_empty),
       .itm_no_tx_i          (itm_no_tx),
       .tpiu_active_i        (tpiu_active_i),
       .tpiu_baud_i          (tpiu_baud_i),
       .emit_valid_i         (emit_valid),
       .ahb_ready_out_o      (ahb_ready_out),
       .ahb_resp_o           (ahb_resp),
       .ahb_rdata_o          (ahb_rdata),
       .if_valid_o           (if_valid),
       .if_overflow_o        (if_overflow),
       .if_size_o            (if_size),
       .if_data_o            (if_data),
       .atb_id_o             (atb_id),
       .dwt_enable_o         (dwt_enable),
       .sync_enable_o        (sync_enable),
       .timestamp_enable_o   (timestamp_enable),
       .timestamp_pulse_o    (timestamp_pulse),
       .timestamp_prescale_o (timestamp_prescale),
       .prg_ts_freq_o        (prg_ts_freq),
       .prg_ts_sync_o        (prg_ts_sync),
       .atb_valid_o          (atb_valid)
       );

  cm3_itm_ts #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_ts
      (// Inputs
       .reset_n              (reset_n),
       .clk                  (clk),
       .timestamp_enable_i   (timestamp_enable),
       .core_halt_i          (dpu_halted_i),
       .timestamp_pulse_i    (timestamp_pulse),
       .timestamp_prescale_i (timestamp_prescale),
       .ready_ts_i           (arb_ready_ts),
       .if_valid_i           (if_valid),
       .ready_if_i           (arb_ready_if),
       .dwt_enable_i         (dwt_enable),
       .dwt_valid_i          (dwt_valid_i),
       .ready_dwt_i          (itm_dwt_ready),
       .dwt_valid_enabled_o  (dwt_valid_enabled),
       .ts_valid_o           (ts_valid),
       .ts_overflow_o        (ts_overflow),
       .ts_size_o            (ts_size),
       .ts_data_o            (ts_data)
       );

  cm3_itm_arb #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_arb
      (// Inputs
       .reset_n              (reset_n),
       .clk                  (clk),
       .fifo_ready_i         (fifo_ready),
       .sync_enable_i        (sync_enable),
       .dist_sync_i          (dwt_dist_sync_i),
       .valid_1_i            (if_valid),
       .overflow_1_i         (if_overflow),
       .size_1_i             (if_size),
       .data_1_i             (if_data),
       .valid_2_i            (dwt_valid_enabled),
       .overflow_2_i         (dwt_overflow_i),
       .size_2_i             (dwt_size_i),
       .data_2_i             (dwt_data_i),
       .valid_3_i            (ts_valid),
       .overflow_3_i         (ts_overflow),
       .size_3_i             (ts_size),
       .data_3_i             (ts_data),
       .arb_valid_o          (arb_valid),
       .arb_valid_ts_o       (arb_valid_ts),
       .arb_size_o           (arb_size),
       .arb_overflow_o       (arb_overflow),
       .arb_data_o           (arb_data),
       .arb_ready_1_o        (arb_ready_if),
       .arb_ready_2_o        (itm_dwt_ready),
       .arb_ready_3_o        (arb_ready_ts),
       .arb_ts_invalid_o     (arb_ts_invalid)

       );

  cm3_itm_fifo #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo
      (// Inputs
       .reset_n              (reset_n),
       .clk                  (clk),
       .emit_ready_i         (emit_ready),
       .arb_valid_i          (arb_valid),
       .arb_size_i           (arb_size),
       .arb_overflow_i       (arb_overflow),
       .arb_data_i           (arb_data),
       .fifo_valid_o         (fifo_valid),
       .fifo_data_o          (fifo_data),
       .fifo_ready_o         (fifo_ready),
       .fifo_empty_o         (fifo_empty)
       );

  cm3_itm_emit #(TRACE_LVL)
    u_itm_emit
      (// Inputs
       .reset_n              (reset_n),
       .clk                  (clk),
       .atb_ready_i          (atb_ready_i),
       .fifo_valid_i         (fifo_valid),
       .fifo_data_i          (fifo_data),
       .ts_value_i           (ts_value_i),
       .ts_clk_change_i      (ts_clk_change_i),
       .prg_ts_freq_i        (prg_ts_freq),
       .prg_ts_sync_i        (prg_ts_sync),
       .dwt_dist_sync_i      (dwt_dist_sync_i),
       .arb_ts_invalid_i     (arb_ts_invalid),
       .arb_valid_ts_i       (arb_valid_ts),
       .emit_valid_o         (emit_valid),
       .emit_ready_o         (emit_ready),
       .atb_data_o           (atb_data),
       .itm_no_tx_o          (itm_no_tx)
       );

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;

  assign ppb_hreadyout_o  = (dsm_trace_lvl > 0) ? ahb_ready_out  : 1'b0;
  assign ppb_hresp_o      = (dsm_trace_lvl > 0) ? ahb_resp       : 1'b0;
  assign ppb_hrdata_o     = (dsm_trace_lvl > 0) ? ahb_rdata      : {32{1'b0}};
  assign atb_id_o         = (dsm_trace_lvl > 0) ? atb_id         : {7{1'b0}};
  assign itm_dwt_ready_o  = (dsm_trace_lvl > 0) ? itm_dwt_ready  : 1'b0;
  assign atb_valid_o      = (dsm_trace_lvl > 0) ? atb_valid      : 1'b0;
  assign atb_data_o       = (dsm_trace_lvl > 0) ? atb_data       : {8{1'b0}};

`else

  assign ppb_hreadyout_o  = (TRACE_LVL > 0) ? ahb_ready_out      : 1'b0;
  assign ppb_hresp_o      = (TRACE_LVL > 0) ? ahb_resp           : 1'b0;
  assign ppb_hrdata_o     = (TRACE_LVL > 0) ? ahb_rdata          : {32{1'b0}};
  assign atb_id_o         = (TRACE_LVL > 0) ? atb_id             : {7{1'b0}};
  assign itm_dwt_ready_o  = (TRACE_LVL > 0) ? itm_dwt_ready      : 1'b0;
  assign atb_valid_o      = (TRACE_LVL > 0) ? atb_valid          : 1'b0;
  assign atb_data_o       = (TRACE_LVL > 0) ? atb_data           : {8{1'b0}};
`endif

`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"IT header and size do not match")
    ITHeaderSize_check(clk, reset_n, if_valid &
        ((if_size==3'b001 & if_data[1:0]!=2'b01 & if_data[1:0]!=2'b00) |
         (if_size==3'b010 & if_data[1:0]!=2'b10 & if_data[1:0]!=2'b00) |
         (if_size==3'b100 & if_data[1:0]!=2'b11 & if_data[1:0]!=2'b00) |
         (if_size==3'b000 & if_data[1:0]==2'b00 & (if_data[7])) |
         (if_size==3'b001 & if_data[1:0]==2'b00 &
          (if_data[15] | ~if_data[7])) |
         (if_size==3'b010 & if_data[1:0]==2'b00 &
          (if_data[23] | ~if_data[15] | ~if_data[7])) |
         (if_size==3'b011 & if_data[1:0]==2'b00 &
          (if_data[31] | ~if_data[23] | ~if_data[15] | ~if_data[7])) |
         (if_size==3'b100 & if_data[1:0]==2'b00 &
          (if_data[39] | ~if_data[31] | ~if_data[23] | ~if_data[15] |
           ~if_data[7]))));

  assert_never #(0,0,"IT header is invalid")
    ITHeader_check(clk, reset_n, if_valid &
        ~((~if_data[2] & if_data[1:0]!=2'b00) | (if_data[3:0]==4'b1000)));

  assert_never #(0,0,"IT size invalid")
    ITSize_check(clk, reset_n, if_valid & if_size > 4);

  assert_never #(0,0,"TS header and size do not match")
    TSHeaderSize_check(clk, reset_n, ts_valid &
        ((ts_size==3'b000 & (ts_data[7])) |
         (ts_size==3'b001 & (ts_data[15] | ~ts_data[7])) |
         (ts_size==3'b010 & (ts_data[23] | ~ts_data[15] | ~ts_data[7])) |
         (ts_size==3'b011 & (ts_data[31] | ~ts_data[23] | ~ts_data[15] |
          ~ts_data[7]))));

  assert_never #(0,0,"TS size invalid")
    TSSize_check(clk, reset_n, ts_valid & ts_size > 4);

  assert_never #(0,0,"TS header is invalid")
    TSHeader_check1(clk, reset_n, ts_valid & ~(ts_data[3:0]==4'h0));

  assert_never #(0,0,"TS header is invalid")
    TSHeader_check2(clk, reset_n, ts_valid & (ts_data[7:0]==8'h80));

  assert_never #(0,0,"TimeStamp small has been incorrectly formed")
    TSmall_check(clk, reset_n, ts_valid & (ts_data[7:0]==8'h70));

  assert_never #(0,0,"Sync header in first byte is invalid")
    SyncHeader_check(clk, reset_n,
                     arb_valid & ~arb_overflow & (arb_data[7:0]==8'h80));

  assert_never #(0,0,"TimeStamp output has been incorrectly formed")
    TimeStamp_check(clk, reset_n, arb_valid & arb_overflow & arb_size!=3'b000 &
                                   (arb_data[7:0]==8'h80));

  assert_next #(0,1,1,0,0,"dwt_dist_sync_i must be pulsed")
    dwt_dist_sync_i_pulsed(clk, reset_n, dwt_dist_sync_i, ~dwt_dist_sync_i);

  assert_never #(0,0,"DWT header and size do not match")
    DWTHeaderSize_check(clk, reset_n, dwt_valid_i &
        ((dwt_size_i==3'b001 & dwt_data_i[1:0]!=2'b01 & dwt_data_i[1:0]!=2'b00) |
         (dwt_size_i==3'b010 & dwt_data_i[1:0]!=2'b10 & dwt_data_i[1:0]!=2'b00) |
         (dwt_size_i==3'b100 & dwt_data_i[1:0]!=2'b11 & dwt_data_i[1:0]!=2'b00) |
         (dwt_size_i==3'b000 & dwt_data_i[1:0]==2'b00 & (dwt_data_i[7])) |
         (dwt_size_i==3'b001 & dwt_data_i[1:0]==2'b00 &
          (dwt_data_i[15] | ~dwt_data_i[7])) |
         (dwt_size_i==3'b010 & dwt_data_i[1:0]==2'b00 &
          (dwt_data_i[23] | ~dwt_data_i[15] | ~dwt_data_i[7])) |
         (dwt_size_i==3'b011 & dwt_data_i[1:0]==2'b00 &
          (dwt_data_i[31] | ~dwt_data_i[23] | ~dwt_data_i[15] | ~dwt_data_i[7])) |
         (dwt_size_i==3'b100 & dwt_data_i[1:0]==2'b00 &
          (dwt_data_i[39] | ~dwt_data_i[31] | ~dwt_data_i[23] |
            ~dwt_data_i[15] | ~dwt_data_i[7]))));

  assert_never #(0,0,"DWT header is invalid")
    DWTHeader_check(clk, reset_n, dwt_valid_i &
        ~((dwt_data_i[2] & dwt_data_i[1:0]!=2'b00) | (dwt_data_i[3:0]==4'b1100)));

  assert_never #(0,0,"DWT size invalid")
    DWTSize_check(clk, reset_n, dwt_valid_i & dwt_size_i > 4);


  assert_never #(0,0,"ITM trace cannot be generated if TRACE_LVL == 0")
    cfg_itmtrace(clk, reset_n, (TRACE_LVL == 0) & atb_valid_o);

`endif

endmodule
