module cm3_itm_arb
  (// Inputs
   reset_n, clk, fifo_ready_i, sync_enable_i, dist_sync_i, valid_1_i,
   overflow_1_i, size_1_i, data_1_i, valid_2_i, overflow_2_i, size_2_i,
   data_2_i, valid_3_i, overflow_3_i, size_3_i, data_3_i,
   arb_valid_o, arb_valid_ts_o, arb_size_o, arb_overflow_o, arb_data_o,
   arb_ready_1_o, arb_ready_2_o, arb_ready_3_o, arb_ts_invalid_o
  );

  parameter     TRACE_LVL      = 1;
  parameter     RESET_ALL_REGS = 0;
  input         reset_n;          // AHB reset
  input         clk;              // AHB clock
  input         fifo_ready_i;     // ATB Ready Master
  input         sync_enable_i;    // Sync Enable
  input         dist_sync_i;      // Distributed Synchronisation
  input         valid_1_i;        // Input valid from stimulus port
  input         overflow_1_i;     // ATB OverFlow Slave 1
  input   [2:0] size_1_i;         // ATB Size Slave 1
  input  [39:0] data_1_i;         // ATB Data Slave 1
  input         valid_2_i;        // Input valid from DWT
  input         overflow_2_i;     // ATB OverFlow Slave 2
  input   [2:0] size_2_i;         // ATB Size Slave 2
  input  [39:0] data_2_i;         // ATB Data Slave 2
  input         valid_3_i;        // Input valid from local timestamp
  input         overflow_3_i;     // ATB OverFlow Slave 3
  input   [2:0] size_3_i;         // ATB Size Slave 3
  input  [39:0] data_3_i;         // ATB Data Slave 3

  output        arb_valid_o;      // ATB Valid Master
  output        arb_valid_ts_o;   // ATB Valid for timestamp
  output  [2:0] arb_size_o;       // ATB Size Master
  output        arb_overflow_o;   // ATB OverFlow Master
  output [39:0] arb_data_o;       // ATB Data Master
  output        arb_ready_1_o;    // ATB Ready Slave 1
  output        arb_ready_2_o;    // ATB Ready Slave 2
  output        arb_ready_3_o;    // ATB Ready Slave 3
  output        arb_ts_invalid_o; // Invalidate timestamp value

  reg           valid_0;
  wire          nxt_valid_0;
  wire          valid_0_we;
  wire    [3:0] arb_sel;          // Slave Select
  wire          arb_overflow;     // Slave Overflow
  wire    [2:0] arb_size;         // Slave Size
  wire   [39:0] arb_data;
  wire          arb_ready_0;
  wire          arb_ready_1;
  wire          arb_ready_2;
  wire          arb_ready_3;
  wire          opt_en;
  wire          full_reset_n;
  wire          arb_valid;
  wire          arb_valid_ts;
  reg           arb_stall;
  wire          nxt_arb_stall;
  wire          arb_stall_we;

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  wire          dsm_reset_all_regs;

  assign full_reset_n   = (dsm_reset_all_regs == 1) ? reset_n : 1'b1;

  assign arb_valid_o    = (dsm_trace_lvl > 0) ? arb_valid    : 1'b0;
  assign arb_valid_ts_o = (dsm_trace_lvl > 0) ? arb_valid_ts : 1'b0;
  assign arb_size_o     = (dsm_trace_lvl > 0) ? arb_size     : {3{1'b0}};
  assign arb_overflow_o = (dsm_trace_lvl > 0) ? arb_overflow : 1'b0;
  assign arb_data_o     = (dsm_trace_lvl > 0) ? arb_data     : {40{1'b0}};
  assign arb_ready_1_o  = (dsm_trace_lvl > 0) ? arb_ready_1  : 1'b0;
  assign arb_ready_2_o  = (dsm_trace_lvl > 0) ? arb_ready_2  : 1'b0;
  assign arb_ready_3_o  = (dsm_trace_lvl > 0) ? arb_ready_3  : 1'b0;
  assign opt_en         = (dsm_trace_lvl > 0) ? 1'b1         : 1'b0;
`else
  assign full_reset_n   = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

  assign arb_valid_o    = (TRACE_LVL > 0) ? arb_valid    : 1'b0;
  assign arb_valid_ts_o = (TRACE_LVL > 0) ? arb_valid_ts : 1'b0;
  assign arb_size_o     = (TRACE_LVL > 0) ? arb_size     : {3{1'b0}};
  assign arb_overflow_o = (TRACE_LVL > 0) ? arb_overflow : 1'b0;
  assign arb_data_o     = (TRACE_LVL > 0) ? arb_data     : {40{1'b0}};
  assign arb_ready_1_o  = (TRACE_LVL > 0) ? arb_ready_1  : 1'b0;
  assign arb_ready_2_o  = (TRACE_LVL > 0) ? arb_ready_2  : 1'b0;
  assign arb_ready_3_o  = (TRACE_LVL > 0) ? arb_ready_3  : 1'b0;
  assign opt_en         = (TRACE_LVL > 0) ? 1'b1         : 1'b0;
`endif

  assign nxt_valid_0 = sync_enable_i & dist_sync_i;
  assign valid_0_we  = opt_en &((nxt_valid_0) | (arb_ready_0));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      valid_0 <= 1'b0;
    else if (valid_0_we)
      valid_0 <= nxt_valid_0;

  assign arb_sel = { (valid_3_i & ~valid_2_i & ~valid_1_i & ~valid_0),
                     (valid_2_i & ~valid_1_i & ~valid_0),
                     (valid_1_i & ~valid_0),
                     (valid_0) };

  assign arb_valid = valid_3_i | valid_2_i | valid_1_i | valid_0;

  assign arb_valid_ts = valid_2_i | valid_1_i;

  assign arb_overflow = ((arb_sel[0]) |
                         (arb_sel[1] & overflow_1_i) |
                         (arb_sel[2] & overflow_2_i) |
                         (arb_sel[3] & overflow_3_i));

  assign arb_size = (//  arb_sel[0] is always 1 byte (==0)
                     ({3{arb_sel[1]}} & size_1_i) |
                     ({3{arb_sel[2]}} & size_2_i) |
                     ({3{arb_sel[3]}} & size_3_i));

  assign arb_data = (// Input select
                     ({40{arb_sel[0]}} & 40'h0000000080) |
                     ({40{arb_sel[1]}} & data_1_i[39:0]) |
                     ({40{arb_sel[2]}} & data_2_i[39:0]) |
                     ({40{arb_sel[3]}} & data_3_i[39:0]));

  assign arb_ready_0 = fifo_ready_i & arb_sel[0];
  assign arb_ready_1 = fifo_ready_i & arb_sel[1];
  assign arb_ready_2 = fifo_ready_i & arb_sel[2];
  assign arb_ready_3 = fifo_ready_i & arb_sel[3];

  assign nxt_arb_stall = ((valid_1_i & ~arb_ready_1) |
                          (valid_2_i & ~arb_ready_2));

  assign arb_stall_we = valid_1_i | valid_2_i | fifo_ready_i;

  always @(posedge clk or negedge full_reset_n)
    if (!full_reset_n)
      arb_stall <= 1'b1;
    else if(arb_stall_we)
      arb_stall <= nxt_arb_stall;

  assign arb_ts_invalid_o = arb_stall | valid_0;


`ifdef ARM_ASSERT_ON

`ifdef ARM_MONITOR_ON

  monitor_itmarb
    ITMArb_monitor(clk, reset_n, fifo_ready_i,
                   valid_0,
                   valid_1_i, overflow_1_i, size_1_i, data_1_i,
                   valid_2_i, overflow_2_i, size_2_i, data_2_i,
                   valid_3_i, overflow_3_i, size_3_i, data_3_i,
                   arb_valid_o, arb_overflow, arb_size, arb_data,
                   arb_ready_0, arb_ready_1, arb_ready_2, arb_ready_3);

`endif

  assert_never #(0,0,"ATSIZE should not be more than 4 when active")
    atsizem_check(clk, reset_n, arb_valid_o & (arb_size > 4));

  assert_never #(0,0,"ATSIZE should not be more than 4 when active")
    atsizes0_check(clk, reset_n, (valid_0 & (arb_size > 4)));

  assert_never #(0,0,"ATSIZE should not be more than 4 when active")
    atsizes1_check(clk, reset_n, (valid_1_i & (size_1_i > 4)));

  assert_never #(0,0,"ATSIZE should not be more than 4 when active")
    atsizes2_check(clk, reset_n, (valid_2_i & (size_2_i > 4)));

  assert_never #(0,0,"ATSIZE should not be more than 4 when active")
    atsizes3_check(clk, reset_n, (valid_3_i & (size_3_i > 4)));

  assert_zero_one_hot #(0,4,0,"arb_sel must be exclusive")
    arb_sel_exclusive(clk, reset_n, arb_sel);

  assert_never #(0,0,"Slave is selected when ATVALID is not")
    slv0_check(clk, reset_n, arb_sel[0] & ~valid_0);

  assert_never #(0,0,"Slave is selected when ATVALID is not")
    slv1_check(clk, reset_n, arb_sel[1] & ~valid_1_i);

  assert_never #(0,0,"Slave is selected when ATVALID is not")
    slv2_check(clk, reset_n, arb_sel[2] & ~valid_2_i);

  assert_never #(0,0,"Slave is selected when ATVALID is not")
    slv3_check(clk, reset_n, arb_sel[3] & ~valid_3_i);

  assert_never #(0,0,"arb_ready_0 without valid_0 not possible")
    arb_ready_0_check(clk, reset_n, arb_ready_0 & ~valid_0);

  assert_never #(0,0,"arb_ready_1 without valid_1_i not possible")
    arb_ready_1_check(clk, reset_n, arb_ready_1 & ~valid_1_i);

  assert_never #(0,0,"arb_ready_2 without valid_2_i not possible")
    arb_ready_2_check(clk, reset_n, arb_ready_2 & ~valid_2_i);

  assert_never #(0,0,"arb_ready_3 without valid_3_i not possible")
    arb_ready_3_check(clk, reset_n, arb_ready_3 & ~valid_3_i);

  assert_never #(0,0,"arb_overflow_o without Slave OverFlow not possible")
    arb_overflow_o_check(clk, reset_n,
                         (arb_overflow & ~(valid_0 | overflow_1_i |
                                           overflow_2_i | overflow_3_i)));

  assert_never_unknown #(0, 1, 0, "valid_0_we must not be X")
  ovl_valid_0_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (valid_0_we));

  assert_never_unknown #(0, 1, 0, "arb_stall_we must not be X")
  ovl_arb_stall_we_x (
    .clk       (clk),
    .reset_n   (reset_n),
    .qualifier (1'b1),
    .test_expr (arb_stall_we));

`endif

endmodule
