module cm3_itm_emit
  (// Inputs
   reset_n, clk, atb_ready_i, fifo_valid_i, fifo_data_i, ts_value_i,
   ts_clk_change_i, prg_ts_freq_i, prg_ts_sync_i, dwt_dist_sync_i,
   arb_ts_invalid_i, arb_valid_ts_i,
   emit_valid_o, emit_ready_o, atb_data_o, itm_no_tx_o
  );

  parameter     TRACE_LVL = 1;

  input         reset_n;          // System Reset
  input         clk;              // System Clock
  input         atb_ready_i;      // ATB Ready Master
  input         fifo_valid_i;     // ATB Valid Slave
  input [8:0]   fifo_data_i;      // ATB Data Slave
  input [47:0]  ts_value_i;       // Binary coded timestamp value
  input         ts_clk_change_i;  // Clock ratio changed
  input [1:0]   prg_ts_freq_i;    // Timestamp control
  input         prg_ts_sync_i;    // Full sync after turning on
  input         dwt_dist_sync_i;  // Debug Synchronisation
  input         arb_ts_invalid_i; // Fifo has stalled a packet
  input         arb_valid_ts_i;   // Fifo has taken a packet

  output        emit_valid_o;     // ATB Valid Master
  output        emit_ready_o;     // ATB Ready Slave
  output  [7:0] atb_data_o;       // ATB Data Master
  output        itm_no_tx_o;      // ITM not transmitting

`define EM_NORMAL   4'b0000
`define EM_OVRFLW   4'b0001
`define EM_SYNC0    4'b0010
`define EM_SYNC1    4'b0011
`define EM_SYNC2    4'b0100
`define EM_SYNC3    4'b0101
`define EM_SYNC4    4'b0110
`define EM_SYNC5    4'b0111
`define EM_TS_1     4'b1000
`define EM_TS_2     4'b1001
`define EM_TS_3     4'b1010
`define EM_TS_4     4'b1011
`define EM_TS_5     4'b1100
`define EM_SPARE1   4'b1101
`define EM_SPARE2   4'b1110
`define EM_SPARE3   4'b1111

  wire    [1:0] control;
  reg     [3:0] emit_state;
  reg     [3:0] nxt_emit_state;
  wire          emit_state_we;
  reg           emit_valid;
  wire          emit_we;
  wire          nxt_emit_valid;
  reg     [7:0] fifo_data_reg;
  reg           ts1_pend;
  wire          nxt_ts1_pend;
  reg           ts2_pend;
  wire          nxt_ts2_pend;
  wire          nxt_ts_clk_pend;
  reg           ts_clk_pend;
  reg           ts_clk_ch;
  reg           ts_wrap;
  wire          nxt_ts_wrap;
  wire          ts_capture;
  wire          ts_request;
  wire          ts_packet_state_we;
  reg           arb_valid_reg;
  reg    [47:0] ts_value;
  reg     [2:0] ts_continue;
  wire    [2:0] nxt_ts_continue;
  reg           ts1_valid;
  wire          nxt_ts1_valid;
  wire          ts_start;
  wire          ts_active;
  wire         ts_changed_46_26;
  wire         ts_changed_25_21;
  wire         ts_changed_20_14;
  wire         ts_changed_12_7;
  wire          emit_ready;
  wire    [7:0] atb_data;
  wire          itm_no_tx;
  wire          opt_en;

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  assign emit_valid_o = (dsm_trace_lvl > 0) ? emit_valid : 1'b0;
  assign emit_ready_o = (dsm_trace_lvl > 0) ? emit_ready : 1'b0;
  assign atb_data_o   = (dsm_trace_lvl > 0) ? atb_data   : {8{1'b0}};
  assign itm_no_tx_o  = (dsm_trace_lvl > 0) ? itm_no_tx  : 1'b0;
  assign opt_en       = (dsm_trace_lvl > 0) ? 1'b1       : 1'b0;
`else
  assign emit_valid_o = (TRACE_LVL > 0)     ? emit_valid : 1'b0;
  assign emit_ready_o = (TRACE_LVL > 0)     ? emit_ready : 1'b0;
  assign atb_data_o   = (TRACE_LVL > 0)     ? atb_data   : {8{1'b0}};
  assign itm_no_tx_o  = (TRACE_LVL > 0)     ? itm_no_tx  : 1'b0;
  assign opt_en       = (TRACE_LVL > 0)     ? 1'b1       : 1'b0;
`endif


  assign ts_changed_46_26 = |(ts_value[47:26] ^ ts_value_i[47:26]);
  assign ts_changed_25_21 = |(ts_value[25:21] ^ ts_value_i[25:21]);
  assign ts_changed_20_14 = |(ts_value[20:14] ^ ts_value_i[20:14]);
  assign ts_changed_12_7  = |(ts_value[12: 7] ^ ts_value_i[12: 7]);


  assign ts_request = (((|prg_ts_freq_i) &
                        (ts_clk_pend |
                         dwt_dist_sync_i |
                         (arb_valid_ts_i & (
                         ts_changed_46_26 |
                         ts_changed_25_21 |
                         ts_changed_20_14 |
                         (ts_value[13] ^ ts_value_i[13]))))) |
                       (prg_ts_freq_i[0] & ts_changed_12_7 & arb_valid_ts_i) |
                       (&prg_ts_freq_i & arb_valid_ts_i));

  assign nxt_ts1_pend = ((ts1_pend & ~(emit_state == `EM_TS_2)) |
                         prg_ts_sync_i | ts_request);

  assign nxt_ts2_pend = ((ts2_pend &
                          ~((emit_state == `EM_TS_1) & ~ts1_pend)) |
                         prg_ts_sync_i |
                         ((|prg_ts_freq_i) &
                          (dwt_dist_sync_i |
                           (arb_valid_ts_i & (
                                              ts_changed_46_26 |
                                              ts_changed_25_21)))));

  assign ts_capture = ts1_pend & arb_valid_ts_i & ~ts_active;

  assign nxt_ts1_valid = (ts1_valid | (ts1_pend & arb_valid_ts_i)) &
                           ~(arb_ts_invalid_i | (emit_state == `EM_TS_5));

  assign nxt_ts_wrap = ((ts_wrap &
                         ~((emit_state == `EM_TS_5) & ~ts1_valid)) |
                        ts_changed_46_26);

  assign nxt_ts_continue[2] = ((ts_continue[2] &
                                ~(emit_state == `EM_TS_5)) |
                               ts_clk_pend |
                               nxt_ts2_pend |
                               ts_changed_25_21);
  assign nxt_ts_continue[1] = ((ts_continue[1] &
                                ~(emit_state == `EM_TS_5)) |
                               ts_clk_pend |
                               nxt_ts2_pend |
                               ts_changed_20_14 |
                               ts_changed_25_21);
  assign nxt_ts_continue[0] = ((ts_continue[0] &
                                ~(emit_state == `EM_TS_5)) |
                               ts_clk_pend |
                               nxt_ts2_pend |
                               ts_changed_12_7 |
                               (ts_value[13] ^ ts_value_i[13])|
                               ts_changed_20_14 |
                               ts_changed_25_21);

  assign ts_packet_state_we = (ts_capture |
                               ((emit_state == `EM_TS_5) & emit_state_we));

  assign nxt_ts_clk_pend = ((ts_clk_pend &
                             ~((emit_state == `EM_TS_5) &
                               ts_clk_ch & ts1_valid)) |
                            ts_clk_change_i);

  assign ts_start = ((ts1_valid | (~ts1_pend & ts2_pend)) &
                     ~arb_valid_ts_i & ~arb_valid_reg & ~fifo_valid_i &
                     (atb_ready_i | ~emit_valid) &
                     (emit_state == `EM_NORMAL));

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      begin
        ts1_pend      <= 1'b0;
        ts2_pend      <= 1'b0;
        ts_clk_pend   <= 1'b0;
        arb_valid_reg <= 1'b0;
      end
    else
      begin
        ts1_pend      <= nxt_ts1_pend;
        ts2_pend      <= nxt_ts2_pend;
        ts_clk_pend   <= nxt_ts_clk_pend;
        arb_valid_reg <= arb_valid_ts_i;
      end

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      begin
        ts_value      <= {48{1'b0}};
        ts_clk_ch     <= 1'b0;
      end
    else if (ts_capture)
      begin
        ts_value      <= ts_value_i;
        ts_clk_ch     <= ts_clk_pend;
      end

  always @(posedge clk or negedge reset_n)
    if(!reset_n)
      begin
        ts_continue   <= {3{1'b0}};
        ts_wrap       <= 1'b0;
        ts1_valid     <= 1'b0; // ASSERT no change whilst TS being output
      end
    else if (ts_packet_state_we)
      begin
        ts_continue    <= nxt_ts_continue;
        ts_wrap        <= nxt_ts_wrap;
        ts1_valid      <= nxt_ts1_valid;
      end

  assign control = { ts_start,
                     (fifo_valid_i & fifo_data_i[8] &
                      (fifo_data_i[7:0]==8'b10000000))};

  assign emit_state_we = (opt_en &
                          ((emit_state == `EM_NORMAL) ?
                           ((fifo_valid_i & emit_ready & fifo_data_i[8]) | ts_start) :
                           atb_ready_i));

  always @ (emit_state or control or ts1_valid or ts_continue)
    case (emit_state)
      `EM_NORMAL :
        case (control)
          2'b00  : nxt_emit_state = `EM_OVRFLW;
          2'b01  : nxt_emit_state = `EM_SYNC0;
          2'b10  : nxt_emit_state = `EM_TS_1;
          2'b11  : nxt_emit_state = `EM_SYNC0; // Unreachable
          default: nxt_emit_state =  4'bxxxx;
        endcase
      `EM_OVRFLW : nxt_emit_state = `EM_NORMAL;
      `EM_SYNC0  : nxt_emit_state = `EM_SYNC1;
      `EM_SYNC1  : nxt_emit_state = `EM_SYNC2;
      `EM_SYNC2  : nxt_emit_state = `EM_SYNC3;
      `EM_SYNC3  : nxt_emit_state = `EM_SYNC4;
      `EM_SYNC4  : nxt_emit_state = `EM_SYNC5;
      `EM_SYNC5  : nxt_emit_state = `EM_NORMAL;
      `EM_TS_1   : nxt_emit_state = `EM_TS_2;
      `EM_TS_2   : nxt_emit_state = (~ts1_valid | ts_continue[0]) ? `EM_TS_3: `EM_TS_5;
      `EM_TS_3   : nxt_emit_state = (~ts1_valid | ts_continue[1]) ? `EM_TS_4: `EM_TS_5;
      `EM_TS_4   : nxt_emit_state = `EM_TS_5;
      `EM_TS_5   : nxt_emit_state = `EM_NORMAL;
      default    : nxt_emit_state = 4'bxxxx;
    endcase

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      emit_state <= `EM_NORMAL;
    else if (emit_state_we)
      emit_state <= nxt_emit_state;

  assign ts_active = emit_state[3];

  assign emit_we  = (opt_en &
                     (nxt_emit_valid |
                      (emit_valid & atb_ready_i &
                       ((emit_state==`EM_NORMAL) |
                        (emit_state==`EM_SYNC5)  |
                        (nxt_emit_state == `EM_TS_5) |
                        (emit_state==`EM_TS_5)))));

  assign nxt_emit_valid = ((fifo_valid_i & emit_ready) | ts_start |
                           ((nxt_emit_state == `EM_TS_5) &
                            (~ts1_valid | ts_continue[2])));

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        emit_valid    <= 1'b0;
        fifo_data_reg <= 8'h00;
      end
    else if (emit_we)
      begin
        emit_valid    <= nxt_emit_valid;
        fifo_data_reg <= fifo_data_i[7:0];
      end

  assign atb_data  = (({8{emit_state==`EM_NORMAL}} & fifo_data_reg) |
                      ({8{emit_state==`EM_TS_1}}   & {2'b10,~ts1_valid,5'b10100}) |
                      ({8{(emit_state==`EM_TS_2)   &  ts1_valid}} & {ts_continue[0],ts_value[6:0]}) |
                      ({8{(emit_state==`EM_TS_2)   & ~ts1_valid}} & {1'b1,ts_value[32:26]}) |
                      ({8{(emit_state==`EM_TS_3)   &  ts1_valid}} & {ts_continue[1],ts_value[13:7]}) |
                      ({8{(emit_state==`EM_TS_3)   & ~ts1_valid}} & {1'b1,ts_value[39:33]}) |
                      ({8{(emit_state==`EM_TS_4)   &  ts1_valid}} & {ts_continue[2],ts_value[20:14]}) |
                      ({8{(emit_state==`EM_TS_4)   & ~ts1_valid}} & {1'b1,ts_value[46:40]}) |
                      ({8{(emit_state==`EM_TS_5)   &  ts1_valid}} & {1'b0,ts_wrap, ts_clk_ch, ts_value[25:21]}) |
                      ({8{(emit_state==`EM_TS_5)   & ~ts1_valid}} & {7'b0000000,ts_value[47]}) |
                      ({8{emit_state==`EM_OVRFLW}} & 8'h70) |
                      ({8{emit_state==`EM_SYNC5}}  & 8'h80));

  assign emit_ready = (emit_state==`EM_NORMAL) & (atb_ready_i | ~emit_valid);

  assign itm_no_tx = (emit_state==`EM_NORMAL) & ~emit_valid;


`ifdef ARM_ASSERT_ON

  reg  [8*20:0] EMTSTATE;
  wire          overflow;
  reg           ATTrack;
  wire          at_start;
  wire          at_end;

  always @ (emit_state)
    case (emit_state)
      `EM_NORMAL : EMTSTATE = {"NORMAL"};
      `EM_OVRFLW : EMTSTATE = {"OVRFLW"};
      `EM_SYNC0  : EMTSTATE = {"SYNC0"};
      `EM_SYNC1  : EMTSTATE = {"SYNC1"};
      `EM_SYNC2  : EMTSTATE = {"SYNC2"};
      `EM_SYNC3  : EMTSTATE = {"SYNC3"};
      `EM_SYNC4  : EMTSTATE = {"SYNC4"};
      `EM_SYNC5  : EMTSTATE = {"SYNC5"};
      `EM_TS_1   : EMTSTATE = {"Timestamp1"};
      `EM_TS_2   : EMTSTATE = {"Timestamp2"};
      `EM_TS_3   : EMTSTATE = {"Timestamp3"};
      `EM_TS_4   : EMTSTATE = {"Timestamp4"};
      `EM_TS_5   : EMTSTATE = {"Timestamp5"};
      default    : EMTSTATE = {"UNKNOWN"};
    endcase

  assign overflow = (emit_state==`EM_OVRFLW);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      ATTrack <= 1'b0;
    else
      ATTrack <= (fifo_valid_i & ~emit_ready_o) | (ATTrack & ~emit_ready_o);

  assign at_start = fifo_valid_i & ~emit_ready_o & ~ATTrack;
  assign at_end   = fifo_valid_i &  emit_ready_o &  ATTrack;

  assert_win_unchange
    #(0,9,0,"Slave ATDATA must not change once ATVALID is asserted")
    fifo_data_i_unchange
      (.clk         (clk),
       .reset_n     (reset_n),
       .start_event (at_start),
       .test_expr   (fifo_data_i),
       .end_event   (at_end));

  assert_never #(0,0,
   "emit_valid_o should be asserted during Sync and Overflow states.")
    u_ovl_validsync_check
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (~emit_valid &
                   ((emit_state==`EM_SYNC0) | (emit_state==`EM_SYNC1) |
                    (emit_state==`EM_TS_1)  | (emit_state==`EM_TS_2)  |
                    (emit_state==`EM_SYNC2) | (emit_state==`EM_SYNC3) |
                    (emit_state==`EM_SYNC4) | (emit_state==`EM_SYNC5) |
                    (emit_state==`EM_OVRFLW))));

  assert_never_unknown #(0,4,0,"EMIT state is invalid")
  u_ovl_state_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (emit_state));

  assert_never #(0,0,
   "emit_state should not reach SPARE states")
    u_ovl_state_spare
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr ((emit_state==`EM_SPARE1) |
                   (emit_state==`EM_SPARE2) |
                   (emit_state==`EM_SPARE3)));

  assert_never_unknown #(0,1,0,"ts_capture is invalid")
  u_ovl_ts_capture_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ts_capture));

  assert_never_unknown #(0,1,0,"ts_packet_state_we is invalid")
  u_ovl_ts_packet_state_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ts_packet_state_we));

  assert_never_unknown #(0,1,0,"emit_state_we is invalid")
  u_ovl_emit_state_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (emit_state_we));

  assert_never_unknown #(0,1,0,"emit_we is invalid")
  u_ovl_emit_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (emit_we));

`ifdef ARM_MONITOR_ON

  monitor_itmemit
    u_itmatb_monitor
      (clk,
       reset_n,
       atb_ready_i,
       fifo_valid_i,
       fifo_data_i,
       emit_valid,
       emit_ready,
       atb_data);

`endif

`endif

endmodule
