module cm3_itm_fifo
  (// Inputs
   reset_n, clk, emit_ready_i, arb_valid_i, arb_size_i, arb_overflow_i,
   arb_data_i,
   fifo_valid_o, fifo_data_o, fifo_ready_o, fifo_empty_o
  );

  parameter     TRACE_LVL      = 1;
  parameter     RESET_ALL_REGS = 0;

  input         reset_n;              // System Reset
  input         clk;                  // System Clock
  input         emit_ready_i;         // ATB Ready Master
  input         arb_valid_i;          // ATB Valid Slave
  input   [2:0] arb_size_i;           // ATB Size Slave
  input         arb_overflow_i;       // ATB OverFlow Slave
  input  [39:0] arb_data_i;           // ATB Data Slave

  output        fifo_valid_o;         // ATB Valid Master
  output  [8:0] fifo_data_o;          // ATB Data Master
  output        fifo_ready_o;         // ATB Ready Slave
  output        fifo_empty_o;         // FIFO empty

  reg     [4:0] write_pointer;        // Fifo In Pointer
  reg     [4:0] nxt_write_pointer;    // Fifo In Next
  wire          write_pointer_we;     // Fifo In Enable
  reg     [4:0] read_pointer;         // Fifo Out Pointer
  wire    [4:0] nxt_read_pointer;     // Fifo Out Next
  wire          read_pointer_we;      // Fifo Out Enable
  wire    [4:0] valid_input_bytes;    // Size vector
  wire    [8:0] write_pointer_extended; // write_pointer Rotate
  wire    [4:0] rot_valid_bytes;      // Fifo Write Vector
  wire    [4:0] fifo_write_enable;    // Fifo Write Enable
  wire    [4:0] fifo_read_enable;     // Fifo Read Enable
  wire    [8:0] byte_0_data_in;       // Fifo Data In 0
  wire    [8:0] byte_1_data_in;       // Fifo Data In 1
  wire    [8:0] byte_2_data_in;       // Fifo Data In 2
  wire    [8:0] byte_3_data_in;       // Fifo Data In 3
  wire    [8:0] byte_4_data_in;       // Fifo Data In 4
  wire    [8:0] byte_0_data_out;      // Fifo Data Out 0
  wire    [8:0] byte_1_data_out;      // Fifo Data Out 1
  wire    [8:0] byte_2_data_out;      // Fifo Data Out 2
  wire    [8:0] byte_3_data_out;      // Fifo Data Out 3
  wire    [8:0] byte_4_data_out;      // Fifo Data Out 4
  wire          byte_0_empty;         // Fifo Empty Byte 0
  wire          byte_1_empty;         // Fifo Empty Byte 1
  wire          byte_2_empty;         // Fifo Empty Byte 2
  wire          byte_3_empty;         // Fifo Empty Byte 3
  wire          byte_4_empty;         // Fifo Empty Byte 4
  wire          byte_0_full;          // Fifo Full Byte 0
  wire          byte_1_full;          // Fifo Full Byte 1
  wire          byte_2_full;          // Fifo Full Byte 2
  wire          byte_3_full;          // Fifo Full Byte 3
  wire          byte_4_full;          // Fifo Full Byte 4
  wire          fifo_valid;
  wire    [8:0] fifo_data;
  wire          fifo_ready;
  wire          fifo_empty;
  wire          opt_en;

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  assign fifo_valid_o = (dsm_trace_lvl > 0) ? fifo_valid    : 1'b0;
  assign fifo_data_o  = (dsm_trace_lvl > 0) ? fifo_data     : {9{1'b0}};
  assign fifo_ready_o = (dsm_trace_lvl > 0) ? fifo_ready    : 1'b0;
  assign fifo_empty_o = (dsm_trace_lvl > 0) ? fifo_empty    : 1'b0;
  assign opt_en       = (dsm_trace_lvl > 0) ? 1'b1          : 1'b0;
`else
  assign fifo_valid_o = (TRACE_LVL > 0)     ? fifo_valid    : 1'b0;
  assign fifo_data_o  = (TRACE_LVL > 0)     ? fifo_data     : {9{1'b0}};
  assign fifo_ready_o = (TRACE_LVL > 0)     ? fifo_ready    : 1'b0;
  assign fifo_empty_o = (TRACE_LVL > 0)     ? fifo_empty    : 1'b0;
  assign opt_en       = (TRACE_LVL > 0)     ? 1'b1          : 1'b0;
`endif
  assign fifo_empty  = (byte_0_empty & byte_1_empty & byte_2_empty &
                        byte_3_empty & byte_4_empty);

  always @(write_pointer or arb_size_i)
    case (arb_size_i)
      3'b000:  nxt_write_pointer = {write_pointer[3:0],write_pointer[4]};
      3'b001:  nxt_write_pointer = {write_pointer[2:0],write_pointer[4:3]};
      3'b010:  nxt_write_pointer = {write_pointer[1:0],write_pointer[4:2]};
      3'b011:  nxt_write_pointer = {write_pointer[0],write_pointer[4:1]};
      3'b100:  nxt_write_pointer = write_pointer;
      3'b101:  nxt_write_pointer = 5'b00000;
      3'b110:  nxt_write_pointer = 5'b00000;
      3'b111:  nxt_write_pointer = 5'b00000;
      default: nxt_write_pointer = 5'bxxxxx;
    endcase

  assign write_pointer_we = arb_valid_i & fifo_ready & opt_en;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      write_pointer <= 5'b00001;
    else if (write_pointer_we)
      write_pointer <= nxt_write_pointer;

  assign nxt_read_pointer = {read_pointer[3:0], read_pointer[4]};
  assign read_pointer_we  = fifo_valid & emit_ready_i & opt_en;

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      read_pointer <= 5'b00001;
    else if (read_pointer_we)
      read_pointer <= nxt_read_pointer;

  assign valid_input_bytes[0] = 1'b1;
  assign valid_input_bytes[1] = arb_size_i[0] | valid_input_bytes[2];
  assign valid_input_bytes[2] = arb_size_i[1] | arb_size_i[2];
  assign valid_input_bytes[3] = (arb_size_i[1:0] == 2'b11) | valid_input_bytes[4];
  assign valid_input_bytes[4] = arb_size_i[2];

  assign write_pointer_extended  = {write_pointer[4:0],write_pointer[4:1]};
  assign rot_valid_bytes = (({5{valid_input_bytes[0]}} & write_pointer_extended[8:4]) |
                            ({5{valid_input_bytes[1]}} & write_pointer_extended[7:3]) |
                            ({5{valid_input_bytes[2]}} & write_pointer_extended[6:2]) |
                            ({5{valid_input_bytes[3]}} & write_pointer_extended[5:1]) |
                            ({5{valid_input_bytes[4]}} & write_pointer_extended[4:0]));

  assign fifo_write_enable = {5{write_pointer_we}} & rot_valid_bytes;

  assign fifo_read_enable  = {5{read_pointer_we}}  & read_pointer;

  assign { byte_4_data_in,byte_3_data_in,
           byte_2_data_in,byte_1_data_in,
           byte_0_data_in } =
         (({45{write_pointer[0]}} &
           {1'b0, arb_data_i[39:32], 1'b0, arb_data_i[31:24],
            1'b0, arb_data_i[23:16], 1'b0, arb_data_i[15:8],
            arb_overflow_i, arb_data_i[7:0] }) |
          ({45{write_pointer[1]}} &
           {1'b0, arb_data_i[31:24], 1'b0, arb_data_i[23:16],
            1'b0, arb_data_i[15:8], arb_overflow_i, arb_data_i[7:0],
            1'b0, arb_data_i[39:32]}) |
          ({45{write_pointer[2]}} &
           {1'b0, arb_data_i[23:16], 1'b0, arb_data_i[15:8],
            arb_overflow_i, arb_data_i[7:0], 1'b0, arb_data_i[39:32],
            1'b0, arb_data_i[31:24]}) |
          ({45{write_pointer[3]}} &
           {1'b0, arb_data_i[15:8], arb_overflow_i, arb_data_i[7:0],
            1'b0,arb_data_i[39:32], 1'b0, arb_data_i[31:24],
            1'b0, arb_data_i[23:16]}) |
          ({45{write_pointer[4]}} &
           {arb_overflow_i, arb_data_i[ 7:0],  1'b0, arb_data_i[39:32],
            1'b0, arb_data_i[31:24], 1'b0, arb_data_i[23:16],
            1'b0, arb_data_i[15:8] }) );

  cm3_itm_fifo_byte #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo_byte_0
      (
       .reset_n        (reset_n),
       .clk            (clk),
       .write_enable_i (fifo_write_enable[0]),
       .read_enable_i  (fifo_read_enable[0]),
       .data_i         (byte_0_data_in),

       .empty_o        (byte_0_empty),
       .full_o         (byte_0_full),
       .data_o         (byte_0_data_out)
    );

  cm3_itm_fifo_byte #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo_byte_1
      (
       .reset_n        (reset_n),
       .clk            (clk),
       .write_enable_i (fifo_write_enable[1]),
       .read_enable_i  (fifo_read_enable[1]),
       .data_i         (byte_1_data_in),

       .empty_o        (byte_1_empty),
       .full_o         (byte_1_full),
       .data_o         (byte_1_data_out)
    );

  cm3_itm_fifo_byte #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo_byte_2
      (
       .reset_n        (reset_n),
       .clk            (clk),
       .write_enable_i (fifo_write_enable[2]),
       .read_enable_i  (fifo_read_enable[2]),
       .data_i         (byte_2_data_in),

       .empty_o        (byte_2_empty),
       .full_o         (byte_2_full),
       .data_o         (byte_2_data_out)
       );

  cm3_itm_fifo_byte #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo_byte_3
      (
       .reset_n        (reset_n),
       .clk            (clk),
       .write_enable_i (fifo_write_enable[3]),
       .read_enable_i  (fifo_read_enable[3]),
       .data_i         (byte_3_data_in),

       .empty_o        (byte_3_empty),
       .full_o         (byte_3_full),
       .data_o         (byte_3_data_out)
       );

  cm3_itm_fifo_byte #(TRACE_LVL, RESET_ALL_REGS)
    u_itm_fifo_byte_4
      (
       .reset_n        (reset_n),
       .clk            (clk),
       .write_enable_i (fifo_write_enable[4]),
       .read_enable_i  (fifo_read_enable[4]),
       .data_i         (byte_4_data_in),

       .empty_o        (byte_4_empty),
       .full_o         (byte_4_full),
       .data_o         (byte_4_data_out)
    );

  assign fifo_valid = ((~byte_0_empty) |
                       (~byte_1_empty) |
                       (~byte_2_empty) |
                       (~byte_3_empty) |
                       (~byte_4_empty));

  assign fifo_data = (({9{read_pointer[0]}} & byte_0_data_out) |
                      ({9{read_pointer[1]}} & byte_1_data_out) |
                      ({9{read_pointer[2]}} & byte_2_data_out) |
                      ({9{read_pointer[3]}} & byte_3_data_out) |
                      ({9{read_pointer[4]}} & byte_4_data_out));

  assign fifo_ready = ~((rot_valid_bytes[0] & byte_0_full) |
                        (rot_valid_bytes[1] & byte_1_full) |
                        (rot_valid_bytes[2] & byte_2_full) |
                        (rot_valid_bytes[3] & byte_3_full) |
                        (rot_valid_bytes[4] & byte_4_full));

`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Invalid arb_size_i")
    atsize_check
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr ((arb_size_i==3'b101) |
                   (arb_size_i==3'b110) |
                   (arb_size_i==3'b111)));

  assert_never_unknown #(0,5,0, "FIFO data invalid")
  fifodata_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (write_pointer));

  assert_never_unknown #(0,1,0, "write_pointer_we invalid")
  fifo_write_pointer_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (write_pointer_we));

  assert_never_unknown #(0,1,0, "read_pointer_we invalid")
  fifo_read_pointer_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (read_pointer_we));

  assert_one_hot #(0,5,0,"Fifo In controls must be exclusive")
    fifoin_exclusive
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (write_pointer));

  assert_one_hot #(0,5,0,"Fifo Out controls must be exclusive")
    fifoout_exclusive
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (read_pointer));

  assert_always #(0,0,"valid_input_bytes should always equal arb_size_i")
    sizevec_check
      (
       .clk       (clk),
       .reset_n   (reset_n),
       .test_expr (({2'b00,valid_input_bytes[0]} +
                    {2'b00,valid_input_bytes[1]} +
                    {2'b00,valid_input_bytes[2]} +
                    {2'b00,valid_input_bytes[3]} +
                    {2'b00,valid_input_bytes[4]}) == (arb_size_i + 3'b001)));

  wire   ovl_full_fifo_valid;

  assign ovl_full_fifo_valid = ((read_pointer[0] & ~byte_0_empty) |
                                (read_pointer[1] & ~byte_1_empty) |
                                (read_pointer[2] & ~byte_2_empty) |
                                (read_pointer[3] & ~byte_3_empty) |
                                (read_pointer[4] & ~byte_4_empty));

  assert_always #(0,0,"fifo_valid should not need read pointer mask")
    ovl_fifo_valid_check
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (ovl_full_fifo_valid == fifo_valid));

`ifdef ARM_MONITOR_ON

  monitor_fifo
    ITMFifo_monitor(clk, reset_n, write_pointer_we,
                    {1'b0,arb_size_i},
                    {1'b0,arb_data_i[39:32],
                     1'b0,arb_data_i[31:24],
                     1'b0,arb_data_i[23:16],
                     1'b0,arb_data_i[15:8],
                     arb_overflow_i,arb_data_i[7:0]},
                    read_pointer_we,
                    4'b0000, fifo_data);

`endif

`endif

endmodule
