module cm3_itm_fifo_byte
  (// Inputs
   reset_n, clk, write_enable_i, read_enable_i, data_i,
   empty_o, full_o, data_o
   );

  parameter     TRACE_LVL      = 1;
  parameter     RESET_ALL_REGS = 0;

  input         reset_n;            // FIFO Reset
  input         clk;                // FIFO Clock
  input         write_enable_i;     // Write Enable;
  input         read_enable_i;      // Read Enable;
  input   [8:0] data_i;             // In Data

  output        empty_o;            // Fifo Empty
  output        full_o;             // Fifo Full
  output  [8:0] data_o;             // OutData

  reg     [1:0] write_pointer;      // In Pointer
  wire    [1:0] nxt_write_pointer;  // In Pointer Next
  reg     [1:0] read_pointer;       // Out Pointer
  wire    [1:0] nxt_read_pointer;   // Out Pointer Next
  wire          data_0_we;          // Write Enable 0
  wire          data_1_we;          // Write Enable 0
  reg     [8:0] data_0;             // Fifo Entry 0
  reg     [8:0] data_1;             // Fifo Entry 1
  reg     [8:0] data_final;         // Output stage
  wire    [8:0] nxt_data_final;     // Output stage
  wire          final_we;
  wire          read_pointer_en;
  wire          write_pointer_en;

  wire          nxt_empty;          // Fifo Empty
  reg           empty;              // Fifo Empty
  wire          full;               // Fifo Full
  wire          opt_en;
  wire          opt_reset_all_n;

`ifdef ARM_DSM

  wire          dsm_reset_all_regs;
  wire [1:0]    dsm_trace_lvl;

  assign opt_reset_all_n = (dsm_reset_all_regs == 1) ? reset_n : 1'b1;

  assign empty_o = (dsm_trace_lvl > 0) ? empty      : 1'b0;
  assign full_o  = (dsm_trace_lvl > 0) ? full       : 1'b0;
  assign data_o  = (dsm_trace_lvl > 0) ? data_final : {9{1'b0}};
  assign opt_en  = (dsm_trace_lvl > 0) ? 1'b1       : 1'b0;

`else

  assign opt_reset_all_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

  assign empty_o = (TRACE_LVL > 0) ? empty      : 1'b0;
  assign full_o  = (TRACE_LVL > 0) ? full       : 1'b0;
  assign data_o  = (TRACE_LVL > 0) ? data_final : {9{1'b0}};
  assign opt_en  = (TRACE_LVL > 0) ? 1'b1       : 1'b0;

`endif

  assign nxt_write_pointer = write_pointer + 2'b01;
  assign nxt_read_pointer  = read_pointer  + 2'b01;
  assign write_pointer_en = write_enable_i & opt_en;
  assign read_pointer_en = (~nxt_empty & (empty | read_enable_i)) & opt_en;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      write_pointer <= 2'b00;
    else if (write_pointer_en)
      write_pointer <= nxt_write_pointer;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      read_pointer <= 2'b00;
    else if (read_pointer_en)
      read_pointer <= nxt_read_pointer;

  assign nxt_empty = write_pointer == read_pointer;
  assign full      = ((write_pointer[1]!=read_pointer[1]) &
                      (write_pointer[0] == read_pointer[0]));

  assign data_0_we = write_enable_i & !write_pointer[0] & opt_en;
  assign data_1_we = write_enable_i &  write_pointer[0] & opt_en;

  always @ (posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      data_0 <= {9{1'b0}};
    else if (data_0_we)
      data_0 <= data_i;

  always @ (posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      data_1 <= {9{1'b0}};
    else if (data_1_we)
      data_1 <= data_i;

  assign nxt_data_final = (({9{!read_pointer[0]}} & data_0) |
                           ({9{ read_pointer[0]}} & data_1));

  assign final_we       = read_enable_i | (~nxt_empty & empty);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      empty <= 1'b1;
    else if (final_we)
      empty <= nxt_empty;

  always @ (posedge clk or negedge opt_reset_all_n)
    if (!opt_reset_all_n)
      data_final <= {9{1'b0}};
    else if (final_we)
      data_final <= nxt_data_final;

`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Fifo should not be written when full")
    full_never
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (write_enable_i & full)
       );

  assert_never #(0,0,"Fifo should not be read when empty")
    empty_never
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (read_enable_i & empty_o)
       );

  assert_never_unknown #(0,1,0, "write_pointer_en invalid")
  fifo_write_pointer_en_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (write_pointer_en));

  assert_never_unknown #(0,1,0, "read_pointer_en invalid")
  fifo_read_pointer_en_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (read_pointer_en));

  assert_never_unknown #(0,1,0, "data_0_we invalid")
  fifo_data_0_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (data_0_we));

  assert_never_unknown #(0,1,0, "data_1_we invalid")
  fifo_data_1_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (data_1_we));

  assert_never_unknown #(0,1,0, "final_we invalid")
  fifo_final_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (final_we));

`endif

endmodule
