module cm3_itm_if
  (// Inputs
   reset_n, clk, dpu_trace_enable_i, ahb_sel_i, ahb_prot_i, ahb_trans_i,
   ahb_write_i, ahb_size_i, ahb_addr_i, ahb_wdata_i, ahb_ready_in_i, ready_if_i,
   atb_ready_i, fifo_empty_i, itm_no_tx_i, tpiu_active_i, tpiu_baud_i,
   emit_valid_i,
   ahb_ready_out_o, ahb_resp_o, ahb_rdata_o, if_valid_o, if_overflow_o,
   if_size_o, if_data_o, atb_id_o, dwt_enable_o, sync_enable_o,
   timestamp_enable_o, timestamp_pulse_o, timestamp_prescale_o,
   prg_ts_freq_o, prg_ts_sync_o, atb_valid_o
  );

  parameter     TRACE_LVL      = 1;
  parameter     RESET_ALL_REGS = 0;
  parameter     EXT_PRESENT    = 0;   // Affects ID registers

  input         reset_n;              // AHB reset
  input         clk;                  // AHB clock
  input         dpu_trace_enable_i;   // Enable Tracing
  input         ahb_sel_i;            // AHB select
  input   [3:0] ahb_prot_i;           // AHB protection
  input   [1:0] ahb_trans_i;          // AHB enable
  input         ahb_write_i;          // AHB write
  input   [1:0] ahb_size_i;           // AHB size
  input  [11:0] ahb_addr_i;           // AHB address
  input         ahb_ready_in_i;       // AHB ready
  input  [31:0] ahb_wdata_i;          // AHB write data
  input         ready_if_i;           // ATB ready
  input         atb_ready_i;          // ATB ready ITM (for integration test)
  input         fifo_empty_i;         // FIFO empty
  input         itm_no_tx_i;          // ITM not transmitting
  input         tpiu_active_i;        // TPIU active
  input         tpiu_baud_i;          // TPIU baud
  input         emit_valid_i;         // Data Valid from emitter

  output        ahb_ready_out_o;      // AHB ready
  output        ahb_resp_o;           // AHB error response
  output [31:0] ahb_rdata_o;          // AHB read data
  output        if_valid_o;           // ATB valid
  output        if_overflow_o;        // ATB overflow
  output  [2:0] if_size_o;            // ATB size
  output [39:0] if_data_o;            // ATB data
  output  [6:0] atb_id_o;             // Identifier
  output        dwt_enable_o;         // DWT output enable
  output        sync_enable_o;        // Sync Enable
  output        timestamp_enable_o;   // TimeStamp Enable
  output        timestamp_pulse_o;    // TimeStamp RefClock
  output  [1:0] timestamp_prescale_o; // TimeStamp Prescale
  output  [1:0] prg_ts_freq_o;        // Global timestamp control
  output        prg_ts_sync_o;        // Sync when timestamps first turned on
  output        atb_valid_o;          // atb_valid with integration mode overlayed


`define TENAREG     10'b1110_0000_00
`define TPRIVREG    10'b1110_0100_00
`define CTRLSTATREG 10'b1110_1000_00
`define INTRREG     10'b1110_1111_00
`define INTWREG     10'b1110_1111_10
`define INTMREG     10'b1111_0000_00
`define LOCKCREG    10'b1111_1011_00
`define LOCKSREG    10'b1111_1011_01
`define PID4        10'b1111_1101_00
`define PID5        10'b1111_1101_01
`define PID6        10'b1111_1101_10
`define PID7        10'b1111_1101_11
`define PID0        10'b1111_1110_00
`define PID1        10'b1111_1110_01
`define PID2        10'b1111_1110_10
`define PID3        10'b1111_1110_11
`define CID0        10'b1111_1111_00
`define CID1        10'b1111_1111_01
`define CID2        10'b1111_1111_10
`define CID3        10'b1111_1111_11

`define CTRLACCKEY  32'hC5ACCE55


  wire          ahb_we;
  reg           ahb_read;
  wire          ahb_tr_valid;
  wire          ahb_rd_valid;
  wire          ahb_wr_valid;
  wire          ahb_word_wr;
  wire          ahb_hword_wr;
  wire          ahb_byte_wr;
  reg     [3:0] ahb_write;
  wire    [3:0] nxt_ahb_write;
  wire   [11:2] nxt_haddr;
  reg    [11:2] haddr;
  wire          nxt_prot;
  reg           hprot;
  wire    [1:0] nxt_hsize;
  reg     [1:0] hsize;

  wire          ctrl_stat_we;         // Control/Status enable
  reg    [39:0] buffer_data;          // Fifo buffer
  wire   [39:0] nxt_buffer_data;      // Fifo buffer DFF input
  wire          buffer_we;            // Fifo buffer enable
  reg           buffer_full;          // Fifo Full
  wire          buffer_empty;         // Fifo empty or about to be empty
  wire          nxt_buffer_full;      // Fifo Full DFF input
  wire          buffer_full_we;       // Fifo Full DFF enable
  wire          itm_busy;             // ITM busy
  reg           integration_mode;     // integration control mode
  wire          integration_mode_en;  // integration control mode DFF enable
  reg           integration_write;    // integration control write
  wire          integration_write_en; // integration control write DFF enable
  reg           locked;               // locked Control
  wire          nxt_locked;           // locked Control DFF enable
  wire          locked_we;            // locked Control DFF input
  reg           overflow;             // instrumentation trace overflow
  wire          nxt_overflow;         // overflow DFF input
  wire          overflow_we;          // overflow DFF enable
  reg [31:0]    read_data;            // Register read data
  reg           stimulus_if;          // stimulus interface
  wire          nxt_stimulus_if;      // stimulus interface
  reg [7:0]     stimulus_mask0;       // Byte 0 of stimulus_mask
  reg [7:0]     stimulus_mask1;       // Byte 1 of stimulus_mask
  reg [7:0]     stimulus_mask2;       // Byte 2 of stimulus_mask
  reg [7:0]     stimulus_mask3;       // Byte 3 of stimulus_mask
  wire [31:0]   stimulus_mask;        // Stimulus port mask
  wire [3:0]    stimulus_mask_we;     // Stimulus port enable
  reg [3:0]     priv_control;         // Privilege control
  wire          priv_control_we;      // Privilege control enable
  wire          stimulus_write;       // stimulus port write
  reg           itm_enable;           // ITM enable
  reg           prg_timestamp_enable; // TimeStamp Enable programmers view
  wire          timestamp_enable;     // TimeStamp Enable programmers view
  reg           swv_enable;           // SWV enable
  reg           tpiu_active_reg ;     // tpiu_active_i sampling DFF
  wire          nxt_tpiu_active_reg;
  reg           tpiu_baud_reg1;       // tpiu_baud_i synchroniser DFF
  reg           tpiu_baud_reg2;       // tpiu_baud_i synchroniser DFF
  reg           tpiu_baud_reg3;       // tpiu_baud_i pulse DFF
  wire          nxt_tpiu_baud_reg2;
  wire          nxt_tpiu_baud_reg3;
  wire          tpiu_sync_enable;     // TPIU DFF Enable
  reg           atb_active;
  wire          atb_active_we;
  wire          nxt_atb_active;
  wire [31:0]   ahb_rdata;
  wire          if_valid;
  wire          if_overflow;
  wire [2:0]    if_size;
  wire [39:0]   if_data;
  reg  [6:0]    atb_id;
  wire          atb_valid;
  wire          ahb_sel;

  reg           dwt_enable;
  reg           sync_enable;
  wire          timestamp_pulse;
  reg  [1:0]    timestamp_prescale;
  reg [1:0]     prg_ts_freq;
  wire          prg_ts_sync;
  wire          opt_en;
  wire          full_reset_n;

`ifdef ARM_DSM

  wire [1:0]    dsm_trace_lvl;
  wire          dsm_reset_all_regs;

  assign full_reset_n = (dsm_reset_all_regs == 1) ? reset_n : 1'b1;

  assign ahb_rdata_o          = (dsm_trace_lvl > 0) ? ahb_rdata          : {32{1'b0}};
  assign if_valid_o           = (dsm_trace_lvl > 0) ? if_valid           : 1'b0;
  assign if_overflow_o        = (dsm_trace_lvl > 0) ? if_overflow        : 1'b0;
  assign if_size_o            = (dsm_trace_lvl > 0) ? if_size            : {3{1'b0}};
  assign if_data_o            = (dsm_trace_lvl > 0) ? if_data            : {40{1'b0}};
  assign atb_id_o             = (dsm_trace_lvl > 0) ? atb_id             : {7{1'b0}};
  assign atb_valid_o          = (dsm_trace_lvl > 0) ? atb_valid          : 1'b0;
  assign dwt_enable_o         = (dsm_trace_lvl > 0) ? dwt_enable         : 1'b0;
  assign sync_enable_o        = (dsm_trace_lvl > 0) ? sync_enable        : 1'b0;
  assign timestamp_enable_o   = (dsm_trace_lvl > 0) ? timestamp_enable   : 1'b0;
  assign timestamp_pulse_o    = (dsm_trace_lvl > 0) ? timestamp_pulse    : 1'b0;
  assign timestamp_prescale_o = (dsm_trace_lvl > 0) ? timestamp_prescale : 2'b00;
  assign prg_ts_freq_o        = (dsm_trace_lvl > 0) ? prg_ts_freq        : 2'b00;
  assign prg_ts_sync_o        = (dsm_trace_lvl > 0) ? prg_ts_sync        : 1'b0;
  assign opt_en               = (dsm_trace_lvl > 0) ? 1'b1               : 1'b0;

`else

  assign full_reset_n = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

  assign ahb_rdata_o          = (TRACE_LVL > 0) ? ahb_rdata          : {32{1'b0}};
  assign if_valid_o           = (TRACE_LVL > 0) ? if_valid           : 1'b0;
  assign if_overflow_o        = (TRACE_LVL > 0) ? if_overflow        : 1'b0;
  assign if_size_o            = (TRACE_LVL > 0) ? if_size            : {3{1'b0}};
  assign if_data_o            = (TRACE_LVL > 0) ? if_data            : {40{1'b0}};
  assign atb_id_o             = (TRACE_LVL > 0) ? atb_id             : {7{1'b0}};
  assign atb_valid_o          = (TRACE_LVL > 0) ? atb_valid          : 1'b0;
  assign dwt_enable_o         = (TRACE_LVL > 0) ? dwt_enable         : 1'b0;
  assign sync_enable_o        = (TRACE_LVL > 0) ? sync_enable        : 1'b0;
  assign timestamp_enable_o   = (TRACE_LVL > 0) ? timestamp_enable   : 1'b0;
  assign timestamp_pulse_o    = (TRACE_LVL > 0) ? timestamp_pulse    : 1'b0;
  assign timestamp_prescale_o = (TRACE_LVL > 0) ? timestamp_prescale : 2'b00;
  assign prg_ts_freq_o        = (TRACE_LVL > 0) ? prg_ts_freq        : 2'b00;
  assign prg_ts_sync_o        = (TRACE_LVL > 0) ? prg_ts_sync        : 1'b0;
  assign opt_en               = (TRACE_LVL > 0) ? 1'b1               : 1'b0;

`endif

  assign ahb_sel          = dpu_trace_enable_i & ahb_sel_i & opt_en;
  assign ahb_tr_valid     = ahb_trans_i[1] & ahb_ready_in_i & ahb_sel;
  assign ahb_we           = ahb_tr_valid | ahb_read | (|ahb_write);
  assign ahb_rd_valid     = ahb_tr_valid & ~ahb_write_i;
  assign ahb_wr_valid     = ahb_tr_valid &  ahb_write_i;
  assign ahb_word_wr      = (ahb_size_i==2'b10);
  assign ahb_hword_wr     = (ahb_size_i==2'b01);
  assign ahb_byte_wr      = (ahb_size_i==2'b00);
  assign nxt_ahb_write[0] = (ahb_wr_valid &
                             (ahb_word_wr |
                              (ahb_hword_wr & ~ahb_addr_i[1]) |
                              (ahb_byte_wr & (ahb_addr_i[1:0]==2'b00))));
  assign nxt_ahb_write[1] = (ahb_wr_valid &
                             (ahb_word_wr |
                              (ahb_hword_wr & ~ahb_addr_i[1]) |
                              (ahb_byte_wr & (ahb_addr_i[1:0]==2'b01))));
  assign nxt_ahb_write[2] = (ahb_wr_valid &
                             (ahb_word_wr |
                              (ahb_hword_wr & ahb_addr_i[1]) |
                              (ahb_byte_wr & (ahb_addr_i[1:0]==2'b10))));
  assign nxt_ahb_write[3] = (ahb_wr_valid &
                             (ahb_word_wr |
                              (ahb_hword_wr & ahb_addr_i[1]) |
                              (ahb_byte_wr & (ahb_addr_i[1:0]==2'b11))));
  assign nxt_prot         = (ahb_tr_valid & ahb_prot_i[1] &
                             ({ahb_prot_i[3:2],ahb_prot_i[0]}==3'b000 |
                              {ahb_prot_i[3:2],ahb_prot_i[0]}!=3'b000));
  assign nxt_haddr        = ahb_addr_i[11:2];
  assign nxt_hsize        = ahb_size_i + 2'b01;
  assign nxt_stimulus_if  = (|nxt_ahb_write) & ~(|nxt_haddr[11:7]);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        ahb_read    <= 1'b1;
        ahb_write   <= 4'b0000;
        haddr       <= {10{1'b1}};
        hprot       <= 1'b0;
        hsize       <= 2'b11;
        stimulus_if <= 1'b1;
      end
    else if (ahb_we)
      begin
        ahb_read    <= ahb_rd_valid;
        ahb_write   <= nxt_ahb_write;
        haddr       <= nxt_haddr;
        hprot       <= nxt_prot;
        hsize       <= nxt_hsize;
        stimulus_if <= nxt_stimulus_if;
      end

  assign stimulus_write = (itm_enable & stimulus_if &
                           stimulus_mask[haddr[6:2]] &
                           (hprot | ~priv_control[haddr[6:5]]));

  assign stimulus_mask_we = (ahb_write & {4{haddr[11:2]==`TENAREG}} &
                             {4{~locked}} & {4{opt_en}} &
                             ({4{hprot}} | ~priv_control));

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      stimulus_mask0 <= 8'h00;
    else if (stimulus_mask_we[0])
      stimulus_mask0 <= ahb_wdata_i[7:0];

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      stimulus_mask1 <= 8'h00;
    else if (stimulus_mask_we[1])
      stimulus_mask1 <= ahb_wdata_i[15:8];

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      stimulus_mask2 <= 8'h00;
    else if (stimulus_mask_we[2])
      stimulus_mask2 <= ahb_wdata_i[23:16];

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      stimulus_mask3 <= 8'h00;
    else if (stimulus_mask_we[3])
      stimulus_mask3 <= ahb_wdata_i[31:24];

  assign stimulus_mask = {stimulus_mask3,stimulus_mask2,
                          stimulus_mask1,stimulus_mask0};

  assign priv_control_we = (ahb_write[0] & (haddr[11:2]==`TPRIVREG) & hprot &
                            ~locked);

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      priv_control[3:0] <= 4'b0000;
    else if (priv_control_we & opt_en)
      priv_control[3:0] <= ahb_wdata_i[3:0];

  assign ctrl_stat_we = ((&ahb_write) & (haddr[11:2]==`CTRLSTATREG) & hprot &
                         ~locked);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        atb_id               <= 7'b0000000;
        itm_enable           <= 1'b0;
        prg_timestamp_enable <= 1'b0;
        sync_enable          <= 1'b0;
        dwt_enable           <= 1'b0;
        swv_enable           <= 1'b0;
        timestamp_prescale   <= 2'b00;
        prg_ts_freq          <= 2'b00;

      end
    else if (ctrl_stat_we & opt_en)
      begin
        atb_id               <= ahb_wdata_i[22:16];
        itm_enable           <= ahb_wdata_i[0];
        prg_timestamp_enable <= ahb_wdata_i[1];
        sync_enable          <= ahb_wdata_i[2];
        dwt_enable           <= ahb_wdata_i[3];
        swv_enable           <= ahb_wdata_i[4];
        timestamp_prescale   <= ahb_wdata_i[9:8];
        prg_ts_freq          <= ahb_wdata_i[11:10];
      end

  assign prg_ts_sync = (ctrl_stat_we & (prg_ts_freq == 2'b00) &
                        (|ahb_wdata_i[11:10]));

  assign itm_busy    = (// FIFO empty status when transmitting
                        ~(fifo_empty_i & itm_no_tx_i) |
                        buffer_full);

  assign integration_mode_en  = (ahb_write[0] & (haddr[11:2]==`INTMREG) &
                                 hprot & ~locked);
  assign integration_write_en = (ahb_write[0] & (haddr[11:2]==`INTWREG) &
                                 hprot & ~locked);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      integration_mode <= 1'b0;
    else if (integration_mode_en & opt_en)
      integration_mode <= ahb_wdata_i[0];

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      integration_write <= 1'b0;
    else if (integration_write_en & opt_en)
      integration_write <= ahb_wdata_i[0];

  assign atb_valid = integration_mode ? integration_write : emit_valid_i;

  assign locked_we  = (&ahb_write) & (haddr[11:2]==`LOCKCREG) & hprot;
  assign nxt_locked = (ahb_wdata_i[31:0]!=`CTRLACCKEY);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      locked <= 1'b1;
    else if (locked_we & opt_en)
      locked <= nxt_locked;

  assign ahb_ready_out_o = 1'b1;
  assign ahb_resp_o      = 1'b0;

  always @ (haddr or buffer_full or stimulus_mask or priv_control or
            itm_enable or itm_busy or atb_id or prg_timestamp_enable or
            timestamp_prescale or prg_ts_freq or atb_ready_i or
            integration_mode or locked or dwt_enable or sync_enable or
            swv_enable)
    case (haddr[11:2])
      0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
      16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31
                    : read_data = {{31{1'b0}},~buffer_full};
      `TENAREG      : read_data = stimulus_mask;
      `TPRIVREG     : read_data = {{28{1'b0}},priv_control};
      `CTRLSTATREG  : read_data = {8'b00000000, itm_busy, atb_id,
                                  4'b0000, prg_ts_freq, timestamp_prescale,
                                  3'b000,swv_enable, dwt_enable, sync_enable,
                                   prg_timestamp_enable,itm_enable};
      `INTRREG      : read_data = {24'h000000,7'b0000000,atb_ready_i};
      `INTMREG      : read_data = {24'h000000,7'b0000000,integration_mode};
      `LOCKSREG     : read_data = {24'h000000,6'b000000,locked,1'b1};
      `PID4         : read_data = {24'h000000,8'h04};
      `PID5         : read_data = {24'h000000,8'h00};
      `PID6         : read_data = {24'h000000,8'h00};
      `PID7         : read_data = {24'h000000,8'h00};
      `PID0         : read_data = {24'h000000,8'h01};
      `PID1         : read_data = {24'h000000,8'hB0};
      `PID2         : read_data = {24'h000000,8'h3B};
      `PID3         : read_data = {24'h000000,8'h00};
      `CID0         : read_data = {24'h000000,8'h0D};
      `CID1         : read_data = {24'h000000,8'hE0};
      `CID2         : read_data = {24'h000000,8'h05};
      `CID3         : read_data = {24'h000000,8'hB1};
      default       : read_data = {32{1'b0}};
    endcase

  assign ahb_rdata = read_data;

  assign buffer_empty    = ready_if_i | ~buffer_full;
  assign buffer_we       = stimulus_write & buffer_empty & opt_en;
  assign nxt_buffer_data = {ahb_wdata_i,haddr[6:2],overflow,hsize};

  always @ (posedge clk or negedge full_reset_n)
    if (!full_reset_n)
      buffer_data <= {40{1'b1}};
    else if (buffer_we)
      buffer_data <= nxt_buffer_data;

  assign buffer_full_we  = ((stimulus_write & ~buffer_full) |
                            (buffer_full & ready_if_i & ~stimulus_write));
  assign nxt_buffer_full = ~buffer_full;

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      buffer_full <= 1'b0;
    else if (buffer_full_we & opt_en)
      buffer_full <= nxt_buffer_full;

  assign overflow_we  = ((stimulus_write &
                          (( buffer_empty &  overflow) |
                           (~buffer_empty & ~overflow))) |
                         (ctrl_stat_we & ~ahb_wdata_i[0]));
  assign nxt_overflow = stimulus_write & ~buffer_empty;

  always @ (negedge reset_n or posedge clk)
    if (!reset_n)
      overflow <= 1'b0;
    else if (overflow_we & opt_en)
      overflow <= nxt_overflow;

  assign if_valid    = buffer_full;
  assign if_overflow = if_valid & buffer_data[2];
  assign if_size     = { buffer_data[1:0]==2'b11,
                         buffer_data[1:0]==2'b10,
                         buffer_data[1:0]==2'b01 };
  assign if_data     = { buffer_data[39:3], 1'b0, buffer_data[1:0] };

  assign tpiu_sync_enable    = (opt_en &
                                (tpiu_active_reg |
                                 (prg_timestamp_enable & swv_enable)));
  assign nxt_tpiu_active_reg = tpiu_active_i & prg_timestamp_enable;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
        tpiu_active_reg <= 1'b0;
    else if (tpiu_sync_enable)
      tpiu_active_reg <= nxt_tpiu_active_reg;

  assign nxt_tpiu_baud_reg2 = (prg_timestamp_enable &
                               (tpiu_baud_reg1 | ~swv_enable));
  assign nxt_tpiu_baud_reg3 = (prg_timestamp_enable & tpiu_baud_reg2 &
                               swv_enable);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        tpiu_baud_reg1 <= 1'b0;
        tpiu_baud_reg2 <= 1'b0;
        tpiu_baud_reg3 <= 1'b0;
      end
    else if (dpu_trace_enable_i)
      begin
        tpiu_baud_reg1 <= tpiu_baud_i;
        tpiu_baud_reg2 <= nxt_tpiu_baud_reg2;
        tpiu_baud_reg3 <= nxt_tpiu_baud_reg3;
      end

  assign atb_active_we  = (// Set
                           nxt_atb_active |
                           (atb_active &
                            (tpiu_active_reg |
                             ~prg_timestamp_enable |
                             ~swv_enable)));
  assign nxt_atb_active = (prg_timestamp_enable &
                           (~swv_enable | (itm_busy & ~tpiu_active_reg)));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      atb_active <= 1'b1;
    else if (atb_active_we & opt_en)
      atb_active <= nxt_atb_active;

  assign timestamp_enable = atb_active | tpiu_active_reg;
  assign timestamp_pulse  = (tpiu_baud_reg3 ^ tpiu_baud_reg2);


`ifdef ARM_ASSERT_ON

  assert_zero_one_hot #(0,3,0,"if_size_o must be one-hot when valid")
    ATSIZE_exclusive(clk, reset_n, {3{if_valid}} & if_size);

  assert_never #(0,0,"buffer_full has been incorrectly formed")
    buffer_full_check(clk, reset_n, (buffer_full_we &
                                     (buffer_full==nxt_buffer_full)));

  assert_never #(0,0,"overflow has been incorrectly formed")
    overflow_check2(clk, reset_n,
     stimulus_write & buffer_we & overflow & ~(overflow_we & ~nxt_overflow));

  assert_never_unknown #(0,1,0, "ahb_we invalid")
  fifo_ahb_we_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ahb_we));

  assert_never_unknown #(0,4,0, "stimulus_mask_we invalid")
  fifo_stimulus_mask_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (stimulus_mask_we));

  assert_never_unknown #(0,1,0, "priv_control_we invalid")
  fifo_priv_control_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (priv_control_we));

  assert_never_unknown #(0,1,0, "ctrl_stat_we invalid")
  fifo_ctrl_stat_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ctrl_stat_we));

  assert_never_unknown #(0,1,0, "integration_mode_en invalid")
  fifo_integration_mode_en_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (integration_mode_en));

  assert_never_unknown #(0,1,0, "integration_write_en invalid")
  fifo_integration_write_en_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (integration_write_en));

  assert_never_unknown #(0,1,0, "locked_we invalid")
  fifo_locked_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (locked_we));

  assert_never_unknown #(0,1,0, "overflow_we invalid")
  fifo_overflow_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (overflow_we));

  assert_never_unknown #(0,1,0, "dpu_trace_enable_i invalid")
  fifo_dpu_trace_enable_i_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (dpu_trace_enable_i));

  assert_never_unknown #(0,1,0, "atb_active_we invalid")
  fifo_atb_active_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (atb_active_we));

`ifdef ARM_MONITOR_ON

  monitor_itmif #(EXT_PRESENT)
    ITMIf_monitor(clk, reset_n, dpu_trace_enable_i, ahb_sel_i, ahb_prot_i,
                  ahb_trans_i, ahb_write_i, ahb_size_i, ahb_addr_i,
                  ahb_ready_in_i, ahb_wdata_i, ahb_ready_out_o, ahb_resp_o,
                  ahb_rdata, opt_en, if_valid, ready_if_i, atb_ready_i,
                  if_overflow, if_size, if_data, sync_enable,
                  prg_timestamp_enable, swv_enable, timestamp_prescale);

`endif

`endif

endmodule
