module cm3_itm_ts
  (// Inputs
   reset_n, clk, timestamp_enable_i, core_halt_i, timestamp_pulse_i,
   timestamp_prescale_i, ready_ts_i, if_valid_i, ready_if_i, dwt_enable_i,
   dwt_valid_i, ready_dwt_i,
   dwt_valid_enabled_o, ts_valid_o, ts_overflow_o, ts_size_o, ts_data_o
   );

  parameter     TRACE_LVL = 1;
  parameter     RESET_ALL_REGS = 0;

  input         reset_n;              // System Reset
  input         clk;                  // System Clock
  input         timestamp_enable_i;   // TimeStamp _weable
  input         core_halt_i;          // Core is halted
  input         timestamp_pulse_i;    // TimeStamp RefClock
  input   [1:0] timestamp_prescale_i; // TimeStamp Prescaler
  input         ready_ts_i;           // ATB Ready
  input         if_valid_i;           // ATB Valid
  input         ready_if_i;           // ATB Ready
  input         dwt_enable_i;         // Local control of DWT packets
  input         dwt_valid_i;          // ATB Valid
  input         ready_dwt_i;          // ATB Ready

  output        dwt_valid_enabled_o;  // DWT valid and enabled
  output        ts_valid_o;           // ATB Valid
  output        ts_overflow_o;        // ATB OverFlow
  output  [2:0] ts_size_o;            // ATB Size
  output [39:0] ts_data_o;            // ATB Data

  reg           ts_valid;
  wire          nxt_ts_valid;
  wire          ts_valid_we;
  reg           timestamp_delayed;
  wire          timestamp_delayed_we;
  wire          nxt_timestamp_delayed;
  reg           packet_delayed;
  wire          packet_delayed_we;
  wire          nxt_packet_delayed;
  reg    [20:0] ts_count;
  wire   [20:0] nxt_ts_count;
  wire          ts_count_we;
  wire          ts_clear;
  reg           ts_overflow;
  wire          nxt_ts_overflow;
  wire          ts_overflow_we;
  wire    [2:0] control_bits;
  wire    [2:0] ts_bytes;
  reg     [5:0] prescale_count;
  reg     [5:0] prescale_max_count;
  wire          prescale_count_wrap;
  wire          prescale_count_set;
  wire          prescale_count_clear;
  wire          prescale_count_en;
  wire    [5:0] nxt_prescale_count;
  wire          ts_count_en;
  wire          dwt_valid_enabled;
  wire   [39:0] ts_data;
  wire          opt_en;
  wire          full_reset_n;
  wire [1:0]    ts_size;
  wire [2:0]    opt_ts_size;
  wire          opt_ts_overflow;

`ifdef ARM_DSM

  wire          dsm_reset_all_regs;
  wire [1:0]    dsm_trace_lvl;

  assign dwt_valid_enabled_o = (dsm_trace_lvl > 0) ? dwt_valid_enabled : 1'b0;
  assign ts_valid_o          = (dsm_trace_lvl > 0) ? ts_valid          : 1'b0;
  assign ts_overflow_o       = (dsm_trace_lvl > 0) ? opt_ts_overflow   : 1'b0;
  assign ts_size_o           = (dsm_trace_lvl > 0) ? opt_ts_size       : 3'b000;
  assign ts_data_o           = (dsm_trace_lvl > 0) ? ts_data           : {40{1'b0}};
  assign opt_en              = (dsm_trace_lvl > 0) ? 1'b1              : 1'b0;

  assign full_reset_n        = (dsm_reset_all_regs == 1) ? reset_n : 1'b1;

`else

  assign dwt_valid_enabled_o = (TRACE_LVL > 0) ? dwt_valid_enabled : 1'b0;
  assign ts_valid_o          = (TRACE_LVL > 0) ? ts_valid          : 1'b0;
  assign ts_overflow_o       = (TRACE_LVL > 0) ? opt_ts_overflow   : 1'b0;
  assign ts_size_o           = (TRACE_LVL > 0) ? opt_ts_size       : 3'b000;
  assign ts_data_o           = (TRACE_LVL > 0) ? ts_data           : {40{1'b0}};
  assign opt_en              = (TRACE_LVL > 0) ? 1'b1              : 1'b0;

  assign full_reset_n        = (RESET_ALL_REGS == 1) ? reset_n : 1'b1;

`endif


  always @(timestamp_prescale_i)
    case (timestamp_prescale_i)
      2'b00:   prescale_max_count = 6'b000000;
      2'b01:   prescale_max_count = 6'b000011;
      2'b10:   prescale_max_count = 6'b001111;
      2'b11:   prescale_max_count = 6'b111111;
      default: prescale_max_count = 6'bxxxxxx;
    endcase // case (ts_prescale_i)

  assign prescale_count_wrap  = (prescale_count == prescale_max_count);
  assign prescale_count_set   = (timestamp_enable_i & (|timestamp_prescale_i) &
                                 timestamp_pulse_i);
  assign prescale_count_clear = ((|prescale_count &
                                  (~(|timestamp_prescale_i) |
                                   ~timestamp_enable_i)));
  assign prescale_count_en    = (prescale_count_set | prescale_count_clear);
  assign nxt_prescale_count   = ({6{prescale_count_set &
                                    ~prescale_count_wrap &
                                    ~prescale_count_clear}} &
                                 (prescale_count + 6'b000001));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      prescale_count <= 6'b111111;
    else if (prescale_count_en & opt_en)
      prescale_count <= nxt_prescale_count;

  assign ts_count_en = (timestamp_enable_i & timestamp_pulse_i &
                        ((timestamp_prescale_i==2'b00) |
                         (prescale_count_en & ~|nxt_prescale_count)));

  assign ts_clear     = ~timestamp_enable_i | ready_ts_i;
  assign nxt_ts_valid = (timestamp_enable_i & ~ts_valid &
                         (// HeartBeat (2,000,000 ticks)
                          (ts_count_en & ts_count==21'h1E847E) |
                          ((if_valid_i | dwt_valid_enabled) & (|ts_count))));
  assign ts_valid_we  = (// Set if IF/DWT write and TimeStamp_nxt!=0
                         nxt_ts_valid |
                         (ts_valid & ts_clear));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      ts_valid <= 1'b0;
    else if (ts_valid_we & opt_en)
      ts_valid <= nxt_ts_valid;

  assign nxt_ts_count = (ts_count + {{20{1'b0}},1'b1}) & {21{~ts_clear}};
  assign ts_count_we  = (opt_en &
                         ((ts_count_en & ~core_halt_i) |
                          (((|ts_bytes) | (|ts_count[2:0]) ) & ts_clear)));

  always @ (posedge clk or negedge full_reset_n)
    if (!full_reset_n)
      ts_count <= {21{1'b1}};
    else if (ts_count_we)
      ts_count <= nxt_ts_count;

  assign ts_bytes[2]  = (|ts_count[20:14]);
  assign ts_bytes[1]  = ts_bytes[2] | (|ts_count[13:7]);
  assign ts_bytes[0]  = (ts_bytes[1] | (|ts_count[6:3]) | (&ts_count[2:0]) |
                         timestamp_delayed | packet_delayed);
  assign ts_size      = ({1'b0,ts_bytes[0]} +
                         {1'b0,ts_bytes[1]} +
                         {1'b0,ts_bytes[2]});
  assign control_bits = (ts_bytes[0] ? {1'b1,packet_delayed,timestamp_delayed} :
                                       ts_count[2:0]);

  assign nxt_ts_overflow = ((&ts_count[20:0]) & ts_count_en & ~ready_ts_i &
                            ~ts_overflow);
  assign ts_overflow_we  = (// Set if timer overflows before being read
                            nxt_ts_overflow |
                            (ts_overflow & ts_clear));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      ts_overflow <= 1'b1;
    else if (ts_overflow_we & opt_en)
      ts_overflow <= nxt_ts_overflow;

  assign dwt_valid_enabled     = dwt_valid_i & dwt_enable_i;

  assign nxt_timestamp_delayed = (ts_valid & ~ready_ts_i &
                                  ~if_valid_i & ~dwt_valid_enabled);
  assign timestamp_delayed_we  = (// Set if timestamp delayed
                                  nxt_timestamp_delayed |
                                  (timestamp_delayed & ts_clear));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      timestamp_delayed <= 1'b1;
    else if (timestamp_delayed_we & opt_en)
      timestamp_delayed <= nxt_timestamp_delayed;

  assign nxt_packet_delayed = (timestamp_enable_i &
                               ((if_valid_i  & ~ready_if_i) |
                                (dwt_valid_enabled & ~ready_dwt_i)));
  assign packet_delayed_we  = (// Set if timestamp delayed
                               nxt_packet_delayed |
                               (packet_delayed & ts_clear));

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      packet_delayed <= 1'b1;
    else if (packet_delayed_we & opt_en)
      packet_delayed <= nxt_packet_delayed;

  assign opt_ts_overflow = ts_overflow;
  assign opt_ts_size     = {1'b0,ts_size};
  assign ts_data         = {{8{1'b0}},
                            {1'b0,      ts_count[20:14]},
                            {ts_bytes[2], ts_count[13:7]},
                            {ts_bytes[1], ts_count[6:0]},
                            {ts_bytes[0], control_bits, 4'b0000} };


`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0,6,0, "TS prescale value must not be unknown")
  ts_prescale_i_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (prescale_max_count));

  assert_never_unknown #(0,1,0, "ts_count_we must not be unknown")
  ts_ts_count_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ts_count_we));

  assert_never_unknown #(0,1,0, "ts_overflow_we must not be unknown")
  ts_ts_overflow_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (ts_overflow_we));

  assert_never_unknown #(0,1,0, "timestamp_delayed_we must not be unknown")
  ts_timestamp_delayed_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (timestamp_delayed_we));

  assert_never_unknown #(0,1,0, "packet_delayed_we must not be unknown")
  ts_packet_delayed_we_unknown
    (.clk       (clk),
     .reset_n   (reset_n),
     .qualifier (1'b1),
     .test_expr (packet_delayed_we));

  assert_never #(0,0,"ts_overflow register must toggle when written")
  ts_overflowToggle_check
    (.clk       (clk),
     .reset_n   (reset_n),
     .test_expr (ts_overflow_we & (ts_overflow==nxt_ts_overflow)));

  assert_never #(0,0,"Valid/Ready incorrect against priority rules")
    validready_check
      (.clk       (clk),
       .reset_n   (reset_n),
       .test_expr (ready_ts_i & (if_valid_i | dwt_valid_enabled)));

  assert_implication #(0,0,"Ready should only be seen if valid asserted")
    validready_check2
      (.clk       (clk),
       .reset_n   (reset_n),
       .antecedent_expr (ready_ts_i),
       .consequent_expr (ts_valid));

`endif

endmodule
