module cm3_mpu
  (// Inputs
   hreset_n, hclk, pclk, mpu_disable_i, ppb_hsel_i, ppb_hwrite_i,
   ppb_hsize_i, ppb_haddr_i, ppb_hprot_i, ppb_hwdata_i, ppb_hready_i,
   dpu_ahb_haddri_i, mtx_dpu_ahb_haddracci_i, dpu_ahb_hproti_i,
   dpu_ahb_htransi_i, mtx_mpu_ahb_hreadyouti_i, dpu_ahb_haddrd_i,
   mtx_dpu_ahb_haddraccd_i, dpu_ahb_hprotd_i, dpu_ahb_hwrited_i,
   dpu_ahb_htransd_i, dpu_ahb_hsized_i, mtx_mpu_ahb_hreadyoutd_i,
   nvic_int_mpu_nmihf_i, nvic_no_def_wbuf_i,
   ppb_hresp_o, ppb_hrdata_o, ppb_hreadyout_o, mpu_mtx_ahb_hproti_o,
   mpu_dpu_ahb_hreadyi_o, mpu_dpu_ahb_hrespi_o, mpu_canceli_o,
   mpu_mtx_ahb_hprotd_o, mpu_dpu_ahb_hreadyd_o, mpu_dpu_ahb_hrespd_o,
   mpu_canceld_o, mpu_mem_attri_o, mpu_mem_attrd_o, mpu_addrd_non_buf_o
  );


  parameter MPU_PRESENT     = 0; // MPU present (0 = not present; 1 = present)
  parameter CLKGATE_PRESENT = 0; // Instantiate architectural clock gating if 1


  input         pclk;                    // PPB clock
  input         hclk;                    // AHB clock
  input         hreset_n;                // AHB reset

  input         mpu_disable_i;           // Disable the MPU (act as default)

  input         ppb_hsel_i;              // PPB select
  input  [11:0] ppb_haddr_i;             // PPB address
  input   [1:0] ppb_hsize_i;             // PPB size
  input   [3:0] ppb_hprot_i;             // PPB protection
  input         ppb_hwrite_i;            // PPB write
  input  [31:0] ppb_hwdata_i;            // PPB write data
  input         ppb_hready_i;            // PPB ready in
  output        ppb_hresp_o;             // PPB error response
  output [31:0] ppb_hrdata_o;            // PPB read data
  output        ppb_hreadyout_o;         // PPB ready out

  input  [31:0] dpu_ahb_haddri_i;        // I-AHB address in
  input   [1:0] dpu_ahb_htransi_i;       // I-AHB trans in
  input   [1:0] dpu_ahb_hproti_i;        // I-AHB protection in
  input         mtx_dpu_ahb_haddracci_i; // I-AHB address accept in
  input         mtx_mpu_ahb_hreadyouti_i;// I-AHB ready in
  output        mpu_dpu_ahb_hreadyi_o;   // I-AHB ready out
  output        mpu_dpu_ahb_hrespi_o;    // I-AHB error response out
  output  [3:0] mpu_mtx_ahb_hproti_o;    // I-AHB protection out

  input  [31:0] dpu_ahb_haddrd_i;        // D-AHB address in
  input   [1:0] dpu_ahb_htransd_i;       // D-AHB trans in
  input   [1:0] dpu_ahb_hsized_i;        // D-AHB size in
  input   [1:0] dpu_ahb_hprotd_i;        // D-AHB protection in
  input         dpu_ahb_hwrited_i;       // D-AHB write in
  input         mtx_dpu_ahb_haddraccd_i; // D-AHB address accept in
  input         mtx_mpu_ahb_hreadyoutd_i;// D-AHB ready in
  output        mpu_dpu_ahb_hreadyd_o;   // D-AHB ready out
  output        mpu_dpu_ahb_hrespd_o;    // D-AHB error response out
  output  [3:0] mpu_mtx_ahb_hprotd_o;    // D-AHB protection out

  input         nvic_int_mpu_nmihf_i;    // In NMI or Hard Fault
  input         nvic_no_def_wbuf_i;      // Write buffer disable

  output        mpu_canceli_o;           // I-AHB Cancel access
  output        mpu_canceld_o;           // D-AHB Cancel access

  output  [1:0] mpu_mem_attri_o;         // I-AHB Memory Attributes
  output  [1:0] mpu_mem_attrd_o;         // D-AHB Memory Attributes
  output        mpu_addrd_non_buf_o;     // D-AHB address non-buffable



  wire          ppb_hresp_full;
  wire   [31:0] ppb_hrdata_full;
  wire          ppb_hreadyout_full;
  wire    [3:0] mpu_mtx_ahb_hproti_full;
  wire          mpu_dpu_ahb_hreadyi_full;
  wire          mpu_dpu_ahb_hrespi_full;
  wire          mpu_canceli_full;
  wire    [1:0] mpu_mem_attri_full;
  wire    [3:0] mpu_mtx_ahb_hprotd_full;
  wire          mpu_dpu_ahb_hreadyd_full;
  wire          mpu_dpu_ahb_hrespd_full;
  wire          mpu_canceld_full;
  wire    [1:0] mpu_mem_attrd_full;
  wire          mpu_addrd_non_buf_full;

  wire          ppb_hresp_dflt;
  wire   [31:0] ppb_hrdata_dflt;
  wire          ppb_hreadyout_dflt;
  wire    [3:0] mpu_mtx_ahb_hproti_dflt;
  wire          mpu_dpu_ahb_hreadyi_dflt;
  wire          mpu_dpu_ahb_hrespi_dflt;
  wire          mpu_canceli_dflt;
  wire    [1:0] mpu_mem_attri_dflt;
  wire    [3:0] mpu_htx_ahb_hprotd_dflt;
  wire          mpu_dpu_ahb_hreadyd_dflt;
  wire          mpu_dpu_ahb_hrespd_dflt;
  wire          mpu_canceld_dflt;
  wire    [1:0] mpu_mem_attrd_dflt;
  wire          mpu_addrd_non_buf_dflt;

  wire          opt_ppb_hresp;
  wire   [31:0] opt_ppb_hrdata;
  wire          opt_ppb_hreadyout;
  wire    [3:0] opt_ahb_hproti;
  wire          opt_ahb_hreadyi;
  wire          opt_ahb_hrespi;
  wire          opt_canceli;
  wire    [1:0] opt_mem_attri;
  wire    [3:0] opt_ahb_hprotd;
  wire          opt_ahb_hreadyd;
  wire          opt_ahb_hrespd;
  wire          opt_canceld;
  wire    [1:0] opt_mem_attrd;
  wire          opt_addrd_non_buf;
  wire          opt_mpu_en;

`ifdef ARM_DSM

  wire              dsm_mpu_present;

  assign opt_mpu_en = (dsm_mpu_present)  ? 1'b1 : 1'b0;

`else

  assign opt_mpu_en = (MPU_PRESENT != 0) ? 1'b1 : 1'b0;

`endif



  assign opt_ppb_hresp     = (opt_mpu_en) ? ppb_hresp_full           : ppb_hresp_dflt;
  assign opt_ppb_hrdata    = (opt_mpu_en) ? ppb_hrdata_full          : ppb_hrdata_dflt;
  assign opt_ppb_hreadyout = (opt_mpu_en) ? ppb_hreadyout_full       : ppb_hreadyout_dflt;
  assign opt_ahb_hproti    = (opt_mpu_en) ? mpu_mtx_ahb_hproti_full  : mpu_mtx_ahb_hproti_dflt;
  assign opt_ahb_hreadyi   = (opt_mpu_en) ? mpu_dpu_ahb_hreadyi_full : mpu_dpu_ahb_hreadyi_dflt;
  assign opt_ahb_hrespi    = (opt_mpu_en) ? mpu_dpu_ahb_hrespi_full  : mpu_dpu_ahb_hrespi_dflt;
  assign opt_canceli       = (opt_mpu_en) ? mpu_canceli_full         : mpu_canceli_dflt;
  assign opt_mem_attri     = (opt_mpu_en) ? mpu_mem_attri_full       : mpu_mem_attri_dflt;
  assign opt_ahb_hprotd    = (opt_mpu_en) ? mpu_mtx_ahb_hprotd_full  : mpu_htx_ahb_hprotd_dflt;
  assign opt_ahb_hreadyd   = (opt_mpu_en) ? mpu_dpu_ahb_hreadyd_full : mpu_dpu_ahb_hreadyd_dflt;
  assign opt_ahb_hrespd    = (opt_mpu_en) ? mpu_dpu_ahb_hrespd_full  : mpu_dpu_ahb_hrespd_dflt;
  assign opt_canceld       = (opt_mpu_en) ? mpu_canceld_full         : mpu_canceld_dflt;
  assign opt_mem_attrd     = (opt_mpu_en) ? mpu_mem_attrd_full       : mpu_mem_attrd_dflt;
  assign opt_addrd_non_buf = (opt_mpu_en) ? mpu_addrd_non_buf_full   : mpu_addrd_non_buf_dflt;


  cm3_mpu_full #(MPU_PRESENT)
    u_cm3_mpu_full
      (// Inputs
       .hclk                      (hclk),
       .pclk                      (pclk),
       .hreset_n                  (hreset_n),
       .mpu_disable_i             (mpu_disable_i),
       .nvic_int_mpu_nmihf_i      (nvic_int_mpu_nmihf_i),
       .ppb_hsel_i                (ppb_hsel_i),
       .ppb_hwrite_i              (ppb_hwrite_i),
       .ppb_hsize_i               (ppb_hsize_i),
       .ppb_haddr_i               (ppb_haddr_i),
       .ppb_hwdata_i              (ppb_hwdata_i),
       .ppb_hprot_i               (ppb_hprot_i),
       .ppb_hready_i              (ppb_hready_i),
       .dpu_ahb_haddri_i          (dpu_ahb_haddri_i),
       .dpu_ahb_hproti_i          (dpu_ahb_hproti_i),
       .dpu_ahb_htransi_i         (dpu_ahb_htransi_i),
       .mtx_mpu_ahb_hreadyouti_i  (mtx_mpu_ahb_hreadyouti_i),
       .mtx_dpu_ahb_haddracci_i   (mtx_dpu_ahb_haddracci_i),
       .dpu_ahb_haddrd_i          (dpu_ahb_haddrd_i),
       .dpu_ahb_hsized_i          (dpu_ahb_hsized_i),
       .dpu_ahb_hprotd_i          (dpu_ahb_hprotd_i),
       .dpu_ahb_hwrited_i         (dpu_ahb_hwrited_i),
       .dpu_ahb_htransd_i         (dpu_ahb_htransd_i),
       .mtx_mpu_ahb_hreadyoutd_i  (mtx_mpu_ahb_hreadyoutd_i),
       .mtx_dpu_ahb_haddraccd_i   (mtx_dpu_ahb_haddraccd_i),
       .nvic_no_def_wbuf_i        (nvic_no_def_wbuf_i),
       .ppb_hresp_o               (ppb_hresp_full),
       .ppb_hrdata_o              (ppb_hrdata_full),
       .ppb_hreadyout_o           (ppb_hreadyout_full),
       .mpu_dpu_ahb_hreadyi_o     (mpu_dpu_ahb_hreadyi_full),
       .mpu_dpu_ahb_hrespi_o      (mpu_dpu_ahb_hrespi_full),
       .mpu_dpu_ahb_hreadyd_o     (mpu_dpu_ahb_hreadyd_full),
       .mpu_dpu_ahb_hrespd_o      (mpu_dpu_ahb_hrespd_full),
       .mpu_mtx_ahb_hproti_o      (mpu_mtx_ahb_hproti_full),
       .mpu_mtx_ahb_hprotd_o      (mpu_mtx_ahb_hprotd_full),
       .mpu_mem_attrd_o           (mpu_mem_attrd_full),
       .mpu_mem_attri_o           (mpu_mem_attri_full),
       .mpu_canceli_o             (mpu_canceli_full),
       .mpu_canceld_o             (mpu_canceld_full),
       .mpu_addrd_non_buf_o       (mpu_addrd_non_buf_full)
      );



  cm3_mpu_default #(MPU_PRESENT)
    u_cm3_mpu_default
      (// Inputs
       .hclk                      (hclk),
       .hreset_n                  (hreset_n),
       .ppb_hsel_i                (ppb_hsel_i),
       .ppb_hprot_i               (ppb_hprot_i),
       .ppb_hready_i              (ppb_hready_i),
       .dpu_ahb_haddri_i          (dpu_ahb_haddri_i),
       .dpu_ahb_hproti_i          (dpu_ahb_hproti_i),
       .dpu_ahb_htransi_i         (dpu_ahb_htransi_i),
       .mtx_mpu_ahb_hreadyouti_i  (mtx_mpu_ahb_hreadyouti_i),
       .mtx_dpu_ahb_haddracci_i   (mtx_dpu_ahb_haddracci_i),
       .dpu_ahb_haddrd_i          (dpu_ahb_haddrd_i),
       .dpu_ahb_hsized_i          (dpu_ahb_hsized_i),
       .dpu_ahb_hprotd_i          (dpu_ahb_hprotd_i),
       .dpu_ahb_htransd_i         (dpu_ahb_htransd_i),
       .mtx_mpu_ahb_hreadyoutd_i  (mtx_mpu_ahb_hreadyoutd_i),
       .mtx_dpu_ahb_haddraccd_i   (mtx_dpu_ahb_haddraccd_i),
       .nvic_no_def_wbuf_i        (nvic_no_def_wbuf_i),
       .ppb_hresp_o               (ppb_hresp_dflt),
       .ppb_hrdata_o              (ppb_hrdata_dflt),
       .ppb_hreadyout_o           (ppb_hreadyout_dflt),
       .mpu_dpu_ahb_hreadyi_o     (mpu_dpu_ahb_hreadyi_dflt),
       .mpu_dpu_ahb_hrespi_o      (mpu_dpu_ahb_hrespi_dflt),
       .mpu_dpu_ahb_hreadyd_o     (mpu_dpu_ahb_hreadyd_dflt),
       .mpu_dpu_ahb_hrespd_o      (mpu_dpu_ahb_hrespd_dflt),
       .mpu_mtx_ahb_hproti_o      (mpu_mtx_ahb_hproti_dflt),
       .mpu_mtx_ahb_hprotd_o      (mpu_htx_ahb_hprotd_dflt),
       .mpu_mem_attrd_o           (mpu_mem_attrd_dflt),
       .mpu_mem_attri_o           (mpu_mem_attri_dflt),
       .mpu_canceli_o             (mpu_canceli_dflt),
       .mpu_canceld_o             (mpu_canceld_dflt),
       .mpu_addrd_non_buf_o       (mpu_addrd_non_buf_dflt)
      );



  assign ppb_hresp_o           = opt_ppb_hresp;
  assign ppb_hrdata_o          = opt_ppb_hrdata;
  assign ppb_hreadyout_o       = opt_ppb_hreadyout;
  assign mpu_mtx_ahb_hproti_o  = opt_ahb_hproti;
  assign mpu_dpu_ahb_hreadyi_o = opt_ahb_hreadyi;
  assign mpu_dpu_ahb_hrespi_o  = opt_ahb_hrespi;
  assign mpu_canceli_o         = opt_canceli;
  assign mpu_mem_attri_o       = opt_mem_attri;
  assign mpu_mtx_ahb_hprotd_o  = opt_ahb_hprotd;
  assign mpu_dpu_ahb_hreadyd_o = opt_ahb_hreadyd;
  assign mpu_dpu_ahb_hrespd_o  = opt_ahb_hrespd;
  assign mpu_canceld_o         = opt_canceld;
  assign mpu_mem_attrd_o       = opt_mem_attrd;
  assign mpu_addrd_non_buf_o   = opt_addrd_non_buf;

endmodule
