`include "cm3_mpu_defs.vp"

module cm3_mpu_align
  (// Inputs
   hclk, hreset_n, dpu_ahb_haddr_i, mtx_dpu_ahb_haddracc_i,
   mpu_dpu_ahb_hready_i, dpu_ahb_hsize_i, dpu_ahb_htrans_i, dpu_ahb_hwrite_i,
   dpu_ahb_hprot_i, mpu_cancel_i,

   aligned_addr_o, check_aligned_o, aligned_write_o, aligned_priv_o,
   mask_cancel_o
  );


  parameter     MPU_PRESENT = 0;   // MPU present (0 = not present; 1 = present)



  input         hclk;                   // AHB clock
  input         hreset_n;               // AHB reset

  input  [31:0] dpu_ahb_haddr_i;        // AHB address in
  input   [1:0] dpu_ahb_htrans_i;       // AHB trans in
  input   [1:0] dpu_ahb_hsize_i;        // AHB size in
  input   [1:0] dpu_ahb_hprot_i;        // AHB protection in
  input         dpu_ahb_hwrite_i;       // AHB write in
  input         mtx_dpu_ahb_haddracc_i; // AHB address accept in
  input         mpu_dpu_ahb_hready_i;   // AHB ready, from MPU to core
  input         mpu_cancel_i;           // AHB Transaction cancelled

  output [31:5] aligned_addr_o;         // Aligned addr for unaligned access
  output        check_aligned_o;        // Need to check an aligned access
  output        aligned_write_o;        // Aligned write for unaligned access
  output        aligned_priv_o;         // Aligned priv for unaligned access
  output        mask_cancel_o;          // Mask mpu_cancel signal


  wire          unaligned;
  wire          trans;
  wire          ppb_access;
  wire   [31:2] nxt_addr;
  wire   [31:0] nxt_aligned_addr;

  reg    [31:5] aligned_addr;
  reg           check_aligned;
  reg           aligned_priv;
  reg           aligned_write;
  reg           mask_cancel;
  wire          nxt_check_aligned;
  wire          nxt_mask_cancel;
  wire          check_aligned_en;
  wire          aligned_addr_en;
  wire          mask_cancel_en;

  wire   [31:5] opt_aligned_addr;
  wire          opt_check_aligned;
  wire          opt_aligned_priv;
  wire          opt_aligned_write;
  wire          opt_mask_cancel;
  wire          opt_mst_en;
  wire          opt_mpu_en;

`ifdef ARM_DSM

  wire          dsm_mpu_present;

  assign opt_mpu_en = (dsm_mpu_present)  ? 1'b1 : 1'b0;

`else

  assign opt_mpu_en = (MPU_PRESENT != 0) ? 1'b1 : 1'b0;

`endif

  assign opt_aligned_addr  = (opt_mpu_en) ? aligned_addr  : {27{1'b0}};
  assign opt_check_aligned = (opt_mpu_en) ? check_aligned : 1'b0;
  assign opt_aligned_priv  = (opt_mpu_en) ? aligned_priv  : 1'b0;
  assign opt_aligned_write = (opt_mpu_en) ? aligned_write : 1'b0;
  assign opt_mask_cancel   = (opt_mpu_en) ? mask_cancel   : 1'b0;
  assign opt_mst_en        = (opt_mpu_en) ? 1'b1          : 1'b0;


  assign trans      = (dpu_ahb_htrans_i[1] & mtx_dpu_ahb_haddracc_i &
                       mpu_dpu_ahb_hready_i);

  assign unaligned  = (((&dpu_ahb_haddr_i[1:0]) & dpu_ahb_hsize_i[0]) |
                       ((|dpu_ahb_haddr_i[1:0]) & dpu_ahb_hsize_i[1]));

  assign ppb_access = (dpu_ahb_haddr_i[31:20] == `MPUPPBREGION);

  assign nxt_mask_cancel = (trans & ~ppb_access &
                            (dpu_ahb_haddr_i[1:0] == 2'b01) &
                            (dpu_ahb_hsize_i == 2'b01));

  assign nxt_check_aligned = trans & unaligned & ~mpu_cancel_i & ~ppb_access;

  assign nxt_addr         = {30{nxt_check_aligned}} & dpu_ahb_haddr_i[31:2];
  assign nxt_aligned_addr = { nxt_addr + {{29{1'b0}}, nxt_check_aligned}, 2'b00 };


  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      check_aligned <= 1'b0;
    else if (check_aligned_en)
      check_aligned <= nxt_check_aligned;

  assign check_aligned_en = (nxt_check_aligned | check_aligned) & opt_mst_en;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      aligned_addr <= {27{1'b0}};
    else if (aligned_addr_en)
      aligned_addr <= nxt_aligned_addr[31:5];

  assign aligned_addr_en = nxt_check_aligned & opt_mst_en;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      aligned_write <= 1'b0;
    else if (aligned_addr_en)
      aligned_write <= dpu_ahb_hwrite_i;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      aligned_priv <= 1'b0;
    else if (aligned_addr_en)
      aligned_priv <= dpu_ahb_hprot_i[1];

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      mask_cancel <= 1'b0;
    else if (mask_cancel_en)
      mask_cancel <= nxt_mask_cancel;

  assign mask_cancel_en = (nxt_mask_cancel | mask_cancel) & opt_mst_en;

  assign aligned_addr_o  = opt_aligned_addr;
  assign check_aligned_o = opt_check_aligned;
  assign aligned_priv_o  = opt_aligned_priv;
  assign aligned_write_o = opt_aligned_write;
  assign mask_cancel_o   = opt_mask_cancel;

endmodule

`include "cm3_mpu_undefs.vp"
