`include "cm3_mpu_defs.vp"

module cm3_mpu_ppb_intf
  (// Inputs
   hreset_n, pclk, hclk, mpu_disable_i, ppb_hready_i, ppb_hsel_i, ppb_hwrite_i,
   ppb_haddr_i, ppb_hsize_i, rgn_hrdata_i, ppb_hwdata_i, ppb_hprot_i,
   nvic_int_mpu_nmihf_i,
   rgn_select_oh_o, rgn_select_wr_o, rgn_base_addr_sel_o, rgn_nxt_wdata_o,
   rgn_attrs_sel_o, rgn_byte_wr_o, ppb_hrdata_o, ppb_hreadyout_o, ppb_hresp_o,
   mpu_enabled_o, def_mem_map_o
  );

  parameter     MPU_PRESENT = 0;   // MPU present (0 = not present; 1 = present)


  input         hreset_n;             // AHB reset
  input         pclk;                 // PPB clock
  input         hclk;                 // AHB clock

  input         mpu_disable_i;        // Disable the MPU (act as default)

  input         ppb_hsel_i;           // PPB select
  input  [11:0] ppb_haddr_i;          // PPB address
  input   [1:0] ppb_hsize_i;          // PPB size
  input   [3:0] ppb_hprot_i;          // PPB protection
  input         ppb_hwrite_i;         // PPB write
  input         ppb_hready_i;         // PPB ready in
  input  [31:0] ppb_hwdata_i;         // PPB write data
  output        ppb_hresp_o;          // PPB error response
  output [31:0] ppb_hrdata_o;         // PPB read data
  output        ppb_hreadyout_o;      // PPB ready out

  input         nvic_int_mpu_nmihf_i; // In an NMI or Hard Fault

  output        mpu_enabled_o;        // Master MPU Enable
  output        def_mem_map_o;        // Enable default memory map

  input  [31:0] rgn_hrdata_i;         // Read Data Specific to a Selected Region
  output  [7:0] rgn_select_oh_o;      // Region Select (one hot)
  output  [7:0] rgn_select_wr_o;      // Region Select for writes (one hot)
  output        rgn_base_addr_sel_o;  // Base Address select
  output        rgn_attrs_sel_o;      // Attribute/Size select
  output [31:0] rgn_nxt_wdata_o;      // Next Write Data for Base or Attributes
  output  [3:0] rgn_byte_wr_o;        // Byte Lane Select for Writes


  wire          mpu_trans;
  wire          ahb_en;
  reg           ahb_rd;
  wire          nxt_ahb_rd;
  wire    [3:0] byte_lanes;
  reg     [3:0] rgn_byte_wr_reg;
  wire    [3:0] rgn_byte_wr;
  wire    [3:0] nxt_rgn_byte_wr;
  wire          align_check;
  wire          ctl_reg_en;
  wire          nxt_priv_fault;
  reg     [5:2] haddr_reg;
  reg           aligned;
  wire          nxt_aligned;

  wire          ppb_hresp;
  reg    [31:0] ppb_hrdata;
  reg           ppb_hreadyout;
  wire          phreadyout_en;
  wire   [31:0] rgn_nxt_wdata;
  wire          mpu_enabled;
  wire          def_mem_map;
  reg           ppb_fault;
  wire          ppb_fault_en;

  wire          rgn_base_addr_sel;
  wire          rgn_attrs_sel;
  wire          rgn_override;

  reg           attr_reg_enable;
  reg           ctl_reg_hfnmiena;
  reg           ctl_reg_privdefena;
  wire    [7:0] nxt_rgn_select_oh;
  reg     [7:0] rgn_select_oh;
  wire    [2:0] rgn_select;
  wire    [7:0] rgn_select_wr;
  wire          rgn_num_wr;

  wire          opt_ppb_hresp;
  wire   [31:0] opt_ppb_hrdata;
  wire          opt_ppb_hreadyout;
  wire    [7:0] opt_rgn_select_oh;
  wire    [7:0] opt_rgn_select_wr;
  wire          opt_rgn_base_addr_sel;
  wire          opt_rgn_attrs_sel;
  wire   [31:0] opt_rgn_nxt_wdata;
  wire    [3:0] opt_rgn_byte_wr;
  wire          opt_mpu_enabled;
  wire          opt_def_mem_map;
  wire          opt_mst_en;

  wire          mst_dsm_en;


`ifdef ARM_DSM

  wire          dsm_mpu_present;

  assign mst_dsm_en = dsm_mpu_present ? 1'b1 : 1'b0;

`else

  assign mst_dsm_en = (MPU_PRESENT == 1) ? 1'b1 : 1'b0;

`endif

  assign opt_ppb_hresp          = (mst_dsm_en) ? ppb_hresp          : 1'b0;
  assign opt_ppb_hrdata         = (mst_dsm_en) ? ppb_hrdata         : {32{1'b0}};
  assign opt_ppb_hreadyout      = (mst_dsm_en) ? ppb_hreadyout      : 1'b1;
  assign opt_rgn_select_oh      = (mst_dsm_en) ? rgn_select_oh      : {8{1'b0}};
  assign opt_rgn_select_wr      = (mst_dsm_en) ? rgn_select_wr      : {8{1'b0}};
  assign opt_rgn_base_addr_sel  = (mst_dsm_en) ? rgn_base_addr_sel  : 1'b0;
  assign opt_rgn_attrs_sel      = (mst_dsm_en) ? rgn_attrs_sel      : 1'b0;
  assign opt_rgn_nxt_wdata      = (mst_dsm_en) ? rgn_nxt_wdata      : {32{1'b0}};
  assign opt_rgn_byte_wr        = (mst_dsm_en) ? rgn_byte_wr        : 4'b0000;
  assign opt_mpu_enabled        = (mst_dsm_en) ? mpu_enabled        : 1'b0;
  assign opt_def_mem_map        = (mst_dsm_en) ? def_mem_map        : 1'b0;
  assign opt_mst_en             = (mst_dsm_en) ? 1'b1               : 1'b0;


  assign mpu_trans     = ppb_hsel_i & ppb_hprot_i[1] & ppb_hready_i;

  assign nxt_ahb_rd    = ~ppb_hwrite_i & mpu_trans & ~mpu_disable_i;

  assign byte_lanes[0] = (mpu_trans & opt_mst_en & ~mpu_disable_i &
                          ((ppb_hsize_i==2'b10) |
                           (ppb_hsize_i==2'b01 & ppb_haddr_i[1]==1'b0) |
                           (ppb_hsize_i==2'b00 & ppb_haddr_i[1:0]==2'b00)));

  assign byte_lanes[1] = (mpu_trans & opt_mst_en & ~mpu_disable_i &
                          ((ppb_hsize_i==2'b10) |
                           (ppb_hsize_i==2'b01 & ppb_haddr_i[1]==1'b0) |
                           (ppb_hsize_i==2'b00 & ppb_haddr_i[1:0]==2'b01)));

  assign byte_lanes[2] = (mpu_trans & opt_mst_en & ~mpu_disable_i &
                          ((ppb_hsize_i==2'b10) |
                           (ppb_hsize_i==2'b01 & ppb_haddr_i[1]==1'b1) |
                           (ppb_hsize_i==2'b00 & ppb_haddr_i[1:0]==2'b10)));

  assign byte_lanes[3] = (mpu_trans & opt_mst_en & ~mpu_disable_i &
                          ((ppb_hsize_i==2'b10) |
                           (ppb_hsize_i==2'b01 & ppb_haddr_i[1]==1'b1) |
                           (ppb_hsize_i==2'b00 & ppb_haddr_i[1:0]==2'b11)));

  assign nxt_rgn_byte_wr = byte_lanes & {4{ppb_hwrite_i}};

  assign nxt_priv_fault = ppb_hready_i & ppb_hsel_i & ~ppb_hprot_i[1];

  assign nxt_aligned    = ~(|ppb_haddr_i[1:0]);

  assign ahb_en = (mpu_trans | (((|rgn_byte_wr) | ahb_rd) & ppb_hready_i)) &
                  opt_mst_en & mst_dsm_en & ~mpu_disable_i;

  always @ (posedge hclk or negedge hreset_n)
     if (!hreset_n)
       begin
         ahb_rd          <= 1'b0;
         rgn_byte_wr_reg <= 4'b0000;
         haddr_reg[5:2]  <= 4'b0000;
         aligned         <= 1'b0;
       end
     else if (ahb_en)
       begin
         ahb_rd          <= nxt_ahb_rd;
         rgn_byte_wr_reg <= nxt_rgn_byte_wr;
         haddr_reg[5:2]  <= ppb_haddr_i[5:2];
         aligned         <= nxt_aligned;
       end

  assign align_check = (aligned & (&rgn_byte_wr_reg)) | ~rgn_base_addr_sel;
  assign rgn_byte_wr = rgn_byte_wr_reg & {4{align_check}};


  assign ctl_reg_en = opt_mst_en &
                      (haddr_reg[5:2] == `MPUCONTROL) & rgn_byte_wr[0];

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      begin
        ctl_reg_privdefena <= 1'b0;
        ctl_reg_hfnmiena   <= 1'b0;
        attr_reg_enable    <= 1'b0;
      end
    else if (ctl_reg_en)
      begin
        ctl_reg_privdefena <= ppb_hwdata_i[2];
        ctl_reg_hfnmiena   <= ppb_hwdata_i[1];
        attr_reg_enable    <= ppb_hwdata_i[0];
      end

  assign rgn_num_wr = (opt_mst_en &
                      (((haddr_reg[5:2] == `MPUREGIONNUM) &
                        rgn_byte_wr[0] & ~(|ppb_hwdata_i[7:3])) |
                       rgn_override));

  assign nxt_rgn_select_oh[0] = (ppb_hwdata_i[2:0] == 3'b000);
  assign nxt_rgn_select_oh[1] = (ppb_hwdata_i[2:0] == 3'b001);
  assign nxt_rgn_select_oh[2] = (ppb_hwdata_i[2:0] == 3'b010);
  assign nxt_rgn_select_oh[3] = (ppb_hwdata_i[2:0] == 3'b011);
  assign nxt_rgn_select_oh[4] = (ppb_hwdata_i[2:0] == 3'b100);
  assign nxt_rgn_select_oh[5] = (ppb_hwdata_i[2:0] == 3'b101);
  assign nxt_rgn_select_oh[6] = (ppb_hwdata_i[2:0] == 3'b110);
  assign nxt_rgn_select_oh[7] = (ppb_hwdata_i[2:0] == 3'b111);

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      rgn_select_oh <= 8'b00000001;
    else if (rgn_num_wr)
      rgn_select_oh <= nxt_rgn_select_oh;

  assign rgn_select = (({3{rgn_select_oh[1]}} & 3'b001) |
                       ({3{rgn_select_oh[2]}} & 3'b010) |
                       ({3{rgn_select_oh[3]}} & 3'b011) |
                       ({3{rgn_select_oh[4]}} & 3'b100) |
                       ({3{rgn_select_oh[5]}} & 3'b101) |
                       ({3{rgn_select_oh[6]}} & 3'b110) |
                       ({3{rgn_select_oh[7]}} & 3'b111));


  assign rgn_base_addr_sel = ((haddr_reg[5:2] == `MPUBASEADDR0) |
                              (haddr_reg[5:2] == `MPUBASEADDR1) |
                              (haddr_reg[5:2] == `MPUBASEADDR2) |
                              (haddr_reg[5:2] == `MPUBASEADDR3));

  assign rgn_attrs_sel     = ((haddr_reg[5:2] == `MPUATTRS0) |
                              (haddr_reg[5:2] == `MPUATTRS1) |
                              (haddr_reg[5:2] == `MPUATTRS2) |
                              (haddr_reg[5:2] == `MPUATTRS3));

  assign rgn_override     = (rgn_base_addr_sel & (&rgn_byte_wr_reg) &
                             aligned & (ppb_hwdata_i[4:3] == 2'b10));

  assign rgn_select_wr[0] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b000) :
                                            rgn_select_oh[0]);
  assign rgn_select_wr[1] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b001) :
                                            rgn_select_oh[1]);
  assign rgn_select_wr[2] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b010) :
                                            rgn_select_oh[2]);
  assign rgn_select_wr[3] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b011) :
                                            rgn_select_oh[3]);
  assign rgn_select_wr[4] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b100) :
                                            rgn_select_oh[4]);
  assign rgn_select_wr[5] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b101) :
                                            rgn_select_oh[5]);
  assign rgn_select_wr[6] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b110) :
                                            rgn_select_oh[6]);
  assign rgn_select_wr[7] = (rgn_override ? (ppb_hwdata_i[2:0] == 3'b111) :
                                            rgn_select_oh[7]);


  assign rgn_nxt_wdata = ppb_hwdata_i;

  always @ (haddr_reg or ahb_rd or attr_reg_enable or ctl_reg_hfnmiena or
            ctl_reg_privdefena or rgn_select or rgn_hrdata_i)
    case (haddr_reg[5:2] & {4{ahb_rd}})
      `MPUTYPE:      ppb_hrdata = {{20{1'b0}}, 4'b1000, {8{1'b0}}};
      `MPUCONTROL:   ppb_hrdata = {{29{1'b0}}, ctl_reg_privdefena,
                                   ctl_reg_hfnmiena, attr_reg_enable};
      `MPUREGIONNUM: ppb_hrdata = {{29{1'b0}}, rgn_select};
      `MPUBASEADDR0: ppb_hrdata = {rgn_hrdata_i[31: 5], 2'b00, rgn_select};
      `MPUBASEADDR1: ppb_hrdata = {rgn_hrdata_i[31: 5], 2'b00, rgn_select};
      `MPUBASEADDR2: ppb_hrdata = {rgn_hrdata_i[31: 5], 2'b00, rgn_select};
      `MPUBASEADDR3: ppb_hrdata = {rgn_hrdata_i[31: 5], 2'b00, rgn_select};
      `MPUATTRS0:    ppb_hrdata = rgn_hrdata_i[31:0];
      `MPUATTRS1:    ppb_hrdata = rgn_hrdata_i[31:0];
      `MPUATTRS2:    ppb_hrdata = rgn_hrdata_i[31:0];
      `MPUATTRS3:    ppb_hrdata = rgn_hrdata_i[31:0];
      default:       ppb_hrdata = 32'h00000000;
    endcase

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      ppb_hreadyout <= 1'b1;
    else if (phreadyout_en)
      ppb_hreadyout <= ~ppb_hreadyout;

  assign phreadyout_en = (nxt_priv_fault | ~ppb_hreadyout) & opt_mst_en;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      ppb_fault <= 1'b0;
    else if (ppb_fault_en)
      ppb_fault <= nxt_priv_fault;

  assign ppb_fault_en = (ppb_hreadyout & (nxt_priv_fault | ppb_fault) &
                         opt_mst_en);
  assign ppb_hresp    = ppb_fault;


  assign mpu_enabled = (attr_reg_enable &
                        (~nvic_int_mpu_nmihf_i | ctl_reg_hfnmiena));

  assign def_mem_map = ctl_reg_privdefena;

  assign rgn_select_oh_o      = opt_rgn_select_oh;
  assign rgn_select_wr_o      = opt_rgn_select_wr;
  assign rgn_base_addr_sel_o  = opt_rgn_base_addr_sel;
  assign rgn_attrs_sel_o      = opt_rgn_attrs_sel;
  assign rgn_nxt_wdata_o      = opt_rgn_nxt_wdata;
  assign rgn_byte_wr_o        = opt_rgn_byte_wr;
  assign ppb_hresp_o          = opt_ppb_hresp;
  assign ppb_hrdata_o         = opt_ppb_hrdata;
  assign ppb_hreadyout_o      = opt_ppb_hreadyout;
  assign mpu_enabled_o        = opt_mpu_enabled;
  assign def_mem_map_o        = opt_def_mem_map;


`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"ppb_hreadyout must only be deasserted when ppb_hresp is set")
    ovl_phreadyout1(hclk, hreset_n, (~ppb_hreadyout & ~ppb_hresp));

  assert_next #(0,1,0,0,0,
    "ppb_hreadyout must only be deasserted for a single cycle")
    ovl_phreadyout2(hclk, hreset_n, (ppb_hresp & ~ppb_hreadyout), ppb_hreadyout);

  assert_zero_one_hot #(0,8,0,"rgn_select_oh must be one-hot")
    ovl_reg_sel_oh(hclk, hreset_n, rgn_select_oh);

  assert_zero_one_hot #(0,8,0,"rgn_select_wr must be one-hot")
    ovl_reg_wr_oh(hclk, hreset_n, rgn_select_wr);

`endif

endmodule

`include "cm3_mpu_undefs.vp"
