module cm3_mpu_region
  (// Inputs
   hclk, pclk, hreset_n, mpu_enabled_i, rgn_select_wr_i, rgn_base_addr_sel_i,
   rgn_nxt_wdata_i, rgn_byte_wr_i, rgn_attrs_sel_i, ahb_haddri_i, ahb_haddrd_i,
   rgn_ihit_o, rgn_dhit_o, rgn_base_addr_reg_o, rgn_attr_size_reg_o
  );


  parameter     MPU_PRESENT = 0;   // MPU present (0 = not present; 1 = present)


  input         pclk;                 // PPB clock
  input         hclk;                 // AHB clock
  input         hreset_n;             // AHB reset

  input  [31:5] ahb_haddri_i;         // I-AHB address
  input  [31:5] ahb_haddrd_i;         // D-AHB address

  input         mpu_enabled_i;        // Master MPU Enable
  input         rgn_select_wr_i;      // Region Selected for write
  input   [3:0] rgn_byte_wr_i;        // Byte Lane Select for Writes
  input         rgn_base_addr_sel_i;  // Base Address select
  input         rgn_attrs_sel_i;      // Attribute/Size Write Enable
  input  [31:0] rgn_nxt_wdata_i;      // Next Write Data for Base or Attributes

  output        rgn_ihit_o;           // ahb_haddri_i is in this enabled region
  output        rgn_dhit_o;           // ahb_haddrd_i is in this enabled region

  output [31:5] rgn_base_addr_reg_o;  // Base Address register
  output [23:0] rgn_attr_size_reg_o;  // Region attribute and size register


  reg    [31:5] base_addr;
  wire    [3:0] base_addr_en;

  reg           region_en;            // Region enable bit
  reg     [4:0] region_size;          // Region size
  reg     [7:0] region_srd;           // Region Sub Region Disable bits
  wire    [7:0] region_perm_srd;      // Permuted Sub Region Disable bits
  wire          rsize_mod_3_0;        // Region size mod 3, result is 0
  wire          rsize_mod_3_1;        // Region size mod 3, result is 1
  wire          rsize_mod_3_2;        // Region size mod 3, result is 2
  wire    [2:0] rsize_mod_3;          // Combined region size mod 3 result
  reg           region_attrs_xn;      // eXecute Never bit
  reg     [2:0] region_attrs_ap;      // Access permissions field
  reg     [2:0] region_attrs_tex;     // Type EXtension field
  reg     [2:0] region_attrs_scb;     // Shareable, cacheable, bufferable bits
  wire    [3:0] attrs_size_en;

  wire          region_comp_en;       // Region comparator enable
  wire   [31:5] region_mask;          // Region address mask

  wire          rgn_ihit;
  wire          rgn_dhit;

  wire   [31:5] rgn_base_addr_reg;
  wire   [23:0] rgn_attr_size_reg;

  wire          opt_rgn_ihit;
  wire          opt_rgn_dhit;
  wire   [31:5] opt_rgn_base_addr_reg;
  wire   [23:0] opt_rgn_attr_size_reg;
  wire          opt_mst_en;
  wire          opt_mpu_en;

`ifdef ARM_DSM

  wire          dsm_mpu_present;

  assign opt_mpu_en = (dsm_mpu_present)  ? 1'b1 : 1'b0;

`else

  assign opt_mpu_en = (MPU_PRESENT != 0) ? 1'b1 : 1'b0;

`endif

  assign opt_rgn_ihit          = (opt_mpu_en) ? rgn_ihit          : 1'b0;
  assign opt_rgn_dhit          = (opt_mpu_en) ? rgn_dhit          : 1'b0;
  assign opt_rgn_base_addr_reg = (opt_mpu_en) ? rgn_base_addr_reg : {27{1'b0}};
  assign opt_rgn_attr_size_reg = (opt_mpu_en) ? rgn_attr_size_reg : {24{1'b0}};
  assign opt_mst_en            = (opt_mpu_en) ? 1'b1              : 1'b0;


  assign base_addr_en = (rgn_byte_wr_i &
                         {4{(rgn_select_wr_i     &
                             rgn_base_addr_sel_i &
                             opt_mst_en)}});

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      base_addr[31:24] <= 8'b00000000;
    else if (base_addr_en[3])
      base_addr[31:24] <= rgn_nxt_wdata_i[31:24];

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      base_addr[23:16] <= 8'b00000000;
    else if (base_addr_en[2])
      base_addr[23:16] <= rgn_nxt_wdata_i[23:16];

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      base_addr[15:8] <= 8'b00000000;
    else if (base_addr_en[1])
      base_addr[15:8] <= rgn_nxt_wdata_i[15:8];

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      base_addr[7:5] <= 3'b000;
    else if (base_addr_en[0])
      base_addr[7:5] <= rgn_nxt_wdata_i[7:5];


  assign attrs_size_en = (rgn_byte_wr_i &
                          {4{(rgn_select_wr_i &
                              rgn_attrs_sel_i &
                              opt_mst_en)}});

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      begin
        region_en   <= 1'b0;
        region_size <= 5'b00000;
      end
    else if (attrs_size_en[0])
      begin
        region_en   <= rgn_nxt_wdata_i[0];
        region_size <= rgn_nxt_wdata_i[5:1];
      end


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      region_srd <= 8'b00000000;
    else if (attrs_size_en[1])
      region_srd <= rgn_nxt_wdata_i[15:8];


  assign rsize_mod_3_0  = ((region_size[4:0] == 5'b01001) | // 1K
                           (region_size[4:0] == 5'b01100) | // 8K
                           (region_size[4:0] == 5'b01111) | // 64K
                           (region_size[4:0] == 5'b10010) | // 512K
                           (region_size[4:0] == 5'b10101) | // 4M
                           (region_size[4:0] == 5'b11000) | // 32M
                           (region_size[4:0] == 5'b11011) | // 256M
                           (region_size[4:0] == 5'b11110)); // 2G

  assign rsize_mod_3_1  = ((region_size[4:0] == 5'b00111) | // 256B
                           (region_size[4:0] == 5'b01010) | // 2K
                           (region_size[4:0] == 5'b01101) | // 16K
                           (region_size[4:0] == 5'b10000) | // 128K
                           (region_size[4:0] == 5'b10011) | // 1M
                           (region_size[4:0] == 5'b10110) | // 8M
                           (region_size[4:0] == 5'b11001) | // 64M
                           (region_size[4:0] == 5'b11100) | // 512M
                           (region_size[4:0] == 5'b11111)); // 4G

  assign rsize_mod_3_2  = ((region_size[4:0] == 5'b01000) | // 512B
                           (region_size[4:0] == 5'b01011) | // 4K
                           (region_size[4:0] == 5'b01110) | // 32K
                           (region_size[4:0] == 5'b10001) | // 256K
                           (region_size[4:0] == 5'b10100) | // 2M
                           (region_size[4:0] == 5'b10111) | // 16M
                           (region_size[4:0] == 5'b11010) | // 128M
                           (region_size[4:0] == 5'b11101)); // 1G

  assign rsize_mod_3    = { rsize_mod_3_2, rsize_mod_3_1, rsize_mod_3_0 };



  assign region_perm_srd = ((// MPU region size DIV 3, result is 0
                             {8{rsize_mod_3_1}} &
                             {region_srd[7], region_srd[6], region_srd[5],
                              region_srd[4], region_srd[3], region_srd[2],
                              region_srd[1], region_srd[0]})               |
                            (// MPU region size DIV 3, result is 1
                             {8{rsize_mod_3_2}} &
                             {region_srd[7], region_srd[3], region_srd[6],
                              region_srd[2], region_srd[5], region_srd[1],
                              region_srd[4], region_srd[0]})               |
                            (// MPU region size DIV 3, result is 2
                             {8{rsize_mod_3_0}} &
                             {region_srd[7], region_srd[5], region_srd[3],
                              region_srd[1], region_srd[6], region_srd[4],
                              region_srd[2], region_srd[0]}));


  assign region_comp_en = mpu_enabled_i & region_en & (|region_size[4:2]);


  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      begin
        region_attrs_xn <= 1'b0;
        region_attrs_ap <= 3'b000;
      end
    else if (attrs_size_en[3])
      begin
        region_attrs_xn <= rgn_nxt_wdata_i[28];
        region_attrs_ap <= rgn_nxt_wdata_i[26:24];
      end

  always @ (posedge pclk or negedge hreset_n)
    if (!hreset_n)
      begin
        region_attrs_tex <= 3'b000;
        region_attrs_scb <= 3'b000;
      end
    else if (attrs_size_en[2])
      begin
        region_attrs_tex <= rgn_nxt_wdata_i[21:19];
        region_attrs_scb <= rgn_nxt_wdata_i[18:16];
      end


  assign rgn_base_addr_reg = base_addr[31:5];

  assign rgn_attr_size_reg = { region_attrs_xn, region_attrs_ap,
                               region_attrs_tex, region_attrs_scb,
                               region_srd, region_size, region_en };


  cm3_mpu_maskgen #(MPU_PRESENT)
    u_cm3_mpu_maskgen
      (// Inputs
       .region_size_i             (region_size),
       .region_comp_en_i          (region_comp_en),
       .region_mask_o             (region_mask)
      );


  cm3_mpu_comp #(MPU_PRESENT)
    u_cm3_mpu_icomp
      (// Inputs
       .region_comp_en_i          (region_comp_en),
       .region_perm_srd_i         (region_perm_srd),
       .rsize_mod_3_i             (rsize_mod_3),
       .base_addr_i               (base_addr),
       .region_mask_i             (region_mask),
       .ahb_haddr_i               (ahb_haddri_i),
       .region_hit_o              (rgn_ihit)
      );


  cm3_mpu_comp #(MPU_PRESENT)
    u_cm3_mpu_dcomp
      (// Inputs
       .region_comp_en_i          (region_comp_en),
       .region_perm_srd_i         (region_perm_srd),
       .rsize_mod_3_i             (rsize_mod_3),
       .base_addr_i               (base_addr),
       .region_mask_i             (region_mask),
       .ahb_haddr_i               (ahb_haddrd_i),
       .region_hit_o              (rgn_dhit)
      );

  assign rgn_ihit_o          = opt_rgn_ihit;
  assign rgn_dhit_o          = opt_rgn_dhit;
  assign rgn_base_addr_reg_o = opt_rgn_base_addr_reg;
  assign rgn_attr_size_reg_o = opt_rgn_attr_size_reg;


`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Cannot hit D-Side region if not enabled")
    ovl_dside_hit(hclk, hreset_n, (~region_comp_en & rgn_dhit));

  assert_never #(0,0,"Cannot hit I-Side region if not enabled")
    ovl_iside_hit(hclk, hreset_n, (~region_comp_en & rgn_ihit));

`endif

endmodule
