module cm3_mpu_regions
  (// Inputs
   hreset_n, hclk, mpu_enabled_i, rgn_select_oh_i, rgn_select_wr_i,
   rgn_base_addr_sel_i, rgn_nxt_wdata_i, rgn_byte_wr_i, aligned_addrd_i,
   rgn_attrs_sel_i, dpu_ahb_haddri_i, dpu_ahb_haddrd_i, check_alignedd_i, pclk,
   dpu_ahb_hwrited_i, dpu_ahb_hprotd_i, aligned_writed_i, aligned_privd_i,
   dside_hit_o, iside_hit_o, attrsd_o, attrsi_o, rgn_hrdata_o, priv_faultd_o
  );

  parameter     MPU_PRESENT = 0;   // MPU present (0 = not present; 1 = present)


  input         pclk;                  // PPB clock
  input         hclk;                  // AHB clock
  input         hreset_n;              // AHB reset

  input   [7:0] rgn_select_oh_i;       // Region Select (one-hot)
  input   [7:0] rgn_select_wr_i;       // Region Select for writes
  input         rgn_base_addr_sel_i;   // Base Address select
  input         rgn_attrs_sel_i;       // Attribute/Size select
  input  [31:0] rgn_nxt_wdata_i;       // Next Write Data for Base or Attributes
  input   [3:0] rgn_byte_wr_i;         // Byte Lane Select for Writes
  output [31:0] rgn_hrdata_o;          // HRDATA for the selected region

  input  [31:0] dpu_ahb_haddri_i;      // I-AHB address
  input  [31:0] dpu_ahb_haddrd_i;      // D-AHB address
  input   [1:0] dpu_ahb_hprotd_i;      // D-AHB protection in
  input         dpu_ahb_hwrited_i;     // D-AHB write in
  output        iside_hit_o;           // IHADDR hit defined & enabled region
  output        dside_hit_o;           // DHADDR hit defined & enabled region

  input         mpu_enabled_i;         // Master MPU Enable
  input  [31:5] aligned_addrd_i;       // Aligned addr for an unaligned access
  input         check_alignedd_i;      // Need to check an aligned access
  input         aligned_writed_i;      // Aligned write for unaligned access
  input         aligned_privd_i;       // Aligned priv for an unaligned access

  output  [9:0] attrsi_o;              // Attributes for the region hit by IHADDR
  output  [8:0] attrsd_o;              // Attributes for the region hit by DHADDR
  output        priv_faultd_o;         // D-side privilege access violation


  wire   [31:5] rgn_base_addr_min;
  wire   [31:0] rgn_base_addr;
  wire   [23:0] rgn_attr_size_min;
  wire   [31:0] rgn_attr_size;
  wire   [31:0] rgn_hrdata;

  wire          d_user;
  wire          d_write;
  wire   [31:5] ahb_haddrd;
  wire   [31:5] ahb_haddri;
  wire    [8:0] attrsd;
  wire    [9:0] attrsi;
  wire          iside_hit;
  wire          dside_hit;
  wire    [9:0] attrsi_7;
  wire    [9:0] attrsi_6;
  wire    [9:0] attrsi_5;
  wire    [9:0] attrsi_4;
  wire    [9:0] attrsi_3;
  wire    [9:0] attrsi_2;
  wire    [9:0] attrsi_1;
  wire    [9:0] attrsi_0;
  wire    [9:0] attrsi_7_6;
  wire    [9:0] attrsi_5_4;
  wire    [9:0] attrsi_3_2;
  wire    [9:0] attrsi_1_0;
  wire    [9:0] attrsi_7_4;
  wire    [9:0] attrsi_3_0;
  wire    [8:0] attrsd_7;
  wire    [8:0] attrsd_6;
  wire    [8:0] attrsd_5;
  wire    [8:0] attrsd_4;
  wire    [8:0] attrsd_3;
  wire    [8:0] attrsd_2;
  wire    [8:0] attrsd_1;
  wire    [8:0] attrsd_0;
  wire    [8:0] attrsd_7_6;
  wire    [8:0] attrsd_5_4;
  wire    [8:0] attrsd_3_2;
  wire    [8:0] attrsd_1_0;
  wire    [8:0] attrsd_7_4;
  wire    [8:0] attrsd_3_0;
  wire    [2:0] attrd_ap_7;
  wire    [2:0] attrd_ap_6;
  wire    [2:0] attrd_ap_5;
  wire    [2:0] attrd_ap_4;
  wire    [2:0] attrd_ap_3;
  wire    [2:0] attrd_ap_2;
  wire    [2:0] attrd_ap_1;
  wire    [2:0] attrd_ap_0;
  wire          faultd_7;
  wire          faultd_6;
  wire          faultd_5;
  wire          faultd_4;
  wire          faultd_3;
  wire          faultd_2;
  wire          faultd_1;
  wire          faultd_0;
  wire          faultd_7_6;
  wire          faultd_5_4;
  wire          faultd_3_2;
  wire          faultd_1_0;
  wire          faultd_7_4;
  wire          faultd_3_0;
  wire          priv_faultd;

  wire   [31:5] rgn_base_addr_reg_0;
  wire   [31:5] rgn_base_addr_reg_1;
  wire   [31:5] rgn_base_addr_reg_2;
  wire   [31:5] rgn_base_addr_reg_3;
  wire   [31:5] rgn_base_addr_reg_4;
  wire   [31:5] rgn_base_addr_reg_5;
  wire   [31:5] rgn_base_addr_reg_6;
  wire   [31:5] rgn_base_addr_reg_7;
  wire   [23:0] rgn_attr_size_reg_0;
  wire   [23:0] rgn_attr_size_reg_1;
  wire   [23:0] rgn_attr_size_reg_2;
  wire   [23:0] rgn_attr_size_reg_3;
  wire   [23:0] rgn_attr_size_reg_4;
  wire   [23:0] rgn_attr_size_reg_5;
  wire   [23:0] rgn_attr_size_reg_6;
  wire   [23:0] rgn_attr_size_reg_7;
  wire    [7:0] rgn_ihit;
  wire    [7:0] rgn_dhit;

  wire          opt_iside_hit;
  wire          opt_dside_hit;
  wire    [8:0] opt_attrsd;
  wire    [9:0] opt_attrsi;
  wire   [31:0] opt_rgn_hrdata;
  wire          opt_mpu_en;

`ifdef ARM_DSM

  wire          dsm_mpu_present;

  assign opt_mpu_en = (dsm_mpu_present)  ? 1'b1 : 1'b0;

`else

  assign opt_mpu_en = (MPU_PRESENT != 0) ? 1'b1 : 1'b0;

`endif

  assign opt_iside_hit  = (opt_mpu_en) ? iside_hit  : 1'b0;
  assign opt_dside_hit  = (opt_mpu_en) ? dside_hit  : 1'b0;
  assign opt_attrsd     = (opt_mpu_en) ? attrsd     : {9{1'b0}};
  assign opt_attrsi     = (opt_mpu_en) ? attrsi     : {10{1'b0}};
  assign opt_rgn_hrdata = (opt_mpu_en) ? rgn_hrdata : {32{1'b0}};


  assign d_user  = (( check_alignedd_i & ~aligned_privd_i) |
                    (~check_alignedd_i & ~dpu_ahb_hprotd_i[1]));

  assign d_write = (( check_alignedd_i & aligned_writed_i) |
                    (~check_alignedd_i & dpu_ahb_hwrited_i));

  assign ahb_haddrd[31:5] = (({27{ check_alignedd_i}} & aligned_addrd_i[31:5]) |
                             ({27{~check_alignedd_i}} & dpu_ahb_haddrd_i[31:5]));

  assign ahb_haddri[31:5] = dpu_ahb_haddri_i[31:5];

  assign attrsi_7 = rgn_attr_size_reg_7[23:14];
  assign attrsi_6 = rgn_attr_size_reg_6[23:14];
  assign attrsi_5 = rgn_attr_size_reg_5[23:14];
  assign attrsi_4 = rgn_attr_size_reg_4[23:14];
  assign attrsi_3 = rgn_attr_size_reg_3[23:14];
  assign attrsi_2 = rgn_attr_size_reg_2[23:14];
  assign attrsi_1 = rgn_attr_size_reg_1[23:14];
  assign attrsi_0 = rgn_attr_size_reg_0[23:14];


  assign attrsi_7_6 = rgn_ihit[7]      ? attrsi_7   : attrsi_6;
  assign attrsi_5_4 = rgn_ihit[5]      ? attrsi_5   : attrsi_4;
  assign attrsi_3_2 = rgn_ihit[3]      ? attrsi_3   : attrsi_2;
  assign attrsi_1_0 = rgn_ihit[1]      ? attrsi_1   : attrsi_0;

  assign attrsi_7_4 = (|rgn_ihit[7:6]) ? attrsi_7_6 : attrsi_5_4;
  assign attrsi_3_0 = (|rgn_ihit[3:2]) ? attrsi_3_2 : attrsi_1_0;

  assign attrsi     = (|rgn_ihit[7:4]) ? attrsi_7_4 : attrsi_3_0;

  assign iside_hit = (|rgn_ihit);

  assign attrsd_7 = rgn_attr_size_reg_7[22:14];
  assign attrsd_6 = rgn_attr_size_reg_6[22:14];
  assign attrsd_5 = rgn_attr_size_reg_5[22:14];
  assign attrsd_4 = rgn_attr_size_reg_4[22:14];
  assign attrsd_3 = rgn_attr_size_reg_3[22:14];
  assign attrsd_2 = rgn_attr_size_reg_2[22:14];
  assign attrsd_1 = rgn_attr_size_reg_1[22:14];
  assign attrsd_0 = rgn_attr_size_reg_0[22:14];


  assign attrsd_7_6 = rgn_dhit[7]      ? attrsd_7   : attrsd_6;
  assign attrsd_5_4 = rgn_dhit[5]      ? attrsd_5   : attrsd_4;
  assign attrsd_3_2 = rgn_dhit[3]      ? attrsd_3   : attrsd_2;
  assign attrsd_1_0 = rgn_dhit[1]      ? attrsd_1   : attrsd_0;

  assign attrsd_7_4 = (|rgn_dhit[7:6]) ? attrsd_7_6 : attrsd_5_4;
  assign attrsd_3_0 = (|rgn_dhit[3:2]) ? attrsd_3_2 : attrsd_1_0;

  assign attrsd     = (|rgn_dhit[7:4]) ? attrsd_7_4 : attrsd_3_0;

  assign attrd_ap_7 = attrsd_7[8:6];
  assign attrd_ap_6 = attrsd_6[8:6];
  assign attrd_ap_5 = attrsd_5[8:6];
  assign attrd_ap_4 = attrsd_4[8:6];
  assign attrd_ap_3 = attrsd_3[8:6];
  assign attrd_ap_2 = attrsd_2[8:6];
  assign attrd_ap_1 = attrsd_1[8:6];
  assign attrd_ap_0 = attrsd_0[8:6];

  assign faultd_7 = ((d_write & attrd_ap_7[2] & (|attrd_ap_7[1:0]))   |
                     (attrd_ap_7 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_7[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_7 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_7 == 3'b010)));

  assign faultd_6 = ((d_write & attrd_ap_6[2] & (|attrd_ap_6[1:0]))   |
                     (attrd_ap_6 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_6[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_6 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_6 == 3'b010)));

  assign faultd_5 = ((d_write & attrd_ap_5[2] & (|attrd_ap_5[1:0]))   |
                     (attrd_ap_5 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_5[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_5 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_5 == 3'b010)));

  assign faultd_4 = ((d_write & attrd_ap_4[2] & (|attrd_ap_4[1:0]))   |
                     (attrd_ap_4 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_4[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_4 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_4 == 3'b010)));

  assign faultd_3 = ((d_write & attrd_ap_3[2] & (|attrd_ap_3[1:0]))   |
                     (attrd_ap_3 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_3[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_3 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_3 == 3'b010)));

  assign faultd_2 = ((d_write & attrd_ap_2[2] & (|attrd_ap_2[1:0]))   |
                     (attrd_ap_2 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_2[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_2 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_2 == 3'b010)));

  assign faultd_1 = ((d_write & attrd_ap_1[2] & (|attrd_ap_1[1:0]))   |
                     (attrd_ap_1 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_1[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_1 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_1 == 3'b010)));

  assign faultd_0 = ((d_write & attrd_ap_0[2] & (|attrd_ap_0[1:0]))   |
                     (attrd_ap_0 == 3'b000)                           |
                     (d_user & ~d_write & (attrd_ap_0[1:0] == 2'b01)) |
                     (d_user & (attrd_ap_0 == 3'b001))                |
                     (d_user & d_write & (attrd_ap_0 == 3'b010)));


  assign faultd_7_6  = rgn_dhit[7]      ? faultd_7   : faultd_6;
  assign faultd_5_4  = rgn_dhit[5]      ? faultd_5   : faultd_4;
  assign faultd_3_2  = rgn_dhit[3]      ? faultd_3   : faultd_2;
  assign faultd_1_0  = rgn_dhit[1]      ? faultd_1   : faultd_0;

  assign faultd_7_4  = (|rgn_dhit[7:6]) ? faultd_7_6 : faultd_5_4;
  assign faultd_3_0  = (|rgn_dhit[3:2]) ? faultd_3_2 : faultd_1_0;

  assign priv_faultd = (|rgn_dhit[7:4]) ? faultd_7_4 : faultd_3_0;

  assign dside_hit   = (|rgn_dhit);


  assign rgn_attr_size_min = (({24{rgn_select_oh_i[0]}} & rgn_attr_size_reg_0) |
                              ({24{rgn_select_oh_i[1]}} & rgn_attr_size_reg_1) |
                              ({24{rgn_select_oh_i[2]}} & rgn_attr_size_reg_2) |
                              ({24{rgn_select_oh_i[3]}} & rgn_attr_size_reg_3) |
                              ({24{rgn_select_oh_i[4]}} & rgn_attr_size_reg_4) |
                              ({24{rgn_select_oh_i[5]}} & rgn_attr_size_reg_5) |
                              ({24{rgn_select_oh_i[6]}} & rgn_attr_size_reg_6) |
                              ({24{rgn_select_oh_i[7]}} & rgn_attr_size_reg_7));

  assign rgn_attr_size     = { 3'b000, rgn_attr_size_min[23],
                               1'b0,   rgn_attr_size_min[22:20],
                               2'b00,  rgn_attr_size_min[19:6],
                               2'b00,  rgn_attr_size_min[5:0] };

  assign rgn_base_addr_min = (({27{rgn_select_oh_i[0]}} & rgn_base_addr_reg_0) |
                              ({27{rgn_select_oh_i[1]}} & rgn_base_addr_reg_1) |
                              ({27{rgn_select_oh_i[2]}} & rgn_base_addr_reg_2) |
                              ({27{rgn_select_oh_i[3]}} & rgn_base_addr_reg_3) |
                              ({27{rgn_select_oh_i[4]}} & rgn_base_addr_reg_4) |
                              ({27{rgn_select_oh_i[5]}} & rgn_base_addr_reg_5) |
                              ({27{rgn_select_oh_i[6]}} & rgn_base_addr_reg_6) |
                              ({27{rgn_select_oh_i[7]}} & rgn_base_addr_reg_7));

  assign rgn_base_addr     = { rgn_base_addr_min, 5'b00000 };

  assign rgn_hrdata        = (({32{ rgn_base_addr_sel_i}} & rgn_base_addr) |
                              ({32{~rgn_base_addr_sel_i}} & rgn_attr_size));


  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_0
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[0]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[0]),
       .rgn_dhit_o                (rgn_dhit[0]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_0),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_0)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_1
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[1]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[1]),
       .rgn_dhit_o                (rgn_dhit[1]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_1),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_1)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_2
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[2]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[2]),
       .rgn_dhit_o                (rgn_dhit[2]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_2),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_2)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_3
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[3]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[3]),
       .rgn_dhit_o                (rgn_dhit[3]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_3),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_3)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_4
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[4]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[4]),
       .rgn_dhit_o                (rgn_dhit[4]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_4),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_4)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_5
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[5]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[5]),
       .rgn_dhit_o                (rgn_dhit[5]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_5),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_5)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_6
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[6]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[6]),
       .rgn_dhit_o                (rgn_dhit[6]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_6),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_6)
      );

  cm3_mpu_region #(MPU_PRESENT)
    u_cm3_mpu_region_7
      (// Inputs
       .hreset_n                  (hreset_n),
       .pclk                      (pclk),
       .hclk                      (hclk),
       .mpu_enabled_i             (mpu_enabled_i),
       .rgn_select_wr_i           (rgn_select_wr_i[7]),
       .rgn_base_addr_sel_i       (rgn_base_addr_sel_i),
       .rgn_nxt_wdata_i           (rgn_nxt_wdata_i),
       .rgn_attrs_sel_i           (rgn_attrs_sel_i),
       .rgn_byte_wr_i             (rgn_byte_wr_i),
       .ahb_haddri_i              (ahb_haddri),
       .ahb_haddrd_i              (ahb_haddrd),
       .rgn_ihit_o                (rgn_ihit[7]),
       .rgn_dhit_o                (rgn_dhit[7]),
       .rgn_base_addr_reg_o       (rgn_base_addr_reg_7),
       .rgn_attr_size_reg_o       (rgn_attr_size_reg_7)
      );

  assign iside_hit_o   = opt_iside_hit;
  assign dside_hit_o   = opt_dside_hit;
  assign attrsd_o      = opt_attrsd;
  assign attrsi_o      = opt_attrsi;
  assign rgn_hrdata_o  = opt_rgn_hrdata;
  assign priv_faultd_o = priv_faultd;


`ifdef ARM_ASSERT_ON

  assert_never #(0,0,"Canot hit a d-side region if MPU is not enabled.")
    ovl_dside_hit(hclk, hreset_n, (dside_hit & ~mpu_enabled_i));

  assert_never #(0,0,"Canot hit an i-side region if MPU is not enabled.")
    ovl_iside_hit(hclk, hreset_n, (iside_hit & ~mpu_enabled_i));

`endif

endmodule
