`include "cm3_mtx_defs.vp"

module cm3_mtx_decode_dap
  (// Inputs
   hclk, dreset_n, mtx_dap_ahb_hreadyout_i, dpu_dbg_snapstall_i, is_dap_addr_i,
   is_dap_unalign_i, is_dap_held_trans_i, os_dap_to_dcode_accept_i,
   os_dap_to_dcode_active_i, mtx_ahb_hreadyd_i, mtx_ahb_hrespd_i,
   mtx_ahb_hrdatad_i, os_dap_to_sys_accept_i, os_dap_to_sys_active_i,
   mtx_ahb_hreadys_i, mtx_ahb_hresps_i, mtx_ahb_hrdatas_i,
   os_dap_to_ppb_accept_i, os_dap_to_ppb_active_i, mtx_ppb_hready_i,
   mtx_ppb_hresp_i, mtx_ppb_hrdata_i, bm_state_bb_wr2_i, bm_dap_active_i,
   bm_dap_active_sel_i, bm_dap_active_data_i, bm_dap_stall_i, bm_readyout_i,
   bm_resp_i, bm_rdata_i,
   dec_dap_sel_dcode_o, dec_dap_sel_sys_o, dec_dap_sel_ppb_o,
   dec_dap_sel_bb_o, dec_dap_sel_dcode_held_o, dec_dap_sel_sys_held_o,
   dec_dap_sel_ppb_held_o, dec_dap_accept_o, dec_dap_active_o,
   dec_dap_readyout_o, dec_dap_resp_o, mtx_dap_ahb_hrdata_o
  );


  parameter     BB_PRESENT = 1;             // Bit banding present if 1


  input         hclk;                       // AHB system clock
  input         dreset_n;                   // Debug reset

  input  [31:0] is_dap_addr_i;              // DAP HADDR
  input   [2:0] is_dap_unalign_i;           // Unaligned transaction encoding
  input         is_dap_held_trans_i;        // DAP transaction held
  input         mtx_dap_ahb_hreadyout_i;    // HREADY output for DAP port

  input         dpu_dbg_snapstall_i;

  input         os_dap_to_dcode_accept_i;   // DAP -> DCode accepted
  input         os_dap_to_dcode_active_i;   // DAP -> DCode active
  input         mtx_ahb_hreadyd_i;
  input   [1:0] mtx_ahb_hrespd_i;
  input  [31:0] mtx_ahb_hrdatad_i;

  input         os_dap_to_sys_accept_i;     // DAP -> SYS accepted
  input         os_dap_to_sys_active_i;     // DAP -> SYS active
  input         mtx_ahb_hreadys_i;
  input   [1:0] mtx_ahb_hresps_i;
  input  [31:0] mtx_ahb_hrdatas_i;

  input         os_dap_to_ppb_accept_i;     // DAP -> PPB accepted
  input         os_dap_to_ppb_active_i;     // DAP -> PPB active
  input         mtx_ppb_hready_i;
  input   [1:0] mtx_ppb_hresp_i;
  input  [31:0] mtx_ppb_hrdata_i;

  input         bm_state_bb_wr2_i;
  input         bm_dap_active_i;
  input   [1:0] bm_dap_active_sel_i;
  input         bm_dap_active_data_i;
  input         bm_dap_stall_i;
  input         bm_readyout_i;
  input         bm_resp_i;
  input  [31:0] bm_rdata_i;

  output        dec_dap_sel_dcode_o;        // DAP select
  output        dec_dap_sel_dcode_held_o;   // DAP select held
  output        dec_dap_sel_sys_o;          // SYS select
  output        dec_dap_sel_sys_held_o;     // SYS select held
  output        dec_dap_sel_ppb_o;          // PPB select
  output        dec_dap_sel_ppb_held_o;     // PPB select held
  output        dec_dap_sel_bb_o;           // Bit-band select (1st cycle)

  output        dec_dap_accept_o;           // Transaction accepted from DAP
  output        dec_dap_active_o;           // Transaction active from DAP
  output        dec_dap_readyout_o;         // HREADY from output stage
  output  [1:0] dec_dap_resp_o;             // HRESP from output stage
  output [31:0] mtx_dap_ahb_hrdata_o;       // HRDATA from output stage



  wire          sel_dcode;                  // DAP address in Code space
  wire          sel_sys;                    // DAP address in System space
  wire          sel_ppb;                    // DAP address in PPB space
  wire          sel_bb;                     // Bit-band access
  wire          mst_bb_en;                  // Logic removal for bit-band
  wire    [2:0] output_port_addr;           // One-hot output port selection

  wire          sel_dcode_held;             // DAP -> DCode held
  wire          sel_sys_held;               // DAP -> SYS held
  wire          sel_ppb_held;               // DAP -> PPB held
  wire          dap_accept;                 // DAP trans accepted by output stage
  wire          dap_active;                 // DAP trans active in output stage
  reg           dap_to_dcode_accept_reg;    // Registered DAP -> DCode accepted
  reg           dap_to_sys_accept_reg;      // Registered DAP -> SYS accepted
  reg           dap_to_ppb_accept_reg;      // Registered DAP -> PPB accepted
  wire          dap_accept_reg_we;          // Registered accept write enable

  wire          nxt_sel_bm_data;            // Next data phase to Bit Master
  wire    [3:0] nxt_output_port_data;
  reg     [3:0] output_port_data;           // Data phase output port selection

  wire   [31:0] dap_rdata;                  // DAP HRDATA mux
  wire    [1:0] dap_resp;                   // DAP HRESP mux
  wire          dap_readyout;               // DAP HREADYOUT mux

`ifdef ARM_DSM

  wire          dsm_bb_present;

  assign mst_bb_en = (dsm_bb_present) ? 1'b1 : 1'b0;

`else

  assign mst_bb_en = (BB_PRESENT == 1) ? 1'b1 : 1'b0;

`endif



  assign sel_dcode = ~(|is_dap_addr_i[31:29]) & is_dap_held_trans_i;
  assign sel_sys   = (is_dap_addr_i[31:20] != 12'hE00) &
                     (|is_dap_addr_i[31:29])           &
                     is_dap_held_trans_i;
  assign sel_ppb   = ~sel_dcode & ~sel_sys;

  assign output_port_addr = {sel_ppb, sel_sys, sel_dcode};

  assign sel_bb = ({is_dap_addr_i[31], is_dap_addr_i[28:25]} == 5'b00001) &
                  (is_dap_addr_i[30] ^ is_dap_addr_i[29]) & mst_bb_en;




  assign sel_dcode_held = dap_to_sys_accept_reg  |
                          dap_to_ppb_accept_reg  |
                          bm_dap_active_i        |
                          bm_dap_active_sel_i[1] |
                          bm_dap_stall_i;

  assign sel_sys_held   = dap_to_dcode_accept_reg |
                          dap_to_ppb_accept_reg   |
                          bm_dap_active_i         |
                          bm_dap_active_sel_i[0]  |
                          bm_dap_stall_i          |
                          bm_state_bb_wr2_i;

  assign sel_ppb_held   = dap_to_dcode_accept_reg |
                          dap_to_sys_accept_reg   |
                          bm_dap_active_i         |
                          bm_dap_active_sel_i[0]  |
                          bm_dap_active_sel_i[1];

  assign dap_accept = (output_port_addr[0] & os_dap_to_dcode_accept_i) |
                      (output_port_addr[1] & os_dap_to_sys_accept_i)   |
                      (output_port_addr[2] & os_dap_to_ppb_accept_i);

  assign dap_active = (output_port_addr[0] & os_dap_to_dcode_active_i) |
                      (output_port_addr[1] & os_dap_to_sys_active_i)   |
                      (output_port_addr[2] & os_dap_to_ppb_active_i);

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      dap_to_dcode_accept_reg <= 1'b0;
    else if (dap_accept_reg_we)
      dap_to_dcode_accept_reg <= os_dap_to_dcode_accept_i;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      dap_to_sys_accept_reg <= 1'b0;
    else if (dap_accept_reg_we)
      dap_to_sys_accept_reg <= os_dap_to_sys_accept_i;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      dap_to_ppb_accept_reg <= 1'b0;
    else if (dap_accept_reg_we)
      dap_to_ppb_accept_reg <= os_dap_to_ppb_accept_i;

  assign dap_accept_reg_we = mtx_dap_ahb_hreadyout_i | dpu_dbg_snapstall_i;




  assign nxt_sel_bm_data = ((is_dap_unalign_i != `MTX_UNALIGN_NORMAL) &
                              ~output_port_addr[2]) |
                           sel_bb                   |
                           bm_dap_active_data_i;

  always @ (posedge hclk or negedge dreset_n)
      if (!dreset_n)
        output_port_data <= 4'b0000;
      else if (dap_accept_reg_we)
        output_port_data <= nxt_output_port_data;

  assign nxt_output_port_data = nxt_sel_bm_data ? 4'b1000 : {1'b0, output_port_addr};



  assign dap_readyout = (output_port_data[0] & mtx_ahb_hreadyd_i) |
                        (output_port_data[1] & mtx_ahb_hreadys_i) |
                        (output_port_data[2] & mtx_ppb_hready_i)  |
                        (output_port_data[3] & bm_readyout_i);

  assign dap_resp     = ({2{output_port_data[0]}} & mtx_ahb_hrespd_i) |
                        ({2{output_port_data[1]}} & mtx_ahb_hresps_i) |
                        ({2{output_port_data[2]}} & mtx_ppb_hresp_i)  |
                        ({2{output_port_data[3]}} & {1'b0, bm_resp_i});

  assign dap_rdata    = ({32{output_port_data[0]}} & mtx_ahb_hrdatad_i) |
                        ({32{output_port_data[1]}} & mtx_ahb_hrdatas_i) |
                        ({32{output_port_data[2]}} & mtx_ppb_hrdata_i)  |
                        ({32{output_port_data[3]}} & bm_rdata_i);



  assign dec_dap_sel_dcode_o       = sel_dcode;
  assign dec_dap_sel_sys_o         = sel_sys;
  assign dec_dap_sel_ppb_o         = sel_ppb;
  assign dec_dap_sel_bb_o          = sel_bb;
  assign dec_dap_sel_dcode_held_o  = sel_dcode_held;
  assign dec_dap_sel_sys_held_o    = sel_sys_held;
  assign dec_dap_sel_ppb_held_o    = sel_ppb_held;

  assign dec_dap_accept_o          = dap_accept;
  assign dec_dap_active_o          = dap_active;

  assign dec_dap_readyout_o        = dap_readyout;
  assign dec_dap_resp_o            = dap_resp;
  assign mtx_dap_ahb_hrdata_o      = dap_rdata;



`ifdef ARM_ASSERT_ON

  assert_implication #(0,0,"Code space decode error.")
   ovl_code_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'h00000000) && (is_dap_addr_i <= 32'h1FFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0001));

  assert_implication #(0,0,"SRAM space decode error.")
   ovl_sram_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'h20000000) && (is_dap_addr_i <= 32'h3FFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0010));

  assert_implication #(0,0,"Peripheral space decode error.")
   ovl_peripheral_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'h40000000) && (is_dap_addr_i <= 32'h5FFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0010));

  assert_implication #(0,0,"External RAM decode error.")
   ovl_external_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'h60000000) && (is_dap_addr_i <= 32'h9FFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0010));

  assert_implication #(0,0,"External Device decode error.")
   ovl_external_device_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'hA0000000) && (is_dap_addr_i <= 32'hDFFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0010));

  assert_implication #(0,0,"PPB decode error.")
   ovl_ppb_mem(hclk, dreset_n,
     (((is_dap_addr_i >= 32'hE0000000) && (is_dap_addr_i <= 32'hE00FFFFF)) || ~is_dap_held_trans_i),
      (output_port_addr == 4'b0100));

  assert_implication #(0,0,"System space decode error.")
   ovl_system_mem(hclk, dreset_n,
     ((is_dap_addr_i >= 32'hE0100000) && (is_dap_addr_i <= 32'hFFFFFFFF) && is_dap_held_trans_i),
       (output_port_addr == 4'b0010));


  assert_zero_one_hot #(0,3,0,"Output stage selection must be exclusive.")
    ovl_output_stage_sel(hclk, dreset_n, {sel_ppb, sel_sys, sel_dcode});


  assert_zero_one_hot #(0,3,0,"output_port_addr must be exclusive.")
    ovl_output_port_addr_exclusive(hclk, dreset_n, output_port_addr);

  assert_never_unknown #(0,3,0,"output_port_addr goes to X.")
    ovl_output_port_addr_x(hclk, dreset_n, 1'b1, output_port_addr);

  assert_zero_one_hot #(0,4,0,"output_port_data must be exclusive.")
    ovl_output_port_data_exclusive(hclk, dreset_n, output_port_data);

  assert_never_unknown #(0,4,0,"output_port_data goes to X.")
    ovl_output_port_data_x(hclk, dreset_n, 1'b1, output_port_data);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
