`include "cm3_mtx_defs.vp"

module cm3_mtx_decode_icore
  (// Inputs
   hclk, hreset_n, mtx_mpu_ahb_hreadyouti_i, fpb_remap_func_i, fpb_matchedi_i,
   d_not_i_trans_i, mtx_ahb_htransd_i, is_icore_addr_i,
   os_icore_to_icode_accept_i, os_icore_to_icode_active_i, mtx_ahb_hreadyi_i,
   mtx_ahb_hrespi_i, mtx_ahb_hrdatai_i, mtx_ahb_hreadys_i, mtx_ahb_hresps_i,
   mtx_ahb_hrdatas_i, bm_icore_fpb_stall_i, bm_icore_active_i,
   bm_icore_active_sel_i, bm_icore_fpb_active_i, bm_icore_stall_i,
   bm_readyout_i, bm_resp_i, bm_rdata_i,
   dec_icore_sel_icode_o, dec_icore_sel_sys_o, dec_icore_sel_icode_held_o,
   dec_icore_sel_sys_held_o, dec_icore_accept_o, dec_icore_active_o,
   dec_icore_readyout_o, dec_icore_resp_o, mtx_dpu_ahb_hrdatai_o
  );


  input         hclk;                        // AHB system clock
  input         hreset_n;                    // AHB system reset

  input  [31:0] is_icore_addr_i;             // ICore HADDR
  input         mtx_mpu_ahb_hreadyouti_i;    // HREADY output for ICore port

  input   [1:0] fpb_remap_func_i;
  input         fpb_matchedi_i;

  input         d_not_i_trans_i;             // DNOTITRANS
  input   [1:0] mtx_ahb_htransd_i;           // Current DCode trans status

  input         os_icore_to_icode_accept_i;  // ICore -> ICode accepted
  input         os_icore_to_icode_active_i;  // ICore -> ICode active
  input         mtx_ahb_hreadyi_i;
  input   [1:0] mtx_ahb_hrespi_i;
  input  [31:0] mtx_ahb_hrdatai_i;

  input         mtx_ahb_hreadys_i;
  input   [1:0] mtx_ahb_hresps_i;
  input  [31:0] mtx_ahb_hrdatas_i;

  input         bm_icore_active_i;
  input         bm_icore_active_sel_i;
  input         bm_icore_stall_i;
  input         bm_icore_fpb_active_i;
  input         bm_icore_fpb_stall_i;
  input         bm_readyout_i;
  input         bm_resp_i;
  input  [31:0] bm_rdata_i;

  output        dec_icore_sel_icode_o;       // ICode select
  output        dec_icore_sel_sys_o;         // SYS select
  output        dec_icore_sel_icode_held_o;  // ICode select held
  output        dec_icore_sel_sys_held_o;    // SYS select held

  output        dec_icore_accept_o;          // Transaction accepted from ICore
  output        dec_icore_active_o;          // Transaction active from ICore
  output        dec_icore_readyout_o;        // HREADY from output stage
  output  [1:0] dec_icore_resp_o;            // HRESP from output stage
  output [31:0] mtx_dpu_ahb_hrdatai_o;       // HRDATA from output stage



  wire          sel_icode;                  // ICore address in Code space
  wire          sel_sys;                    // ICore address in System space
  wire          sel_xn;                     // ICore address in fixed XN region
  wire    [2:0] output_port_addr;           // One-hot output port selection

  wire          suppress_i_trans;           // Suppress ICode HTRANS
  wire          sel_icode_held;             // ICore -> ICode held
  wire          sel_sys_held;               // ICore -> SYS held
  wire          icore_accept;               // ICore trans accepted by output stage
  wire          icore_active;               // ICore trans active in output stage
  reg           icore_to_icode_accept_reg;  // Registered ICore -> ICode accepted

  reg     [2:0] output_port_addr_reg;       // Data phase of output_port_addr
  reg     [4:0] output_port_data;           // Data phase output port selection
  wire          no_port_data;               // No port selected in data phase

  wire          bkpt_remap_lo;              // Remap BKPT instr to lower halfword
  wire          bkpt_remap_hi;              // Remap BKPT instr to upper halfword
  wire          bkpt_remap_all;             // Remap BKPT instr to both halfwords
  wire    [1:0] bkpt_resp_mux;              // BKPT insertion HRESP mux
  wire   [31:0] bkpt_rdata_mux;             // BKPT insertion HRDATA mux
  wire          icore_readyout;             // ICore HREADYOUT mux
  wire    [1:0] icore_resp;                 // ICore HRESP mux
  wire   [31:0] icore_rdata;                // ICore HRDATA




  assign sel_icode = ~(|is_icore_addr_i[31:29]); // 0x00000000 - 0x1FFFFFFF
  assign sel_xn    = &is_icore_addr_i[31:29];    // 0xE0000000 - 0xFFFFFFFF
  assign sel_sys   = ~sel_icode & ~sel_xn;       // 0x20000000 - 0xDFFFFFFF

  assign output_port_addr = {sel_xn, sel_sys, sel_icode};




  assign suppress_i_trans = mtx_ahb_htransd_i[1] & d_not_i_trans_i;



  assign sel_icode_held = suppress_i_trans     |
                          bm_icore_fpb_stall_i |
                          bm_icore_active_i    |
                          bm_icore_active_sel_i;

  assign sel_sys_held   = icore_to_icode_accept_reg |
                          bm_icore_fpb_stall_i      |
                          bm_icore_active_i         |
                          bm_icore_stall_i;


  assign icore_accept = output_port_addr[0] & os_icore_to_icode_accept_i;


  assign icore_active = (output_port_addr[0] & os_icore_to_icode_active_i) |
                        (output_port_addr[1] &
                          (~bm_icore_stall_i & ~icore_to_icode_accept_reg & bm_readyout_i)) |
                        (output_port_addr[2]);

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      icore_to_icode_accept_reg <= 1'b0;
    else if (mtx_mpu_ahb_hreadyouti_i)
      icore_to_icode_accept_reg <= os_icore_to_icode_accept_i;



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      output_port_addr_reg <= 3'b000;
    else if (mtx_mpu_ahb_hreadyouti_i)
      output_port_addr_reg <= output_port_addr;


  always @ (bm_icore_fpb_stall_i or bm_icore_fpb_active_i or
            output_port_addr_reg)
    case ({bm_icore_fpb_stall_i, bm_icore_fpb_active_i})
      2'b10, 2'b11: output_port_data = {5'b01_000};
      2'b01:        output_port_data = {5'b10_000};
      2'b00:        output_port_data = {2'b00, output_port_addr_reg};
      default:      output_port_data = 5'bXXXXX;
    endcase

  assign no_port_data = ~(|output_port_data);



  assign bkpt_remap_lo  = fpb_remap_func_i[0] & fpb_matchedi_i;
  assign bkpt_remap_hi  = fpb_remap_func_i[1] & fpb_matchedi_i;
  assign bkpt_remap_all = bkpt_remap_lo & bkpt_remap_hi;

  assign bkpt_resp_mux         = bkpt_remap_all ? `MTX_RSP_OKAY : mtx_ahb_hrespi_i;
  assign bkpt_rdata_mux[31:16] = bkpt_remap_hi  ? 16'hBE00 : mtx_ahb_hrdatai_i[31:16];
  assign bkpt_rdata_mux[15:0]  = bkpt_remap_lo  ? 16'hBE00 : mtx_ahb_hrdatai_i[15:0];


  assign icore_readyout = (output_port_data[0] & mtx_ahb_hreadyi_i) |
                          (output_port_data[1] & mtx_ahb_hreadys_i) |
                          (output_port_data[2] & 1'b1)              |
                          (output_port_data[4] & bm_readyout_i)     |
                          (no_port_data        & mtx_ahb_hreadyi_i);

  assign icore_resp     = ({2{output_port_data[0]}} & bkpt_resp_mux)     |
                          ({2{output_port_data[1]}} & mtx_ahb_hresps_i)  |
                          ({2{output_port_data[2]}} & `MTX_RSP_OKAY)     |
                          ({2{output_port_data[3]}} & `MTX_RSP_OKAY)     |
                          ({2{output_port_data[4]}} & {1'b0, bm_resp_i}) |
                          ({2{no_port_data}}        & mtx_ahb_hrespi_i);

  assign icore_rdata    = ({32{output_port_data[0]}} & bkpt_rdata_mux)    |
                          ({32{output_port_data[1]}} & mtx_ahb_hrdatas_i) |
                          ({32{output_port_data[2]}} & mtx_ahb_hrdatai_i) |
                          ({32{output_port_data[3]}} & mtx_ahb_hrdatai_i) |
                          ({32{output_port_data[4]}} & bm_rdata_i);



   assign dec_icore_sel_icode_o       = sel_icode;
   assign dec_icore_sel_sys_o         = sel_sys;
   assign dec_icore_sel_icode_held_o  = sel_icode_held;
   assign dec_icore_sel_sys_held_o    = sel_sys_held;

   assign dec_icore_accept_o          = icore_accept;
   assign dec_icore_active_o          = icore_active;

   assign dec_icore_readyout_o        = icore_readyout;
   assign dec_icore_resp_o            = icore_resp;
   assign mtx_dpu_ahb_hrdatai_o       = icore_rdata;



`ifdef ARM_ASSERT_ON

  assert_implication #(0,0,
    "output_port_data must be ICore when fpb_remap_func_i != 2'b00 & fpb_matchedi_i.")
    ovl_remap_check(hclk, hreset_n, (fpb_remap_func_i != 2'b00 & fpb_matchedi_i),
       (output_port_data == 5'b00001));


  assert_implication #(0,0,"Code space decode error.")
    ovl_code_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'h00000000) && (is_icore_addr_i <= 32'h1FFFFFFF)),
        (output_port_addr == 5'b00001));

  assert_implication #(0,0,"SRAM space decode error.")
    ovl_sram_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'h20000000) && (is_icore_addr_i <= 32'h3FFFFFFF)),
        (output_port_addr == 5'b00010));

  assert_implication #(0,0,"Peripheral space decode error.")
    ovl_peripheral_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'h40000000) && (is_icore_addr_i <= 32'h5FFFFFFF)),
        (output_port_addr == 5'b00010));

  assert_implication #(0,0,"External RAM decode error.")
    ovl_external_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'h60000000) && (is_icore_addr_i <= 32'h9FFFFFFF)),
        (output_port_addr == 5'b00010));

  assert_implication #(0,0,"External Device decode error.")
    ovl_external_device_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'hA0000000) && (is_icore_addr_i <= 32'hDFFFFFFF)),
        (output_port_addr == 5'b00010));

  assert_implication #(0,0,"PPB decode error.")
    ovl_ppb_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'hE0000000) && (is_icore_addr_i <= 32'hE00FFFFF)),
        (output_port_addr == 5'b00100));

  assert_implication #(0,0,"System space decode error.")
    ovl_system_mem(hclk, hreset_n,
      ((is_icore_addr_i >= 32'hE0100000) && (is_icore_addr_i <= 32'hFFFFFFFF)),
        (output_port_addr == 5'b00100));


  assert_zero_one_hot #(0,2,0,"Output stage selection must be exclusive.")
    ovl_output_stage_sel(hclk, hreset_n, {sel_sys, sel_icode});


  assert_zero_one_hot #(0,3,0,"output_port_addr must be exclusive.")
    ovl_output_port_addr_exclusive(hclk, hreset_n, output_port_addr);

  assert_never_unknown #(0,3,0,"output_port_addr goes to X.")
    ovl_output_port_addr_x(hclk, hreset_n, 1'b1, output_port_addr);

  assert_zero_one_hot #(0,5,0,"output_port_data must be exclusive.")
    ovl_output_port_data_exclusive(hclk, hreset_n, output_port_data);

  assert_never_unknown #(0,5,0,"output_port_data goes to X.")
    ovl_output_port_data_x(hclk, hreset_n, 1'b1, output_port_data);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
