`include "cm3_mtx_defs.vp"

module cm3_mtx_input_stage_dap
  (// Inputs
   hclk, dreset_n, dap_ahb_hmaster_i, dap_ahb_haddr_i, dap_ahb_htrans_i,
   dap_ahb_hwrite_i, dap_ahb_hsize_i, dap_ahb_hburst_i, dap_ahb_hprot_i,
   dap_ahb_habort_i, dec_dap_accept_i, dec_dap_active_i, dec_dap_readyout_i,
   dec_dap_resp_i,
   mtx_dap_ahb_hreadyout_o, mtx_dap_ahb_haddracc_o, mtx_dap_ahb_hresp_o,
   is_dap_master_o, is_dap_addr_o, is_dap_trans_o, is_dap_write_o,
   is_dap_size_o, is_dap_size_unalign_o, is_dap_burst_o, is_dap_prot_o,
   is_dap_held_trans_o, is_dap_unalign_o
  );


  parameter DEBUG_LVL = 3;



  input         hclk;                     // AHB system clock
  input         dreset_n;                 // Debug reset

  input   [1:0] dap_ahb_hmaster_i;
  input  [31:0] dap_ahb_haddr_i;
  input   [1:0] dap_ahb_htrans_i;
  input         dap_ahb_hwrite_i;
  input   [1:0] dap_ahb_hsize_i;
  input   [2:0] dap_ahb_hburst_i;
  input   [3:0] dap_ahb_hprot_i;
  input         dap_ahb_habort_i;
  output        mtx_dap_ahb_hreadyout_o;
  output        mtx_dap_ahb_haddracc_o;
  output  [1:0] mtx_dap_ahb_hresp_o;

  input         dec_dap_accept_i;         // DAP transfer accepted
  input         dec_dap_active_i;         // DAP transfer active
  input         dec_dap_readyout_i;       // HREADYOUT from selected slave
  input   [1:0] dec_dap_resp_i;           // HRESP from selected slave

  output [31:0] is_dap_addr_o;            // HADDR
  output  [1:0] is_dap_master_o;          // HMASTER
  output  [1:0] is_dap_trans_o;           // HTRANS
  output        is_dap_write_o;           // HWRITE
  output  [1:0] is_dap_size_o;            // HSIZE
  output  [1:0] is_dap_size_unalign_o;    // HSIZE modified for unaligned access
  output  [2:0] is_dap_burst_o;           // HBURST
  output  [3:0] is_dap_prot_o;            // HPROT
  output        is_dap_held_trans_o;      // Holding register active
  output  [2:0] is_dap_unalign_o;         // Unaligned access encoding



  wire          addr_valid;           // Address phase of valid transfer
  reg           data_valid;           // Data phase of valid transfer
  wire          nxt_data_valid;
  wire          data_valid_we;
  wire          data_valid_split_en;
  wire          nxt_split_trans;
  reg           split_trans;

  wire    [1:0] is_icore_trans;
  wire          is_dap_held_trans;
  reg     [1:0] is_dap_size_unalign;
  reg     [2:0] is_dap_unalign;

  wire          ahb_hreadyout;
  wire    [1:0] ahb_hresp;
  wire          ahb_haddracc;

  wire          opt_ahb_hreadyout;
  wire          opt_ahb_haddracc;
  wire    [1:0] opt_ahb_hresp;
  wire    [1:0] opt_master;
  wire   [31:0] opt_addr;
  wire    [1:0] opt_trans;
  wire          opt_write;
  wire    [1:0] opt_size;
  wire    [1:0] opt_size_unalign;
  wire    [2:0] opt_burst;
  wire    [3:0] opt_prot;
  wire          opt_held_trans;
  wire    [2:0] opt_unalign;
  wire          opt_mst_en;



`ifdef ARM_DSM

  wire [1:0]    dsm_debug_lvl;

  assign opt_ahb_hreadyout = (dsm_debug_lvl > 0) ? ahb_hreadyout       : 1'b1;
  assign opt_ahb_haddracc  = (dsm_debug_lvl > 0) ? ahb_haddracc        : 1'b1;
  assign opt_ahb_hresp     = (dsm_debug_lvl > 0) ? ahb_hresp           : 2'b00;
  assign opt_master        = (dsm_debug_lvl > 0) ? dap_ahb_hmaster_i   : 2'b00;
  assign opt_addr          = (dsm_debug_lvl > 0) ? dap_ahb_haddr_i     : {32{1'b0}};
  assign opt_trans         = (dsm_debug_lvl > 0) ? is_icore_trans      : 2'b00;
  assign opt_write         = (dsm_debug_lvl > 0) ? dap_ahb_hwrite_i    : 1'b0;
  assign opt_size          = (dsm_debug_lvl > 0) ? dap_ahb_hsize_i     : 2'b00;
  assign opt_size_unalign  = (dsm_debug_lvl > 0) ? is_dap_size_unalign : 2'b00;
  assign opt_burst         = (dsm_debug_lvl > 0) ? dap_ahb_hburst_i    : 3'b000;
  assign opt_prot          = (dsm_debug_lvl > 0) ? dap_ahb_hprot_i     : 4'b0000;
  assign opt_held_trans    = (dsm_debug_lvl > 0) ? is_dap_held_trans   : 1'b0;
  assign opt_unalign       = (dsm_debug_lvl > 0) ? is_dap_unalign      : 3'b000;
  assign opt_mst_en        = (dsm_debug_lvl > 0) ? 1'b1                : 1'b0;

`else

  assign opt_ahb_hreadyout = (DEBUG_LVL > 0) ? ahb_hreadyout       : 1'b1;
  assign opt_ahb_haddracc  = (DEBUG_LVL > 0) ? ahb_haddracc        : 1'b1;
  assign opt_ahb_hresp     = (DEBUG_LVL > 0) ? ahb_hresp           : 2'b00;
  assign opt_master        = (DEBUG_LVL > 0) ? dap_ahb_hmaster_i   : 2'b00;
  assign opt_addr          = (DEBUG_LVL > 0) ? dap_ahb_haddr_i     : {32{1'b0}};
  assign opt_trans         = (DEBUG_LVL > 0) ? is_icore_trans      : 2'b00;
  assign opt_write         = (DEBUG_LVL > 0) ? dap_ahb_hwrite_i    : 1'b0;
  assign opt_size          = (DEBUG_LVL > 0) ? dap_ahb_hsize_i     : 2'b00;
  assign opt_size_unalign  = (DEBUG_LVL > 0) ? is_dap_size_unalign : 2'b00;
  assign opt_burst         = (DEBUG_LVL > 0) ? dap_ahb_hburst_i    : 3'b000;
  assign opt_prot          = (DEBUG_LVL > 0) ? dap_ahb_hprot_i     : 4'b0000;
  assign opt_held_trans    = (DEBUG_LVL > 0) ? is_dap_held_trans   : 1'b0;
  assign opt_unalign       = (DEBUG_LVL > 0) ? is_dap_unalign      : 3'b000;
  assign opt_mst_en        = (DEBUG_LVL > 0) ? 1'b1                : 1'b0;

`endif



  assign addr_valid = dap_ahb_htrans_i[1];

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      data_valid <= 1'b0;
    else if (data_valid_we)
      data_valid <= nxt_data_valid;

  assign nxt_data_valid = addr_valid & ahb_haddracc;
  assign data_valid_we  = opt_mst_en & (ahb_hreadyout | dap_ahb_habort_i);


  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      split_trans <= 1'b0;
    else
      split_trans <= nxt_split_trans;

  assign data_valid_split_en = (ahb_hreadyout & ahb_haddracc) | dap_ahb_habort_i;

  assign nxt_split_trans     = (~split_trans & ~dec_dap_accept_i) |
                               ( split_trans & ~data_valid_split_en);



  assign is_dap_held_trans = addr_valid;



  always @ (dap_ahb_hsize_i or dap_ahb_haddr_i)
    case ({dap_ahb_hsize_i[1:0], dap_ahb_haddr_i[1:0]})
      4'b00_00, 4'b00_01, 4'b00_10, 4'b00_11, 4'b01_00, 4'b01_10, 4'b10_00:
        begin
          is_dap_unalign      = `MTX_UNALIGN_NORMAL;
          is_dap_size_unalign = dap_ahb_hsize_i;
        end
      4'b01_01:
        begin
          is_dap_unalign      = `MTX_UNALIGN_HWRD_01;
          is_dap_size_unalign = `MTX_SZ_BYTE;
        end
      4'b01_11:
        begin
          is_dap_unalign      = `MTX_UNALIGN_HWRD_11;
          is_dap_size_unalign = `MTX_SZ_BYTE;
        end

      4'b10_01:
        begin
          is_dap_unalign      = `MTX_UNALIGN_WORD_01;
          is_dap_size_unalign = `MTX_SZ_BYTE;
        end
      4'b10_10:
        begin
          is_dap_unalign      = `MTX_UNALIGN_WORD_10;
          is_dap_size_unalign = `MTX_SZ_HALF;
        end
      4'b10_11:
        begin
          is_dap_unalign      = `MTX_UNALIGN_WORD_11;
          is_dap_size_unalign = `MTX_SZ_BYTE;
         end


      default:
        begin
          is_dap_unalign      = 3'bXXX;
          is_dap_size_unalign = 2'bXX;
        end
    endcase


  assign is_icore_trans = split_trans ? {dap_ahb_htrans_i[1], 1'b0} :
                                        dap_ahb_htrans_i;



  assign ahb_hreadyout = ~data_valid | dec_dap_readyout_i;
  assign ahb_hresp     = data_valid ? dec_dap_resp_i : `MTX_RSP_OKAY;




  assign ahb_haddracc = ~is_dap_held_trans | (ahb_hreadyout & dec_dap_active_i);



   assign mtx_dap_ahb_hreadyout_o = opt_ahb_hreadyout;
   assign mtx_dap_ahb_haddracc_o  = opt_ahb_haddracc;
   assign mtx_dap_ahb_hresp_o     = opt_ahb_hresp;

   assign is_dap_master_o         = opt_master;
   assign is_dap_addr_o           = opt_addr;
   assign is_dap_trans_o          = opt_trans;
   assign is_dap_write_o          = opt_write;
   assign is_dap_size_o           = opt_size;
   assign is_dap_size_unalign_o   = opt_size_unalign;
   assign is_dap_burst_o          = opt_burst;
   assign is_dap_prot_o           = opt_prot;
   assign is_dap_held_trans_o     = opt_held_trans;
   assign is_dap_unalign_o        = opt_unalign;



`ifdef ARM_ASSERT_ON

  assert_always #(0,0,"opt_ahb_haddracc must only be de-asserted if HTRANS != IDLE.")
    ovl_haddracc_htrans(hclk, dreset_n,
      (opt_ahb_haddracc == 1'b0) ? (dap_ahb_htrans_i != `MTX_TRN_IDLE) : 1'b1);

  assert_never #(0,0,"dap_ahb_hsize_i = DWORD is not allowed.")
    ovl_hsize_dword(hclk, dreset_n, (dap_ahb_hsize_i[1:0] == 2'b11));

  assert_never #(0,0,"opt_unalign must never be b001.")
    ovl_unalign_001(hclk, dreset_n, (opt_unalign == 3'b001));

  assert_never #(0,0,"opt_unalign must never be b111.")
    ovl_unalign_111(hclk, dreset_n, (opt_unalign == 3'b111));

  assert_never_unknown #(0,3,0,"opt_unalign goes to X.")
    opt_unalign_x(hclk, dreset_n, 1'b1, opt_unalign);

  assert_never_unknown #(0,2,0,"is_dap_size_unalign goes to X.")
    opt_size_unalign_x(hclk, dreset_n, 1'b1, is_dap_size_unalign);

 assert_never_unknown #(0,1,0,"split_trans goes to X.")
    opt_split_trans_x(hclk, dreset_n, 1'b1, split_trans);

 assert_never_unknown #(0,2,0,"dap_ahb_hsize_i goes to X.")
    ovl_hsize_x(hclk, dreset_n, 1'b1, dap_ahb_hsize_i);

 assert_never_unknown #(0,32,0,"dap_ahb_haddr_i goes to X.")
    ovl_haddr_x(hclk, dreset_n, 1'b1, dap_ahb_haddr_i);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
