`include "cm3_mtx_defs.vp"

module cm3_mtx_input_stage_dcore
  (// Inputs
   hclk, hreset_n, dpu_ahb_haddrd_i, dpu_ahb_htransd_i, dpu_ahb_hwrited_i,
   dpu_ahb_hsized_i, dpu_ahb_hburstd_i, dpu_ahb_hprotd_i, mpu_mem_attrd_i,
   dpu_exreqd_i, dpu_dbg_snapstall_i, dpu_force_non_buf_i,
   mpu_addrd_non_buf_i, mpu_canceld_i, buff_write_accept_i,
   dec_dcore_sel_bb_i, dec_dcore_sel_ppb_i, dec_dcore_accept_i,
   dec_dcore_active_i, dec_dcore_readyout_i, dec_dcore_resp_i,
   dec_dcore_exresp_i, dec_dcore_sel_dcode_i, dec_dcore_sel_sys_i,
   mtx_mpu_ahb_hreadyoutd_o, mtx_dpu_ahb_haddraccd_o, mtx_dpu_ahb_hrespd_o,
   mtx_dpu_exrespd_o, is_dcore_master_o, is_dcore_addr_o, is_dcore_trans_o,
   is_dcore_write_o, is_dcore_size_o, is_dcore_size_unalign_o,
   is_dcore_burst_o, is_dcore_prot_o, is_dcore_mem_attr_o, is_dcore_exreq_o,
   is_dcore_held_trans_o, is_dcore_unalign_o, is_dcore_buffered_write_o,
   mtx_mpu_canceld_reg_o
  );


  input         hclk;
  input         hreset_n;

  input  [31:0] dpu_ahb_haddrd_i;
  input   [1:0] dpu_ahb_htransd_i;
  input         dpu_ahb_hwrited_i;
  input   [1:0] dpu_ahb_hsized_i;
  input   [2:0] dpu_ahb_hburstd_i;
  input   [3:0] dpu_ahb_hprotd_i;
  output        mtx_mpu_ahb_hreadyoutd_o; // HREADY seen by MPU before core
  output        mtx_dpu_ahb_haddraccd_o;
  output  [1:0] mtx_dpu_ahb_hrespd_o;

  input   [1:0] mpu_mem_attrd_i;          // MPU memory attributes
  input         mpu_addrd_non_buf_i;      // Address non-bufferable
  input         dpu_force_non_buf_i;      // Force non-bufferable
  input         dpu_dbg_snapstall_i;      // SNAPSTALL from debug
  input         dpu_exreqd_i;             // Exclusive Request
  output        mtx_dpu_exrespd_o;        // Exclusive response

  input         mpu_canceld_i;            // MPU lookup fail
  output        mtx_mpu_canceld_reg_o;    // Cancel transaction following mpu_canceld_i

  output  [1:0] is_dcore_buffered_write_o;// Trans is buffered write
  input         buff_write_accept_i;      // Buffered write accepted by BitMaster

  output [31:0] is_dcore_addr_o;          // HADDR
  output  [1:0] is_dcore_master_o;        // HMASTER
  output  [1:0] is_dcore_trans_o;         // HTRANS
  output        is_dcore_write_o;         // HWRITE
  output  [1:0] is_dcore_size_o;          // HSIZE
  output  [1:0] is_dcore_size_unalign_o;  // HSIZE, modified for unaligned access
  output  [2:0] is_dcore_burst_o;         // HBURST
  output  [3:0] is_dcore_prot_o;          // HPROT
  output  [1:0] is_dcore_mem_attr_o;      // Memory attributes
  output        is_dcore_exreq_o;         // Exclusive Request output
  output        is_dcore_held_trans_o;    // Holding register active flag

  input         dec_dcore_sel_dcode_i;    // Access to D-Code port
  input         dec_dcore_sel_sys_i;      // Access to System port
  input         dec_dcore_sel_bb_i;       // Bit-band access
  input         dec_dcore_sel_ppb_i;      // PPB access
  input         dec_dcore_accept_i;       // Active signal
  input         dec_dcore_active_i;       // Active signal - address phase
  input         dec_dcore_readyout_i;     // HREADYOUT input
  input   [1:0] dec_dcore_resp_i;         // HRESP input
  input         dec_dcore_exresp_i;       // EXRESP input

  output  [2:0] is_dcore_unalign_o;


  wire        addr_valid;             // Indicates address phase of valid transfer
  reg         data_valid;             // Indicates data phase of valid transfer
  wire        nxt_data_valid;
  wire        data_valid_we;
  wire        data_valid_split_en;
  reg         split_trans;
  wire        nxt_split_trans;

  reg   [1:0] is_dcore_size_unalign;  // Modified HSIZE for unaligned accesses
  reg   [2:0] is_dcore_unalign;       // Unaligned access encoding

  reg   [1:0] buffered_write;         // Buffered Write
  wire  [1:0] nxt_buffered_write;     // Buffered Write - address phase
  wire        buffered_write_we;

  reg         mpu_canceld_reg;        // Registered mpu_canceld_i
  reg         mpu_unalign_lookup;     // MPU Unaligned lookup in next cycle
  wire        nxt_mpu_canceld_reg;    // Next mpu_canceld_reg
  wire        nxt_mpu_unalign_lookup; // Next mpu_unalign_lookup

  wire  [1:0] is_dcore_trans;
  wire        is_dcore_held_trans;

  wire        i_hreadyout;
  wire        mpu_ahb_hreadyoutd;
  wire        dpu_ahb_haddraccd;
  wire  [1:0] dpu_ahb_hrespd;
  wire        dpu_exrespd;



  assign addr_valid = dpu_ahb_htransd_i[1];

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      data_valid <= 1'b0;
    else if (data_valid_we)
      data_valid <= nxt_data_valid;

  assign data_valid_we  = mpu_ahb_hreadyoutd | dpu_dbg_snapstall_i;
  assign nxt_data_valid = addr_valid     & dpu_ahb_haddraccd &
                          ~mpu_canceld_i & ~mpu_canceld_reg;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      split_trans <= 1'b0;
    else
      split_trans <= nxt_split_trans;

  assign data_valid_split_en = (mpu_ahb_hreadyoutd & dpu_ahb_haddraccd) |
                               dpu_dbg_snapstall_i;

  assign nxt_split_trans     = (~split_trans & ~dec_dcore_accept_i) |
                               ( split_trans & ~data_valid_split_en);




  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        mpu_canceld_reg    <= 1'b0;
        mpu_unalign_lookup <= 1'b0;
      end
    else
      begin
        mpu_canceld_reg    <= nxt_mpu_canceld_reg;
        mpu_unalign_lookup <= nxt_mpu_unalign_lookup;
      end

  assign nxt_mpu_canceld_reg = (mpu_ahb_hreadyoutd & dpu_ahb_haddraccd &
                                mpu_canceld_i      & addr_valid) |
                               (mpu_unalign_lookup & mpu_canceld_i);

  assign nxt_mpu_unalign_lookup = mpu_ahb_hreadyoutd  &
                                  dpu_ahb_haddraccd   &
                                  ~mpu_canceld_i      &
                                  (is_dcore_unalign != `MTX_UNALIGN_NORMAL) &
                                  addr_valid          &
                                  ~dec_dcore_sel_bb_i &
                                  ~dec_dcore_sel_ppb_i;

  assign is_dcore_held_trans = addr_valid & ~mpu_canceld_reg;




  always @ (dpu_ahb_hsized_i or dpu_ahb_haddrd_i)
    case ({dpu_ahb_hsized_i[1:0], dpu_ahb_haddrd_i[1:0]})
      4'b00_00, 4'b00_01, 4'b00_10, 4'b00_11, 4'b01_00, 4'b01_10, 4'b10_00:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_NORMAL;
          is_dcore_size_unalign = dpu_ahb_hsized_i;
        end
      4'b01_01:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_HWRD_01;
          is_dcore_size_unalign = `MTX_SZ_BYTE;
        end
      4'b01_11:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_HWRD_11;
          is_dcore_size_unalign = `MTX_SZ_BYTE;
        end

      4'b10_01:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_WORD_01;
          is_dcore_size_unalign = `MTX_SZ_BYTE;
        end
      4'b10_10:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_WORD_10;
          is_dcore_size_unalign = `MTX_SZ_HALF;
        end
      4'b10_11:
        begin
          is_dcore_unalign      = `MTX_UNALIGN_WORD_11;
          is_dcore_size_unalign = `MTX_SZ_BYTE;
        end


      default:
        begin
          is_dcore_unalign      = 3'bXXX;
          is_dcore_size_unalign = 2'bXX;
        end
    endcase

  assign is_dcore_trans = split_trans ? {dpu_ahb_htransd_i[1], 1'b0} :
                                        dpu_ahb_htransd_i;




  assign nxt_buffered_write = {2{(dpu_ahb_hwrited_i    &
                                  dpu_ahb_hprotd_i[2]  &
                                  is_dcore_trans[1]    &
                                  dec_dcore_accept_i   &
                                  ~dpu_force_non_buf_i &
                                  ~mpu_addrd_non_buf_i &
                                  (is_dcore_unalign == `MTX_UNALIGN_NORMAL) &
                                  dpu_ahb_haddraccd    &
                                  ~mpu_canceld_i       &
                                  ~dpu_exreqd_i)}}     &
                              {dec_dcore_sel_sys_i, dec_dcore_sel_dcode_i};

  assign buffered_write_we  = mpu_ahb_hreadyoutd;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      buffered_write <= 2'b00;
    else if (buffered_write_we)
      buffered_write <= nxt_buffered_write;




  assign i_hreadyout        = ~data_valid | dec_dcore_readyout_i;
  assign dpu_exrespd        = dec_dcore_exresp_i;

  assign mpu_ahb_hreadyoutd = buff_write_accept_i | i_hreadyout;
  assign dpu_ahb_hrespd     = {2{(~buff_write_accept_i) & data_valid}} &
                              dec_dcore_resp_i;




  assign dpu_ahb_haddraccd = ~is_dcore_held_trans |
                             (i_hreadyout & dec_dcore_active_i);



  assign mtx_mpu_ahb_hreadyoutd_o  = mpu_ahb_hreadyoutd;
  assign mtx_dpu_ahb_haddraccd_o   = dpu_ahb_haddraccd;
  assign mtx_dpu_ahb_hrespd_o      = dpu_ahb_hrespd;
  assign mtx_dpu_exrespd_o         = dpu_exrespd;

  assign is_dcore_addr_o           = dpu_ahb_haddrd_i;
  assign is_dcore_master_o         = 2'b00;
  assign is_dcore_trans_o          = is_dcore_trans;
  assign is_dcore_write_o          = dpu_ahb_hwrited_i;
  assign is_dcore_size_o           = dpu_ahb_hsized_i;
  assign is_dcore_size_unalign_o   = is_dcore_size_unalign;
  assign is_dcore_burst_o          = dpu_ahb_hburstd_i;
  assign is_dcore_prot_o           = dpu_ahb_hprotd_i;
  assign is_dcore_mem_attr_o       = mpu_mem_attrd_i;
  assign is_dcore_exreq_o          = dpu_exreqd_i;
  assign is_dcore_held_trans_o     = is_dcore_held_trans;

  assign is_dcore_unalign_o        = is_dcore_unalign;
  assign is_dcore_buffered_write_o = buffered_write;
  assign mtx_mpu_canceld_reg_o     = mpu_canceld_reg;



`ifdef ARM_ASSERT_ON

  assert_always #(0,0,"dpu_ahb_haddracci must only be de-asserted if HTRANS != IDLE.")
    ovl_haddracc_htrans(hclk, hreset_n,
      (dpu_ahb_haddraccd == 1'b0) ? (dpu_ahb_htransd_i != `MTX_TRN_IDLE) : 1'b1);

  assert_zero_one_hot #(0,2,0,"buffered_write must be exclusive.")
    ovl_buffered_write_exclusive(hclk, hreset_n, buffered_write);

  assert_never #(0,0,"dpu_ahb_hsized_i = DWORD not allowed.")
    ovl_hsize_dword(hclk, hreset_n, dpu_ahb_hsized_i[1:0] == 2'b11);

  assert_never #(0,0,"is_dcore_unalign must never be b001.")
    ovl_unalign_001(hclk, hreset_n, is_dcore_unalign == 3'b001);

  assert_never #(0,0,"is_dcore_unalign must never be b111.")
    ovl_unalign_111(hclk, hreset_n, is_dcore_unalign == 3'b111);

  assert_never_unknown #(0,3,0,"is_dcore_unalign goes to x.")
    ovl_unalign_x (hclk, hreset_n, 1'b1, is_dcore_unalign);

  assert_never_unknown #(0,2,0,"is_dcore_size_unalign goes to x.")
    ovl_size_unaligned_x (hclk, hreset_n, 1'b1, is_dcore_size_unalign);

  assert_never_unknown #(0,1,0,"split_trans goes to x.")
    ovl_split_transaction_x (hclk, hreset_n, 1'b1, split_trans);

  assert_never_unknown #(0,2,0,"dpu_ahb_hsized_i goes to x.")
    ovl_hsized_x (hclk, hreset_n, 1'b1, dpu_ahb_hsized_i);

 assert_never_unknown #(0,32,0,"dpu_ahb_haddrd_i goes to x.")
    ovl_haddrd_x (hclk, hreset_n, 1'b1, dpu_ahb_haddrd_i);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
