module cm3_mtx_input_stage_icore
  (// Inputs
   hclk, hreset_n, dpu_ahb_haddri_i, dpu_ahb_htransi_i, dpu_ahb_hsizei_i,
   dpu_ahb_hbursti_i, dpu_ahb_hproti_i, mpu_mem_attri_i, mpu_canceli_i,
   bm_icore_to_sys_i, dec_icore_accept_i, dec_icore_active_i,
   dec_icore_readyout_i, dec_icore_resp_i,
   mtx_mpu_ahb_hreadyouti_o, mtx_dpu_ahb_haddracci_o, mtx_dpu_ahb_hrespi_o,
   is_icore_addr_o, is_icore_trans_o, is_icore_size_o, is_icore_burst_o,
   is_icore_prot_o, is_icore_mem_attr_o, is_icore_held_trans_o,
   mtx_mpu_canceli_reg_o
  );


  input         hclk;
  input         hreset_n;

  input  [31:0] dpu_ahb_haddri_i;
  input   [1:0] dpu_ahb_htransi_i;
  input   [1:0] dpu_ahb_hsizei_i;
  input   [2:0] dpu_ahb_hbursti_i;
  input   [3:0] dpu_ahb_hproti_i;
  output        mtx_mpu_ahb_hreadyouti_o; // HREADY response through MPU
  output        mtx_dpu_ahb_haddracci_o;
  output  [1:0] mtx_dpu_ahb_hrespi_o;

  input   [1:0] mpu_mem_attri_i;          // MPU memory attributes

  input         mpu_canceli_i;            // MPU lookup fail
  output        mtx_mpu_canceli_reg_o;    // Cancel transaction following MPU cancel

  input         bm_icore_to_sys_i;        // Pipelined I-Core to System access

  input         dec_icore_accept_i;       // ICore transaction accepted
  input         dec_icore_active_i;       // ICore transaction active
  input         dec_icore_readyout_i;     // HREADYOUT from selected slave
  input   [1:0] dec_icore_resp_i;         // HRESP from selected slave

  output [31:0] is_icore_addr_o;          // HADDR
  output  [1:0] is_icore_trans_o;         // HTRANS
  output  [1:0] is_icore_size_o;          // HSIZE
  output  [2:0] is_icore_burst_o;         // HBURST
  output  [3:0] is_icore_prot_o;          // HPROT
  output  [1:0] is_icore_mem_attr_o;      // Memory attributes
  output        is_icore_held_trans_o;    // Holding register active flag



  wire         mpu_ahb_hreadyouti;
  wire         dpu_ahb_haddracci;
  wire   [1:0] dpu_ahb_hrespi;

  wire   [1:0] is_icore_trans;        // HTRANS output
  wire         is_icore_held_trans;   // Holding register active flag

  wire         addr_valid;            // Address phase of valid transfer
  reg          data_valid;            // Data phase of valid transfer
  wire         nxt_data_valid;
  wire         data_valid_we;
  wire         data_valid_split_en;
  reg          split_trans;
  wire         nxt_split_trans;

  reg          mpu_canceli_reg;       // Registered mpu_canceli_i
  wire         nxt_mpu_canceli_reg;   // Next mpu_canceli_reg



  assign addr_valid = dpu_ahb_htransi_i[1];

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      data_valid <= 1'b0;
    else if (data_valid_we)
      data_valid <= nxt_data_valid;

  assign data_valid_we  = mpu_ahb_hreadyouti;
  assign nxt_data_valid = addr_valid                             &
                          dpu_ahb_haddracci                      &
                          ~(mpu_canceli_i & ~bm_icore_to_sys_i)  &
                          ~mpu_canceli_reg;

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      split_trans <= 1'b0;
    else
      split_trans <= nxt_split_trans;

  assign data_valid_split_en = mpu_ahb_hreadyouti & dpu_ahb_haddracci;

  assign nxt_split_trans     = (~split_trans & ~dec_icore_accept_i) |
                               ( split_trans & ~data_valid_split_en);


  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      mpu_canceli_reg <= 1'b0;
    else
      mpu_canceli_reg <= nxt_mpu_canceli_reg;

  assign nxt_mpu_canceli_reg = mpu_ahb_hreadyouti &
                               dpu_ahb_haddracci  &
                               mpu_canceli_i      &
                               addr_valid;

  assign is_icore_held_trans = addr_valid & ~mpu_canceli_reg;


  assign is_icore_trans = split_trans ? {dpu_ahb_htransi_i[1], 1'b0} :
                                        dpu_ahb_htransi_i;




  assign mpu_ahb_hreadyouti = ~data_valid | dec_icore_readyout_i;
  assign dpu_ahb_hrespi     = {2{data_valid}} & dec_icore_resp_i;




  assign dpu_ahb_haddracci = ~is_icore_held_trans |
                             (mpu_ahb_hreadyouti & dec_icore_active_i);



  assign mtx_mpu_ahb_hreadyouti_o = mpu_ahb_hreadyouti;
  assign mtx_dpu_ahb_haddracci_o  = dpu_ahb_haddracci;
  assign mtx_dpu_ahb_hrespi_o     = dpu_ahb_hrespi;

  assign is_icore_addr_o          = dpu_ahb_haddri_i;
  assign is_icore_trans_o         = is_icore_trans;
  assign is_icore_size_o          = dpu_ahb_hsizei_i;
  assign is_icore_burst_o         = dpu_ahb_hbursti_i;
  assign is_icore_prot_o          = dpu_ahb_hproti_i;
  assign is_icore_mem_attr_o      = mpu_mem_attri_i;
  assign is_icore_held_trans_o    = is_icore_held_trans;

  assign mtx_mpu_canceli_reg_o    = mpu_canceli_reg;



`ifdef ARM_ASSERT_ON

  `include "cm3_mtx_defs.v"

  assert_always #(0,0,"dpu_ahb_haddracci must only be de-asserted if HTRANS != IDLE.")
    ovl_haddracc_htrans(hclk, hreset_n,
      (dpu_ahb_haddracci == 1'b0) ? (dpu_ahb_htransi_i != `MTX_TRN_IDLE) : 1'b1);

  assert_never #(0,0,"dpu_ahb_hsizei_i = DWORD not allowed.")
    ovl_hsize_dword(hclk, hreset_n, dpu_ahb_hsizei_i[1:0] == 2'b11);

  assert_never_unknown #(0,1,0,"split_trans goes to X.")
    ovl_split_trans_x (hclk, hreset_n, 1'b1, split_trans);

  `include "cm3_mtx_undefs.v"

`endif

endmodule
