`include "cm3_mtx_defs.vp"

module cm3_mtx_output_stage_icode
  (// Inputs
   hclk, hreset_n, d_not_i_trans_i, mtx_mpu_canceli_reg_i,
   dec_icore_sel_icode_i, dec_icore_sel_icode_held_i, is_icore_addr_i,
   is_icore_trans_i, is_icore_size_i, is_icore_burst_i, is_icore_prot_i,
   is_icore_held_trans_i, mtx_ahb_hreadyi_i,
   os_icore_to_icode_accept_o, os_icore_to_icode_active_o,
   os_icore_to_icode_held_o, mtx_ahb_haddri_o, mtx_ahb_htransi_o,
   mtx_ahb_hsizei_o, mtx_ahb_hbursti_o, mtx_ahb_hproti_o, mtx_mem_attri_o
  );

  parameter CONST_AHB_CTRL = 0;  // Constant AHB control info if 1


  input         hclk;                       // AHB system clock
  input         hreset_n;                   // AHB system reset

  input         d_not_i_trans_i;

  input         dec_icore_sel_icode_i;      // ICore -> ICode select
  input         dec_icore_sel_icode_held_i; // ICore -> ICode select held

  input  [31:0] is_icore_addr_i;
  input   [1:0] is_icore_trans_i;
  input   [1:0] is_icore_size_i;
  input   [2:0] is_icore_burst_i;
  input   [3:0] is_icore_prot_i;
  input         is_icore_held_trans_i;
  output        os_icore_to_icode_accept_o; // ICore -> ICode accepted
  output        os_icore_to_icode_active_o; // ICore -> ICode active
  output        os_icore_to_icode_held_o;   // ICore -> ICode waited request

  input         mtx_mpu_canceli_reg_i;      // MPU instruction lookup fail

  output [31:0] mtx_ahb_haddri_o;
  output  [1:0] mtx_ahb_htransi_o;
  output  [2:0] mtx_ahb_hsizei_o;
  output  [2:0] mtx_ahb_hbursti_o;
  output  [3:0] mtx_ahb_hproti_o;
  output  [1:0] mtx_mem_attri_o;
  input         mtx_ahb_hreadyi_i;



  wire          icore_req;      // Request from ICore
  wire          no_port_sel;    // No port selected

  wire          icore_to_icode_accept;
  wire          icore_to_icode_active;

  wire          ahb_ctrl_valid_we;
  reg           ahb_ctrl_valid;
  wire          icode_held_req;
  wire [31:0]   ahb_haddr;
  wire  [1:0]   ahb_htrans;
  wire  [2:0]   ahb_hsize;
  wire  [2:0]   ahb_hburst;
  wire  [3:0]   ahb_hprot;
  wire  [1:0]   mem_attr;       // Memory attributes

  wire          opt_const_en;

`ifdef ARM_DSM

  wire          dsm_const_ahb_ctrl;

  assign opt_const_en = (dsm_const_ahb_ctrl > 0) ? 1'b1 : 1'b0;

`else

  assign opt_const_en = (CONST_AHB_CTRL > 0) ? 1'b1 : 1'b0;

`endif


  assign icore_req   = is_icore_held_trans_i & dec_icore_sel_icode_i;
  assign no_port_sel = ~(icore_req & ~dec_icore_sel_icode_held_i);



  assign icore_to_icode_accept = ~no_port_sel;
  assign icore_to_icode_active = ~dec_icore_sel_icode_held_i & mtx_ahb_hreadyi_i;



  assign ahb_haddr  = {3'b000, is_icore_addr_i[28:0]};
  assign ahb_htrans = ((~icode_held_req &
                        (dec_icore_sel_icode_held_i |
                         ~dec_icore_sel_icode_i     |
                         mtx_mpu_canceli_reg_i)) ?
                       `MTX_TRN_IDLE : is_icore_trans_i);
  assign ahb_hsize  = {1'b0, is_icore_size_i};
  assign ahb_hburst = is_icore_burst_i;
  assign ahb_hprot  = {2'b10, is_icore_prot_i[1:0]}; // Cacheable, non-bufferable
  assign mem_attr   = 2'b01;                         // Non-shareable, allocate

  assign ahb_ctrl_valid_we = (opt_const_en & d_not_i_trans_i &
                              (ahb_ctrl_valid |
                               ((|ahb_htrans) & ~mtx_ahb_hreadyi_i)));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      ahb_ctrl_valid  <= 1'b0;
    else if (ahb_ctrl_valid_we)
      ahb_ctrl_valid  <= (|ahb_htrans) & ~mtx_ahb_hreadyi_i & d_not_i_trans_i;

  assign icode_held_req = ahb_ctrl_valid & opt_const_en;


  assign os_icore_to_icode_accept_o   = icore_to_icode_accept;
  assign os_icore_to_icode_active_o   = icore_to_icode_active;
  assign os_icore_to_icode_held_o     = icode_held_req;
  assign mtx_ahb_haddri_o             = ahb_haddr;
  assign mtx_ahb_htransi_o            = ahb_htrans;
  assign mtx_ahb_hsizei_o             = ahb_hsize;
  assign mtx_ahb_hbursti_o            = ahb_hburst;
  assign mtx_ahb_hproti_o             = ahb_hprot;
  assign mtx_mem_attri_o              = mem_attr;


`ifdef ARM_ASSERT_ON

  assert_always #(0,0,"ahb_hprot[3:2] must always be 2'b10.")
    ovl_hprot32_valid(hclk, hreset_n, (ahb_hprot[3:2] == 2'b10));

  assert_always #(0,0,"mem_attr must always be 2'b01.")
    ovl_mem_attr_valid(hclk, hreset_n, (mem_attr == 2'b01));

  assert_always #(0,0,"is_icore_addr_i[31:29] must be 0 when port is active.")
    ovl_icore_addr_valid(hclk, hreset_n,
      (no_port_sel ? 1'b1 : (is_icore_addr_i[31:29] == 3'b000)));

  assert_never_unknown #(0,32,0,"ahb_haddr goes to X.")
    ovl_haddr_x(hclk, hreset_n, (ahb_htrans != `MTX_TRN_IDLE), ahb_haddr);

  assert_never_unknown #(0,3,0,"ahb_hsize goes to X.")
    ovl_hsize_x(hclk, hreset_n, (ahb_htrans != `MTX_TRN_IDLE), ahb_hsize);

  assert_never_unknown #(0,3,0,"ahb_hburst goes to X.")
    ovl_hburst_x(hclk, hreset_n, (ahb_htrans != `MTX_TRN_IDLE), ahb_hburst);

  assert_never_unknown #(0,4,0,"ahb_hprot goes to X.")
    ovl_hprot_x(hclk, hreset_n, (ahb_htrans != `MTX_TRN_IDLE), ahb_hprot);

  assert_never_unknown #(0,2,0,"mem_attr goes to X.")
    ovl_mem_attr_x(hclk, hreset_n, (ahb_htrans != `MTX_TRN_IDLE), mem_attr);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
