`include "cm3_mtx_defs.vp"

module cm3_mtx_output_stage_ppb
  (// Inputs
   hclk, hreset_n, mpu_canceld_i, dec_dcore_sel_ppb_i, dec_dcore_sel_ppb_held_i,
   is_dcore_master_i, is_dcore_addr_i, is_dcore_trans_i, is_dcore_write_i,
   is_dcore_size_i, is_dcore_burst_i, is_dcore_prot_i, dpu_ahb_hwdatad_i,
   is_dcore_held_trans_i, dec_dap_sel_ppb_i, dec_dap_sel_ppb_held_i,
   is_dap_master_i, is_dap_addr_i, is_dap_trans_i, is_dap_write_i,
   is_dap_size_i, is_dap_burst_i, is_dap_prot_i, dap_ahb_hwdata_i,
   is_dap_held_trans_i, mtx_ppb_hready_i,
   os_dcore_to_ppb_accept_o, os_dap_to_ppb_accept_o, os_dcore_to_ppb_active_o,
   os_dap_to_ppb_active_o, mtx_ppb_haddr_o, mtx_ppb_hmaster_o,
   mtx_ppb_htrans_o, mtx_ppb_hwrite_o, mtx_ppb_hsize_o, mtx_ppb_hburst_o,
   mtx_ppb_hprot_o, mtx_ppb_hwdata_o
  );


  input         hclk;                     // AHB system clock
  input         hreset_n;                 // AHB system reset

  input         dec_dcore_sel_ppb_i;      // DCore -> PPB select
  input         dec_dcore_sel_ppb_held_i; // DCore -> PPB select held
  input         dec_dap_sel_ppb_i;        // DAP -> PPB select
  input         dec_dap_sel_ppb_held_i;   // DAP -> PPB select held

  input  [31:0] is_dcore_addr_i;
  input   [1:0] is_dcore_master_i;
  input   [1:0] is_dcore_trans_i;
  input         is_dcore_write_i;
  input   [1:0] is_dcore_size_i;
  input   [2:0] is_dcore_burst_i;
  input   [3:0] is_dcore_prot_i;
  input  [31:0] dpu_ahb_hwdatad_i;
  input         is_dcore_held_trans_i;
  output        os_dcore_to_ppb_accept_o; // DCore -> PPB accepted
  output        os_dcore_to_ppb_active_o; // DCore -> PPB active

  input  [1:0]  is_dap_master_i;
  input [31:0]  is_dap_addr_i;
  input  [1:0]  is_dap_trans_i;
  input         is_dap_write_i;
  input  [1:0]  is_dap_size_i;
  input  [2:0]  is_dap_burst_i;
  input  [3:0]  is_dap_prot_i;
  input [31:0]  dap_ahb_hwdata_i;
  input         is_dap_held_trans_i;
  output        os_dap_to_ppb_accept_o;   // DAP -> PPB accepted
  output        os_dap_to_ppb_active_o;   // DAP -> PPB active

  input         mpu_canceld_i;            // MPU D-Side lookup failed

  output [31:0] mtx_ppb_haddr_o;
  output  [1:0] mtx_ppb_hmaster_o;
  output  [1:0] mtx_ppb_htrans_o;
  output        mtx_ppb_hwrite_o;
  output  [2:0] mtx_ppb_hsize_o;
  output  [2:0] mtx_ppb_hburst_o;
  output  [3:0] mtx_ppb_hprot_o;
  output [31:0] mtx_ppb_hwdata_o;
  input         mtx_ppb_hready_i;


  wire          dcore_req;           // DCore request
  wire          dap_req;             // DAP request
  reg     [1:0] input_port_addr;     // Prioritized input port address
  reg           no_port_sel;         // No port selected

  wire          dcore_to_ppb_accept; // DCore -> PPB transaction accepted
  wire          dcore_to_ppb_active; // DCore -> PPB transaction active
  wire          dap_to_ppb_accept;   // DAP -> PPB transaction accepted
  wire          dap_to_ppb_active;   // DAP -> PPB transaction active

  reg    [31:0] ppb_haddr;
  reg     [1:0] ppb_hmaster;
  reg     [1:0] ppb_htrans;
  reg           ppb_hwrite;
  reg     [2:0] ppb_hsize;
  reg     [2:0] ppb_hburst;
  reg     [3:0] ppb_hprot;
  wire   [31:0] ppb_hwdata;

  reg     [1:0] input_port_data;     // Input port for data phase response
  wire    [1:0] nxt_input_port_data;
  reg           no_port_sel_data;



  assign dcore_req = is_dcore_held_trans_i & dec_dcore_sel_ppb_i;
  assign dap_req   = is_dap_held_trans_i & dec_dap_sel_ppb_i;

  always @ (dcore_req or dap_req)
    case ({dcore_req, dap_req})
      2'b10, 2'b11: input_port_addr = 2'b01;  // DCore
      2'b01:        input_port_addr = 2'b10;  // DAP
      2'b00:        input_port_addr = 2'b00;  // No input port request
      default:      input_port_addr = 2'bXX;  // Propagate X
    endcase

  always @ (dcore_req or dap_req or dec_dcore_sel_ppb_held_i or
            dec_dap_sel_ppb_held_i)
    case ({dcore_req, dap_req})
      2'b10, 2'b11: no_port_sel = dec_dcore_sel_ppb_held_i;
      2'b01:        no_port_sel = dec_dap_sel_ppb_held_i;
      2'b00:        no_port_sel = 1'b1;
      default:      no_port_sel = 1'bX;
    endcase



  assign dcore_to_ppb_accept = input_port_addr[0] & ~no_port_sel;
  assign dap_to_ppb_accept   = input_port_addr[1] & ~no_port_sel;

  assign dcore_to_ppb_active = ~dec_dcore_sel_ppb_held_i & mtx_ppb_hready_i;
  assign dap_to_ppb_active   = dap_to_ppb_accept & mtx_ppb_hready_i;



  always @ (is_dcore_addr_i or is_dcore_trans_i or is_dcore_write_i or
            is_dcore_size_i or is_dcore_burst_i or is_dcore_prot_i or
            is_dcore_master_i or is_dap_addr_i or is_dap_trans_i or
            is_dap_write_i or is_dap_size_i or is_dap_burst_i or
            is_dap_prot_i or is_dap_master_i or input_port_addr or
            no_port_sel or mpu_canceld_i)
    case (input_port_addr)
      2'b01:
        begin
          ppb_hmaster = is_dcore_master_i;
          ppb_haddr   = {is_dcore_addr_i[31:29], 9'b000000000, is_dcore_addr_i[19:0]};
          ppb_htrans  = (mpu_canceld_i | no_port_sel) ? `MTX_TRN_IDLE : is_dcore_trans_i;
          ppb_hwrite  = is_dcore_write_i;
          ppb_hsize   = {1'b0, is_dcore_size_i};
          ppb_hburst  = is_dcore_burst_i;
          ppb_hprot   = is_dcore_prot_i;
        end

      2'b10:
        begin
          ppb_hmaster = is_dap_master_i;
          ppb_haddr   = {is_dap_addr_i[31:29], 9'b000000000, is_dap_addr_i[19:0]};
          ppb_htrans  = no_port_sel ? `MTX_TRN_IDLE : is_dap_trans_i;
          ppb_hwrite  = is_dap_write_i;
          ppb_hsize   = {1'b0, is_dap_size_i};
          ppb_hburst  = is_dap_burst_i;
          ppb_hprot   = is_dap_prot_i;
        end

      2'b00, 2'b11:
        begin
          ppb_hmaster = 2'b00;
          ppb_haddr   = 32'h00000000;
          ppb_htrans  = 2'b00;
          ppb_hwrite  = 1'b0;
          ppb_hsize   = 3'b000;
          ppb_hburst  = 3'b000;
          ppb_hprot   = 4'b0000;
        end

      default:
        begin
          ppb_hmaster = 2'bXX;
          ppb_haddr   = 32'hXXXXXXXX;
          ppb_htrans  = 2'bXX;
          ppb_hwrite  = 1'bX;
          ppb_hsize   = 3'bXXX;
          ppb_hburst  = 3'bXXX;
          ppb_hprot   = 4'bXXXX;
        end

    endcase



  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      begin
        input_port_data  <= 2'b00;
        no_port_sel_data <= 1'b1;
      end
    else if (mtx_ppb_hready_i)
      begin
        input_port_data  <= nxt_input_port_data;
        no_port_sel_data <= no_port_sel;
      end

  assign nxt_input_port_data = input_port_addr & {2{ppb_hwrite}};

  assign ppb_hwdata = ({32{input_port_data[0] & ~no_port_sel_data}} & dpu_ahb_hwdatad_i) |
                      ({32{input_port_data[1] & ~no_port_sel_data}} & dap_ahb_hwdata_i);



  assign os_dcore_to_ppb_accept_o = dcore_to_ppb_accept;
  assign os_dap_to_ppb_accept_o   = dap_to_ppb_accept;
  assign os_dcore_to_ppb_active_o = dcore_to_ppb_active;
  assign os_dap_to_ppb_active_o   = dap_to_ppb_active;
  assign mtx_ppb_haddr_o          = ppb_haddr;
  assign mtx_ppb_hmaster_o        = ppb_hmaster;
  assign mtx_ppb_htrans_o         = ppb_htrans;
  assign mtx_ppb_hwrite_o         = ppb_hwrite;
  assign mtx_ppb_hsize_o          = ppb_hsize;
  assign mtx_ppb_hburst_o         = ppb_hburst;
  assign mtx_ppb_hprot_o          = ppb_hprot;
  assign mtx_ppb_hwdata_o         = ppb_hwdata;



`ifdef ARM_ASSERT_ON

  assert_zero_one_hot #(0,2,0,"input_port_addr must be exclusive.")
    ovl_input_port_addr_exclusive(hclk, hreset_n, input_port_addr);

  assert_never_unknown #(0,2,0,"input_port_addr goes to X.")
    ovl_input_port_addr_x (hclk, hreset_n, 1'b1, input_port_addr);

  assert_zero_one_hot #(0,2,0,"input_port_data must be exclusive.")
    ovl_input_port_data_exclusive(hclk, hreset_n, input_port_data);

  assert_never_unknown #(0,2,0,"input_port_data goes to X.")
    ovl_input_port_data_x (hclk, hreset_n, 1'b1, input_port_data);

  assert_never_unknown #(0,1,0,"no_port_sel_data goes to X.")
    ovl_no_port_sel_data_x (hclk, hreset_n, 1'b1, no_port_sel_data);

  assert_always #(0,0,"is_dcore_addr_i[28:20] must be 0 when port is active.")
    ovl_dcore_addr_valid(hclk, hreset_n,
      (((input_port_addr == 2'b01) & ~no_port_sel) ?
         (is_dcore_addr_i[28:20] == 9'b000000000) : 1'b1));

  assert_always #(0,0,"is_dap_addr_i[28:20] must be 0 when port is active.")
    ovl_dap_addr_valid(hclk, hreset_n,
      (((input_port_addr == 2'b10) & ~no_port_sel) ?
         (is_dap_addr_i[28:20] == 9'b000000000) : 1'b1));

  assert_always #(0,0,"ppb_hmaster must be 00 when DCore port is active.")
    ovl_dcore_master_valid(hclk, hreset_n,
      (((input_port_addr == 2'b00) & ~no_port_sel) ?
         (ppb_hmaster == 2'b00) : 1'b1));


  assert_never_unknown #(0,32,0,"ppb_haddr goes to x during transaction.")
    ovl_haddr_x (hclk, hreset_n, (ppb_htrans != `MTX_TRN_IDLE), ppb_haddr);

  assert_never_unknown #(0,1,0,"ppb_hwrite goes to x during transaction.")
    ovl_hwrite_x (hclk, hreset_n, (ppb_htrans != `MTX_TRN_IDLE), ppb_hwrite);

  assert_never_unknown #(0,3,0,"ppb_hsize goes to x during transaction.")
    ovl_hsize_x (hclk, hreset_n, (ppb_htrans != `MTX_TRN_IDLE), ppb_hsize);

  assert_never_unknown #(0,3,0,"ppb_hburst goes to x during transaction.")
    ovl_hburst_x (hclk, hreset_n, (ppb_htrans != `MTX_TRN_IDLE), ppb_hburst);

  assert_never_unknown #(0,4,0,"ppb_hprot goes to x during transaction.")
    ovl_hprot_x (hclk, hreset_n, (ppb_htrans != `MTX_TRN_IDLE), ppb_hprot);

`endif

endmodule

`include "cm3_mtx_undefs.vp"
