`include "cm3_nvic_defs.vp"

module cm3_nvic_int_state
  (// Inputs
   hreset_n, fclk, hclk, int_pend_set_i, int_pend_clr_i, int_actv_set_i,
   int_actv_clr_i, int_pend_i, dpu_int_entry_i, isr_entry_i, dpu_int_exit_i,
   isr_exit_i,
   pend_state_o, actv_state_o, slp_wfe_pend_o
   );

  parameter INTS            = 1;                    // Number of interrupts

  input                    hreset_n;                // Reset (main NVIC reset)
  input                    fclk;                    // Clock (free-running)
  input                    hclk;                    // Clock (clock gated)
  input [INTS-1:0]         int_pend_set_i;          // Interrupt pending set
  input [INTS-1:0]         int_pend_clr_i;          // Interrupt pending clear
  input [INTS-1:0]         int_actv_set_i;          // Interrupt active set
  input [INTS-1:0]         int_actv_clr_i;          // Interrupt active clear
  input [INTS-1:0]         int_pend_i;              // Interrupt pending set
  input                    dpu_int_entry_i;         // Interrupt entry
  input [INTS-1:0]         isr_entry_i;             // ISR entry
  input                    dpu_int_exit_i;          // Interrupt exit
  input [INTS-1:0]         isr_exit_i;              // ISR exit

  output [INTS-1:0]        pend_state_o;            // Interrupt pend status
  output [INTS-1:0]        actv_state_o;            // Interrupt active status
  output                   slp_wfe_pend_o;          // WFE newly pend event

  wire                     int_pend_we;
  wire [INTS-1:0]          int_pend_pulse;
  reg  [INTS-1:0]          pend_state;
  reg  [INTS-1:0]          actv_state;
  reg  [INTS-1:0]          next_pend_state;
  wire                     update_actv_state;
  wire                     update_pend_state;
  reg  [INTS-1:0]          next_actv_state;
  reg  [INTS-1:0]          prev_int_pend;
  reg                      slp_wfe_pend;
  wire                     nxt_slp_wfe_pend;
  wire                     update_slp_wfe_pend;
  wire [INTS-1:0]          opt_int_pend_set;
  wire [INTS-1:0]          opt_int_actv_set;
  wire [INTS-1:0]          opt_int_pend;

  integer                  i;

`ifdef ARM_DSM

  wire [7:0]               dsm_num_irq;
  reg  [239:0]             dsm_imask;
  integer                  j;

  always @ (dsm_num_irq)
    for (j = 0; j < 240; j = j + 1)
      dsm_imask[j] = (j < dsm_num_irq) ? 1'b1 : 1'b0;

  assign opt_int_pend_set = int_pend_set_i & {dsm_imask, {`num_non_irq{1'b1}}};
  assign opt_int_pend     = int_pend_i     & {dsm_imask, {`num_non_irq{1'b1}}};
  assign opt_int_actv_set = int_actv_set_i & {dsm_imask, {`num_non_irq{1'b1}}};

`else

  assign opt_int_pend_set = int_pend_set_i;
  assign opt_int_pend     = int_pend_i;
  assign opt_int_actv_set = int_actv_set_i;

`endif

  assign int_pend_we = (opt_int_pend!=prev_int_pend);

  always @ (posedge fclk or negedge hreset_n)
    if (!hreset_n)
      prev_int_pend <= {INTS{1'b0}};
    else if (int_pend_we)
      prev_int_pend <= opt_int_pend;

  assign int_pend_pulse = opt_int_pend & ~prev_int_pend;

  always @ (pend_state or actv_state or opt_int_pend or int_pend_pulse or
            isr_entry_i or opt_int_pend_set or int_pend_clr_i or
            opt_int_actv_set or int_actv_clr_i or isr_exit_i)
    for (i = 0; i < INTS; i = i + 1)
      begin
        case ({ actv_state[i], pend_state[i] })
          2'b00 : begin
            next_pend_state[i] =  opt_int_pend[i] | opt_int_pend_set[i];
            next_actv_state[i] =  isr_entry_i[i]  | opt_int_actv_set[i];
          end
          2'b01 : begin
            next_pend_state[i] = ~(isr_entry_i[i] | int_pend_clr_i[i]);
            next_actv_state[i] =  isr_entry_i[i]  | opt_int_actv_set[i];
          end
          2'b10 : begin
            next_pend_state[i] =  int_pend_pulse[i] | opt_int_pend_set[i];
            next_actv_state[i] = ~(isr_exit_i[i]    | int_actv_clr_i[i]);
          end
          2'b11 : begin
            next_pend_state[i] = ~int_pend_clr_i[i];
            next_actv_state[i] = ~(isr_exit_i[i]    | int_actv_clr_i[i]);
          end
          default : begin
            next_pend_state[i] = 1'bx;
            next_actv_state[i] = 1'bx;
          end
        endcase
      end

  assign update_pend_state = (dpu_int_entry_i |
                              (|opt_int_pend)     | (|int_pend_pulse) |
                              (|opt_int_pend_set) | (|int_pend_clr_i));

  always @ (posedge fclk or negedge hreset_n)
    if (!hreset_n)
      pend_state <= {INTS{1'b0}};
    else if (update_pend_state)
      pend_state <= next_pend_state;

  assign update_actv_state = (dpu_int_entry_i | dpu_int_exit_i |
                              (|opt_int_actv_set) | (|int_actv_clr_i));

  always @ (posedge hclk or negedge hreset_n)
    if (!hreset_n)
      actv_state <= {INTS{1'b0}};
    else if (update_actv_state)
      actv_state <= next_actv_state;

  assign nxt_slp_wfe_pend    = update_pend_state & (|(next_pend_state & ~pend_state));
  assign update_slp_wfe_pend = nxt_slp_wfe_pend | slp_wfe_pend;

  always @ (posedge fclk or negedge hreset_n)
    if (!hreset_n)
      slp_wfe_pend <= 1'b0;
    else if (update_slp_wfe_pend)
      slp_wfe_pend <= nxt_slp_wfe_pend;

  assign pend_state_o   = pend_state;
  assign actv_state_o   = actv_state;
  assign slp_wfe_pend_o = slp_wfe_pend;

endmodule

`include "cm3_nvic_undefs.vp"
