`include "cm3_nvic_defs.vp"

module cm3_nvic_preempt
  (// Inputs
   fclk, hreset_n, irq_en_i, irq_lvl_i, int_pend_v_i, int_actv_v_i,
   nmi_active_i, curr_set_lvl_i, core_at_base_i, prem_mask_i, int_basepri_i,
   int_no_basepri_i, int_fmask_i, int_fmask_clr_i, int_pmask_i, int_stk_actv_i,
   dpu_int_exit_i, scr_wfeonpend_i, scr_deepsleep_i, slp_wfe_valid_i,
   slp_wfi_valid_i, slp_asleep_i, wic_ds_req_n_i, dbg_mon_int_allow_i,
   dbg_mon_exit_i,
   wic_ds_ack_n_o, wic_mask_o
  );

  parameter WIC_PRESENT     = 0;
  parameter WIC_LINES       = 3;
  parameter NUM_IRQ         = 16;
  parameter LVL_WIDTH       = 8;
  parameter IRQ_WIDTH       = NUM_IRQ * LVL_WIDTH;
  parameter NUM_INTS        = `num_non_irq + NUM_IRQ;

  parameter IRQ_LINES       = (WIC_LINES == 3) ? 1 : (WIC_LINES - 3);


  input                     fclk;                   // Free running clock
  input                     hreset_n;               // NVIC reset
  input                     wic_ds_req_n_i;         // WIC mode request
  input [NUM_IRQ-1:0]       irq_en_i;               // IRQ enables
  input [IRQ_WIDTH-1:0]     irq_lvl_i;              // IRQ Priority levels
  input                     nmi_active_i;           // NMI is active
  input [LVL_WIDTH-1:0]     curr_set_lvl_i;         // Current active ISR level
  input [LVL_WIDTH-1:0]     int_basepri_i;          // Base priority
  input                     int_stk_actv_i;         // Stack is active
  input [7:0]               prem_mask_i;            // Pre-emptive priority mask
  input                     int_fmask_i;            // Fault mask
  input                     int_pmask_i;            // Priority mask
  input                     core_at_base_i;         // In thread
  input                     int_no_basepri_i;       // int_basepri is disabled
  input                     slp_wfe_valid_i;        // Sleep due to WFE
  input                     slp_wfi_valid_i;        // Sleep due to WFI
  input                     scr_wfeonpend_i;        // Wake WFE on pended ints
  input                     slp_asleep_i;           // DPU is sleeping
  input                     dpu_int_exit_i;         // ISR has completed
  input                     int_fmask_clr_i;        // int_fmask clear
  input                     dbg_mon_exit_i;         // Dbg mon Sleep on exit case
  input                     dbg_mon_int_allow_i;    // dbg_mon_allow (-int_pmask)
  input [NUM_INTS-1:0]      int_pend_v_i;           // Pend Vector
  input [NUM_INTS-1:0]      int_actv_v_i;           // Reduced Active Vector
  input                     scr_deepsleep_i;        // deepsleep set

  output                    wic_ds_ack_n_o;         // WIC mode acknowledge
  output [WIC_LINES-1:0]    wic_mask_o;             // WIC mask sensitivity list

  wire                      mst_wic_en;             // Enable WIC
  reg                       wic_ds_ack;             // WIC mode acknowledge
  wire                      wic_ds_ack_we;          // - reg enable
  wire                      set_wic_ds_ack;         // - set reg
  wire                      clr_wic_ds_ack;         // - clear reg
  wire                      wic_ds_ack_n;           // - inverted and masked
  reg  [IRQ_LINES-1:0]      irq_pri_ok;             // IRQ pri higher then curr
  wire [WIC_LINES-1:0]      wic_mask;               // WIC mask sensitivity list
  wire [WIC_LINES:0]        mask;                   // - generate above
  wire                      masked_wfe;             // pmasked WFE without SEVONPEND
  wire                      wfeonpend_sleep;        // SEVONPEND WFE
  wire                      ign_basepri;            // flag to ignore base pri
  wire                      to_base_exit;           // return to thread
  wire                      irq_mask_en;            // allow irq masking
  wire [LVL_WIDTH-1:0]      comp_lvl;               // pri level to compare against
  reg  [LVL_WIDTH-1:0]      this_lvl;               // temp variable
  integer                   i,j;                    // loop index variables

  wire [WIC_LINES-1:0]      opt_wic_mask;           // logic optimisation term

`ifdef ARM_DSM

  wire [7:0]                dsm_num_irq;
  wire                      dsm_wic_present;
  wire [7:0]                dsm_wic_lines;
  reg  [239:0]              dsm_imask;
  reg  [242:0]              dsm_wmask;
  integer                   k;
  integer                   m;

  assign opt_wic_mask  = dsm_wic_present ? wic_mask    : {WIC_LINES{1'b0}};
  assign wic_ds_ack_n  = dsm_wic_present ? ~wic_ds_ack : 1'b1;
  assign mst_wic_en    = dsm_wic_present ? 1'b1        : 1'b0;

  always @ (dsm_num_irq)
    for (k = 0; k < 240; k = k + 1)
      dsm_imask[k] = (k < dsm_num_irq) ? 1'b1 : 1'b0;

  always @ (dsm_wic_lines)
    for (m = 0; m < 243; m = m + 1)
      dsm_wmask[m] = (m < dsm_wic_lines) ? 1'b1 : 1'b0;

`else

  assign opt_wic_mask = (WIC_PRESENT != 0) ? wic_mask    : {WIC_LINES{1'b0}};
  assign wic_ds_ack_n = (WIC_PRESENT != 0) ? ~wic_ds_ack : 1'b1;
  assign mst_wic_en   = (WIC_PRESENT != 0) ? 1'b1        : 1'b0;

`endif


  function a_lvl_lt_b_lvl;

    input [LVL_WIDTH-1:0]   a_lvl;
    input [LVL_WIDTH-1:0]   b_lvl;
    input [7:0]             mask;

    reg   [8:0]             full_a_lvl; // ALvl padded out to maximum
    reg   [8:0]             full_b_lvl; // BLvl padded out to maximum
    reg   [7:0]             mask_a_lvl; // Masked FullALvl
    reg   [7:0]             mask_b_lvl; // Masked FullBLvl
    reg                     a_lt_b;     // ALvl < BLvl
    reg                     a_l_b_76;   // ALvl[7:6] < BLvl[7:6]
    reg                     a_e_b_76;   // ALvl[7:6] = BLvl[7:6]
    reg                     a_l_b_54;   // ALvl[5:4] < BLvl[5:4]
    reg                     a_e_b_54;   // ALvl[5:4] = BLvl[5:4]
    reg                     a_l_b_32;   // ALvl[3:2] < BLvl[3:2]
    reg                     a_e_b_32;   // ALvl[3:2] = BLvl[3:2]
    reg                     a_l_b_10;   // ALvl[1:0] < BLvl[1:0]

    begin
      full_a_lvl[8:0] = {{9-LVL_WIDTH{1'b0}}, a_lvl} << (8 - LVL_WIDTH);
      full_b_lvl[8:0] = {{9-LVL_WIDTH{1'b0}}, b_lvl} << (8 - LVL_WIDTH);

      mask_a_lvl      = full_a_lvl[7:0] & mask;
      mask_b_lvl      = full_b_lvl[7:0] & mask;

      a_l_b_76 = ((~mask_a_lvl[7] & (mask_b_lvl[7] | (~mask_a_lvl[6] & mask_b_lvl[6]))) |
                  (~mask_a_lvl[6] & mask_b_lvl[7] & mask_b_lvl[6]));

      a_e_b_76 = mask_a_lvl[7:6] == mask_b_lvl[7:6];

      a_l_b_54 = ((~mask_a_lvl[5] & (mask_b_lvl[5] | (~mask_a_lvl[4] & mask_b_lvl[4]))) |
                  (~mask_a_lvl[4] & mask_b_lvl[5] & mask_b_lvl[4]));

      a_e_b_54 = mask_a_lvl[5:4] == mask_b_lvl[5:4];

      a_l_b_32 = ((~mask_a_lvl[3] & (mask_b_lvl[3] | (~mask_a_lvl[2] & mask_b_lvl[2]))) |
                  (~mask_a_lvl[2] & mask_b_lvl[3] & mask_b_lvl[2]));

      a_e_b_32 = mask_a_lvl[3:2] == mask_b_lvl[3:2];

      a_l_b_10 = ((~mask_a_lvl[1] & (mask_b_lvl[1] | (~mask_a_lvl[0] & mask_b_lvl[0]))) |
                  (~mask_a_lvl[0] & mask_b_lvl[1] & mask_b_lvl[0]));

      a_lt_b   = ((a_l_b_76) |
                  (a_e_b_76 & a_l_b_54) |
                  (a_e_b_76 & a_e_b_54 & a_l_b_32) |
                  (a_e_b_76 & a_e_b_54 & a_e_b_32 & a_l_b_10));

      a_lvl_lt_b_lvl = a_lt_b;

    end
  endfunction


  assign set_wic_ds_ack = (~wic_ds_req_n_i & ~slp_asleep_i &
                           wic_ds_ack_n & ~(slp_wfi_valid_i | slp_wfe_valid_i));

  assign clr_wic_ds_ack = (wic_ds_req_n_i & ~wic_ds_ack_n &
                           ~(slp_asleep_i & scr_deepsleep_i));

  assign wic_ds_ack_we  = (mst_wic_en & (set_wic_ds_ack | clr_wic_ds_ack));

  always @(posedge fclk or negedge hreset_n)
    if (~hreset_n)
      wic_ds_ack <= 1'b0;
    else if (wic_ds_ack_we)
      wic_ds_ack <= set_wic_ds_ack;


  assign masked_wfe      = slp_wfe_valid_i & int_pmask_i & ~scr_wfeonpend_i;

  assign wfeonpend_sleep = scr_wfeonpend_i & slp_wfe_valid_i;

  assign ign_basepri     = ((core_at_base_i | to_base_exit) &
                            int_no_basepri_i & ~masked_wfe &
                            (int_fmask_clr_i | ~int_fmask_i));

  assign to_base_exit    = ~int_stk_actv_i & dpu_int_exit_i;

  assign comp_lvl        = to_base_exit ? int_basepri_i : curr_set_lvl_i;

  assign irq_mask_en     = ~masked_wfe & (int_fmask_clr_i | ~int_fmask_i);

  always @(irq_lvl_i or comp_lvl or prem_mask_i)
    for(i=0;i<IRQ_LINES;i=i+1)
      begin
        for(j=0;j<LVL_WIDTH;j=j+1)
          this_lvl[j] = irq_lvl_i[(i*LVL_WIDTH) + j];

        irq_pri_ok[i] = a_lvl_lt_b_lvl(this_lvl,comp_lvl,prem_mask_i);
      end

  assign mask = {
    ({IRQ_LINES{irq_mask_en}} & irq_pri_ok & irq_en_i[IRQ_LINES-1:0]) |
    ({IRQ_LINES{wfeonpend_sleep}} & ~int_pend_v_i[`NVIC_IRQBASE+IRQ_LINES-1:`NVIC_IRQBASE]) |
    ({IRQ_LINES{ign_basepri}} & irq_en_i[IRQ_LINES-1:0]),

    (// PriMask case
     (dbg_mon_int_allow_i &  int_pmask_i & slp_wfi_valid_i) |
     (dbg_mon_int_allow_i & ~int_pmask_i) |
     dbg_mon_exit_i & to_base_exit & (~int_pmask_i | slp_wfi_valid_i)),

    ~nmi_active_i |
    to_base_exit |
    (wfeonpend_sleep & ~int_pend_v_i[`NVIC_NMINUM]),

    slp_wfe_valid_i };

`ifdef ARM_DSM
  assign wic_mask = mask[WIC_LINES-1:0] & {dsm_imask,3'b111} & dsm_wmask;
`else
  assign wic_mask = mask[WIC_LINES-1:0];
`endif

  assign wic_mask_o     = opt_wic_mask;
  assign wic_ds_ack_n_o = wic_ds_ack_n;

endmodule

`include "cm3_nvic_undefs.vp"
