module cm3_ppb_ahb_to_apb
  (// Inputs
   hclk, dreset_n, ahb_hsel_i, ahb_hmaster_i, ahb_haddr_i, ahb_htrans_i,
   ahb_hwrite_i, ahb_hready_i, ahb_hwdata_i, apb_prdata_i, apb_pready_i,
   apb_pslverr_i,
   ahb_hreadyout_o, ahb_hresp_o, ahb_hrdata_o, apb_paddr31_o, apb_paddr_o,
   apb_psel_o, apb_penable_o, apb_pwrite_o, apb_pwdata_o
  );


  input         hclk;             // System clock
  input         dreset_n;         // Debug reset

  input         ahb_hsel_i;       // AHB peripheral select
  input         ahb_hmaster_i;    // AHB master
  input  [19:2] ahb_haddr_i;      // AHB address bus
  input   [1:0] ahb_htrans_i;     // AHB transfer type
  input         ahb_hwrite_i;     // AHB transfer direction
  input         ahb_hready_i;     // AHB ready input
  input  [31:0] ahb_hwdata_i;     // AHB write data bus
  output        ahb_hreadyout_o;  // AHB slave ready output
  output        ahb_hresp_o;      // AHB slave response
  output [31:0] ahb_hrdata_o;     // AHB read data bus

  output        apb_psel_o;       // APB peripheral select
  output [19:2] apb_paddr_o;      // APB address bus
  output        apb_paddr31_o;    // APB address bus - bit31
  output        apb_penable_o;    // APB peripheral enable
  output        apb_pwrite_o;     // APB write
  output [31:0] apb_pwdata_o;     // APB write data bus
  input  [31:0] apb_prdata_i;     // APB read data bus
  input         apb_pready_i;     // APB ready
  input         apb_pslverr_i;    // APB slave error



  wire          apb_psel;
  reg    [19:2] apb_paddr;
  reg           apb_paddr31;
  reg           apb_penable;
  reg           apb_pwrite;

  wire   [31:0] ahb_hrdata;
  wire          ahb_hresp;
  wire          ahb_hreadyout;

  wire          ahb_trans_valid;    // Valid transfer on AHB
  reg           apb_trans_setup;    // APB in setup phase
  wire          apb_reg_we;         // Transaction register write enable
  wire          nxt_apb_pwrite;     // Next PWRITE
  wire          err_resp;           // First cycle of error response
  reg           err_resp_reg;       // Second cycle of error response
  wire          err_resp_reg_we;    // Error response reg write enable



  assign ahb_trans_valid = ahb_hsel_i & ahb_hready_i & ahb_htrans_i[1];



  assign apb_psel = apb_trans_setup | apb_penable;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      begin
        apb_trans_setup  <= 1'b0;
        apb_penable      <= 1'b0;
        apb_pwrite       <= 1'b0;
      end
    else if (apb_reg_we)
      begin
        apb_trans_setup  <= ahb_trans_valid;
        apb_penable      <= apb_trans_setup;
        apb_pwrite       <= nxt_apb_pwrite;
      end

  assign nxt_apb_pwrite = (ahb_trans_valid & ahb_hwrite_i) |
                          (apb_trans_setup & apb_pwrite);

  assign apb_reg_we = ahb_trans_valid |
                      apb_trans_setup |
                      (apb_penable & apb_pready_i);


  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      begin
        apb_paddr   <= {18{1'b0}};
        apb_paddr31 <= 1'b0;
      end
    else if (ahb_trans_valid)
      begin
        apb_paddr   <= ahb_haddr_i;
        apb_paddr31 <= ahb_hmaster_i;
      end



  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      err_resp_reg <= 1'b0;
    else if (err_resp_reg_we)
      err_resp_reg <= err_resp;

  assign err_resp        = apb_penable & apb_pready_i & apb_pslverr_i;
  assign err_resp_reg_we = err_resp | err_resp_reg;

  assign ahb_hrdata    = apb_prdata_i;
  assign ahb_hresp     = err_resp | err_resp_reg;
  assign ahb_hreadyout = ~(apb_trans_setup | (apb_penable & ~apb_pready_i) | err_resp);



  assign apb_psel_o      = apb_psel;
  assign apb_paddr_o     = apb_paddr;
  assign apb_paddr31_o   = apb_paddr31;
  assign apb_penable_o   = apb_penable;
  assign apb_pwrite_o    = apb_pwrite;
  assign apb_pwdata_o    = ahb_hwdata_i;

  assign ahb_hrdata_o    = ahb_hrdata;
  assign ahb_hresp_o     = ahb_hresp;
  assign ahb_hreadyout_o = ahb_hreadyout;



`ifdef AHB_ASSERT_ON

  wire [19:0] apb_paddr_pc;
  assign      apb_paddr_pc = {apb_paddr, 2'b00};

  ApbPC #(20, 32, 1, "ppb.log")
    uApbPC
    (// Inputs
     .PCLK                        (hclk),
     .PCLKEN                      (1'b1),
     .PRESETn                     (dreset_n),
     .PSEL                        (apb_psel),
     .PENABLE                     (apb_penable),
     .PREADY                      (apb_pready_i),
     .PSLVERR                     (apb_pslverr_i),
     .PADDR                       (apb_paddr_pc),
     .PWRITE                      (apb_pwrite),
     .PWDATA                      (ahb_hwdata_i),
     .PRDATA                      (apb_prdata_i)
    );

`endif

endmodule
