module cm3_ppb_decoder
  (// Inputs
   dreset_n, hclk, ppb_lock_i, ppb_hmaster_i, ppb_htrans_i, ppb_hprot_i,
   ppb_haddr_i, ppb_hwrite_i, itm_ppb_hreadyout_i, dwt_ppb_hreadyout_i,
   fpb_ppb_hreadyout_i, ext_ppb_hreadyout_i, nvic_ppb_hreadyout_i,
   mpu_ppb_hreadyout_i, itm_ppb_hresp_i, dwt_ppb_hresp_i, fpb_ppb_hresp_i,
   ext_ppb_hresp_i, nvic_ppb_hresp_i, mpu_ppb_hresp_i, itm_ppb_hrdata_i,
   dwt_ppb_hrdata_i, fpb_ppb_hrdata_i, ext_ppb_hrdata_i, nvic_ppb_hrdata_i,
   mpu_ppb_hrdata_i,
   ppb_hsel_itm_o, ppb_hsel_dwt_o, ppb_hsel_fpb_o, ppb_hsel_ext_o,
   ppb_hsel_nvic_o, ppb_hsel_mpu_o, ppb_pclken_o, ppb_hrdata_o,
   ppb_hready_o, ppb_hresp_o
  );


  input         hclk;
  input         dreset_n;

  input  [31:0] ppb_haddr_i;
  input         ppb_hmaster_i;
  input   [1:0] ppb_htrans_i;
  input   [3:0] ppb_hprot_i;
  input         ppb_hwrite_i;
  output [31:0] ppb_hrdata_o;
  output        ppb_hready_o;
  output        ppb_hresp_o;
  input   [5:0] ppb_lock_i;             // Lock selected PPB transactions

  input         itm_ppb_hreadyout_i;
  input         itm_ppb_hresp_i;
  input  [31:0] itm_ppb_hrdata_i;

  input         dwt_ppb_hreadyout_i;
  input         dwt_ppb_hresp_i;
  input  [31:0] dwt_ppb_hrdata_i;

  input         fpb_ppb_hreadyout_i;
  input         fpb_ppb_hresp_i;
  input  [31:0] fpb_ppb_hrdata_i;

  input         mpu_ppb_hreadyout_i;
  input         mpu_ppb_hresp_i;
  input  [31:0] mpu_ppb_hrdata_i;

  input         nvic_ppb_hreadyout_i;
  input         nvic_ppb_hresp_i;
  input  [31:0] nvic_ppb_hrdata_i;

  input         ext_ppb_hreadyout_i;
  input         ext_ppb_hresp_i;
  input  [31:0] ext_ppb_hrdata_i;

  output        ppb_hsel_itm_o;
  output        ppb_hsel_dwt_o;
  output        ppb_hsel_fpb_o;
  output        ppb_hsel_ext_o;
  output        ppb_hsel_nvic_o;
  output        ppb_hsel_mpu_o;

  output        ppb_pclken_o;



  wire          ppb_trans;            // Transaction detection
  wire    [6:0] ppb_sel;              // Internal PPB slave select
  wire          nvic_addrs;           // Address is within NVIC range
  wire          mpu_addrs;            // Address is within MPU range
  wire          dbg_addrs;            // Address is within debug regiser range

  wire          ppb_hsel_nvic_icsr;   // NVIC ICSR reg HSEL, regardless of lock
  wire          ppb_hsel_nvic_stir;   // NVIC STIR reg HSEL, regardless of lock
  wire          ppb_hsel_dbg_dcrd;    // Debug DCRD reg HSEL, regardless of lock
  wire          ppb_hsel_itm;         // ITM HSEL
  wire          ppb_hsel_dwt;         // DWT HSEL
  wire          ppb_hsel_fpb;         // FPB HSEL
  wire          ppb_hsel_ext;         // External HSEL
  wire          ppb_hsel_nvic;        // NVIC HSEL
  wire          ppb_hsel_mpu;         // MPU HSEL

  wire          dcore_priv_write;     // PPB write is privileged from DCore
  reg     [5:0] ppb_lock_reg;         // Registered PPB lock
  wire          block_itm;            // Block ITM access
  wire          block_dwt;            // Block DWT access
  wire          block_fpb;            // Block FPB access
  wire          block_nvic;           // Block NVIC access
  wire          block_mpu;            // Block MPU access
  wire          block_dbg;            // Block debug register access

  reg     [6:0] ppb_sel_data;         // Data phase select
  wire          ppb_sel_data_we;      // Data phase select write enable
  wire   [31:0] ppb_hrdata;           // Muxed slave HRDATA
  wire          ppb_hready;           // Muxed slave HREADYOUT; bus HREADY
  wire          ppb_hresp;            // Muxed slave HRESP

  wire          ppb_priv_fault;       // Insert fault
  reg           dec_ppb_hreadyout;    // Decoder HREADYOUT response
  wire          nxt_dec_ppb_hreadyout;
  reg           dec_ppb_hresp;        // Decoder HRESP response
  wire          dec_ppb_hreadyout_we; // Decoder HREADYOUT write enable
  wire          dec_ppb_hresp_we;     // Decoder HRESP write enable

  reg           ppb_pclken;



  assign ppb_trans   = ppb_htrans_i[1];

  assign nvic_addrs  = (ppb_haddr_i[19:12] == 8'h0E);
  assign mpu_addrs   = (ppb_haddr_i[11:6]  == 6'b1101_10) & (|ppb_haddr_i[5:4]);
  assign dbg_addrs   = (ppb_haddr_i[11:4]  == 8'hDF);

  assign ppb_sel[0] = (ppb_haddr_i[19:12] == 8'b00000000);   // ITM
  assign ppb_sel[1] = (ppb_haddr_i[19:12] == 8'b00000001);   // DWT
  assign ppb_sel[2] = (ppb_haddr_i[19:12] == 8'b00000010);   // FPB
  assign ppb_sel[3] = (|ppb_haddr_i[19:18]);                 // External
  assign ppb_sel[4] = nvic_addrs & ~mpu_addrs & ~dbg_addrs;  // NVIC
  assign ppb_sel[5] = nvic_addrs &  mpu_addrs;               // MPU
  assign ppb_sel[6] = nvic_addrs &  dbg_addrs;               // Debug



  assign ppb_hsel_nvic_stir = nvic_addrs & (ppb_haddr_i[11:2] == 10'b1111000000);
  assign ppb_hsel_nvic_icsr = nvic_addrs & (ppb_haddr_i[11:2] == 10'b1101000001);
  assign ppb_hsel_dbg_dcrd  = nvic_addrs & (ppb_haddr_i[11:2] == 10'b1101111110);



  assign dcore_priv_write = ~ppb_hmaster_i & ppb_hprot_i[1] & ppb_hwrite_i;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      ppb_lock_reg <= 6'b000000;
    else
      ppb_lock_reg <= ppb_lock_i;

  assign block_itm  = dcore_priv_write & ppb_lock_reg[0];
  assign block_dwt  = dcore_priv_write & ppb_lock_reg[1];
  assign block_fpb  = dcore_priv_write & ppb_lock_reg[2];
  assign block_nvic = dcore_priv_write & ppb_lock_reg[3] &
                        ~ppb_hsel_nvic_icsr & ~ppb_hsel_nvic_stir;
  assign block_mpu  = dcore_priv_write & ppb_lock_reg[4];
  assign block_dbg  = dcore_priv_write & ppb_lock_reg[5] & ~ppb_hsel_dbg_dcrd;



  assign ppb_hsel_itm  = ppb_sel[0] & ppb_trans & ~block_itm;
  assign ppb_hsel_dwt  = ppb_sel[1] & ppb_trans & ~block_dwt & ppb_hprot_i[1] & ppb_hready;
  assign ppb_hsel_fpb  = ppb_sel[2] & ppb_trans & ~block_fpb & ppb_hprot_i[1] & ppb_hready;
  assign ppb_hsel_ext  = ppb_sel[3] & ppb_trans & ppb_hprot_i[1] & ppb_hready;
  assign ppb_hsel_nvic = ((ppb_sel[4] & ~block_nvic) | (ppb_sel[6] & ~block_dbg)) & ppb_trans;
  assign ppb_hsel_mpu  = ppb_sel[5] & ppb_trans & ~block_mpu;



  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      ppb_sel_data <= 7'b0000000;
    else if (ppb_sel_data_we)
      ppb_sel_data <= ppb_sel;

  assign ppb_sel_data_we = ppb_hready & (ppb_htrans_i[1] | (|ppb_sel_data));

  assign ppb_hrdata = ({32{ppb_sel_data[0]}} &  itm_ppb_hrdata_i) |
                      ({32{ppb_sel_data[1]}} &  dwt_ppb_hrdata_i) |
                      ({32{ppb_sel_data[2]}} &  fpb_ppb_hrdata_i) |
                      ({32{ppb_sel_data[3]}} &  ext_ppb_hrdata_i) |
                      ({32{ppb_sel_data[4]}} & nvic_ppb_hrdata_i) |
                      ({32{ppb_sel_data[5]}} &  mpu_ppb_hrdata_i) |
                      ({32{ppb_sel_data[6]}} & nvic_ppb_hrdata_i);

  assign ppb_hready = (ppb_sel_data[6] & nvic_ppb_hreadyout_i)                    |
                      (ppb_sel_data[5] & mpu_ppb_hreadyout_i)                     |
                      (ppb_sel_data[4] & nvic_ppb_hreadyout_i)                    |
                      (ppb_sel_data[3] & ext_ppb_hreadyout_i & dec_ppb_hreadyout) |
                      (ppb_sel_data[2] & dec_ppb_hreadyout)                       |
                      (ppb_sel_data[1] & dec_ppb_hreadyout)                       |
                      (ppb_sel_data[0] & dec_ppb_hreadyout)                       |
                      ~(|ppb_sel_data[6:0]);

  assign ppb_hresp  = (ppb_sel_data[6] & nvic_ppb_hresp_i)                  |
                      (ppb_sel_data[5] & mpu_ppb_hresp_i)                   |
                      (ppb_sel_data[4] & nvic_ppb_hresp_i)                  |
                      (ppb_sel_data[3] & (ext_ppb_hresp_i | dec_ppb_hresp)) |
                      (ppb_sel_data[2] & dec_ppb_hresp)                     |
                      (ppb_sel_data[1] & dec_ppb_hresp)                     |
                      (ppb_sel_data[0] & dec_ppb_hresp);



  assign ppb_priv_fault = ppb_htrans_i[1] &
                          ~ppb_hprot_i[1] &
                          (|ppb_sel[3:1]) &
                          ppb_hready;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      dec_ppb_hreadyout <= 1'b1;
    else if (dec_ppb_hreadyout_we)
      dec_ppb_hreadyout <= nxt_dec_ppb_hreadyout;

  assign dec_ppb_hreadyout_we  = ppb_priv_fault | ~dec_ppb_hreadyout;
  assign nxt_dec_ppb_hreadyout = ~dec_ppb_hreadyout;

  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      dec_ppb_hresp <= 1'b0;
    else if (dec_ppb_hresp_we)
      dec_ppb_hresp <= ppb_priv_fault;

  assign dec_ppb_hresp_we  = ppb_priv_fault | (ppb_hready & dec_ppb_hresp);



  always @ (posedge hclk or negedge dreset_n)
    if (!dreset_n)
      ppb_pclken <= 1'b1;
    else
      ppb_pclken <= ppb_hwrite_i & ppb_trans;



  assign ppb_hsel_itm_o  = ppb_hsel_itm;
  assign ppb_hsel_dwt_o  = ppb_hsel_dwt;
  assign ppb_hsel_fpb_o  = ppb_hsel_fpb;
  assign ppb_hsel_ext_o  = ppb_hsel_ext;
  assign ppb_hsel_nvic_o = ppb_hsel_nvic;
  assign ppb_hsel_mpu_o  = ppb_hsel_mpu;

  assign ppb_hrdata_o    = ppb_hrdata;
  assign ppb_hready_o    = ppb_hready;
  assign ppb_hresp_o     = ppb_hresp;

  assign ppb_pclken_o    = ppb_pclken;



`ifdef ARM_ASSERT_ON

  assert_never #(0,0,
    "ppb_trans must only be asserted for ppb_haddr_i range 0xE0000000 - 0xE00FFFFF.")
    ovl_ppb_addr_valid(hclk, dreset_n,
      (ppb_trans & ((ppb_haddr_i >= 32'hE0100000) | (ppb_haddr_i <= 32'hDFFFFFFF))));

  assert_never #(0,0,
    "ppb_hsel_itm should only be asserted for ppb_haddr_i range 0xE0000000 - 0xE0000FFF.")
      ovl_itm_addr_valid(hclk, dreset_n,
        (ppb_hsel_itm & ((ppb_haddr_i > 32'hE0000FFF) | (ppb_haddr_i < 32'hE0000000))));

  assert_never #(0,0,
    "ppb_hsel_dwt should only be asserted for ppb_haddr_i range 0xE0001000 - 0xE0001FFF.")
      ovl_dwt_addr_valid(hclk, dreset_n,
        (ppb_hsel_dwt & ((ppb_haddr_i > 32'hE0001FFF) | (ppb_haddr_i < 32'hE0001000))));

  assert_never #(0,0,
    "ppb_hsel_fpb should only be asserted for ppb_haddr_i range 0xE0002000 - 0xE0002FFF.")
      ovl_fpb_addr_valid(hclk, dreset_n,
        (ppb_hsel_fpb & ((ppb_haddr_i > 32'hE0002FFF) | (ppb_haddr_i < 32'hE0002000))));

  assert_never #(0,0,
    "ppb_hsel_nvic should only be asserted for ppb_haddr_i range 0xE000E000 - 0xE000EFFF.")
      ovl_nvic_addr_valid(hclk, dreset_n,
        (ppb_hsel_nvic & ((ppb_haddr_i > 32'hE000EFFF) | (ppb_haddr_i < 32'hE000E000))));

  assert_never #(0,0,
    "ppb_hsel_nvic_icsr should only be asserted for ppb_haddr_i = 0xE000ED04.")
      ovl_icsr_addr_valid(hclk, dreset_n,
        (ppb_hsel_nvic_icsr & (ppb_haddr_i[31:2] != 30'h38003b41)));

  assert_never #(0,0,
    "ppb_hsel_nvic_stir should only be asserted for ppb_haddr_i = 0xE000EF00.")
      ovl_stir_addr_valid(hclk, dreset_n,
        (ppb_hsel_nvic_stir & (ppb_haddr_i[31:2] != 30'h38003BC0)));

  assert_never #(0,0,
    "ppb_hsel_dbg_dcrd should only be asserted for ppb_haddr_i = 0xE000EDF8.")
      ovl_dcrd_addr_valid(hclk, dreset_n,
        (ppb_hsel_dbg_dcrd & (ppb_haddr_i[31:2] != 30'h38003b7e)));

  assert_never #(0,0,
    "ppb_hsel_ext should only be asserted for ppb_haddr_i range 0xE0040000 - 0xE00FFFFF.")
      ovl_ext_addr_valid(hclk, dreset_n,
        (ppb_hsel_ext & ((ppb_haddr_i > 32'hE00FFFFF) | (ppb_haddr_i < 32'hE0040000))));

  assert_zero_one_hot #(0,7,0,"ppb_sel must be one-hot")
    ovl_onehot_ppb_sel(hclk, dreset_n, ppb_sel[6:0]);

  assert_zero_one_hot #(0,7,0,"ppb_sel_data must be one-hot")
    ovl_onehot_ppb_sel_data(hclk, dreset_n, ppb_sel_data[6:0]);

  assert_never #(0,0,"Unused HRESP/HREADYOUT nets driven to unexpected response.")
    ovl_unused_check(hclk, dreset_n,
      (~itm_ppb_hreadyout_i | ~dwt_ppb_hreadyout_i | ~fpb_ppb_hreadyout_i |
        itm_ppb_hresp_i     | dwt_ppb_hresp_i      | fpb_ppb_hresp_i));

  assert_never #(0,0,"dec_ppb_hreadyout must only be deasserted when dec_ppb_hresp is set.")
    ovl_dec_hready_check(hclk, dreset_n, (~dec_ppb_hreadyout & ~dec_ppb_hresp));

  assert_never #(0,0,"DAP->ITM accesses must not be blocked by ppb_lock_i.")
    ovl_block_itm_check(hclk, dreset_n, (ppb_hmaster_i & block_itm));
  assert_never #(0,0,"DAP->DWT accesses must not be blocked by ppb_lock_i.")
    ovl_block_dwt_check(hclk, dreset_n, (ppb_hmaster_i & block_dwt));
  assert_never #(0,0,"DAP->FPB accesses must not be blocked by ppb_lock_i.")
    ovl_block_fpb_check(hclk, dreset_n, (ppb_hmaster_i & block_fpb));
  assert_never #(0,0,"DAP->NVIC accesses must not be blocked by ppb_lock_i.")
    ovl_block_nvic_check(hclk, dreset_n, (ppb_hmaster_i & block_nvic));
  assert_never #(0,0,"DAP->MPU accesses must not be blocked by ppb_lock_i.")
    ovl_block_mpu_check(hclk, dreset_n, (ppb_hmaster_i & block_mpu));
  assert_never #(0,0,"DAP->DEBUG accesses must not be blocked by ppb_lock_i.")
    ovl_block_db_check(hclk, dreset_n, (ppb_hmaster_i & block_dbg));

  assert_never #(0,0,"ITM read accesses must not be blocked by ppb_lock_i.")
    ovl_block_itm_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_itm));
  assert_never #(0,0,"DWT read accesses must not be blocked by ppb_lock_i.")
    ovl_block_dwt_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_dwt));
  assert_never #(0,0,"FPB read accesses must not be blocked by ppb_lock_i.")
    ovl_block_fpb_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_fpb));
  assert_never #(0,0,"NVIC read accesses must not be blocked by ppb_lock_i.")
    ovl_block_nvic_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_nvic));
  assert_never #(0,0,"MPU read accesses must not be blocked by ppb_lock_i.")
    ovl_block_mpu_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_mpu));
  assert_never #(0,0,"DEBUG read accesses must not be blocked by ppb_lock_i.")
    ovl_block_dbg_rd_check(hclk, dreset_n, (~ppb_hwrite_i & block_dbg));

  assert_never #(0,0,
    "Interrupt Control State Register must not be blocked by ppb_lock_reg[3].")
      ovl_block_icsr_check(hclk, dreset_n,
        (ppb_lock_reg[3] & ppb_hsel_nvic_icsr & block_nvic));

  assert_never #(0,0,
    "Interrupt Control State Register must not be blocked by ppb_lock_reg[3].")
      ovl_block_stir_check(hclk, dreset_n,
        (ppb_lock_reg[3] & ppb_hsel_nvic_stir & block_nvic));

  assert_never #(0,0,
    "Debug Control Register Data must not be blocked by ppb_lock_reg[5].")
      ovl_block_dcrd_check(hclk, dreset_n,
        (ppb_lock_reg[5] & ppb_hsel_dbg_dcrd & block_dbg));

`endif

endmodule
