module cm3_tpiu (
  CLK, CLKEN, TRACECLKIN, RESETn, TRESETn, PWRITE, PENABLE, PSEL,
  PADDR, PWDATA, ATVALID1S, ATID1S, ATDATA1S, ATVALID2S,
  ATID2S, ATDATA2S, SYNCREQ, MAXPORTSIZE, TRIGGER,
  PRDATA, ATREADY1S, ATREADY2S, TRACECLK,
  TRACEDATA, TRACESWO, SWOACTIVE, TPIUACTV, TPIUBAUD
  );

  parameter TRACE_LVL     = 1;
  parameter CP_PRESENT    = 0;  // Affects ID registers

  parameter PORT1_PRESENT = TRACE_LVL > 0 ? 1 : 0;
  parameter PORT2_PRESENT = TRACE_LVL > 1 ? 1 : 0;

  input         CLK;           // ATB and APB clock
  input         CLKEN;         // ATB clock enable
  input         TRACECLKIN;    // TRACEIN clock
  input         RESETn;        // ATB asychronous reset active low
  input         TRESETn;       // TRACECLKIN asychronous reset active low

  input         PWRITE;        // APB write or read
  input         PENABLE;       // APB enable
  input         PSEL;          // APB slave select
  input [11:2]  PADDR;         // APB address bus
  input [12:0]  PWDATA;        // APB write data bus bits [12:0] only

  input         ATVALID1S;     // ATB valid signal
  input [6:0]   ATID1S;        // ATB ID
  input [7:0]   ATDATA1S;      // ATB data

  input         ATVALID2S;     // ATB valid signal
  input [6:0]   ATID2S;        // ATB ID
  input [7:0]   ATDATA2S;      // ATB data
  input         SYNCREQ;       // Global synchronisation request
  input [1:0]   MAXPORTSIZE;   // External tie-off (binary)
  input         TRIGGER;       // Trigger from ETM

  output [31:0] PRDATA;        // APB read data bus
  output        ATREADY1S;     // ATB ready signal
  output        ATREADY2S;     // ATB ready signal

  output        TRACECLK;      // TRACECLK output
  output [3:0]  TRACEDATA;     // Trace Data
  output        TRACESWO;      // Single wire output
  output        SWOACTIVE;     // SWO mode selected
  output        TPIUACTV;      // Trace output is active
  output        TPIUBAUD;      // SWO mode frequency indicator (in TRACECLK domain)

  wire [7:0]    fifo1_data0;    // Fifo1 data byte 0
  wire [7:0]    fifo1_data1;    // Fifo1 data byte 1
  wire [7:0]    fifo1_data2;    // Fifo1 data byte 2
  wire [7:0]    fifo1_data3;    // Fifo1 data byte 3
  wire          fifo1_valid;    // Fifo1 data valid
  wire [7:0]    fifo1_read_data;          // Fifo1 data output
  wire [2:0]    read_pointer1_gray_async; // Fifo1 read pointer (TRACECLKIN)
  wire [2:0]    read_pointer1_gray;       // Fifo1 read pointer (CLK)
  wire [2:0]    write_pointer1_gray_async;// Fifo1 write pointer (CLK)
  wire [1:0]    write_pointer1_1_0;       // Fifo1 write pointer (CLK)
  wire [2:0]    write_pointer1_gray;      // Fifo1 write pointer (TRACECLKIN)

  wire [7:0]    fifo2_data0;    // Fifo2 data byte 0
  wire [7:0]    fifo2_data1;    // Fifo2 data byte 1
  wire [7:0]    fifo2_data2;    // Fifo2 data byte 2
  wire [7:0]    fifo2_data3;    // Fifo2 data byte 3

  wire          fifo2_valid;    // Fifo2 data valid
  wire [7:0]    fifo2_read_data;// Fifo2 data output
  wire          fifo2_pop;      // Byte taken from fifo2 by formatter
  wire [2:0]    read_pointer2_gray_async; // Fifo2 read pointer (TRACECLKIN)
  wire [2:0]    read_pointer2_gray;       // Fifo2 read pointer (CLK)
  wire [2:0]    write_pointer2_gray_async;// Fifo2 write pointer (CLK)
  wire [1:0]    write_pointer2_1_0;       // Fifo2 write pointer (CLK)
  wire [2:0]    write_pointer2_gray;      // Fifo2 write pointer (TRACECLKIN)

  wire [7:0]    format_data;              // Formatter output byte
  wire          format_data_ready;        // Formatter byte accepted by TaceOut
  wire          format_data_valid;        // Formatter output valid
  wire          fifo_test_mode;           // Fifo test mode (no output)
  wire          integration_fifo1_read;   // Fifo test read from 1
  wire          integration_fifo2_read;   // Fifo test read from 2

  wire [1:0]    prg_port_mode;            // Port mode selection
  wire [2:0]    prg_port_size_decoded;    // Synchronous mode port size
  wire [12:0]   prg_baud_div;             // Async mode divisor
  wire          baud_div_reset_async;     // Reload counter when written
  wire          baud_div_reset;           // Synchronised to TraceClockIn
  wire          prg_formatter_on;         // Use formatter (if fitted)

  wire          fifo1_ready;   // From formatter, before Int Test
  wire          fifo2_ready;   // From formatter, before Int Test
  wire          port1_valid;   // Inputs masked by Int Test Mode
  wire          port2_valid;   // Inputs masked by Int Test Mode

  wire          sync_valid_async; // Sync request to sync to TRACECLKIN
  wire          sync_valid;       // Sync request in TRACECLK domain
  wire          sync_ack_async;   // Sync ack to synchroniser
  wire          sync_ack;         // Sync ack, resynced to CLK

  wire          trigger_valid_async; // Trig request to sync to TRACECLKIN
  wire          trigger_valid;       // Trig request in TRACECLK domain
  wire          trigger_ack_async;   // Trig ack to synchroniser
  wire          trigger_ack;         // Trig ack, resynced to CLK

  wire          fifo1_pop;           // Btye taken from fifo1 by formatter
  wire          fifo_empty1;
  wire          fifo_empty2;
  wire          trace_out_idle_async;// TraceOut is idle
  wire          trace_out_idle;      // Syncronised to CLK
  wire          tpiu_baud_async;     // TPIU_baud in TRACECLKIN


  wire          opt_fifo1_ready;
  wire [2:0]    opt_write_pointer1_gray_async;
  wire          opt_fifo_empty1;
  wire [1:0]    opt_write_pointer1_1_0;
  wire [7:0]    opt_fifo1_data0;
  wire [7:0]    opt_fifo1_data1;
  wire [7:0]    opt_fifo1_data2;
  wire [7:0]    opt_fifo1_data3;
  wire          opt_fifo2_ready;
  wire [2:0]    opt_write_pointer2_gray_async;
  wire          opt_fifo_empty2;
  wire [1:0]    opt_write_pointer2_1_0;
  wire [7:0]    opt_fifo2_data0;
  wire [7:0]    opt_fifo2_data1;
  wire [7:0]    opt_fifo2_data2;
  wire          opt_fifo1_valid;
  wire [2:0]    iread_pointer1_gray_async;
  wire [7:0]    opt_fifo1_read_data;
  wire          opt_fifo2_valid;
  wire [2:0]    iread_pointer2_gray_async;
  wire [7:0]    opt_fifo2_read_data;

  wire [2:0]    read_pointer1_gray_async_cdc;
  wire [2:0]    write_pointer1_gray_async_cdc;
  wire [2:0]    read_pointer2_gray_async_cdc;
  wire [2:0]    write_pointer2_gray_async_cdc;
  wire          baud_div_reset_async_cdc;
  wire          sync_valid_async_cdc;
  wire          sync_ack_async_cdc;
  wire          trigger_valid_async_cdc;
  wire          trigger_ack_async_cdc;
  wire          trace_out_idle_async_cdc;
  wire          tpiu_baud_async_cdc;

  cm3_tpiu_apb_if #(TRACE_LVL,CP_PRESENT)
    u_cm3_tpiu_apb_if
    (
     .clk                       (CLK),
     .clk_en                    (CLKEN),
     .reset_n                   (RESETn),
     .apb_write_i               (PWRITE),
     .apb_enable_i              (PENABLE),
     .apb_sel_i                 (PSEL),
     .apb_addr_i                (PADDR[11:2]),
     .max_port_size_i           (MAXPORTSIZE),
     .atb_valid1_i              (ATVALID1S),
     .atb_valid2_i              (ATVALID2S),
     .apb_wdata_i               (PWDATA),
     .fifo1_ready_i             (opt_fifo1_ready),
     .fifo2_ready_i             (opt_fifo2_ready),
     .sync_req_i                (SYNCREQ),
     .trigger_i                 (TRIGGER),
     .sync_ack_i                (sync_ack),
     .trigger_ack_i             (trigger_ack),
     .fifo_empty1_i             (opt_fifo_empty1),
     .fifo_empty2_i             (opt_fifo_empty2),
     .trace_out_idle_i          (trace_out_idle),
     .fifo1_data0_i             (opt_fifo1_data0[7:0]),
     .fifo1_data1_i             (opt_fifo1_data1[7:0]),
     .fifo1_data2_i             (opt_fifo1_data2[7:0]),
     .fifo2_data0_i             (opt_fifo2_data0[7:0]),
     .fifo2_data1_i             (opt_fifo2_data1[7:0]),
     .fifo2_data2_i             (opt_fifo2_data2[7:0]),
     .write_pointer1_1_0_i      (opt_write_pointer1_1_0),
     .write_pointer2_1_0_i      (opt_write_pointer2_1_0),
     .prg_port_size_decoded_o   (prg_port_size_decoded),
     .prg_port_mode_o           (prg_port_mode),
     .prg_baud_div_o            (prg_baud_div),
     .baud_div_reset_o          (baud_div_reset_async),
     .prg_formatter_on_o        (prg_formatter_on),
     .apb_rdata_o               (PRDATA),
     .atb_ready1_o              (ATREADY1S),
     .atb_ready2_o              (ATREADY2S),
     .port1_valid_o             (port1_valid),
     .port2_valid_o             (port2_valid),
     .sync_valid_async_o        (sync_valid_async),
     .trigger_valid_async_o     (trigger_valid_async),
     .fifo_test_mode_o          (fifo_test_mode),
     .integration_fifo1_read_o  (integration_fifo1_read),
     .integration_fifo2_read_o  (integration_fifo2_read),
     .tpiu_active_o             (TPIUACTV),
     .swo_active_o              (SWOACTIVE)
     );


  cm3_tpiu_atb_fifo #(PORT1_PRESENT)
    u_cm3_tpiu_atb_fifo1
    (
     .clk                       (CLK),
     .clk_en                    (CLKEN),
     .reset_n                   (RESETn),
     .atb_data_i                (ATDATA1S),
     .atb_valid_i               (port1_valid),
     .read_pointer_gray_i       (read_pointer1_gray[2:0]),
     .fifo_test_mode_i          (fifo_test_mode),
     .test_read_i               (integration_fifo1_read),
     .atb_ready_o               (fifo1_ready),
     .fifo_empty_o              (fifo_empty1),
     .write_pointer_gray_async_o(write_pointer1_gray_async[2:0]),
     .write_pointer_1_0_o       (write_pointer1_1_0[1:0]),
     .fifo_data0_o              (fifo1_data0[7:0]),
     .fifo_data1_o              (fifo1_data1[7:0]),
     .fifo_data2_o              (fifo1_data2[7:0]),
     .fifo_data3_o              (fifo1_data3[7:0]));

  assign opt_fifo1_ready  = PORT1_PRESENT != 0 ? fifo1_ready : 1'b0;
  assign opt_write_pointer1_gray_async  = PORT1_PRESENT != 0 ? write_pointer1_gray_async  : 3'b000;
  assign opt_fifo_empty1  = PORT1_PRESENT != 0 ? fifo_empty1  : 1'b1;
  assign opt_write_pointer1_1_0 = PORT1_PRESENT != 0 ? write_pointer1_1_0 : 2'b00;
  assign opt_fifo1_data0  = PORT1_PRESENT != 0 ? fifo1_data0  : {8{1'b0}};
  assign opt_fifo1_data1  = PORT1_PRESENT != 0 ? fifo1_data1  : {8{1'b0}};
  assign opt_fifo1_data2  = PORT1_PRESENT != 0 ? fifo1_data2  : {8{1'b0}};
  assign opt_fifo1_data3  = PORT1_PRESENT != 0 ? fifo1_data3  : {8{1'b0}};

  cm3_tpiu_atb_fifo #(PORT2_PRESENT)
    u_cm3_tpiu_atb_fifo2
    (
     .clk                       (CLK),
     .clk_en                    (CLKEN),
     .reset_n                   (RESETn),
     .atb_data_i                (ATDATA2S),
     .atb_valid_i               (port2_valid),
     .read_pointer_gray_i       (read_pointer2_gray[2:0]),
     .fifo_test_mode_i          (fifo_test_mode),
     .test_read_i               (integration_fifo2_read),
     .atb_ready_o               (fifo2_ready),
     .fifo_empty_o              (fifo_empty2),
     .write_pointer_gray_async_o(write_pointer2_gray_async[2:0]),
     .write_pointer_1_0_o       (write_pointer2_1_0[1:0]),
     .fifo_data0_o              (fifo2_data0[7:0]),
     .fifo_data1_o              (fifo2_data1[7:0]),
     .fifo_data2_o              (fifo2_data2[7:0]),
     .fifo_data3_o              (fifo2_data3[7:0]));

  assign opt_fifo2_ready = PORT2_PRESENT != 0 ? fifo2_ready : 1'b0;
  assign opt_write_pointer2_gray_async  = PORT2_PRESENT != 0 ? write_pointer2_gray_async  : 3'b000;
  assign opt_fifo_empty2  = PORT2_PRESENT != 0 ? fifo_empty2  : 1'b1;
  assign opt_write_pointer2_1_0 = PORT2_PRESENT != 0 ? write_pointer2_1_0 : 2'b00;
  assign opt_fifo2_data0  = PORT2_PRESENT != 0 ? fifo2_data0  : {8{1'b0}};
  assign opt_fifo2_data1  = PORT2_PRESENT != 0 ? fifo2_data1  : {8{1'b0}};
  assign opt_fifo2_data2  = PORT2_PRESENT != 0 ? fifo2_data2  : {8{1'b0}};


  cm3_tpiu_trace_fifo #(PORT1_PRESENT)
    u_cm3_tpiu_trace_fifo1
    (
     .clk                       (TRACECLKIN),
     .reset_n                   (TRESETn),
     .pop_fifo_i                (fifo1_pop),
     .write_pointer_gray_i      (write_pointer1_gray[2:0]),
     .fifo_data0_i              (opt_fifo1_data0[7:0]),
     .fifo_data1_i              (opt_fifo1_data1[7:0]),
     .fifo_data2_i              (opt_fifo1_data2[7:0]),
     .fifo_data3_i              (opt_fifo1_data3[7:0]),
     .fifo_valid_o              (opt_fifo1_valid),
     .read_pointer_gray_async_o (iread_pointer1_gray_async[2:0]),
     .fifo_read_data_o          (opt_fifo1_read_data[7:0]));

  assign read_pointer1_gray_async = PORT1_PRESENT != 0 ? iread_pointer1_gray_async : 3'b000;
  assign fifo1_read_data    = PORT1_PRESENT != 0 ? opt_fifo1_read_data : 8'h00;
  assign fifo1_valid        = PORT1_PRESENT != 0 ? opt_fifo1_valid : 1'b0;

  cm3_tpiu_trace_fifo #(PORT2_PRESENT)
    u_cm3_tpiu_trace_fifo2
    (
     .clk                       (TRACECLKIN),
     .reset_n                   (TRESETn),
     .pop_fifo_i                (fifo2_pop),
     .write_pointer_gray_i      (write_pointer2_gray[2:0]),
     .fifo_data0_i              (fifo2_data0[7:0]),
     .fifo_data1_i              (fifo2_data1[7:0]),
     .fifo_data2_i              (fifo2_data2[7:0]),
     .fifo_data3_i              (fifo2_data3[7:0]),
     .fifo_valid_o              (opt_fifo2_valid),
     .read_pointer_gray_async_o (iread_pointer2_gray_async[2:0]),
     .fifo_read_data_o          (opt_fifo2_read_data[7:0]));

  assign read_pointer2_gray_async = PORT2_PRESENT != 0 ? iread_pointer2_gray_async : 3'b000;
  assign fifo2_read_data    = PORT2_PRESENT != 0 ? opt_fifo2_read_data : 8'h00;
  assign fifo2_valid        = PORT2_PRESENT != 0 ? opt_fifo2_valid : 1'b0;



  cm3_tpiu_formatter #(TRACE_LVL)
    u_tpiu_formatter
    (
     .clk                       (TRACECLKIN),
     .reset_n                   (TRESETn),
     .prg_formatter_on_i        (prg_formatter_on),
     .fifo1_valid_i             (fifo1_valid),
     .atb_id1_i                 (ATID1S[6:0]),
     .fifo1_data_i              (fifo1_read_data),
     .fifo2_valid_i             (fifo2_valid),
     .atb_id2_i                 (ATID2S[6:0]),
     .fifo2_data_i              (fifo2_read_data),
     .format_data_ready_i       (format_data_ready),
     .sync_valid_i              (sync_valid),
     .trigger_valid_i           (trigger_valid),
     .prg_port_mode_i           (prg_port_mode),
     .sync_ack_async_o          (sync_ack_async),
     .trigger_ack_async_o       (trigger_ack_async),
     .fifo1_pop_o               (fifo1_pop),
     .fifo2_pop_o               (fifo2_pop),
     .format_data_valid_o       (format_data_valid),
     .format_data_o             (format_data[7:0])
     );


  cm3_tpiu_trace_out #(TRACE_LVL)
    u_cm3_tpiu_trace_out
    (
     .clk                       (TRACECLKIN),
     .reset_n                   (TRESETn),
     .prg_port_size_decoded_i   (prg_port_size_decoded),
     .prg_uart_mode_i           (prg_port_mode[1]),
     .prg_swo_mode_i            (prg_port_mode[0]),
     .format_data_valid_i       (format_data_valid),
     .format_data_i             (format_data),
     .prg_baud_div_i            (prg_baud_div[12:0]),
     .baud_div_reset_i          (baud_div_reset),
     .trace_data_o              (TRACEDATA[3:0]),
     .trace_swo_o               (TRACESWO),
     .tpiu_baud_o               (tpiu_baud_async),
     .trace_out_idle_async_o    (trace_out_idle_async),
     .format_data_ready_o       (format_data_ready));

  cm3_tpiu_trace_clk #(TRACE_LVL)
    u_cm3_tpiu_trace_clk
    (
     .clk                       (TRACECLKIN),
     .reset_n                   (TRESETn),
     .trace_clk_o               (TRACECLK)
     );

  cm3_cdc_connect #(2,0)
    u_connect_read_pointer1_gray
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (read_pointer1_gray_async),
       .data_o                   (read_pointer1_gray_async_cdc)
       );
  cm3_cdc_connect #(2,0)
    u_connect_read_pointer2_gray
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (read_pointer2_gray_async),
       .data_o                   (read_pointer2_gray_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_sync_ack
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (sync_ack_async),
       .data_o                   (sync_ack_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_trigger_ack
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (trigger_ack_async),
       .data_o                   (trigger_ack_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_tpiu_baud
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (tpiu_baud_async),
       .data_o                   (tpiu_baud_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_trace_out_idle
      (
       .clk_launch               (TRACECLKIN),
       .clk_capture              (CLK),
       .data_i                   (trace_out_idle_async),
       .data_o                   (trace_out_idle_async_cdc)
       );
  cm3_cdc_connect #(2,0)
    u_connect_write_pointer1_gray
      (
       .clk_launch               (CLK),
       .clk_capture              (TRACECLKIN),
       .data_i                   (opt_write_pointer1_gray_async),
       .data_o                   (write_pointer1_gray_async_cdc)
       );
  cm3_cdc_connect #(2,0)
    u_connect_write_pointer2_gray
      (
       .clk_launch               (CLK),
       .clk_capture              (TRACECLKIN),
       .data_i                   (opt_write_pointer2_gray_async),
       .data_o                   (write_pointer2_gray_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_sync_valid
      (
       .clk_launch               (CLK),
       .clk_capture              (TRACECLKIN),
       .data_i                   (sync_valid_async),
       .data_o                   (sync_valid_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_trigger_valid
      (
       .clk_launch               (CLK),
       .clk_capture              (TRACECLKIN),
       .data_i                   (trigger_valid_async),
       .data_o                   (trigger_valid_async_cdc)
       );
  cm3_cdc_connect #(0,0)
    u_connect_write_baud_div_reset
      (
      .clk_launch               (CLK),
      .clk_capture              (TRACECLKIN),
      .data_i                   (baud_div_reset_async),
      .data_o                   (baud_div_reset_async_cdc)
       );

  cm3_tpiu_atb_sync #(TRACE_LVL)
    u_cm3_tpiu_atb_sync
    (
     .clk                       (CLK),
     .reset_n                   (RESETn),
     .read_pointer1_gray_async_i(read_pointer1_gray_async_cdc),
     .read_pointer2_gray_async_i(read_pointer2_gray_async_cdc),
     .sync_ack_async_i          (sync_ack_async_cdc),
     .trigger_ack_async_i       (trigger_ack_async_cdc),
     .tpiu_baud_async_i         (tpiu_baud_async_cdc),
     .trace_out_idle_async_i    (trace_out_idle_async_cdc),
     .read_pointer1_gray_o      (read_pointer1_gray),
     .read_pointer2_gray_o      (read_pointer2_gray),
     .sync_ack_o                (sync_ack),
     .trigger_ack_o             (trigger_ack),
     .tpiu_baud_o               (TPIUBAUD),
     .trace_out_idle_o          (trace_out_idle)
     );

  cm3_tpiu_trace_sync #(TRACE_LVL)
    u_cm3_tpiu_trace_sync
  (
   .clk                         (TRACECLKIN),
   .reset_n                     (TRESETn),
   .write_pointer1_gray_async_i (write_pointer1_gray_async_cdc),
   .write_pointer2_gray_async_i (write_pointer2_gray_async_cdc),
   .sync_valid_async_i          (sync_valid_async_cdc),
   .trigger_valid_async_i       (trigger_valid_async_cdc),
   .baud_div_reset_async_i      (baud_div_reset_async_cdc),
   .write_pointer1_gray_o       (write_pointer1_gray),
   .write_pointer2_gray_o       (write_pointer2_gray),
   .sync_valid_o                (sync_valid),
   .trigger_valid_o             (trigger_valid),
   .baud_div_reset_o            (baud_div_reset)
   );

endmodule // cm3_tpiu
