`include "cm3_tpiu_defs.vp"

module cm3_tpiu_apb_if  (
  clk, clk_en, reset_n, apb_write_i, apb_enable_i, apb_sel_i, apb_addr_i, apb_wdata_i, atb_valid1_i,
  atb_valid2_i, max_port_size_i, fifo1_ready_i, fifo2_ready_i, sync_req_i, sync_ack_i,
  trigger_i, trigger_ack_i, fifo_empty1_i, fifo_empty2_i, trace_out_idle_i,
  fifo1_data0_i, fifo1_data1_i, fifo1_data2_i, fifo2_data0_i, fifo2_data1_i, fifo2_data2_i,
  write_pointer1_1_0_i,write_pointer2_1_0_i,
  prg_port_size_decoded_o, prg_port_mode_o, prg_baud_div_o, baud_div_reset_o, prg_formatter_on_o, apb_rdata_o, tpiu_active_o,
  swo_active_o, atb_ready1_o, atb_ready2_o, port1_valid_o, port2_valid_o, sync_valid_async_o,
  trigger_valid_async_o, fifo_test_mode_o, integration_fifo1_read_o, integration_fifo2_read_o
  );

  parameter TRACE_LVL      = 1;
  parameter CP_PRESENT     = 0;  // Affects ID registers


  input         clk;              // APB clock
  input         clk_en;            // APB clock enable
  input         reset_n;           // APB asychronous reset active low
  input         apb_write_i;           // APB write or read
  input         apb_enable_i;          // APB enable
  input         apb_sel_i;             // APB slave select
  input [11:2]  apb_addr_i;            // APB address bus
  input [12:0]  apb_wdata_i;           // APB Write Data
  input [1:0]   max_port_size_i;      // Configuration tie off
  input         atb_valid1_i;        // ATVALID from port 1
  input         atb_valid2_i;        // ATVALID from port 2
  input         fifo1_ready_i;      // ATREADY input from formatter
  input         fifo2_ready_i;      // ATREADY input from formatter

  input         sync_req_i;          // Sync trigger input
  input         sync_ack_i;       // Re-synchronised ack for sync_req
  input         trigger_i;          // Trigger request input
  input         trigger_ack_i;       // Re-synchronised ack for TRIGGER
  input         fifo_empty1_i;
  input         fifo_empty2_i;
  input         trace_out_idle_i;
  input [7:0]   fifo1_data0_i;
  input [7:0]   fifo1_data1_i;
  input [7:0]   fifo1_data2_i;
  input [7:0]   fifo2_data0_i;
  input [7:0]   fifo2_data1_i;
  input [7:0]   fifo2_data2_i;
  input [1:0]   write_pointer1_1_0_i;  // Lower write pointer bits
  input [1:0]   write_pointer2_1_0_i;  // Lower write pointer bits

  output [2:0]  prg_port_size_decoded_o;         // Selected port size
  output [1:0]  prg_port_mode_o;         // Selected Port Mode
  output [12:0] prg_baud_div_o;          // Baud Divisor
  output        baud_div_reset_o;       // Baud divisor re-programed
  output        prg_formatter_on_o;        // Formatter enabled (2 port only)
  output        tpiu_active_o;         // Data is present in TPIU
  output        swo_active_o;        // SWO mode is selected

  output [31:0] apb_rdata_o;           // APB read data bus
  output        atb_ready1_o;        // ATREADY after integration reg.
  output        atb_ready2_o;        // ATREADY after integration reg.
  output        port1_valid_o;      // ATVALID masked in int test mode
  output        port2_valid_o;      // ATVALID masked in int test mode
  output        sync_valid_async_o;         // Latched Sync request
  output        trigger_valid_async_o;         // Latched trigger request
  output        fifo_test_mode_o;     // Fifo test mode selected
  output        integration_fifo1_read_o;      // Test mode fifo1 read
  output        integration_fifo2_read_o;      // Test mode fifo2 read


  wire          register_write;              // Global register write
  wire          nxt_register_read;           // Global register write
  wire          register_read_we;            // Enable for read register
  reg           register_read;               // Global register write
  wire          formatter_control_read;      // Read Enable
  wire          formatter_control_we;        // Write enable
  wire [31:0]   formatter_control_read_data; // Read Data

  wire          formatter_status_read;       // Read Enable
  wire [31:0]   formatter_status_read_data;  // Read Data
  wire          supported_port_size_read;    // Read Enable
  wire [31:0]   supported_port_size_read_data;// Read Data

  wire          prg_port_size_read;          // Read Enable
  wire          prg_port_size_we;            // Write Enable
  wire [31:0]   prg_port_size_read_data;     // Read Data
  reg  [1:0]    prg_port_size_1_0;           // Register storage [1:0]
  wire [1:0]    nxt_prg_port_size_1_0;       // Register storage [1:0]
  reg           prg_port_size_3;             // Register storage [3]
  wire          nxt_prg_port_size_3;         // Register storage [3]
  wire [2:0]    prg_port_size_decoded;       // Port size info for trace_out block

  reg [1:0]     integration_mode;            // Register storage
  wire          integration_mode_we;         // Write Enable
  wire          integration_mode_read;       // Read Enable
  wire [31:0]   integration_mode_read_data;  // Read Data

  wire          integration_atb_cr0_read;    // Read Enable
  wire [31:0]   integration_atb_cr0read_data;// Read Data
  wire          integration_flush_read;      // Reaintegration_flush_readd Enable
  wire [31:0]   integration_flush_read_data; // Read Data
  wire          integration_fifo1_read;      // Read Enable
  wire [31:0]   fifo1_read_data;             // Read Data
  wire          integration_fifo2_read;      // Read Enable
  wire [31:0]   fifo2_read_data;             // Read Data

  reg           integration_atb_cr2;         // Register storage
  wire          integration_atb_cr2_we;      // Write Enable

  wire          claim_tag_we;                // Write Enable
  wire          claim_set_read;              // Set register Read Enable
  wire          claim_clear_read;            // Clear register Read Enable
  reg [3:0]     claim_tag;                   // Register storage
  wire [3:0]    nxt_claim_tag;               // D-input for register
  wire [31:0]   claim_set_read_data;         // Read Data
  wire [31:0]   claim_clear_read_data;       // Read Data

  wire          device_id_read;              // Read Enable
  wire [31:0]   device_id_read_data;         // Read Data

  wire          device_type_read;            // Read Enable
  wire [31:0]   device_type_read_data;       // Read Data

  wire          peripheral_id0_read;         // Read Enable
  wire          peripheral_id1_read;         // Read Enable
  wire          peripheral_id2_read;         // Read Enable
  wire          peripheral_id4_read;         // Read Enable
  wire [31:0]   peripheral_id0_read_data;    // Read Data
  wire [31:0]   peripheral_id1_read_data;    // Read Data
  wire [31:0]   peripheral_id2_read_data;    // Read Data
  wire [31:0]   peripheral_id4_read_data;    // Read Data

  wire          component_id0_read;          // Read Enable
  wire          component_id1_read;          // Read Enable
  wire          component_id2_read;          // Read Enable
  wire          component_id3_read;          // Read Enable
  wire [31:0]   component_id0_read_data;     // Read Data
  wire [31:0]   component_id1_read_data;     // Read Data
  wire [31:0]   component_id2_read_data;     // Read Data
  wire [31:0]   component_id3_read_data;     // Read Data

  wire          prg_baud_div_we;             // Write Enable
  wire          prg_baud_div_read;           // Read Enable
  wire [31:0]   prg_baud_div_read_data;      // Read Data
  reg [12:0]    prg_baud_div;                // Register storage

  wire          prg_port_mode_we;            // Write Enable
  wire          prg_port_mode_read;          // Read Enable
  wire [31:0]   prg_port_mode_read_data;     // Read Data
  reg  [1:0]    prg_port_mode;               // Register storage
  reg           formatter_control;           // Formatter control register

  wire [31:0]   read_data;                   // Read Mux output

  reg           sync_valid_async;            // Sync request seen
  wire          nxt_sync_valid_async;        // D input for SyncReg
  reg           trigger_valid_async;         // Trigger tequest seen
  wire          nxt_trigger_valid_async;     // D input for TrigReg

  reg           baud_div_reset;
  wire          prg_formatter_on;
  wire          tpiu_active;
  wire          swo_active;
  wire          port1_valid;
  wire          port2_valid;
  wire          fifo_test_mode;
  wire          opt_tpiu_en;

  assign prg_port_size_decoded_o   = (TRACE_LVL > 0) ? prg_port_size_decoded : 3'b000;
  assign prg_port_mode_o           = (TRACE_LVL > 0) ? prg_port_mode     : 2'b00;
  assign prg_baud_div_o            = (TRACE_LVL > 0) ? prg_baud_div      : {13{1'b0}};
  assign baud_div_reset_o          = (TRACE_LVL > 0) ? baud_div_reset   : 1'b0;
  assign prg_formatter_on_o        = (TRACE_LVL > 0) ? prg_formatter_on    : 1'b0;
  assign tpiu_active_o             = (TRACE_LVL > 0) ? tpiu_active     : 1'b0;
  assign swo_active_o              = (TRACE_LVL > 0) ? swo_active    : 1'b0;
  assign apb_rdata_o               = (TRACE_LVL > 0) ? read_data     : {32{1'b0}};
  assign port1_valid_o             = (TRACE_LVL > 0) ? port1_valid  : 1'b0;
  assign port2_valid_o             = (TRACE_LVL > 0) ? port2_valid  : 1'b0;
  assign sync_valid_async_o        = (TRACE_LVL > 0) ? sync_valid_async     : 1'b0;
  assign trigger_valid_async_o     = (TRACE_LVL > 0) ? trigger_valid_async     : 1'b0;
  assign fifo_test_mode_o          = (TRACE_LVL > 0) ? fifo_test_mode : 1'b0;
  assign integration_fifo1_read_o  = (TRACE_LVL > 0) ? integration_fifo1_read  : 1'b0;
  assign integration_fifo2_read_o  = (TRACE_LVL > 0) ? integration_fifo2_read  : 1'b0;
  assign opt_tpiu_en               = (TRACE_LVL > 0) ? 1'b1          : 1'b0;


  assign nxt_sync_valid_async  = sync_req_i |(sync_valid_async & ~sync_ack_i);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      sync_valid_async <= 1'b0;
    else
      sync_valid_async <= nxt_sync_valid_async;

  assign nxt_trigger_valid_async  = trigger_i |(trigger_valid_async & ~trigger_ack_i);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      trigger_valid_async <= 1'b0;
    else
      trigger_valid_async <= nxt_trigger_valid_async;

  assign tpiu_active = ~(fifo_empty1_i & fifo_empty2_i & trace_out_idle_i);

  assign register_write = clk_en & opt_tpiu_en & apb_enable_i & apb_sel_i & apb_write_i;
  assign nxt_register_read  = ~apb_enable_i & apb_sel_i & ~apb_write_i;
  assign register_read_we = clk_en & opt_tpiu_en & apb_sel_i;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      register_read <= 1'b0;
    else if (register_read_we)
      register_read <= nxt_register_read;

  reg [11:2] apb_addr_reg;

  always @(posedge clk)
    if (apb_sel_i)
      apb_addr_reg <= apb_addr_i;

  assign supported_port_size_read   = register_read & (apb_addr_reg == `CM_TPIU_SUPPORTSIZE_ADD);
  assign supported_port_size_read_data = {{28{1'b0}},(&max_port_size_i),1'b0,
                              (|max_port_size_i),1'b1};

  assign nxt_prg_port_size_1_0 = {(|max_port_size_i)&apb_wdata_i[1],apb_wdata_i[0]};
  assign nxt_prg_port_size_3 = &{apb_wdata_i[3],max_port_size_i};

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        prg_port_size_1_0 <= 2'b01;
        prg_port_size_3  <= 1'b0;
      end
    else if (prg_port_size_we)
      begin
        prg_port_size_1_0 <= nxt_prg_port_size_1_0;
        prg_port_size_3   <= nxt_prg_port_size_3;
      end

  assign prg_port_size_read   = register_read  & (apb_addr_reg == `CM_TPIU_CURPORTSIZE_ADD);
  assign prg_port_size_we   = register_write & (apb_addr_reg == `CM_TPIU_CURPORTSIZE_ADD);
  assign prg_port_size_read_data = {{28{1'b0}},prg_port_size_3,1'b0,prg_port_size_1_0};
  assign prg_port_size_decoded = {2'b00,(|prg_port_mode)} |
                                   ({{2{~(|prg_port_mode)}}, 1'b1} &
                                    {prg_port_size_3, prg_port_size_1_0});

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        prg_baud_div   <= 13'h0000;
        baud_div_reset <= 1'b0;
      end
    else if (prg_baud_div_we)
      begin
        prg_baud_div   <= apb_wdata_i[12:0];
        baud_div_reset <= ~baud_div_reset;
      end

  assign prg_baud_div_we   = register_write & (apb_addr_reg == `CM_TPIU_CURSPDDIV_ADD);
  assign prg_baud_div_read   = register_read & (apb_addr_reg == `CM_TPIU_CURSPDDIV_ADD);
  assign prg_baud_div_read_data = {{19{1'b0}}, prg_baud_div};

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      prg_port_mode <= 2'b01;
    else if (prg_port_mode_we)
      prg_port_mode <= apb_wdata_i[1:0];

  assign prg_port_mode_we   = register_write & (apb_addr_reg == `CM_TPIU_SELPINPROT_ADD);
  assign prg_port_mode_read   = register_read & (apb_addr_reg == `CM_TPIU_SELPINPROT_ADD);
  assign prg_port_mode_read_data = {{30{1'b0}}, prg_port_mode};
  assign swo_active       = |prg_port_mode;

  assign formatter_status_read   = register_read & (apb_addr_reg == `CM_TPIU_FMFLSTAT_ADD);
  assign formatter_status_read_data = `CM_TPIU_FMFLSTAT_VAL;

  assign formatter_control_read    = register_read & (apb_addr_reg == `CM_TPIU_FFCR_ADD);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      formatter_control  <= 1'b1;
    else if (formatter_control_we)
      formatter_control <= apb_wdata_i[1];

  assign formatter_control_we = register_write & (apb_addr_reg == `CM_TPIU_FFCR_ADD);

  assign prg_formatter_on     = formatter_control | (prg_port_mode==2'b00);
  assign formatter_control_read_data  = {28'h0000010,2'b00,prg_formatter_on,1'b0};

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      integration_atb_cr2 <= 1'b0;
    else if (integration_atb_cr2_we)
        integration_atb_cr2 <= apb_wdata_i[0];

  assign integration_atb_cr2_we   = register_write & (apb_addr_reg == `CM_TPIU_ITATBCR2_ADD);

  assign integration_atb_cr0_read   = register_read & (apb_addr_reg == `CM_TPIU_ITATBCR0_ADD);
  assign integration_atb_cr0read_data = { {31{1'b0}}, atb_valid1_i | atb_valid2_i};

  assign integration_flush_read   = register_read & (apb_addr_reg == `CM_TPIU_ITTRFLIN_ADD);
  assign integration_flush_read_data = {{31{1'b0}}, trigger_i};

  assign integration_fifo1_read   = fifo_test_mode & register_read &
                           (apb_addr_reg == `CM_TPIU_ITFTTDR0_ADD);
  assign fifo1_read_data  = { 2'b00,atb_valid2_i,write_pointer2_1_0_i,atb_valid1_i,write_pointer1_1_0_i,
                            fifo1_data2_i, fifo1_data1_i, fifo1_data0_i};

  assign integration_fifo2_read   = fifo_test_mode & register_read &
                           (apb_addr_reg == `CM_TPIU_ITFTTDR1_ADD);
  assign fifo2_read_data  = { 2'b00,atb_valid2_i,write_pointer2_1_0_i,atb_valid1_i,write_pointer1_1_0_i,
                            fifo2_data2_i, fifo2_data1_i, fifo2_data0_i};

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      integration_mode <= 2'b00;
    else if (integration_mode_we)
      integration_mode <= apb_wdata_i[1:0];

  assign integration_mode_we        = register_write & (apb_addr_reg == `CM_TPIU_ITCTRL_ADD);
  assign integration_mode_read      = register_read & (apb_addr_reg == `CM_TPIU_ITCTRL_ADD);
  assign integration_mode_read_data = { {30{1'b0}}, integration_mode};
  assign fifo_test_mode             = integration_mode[1];

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      claim_tag <= 4'h0;
    else if (claim_tag_we)
        claim_tag <= nxt_claim_tag;

  assign claim_tag_we     = (register_write &
                             ( (apb_addr_reg == `CM_TPIU_CLAIMSET_ADD) |
                               (apb_addr_reg == `CM_TPIU_CLAIMCLR_ADD) ) );
  assign nxt_claim_tag  = (apb_addr_reg == `CM_TPIU_CLAIMSET_ADD) ?
                           (claim_tag |  apb_wdata_i[3:0]) :
                           (claim_tag & ~apb_wdata_i[3:0]) ;

  assign claim_set_read     = register_read & (apb_addr_reg == `CM_TPIU_CLAIMSET_ADD);
  assign claim_clear_read     = register_read & (apb_addr_reg == `CM_TPIU_CLAIMCLR_ADD);
  assign claim_set_read_data   = {{28{1'b0}}, 4'hF};
  assign claim_clear_read_data   = {{28{1'b0}}, claim_tag[3:0]};

  assign device_id_read        = register_read & (apb_addr_reg == `CM_TPIU_DEVID_ADD);
  assign device_id_read_data      = TRACE_LVL > 1 ? `CM_TPIU_DEVID_VAL_ETM :
                                            `CM_TPIU_DEVID_VAL_NON;

  assign device_type_read      = register_read & (apb_addr_reg == `CM_TPIU_DEVTYPE_ADD);
  assign device_type_read_data    = {{24{1'b0}},`CM_TPIU_DEVTYPE_VAL};

  assign peripheral_id4_read    = register_read & (apb_addr_reg == `CM_TPIU_PERIPHID4_ADD);
  assign peripheral_id4_read_data  = {{24{1'b0}},`CM_TPIU_PERIPHID4_VAL};

  assign peripheral_id0_read    = register_read & (apb_addr_reg == `CM_TPIU_PERIPHID0_ADD);
  assign peripheral_id0_read_data  = {{24{1'b0}}, (CP_PRESENT == 0) ? `CM3_TPIU_PERIPHID0_VAL : `CM4_TPIU_PERIPHID0_VAL };

  assign peripheral_id1_read    = register_read & (apb_addr_reg == `CM_TPIU_PERIPHID1_ADD);
  assign peripheral_id1_read_data  = {{24{1'b0}},`CM_TPIU_PERIPHID1_VAL};

  assign peripheral_id2_read    = register_read & (apb_addr_reg == `CM_TPIU_PERIPHID2_ADD);
  assign peripheral_id2_read_data  = {{24{1'b0}},(CP_PRESENT == 0) ? `CM3_TPIU_PERIPHID2_VAL : `CM4_TPIU_PERIPHID2_VAL };

  assign component_id0_read      = register_read & (apb_addr_reg == `CM_TPIU_COMPONID0_ADD);
  assign component_id0_read_data = {{24{1'b0}},`CM_TPIU_COMPONID0_VAL};

  assign component_id1_read      = register_read & (apb_addr_reg == `CM_TPIU_COMPONID1_ADD);
  assign component_id1_read_data = {{24{1'b0}},`CM_TPIU_COMPONID1_VAL};

  assign component_id2_read      = register_read & (apb_addr_reg== `CM_TPIU_COMPONID2_ADD);
  assign component_id2_read_data = {{24{1'b0}},`CM_TPIU_COMPONID2_VAL};

  assign component_id3_read      = register_read & (apb_addr_reg == `CM_TPIU_COMPONID3_ADD);
  assign component_id3_read_data = {{24{1'b0}},`CM_TPIU_COMPONID3_VAL};

  assign read_data = (({32{supported_port_size_read}} & supported_port_size_read_data) |
                      ({32{prg_port_size_read}}       & prg_port_size_read_data)       |
                      ({32{prg_baud_div_read}}        & prg_baud_div_read_data)        |
                      ({32{prg_port_mode_read}}       & prg_port_mode_read_data)       |
                      ({32{formatter_status_read}}    & formatter_status_read_data)    |
                      ({32{formatter_control_read}}   & formatter_control_read_data)   |
                      ({32{integration_flush_read}}   & integration_flush_read_data)   |
                      ({32{integration_fifo1_read}}   & fifo1_read_data)               |
                      ({32{integration_atb_cr0_read}} & integration_atb_cr0read_data)  |
                      ({32{integration_fifo2_read}}   & fifo2_read_data)               |
                      ({32{claim_set_read}}           & claim_set_read_data)           |
                      ({32{claim_clear_read}}         & claim_clear_read_data)         |
                      ({32{integration_mode_read}}    & integration_mode_read_data)    |
                      ({32{device_id_read}}           & device_id_read_data)           |
                      ({32{device_type_read}}         & device_type_read_data)         |
                      ({32{peripheral_id0_read}}      & peripheral_id0_read_data)      |
                      ({32{peripheral_id1_read}}      & peripheral_id1_read_data)      |
                      ({32{peripheral_id2_read}}      & peripheral_id2_read_data)      |
                      ({32{peripheral_id4_read}}      & peripheral_id4_read_data)      |
                      ({32{component_id0_read}}       & component_id0_read_data)       |
                      ({32{component_id1_read}}       & component_id1_read_data)       |
                      ({32{component_id2_read}}       & component_id2_read_data)       |
                      ({32{component_id3_read}}       & component_id3_read_data));

  assign atb_ready1_o   = integration_mode[0]? integration_atb_cr2 : fifo1_ready_i;
  assign atb_ready2_o   = integration_mode[0]? integration_atb_cr2 : fifo2_ready_i;

  assign port1_valid = ~integration_mode[0] & atb_valid1_i;
  assign port2_valid = ~integration_mode[0] & atb_valid2_i;

endmodule // cm3_tpiu_apb_if
