`include "cm3_tpiu_defs.vp"

module cm3_tpiu_atb_fifo (
  clk, clk_en, reset_n, atb_valid_i, read_pointer_gray_i, atb_data_i,
  fifo_test_mode_i, test_read_i,
  atb_ready_o, fifo_empty_o, write_pointer_gray_async_o, write_pointer_1_0_o,
  fifo_data0_o, fifo_data1_o, fifo_data2_o, fifo_data3_o
  );

  parameter PORT_PRESENT = 1;


  input         clk;               // ATB clock
  input         clk_en;            // ATB Clock enable
  input         reset_n;           // ATB asychronous reset active low
  input         atb_valid_i;       // Data valid signal from ATB
  input  [2:0]  read_pointer_gray_i;  // Read pointer from the synchronizer
  input  [7:0]  atb_data_i;        // Write data to the FIFO
  input         fifo_test_mode_i;  // Test mode, fifo does not drain to traceport
  input         test_read_i;       // Test mode, fifo has been flushed.

  output        atb_ready_o;       // Ready to accept data from ATB port
  output        fifo_empty_o;      // Fifo has no data
  output [2:0]  write_pointer_gray_async_o;      // Write pointer to the read clock domain
  output [1:0]  write_pointer_1_0_o;     // For fifo test mode, shows depth

  output [7:0]  fifo_data0_o;      // Output all the data registers for use
  output [7:0]  fifo_data1_o;      // in the TRACECLKIN domain
  output [7:0]  fifo_data2_o;      // Fifo byte2
  output [7:0]  fifo_data3_o;      // Fifo byte3


  wire         write_pointer_we;    // Global write enable
  wire         fifo_data_we_0;      // Byte 0 write enable
  wire         fifo_data_we_1;      // Byte 1 write enable
  wire         fifo_data_we_2;      // Byte 2 write enable
  wire         fifo_data_we_3;      // Byte 3 write enable
  wire [2:0]   read_pointer;        // Read pointer (binary)
  wire [2:0]   nxt_write_pointer;   // D-input for write_pointer
  wire [2:0]   nxt_write_pointer_gray_async;    // D-input for write_pointer_gray_async
  reg  [7:0]   fifo_data0;          // Fifo Data register
  reg  [7:0]   fifo_data1;          // Fifo Data register
  reg  [7:0]   fifo_data2;          // Fifo Data register
  reg  [7:0]   fifo_data3;          // Fifo Data register
  wire [7:0]   nxt_fifo_data3;      // Fifo Data register
  wire         atb_ready;
  wire         fifo_empty;

  wire [2:0]   write_pointer_decode;// Write pointer (binary)
  wire [2:0]   write_pointer;       // Write pointer, even in test mode

  reg  [2:0]   iwrite_pointer_gray_async;
  wire [1:0]   iwrite_pointer_1_0;
  wire         opt_port_en;

  assign opt_port_en  = (PORT_PRESENT != 0) ? 1'b1        : 1'b0;


  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      fifo_data0 <= {8{1'b0}};
    else if (fifo_data_we_0)
      fifo_data0 <= atb_data_i;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      fifo_data1 <= {8{1'b0}};
    else if (fifo_data_we_1)
      fifo_data1 <= atb_data_i;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      fifo_data2 <= {8{1'b0}};
    else if (fifo_data_we_2)
      fifo_data2 <= atb_data_i;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      fifo_data3 <= {8{1'b0}};
    else if (fifo_data_we_3)
      fifo_data3 <= nxt_fifo_data3;

  assign write_pointer_we = clk_en & opt_port_en & (
                                                   (atb_valid_i & atb_ready) |
                                                    test_read_i);

  assign fifo_data_we_0 = (write_pointer_we & (write_pointer[1:0] == 2'b00));
  assign fifo_data_we_1 = (write_pointer_we & (write_pointer[1:0] == 2'b01));
  assign fifo_data_we_2 = (write_pointer_we & (write_pointer[1:0] == 2'b10));
  assign fifo_data_we_3 = (write_pointer_we & ((write_pointer_decode[1:0] == 2'b11)
                                               | fifo_test_mode_i));

  assign write_pointer[2] = write_pointer_decode[2];
  assign write_pointer[1:0] = fifo_test_mode_i ? fifo_data3[1:0] : write_pointer_decode[1:0];

  assign nxt_fifo_data3[7:2] = atb_data_i[7:2];
  assign nxt_fifo_data3[1:0] = fifo_test_mode_i ? nxt_write_pointer[1:0] : atb_data_i[1:0];

  assign nxt_write_pointer = {3{~test_read_i}} &
                              write_pointer + 3'b001;

  assign atb_ready = fifo_test_mode_i ? ((fifo_data3[1:0] != 2'b11) & ~test_read_i) :
                      ~((write_pointer_decode[1:0] == read_pointer[1:0]) &
                        (write_pointer_decode[2]   != read_pointer[2]));

  assign fifo_empty = (write_pointer[2:0] == read_pointer[2:0]);

  assign nxt_write_pointer_gray_async = {3{~fifo_test_mode_i}} &
                         ({1'b0, nxt_write_pointer[2:1]} ^ nxt_write_pointer[2:0]);

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      iwrite_pointer_gray_async   <= {3{1'b0}};
    else if (write_pointer_we)
      iwrite_pointer_gray_async  <= nxt_write_pointer_gray_async;

  assign write_pointer_decode[0] = iwrite_pointer_gray_async[0] ^
                              iwrite_pointer_gray_async[1] ^
                              iwrite_pointer_gray_async[2];
  assign write_pointer_decode[1] = iwrite_pointer_gray_async[1] ^
                              iwrite_pointer_gray_async[2];
  assign write_pointer_decode[2] = iwrite_pointer_gray_async[2];

  assign iwrite_pointer_1_0 = fifo_data3[1:0];

  assign read_pointer[0] = read_pointer_gray_i[0] ^ read_pointer_gray_i[1] ^
                             read_pointer_gray_i[2];
  assign read_pointer[1] = read_pointer_gray_i[1] ^ read_pointer_gray_i[2];
  assign read_pointer[2] = read_pointer_gray_i[2];

  assign atb_ready_o  = (PORT_PRESENT != 0) ? atb_ready   : 1'b0;
  assign fifo_empty_o = (PORT_PRESENT != 0) ? fifo_empty  : 1'b0;
  assign write_pointer_gray_async_o  = (PORT_PRESENT != 0) ? iwrite_pointer_gray_async  : 3'b000;
  assign write_pointer_1_0_o         = (PORT_PRESENT != 0) ? iwrite_pointer_1_0 : 2'b00;
  assign fifo_data0_o = (PORT_PRESENT != 0) ? fifo_data0  : {8{1'b0}};
  assign fifo_data1_o = (PORT_PRESENT != 0) ? fifo_data1  : {8{1'b0}};
  assign fifo_data2_o = (PORT_PRESENT != 0) ? fifo_data2  : {8{1'b0}};
  assign fifo_data3_o = (PORT_PRESENT != 0) ? fifo_data3  : {8{1'b0}};


endmodule // cm3_tpiu_atb_fifo
