`include "cm3_tpiu_defs.vp"

module cm3_tpiu_formatter  (
  clk, reset_n, prg_formatter_on_i, fifo1_valid_i, atb_id1_i, fifo1_data_i,
  fifo2_valid_i, atb_id2_i, fifo2_data_i, format_data_ready_i, sync_valid_i,
  trigger_valid_i, prg_port_mode_i,
  fifo1_pop_o, fifo2_pop_o, format_data_valid_o, format_data_o,
  sync_ack_async_o, trigger_ack_async_o
  );

  parameter TRACE_LVL = 1;


  input         clk;                   // ATB/Core clock
  input         reset_n;               // ATB/Core sychronous reset
  input         prg_formatter_on_i;    // Formatter in use
  input         fifo1_valid_i;         // VAlid signals present from source
  input [6:0]   atb_id1_i;             // Current trace source ID
  input [7:0]   fifo1_data_i;          // Trace data
  input         fifo2_valid_i;         // VAlid signals present from source
  input [6:0]   atb_id2_i;             // Current trace source ID
  input [7:0]   fifo2_data_i;          // Trace data
  input         format_data_ready_i;   // Formatter output interface ready
  input         sync_valid_i;          // Sync is required asap - inter frame
  input         trigger_valid_i;       // Sync is required asap - intra frame
  input [1:0]   prg_port_mode_i;       // Port mode register value


  output        fifo1_pop_o;           // Data taken from fifo1
  output        fifo2_pop_o;           // Data taken from fifo2
  output        format_data_valid_o;   // Formatter output data valid
  output [7:0]  format_data_o;         // Formatter output data
  output        sync_ack_async_o;      // Sync has been captured here
  output        trigger_ack_async_o;   // Trigger has been captured here

  reg  [1:0]    nxt_data_source;       // select between IDs
  wire [3:0]    nxt_packet_count;      // Position in packet
  wire          nxt_packet_last_byte;  // Identifies last byte of packet.
  wire [6:0]    nxt_last_byte_store;   // Collect LSBs as packet is built up
  wire          new_store_bit;         // LSB from current packet byte.

  wire          byte_taken;            // Byte from formatter (not sync) taken
  wire          byte_taken_en;         // Byte taken derived flop enable
  wire [7:0]    id_byte;               // ID value to insert in stream
  wire          nxt_source_change;     // Src change D input
  wire [7:0]    nxt_format_frame_data; // format_data D input
  wire          nxt_format_data_valid; // format_data_valid D input
  wire [7:0]    nxt_input_buffer_data; // input_buffer_data D input
  wire          input_buffer_we;       // common signal for input_buffer used
  wire          input_buffer_reg_we;   // Enable for input_buffer registers


  wire          nxt_input_buffer_valid;// D input for input_buffer_valid
  wire          input_buffer_valid_set;// Data is valid
  wire          input_buffer_used;     // Data taken came from holding reg.
  wire          id_late_change_set;    // bit to identify byte following ID pkt
  wire          nxt_id_late_change;    // D input for IDXBitReg

  wire [1:0]    selected_id_source;    // delayed source for ID bytes
  wire          nxt_null_id_selected;  // Last ouput ID was null.

  wire          last_byte_cycle;       // Cycle is last in packet
  wire          id_cycle;              // Cycle is ID change
  wire          data_cycle;            // Cycle is data byte
  wire          data_formatted_bit0;   // Bit[0] for data bytes
  wire          force_periodic_id;     // An ID is required every 16 frames
  wire          format_frame_data_we;  // Formatter output register enable
  wire          format_frame_data_en;  // Formatter output data enable
  wire          skip_null;             // Skip data, ID was null.
  wire          nxt_id_in_this_frame;  // This frame has an ID in it.
  wire          packet_start;          // nxt_packet_count = 0

  reg [1:0]     data_source;           // Selected input source
  reg           packet_last_byte;      // Identifies last byte of packet
  reg [7:0]     format_frame_data;     // Formatter output folding register
  wire [7:0]    format_data;           // Frame Data or sync data
  reg [6:0]     last_byte_store;       // Collect LSB bits for end of packet
  reg           last_byte_store_7;     // seperate top bit for reset
  reg           sync_pending;
  wire          nxt_sync_pending;

  reg           format_data_valid;     // Formatter data is valid
  reg [3:0]     packet_count;          // Byte in packet count
  reg           source_change;         // Source of data has changed.
  reg           source_change_set;     // Source of data has changed.
  reg [7:0]     input_buffer_data;     // Input holding register
  reg           input_buffer_valid;    // Data which hasn't been used.
  reg           id_late_change;
  reg           null_id_selected;      // Last ID was null header
  reg           id_cycle_f1;           // ID cycle flag pipe stage 1
  reg           id_cycle_f2;           // ID cycle flag pipe stage 2
  reg [3:0]     frames_without_id;     // Frames since last ID output
  wire [3:0]    nxt_frames_without_id; // Frames since last ID output
  reg           id_in_this_frame;      // This frame contains an ID

  wire          sync_to_fill;          // Sync required to fill stream
  wire          sync_request_accepted; // Sync generated from request
  reg [1:0]     sync_count;            // Sync packet counter
  wire [1:0]    nxt_sync_count;        // D-input for SyncCountReg
  wire          sync_insert_active;    // Sync is active
  wire          sync_packet_7f;        // Sync packet required, is 7F (final)
  wire          sync_packet_ff;        // Sync packet required, is FF
  wire          nxt_sync_still_active; // D-input for SyncEn
  reg           sync_still_active;     // SyncEn from last cycle
  wire          sync_count_we;         // Enable for SyncCountReg
  wire          insert_trigger;        // Insert trigger now

  reg           trigger_ack_async;     // Trigger syncroniser acknowledge
  wire          trigger_ack_async_we;  // Enable for trigger_ack_async Reg
  wire          nxt_trigger_ack_async; // D input for trigger_ack_async Reg
  wire          itm_valid;
  wire [7:0]    itm_data;

  wire          fifo1_pop;
  wire          fifo2_pop;
  wire          opt_tpiu_en;
  wire          allow_sync_start;
  wire [3:0]    sync_mask;
  wire          id_bit;

  assign fifo1_pop_o         = (TRACE_LVL > 0) ? fifo1_pop    : 1'b0;
  assign fifo2_pop_o         = (TRACE_LVL > 0) ? fifo2_pop    : 1'b0;
  assign format_data_o       = (TRACE_LVL > 0) ? format_data  : {8{1'b0}};
  assign sync_ack_async_o    = (TRACE_LVL > 0) ? sync_pending : 1'b0;
  assign trigger_ack_async_o = (TRACE_LVL > 0) ? trigger_ack_async : 1'b0;
  assign opt_tpiu_en         = (TRACE_LVL > 0) ? 1'b1         : 1'b0;


  assign itm_valid = (TRACE_LVL == 1) ? fifo1_valid_i : fifo2_valid_i;
  assign itm_data  = (TRACE_LVL == 1) ? fifo1_data_i  : fifo2_data_i;


  assign byte_taken = format_data_valid &
                      format_data_ready_i & ~sync_insert_active;
  assign byte_taken_en = byte_taken & prg_formatter_on_i & opt_tpiu_en;

  assign format_frame_data_en = (format_data_ready_i &
                                  ~sync_insert_active) |
                                 ~format_data_valid;
  assign format_frame_data_we =  opt_tpiu_en & prg_formatter_on_i & format_frame_data_en;

  assign nxt_packet_count     = packet_count + 4'b0001;
  assign packet_start         = (packet_count == 4'b1111);
  assign nxt_packet_last_byte = (packet_count == 4'b1110);

  assign nxt_input_buffer_data = ({8{nxt_data_source==`CM_TPIU_SRC_SRC1}} & fifo1_data_i) |
                                   ({8{nxt_data_source==`CM_TPIU_SRC_SRC2}} & fifo2_data_i);

  assign selected_id_source = (packet_count[0]&(source_change & ~skip_null)) ?
                                data_source : nxt_data_source;

  assign nxt_null_id_selected = selected_id_source == `CM_TPIU_SRC_IDLE;

  assign id_byte[7:1] = ({7{(selected_id_source==`CM_TPIU_SRC_SRC1)}} & atb_id1_i)|
                          ({7{(selected_id_source==`CM_TPIU_SRC_SRC2)}} & atb_id2_i)|
                            ({7{(selected_id_source==`CM_TPIU_SRC_TRIG)}} & `CM_TPIU_ID_TRIGGER);
  assign id_byte[0] = 1'b1;

  assign insert_trigger = (((byte_taken & ~packet_count[0]) | (format_data_ready_i
                                                               & ~format_data_valid
                                                               & (&sync_count)))
                           & trigger_valid_i &
                           ~trigger_ack_async & ~(packet_count[3:2] == 2'b11));

  assign nxt_trigger_ack_async  = insert_trigger | (trigger_ack_async & trigger_valid_i);

  assign trigger_ack_async_we = (insert_trigger | trigger_ack_async) & opt_tpiu_en;

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      trigger_ack_async <= 1'b0;
    else if (trigger_ack_async_we)
      trigger_ack_async <= nxt_trigger_ack_async;

  assign fifo1_pop = prg_formatter_on_i ? (input_buffer_we & (nxt_data_source==`CM_TPIU_SRC_SRC1)):
                                         (format_data_ready_i | (TRACE_LVL > 1));
  assign fifo2_pop = prg_formatter_on_i ? (input_buffer_we & (nxt_data_source==`CM_TPIU_SRC_SRC2)) :
                                          format_data_ready_i;

  assign input_buffer_valid_set = ((nxt_data_source==`CM_TPIU_SRC_SRC1) & fifo1_valid_i) |
                              ((nxt_data_source==`CM_TPIU_SRC_SRC2) & fifo2_valid_i)|
                                (nxt_data_source==`CM_TPIU_SRC_TRIG);

  assign input_buffer_used = (format_frame_data_en & data_cycle) &
                               (~null_id_selected | id_late_change);

  assign input_buffer_we = (~input_buffer_valid | input_buffer_used);
  assign input_buffer_reg_we = input_buffer_we & prg_formatter_on_i & opt_tpiu_en;

  assign nxt_input_buffer_valid = input_buffer_valid_set |
                                  (input_buffer_valid & ~input_buffer_used);

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      input_buffer_data <= 8'h00;
    else if (input_buffer_reg_we)
      input_buffer_data <= nxt_input_buffer_data;

  assign skip_null = (data_source==`CM_TPIU_SRC_IDLE ) & source_change & ~id_cycle_f1;

  always @(fifo1_valid_i or fifo2_valid_i or byte_taken
           or id_cycle_f1 or format_data_valid or skip_null or data_source
           or source_change or insert_trigger or packet_count or last_byte_cycle)
    case ({((~last_byte_cycle & byte_taken) | ~format_data_valid | skip_null ),data_source})
      3'b100: //No DATA
        begin
          nxt_data_source = {2{~(~skip_null & source_change & ~id_cycle_f1)}} & // else idle
                            (({2{insert_trigger}} & `CM_TPIU_SRC_TRIG ) |
                             ({2{fifo1_valid_i}} & `CM_TPIU_SRC_SRC1 ) |
                             ({2{fifo2_valid_i & ~fifo1_valid_i}} & `CM_TPIU_SRC_SRC2));
          source_change_set = (fifo1_valid_i | fifo2_valid_i | insert_trigger) &
                              ~(source_change & ~id_cycle_f1);
        end
      3'b101: // Input 1
        begin
          nxt_data_source = ((source_change & ~id_cycle_f1)) ? `CM_TPIU_SRC_SRC1 :
                            (({2{insert_trigger}} & `CM_TPIU_SRC_TRIG) |
                             ({2{fifo1_valid_i}} & `CM_TPIU_SRC_SRC1) |
                             ({2{fifo2_valid_i & ~fifo1_valid_i}} & `CM_TPIU_SRC_SRC2));
          source_change_set = (~fifo1_valid_i | insert_trigger) &
                              ~(source_change & ~id_cycle_f1);
        end
      3'b110: // Input 2
        begin
          nxt_data_source = (source_change & ~id_cycle_f1) ? `CM_TPIU_SRC_SRC2 :
                            (({2{insert_trigger}} & `CM_TPIU_SRC_TRIG) |
                             ({2{fifo2_valid_i}} & `CM_TPIU_SRC_SRC2) |
                             ({2{fifo1_valid_i & ~fifo2_valid_i}} & `CM_TPIU_SRC_SRC1));
          source_change_set = (~fifo2_valid_i | insert_trigger) &
                              ~(source_change & ~id_cycle_f1);
        end
      3'b111: // Trigger
        begin
          nxt_data_source = {2{(~skip_null & source_change & ~id_cycle_f1)}} | // trigger
                            ({2{packet_count[0]}} & `CM_TPIU_SRC_TRIG) |
                            ({2{fifo1_valid_i}} & `CM_TPIU_SRC_SRC1) |
                            ({2{fifo2_valid_i & ~fifo1_valid_i}} & `CM_TPIU_SRC_SRC2);
          source_change_set = ~(source_change & ~id_cycle_f1);
        end
      3'b000,3'b001,3'b010,3'b011:
        begin
          nxt_data_source = data_source;
          source_change_set = 1'b0;
        end
      default:
        begin
          nxt_data_source = 2'bxx;
          source_change_set = 1'bx;
        end
    endcase // case({((~last_byte_cycle & byte_taken) | ~format_data_valid | skip_null ),data_source})


  assign nxt_source_change = force_periodic_id | source_change_set |
                             (source_change & ~id_cycle_f1);

  assign last_byte_cycle = nxt_packet_last_byte;

  assign id_cycle  = (~nxt_packet_count[0] & nxt_source_change &
                      ~(source_change_set & ~source_change & (packet_count==4'hD))) |
                      ~format_data_valid;

  assign data_cycle = ~(last_byte_cycle | id_cycle);

  assign id_late_change_set = ~((packet_count==4'hD) | (packet_count==4'hE & data_source[0])) &
                                source_change_set &
                                ((~source_change & packet_count[0] &  ~(id_cycle_f2)) |
                                 last_byte_cycle |
                                  (packet_count[0] & id_cycle_f2 & ~input_buffer_we));

  assign nxt_id_late_change = id_late_change_set | (id_late_change &(~id_cycle_f1 | ~byte_taken));

  assign data_formatted_bit0 = nxt_packet_count[0] & input_buffer_data[0];

  assign sync_packet_7f  = sync_insert_active &  (sync_count==2'b11);
  assign sync_packet_ff  = sync_insert_active & ~(sync_count==2'b11);

  assign nxt_format_frame_data = ({8{last_byte_cycle}} & {last_byte_store_7,last_byte_store} )|
                                   ({8{id_cycle}} & id_byte)|
                                     ({8{data_cycle}} & {input_buffer_data[7:1],
                                                         data_formatted_bit0});
  assign id_bit = packet_count[0] & input_buffer_valid &
                  (force_periodic_id | source_change_set | (source_change & ~id_cycle_f1)) &
                    (packet_count[3:1] != 3'b110) &
         (source_change_set & ((~source_change &  ~(id_cycle_f2)) |
                               last_byte_cycle |
                               (id_cycle_f2 & ~input_buffer_we)) |
          (id_late_change & (~id_cycle_f1 | ~byte_taken))
          );

  assign new_store_bit = id_bit | (~id_cycle & input_buffer_data[0]);
  assign nxt_last_byte_store = {last_byte_store_7,last_byte_store[6:1]};

  assign sync_mask = {{2{~(sync_pending|trigger_valid_i)}},~(sync_pending|trigger_valid_i) &
                      ~(&packet_count[3:2]),1'b0};
  assign allow_sync_start = (sync_mask | packet_count) == 4'b1111;
  assign nxt_format_data_valid = (fifo1_valid_i | fifo2_valid_i | input_buffer_valid | insert_trigger |
                                  (|packet_count & ~(&packet_count) & sync_pending & sync_valid_i)) |
                                   (~(byte_taken & allow_sync_start) &
                                      format_data_valid);

  assign nxt_id_in_this_frame = packet_last_byte ? 1'b0 : (id_in_this_frame | id_cycle);

  assign nxt_frames_without_id = ({4{~id_in_this_frame}} &
                                  (frames_without_id + {3'b000,nxt_packet_last_byte &
                                                               ~id_in_this_frame}));
  assign force_periodic_id = (frames_without_id == 4'b1111) & packet_start;

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        packet_count        <= 4'b0000;
        packet_last_byte    <= 1'b0;
        frames_without_id   <= 4'b0000;
        id_in_this_frame    <= 1'b0;
      end
    else if (byte_taken_en)
      begin
        packet_count        <= nxt_packet_count;
        packet_last_byte    <= nxt_packet_last_byte;
        frames_without_id   <= nxt_frames_without_id;
        id_in_this_frame    <= nxt_id_in_this_frame;
      end

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      last_byte_store_7 <= 1'b0;
    else if (byte_taken_en & ~nxt_packet_count[0])
      last_byte_store_7 <= new_store_bit;

  always@(posedge clk)
    if (byte_taken_en & ~nxt_packet_count[0])
      last_byte_store[6:0] <= nxt_last_byte_store[6:0];

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        data_source         <= 2'b00;
        source_change       <= 1'b0;
        format_data_valid   <= 1'b0;
        input_buffer_valid  <= 1'b0;
        null_id_selected    <= 1'b0;
        id_late_change      <= 1'b0;
      end
    else if (prg_formatter_on_i & opt_tpiu_en)
      begin
        data_source         <= nxt_data_source;
        source_change       <= nxt_source_change;
        format_data_valid   <= nxt_format_data_valid;
        input_buffer_valid  <= nxt_input_buffer_valid;
        null_id_selected    <= nxt_null_id_selected;
        id_late_change      <= nxt_id_late_change;
      end

  always@(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        format_frame_data <= 8'h00;
        id_cycle_f1    <= 1'b1;
        id_cycle_f2    <= 1'b1;
      end
    else if (format_frame_data_we)
      begin
        format_frame_data <= nxt_format_frame_data;
        id_cycle_f1    <= id_cycle;
        id_cycle_f2    <= id_cycle_f1;
      end


  assign sync_to_fill = ~format_data_valid &
                        ~(|prg_port_mode_i);

  assign nxt_sync_pending = sync_valid_i | (sync_pending & ~sync_request_accepted);
  assign nxt_sync_count   = {|packet_count & ~(&packet_count),1'b0} | (sync_count + 2'b01);
  assign sync_count_we    = sync_insert_active & format_data_ready_i  & opt_tpiu_en;

  assign sync_request_accepted = (sync_pending) &
                                   ((&packet_count & format_data_ready_i) |
                                    ~(format_data_valid | (|packet_count)));

  assign nxt_sync_still_active = (sync_still_active & ~(format_data_ready_i & (&sync_count)))|
                                   sync_to_fill | sync_request_accepted;

  assign sync_insert_active = sync_still_active | sync_to_fill;

  assign format_data_valid_o = prg_formatter_on_i ? (sync_insert_active |
                                                     format_data_valid) :
                                                      itm_valid;

  assign format_data = prg_formatter_on_i ? (({8{sync_packet_ff}} & 8'hFF) |
                                               ({8{sync_packet_7f}} & 8'h7F) |
                                               ({8{~sync_insert_active}} &
                                                format_frame_data)) :
                                              itm_data;
  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      sync_count <= 2'b00;
    else if (sync_count_we)
      sync_count <= nxt_sync_count;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        sync_still_active <= 1'b0;
        sync_pending      <= 1'b1;
      end
    else
      begin
        sync_still_active <= nxt_sync_still_active;
        sync_pending      <= nxt_sync_pending;
      end

endmodule // cm3_formatter
