module cm3_tpiu_trace_fifo (/*AUTOARG*/
  clk, reset_n, pop_fifo_i, write_pointer_gray_i, fifo_data0_i, fifo_data1_i,
  fifo_data2_i, fifo_data3_i,
  fifo_valid_o, read_pointer_gray_async_o, fifo_read_data_o
  );

  parameter PORT_PRESENT = 1;


  input        clk;                 // Trace clock
  input        reset_n;             // Trace asychronous reset active low
  input        pop_fifo_i;          // Pop signal for reading from the FIFO
  input [2:0]  write_pointer_gray_i;// Write pointer from the synchronizer

  input [7:0]  fifo_data0_i;        // Input all the data registers for use
  input [7:0]  fifo_data1_i;        // in the read mux
  input [7:0]  fifo_data2_i;        // fifo data 2
  input [7:0]  fifo_data3_i;        // fifo data 3

  output        fifo_valid_o;       // FIFO data present
  output [2:0]  read_pointer_gray_async_o;// Read pointer to the write clock domain
  output [7:0]  fifo_read_data_o;   // Output the read data values

  wire          pop_fifo_i;         // Read an item from FIFO
  wire [2:0]    nxt_read_pointer;   // D input for read_pointer
  wire [2:0]    nxt_read_pointer_gray_async;   // D input for read_pointer_gray_async
  wire          read_pointer_we;    // enable for read_pointer
  reg  [2:0]    read_pointer_gray_async;// Gray coded read pointer
  wire [2:0]    read_pointer;       // Fifo read pointer
  reg  [7:0]    fifo_read_data;     // Fifo output register
  wire          fifo_empty;         // Fifo is empty

  wire          fifo_valid;
  wire          opt_port_we;

  assign opt_port_we      = (PORT_PRESENT != 0) ? 1'b1       : 1'b0;



  always @(fifo_data0_i or fifo_data1_i or fifo_data2_i
           or fifo_data3_i or read_pointer_gray_async)
    begin : c_fifo_read_data
      case (read_pointer_gray_async[2:0])
        3'b000,3'b110  :  fifo_read_data = fifo_data0_i[7:0];
        3'b001,3'b111  :  fifo_read_data = fifo_data1_i[7:0];
        3'b011,3'b101  :  fifo_read_data = fifo_data2_i[7:0];
        3'b010,3'b100  :  fifo_read_data = fifo_data3_i[7:0];
        default        :  fifo_read_data = {8{1'bx}};
      endcase // case
    end // block: c_fifo_read_data

  assign nxt_read_pointer  = read_pointer + 3'b001;
  assign read_pointer_we   = (pop_fifo_i & ~fifo_empty) & opt_port_we;

  assign fifo_empty  = (write_pointer_gray_i == read_pointer_gray_async);
  assign fifo_valid = ~fifo_empty;

  assign nxt_read_pointer_gray_async = {1'b0,nxt_read_pointer[2:1]} ^ nxt_read_pointer[2:0];

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      read_pointer_gray_async  <= 3'b000;
    else if (read_pointer_we)
      read_pointer_gray_async  <= nxt_read_pointer_gray_async;

  assign read_pointer[0] = read_pointer_gray_async[0] ^ read_pointer_gray_async[1] ^ read_pointer_gray_async[2];
  assign read_pointer[1] = read_pointer_gray_async[1] ^ read_pointer_gray_async[2];
  assign read_pointer[2] = read_pointer_gray_async[2];

  assign fifo_valid_o     = (PORT_PRESENT != 0) ? fifo_valid : 1'b0;
  assign read_pointer_gray_async_o = (PORT_PRESENT != 0) ? read_pointer_gray_async : 3'b000;
  assign fifo_read_data_o = (PORT_PRESENT != 0) ? fifo_read_data : {8{1'b0}};

`ifdef ARM_ASSERT_ON

  assert_never_unknown #(0,1,0,"TPIU fifo_valid must not be X")
  fifo_validX  (
   .clk         (clk),
   .reset_n     (reset_n),
   .qualifier   (1'b1),
   .test_expr   (fifo_valid)
   );

  assert_never_unknown #(0,8,0,"TPIU Fifo output must not be X")
  FifoRdataX (
   .clk         (clk),
   .reset_n     (reset_n),
   .qualifier   (1'b1),
   .test_expr   (fifo_read_data)
   );

`endif

endmodule // cm3_tpiu_trace_fifo
