`include "cm3_tpiu_defs.vp"

module cm3_tpiu_trace_out (
  clk, reset_n, prg_port_size_decoded_i, prg_uart_mode_i, prg_swo_mode_i, format_data_valid_i,
  format_data_i, prg_baud_div_i, baud_div_reset_i,
  trace_data_o, trace_swo_o, tpiu_baud_o, format_data_ready_o, trace_out_idle_async_o
  );

  parameter TRACE_LVL = 1;


  input           clk;    // Trace clock
  input           reset_n;       // Trace asychronous reset active low

  input          prg_uart_mode_i;          // Select UART mode
  input          prg_swo_mode_i;           // Select SWO mode
  input          format_data_valid_i;      // Fifo input is valid
  input [7:0]    format_data_i;            // Fifo data input
  input [12:0]   prg_baud_div_i;           // Prescaler for single wire o/p
  input          baud_div_reset_i;         // prg_baud_div has been changed
  input [2:0]    prg_port_size_decoded_i;  // Selected Trace port size. (One Hot)

  output [3:0]   trace_data_o;             // Traceport data out
  output         trace_swo_o;              // Single wire modes output
  output         tpiu_baud_o;              // Toggles at baud rate frequency in SVW modes
  output         format_data_ready_o;      // Formatter data handshaking
  output         trace_out_idle_async_o;   // State machine is in idle state


  reg   [3:0]    nxt_trace_data;     // D input for trace_data
  reg   [3:0]    trace_data;         // The output register bank
  wire  [2:0]    nxt_bit_count;      // D-input for bit_count
  wire  [2:0]    bit_count_add;      // Bit count calculation
  reg   [2:0]    bit_count;          // Bit count register
  reg  [12:0]    baud_count;         // Baud counter register
  wire [12:0]    nxt_baud_count;     // D-input for baud_count
  wire           nxt_bit_enable;     // Bit clock edge
  reg            bit_enable;         // Bit clock edge
  wire           uart_serial_data;   // Uart mode data
  wire           swo_serial_data;    // Manchester mode data
  reg            baud_div_reset_reg; // detect change in prescaler
  wire           count_reload;       // reload baud counter when reloaded
  wire           baud_count_we;      // Enable for baud counter
  reg            serial_state_we;    // Enable for state register
  wire           bit_count_we;       // Enable for Bit counter

  reg [1:0]      serial_state;       // Protocol state machine
  reg [1:0]      nxt_serial_state;   // D-input for serial_state
  reg            encode_state;       // Manchester mode encoding
  wire           nxt_encode_state;   // D-input for encode_state
  wire           encode_state_we;    // Enable for encode_state

  reg            trace_swo;          // Single wire modes output register
  wire           nxt_trace_swo;      // D-input for trace_swo
  wire           trace_swo_we;       // Enable for trace_swo
  wire           trace_data_we;      // Enable for trace data output reg
  reg [2:0]      swo_byte_run;       // Manchester mode byte run length
  wire           swo_byte_run_we;    // Enable for swo_byte_run
  wire[2:0]      nxt_swo_byte_run;   // D-input for swo_byte_run
  wire           tpiu_baud_we;       // Enable for TPIUBaud
  reg            tpiu_baud;          // SWV and Uart mode baud rate reference

  wire           format_data_ready;
  reg            trace_out_idle_async;
  wire           nxt_trace_out_idle_async;
  wire           opt_tpiu_en;

  assign trace_data_o           = (TRACE_LVL > 0) ? trace_data           : 4'b0000;
  assign trace_swo_o            = (TRACE_LVL > 0) ? trace_swo            : 1'b0;
  assign tpiu_baud_o            = (TRACE_LVL > 0) ? tpiu_baud            : 1'b0;
  assign format_data_ready_o    = (TRACE_LVL > 0) ? format_data_ready    : 1'b0;
  assign trace_out_idle_async_o = (TRACE_LVL > 0) ? trace_out_idle_async : 1'b0;
  assign opt_tpiu_en            = (TRACE_LVL > 0) ? 1'b1                 : 1'b0;



  always @(bit_enable or bit_count or format_data_valid_i or prg_swo_mode_i or swo_byte_run or
           serial_state or encode_state or prg_uart_mode_i or opt_tpiu_en)
  begin
    case (serial_state)
      `ST_CM_TPIU_IDLE:
        begin
          if (prg_uart_mode_i | prg_swo_mode_i)
            begin
              nxt_serial_state = `ST_CM_TPIU_STRT;
              serial_state_we = (format_data_valid_i & bit_enable) & opt_tpiu_en;
            end
          else
            begin
              serial_state_we = opt_tpiu_en;
              nxt_serial_state = `ST_CM_TPIU_DATA;
            end
        end // case: `ST_CM_TPIU_IDLE
      `ST_CM_TPIU_STRT:
        begin
          serial_state_we = (!encode_state & bit_enable) & opt_tpiu_en;
          nxt_serial_state = `ST_CM_TPIU_DATA;
        end
      `ST_CM_TPIU_DATA:
        begin
          case ({prg_uart_mode_i, prg_swo_mode_i})
            2'b00: // TracePort Mode. Exit only on change of mode
              begin
                serial_state_we = 1'b0;
                nxt_serial_state = `ST_CM_TPIU_DATA;
              end
            2'b10: // Uart Mode
              begin
                serial_state_we = ((bit_count == 3'b111) & bit_enable) & opt_tpiu_en;
                nxt_serial_state = (bit_count == 3'b111) ? `ST_CM_TPIU_STOP : `ST_CM_TPIU_DATA;
              end
            2'b01: // SWO mode
              begin
                serial_state_we =  ((!format_data_valid_i & (bit_count == 3'b000) & ~(bit_enable & !encode_state)) |
                              ((bit_count == 3'b111)&
                               (!encode_state & bit_enable) & (swo_byte_run == 3'b111))) & opt_tpiu_en;
                nxt_serial_state = (!format_data_valid_i & (bit_count == 3'b000)) |
                            ((bit_count == 3'b111)&
                                 (!encode_state & bit_enable) & (swo_byte_run == 3'b111))?
                           `ST_CM_TPIU_STOP:`ST_CM_TPIU_DATA;
              end
            2'b11: // Invalid mode setting
              begin
                serial_state_we = 1'b0;
                nxt_serial_state = `ST_CM_TPIU_DATA;
              end
            default:
              begin
                serial_state_we = 1'bx;
                nxt_serial_state = 2'bxx;
              end
          endcase // case{prg_uart_mode_i,prg_swo_mode_i}
        end // case: `ST_CM_TPIU_DATA

      `ST_CM_TPIU_STOP:
        begin
          serial_state_we  = (bit_enable & !encode_state)  & opt_tpiu_en;
          nxt_serial_state = format_data_valid_i?`ST_CM_TPIU_STRT:`ST_CM_TPIU_IDLE;
        end // case: `ST_CM_TPIU_STOP
      default:
        begin
          serial_state_we = 1'bx;
          nxt_serial_state = 2'bxx;
        end
    endcase // caseserial_state
  end // always @ (...

  assign bit_count_we = (serial_state == `ST_CM_TPIU_DATA) &
                        bit_enable & (!encode_state);

  assign nxt_swo_byte_run = {3{(serial_state == `ST_CM_TPIU_DATA) & prg_swo_mode_i}} &
                          (swo_byte_run + 3'b001);
  assign swo_byte_run_we = (((serial_state == `ST_CM_TPIU_DATA) & prg_swo_mode_i &
                           (bit_count == 3'b111)) |
                             (serial_state == `ST_CM_TPIU_STRT)) & bit_enable & !encode_state;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      serial_state <= `ST_CM_TPIU_IDLE;
    else if (serial_state_we)
      serial_state <= nxt_serial_state;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      trace_out_idle_async <= 1'b1;
    else if (serial_state_we & opt_tpiu_en)
      trace_out_idle_async <= nxt_trace_out_idle_async;

  assign nxt_trace_out_idle_async = opt_tpiu_en ? (nxt_serial_state == `ST_CM_TPIU_IDLE) : 1'b1;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      swo_byte_run <= 3'b000;
    else if (swo_byte_run_we & opt_tpiu_en)
      swo_byte_run <= nxt_swo_byte_run;

  assign bit_count_add = (bit_count + 3'b001);

  assign nxt_bit_count[0] = bit_count_add[0] |
                            prg_port_size_decoded_i[2] |
                            prg_port_size_decoded_i[1];
  assign nxt_bit_count[1] = bit_count_add[1] |
                             prg_port_size_decoded_i[2];
  assign nxt_bit_count[2] = bit_count_add[2];
  assign format_data_ready = (bit_count == 3'b111) & bit_count_we;

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      bit_count  <= 3'b000;
    else if (bit_count_we & opt_tpiu_en)
      bit_count <= nxt_bit_count;

  assign count_reload = (baud_div_reset_i != baud_div_reset_reg);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      baud_div_reset_reg  <= 1'b0;
    else if (count_reload & opt_tpiu_en)
      baud_div_reset_reg  <= baud_div_reset_i;

  assign baud_count_we = count_reload | prg_uart_mode_i | prg_swo_mode_i | ~bit_enable;

  assign nxt_baud_count = ((baud_count == prg_baud_div_i)|count_reload) ? {13{1'b0}} :
                       baud_count + {{12{1'b0}},1'b1};

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
        baud_count <= {13{1'b0}};
        bit_enable <= 1'b0;
      end
    else if (baud_count_we & opt_tpiu_en)
      begin
        baud_count <= nxt_baud_count;
        bit_enable <= nxt_bit_enable;
      end

  assign nxt_bit_enable = ~(|baud_count) | (~prg_uart_mode_i & ~prg_swo_mode_i);

  assign tpiu_baud_we = (prg_swo_mode_i | prg_uart_mode_i) & bit_enable;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      tpiu_baud <= 1'b0;
    else if (tpiu_baud_we & opt_tpiu_en)
      tpiu_baud <= ~tpiu_baud;

  always @(bit_count or prg_port_size_decoded_i or format_data_i)
    begin
      case (bit_count)
        3'b000:
          begin
            nxt_trace_data = {3'b000,format_data_i[0]};
          end
        3'b001:
          begin
            nxt_trace_data[3] = 1'b0;
            nxt_trace_data[2] = 1'b0;
            nxt_trace_data[1] = (prg_port_size_decoded_i[1] & format_data_i[1]);
            nxt_trace_data[0] = |(prg_port_size_decoded_i[1:0] & {format_data_i[0],format_data_i[1]});
          end
        3'b010:
          begin
            nxt_trace_data = {3'b000,format_data_i[2]};
          end
        3'b011:
          begin
            nxt_trace_data[0] = |(prg_port_size_decoded_i[2:0] & {format_data_i[0],format_data_i[2],format_data_i[3]});
            nxt_trace_data[1] = |(prg_port_size_decoded_i[2:1] & {format_data_i[1],format_data_i[3]});
            nxt_trace_data[3:2] = {2{prg_port_size_decoded_i[2]}} & format_data_i[3:2];
          end
        3'b100:
          begin
            nxt_trace_data[0] = format_data_i[4];
            nxt_trace_data[3:1] = 3'b000;
          end
        3'b101:
          begin
            nxt_trace_data[0] = |(prg_port_size_decoded_i[1:0] & {format_data_i[4],format_data_i[5]});
            nxt_trace_data[1] = (prg_port_size_decoded_i[1] & format_data_i[5]);
            nxt_trace_data[3:2] = 2'b00;
          end
        3'b110:
          begin
            nxt_trace_data[0] = format_data_i[6];
            nxt_trace_data[3:1] = 3'b000;
          end
        3'b111:
          begin
            nxt_trace_data[0] = |(prg_port_size_decoded_i[2:0] & {format_data_i[4],format_data_i[6],format_data_i[7]});
            nxt_trace_data[1] = |(prg_port_size_decoded_i[2:1] & {format_data_i[5],format_data_i[7]});
            nxt_trace_data[3:2] = {2{prg_port_size_decoded_i[2]}} & format_data_i[7:6];
          end
        default:
          begin
            nxt_trace_data = 4'bXXXX;
          end
      endcase // case(bit_count)
    end // always @ (...

  assign trace_data_we = (!(prg_uart_mode_i|prg_swo_mode_i) |(|trace_data))  & opt_tpiu_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      trace_data <= {4{1'b0}};
    else if (trace_data_we)
      trace_data <= nxt_trace_data;

  assign nxt_encode_state  = (~encode_state | (serial_state == `ST_CM_TPIU_IDLE)) & prg_swo_mode_i;
  assign encode_state_we = ((prg_swo_mode_i & bit_enable & !(serial_state == `ST_CM_TPIU_IDLE)) |
                             ((encode_state ^ nxt_encode_state) & ((serial_state == `ST_CM_TPIU_IDLE) |
                                                                   ~prg_swo_mode_i))) & opt_tpiu_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      encode_state <= 1'b1;
    else if (encode_state_we)
      encode_state <= nxt_encode_state;

  assign uart_serial_data = ((serial_state==`ST_CM_TPIU_DATA) & nxt_trace_data[0])|
                     (serial_state==`ST_CM_TPIU_STOP)|
                       (serial_state==`ST_CM_TPIU_IDLE);

  assign swo_serial_data  = prg_swo_mode_i & (format_data_valid_i & (!(encode_state ^ nxt_trace_data[0]) &
                                                                     (serial_state == `ST_CM_TPIU_DATA)) |
                                              (encode_state & (serial_state == `ST_CM_TPIU_STRT)));


  assign nxt_trace_swo = prg_uart_mode_i ? uart_serial_data : swo_serial_data;
  assign trace_swo_we = (prg_swo_mode_i | prg_uart_mode_i | trace_swo) & opt_tpiu_en;

  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      trace_swo <= 1'b0;
    else if (trace_swo_we)
      trace_swo <= nxt_trace_swo;


`ifdef ARM_ASSERT_ON

  reg [8*20:0] debug_trace_state;

  always @ (serial_state or prg_uart_mode_i or bit_count)
    case (serial_state)
      `ST_CM_TPIU_IDLE : debug_trace_state = {"IDLE"};
      `ST_CM_TPIU_STRT : debug_trace_state = {"START"};
      `ST_CM_TPIU_DATA : begin
        if (prg_uart_mode_i)
          case (bit_count)
            3'b000  : debug_trace_state = {"DATA0"};
            3'b001  : debug_trace_state = {"DATA1"};
            3'b010  : debug_trace_state = {"DATA2"};
            3'b011  : debug_trace_state = {"DATA3"};
            3'b100  : debug_trace_state = {"DATA4"};
            3'b101  : debug_trace_state = {"DATA5"};
            3'b110  : debug_trace_state = {"DATA6"};
            3'b111  : debug_trace_state = {"DATA7"};
            default : debug_trace_state = {"UNKNOWN"};
          endcase
        else
          debug_trace_state = {"DATA"};
      end
      `ST_CM_TPIU_STOP : debug_trace_state = {"STOP"};
      default       : debug_trace_state = {"UNKNOWN"};
    endcase

  reg  [2:0] prg_port_size_reg1;
  reg  [2:0] prg_port_size_reg2;
  wire [2:0] prg_port_size_any;
  wire       ovl_reg_enable = (bit_count == 3'b000) | ~(prg_swo_mode_i | prg_uart_mode_i);

  always @ (posedge clk or negedge reset_n)
    if (!reset_n)
      begin
      prg_port_size_reg1 <= 4'b0001;
      prg_port_size_reg2 <= 4'b0001;
      end
    else if (ovl_reg_enable)
      begin
      prg_port_size_reg1 <= prg_port_size_decoded_i ;
      prg_port_size_reg2 <= prg_port_size_reg1;
      end

  assign prg_port_size_any = prg_port_size_decoded_i | prg_port_size_reg1 | prg_port_size_reg2;


  assert_never #(0,0,"FifoData must be valid in TracePort mode")
  format_data_valid_iTPmode (
                   .clk       (clk),
                   .reset_n   (reset_n),
                   .test_expr (!prg_uart_mode_i&!prg_swo_mode_i&
                               (serial_state ==`ST_CM_TPIU_DATA)&!format_data_valid_i)
   );

  assert_never #(0,0,"trace_data[3] active but prg_port_size is not 4")
  TraceData3_check (
                   .clk       (clk),
                   .reset_n   (reset_n),
                   .test_expr (trace_data[3] & (~prg_port_size_any[2]))
   );
  assert_never #(0,0,"trace_data[2] active but prg_port_size is not 4")
  TraceData2_check (
                   .clk       (clk),
                   .reset_n   (reset_n),
                   .test_expr (trace_data[2] & (~(prg_port_size_any[2])))
   );
  assert_never #(0,0,"trace_data[1] active but prg_port_size is not 2 or 4")
  TraceData1_check (
                   .clk       (clk),
                   .reset_n   (reset_n),
                   .test_expr (trace_data[1] & (~(|prg_port_size_any[2:1])))
   );

  reg    prg_swo_mode_reg;
  always @(posedge clk or negedge reset_n)
    if (!reset_n)
      prg_swo_mode_reg <= 1'b1;
    else
      prg_swo_mode_reg <= prg_swo_mode_i;

  assert_implication #(0,0,"Encode state should be low unless in manchester mode")
  u_ovl_encode_state_low
    (
     .clk       (clk),
     .reset_n   (reset_n),
     .antecedent_expr (encode_state & opt_tpiu_en),
     .consequent_expr (prg_swo_mode_reg)
   );


`endif

endmodule // cm3_tpiu_trace_out
