module cm3_wic
  (// Inputs
   FCLK, RESETn, WICLOAD, WICCLEAR, WICINT, WICMASK, WICENREQ,
   WICDSACKn,
   WAKEUP, WICSENSE, WICPEND, WICDSREQn, WICENACK
  );

  parameter WIC_PRESENT = 0;  // WIC present if 1.
  parameter WIC_LINES   = 8;  // Number of WIC lines. Min value is 3.


  input                  FCLK;
  input                  RESETn;

  input                  WICLOAD;     // WIC mask load from core
  input                  WICCLEAR;    // WIC mask clear from core
  input  [WIC_LINES-1:0] WICINT;      // Interrupt request from system
  input  [WIC_LINES-1:0] WICMASK;     // Mask from core
  input                  WICENREQ;    // WIC enable request from PMU
  input                  WICDSACKn;   // WIC enable ack from core
  output [WIC_LINES-1:0] WICSENSE;    // Input lines that can generate WAKEUP
  output [WIC_LINES-1:0] WICPEND;     // Pended interrupt request
  output                 WICDSREQn;   // WIC enable request to core
  output                 WICENACK;    // WIC enable ack to PMU
  output                 WAKEUP;      // Wake up request to PMU


  reg    [WIC_LINES-1:0] wic_sense;        // Interrupt mask register
  wire   [WIC_LINES-1:0] nxt_wic_sense;
  wire                   wic_sense_we;
  reg    [WIC_LINES-1:0] wic_pend;         // Interrupt pend register
  wire   [WIC_LINES-1:0] nxt_wic_pend;
  wire                   wic_pend_we;
  reg                    wic_active;       // Active status register
  wire                   wic_active_we;

  reg                    wic_ds_req;
  wire                   set_wic_ds_req;
  wire                   clr_wic_ds_req;
  wire                   wic_ds_req_we;
  reg                    wic_en_ack;
  wire                   set_wic_en_ack;
  wire                   clr_wic_en_ack;
  wire                   wic_en_ack_we;

  wire                   wakeup_pmu;       // Request wakeup to PMU

  wire                   opt_mst_wic_en;
  wire                   opt_wakeup;
  wire   [WIC_LINES-1:0] opt_wic_sense;
  wire   [WIC_LINES-1:0] opt_wic_pend;
  wire                   opt_wic_ds_req_n;
  wire                   opt_wic_en_ack;

  assign opt_wakeup       = (WIC_PRESENT != 0) ? wakeup_pmu  : 1'b0;
  assign opt_wic_sense    = (WIC_PRESENT != 0) ? wic_sense   : {WIC_LINES{1'b0}};
  assign opt_wic_pend     = (WIC_PRESENT != 0) ? wic_pend    : {WIC_LINES{1'b0}};
  assign opt_wic_ds_req_n = (WIC_PRESENT != 0) ? ~wic_ds_req : 1'b1;
  assign opt_wic_en_ack   = (WIC_PRESENT != 0) ? wic_en_ack  : 1'b0;
  assign opt_mst_wic_en   = (WIC_PRESENT != 0) ? 1'b1        : 1'b0;


  assign set_wic_ds_req = WICENREQ & opt_wic_ds_req_n;
  assign clr_wic_ds_req = ~opt_wic_ds_req_n & ~WICENREQ;
  assign wic_ds_req_we  = opt_mst_wic_en & (set_wic_ds_req | clr_wic_ds_req);

  always @ (posedge FCLK or negedge RESETn)
    if (!RESETn)
      wic_ds_req <= 1'b0;
    else if (wic_ds_req_we)
      wic_ds_req <= set_wic_ds_req;

  assign set_wic_en_ack = ~WICDSACKn & ~opt_wic_en_ack;
  assign clr_wic_en_ack = opt_wic_en_ack & WICDSACKn;
  assign wic_en_ack_we  = opt_mst_wic_en & (set_wic_en_ack | clr_wic_en_ack);

  always @ (posedge FCLK or negedge RESETn)
    if (!RESETn)
      wic_en_ack <= 1'b0;
    else if (wic_en_ack_we)
      wic_en_ack <= set_wic_en_ack;


  assign nxt_wic_sense = {WIC_LINES{WICLOAD}} & WICMASK;
  assign wic_sense_we  = opt_mst_wic_en & (WICCLEAR | WICLOAD);

  always @ (posedge FCLK or negedge RESETn)
    if (!RESETn)
      wic_sense <= {WIC_LINES{1'b0}};
    else if (wic_sense_we)
      wic_sense <= nxt_wic_sense;


  assign nxt_wic_pend = {WIC_LINES{~WICCLEAR}} & (WICINT | wic_pend);
  assign wic_pend_we  = opt_mst_wic_en & (WICLOAD | wic_active) & (WICCLEAR | (|WICINT));

  always @( posedge FCLK or negedge RESETn)
    if (!RESETn)
      wic_pend <= {WIC_LINES{1'b0}};
    else if (wic_pend_we)
      wic_pend <= nxt_wic_pend;


  assign wic_active_we = opt_mst_wic_en & (WICLOAD | WICCLEAR);

  always @ (posedge FCLK or negedge RESETn)
    if (!RESETn)
      wic_active <= 1'b0;
    else if (wic_active_we)
      wic_active <= WICLOAD;


  assign wakeup_pmu = (|(wic_pend & wic_sense));

  assign WAKEUP    = opt_wakeup;
  assign WICSENSE  = opt_wic_sense;
  assign WICPEND   = opt_wic_pend;
  assign WICDSREQn = opt_wic_ds_req_n;
  assign WICENACK  = opt_wic_en_ack;



`ifdef ARM_ASSERT_ON
  assert_never_unknown #(0,1,0, "wic_ds_req_we must not be unknown")
  ts_wic_ds_req_we_unknown
    (.clk       (FCLK),
     .reset_n   (RESETn),
     .qualifier (1'b1),
     .test_expr (wic_ds_req_we));

  assert_never_unknown #(0,1,0, "wic_sense_we must not be unknown")
  ts_wic_sense_we_unknown
    (.clk       (FCLK),
     .reset_n   (RESETn),
     .qualifier (1'b1),
     .test_expr (wic_sense_we));

  assert_never_unknown #(0,1,0, "wic_pend_we must not be unknown")
  ts_wic_pend_we_unknown
    (.clk       (FCLK),
     .reset_n   (RESETn),
     .qualifier (1'b1),
     .test_expr (wic_pend_we));

  assert_never_unknown #(0,1,0, "wic_active_we must not be unknown")
  ts_wic_active_we_unknown
    (.clk       (FCLK),
     .reset_n   (RESETn),
     .qualifier (1'b1),
     .test_expr (wic_active_we));

`endif
endmodule