module input_stage_t (
    hclk,
    hreset_n,
    hsels_i,
    haddrs_i,
    htranss_i,
    hwrites_i,
    hsizes_i,
    hbursts_i,
    hprots_i,
    hmasters_i,
    hmastlocks_i,
    hreadys_i,

    active_i,
    readyout_i,
    resp_i,

    hreadyouts_o,
    hresps_o,

    sel_o,
    addr_o,
    trans_o,
    write_o,
    size_o,
    burst_o,
    prot_o,
    master_o,
    mastlock_o,
    held_tran_o
    );

    input         hclk;              // AHB System Clock
    input         hreset_n;          // AHB System Reset
    input         hsels_i;           // Slave Select from AHB
    input  [31:0] haddrs_i;          // Address bus from AHB
    input   [1:0] htranss_i;         // Transfer type from AHB
    input         hwrites_i;         // Transfer direction from AHB
    input   [2:0] hsizes_i;          // Transfer size from AHB
    input   [2:0] hbursts_i;         // Burst type from AHB
    input   [3:0] hprots_i;          // Protection control from AHB
    input   [3:0] hmasters_i;        // Master number from AHB
    input         hmastlocks_i;      // Locked Sequence  from AHB
    input         hreadys_i;         // Transfer done from AHB
    input         active_i;          // Active signal
    input         readyout_i;        // HREADYOUT input
    input         resp_i;            // HRESP input

    output        hreadyouts_o;      // HREADY feedback to AHB
    output        hresps_o;          // Transfer response to AHB
    output        sel_o;             // HSEL output
    output [31:0] addr_o;            // HADDR output
    output  [1:0] trans_o;           // HTRANS output
    output        write_o;           // HWRITE output
    output  [2:0] size_o;            // HSIZE output
    output  [2:0] burst_o;           // HBURST output
    output  [3:0] prot_o;            // HPROT output
    output [3:0]  master_o;          // HMASTER output
    output        mastlock_o;        // HMASTLOCK output
    output        held_tran_o;       // Holding register active flag


    reg         hreadyouts;      // HREADY feedback to AHB
    reg         hresps;          // Transfer response to AHB
    reg         sel;             // HSEL output
    reg  [31:0] addr;            // HADDR output
    wire  [1:0] trans;           // HTRANS output
    reg         write;           // HWRITE output
    reg   [2:0] size;            // HSIZE output
    wire  [2:0] burst;           // HBURST output
    reg   [3:0] prot;            // HPROT output
    reg   [3:0] master;          // HMASTER output
    reg         mastlock;        // HMASTLOCK output
    wire        held_tran;       // Holding register active flag


`define FM_TRN_IDLE    2'b00     // Idle Transfer
`define FM_TRN_BUSY    2'b01     // Busy Transfer
`define FM_TRN_NONSEQ  2'b10     // Nonsequential transfer
`define FM_TRN_SEQ     2'b11     // Sequential transfer

`define FM_SZ_BYTE     3'b000    // 8-bit
`define FM_SZ_HALF     3'b001    // 16-bit
`define FM_SZ_WORD     3'b010    // 32-bit

`define FM_BUR_SINGLE  3'b000    // Single BURST
`define FM_BUR_INCR    3'b001    // Incremental BURSTS
`define FM_BUR_WRAP4   3'b010    // 4-beat wrap
`define FM_BUR_INCR4   3'b011    // 4-beat incr
`define FM_BUR_WRAP8   3'b100    // 8-beat wrap
`define FM_BUR_INCR8   3'b101    // 8-beat incr
`define FM_BUR_WRAP16  3'b110    // 16-beat wrap
`define FM_BUR_INCR16  3'b111    // 16-beat incr

    wire        load_reg;             // Holding register load flag
    wire        pend_tran;            // An active transfer cannot complete
    reg         pend_tran_reg;        // Registered version of PendTran
    wire        addr_valid;           // Indicates address phase of
    reg         data_valid;           // Indicates data phase of
    reg   [1:0] reg_trans;            // Registered HTRANSS
    reg  [31:0] reg_addr;             // Registered HADDRS
    reg         reg_write;            // Registered HWRITES
    reg   [2:0] reg_size;             // Registered HSIZES
    reg   [2:0] reg_burst;            // Registered HBURSTS
    reg   [3:0] reg_prot;             // Registered HPROTS
    reg   [3:0] reg_master;           // Registerd HMASTERS
    reg         reg_mastlock;         // Registered HMASTLOCKS
    reg   [1:0] trans_b;              // HTRANS output used for burst information
    reg   [1:0] trans_int;            // HTRANS output
    reg   [2:0] burst_int;            // HBURST output
    reg   [3:0] offset_addr;          // Address offset for boundary logic
    reg   [3:0] check_addr;           // Address check for wrapped bursts
    reg         burst_override;       // Registered BurstOverrideNext
    wire        next_burst_override;  // Indicates burst has been over-ridden
    reg         bound;                // Registered version of BoundNext
    wire        next_bound;           // Indicates boundary wrapping
    wire        bound_en;             // Clock-enable for Bound register


  always @ (posedge hclk or negedge hreset_n)
  begin : p_holding_reg_seq1
    if  ( hreset_n ==1'b0 )
    begin
      reg_trans    <= 2'b00;
      reg_addr     <= {32{1'b0}};
      reg_write    <= 1'b0 ;
      reg_size     <= 3'b000;
      reg_burst    <= 3'b000;
      reg_prot     <= {4{1'b0}};
      reg_master   <= 4'b0000;
      reg_mastlock <= 1'b0 ;
    end
    else
    begin
      if  (load_reg)
      begin
        reg_trans    <= htranss_i;
        reg_addr     <= haddrs_i;
        reg_write    <= hwrites_i;
        reg_size     <= hsizes_i;
        reg_burst    <= hbursts_i;
        reg_prot     <= hprots_i;
        reg_master   <= hmasters_i;
        reg_mastlock <= hmastlocks_i;
      end
    end
  end

  assign addr_valid = (hsels_i & htranss_i[1]);

  assign load_reg = (addr_valid & hreadys_i);

  always @ (posedge hclk or negedge hreset_n)
  begin : p_data_valid
    if  (hreset_n == 1'b0)
      data_valid <= 1'b0;
    else
    begin
     if (hreadys_i)
      data_valid  <= addr_valid;
    end
  end


  assign pend_tran = (load_reg & (!active_i)) ? 1'b1 :
                    ((active_i & readyout_i) ? 1'b0 : pend_tran_reg);

  always @ (posedge hclk or negedge hreset_n)
  begin : p_pend_tran_reg
    if (hreset_n == 1'b0)
     pend_tran_reg  <= 1'b0;
    else
     pend_tran_reg  <= pend_tran;
  end

  assign  held_tran  = (load_reg | pend_tran_reg);


  always @ (pend_tran_reg or hsels_i or htranss_i or haddrs_i or hwrites_i
            or hsizes_i or hbursts_i or hprots_i or hmastlocks_i or reg_addr
            or reg_write or reg_size or reg_burst or reg_prot or reg_mastlock or
            hmasters_i or reg_master)
  begin : p_mux_comb
    if  ( pend_tran_reg ==1'b0 )
    begin
      sel       = hsels_i;
      trans_int = htranss_i;
      addr      = haddrs_i;
      write     = hwrites_i;
      size      = hsizes_i;
      burst_int = hbursts_i;
      prot      = hprots_i;
      master    = hmasters_i;
      mastlock  = hmastlocks_i;
    end
    else
    begin
      sel       = 1'b1;
      trans_int = `FM_TRN_NONSEQ;
      addr      = reg_addr;
      write     = reg_write;
      size      = reg_size;
      burst_int = reg_burst;
      prot      = reg_prot;
      master    = reg_master;
      mastlock  = reg_mastlock;
    end
  end


  always @ (pend_tran_reg or htranss_i or reg_trans)
  begin : p_trans_b_comb
    if  ( pend_tran_reg ==1'b0 )
      trans_b = htranss_i;
    else
      trans_b = reg_trans;
  end // block: p_TransbComb


  assign trans  = (burst_override & bound) ? {trans_int[1], 1'b0}
                : trans_int;

  assign burst  = (burst_override & (trans_b != `FM_TRN_NONSEQ)) ? `FM_BUR_INCR
                : burst_int;


  always @ (data_valid or pend_tran_reg or readyout_i or resp_i)
  begin : p_ready_comb
    if  (data_valid == 1'b0)
      begin
        hreadyouts = 1'b1;
        hresps     = 1'b0;
      end
    else if (pend_tran_reg)
      begin
        hreadyouts = 1'b0;
        hresps     = 1'b0;
      end
    else
      begin
        hreadyouts = readyout_i;
        hresps     = resp_i;
      end
  end // block: p_ReadyComb


  assign next_burst_override  = ((htranss_i == `FM_TRN_NONSEQ) |
                               (htranss_i == `FM_TRN_IDLE)) ? 1'b0
                              : (((load_reg == 1'b1) &&
                                  (active_i == 1'b0) &&
                                  (htranss_i ==`FM_TRN_SEQ)) ? 1'b1
                                 : burst_override) ;

  always @ (posedge hclk or negedge hreset_n)
  begin : p_burst_override_seq
    if  ( hreset_n ==1'b0 )
      burst_override  <= 1'b0 ;
    else
      begin
        if (hreadys_i)
          burst_override  <= next_burst_override ;
      end
  end // block: p_BurstOverrideSeq

  always @ (haddrs_i or hsizes_i)
  begin : p_offset_addr_comb
    case (hsizes_i )
      3'b000:  offset_addr  = haddrs_i[3:0];
      3'b001:  offset_addr  = haddrs_i[4:1];
      3'b010:  offset_addr  = haddrs_i[5:2];
      3'b011:  offset_addr  = haddrs_i[6:3];
      default: offset_addr  = haddrs_i[3:0];
    endcase
  end

  always @ (offset_addr or hbursts_i)
  begin : p_check_addr_comb
    case (hbursts_i )
      `FM_BUR_WRAP4 : begin
        check_addr[1:0] = offset_addr[1:0];
        check_addr[3:2] = 2'b11;
      end

      `FM_BUR_WRAP8 : begin
        check_addr[2:0] = offset_addr[2:0];
        check_addr[3]   = 1'b1 ;
      end

      `FM_BUR_WRAP16 : begin
        check_addr[3:0] = offset_addr[3:0];
      end

      default: begin
        check_addr[3:0] = 4'b0000;
      end
    endcase
  end

  assign next_bound  = (check_addr == 4'b1111) ? 1'b1 : 1'b0;

  assign bound_en  = ((htranss_i [1]) & hreadys_i) ? 1'b1 : 1'b0;

  always @ (posedge hclk or negedge hreset_n)
  begin : p_bound_seq
    if  ( hreset_n ==1'b0 )
      bound  <= 1'b0 ;
    else
    begin
      if  (bound_en)
        bound  <= next_bound ;
    end
  end

  assign hreadyouts_o  = hreadyouts;
  assign hresps_o      = hresps;
  assign sel_o         = sel;
  assign addr_o        = addr;
  assign trans_o       = trans;
  assign write_o       = write;
  assign size_o        = size;
  assign burst_o       = burst;
  assign prot_o        = prot;
  assign master_o      = master;
  assign mastlock_o    = mastlock;
  assign held_tran_o   = held_tran;

endmodule
