module output_stage_t (
    hclk,
    hreset_n,

    sel0_i,
    addr0_i,
    trans0_i,
    write0_i,
    size0_i,
    burst0_i,
    prot0_i,
    master0_i,
    mastlock0_i,
    wdata0_i,
    held_tran0_i,

    sel1_i,
    addr1_i,
    trans1_i,
    write1_i,
    size1_i,
    burst1_i,
    prot1_i,
    master1_i,
    mastlock1_i,
    wdata1_i,
    held_tran1_i,

    hreadyoutm_i,

    active0_o,
    active1_o,

    hselm_o,
    haddrm_o,
    htransm_o,
    hwritem_o,
    hsizem_o,
    hburstm_o,
    hprotm_o,
    hmasterm_o,
    hmastlockm_o,
    hreadymuxm_o,
    hwdatam_o
    );


    input         hclk;       // AHB system clock
    input         hreset_n;    // AHB system reset

    input         sel0_i;       // Port 0 HSEL signal
    input [31:0]  addr0_i;      // Port 0 HADDR signal
    input  [1:0]  trans0_i;     // Port 0 HTRANS signal
    input         write0_i;     // Port 0 HWRITE signal
    input  [2:0]  size0_i;      // Port 0 HSIZE signal
    input  [2:0]  burst0_i;     // Port 0 HBURST signal
    input  [3:0]  prot0_i;      // Port 0 HPROT signal
    input  [3:0]  master0_i;    // Port 0 HMASTER signal
    input         mastlock0_i;  // Port 0 HMASTLOCK signal
    input [31:0]  wdata0_i;     // Port 0 HWDATA signal
    input         held_tran0_i; // Port 0 HeldTran signal

    input         sel1_i;       // Port 1 HSEL signal
    input [31:0]  addr1_i;      // Port 1 HADDR signal
    input  [1:0]  trans1_i;     // Port 1 HTRANS signal
    input         write1_i;     // Port 1 HWRITE signal
    input  [2:0]  size1_i;      // Port 1 HSIZE signal
    input  [2:0]  burst1_i;     // Port 1 HBURST signal
    input  [3:0]  prot1_i;      //  Port 1 HPROT signal
    input  [3:0]  master1_i;    // Port 1 HMASTER signal
    input         mastlock1_i;  // Port 1 HMASTLOCK signal
    input [31:0]  wdata1_i;     // Port 1 HWDATA signal
    input         held_tran1_i; // Port 1 HeldTran signal

    input         hreadyoutm_i; // HREADY feedback

    output        active0_o;    // Port 0 Active signal
    output        active1_o;    // Port 1 Active signal

    output        hselm_o;      // Slave select line
    output [31:0] haddrm_o;     // Address
    output  [1:0] htransm_o;    // Transfer type
    output        hwritem_o;    // Transfer direction
    output  [2:0] hsizem_o;     // Transfer size
    output  [2:0] hburstm_o;    // Burst type
    output  [3:0] hprotm_o;     // Protection control
    output  [3:0] hmasterm_o;   // Master ID
    output        hmastlockm_o; // Locked transfer
    output        hreadymuxm_o; // Transfer done
    output [31:0] hwdatam_o;    // Write data


    reg         active0;    // Port 0 Active signal

    reg         active1;    // Port 1 Active signal

    reg  [31:0] haddrm;     // Address
    reg         hwritem;    // Transfer direction
    reg   [2:0] hsizem;     // Transfer size
    reg   [3:0] hprotm;     // Protection control
    reg   [3:0] hmasterm;   // Master ID
    reg  [31:0] hwdatam;    // Write data


    wire        req_port0;      // Port 0 request signal
    wire        req_port1;      // Port 1 request signal

    wire  [3:0] addr_in_port;   // Address input port
    reg   [3:0] data_in_port;   // Data input port
    wire        no_port;        // No port selected signal
    reg         slave_sel;      // Slave select signal

    reg         hsel_lock;      //  Held HSELS during locked sequence
    wire        next_hsel_lock; //  Pre-registered HselLock
    wire        hlock_arb;      //  HMASTLOCK modified by HSEL for arbitration

    reg         hselm;      // internal HSELM
    reg   [1:0] htransm;    // internal HTRANSM
    reg   [2:0] hburstm;    // internal HBURSTM
    reg         hreadymuxm; // internal HREADYMUXM
    reg         hmastlockm; // internal HMASTLOCKM


  assign req_port0  = (held_tran0_i & sel0_i) ? 1'b1 : 1'b0 ;
  assign req_port1  = (held_tran1_i & sel1_i) ? 1'b1 : 1'b0 ;

  output_arb_t  u_output_arb
        (
        .hclk           (hclk),
        .hreset_n       (hreset_n),
        .req_port0_i    (req_port0),
        .req_port1_i    (req_port1),
        .hreadym_i      (hreadymuxm),
        .hselm_i        (hselm),
        .htransm_i      (htransm),
        .hburstm_i      (hburstm),
        .hmastlockm_i   (hlock_arb),

        .addr_in_port_o (addr_in_port),
        .no_port_o      (no_port)
        );

  always @ (addr_in_port or no_port)
  begin : p_active_comb
    active0  = 1'b0 ;
    active1  = 1'b0 ;
    if  ( no_port ==1'b0 )
    begin
      case (addr_in_port )
        4'b0000 : active0  = 1'b1 ;
        4'b0001 : active1  = 1'b1 ;
        default : begin
        end
      endcase // case(AddrInPort )
    end
  end // block: p_ActiveComb


  always @(sel0_i or addr0_i or trans0_i or write0_i or size0_i or burst0_i or prot0_i or
           master0_i or mastlock0_i or
         sel1_i or addr1_i or trans1_i or write1_i or size1_i or burst1_i or prot1_i or
           master1_i or mastlock1_i or
            addr_in_port or no_port)
  begin : p_addr_mux
    if  ( no_port )
      begin
        hselm      = 1'b0 ;
        haddrm     = {32{1'b0}};
        htransm    = {2{1'b0}};
        hwritem    = 1'b0 ;
        hsizem     = {3{1'b0}};
        hburstm    = {3{1'b0}};
        hprotm     = {4{1'b0}};
        hmasterm   = {4{1'b0}};
        hmastlockm = 1'b0 ;
      end
    else
      begin
        case (addr_in_port)

          4'b0000 :
            begin
              hselm      = sel0_i ;
              haddrm     = addr0_i ;
              htransm    = trans0_i ;
              hwritem    = write0_i ;
              hsizem     = size0_i ;
              hburstm    = burst0_i ;
              hprotm     = prot0_i ;
              hmasterm   = master0_i;
              hmastlockm = mastlock0_i ;
            end // case: 4'b0000

          4'b0001 :
            begin
              hselm      = sel1_i;
              haddrm     = addr1_i;
              htransm    = trans1_i;
              hwritem    = write1_i;
              hsizem     = size1_i;
              hburstm    = burst1_i;
              hprotm     = prot1_i;
              hmasterm   = master1_i;
              hmastlockm = mastlock1_i;
            end // case: 4'b0001

          default :
            begin
              hselm      = sel0_i ;
              haddrm     = addr0_i ;
              htransm    = trans0_i ;
              hwritem    = write0_i ;
              hsizem     = size0_i ;
              hburstm    = burst0_i ;
              hprotm     = prot0_i ;
              hmasterm   = master0_i;
              hmastlockm = mastlock0_i ;
            end // case: default
        endcase // case(AddrInPort)
      end // else: !if( NoPort )
  end // block: p_AddrMux

  assign next_hsel_lock = ((hselm & htransm[1] & hmastlockm) ? 1'b1
                         : ((hmastlockm == 1'b0) ? 1'b0
                            : hsel_lock));

  always @ (posedge hclk or negedge hreset_n)
    begin : p_hsel_lock
      if (hreset_n == 1'b0)
        begin
          hsel_lock <= 1'b0;
        end
      else
        begin
          if (hreadymuxm)
            begin
              hsel_lock <= next_hsel_lock;
            end
        end
    end

  assign hlock_arb = hmastlockm & (hsel_lock | hselm);

  always @ (posedge hclk or negedge hreset_n)
  begin : p_data_in_port_reg
    if  (hreset_n == 1'b0)
     data_in_port  <= 4'b0000;
    else
      begin
        if  (hreadymuxm)
          data_in_port <= addr_in_port ;
      end // else: !if(HRESETn == 1'b0)
  end // block: p_DataInPortReg

  always @ (wdata0_i or
            wdata1_i or
            data_in_port)
  begin : p_data_mux
    case (data_in_port)
      4'b0000 : hwdatam  = wdata0_i ;
      4'b0001 : hwdatam  = wdata1_i;
      default : hwdatam  = wdata0_i ;
    endcase // case(DataInPort )
  end // block: p_DataMux


  always @ (posedge hclk or negedge hreset_n)
  begin : p_slave_sel_reg
    if  ( hreset_n ==1'b0 )
      slave_sel  <= 1'b0 ;
    else
    begin
      if  ( hreadymuxm )
        slave_sel  <= hselm ;
    end
  end

  always @ (slave_sel or hreadyoutm_i)
  begin : p_hreadys_comb
    if  ( slave_sel )
      hreadymuxm  = hreadyoutm_i ;
    else
      hreadymuxm  = 1'b1 ;
  end

  assign active0_o     = active0;
  assign active1_o     = active1;

  assign hselm_o       = hselm;
  assign haddrm_o      = haddrm;
  assign htransm_o     = htransm;
  assign hwritem_o     = hwritem;
  assign hsizem_o      = hsizem;
  assign hburstm_o     = hburstm;
  assign hprotm_o      = hprotm;
  assign hmasterm_o    = hmasterm;
  assign hmastlockm_o  = hmastlockm;
  assign hreadymuxm_o  = hreadymuxm;
  assign hwdatam_o     = hwdatam;

endmodule
